/*
 *  ./acpica/busmgr/bm.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __BM_C__

#include <acpi.h>

#include "busmgr.h"
#include "bm.h"
#include "utility.h"

#define _COMPONENT          BUS_MANAGER
        MODULE_NAME         ("bm");


// TODO: Remove
#include <KernelExport.h>



// TODO: Use a dynamically sizeable array versus fixed size one.
extern BM_INTERNAL_DEVICE_LIST DeviceList;




ACPI_STATUS
BmPrintDevice (
    BM_INTERNAL_DEVICE      *Device)
{
    ACPI_BUFFER             Buffer;

    if (!Device)
    {
        return(AE_BAD_PARAMETER);
    }

    Buffer.Length = 256;
    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return(AE_NO_MEMORY);
    }

    AcpiGetName(Device->Info.AcpiHandle, ACPI_FULL_PATHNAME, &Buffer);

    OsdPrintf("[%s]: Handle[0x%08X] AcpiHandle[0x%08X] Name[%s] \n", Device->Info.Name, Device->Info.DeviceHandle, Device->Info.AcpiHandle, Buffer.Pointer);
    /*
    OsdPrintf("  Info:    Class[%p] HardwareId[%s] Flags[%p] Status[%p]\n", Device->Info.Class, Device->Info.HardwareId, Device->Info.Flags, Device->Info.Status);
    OsdPrintf("  Control: Request[%p] Notify[%p] Context[%p]\n", Device->Control.Request, Device->Control.Notify, Device->Control.Context);
    OsdPrintf("  Linkage: Parent[%p] Next[%p] Scope.Head[%p] Scope.Tail[%p]\n", Device->Parent, Device->Next, Device->Scope.Head, Device->Scope.Tail);
    */
    OsdFree(Buffer.Pointer);

    return(AE_OK);
}



ACPI_STATUS
BmCalculateDeviceStatus (
    BM_INTERNAL_DEVICE      *Device)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             AcpiHandle = NULL;

    if (!Device)
    {
        return AE_BAD_PARAMETER;
    }

    Device->Info.Status = BM_STATUS_UNKNOWN;

    /*
     * If this device doesn't contains an _STA object return now
     * with an 'unknown' status.  Note that this is not an error...
     */
    if (!Device->Info.Flags & BM_FLAGS_DYNAMIC_STATUS)
    {
        return AE_OK;
    }

    /*
     * Evaluate _STA...
     */
    if (ACPI_SUCCESS(Status = AcpiGetHandle(Device->Info.AcpiHandle, "_STA", &AcpiHandle)))
    {
        Status = UtEvaluateSimpleInteger(AcpiHandle, &(Device->Info.Status));
    }

    return Status;
}



ACPI_STATUS
BmCalculateDeviceInfo (
    ACPI_OBJECT_TYPE        AcpiType,
    BM_INTERNAL_DEVICE      *Device)
{
    ACPI_DEVICE_INFO        Info;
    ACPI_HANDLE             ChildAcpiHandle = NULL;
    ACPI_BUFFER             Buffer;

    if (!Device)
    {
        return AE_BAD_PARAMETER;
    }

    /*
     * Device Name:
     * ------------
     */
    Buffer.Length = sizeof(Device->Info.Name);
    Buffer.Pointer = &(Device->Info.Name);
    AcpiGetName(Device->Info.AcpiHandle, ACPI_SINGLE_NAME, &Buffer);

    /*
     * Device Class:
     * -------------
     */
    switch (AcpiType)
    {
    case INTERNAL_TYPE_Scope:
        Device->Info.Class = BM_CLASS_SCOPE;
        break;
    case ACPI_TYPE_Processor:
        Device->Info.Class = BM_CLASS_PROCESSOR;
        break;
    case ACPI_TYPE_Thermal:
        Device->Info.Class = BM_CLASS_THERMAL_ZONE;
        break;
    case ACPI_TYPE_Power:
        Device->Info.Class = BM_CLASS_POWER_RESOURCE;
        break;
    case ACPI_TYPE_Device:
        Device->Info.Class = BM_CLASS_DEVICE;
        break;
    }

    /* 
     * Device Flags:
     * -------------
     */
    switch (Device->Info.Class)
    {
    case BM_CLASS_DEVICE:
        /*
         * Presence of _DCK indicates a docking station.
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_DCK", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_DOCKING_STATION;
        }

        /*
         * Presence of _EJD and/or _EJx indicates 'ejectable'.
         * TODO: _EJx...
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_EJD", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_EJECTABLE;
        }

        /*
         * Presence of _PR0 or _PS0 indicates 'power manageable'.
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_PR0", &ChildAcpiHandle)) ||
            ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_PS0", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_POWER_MANAGEABLE;
        }

        /*
         * Presence of _CRS indicates 'configurable'.
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_CRS", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_CONFIGURABLE;
        }

        /* Fall through to the next case statement... */

    case BM_CLASS_PROCESSOR:
    case BM_CLASS_THERMAL_ZONE:
    case BM_CLASS_POWER_RESOURCE:

        /*
         * Power resources are by definition power manageable.
         */
        if (Device->Info.Class == BM_CLASS_POWER_RESOURCE)
        {
            Device->Info.Flags |= BM_FLAGS_POWER_MANAGEABLE;
        }

        /*
         * Presence of _HID or _ADR indicates 'identifiable'.
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_HID", &ChildAcpiHandle)) ||
           ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_ADR", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_IDENTIFIABLE;
        }

        /*
         * Presence of _STA indicates 'dynamic status'.
         */
        if (ACPI_SUCCESS(AcpiGetHandle(Device->Info.AcpiHandle, "_STA", &ChildAcpiHandle)))
        {
            Device->Info.Flags |= BM_FLAGS_DYNAMIC_STATUS;
        }
        
        break;
    }

    /*
     * Hardware ID & Status:
     * ---------------------
     * TODO: Note that AcpiGetObjectInfo() currently only works with 
     *       'real' devices (not thermal zones, processors, etc.), but
     *       these object may be assigned _HID, _STA, etc. in ACPI 2.0.
     */
    if (Device->Info.Class == BM_CLASS_DEVICE)
    {
        if (ACPI_SUCCESS(AcpiGetObjectInfo(Device->Info.AcpiHandle, &Info)))
        {
            /*
             * Hardware ID:
             * ------------
             */ 
            if (Info.Valid & ACPI_VALID_HID)
            {
                MEMCPY((void*)Device->Info.HardwareId, (void*)Info.HardwareId, sizeof(Device->Info.HardwareId));
            }

            /*
             * Status:
             * -------
             * Note that devices w/o dynamic status (no _STA) are assigned the 
             * default status (present, enabled, show in UI, and functioning).
             */
            if (Device->Info.Flags & BM_FLAGS_DYNAMIC_STATUS)
            {
                if (Info.Valid & ACPI_VALID_STA)
                {
                    Device->Info.Status = Info.CurrentStatus;
                }
                else
                {
                    Device->Info.Status = BM_STATUS_UNKNOWN;
                }
            }
            else
            {
                Device->Info.Status = BM_STATUS_DEFAULT;
            }
        }
    }

    return AE_OK;
}



ACPI_STATUS
BmAddDevice (
    ACPI_HANDLE             AcpiHandle,
    ACPI_OBJECT_TYPE        AcpiType,
    BM_INTERNAL_DEVICE      *ParentDevice,
    BM_INTERNAL_DEVICE      **ChildDevice)
{
    ACPI_STATUS             Status = AE_OK;
    BM_INTERNAL_DEVICE      *NewDevice = NULL;

    if (DeviceList.Count > BM_MAX_HANDLES)
    {
        return AE_NO_MEMORY;
    }

    /*
     * Allocate the new device and add to the device array.
     */ 
    NewDevice = OsdCallocate(sizeof(BM_INTERNAL_DEVICE));
    if (!NewDevice)
    {
        return AE_NO_MEMORY;
    }

    NewDevice->Info.DeviceHandle = DeviceList.Count;
    NewDevice->Info.AcpiHandle = AcpiHandle;

    /*
     * Add to the DeviceList.
     */
    DeviceList.Devices[DeviceList.Count++] = NewDevice;

    /*
     * Evaluate _INI:
     * --------------
     * Run the _INI control method, if present, before proceeding.  Note
     * that this MUST be executed before any other objects defined within 
     * the scope of the object are evaluated.
     */
    AcpiEvaluateObject(NewDevice->Info.AcpiHandle, "_INI", NULL, NULL);

    /*
     * Get Device Info:
     * ----------------
     */
    BmCalculateDeviceInfo(AcpiType, NewDevice);

    /*
     * Formulate Hierarchy:
     * --------------------
     * Arrange within the namespace by assigning the parent and
     * adding to the parent device's list of children (scope).
     */
    NewDevice->Parent = ParentDevice;
    NewDevice->Next = NULL;

    if (ParentDevice)
    {
        if (!ParentDevice->Scope.Head)
        {
            ParentDevice->Scope.Head = NewDevice;
        }
        else
        {
            if (!ParentDevice->Scope.Tail)
            {
                (ParentDevice->Scope.Head)->Next = NewDevice;
            }
            else
            {
                (ParentDevice->Scope.Tail)->Next = NewDevice;
            }

            ParentDevice->Scope.Tail = NewDevice;
        }
    }

    (*ChildDevice) = NewDevice;


#ifdef ACPI_DEBUG
    BmPrintDevice(NewDevice);
#endif

    return Status;
}


ACPI_STATUS
BmEnumerateNamespace(void)
{
    ACPI_HANDLE             ParentAcpiHandle = ACPI_ROOT_OBJECT;
    ACPI_HANDLE             ChildAcpiHandle = NULL;
    BM_INTERNAL_DEVICE      *ParentDevice = NULL;
    BM_INTERNAL_DEVICE      *ChildDevice = NULL;
    ACPI_OBJECT_TYPE        AcpiType = 0;
    UINT32                  Level = 1;

    FUNCTION_TRACE("BmEnumerateNamespace");

    /*
     * Root Object:
     * ------------
     * Fabricate the root object, which happens to always get a 
     * DeviceHandle of zero.
     */
    DeviceList.Devices[0] = OsdCallocate(sizeof(BM_INTERNAL_DEVICE));
    if (NULL == (DeviceList.Devices[0]))
    {
        return AE_NO_MEMORY;
    }

    DeviceList.Devices[0]->Info.DeviceHandle = BM_HANDLE_UNKNOWN;
    DeviceList.Devices[0]->Info.AcpiHandle = ACPI_ROOT_OBJECT;
    DeviceList.Devices[0]->Info.Class = BM_CLASS_SCOPE;
    DeviceList.Devices[0]->Info.HardwareId[0] = BM_HID_UNKNOWN;
    DeviceList.Devices[0]->Info.Flags = BM_FLAGS_UNKNOWN;
    DeviceList.Devices[0]->Info.Status = BM_STATUS_DEFAULT;

    ParentDevice = DeviceList.Devices[0];

    DeviceList.Count = 1;

    /*
     * Device Hierarchy:
     * -----------------
     * Parse through the ACPI namespace, identify all 'devices', 
     * and create a new entry for each in our collection.
     */ 
    while (Level > 0)
    {
        if (ACPI_SUCCESS(AcpiGetNextObject(ACPI_TYPE_Any, ParentAcpiHandle, ChildAcpiHandle, &ChildAcpiHandle)))
        {
           /*
            * TODO: This is a hack to get around the problem identifying Scope 
            *       objects.  Scopes somehow need to be uniquely identified...
            */
            if (ACPI_SUCCESS(AcpiGetType(ChildAcpiHandle, &AcpiType)))
            {
                if (AcpiType == ACPI_TYPE_Any)
                {
                    if (ACPI_SUCCESS(AcpiGetNextObject(ACPI_TYPE_Any, ChildAcpiHandle, 0, NULL)))
                    {
                        AcpiType = INTERNAL_TYPE_Scope;
                    }
                }
            }            

            switch (AcpiType)
            {
            case INTERNAL_TYPE_Scope:
            case ACPI_TYPE_Device:
            case ACPI_TYPE_Processor:
            case ACPI_TYPE_Thermal:
            case ACPI_TYPE_Power:
                /*
                 * Device?
                 * -------
                 * If this object is a 'device', insert into the ACPI Bus 
                 * Manager's local hierarchy and search the object's scope for 
                 * any child devices (depth-first).
                 */
                BmAddDevice(ChildAcpiHandle, AcpiType, ParentDevice, &ChildDevice);

                if (ACPI_SUCCESS(AcpiGetNextObject(ACPI_TYPE_Any, ChildAcpiHandle, 0, NULL)))
                {
                    Level++;
                    ParentAcpiHandle = ChildAcpiHandle;
                    ChildAcpiHandle = 0;
                    ParentDevice = ChildDevice;
                }
                break;
            }
        }

        else
        {
            /* 
             * Scope Exhausted:
             * ----------------
             * No more children in this object's scope (AcpiGetNextObject failed), 
             * Go back upwards in the namespace tree to the object's parent.
             */
            Level--;
            ChildAcpiHandle = ParentAcpiHandle;
            AcpiGetParent(ParentAcpiHandle, &ParentAcpiHandle);

            if (ParentDevice)
            {
                ParentDevice = ParentDevice->Parent;
            }
            else
            {
                /*
                 * We should never get here unless the namespace is really 
                 * messed up - punt.
                 */
                DEBUG_PRINT(ACPI_ERROR, ("Invalid parent for device [%p]", ParentDevice));
                return AE_NULL_ENTRY;
            }
        }
    }

    return_ACPI_STATUS(AE_OK);
}


ACPI_STATUS
BmGetDeviceStatus (
    BM_HANDLE               DeviceHandle,
    BM_DEVICE_STATUS        *DeviceStatus)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("BmGetDeviceStatus");

    if ((DeviceHandle >= DeviceList.Count) || !DeviceStatus)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    (*DeviceStatus) = BM_STATUS_UNKNOWN;

    if (ACPI_FAILURE(Status = BmCalculateDeviceStatus(DeviceList.Devices[DeviceHandle])))
    {
        return_ACPI_STATUS(Status);
    }

    (*DeviceStatus) = DeviceList.Devices[DeviceHandle]->Info.Status;
    
    return_ACPI_STATUS(Status);
}


ACPI_STATUS
BmGetDeviceInfo (
    BM_HANDLE               DeviceHandle,
    BM_DEVICE_INFO          **DeviceInfo)
{
    FUNCTION_TRACE("BmGetDeviceInfo");

    if ((DeviceHandle >= BM_MAX_HANDLES) || !DeviceInfo)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    *DeviceInfo = &(DeviceList.Devices[DeviceHandle]->Info);

    return_ACPI_STATUS(AE_OK);
}


ACPI_STATUS
BmGetDeviceHandle (
    ACPI_HANDLE             AcpiHandle,
    BM_HANDLE               *DeviceHandle)
{
    BM_INTERNAL_DEVICE      *Device = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("BmGetDeviceHandle");

    if (!DeviceHandle)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    for (i = 0; i < DeviceList.Count; i++)
    {
        Device = DeviceList.Devices[i];

        if (Device->Info.AcpiHandle == AcpiHandle)
        {
            (*DeviceHandle) = Device->Info.DeviceHandle;
            return_ACPI_STATUS(AE_OK);
        }
    }

    return_ACPI_STATUS(AE_NOT_FOUND);
}


ACPI_STATUS
BmGetDeviceContext (
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    FUNCTION_TRACE("BmGetDeviceContext");

    if ((DeviceHandle >= BM_MAX_HANDLES) || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    if (!(*Context = DeviceList.Devices[DeviceHandle]->Control.Context))
    {
        BmPrintDevice(DeviceList.Devices[DeviceHandle]);
        return_ACPI_STATUS(AE_NULL_ENTRY);
    }

    return_ACPI_STATUS(AE_OK);
}


ACPI_STATUS
BmSearch(
    BM_HANDLE               StartDeviceHandle,
    BM_DEVICE_INFO          *Criteria,
    BM_HANDLE_LIST          *Results)
{
    ACPI_STATUS             Status = AE_NOT_FOUND;
    BM_INTERNAL_DEVICE      *Device = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("BmSearch");

    if ((StartDeviceHandle >= BM_MAX_HANDLES) || !Criteria || !Results)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Results->Count = 0;

    // TODO: Use StartDeviceHandle to only search a branch of the device
    //       heirarchy (versus the whole thing).

    //if (StartDevice == DeviceList.Devices[0])
    {
        /*
         * Parse through the device array, looking for matches.
         */
        for (i = 0; i < DeviceList.Count; i++)
        {
            Device = DeviceList.Devices[i];
            if (!Device)
                continue;

            /*
             * Look for a match.
             */
            if (Criteria->Class && !(Criteria->Class & Device->Info.Class))
             continue;

            if (Criteria->Flags && !(Criteria->Flags & Device->Info.Flags))
                continue;

            // TODO: What about searches on DeviceInfo.Status???
            //       Problem is that a value of 0x00 means something other than UNKNOWN.

            if ((Criteria->HardwareId[0] != BM_HID_UNKNOWN) && (0 != STRNCMP(Criteria->HardwareId, Device->Info.HardwareId, sizeof(BM_DEVICE_HID))))
            {
                continue;
            }

            /*
             * Found a match.
             */
            Results->Handles[Results->Count++] = Device->Info.DeviceHandle;

            if (Results->Count > BM_MAX_HANDLES)
            {
                DEBUG_PRINT(ACPI_WARN, ("BmSearch: Found maximum number of matches - terminating.\n"));
                break;
            }
            else
            {
                DEBUG_PRINT(ACPI_INFO, ("BmSearch: Found DeviceHandle[0x%08X].\n", Device->Info.DeviceHandle));
            }
        }
    }

    if (Results->Count == 0)
    {
        DEBUG_PRINT(ACPI_INFO, ("BmSearch: Found no matching devices.\n"));
        Status = AE_NOT_FOUND;
    }
    else
    {
        Status = AE_OK;
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
BmRegisterControl (
    BM_DEVICE_INFO          *Criteria,
    BM_CONTROL_INFO         *Control)
{
    ACPI_STATUS             Status = AE_NOT_FOUND;
    BM_HANDLE_LIST          MatchingDevices;
    BM_INTERNAL_DEVICE      *Device = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("BmRegisterControl");

    if (!Criteria || !Control || !Control->Notify || !Control->Request)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&MatchingDevices, 0, sizeof(BM_HANDLE_LIST));

    /*
     * Find Matches:
     * -------------
     * Search through the entire device hierarchy for matches (given this
     * device Criteria).
     */
    if (ACPI_FAILURE(Status = BmSearch(BM_HANDLE_ROOT, Criteria, &MatchingDevices)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Install Control:
     * ----------------
     * For each match, record the control module information and execute
     * the control's Notify() funciton (if present) to notify the control
     * module of the device's presence.
     */ 
    for (i = 0; i < MatchingDevices.Count; i++)
    {
        Device = DeviceList.Devices[MatchingDevices.Handles[i]];
        if (!Device)
        {
            DEBUG_PRINT(ACPI_ERROR, ("BmRegisterControl: NULL device entry [0x%08X].\n", Device->Info.DeviceHandle));
            continue;
        }

        DEBUG_PRINT(ACPI_INFO, ("BmRegisterControl: Registering control for device [0x%08X].\n", Device->Info.DeviceHandle));

        MEMCPY(&(Device->Control), Control, sizeof(BM_CONTROL_INFO));

        if (ACPI_FAILURE(Device->Control.Notify(BM_NOTIFY_DEVICE_ADDED, Device->Info.DeviceHandle, &(Device->Control.Context))))
        {
            MEMSET(&(Device->Control), 0, sizeof(BM_CONTROL_INFO));
        }
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
BmUnregisterControl (
    BM_DEVICE_INFO          *Criteria,
    BM_CONTROL_INFO         *Control)
{
    ACPI_STATUS             Status = AE_NOT_FOUND;
    BM_HANDLE_LIST          MatchingDevices;
    BM_INTERNAL_DEVICE      *Device = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("BmUnregisterControl");

    if (!Criteria || !Control || !Control->Notify || !Control->Request)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&MatchingDevices, 0, sizeof(BM_HANDLE_LIST));

    /*
     * Find Matches:
     * -------------
     * Search through the entire device hierarchy for matches (given this
     * device Criteria).
     */
    if (ACPI_FAILURE(Status = BmSearch(BM_HANDLE_ROOT, Criteria, &MatchingDevices)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Remove Control:
     * ---------------
     * For each match, execute the control's Notify() function (if present)
     * to allow the control module to cleanup and remove all control-
     * specific information from the device.
     */ 
    for (i = 0; i < MatchingDevices.Count; i++)
    {
        Device = DeviceList.Devices[MatchingDevices.Handles[i]];

        if (!Device)
        {
            DEBUG_PRINT(ACPI_ERROR, ("BmUnregisterControl: NULL device entry [0x%08X].\n", Device->Info.DeviceHandle));
            continue;
        }

        DEBUG_PRINT(ACPI_INFO, ("BmUnregisterControl: Unregistering control for device [0x%08X].\n", Device->Info.DeviceHandle));

        Device->Control.Notify(BM_NOTIFY_DEVICE_REMOVED, Device->Info.DeviceHandle, &Device->Control.Context);

        MEMSET(&(Device->Control), 0, sizeof(BM_CONTROL_INFO));
    }

    return_ACPI_STATUS(AE_OK);
}

