/*
 *  ./acpica/busmgr/busmgr.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __BUSMGR_C__

#include <acpi.h>

#include "busmgr.h"
#include "bm.h"
#include "utility.h"


#define _COMPONENT          BUS_MANAGER
        MODULE_NAME         ("busmgr");



// TODO: Use a dynamically sizeable array versus fixed size one.
BM_INTERNAL_DEVICE_LIST     DeviceList;




ACPI_STATUS
BmInitialize(void)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("BmInitialize");

    MEMSET(&DeviceList, 0, sizeof(BM_HANDLE_LIST));

    /*
     * Enumerate the ACPI namespace, building an internal representation of
     * all 'devices'.
     */
    if (ACPI_FAILURE(Status = BmEnumerateNamespace()))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Register for all notifications.
     */
    if (ACPI_FAILURE(Status = AcpiInstallNotifyHandler(ACPI_ROOT_OBJECT, ACPI_SYSTEM_NOTIFY, &BmNotify, NULL)))
    {
        return_ACPI_STATUS(Status);
    }
    if (ACPI_FAILURE(Status = AcpiInstallNotifyHandler(ACPI_ROOT_OBJECT, ACPI_DEVICE_NOTIFY, &BmNotify, NULL)))
    {
        return_ACPI_STATUS(Status);
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
BmTerminate(void)
{
    UINT32                  i = 0;

    FUNCTION_TRACE("BmTerminate");

    /*
     * Parse through the device array, freeing all entries.
     */
    for (i = 0; i < DeviceList.Count; i++)
    {
        if (DeviceList.Devices[i])
        {
            //UtPrintDevice(DeviceList.Devices[i]);
            OsdFree(DeviceList.Devices[i]);
        }
    }

    return_ACPI_STATUS(AE_OK);
}



void
BmNotify (
    ACPI_HANDLE         AcpiHandle,
    UINT32              NotifyValue,
    void                *Context)
{
    BM_HANDLE           DeviceHandle = BM_HANDLE_UNKNOWN;
    BM_INTERNAL_DEVICE  *Device = NULL;

    FUNCTION_TRACE("BmNotify");

    DEBUG_PRINT(ACPI_INFO, ("BmNotify: Recieved notify [0x%08X]:[0x%08X]\n", AcpiHandle, NotifyValue));

    /*
     * Resolve the AcpiHandle to an internal device structure.
     */
    if (ACPI_FAILURE(BmGetDeviceHandle(AcpiHandle, &DeviceHandle)))
    {
        DEBUG_PRINT(ACPI_WARN, ("BmNotify: Unable to locate a device with this ACPI handle.\n"));
        return_VOID;
    }

    Device = DeviceList.Devices[DeviceHandle];
    if (!Device)
    {
        DEBUG_PRINT(ACPI_WARN, ("BmNotify: Invalid (NULL) device entry.\n"));
        return_VOID;
    }

    /*
     * Device-Specific Notify:
     * -----------------------
     * Forward this notify to the control module's Notify() function, if 
     * present.
     */
    if (NotifyValue > 0x7F)
    {
        if (Device->Control.Notify)
        {
            Device->Control.Notify(NotifyValue, Device->Info.DeviceHandle, &(Device->Control.Context));
        }
    }

    /*
     * Standard Notify:
     * ----------------
     */
    else
    {
        switch (NotifyValue)
        {
        case 0x00:
            /* Enumerate bus */
            // TODO...
            break;

        case 0x01:
            /* Check device (device has come or gone) */
            // TODO...
            break;

        case 0x02:
            /* Device asserting wake */
            // TODO...
            break;

        case 0x03:
            /* Device requesting eject */
            // TODO...
            break;

        default:
            break;
        }
    }

    return_VOID;
}



ACPI_STATUS
BmRequest (
    BM_REQUEST_INFO         *RequestInfo)
{
    ACPI_STATUS             Status = AE_OK;
    BM_INTERNAL_DEVICE      *Device = NULL;

    FUNCTION_TRACE("BmRequest");

    /*
     * Must have a valid request structure.
     */
    if (!RequestInfo)
    {
        return AE_BAD_PARAMETER;
    }

    DEBUG_PRINT(ACPI_INFO, ("BmRequest: DeviceHandle[0x%08X] Command[0x%08X]\n", RequestInfo->DeviceHandle, RequestInfo->Command));

    /*
     * Valid Device?
     */
    if ((RequestInfo->DeviceHandle > BM_MAX_HANDLES) || (!DeviceList.Devices[RequestInfo->DeviceHandle]))
    {
        RequestInfo->Status = AE_NOT_FOUND;
        return AE_NOT_FOUND;
    }

    Device = DeviceList.Devices[RequestInfo->DeviceHandle];

    /*
     * Device-Specific Request?
     * ------------------------
     * If a device-specific command (>=0x80), forward to the appropriate
     * control module (if exists).  
     */
    if (RequestInfo->Command & BM_COMMAND_DEVICE_SPECIFIC)
    {
        if (Device->Control.Request)
        {
            Status = Device->Control.Request(RequestInfo, Device->Control.Context);
        }
        else
        {
            Status = AE_SUPPORT;
        }
    }

    /*
     * Standard Requests:
     * ------------------
     * Handle commands specific to the ACPI Bus Manager (<0x80).
     */
    else
    {
        switch (RequestInfo->Command)
        {
        case BM_COMMAND_SEARCH:
            {
                BM_SEARCH_INFO *SearchInfo = NULL;

                if (ACPI_SUCCESS(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&SearchInfo, sizeof(BM_SEARCH_INFO))))
                {
                    Status = BmSearch(Device->Info.DeviceHandle, &(SearchInfo->Criteria), &(SearchInfo->Results));
                }
            }
            break;

        case BM_COMMAND_GET_INFO:
            Status = UtCopyToBuffer(&(RequestInfo->Buffer), &(Device->Info), sizeof(BM_DEVICE_INFO));
            break;

        case BM_COMMAND_GET_STATUS:
            {
                BM_DEVICE_STATUS DeviceStatus = BM_STATUS_UNKNOWN;

                if(ACPI_SUCCESS(Status = BmGetDeviceStatus(Device->Info.DeviceHandle, &DeviceStatus)))
                {
                    Status = UtCopyToBuffer(&(RequestInfo->Buffer), &DeviceStatus, sizeof(BM_DEVICE_STATUS));
                }
            }
            break;

        case BM_COMMAND_GET_CURRENT_RESOURCES:
            if (Device->Info.Flags & BM_FLAGS_CONFIGURABLE)
            {
                Status = AcpiGetCurrentResources(Device->Info.AcpiHandle, &(RequestInfo->Buffer));
            }
            else
            {
                Status = AE_SUPPORT;
            }
            break;

        case BM_COMMAND_GET_POSSIBLE_RESOURCES:
            if (Device->Info.Flags & BM_FLAGS_CONFIGURABLE)
            {
                Status = AcpiGetPossibleResources(Device->Info.AcpiHandle, &(RequestInfo->Buffer));
            }
            else
            {
                Status = AE_SUPPORT;
            }
            break;

        case BM_COMMAND_SET_CURRENT_RESOURCES:
            if (Device->Info.Flags & BM_FLAGS_CONFIGURABLE)
            {
                Status = AcpiSetCurrentResources(Device->Info.AcpiHandle, &(RequestInfo->Buffer));
            }
            else
            {
                Status = AE_SUPPORT;
            }
            break;

        default:
            Status = AE_SUPPORT;
            break;
        }

        RequestInfo->Status = Status;
    }

    return_ACPI_STATUS(Status);
}