/*
 *  ./acpica/busmgr/utility.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __UTILITY_C__

#include <acpi.h>

#include "busmgr.h"
#include "bm.h"


#define _COMPONENT          BUS_MANAGER
        MODULE_NAME         ("utility");




void
UtPrintObject (
    ACPI_HANDLE             AcpiHandle)
{
    ACPI_BUFFER             Buffer;
    ACPI_HANDLE             Parent;
    ACPI_OBJECT_TYPE        Type;

    Buffer.Length = 256;
    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return;
    }

    AcpiGetName(AcpiHandle, ACPI_FULL_PATHNAME, &Buffer);
    AcpiGetParent(AcpiHandle, &Parent);
    AcpiGetType(AcpiHandle, &Type);

    /*
     * TODO: Hack to get around scope identification problem.
     */
    if (Type == ACPI_TYPE_Any)
    {
        if (ACPI_SUCCESS(AcpiGetNextObject(ACPI_TYPE_Any, AcpiHandle, 0, NULL)))
        {
            Type = INTERNAL_TYPE_Scope;
        }
    }
    
    switch (Type)
    {
    case INTERNAL_TYPE_Scope:
        OsdPrintf("SCOPE: ");
        break;
    case ACPI_TYPE_Number:
        OsdPrintf("SIMPLE (Number): ");
        break;
    case ACPI_TYPE_String:
        OsdPrintf("SIMPLE (String): ");
        break;
    case ACPI_TYPE_Buffer:
        OsdPrintf("SIMPLE (Buffer): ");
        break;
    case ACPI_TYPE_Package:
        OsdPrintf("SIMPLE (Package): ");
        break;
    case ACPI_TYPE_FieldUnit:
        OsdPrintf("FIELD UNIT: ");
        break;
    case ACPI_TYPE_Device:
        OsdPrintf("DEVICE: ");
        break;
    case ACPI_TYPE_Event:
        OsdPrintf("EVENT: ");
        break;
    case ACPI_TYPE_Method:
        OsdPrintf("CONTROL METHOD: ");
        break;
    case ACPI_TYPE_Mutex:
        OsdPrintf("MUTEX: ");
        break;
    case ACPI_TYPE_Region:
        OsdPrintf("OPERATION REGION: ");
        break;
    case ACPI_TYPE_Power:
        OsdPrintf("POWER RESOURCE: ");
        break;
    case ACPI_TYPE_Processor:
        OsdPrintf("PROCESSOR: ");
        break;
    case ACPI_TYPE_Thermal:
        OsdPrintf("THERMAL ZONE: ");
        break;
    case ACPI_TYPE_BufferField:
        OsdPrintf("BUFFER FIELD: ");
        break;
    case ACPI_TYPE_DdbHandle:
        OsdPrintf("DDB HANDLE: ");
        break;
    default:
        OsdPrintf("OTHER (%d): ", Type);
        break;
    }

    OsdPrintf("Object[%p][%s] Parent[%p].\n", AcpiHandle, (char*)Buffer.Pointer, Parent);

    OsdFree(Buffer.Pointer);
}



ACPI_STATUS
UtCopyToBuffer(
    ACPI_BUFFER             *Buffer,
    void                    *Data,
    UINT32                  Length)
{
    FUNCTION_TRACE("UtCopyToBuffer");

    if (!Buffer || (!Buffer->Pointer) || !Data || (Length == 0))
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    if (Length > Buffer->Length)
    {
        Buffer->Length = Length;
        return_ACPI_STATUS(AE_BUFFER_OVERFLOW);
    }

    Buffer->Length = Length;
    MEMCPY(Buffer->Pointer, Data, Length);

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
UtCastBuffer(
    ACPI_BUFFER             *Buffer,
    void                    **Pointer,
    UINT32                  Length)
{
    FUNCTION_TRACE("UtCastBuffer");

    if (!Buffer || !Buffer->Pointer || !Pointer || Length == 0)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    if (Length > Buffer->Length)
    {
        return_ACPI_STATUS(AE_BAD_DATA);
    }

    *Pointer = Buffer->Pointer;

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
UtExtractPackageData(
    ACPI_OBJECT             *Package,
    ACPI_BUFFER             *PackageFormat,
    ACPI_BUFFER             *Buffer)
{
    ACPI_STATUS             Status = AE_OK;
    UINT8                   *Head = NULL;
    UINT8                   *Tail = NULL;
    UINT8                   **Pointer = NULL;
    UINT32                  TailOffset = 0;
    ACPI_OBJECT             *Element = NULL;
    UINT32                  SizeRequired = 0;
    char*                   Format = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("UtExtractPackageData");

    if (!Package || (Package->Type != ACPI_TYPE_Package) || (Package->Package.Count == 0) ||
        !PackageFormat || (PackageFormat->Length < 1) || (!PackageFormat->Pointer) ||
        !Buffer)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    if ((PackageFormat->Length - 1) > Package->Package.Count)
    {
        return_ACPI_STATUS(AE_BAD_DATA);
    }

    Format = (char*)PackageFormat->Pointer;

    /*
     * Calculate SizeRequired:
     */
    for (i = 0; i < (PackageFormat->Length - 1); i++)
    {
        Element = &(Package->Package.Elements[i]);

        switch (Element->Type)
        {
        case ACPI_TYPE_Number:
            if (Format[i] != 'N')
            {
                return_ACPI_STATUS(AE_BAD_DATA);
            }
            SizeRequired += sizeof(UINT32);
            TailOffset += sizeof(UINT32);
            break;

        case ACPI_TYPE_String:
            if (Format[i] != 'S')
            {
                return_ACPI_STATUS(AE_BAD_DATA);
            }
            SizeRequired += sizeof(UINT8*) + Element->String.Length + 1;
            TailOffset += sizeof(UINT8*);
            break;

        case ACPI_TYPE_Buffer:
            if (Format[i] != 'B')
            {
                return_ACPI_STATUS(AE_BAD_DATA);
            }
            SizeRequired += sizeof(UINT8*) + Element->Buffer.Length;
            TailOffset += sizeof(UINT8*);
            break;

        case ACPI_TYPE_Package:
        default:
            // TODO: Handle nexted packages...
            return_ACPI_STATUS(AE_SUPPORT);
            break;
        }
    }

    if (SizeRequired > Buffer->Length)
    {
        Buffer->Length = SizeRequired;
        return_ACPI_STATUS(AE_BUFFER_OVERFLOW);
    }

    Buffer->Length = SizeRequired;

    if (!Buffer->Pointer)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Head = Buffer->Pointer;
    Tail = Buffer->Pointer + TailOffset;

    /*
     * Extract Package Data:
     */
    for (i = 0; i < (PackageFormat->Length - 1); i++)
    {
        Element = &(Package->Package.Elements[i]);

        switch (Element->Type)
        {
        case ACPI_TYPE_Number:
            *((UINT32*)Head) = Element->Number.Value;
            Head += sizeof(UINT32);
            break;

        case ACPI_TYPE_String:
            Pointer = (UINT8**)Head;
            *Pointer = Tail;
            memcpy(Tail, Element->String.Pointer, Element->String.Length);
            Head += sizeof(UINT8*);
            Tail += Element->String.Length;

            /* NULL terminate string */
            *Tail = 0;
            Tail++;
            break;

        case ACPI_TYPE_Buffer:
            Pointer = (UINT8**)Head;
            *Pointer = Tail;
            memcpy(Tail, Element->Buffer.Pointer, Element->Buffer.Length);
            Head += sizeof(UINT8*);
            Tail += Element->Buffer.Length;
            break;

        case ACPI_TYPE_Package:
            // TODO: Handle nexted packages...
            break;
        }
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
UtEvaluateSimpleInteger(
    ACPI_HANDLE             AcpiHandle,
    UINT32                  *Data)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT             *Element = NULL;
    ACPI_BUFFER             Buffer;

    FUNCTION_TRACE("UtEvaluateSimpleInteger");

    if (!Data)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Buffer.Length = sizeof(ACPI_OBJECT);
    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    if (ACPI_FAILURE(Status = AcpiEvaluateObject(AcpiHandle, NULL, NULL, &Buffer)))
    {
        goto Cleanup;
    }

    if (ACPI_FAILURE(Status = UtCastBuffer(&Buffer, (void**)&Element, sizeof(ACPI_OBJECT))))
    {
        goto Cleanup;
    }

    if (Element->Type != ACPI_TYPE_Number)
    {
        Status = AE_BAD_DATA;
    }
    else
    {
        *Data = Element->Number.Value;
    }

Cleanup:

    OsdFree(Buffer.Pointer);
    
    return_ACPI_STATUS(Status);
}



ACPI_STATUS
UtEvaluateReferenceList(
    ACPI_HANDLE             AcpiHandle,
    BM_HANDLE_LIST          *HandleList)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT             *Package = NULL;
    ACPI_OBJECT             *Element = NULL;
    ACPI_BUFFER             Buffer;
    UINT32                  i = 0;
    
    FUNCTION_TRACE("UtEvaluateReferenceList");

    if (!HandleList)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&Buffer, 0, sizeof(ACPI_BUFFER));

    /*
     * Get Buffer Size:
     * ----------------
     * First AcpiEvaluateObject() attempt is simply to get the correct
     * buffer size.
     */
    Status = AcpiEvaluateObject(AcpiHandle, NULL, NULL, &Buffer);

    if (Status == AE_BUFFER_OVERFLOW)
    {
        Buffer.Pointer = OsdCallocate(Buffer.Length);
        if (!Buffer.Pointer)
        {
            return_ACPI_STATUS(AE_NO_MEMORY);
        }
    }
    else
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Evaluate the object.
     */
    if (ACPI_FAILURE(Status = AcpiEvaluateObject(AcpiHandle, NULL, NULL, &Buffer)))
    {
        goto Cleanup;
    }

    /*
     * Validate the package.
     */
    if (ACPI_FAILURE(Status = UtCastBuffer(&Buffer, (void**)&Package, sizeof(ACPI_OBJECT))))
    {
        goto Cleanup;
    }

    if (Package->Type != ACPI_TYPE_Package)
    {
        Status = AE_BAD_DATA;
        goto Cleanup;
    }

    if (Package->Package.Count > BM_MAX_HANDLES)
    {
        Package->Package.Count = BM_MAX_HANDLES;
    }

    /*
     * Parse the package.
     */
    for (i = 0; i < Package->Package.Count; i++)
    {
        Element = &(Package->Package.Elements[i]);

        if (!Element || (Element->Type != ACPI_TYPE_Any))
        {
            Status = AE_BAD_DATA;
            DEBUG_PRINT(ACPI_WARN, ("UtEvaluateReferenceList: Invalid element (not a reference) in package.\n"));
            // TODO: continue or break?
            continue;
        }

        if (ACPI_FAILURE(Status = BmGetDeviceHandle(Element->Reference.Handle, &(HandleList->Handles[i]))))
        {
            DEBUG_PRINT(ACPI_WARN, ("UtEvaluateReferenceList: Unable to locate referenced device w/ ACPI handle [0x%08X] in device hierarchy.\n", Element->Reference.Handle));
            // TODO: continue or break?
            continue;
        }

        DEBUG_PRINT(ACPI_INFO, ("UtEvaluateReferenceList: Found object reference [0x%08X]:[0x%08X]\n", Element->Reference.Handle, HandleList->Handles[i]));

        (HandleList->Count)++;
    }

Cleanup:

    OsdFree(Buffer.Pointer);

    return_ACPI_STATUS(Status);
}


