/*
 *  ./acpica/control/ec/ec.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * TODO:    1. Resolve Fujitsu GPE problem (if status bit is set when enable
 *             occurs then the system goes out to lunch).
 *          2, Make full use of EC-specific semaphores.
 */


#define __EC_C__

#include <acpi.h>

#include <bm.h>
#include <utility.h>

#include "ec.h"

#define _COMPONENT          EMBEDDED_CONTROLLER
        MODULE_NAME         ("ec");




ACPI_STATUS
EcGetPortValues(
    EC_INFO                 *EmbeddedController)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_BUFFER             Buffer;
    UINT8                   *Pointer = NULL;
    RESOURCE                *FirstResource = NULL;
    RESOURCE                *SecondResource = NULL;

    FUNCTION_TRACE("EcGetPortValues");

    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Buffer.Length = 0;
    Buffer.Pointer = NULL;

    Status = AcpiGetCurrentResources(EmbeddedController->AcpiHandle, &Buffer);
    if (Status != AE_BUFFER_OVERFLOW)
    {
        return_ACPI_STATUS(Status);
    }

    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    if (ACPI_FAILURE(Status = AcpiGetCurrentResources(EmbeddedController->AcpiHandle, &Buffer)))
    {
        goto End;
    }

    Pointer = Buffer.Pointer;
    FirstResource = (RESOURCE*)(Pointer);

    Pointer += FirstResource->Length;
    SecondResource = (RESOURCE*)(Pointer);

    if ((FirstResource->Id != Io) || (SecondResource->Id != Io))
    {
        Status = AE_BAD_DATA;
        goto End;
    }

    EmbeddedController->DataPort = FirstResource->Data.Io.MinBaseAddress;
    EmbeddedController->StatusPort = EmbeddedController->CommandPort = SecondResource->Data.Io.MinBaseAddress;

End:
    OsdFree(Buffer.Pointer);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcAddDevice(
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    EC_INFO                 *EmbeddedController = NULL;
    BM_DEVICE_INFO          *DeviceInfo = NULL;

    FUNCTION_TRACE("EcAddDevice");

    DEBUG_PRINT(ACPI_INFO, ("EcAddDevice: EC device [0x%08X].\n", DeviceHandle));

    if (!Context || *Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Get information on this device.
     */
    if (ACPI_FAILURE(Status = BmGetDeviceInfo(DeviceHandle, &DeviceInfo)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Allocate a new EC_INFO structure.
     */ 
    EmbeddedController = OsdCallocate(sizeof(EC_INFO));
    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    EmbeddedController->DeviceHandle = DeviceInfo->DeviceHandle;
    EmbeddedController->AcpiHandle = DeviceInfo->AcpiHandle;

    /*
     * Create a semaphore for serializing transactions to this EC.
     */
    if (ACPI_FAILURE(Status = OsdCreateSemaphore(1, &(EmbeddedController->Semaphore))))
    {
        goto End;
    }

    /*
     * Get the I/O port addresses for the command/status and data ports.
     */
    if (ACPI_FAILURE(Status = EcGetPortValues(EmbeddedController)))
    {
        goto End;
    }

    /*
     * Install a handler for servicing this EC's GPE.
     */
    if (ACPI_FAILURE(Status = EcInstallGpeHandler(EmbeddedController)))
    {
        goto End;
    }

    EmbeddedController->GpeHandler = TRUE;

    /*
     * Install a handler for servicing this EC's address space.
     */
    if (ACPI_FAILURE(Status = EcInstallSpaceHandler(EmbeddedController)))
    {
        goto End;
    }

    EmbeddedController->SpaceHandler = TRUE;

    /*
     * Context now contains information specific to this EC.  Note
     * that we'll get this pointer back on every Request() and Notify().
     */ 
    *Context = EmbeddedController;

End:
    if (ACPI_FAILURE(Status))
    {
        if (EmbeddedController->Semaphore)
        {
            OsdDeleteSemaphore(EmbeddedController->Semaphore);
        }

        if (EmbeddedController->GpeHandler)
        {
            EcRemoveGpeHandler(EmbeddedController);
        }

        if (EmbeddedController->SpaceHandler)
        {
            EcRemoveSpaceHandler(EmbeddedController);
        }

        OsdFree(EmbeddedController);
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcRemoveDevice(
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    EC_INFO                 *EmbeddedController = NULL;

    FUNCTION_TRACE("EcRemoveDevice");

    if (!Context || !*Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Context contains information specific to this EC.
     */ 
    EmbeddedController = (EC_INFO*)*Context;

    DEBUG_PRINT(ACPI_INFO, ("EcRemoveDevice: EC device [0x%08X].\n", EmbeddedController->DeviceHandle));

    /*
     * Remove the address space and GPE handlers.
     */ 
    EcRemoveSpaceHandler(EmbeddedController);
    EcRemoveGpeHandler(EmbeddedController);

    OsdDeleteSemaphore(EmbeddedController->Semaphore);

    OsdFree(EmbeddedController);

    *Context = NULL;

    return_ACPI_STATUS(Status);
}




ACPI_STATUS
EcInitialize(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("EcInitialize");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    MEMCPY(Criteria.HardwareId, BM_HID_EMBEDDED_CONTROLLER, sizeof(BM_HID_EMBEDDED_CONTROLLER));

    Control.Notify = &EcNotify;
    Control.Request = &EcRequest;

    Status = BmRegisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcTerminate(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("EcTerminate");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    MEMCPY(Criteria.HardwareId, BM_HID_EMBEDDED_CONTROLLER, sizeof(BM_HID_EMBEDDED_CONTROLLER));

    Control.Notify = &EcNotify;
    Control.Request = &EcRequest;

    if (ACPI_FAILURE(Status = BmUnregisterControl(&Criteria, &Control)))
    {
        DEBUG_PRINT(ACPI_WARN, ("EcTerminate: BmUnregister() returned error [0x%08X].\n", Status));
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS 
EcNotify (
    BM_NOTIFY               NotifyType,
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("EcNotify");

    switch (NotifyType)
    {
    case BM_NOTIFY_DEVICE_ADDED:
        Status = EcAddDevice(DeviceHandle, Context);
        break;

    case BM_NOTIFY_DEVICE_REMOVED:
        DEBUG_PRINT(ACPI_INFO, ("EcNotify: Removing EC device [0x%08X].\n", DeviceHandle));
        Status = EcRemoveDevice(Context);
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    return_ACPI_STATUS(Status);
}    



ACPI_STATUS
EcRequest (
    BM_REQUEST_INFO         *RequestInfo,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    EC_REQUEST              *EcRequest = NULL;
    EC_INFO                 *EmbeddedController = NULL;

    FUNCTION_TRACE("EcRequest");

    /*
     * Must have a valid request structure and Context.
     */
    if (!RequestInfo || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Buffer must contain a valid EC_REQUEST structure.
     */
    if (ACPI_FAILURE(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&EcRequest, sizeof(EC_REQUEST))))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Context contains information specific to this EC.
     */ 
    EmbeddedController = (EC_INFO*)Context;

    /*
     * Obtaining semaphore to enforce serialization of all EC transactions.
     */
    if (ACPI_FAILURE(Status = OsdWaitSemaphore(EmbeddedController->Semaphore, 1, EC_DEFAULT_TIMEOUT)))
    {
        goto End;
    }

    /*
     * Perform the Transaction.
     */
    Status = EcTransaction(EmbeddedController, EcRequest);

    /*
     * Signal the semaphore to indicate transaction completion.
     */
    OsdSignalSemaphore(EmbeddedController->Semaphore, 1);

End:
    RequestInfo->Status = Status;

    return_ACPI_STATUS(Status);
}
