/*
 *  ./acpica/control/ec/ecgpe.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ECGPE_C__

#include <acpi.h>

#include "ec.h"

#define _COMPONENT          EMBEDDED_CONTROLLER
        MODULE_NAME         ("ecgpe");



static char                 hex[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};




void
EcGpeHandler(
    void                    *Context)
{
    ACPI_STATUS             Status = AE_ERROR;
    EC_INFO                 *EmbeddedController = NULL;
    EC_STATUS               EcStatus = 0;
    UINT8                   Data = 0;
    char                    ObjectName[5] = {'_', 'Q', '0', '0', '\0'};

    FUNCTION_TRACE("EcGpeHandler");

    if (!Context)
    {
        DEBUG_PRINT(ACPI_ERROR, ("EcGpeHandler: Invalid (NULL) context.\n"));
        return_VOID;
    }

    EmbeddedController = (EC_INFO*)Context;

    /*
     * Obtaining semaphore to enforce serialization of all EC transactions.
     */
    if (ACPI_FAILURE(OsdWaitSemaphore(EmbeddedController->Semaphore, 1, EC_DEFAULT_TIMEOUT)))
    {
        return_VOID;
    }

    /*
     * Check EC_SCI:
     * -------------
     * Make sure this is a 'real' EC SCI by checking the EC_SCI bit of the
     * status register.  Note that this function should only be called when
     * this bit is set (polling is used to detect IBE/OBF events).
     */
    EcStatus = OsdIn8(EmbeddedController->StatusPort);
    if (EcStatus & EC_EVENT_SCI)
    {
        /*
         * Perform the Query
         */
        Status = EcQuery(EmbeddedController, &Data);
    }

    /*
     * Signal the semaphore to indicate transaction completion.
     */
    OsdSignalSemaphore(EmbeddedController->Semaphore, 1);

    /*
     * Evaluate _Qxx:
     * --------------
     */
    if (ACPI_SUCCESS(Status))
    {
        ObjectName[2] = hex[((Data>>4) & 0x0F)];
        ObjectName[3] = hex[(Data & 0x0F)];

        DEBUG_PRINT(ACPI_INFO, ("EcGpeHandler: Evaluating [%s].\n", ObjectName));

        Status = AcpiEvaluateObject(EmbeddedController->AcpiHandle, ObjectName, NULL, NULL);
    }

    return_VOID;
}



ACPI_STATUS
EcInstallGpeHandler(
    EC_INFO                 *EmbeddedController)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  *DataType = NULL;
    UINT32                  *DataValue = NULL; 
    ACPI_BUFFER             Buffer;
    
    FUNCTION_TRACE("EcInstallGpeHandler");

    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Buffer.Length = 16;
    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    /*
     * Evaluate _GPE:
     * --------------
     * Evaluate the "_GPE" object (required) to find out which GPE bit
     * is used by this EC to signal events (SCIs).
     */
    if (ACPI_FAILURE(Status = AcpiEvaluateObject(EmbeddedController->AcpiHandle, "_GPE", NULL, &Buffer)))
    {
        goto Cleanup;
    }
        
    DataType = (UINT32*)Buffer.Pointer;
    DataValue = (UINT32*)(Buffer.Pointer + sizeof(UINT32));

    if ((*DataType) != ACPI_TYPE_Number)
    {
        goto Cleanup;
    }
 
    EmbeddedController->GpeBit = (*DataValue);

    /*
     * Install GPE Handler:
     * --------------------
     * Install a handler for this EC's GPE bit.  Note that EC SCIs are 
     * treated as both edge- and level-triggered interrupts; in other words
     * we clear the status bit immediately after getting an EC-SCI, then
     * again after we're done processing the event.  This guarantees that
     * events we cause while performing a transaction (e.g. IBE/OBF) get 
     * cleared before re-enabled the GPE.
     */
    if (ACPI_FAILURE(Status = AcpiInstallGpeHandler(EmbeddedController->GpeBit, EVENT_LEVEL_TRIGGERED | EVENT_EDGE_TRIGGERED, &EcGpeHandler, EmbeddedController)))
    {
        DEBUG_PRINT(ACPI_ERROR, ("EcInstallGpeHandler: AcpiInstallGpeHandler() failed for GPE bit [0x%02x] with status [0x%08X].\n", EmbeddedController->GpeBit, Status));
        EmbeddedController->GpeBit = EC_GPE_UNKNOWN;
        goto Cleanup;
    }

    DEBUG_PRINT(ACPI_INFO, ("EcInstallGpeHandler: Handler installed for GPE bit [0x%02x]\n", EmbeddedController->GpeBit));

Cleanup:
    OsdFree(Buffer.Pointer);
    
    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcRemoveGpeHandler(
    EC_INFO                 *EmbeddedController)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("EcRemoveGpeHandler");

    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Status = AcpiRemoveGpeHandler(EmbeddedController->GpeBit, &EcGpeHandler);
    
    return_ACPI_STATUS(Status);
}

    
