/*
 *  ./acpica/control/ec/ecspace.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ECSPACE_C__

#include <acpi.h>

#include "ec.h"

#define _COMPONENT          EMBEDDED_CONTROLLER
        MODULE_NAME         ("ecspace");






ACPI_STATUS 
EcSpaceSetup (
    ACPI_HANDLE                 RegionHandle,
    UINT32                      Function,
    void                        *HandlerContext,
    void                        **ReturnContext)
{

 /* 
  * The EmbeddedController object is in the Handler context and is needed
  * when calling the ExSpaceHandler
  */
 *ReturnContext = HandlerContext;

    return AE_OK;
}




ACPI_STATUS
EcSpaceHandler (
    UINT32                  Function,
    UINT32                  Address,
    UINT32                  BitWidth,
    UINT32                  *Value,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    EC_INFO                 *EmbeddedController = NULL;
    EC_REQUEST              EcRequest;

    FUNCTION_TRACE("EcSpaceHandler");

    if (Address > 0xFF || BitWidth != 8 || !Value || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    EmbeddedController = (EC_INFO*)Context;

    switch (Function)
    {
    case ADDRESS_SPACE_READ:
        EcRequest.Command = EC_COMMAND_READ;
        EcRequest.Address = Address;
        EcRequest.Data = 0;
        break;

    case ADDRESS_SPACE_WRITE:
        EcRequest.Command = EC_COMMAND_WRITE;
        EcRequest.Address = Address;
        EcRequest.Data = (UINT8)(*Value);
        break;

    default:
        DEBUG_PRINT(ACPI_WARN, ("EcSpaceHandler: Received request with invalid function [0x%08X].\n", Function));
        return_ACPI_STATUS(AE_BAD_PARAMETER);
        break;
    }

    /*
     * Obtaining semaphore to enforce serialization of all EC transactions.
     */
    if (ACPI_FAILURE(Status = OsdWaitSemaphore(EmbeddedController->Semaphore, 1, EC_DEFAULT_TIMEOUT)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Perform the Transaction.
     */
    if (ACPI_SUCCESS(Status = EcTransaction(EmbeddedController, &EcRequest)))
    {
        (*Value) = (UINT32)EcRequest.Data;
    }

    /*
     * Signal the semaphore to indicate transaction completion.
     */
    OsdSignalSemaphore(EmbeddedController->Semaphore, 1);

    DEBUG_PRINT(ACPI_INFO, ("EcSpaceHandler: DeviceHandle[0x%08X] Command[0x%08X] Address[0x%08X] Data[0x%08X].\n", EmbeddedController->DeviceHandle, EcRequest.Command, EcRequest.Address, EcRequest.Data));

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcInstallSpaceHandler(
    EC_INFO                 *EmbeddedController)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("EcInstallSpaceHandler");

    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Status = AcpiInstallAddressSpaceHandler (EmbeddedController->AcpiHandle, ADDRESS_SPACE_EC, 
                    &EcSpaceHandler, &EcSpaceSetup, EmbeddedController);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
EcRemoveSpaceHandler(
    EC_INFO                 *EmbeddedController)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("EcRemoveSpaceHandler");

    if (!EmbeddedController)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Status = AcpiRemoveAddressSpaceHandler(EmbeddedController->AcpiHandle, ADDRESS_SPACE_EC, &EcSpaceHandler);

    return_ACPI_STATUS(Status);
}

