/*
 *  ./acpica/control/power/pwdevice.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __PWDEVICE_C__

#include <acpi.h>

#include <utility.h>

#include "power.h"


#define _COMPONENT          POWER_CONTROL
        MODULE_NAME         ("pwdevice");




ACPI_STATUS
PwDevicePrint(
    PW_DEVICE_INFO          *PowerDevice)
{
    FUNCTION_TRACE("PwDevicePrint");

    if (!PowerDevice)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
PwDeviceGetInfo(
    PW_DEVICE_INFO          *PowerDevice,
    ACPI_BUFFER             *Buffer)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PwDeviceGetInfo");

    if (!PowerDevice || !Buffer)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Status = UtCopyToBuffer(Buffer, PowerDevice, sizeof(PW_DEVICE_INFO));

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceGetInferredPowerState(
    PW_DEVICE_INFO          *PowerDevice,
    PW_POWER_STATE          *PowerState)
{
    ACPI_HANDLE             AcpiHandle = NULL;
    BM_HANDLE_LIST          HandleList;
    PW_POWER_STATE          ResourceListPowerState = PW_POWER_STATE_UNKNOWN;
    char                    ObjectName[5] = {'_', 'P', 'R', '0', '\0'};
    UINT32                  i = 0;

    FUNCTION_TRACE("PwDeviceGetInferredPowerState");

    if (!PowerDevice || !PowerState)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&HandleList, 0, sizeof(BM_HANDLE_LIST));

    (*PowerState) = PW_DEVICE_POWER_STATE_D3;

    /*
     * Calculate Power State:
     * ----------------------
     * Try to infer the PowerDevice's power state by checking the state of 
     * the PowerDevice's power resources.  We start by evaluating _PR0 (resource
     * requirements at D0) and work through _PR1 and _PR2.  We know the
     * current PowerDevice power state when all resources (for a state) are ON.
     * If, however, no power resources are fully on then the device is assumed
     * to be off (D3).
     */ 
    for (i = PW_DEVICE_POWER_STATE_D0; i < PW_DEVICE_POWER_STATE_D3; i++)
    {
        if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, ObjectName, &AcpiHandle)))
        {
            if (ACPI_SUCCESS(UtEvaluateReferenceList(AcpiHandle, &HandleList)))
            {
                if (ACPI_SUCCESS(PwResourceListGetPowerState(&HandleList, &ResourceListPowerState)))
                {
                    if (ResourceListPowerState == PW_RESOURCE_POWER_STATE_ON)
                    {
                        (*PowerState) = i;
                        break;
                    }
                }
            }
        }

        ObjectName[3]++;
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
PwDeviceGetPowerState(
    PW_DEVICE_INFO          *PowerDevice,
    PW_POWER_STATE          *PowerState)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             AcpiHandle = NULL;

    FUNCTION_TRACE("PwDeviceGetPowerState");

    if (!PowerDevice || !PowerState)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    (*PowerState) = PW_POWER_STATE_UNKNOWN;

    if (PowerDevice->Flags & PW_FLAGS_PSC)
    {
        if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PSC", &AcpiHandle)))
        {
            Status = UtEvaluateSimpleInteger(AcpiHandle, PowerState);
        }
    }
    else
    {
        Status = PwDeviceGetInferredPowerState(PowerDevice, PowerState);
    }

    DEBUG_PRINT(ACPI_INFO, ("Current power state is [%d]\n", *PowerState));

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceSetPowerState(
    PW_DEVICE_INFO          *PowerDevice,
    PW_POWER_STATE          CurrentPowerState,
    PW_POWER_STATE          TargetPowerState)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             AcpiHandle = NULL;
    BM_HANDLE_LIST          CurrentResourceList;
    BM_HANDLE_LIST          TargetResourceList;
    char                    ObjectName[5] = {'_', 'P', 'R', '0', '\0'};

    FUNCTION_TRACE("PwDeviceSetPowerState");

    if (!PowerDevice || (TargetPowerState > PW_DEVICE_POWER_STATE_D3))
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&CurrentResourceList, 0, sizeof(BM_HANDLE_LIST));
    MEMSET(&TargetResourceList, 0, sizeof(BM_HANDLE_LIST));

    /*
     * Current Power State:
     * --------------------
     * Get the list of power resources associated with the device's current
     * power state. 
     */
    if (CurrentPowerState != PW_POWER_STATE_UNKNOWN)
    {
        ObjectName[3] = '0' + CurrentPowerState;

        if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, ObjectName, &AcpiHandle)))
        {
            UtEvaluateReferenceList(AcpiHandle, &CurrentResourceList);
        }
    }

    /*
     * Target Power State:
     * -------------------
     * Get the list of power resources associated with the device's target
     * power state.
     */
    ObjectName[3] = '0' + TargetPowerState;

    if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, ObjectName, &AcpiHandle)))
    {
        UtEvaluateReferenceList(AcpiHandle, &TargetResourceList);
    }

    /*
     * Transition Power Resources:
     * ---------------------------
     */
    if (CurrentResourceList.Count || TargetResourceList.Count)
    {
        Status = PwResourceListTransition(PowerDevice, &CurrentResourceList, &TargetResourceList);
    }

    /*
     * Execute _PSx:
     * -------------
     * Execute the _PSx method corresponding to the target Dx state, if
     * it exists.
     */
    ObjectName[2] = 'S';
    ObjectName[3] = '0' + TargetPowerState;

    Status = AcpiEvaluateObject(PowerDevice->AcpiHandle, ObjectName, NULL, NULL);
    Status = AE_OK;

    /*
     * Update Power State:
     * -------------------
     * TODO: What to do if any of the above fails???
     */
    if (ACPI_SUCCESS(Status))
    {
        PowerDevice->PowerState = TargetPowerState;
        DEBUG_PRINT(ACPI_INFO, ("Power state was [%d], now [%d]\n", CurrentPowerState, TargetPowerState));
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceCalculateInfo(
    PW_DEVICE_INFO          *PowerDevice)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             AcpiHandle = NULL;
    PW_POWER_STATE          DxSupported = PW_POWER_STATE_UNKNOWN;
    char                    ObjectName[5] = {'_', 'S', '0', 'D', '\0'};
    UINT32                  i = 0;

    FUNCTION_TRACE("PwDeviceCalculateInfo");

    if (!PowerDevice)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /* 
     * Power Management Flags:
     * -----------------------
     */
    if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PSC", &AcpiHandle)))
    {
        PowerDevice->Flags |= PW_FLAGS_PSC;
    }

    if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_IRC", &AcpiHandle)))
    {
        PowerDevice->Flags |= PW_FLAGS_IRC;
    }

    if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PRW", &AcpiHandle)))
    {
        PowerDevice->Flags |= PW_FLAGS_PRW;
    }

    /*
     * Dx States Supported:
     * --------------------
     * Figure out which Dx states are supported by this device for the
     * S0 (working) state.  Note that D0 and D3 are required (assumed).
     */
    PowerDevice->DxSupported[PW_SYSTEM_POWER_STATE_S0] = PW_FLAGS_D0 | PW_FLAGS_D3;

    if ((ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PR1", &AcpiHandle))) ||
        (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PS1", &AcpiHandle))))
    {
        PowerDevice->DxSupported[PW_SYSTEM_POWER_STATE_S0] |= PW_FLAGS_D1;
    }

    if ((ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PR2", &AcpiHandle))) ||
        (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, "_PS2", &AcpiHandle))))
    {
        PowerDevice->DxSupported[PW_SYSTEM_POWER_STATE_S0] |= PW_FLAGS_D2;
    }

    /*
     * Dx States Supported:
     * --------------------
     * Figure out which Dx states are supported by this device for 
     * all other Sx states.  Note that D3 is required (assumed).
     */
    for (i = PW_SYSTEM_POWER_STATE_S1; i <= PW_SYSTEM_POWER_STATE_S5; i++)
    {
        PowerDevice->DxSupported[i] |= PW_FLAGS_D3;

        if (ACPI_SUCCESS(AcpiGetHandle(PowerDevice->AcpiHandle, ObjectName, &AcpiHandle)))
        {
            if (ACPI_SUCCESS(UtEvaluateSimpleInteger(AcpiHandle, &DxSupported)))
            {
                /* 
                 * DxSupported now contains the HIGHEST Dx supported by this 
                 * Sx state -- convert this value to a bit mask of which Dx 
                 * states are supported (conceptually simpler).
                 */ 
                switch (DxSupported)
                {
                case 0:
                    PowerDevice->DxSupported[i] |= PW_FLAGS_D0;
                case 1:
                    PowerDevice->DxSupported[i] |= PW_FLAGS_D1;
                case 2:
                    PowerDevice->DxSupported[i] |= PW_FLAGS_D2;
                case 3:
                    PowerDevice->DxSupported[i] |= PW_FLAGS_D3;
                    break;
                }

                /*
                 * Mask of any states that _SxD falsely advertises (e.g.
                 * claims D1 support but neither _PR2 or _PS2 exist).  In
                 * other words, S1-S5 can't offer a Dx state that isn't
                 * supported by S0.
                 */
                PowerDevice->DxSupported[i] &= PowerDevice->DxSupported[PW_SYSTEM_POWER_STATE_S0];
            }
        }

        ObjectName[2]++;
    }

    /*
     * Current Power State:
     * --------------------
     */
    if (ACPI_FAILURE(Status = PwDeviceGetPowerState(PowerDevice, &(PowerDevice->PowerState))))
    {
        return_ACPI_STATUS(Status);
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceAdd(
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    PW_DEVICE_INFO          *PowerDevice = NULL;
    BM_DEVICE_INFO          *DeviceInfo = NULL;

    FUNCTION_TRACE("PwDeviceAdd");

    DEBUG_PRINT(ACPI_INFO, ("PwDeviceAdd: Power-manageable PowerDevice [0x%08X].\n", DeviceHandle));

    if (!Context || *Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Get details on this device.
     */
    if (ACPI_FAILURE(Status = BmGetDeviceInfo(DeviceHandle, &DeviceInfo)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Allocate a new PW_DEVICE_INFO structure.
     */
    PowerDevice = OsdCallocate(sizeof(PW_DEVICE_INFO));
    if (!PowerDevice)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    PowerDevice->DeviceHandle = DeviceInfo->DeviceHandle;
    PowerDevice->AcpiHandle = DeviceInfo->AcpiHandle;

    /*
     * Get information on this power-manageable device.
     */
    if (ACPI_FAILURE(Status = PwDeviceCalculateInfo(PowerDevice)))
    {
        goto End;
    }
    
    /*
     * Set the initial power resource references, if any.
     */
    Status = PwDeviceSetPowerState(PowerDevice, PW_POWER_STATE_UNKNOWN, PowerDevice->PowerState);

End:
    if (ACPI_FAILURE(Status))
    {
        OsdFree(PowerDevice);
    }
    else
    {
        *Context = PowerDevice;
    }
    
    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
PwDeviceRemove(
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    PW_DEVICE_INFO          *PowerDevice = NULL;

    FUNCTION_TRACE("PwDeviceRemove");

    if (!Context || !*Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Context contains information specific to this power-manageable device.
     */ 
    PowerDevice = (PW_DEVICE_INFO*)*Context;

    DEBUG_PRINT(ACPI_INFO, ("PwDeviceRemove: Power-manageable PowerDevice [0x%08X].\n", PowerDevice->DeviceHandle));

    OsdFree(PowerDevice);

    return_ACPI_STATUS(Status);
}




ACPI_STATUS
PwDeviceInitialize(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("PwDeviceInitialize");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_DEVICE;
    Criteria.Flags = BM_FLAGS_POWER_MANAGEABLE;

    Control.Notify = &PwDeviceNotify;
    Control.Request = &PwDeviceRequest;

    Status = BmRegisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceTerminate(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("PwDeviceTerminate");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_DEVICE;
    Criteria.Flags = BM_FLAGS_POWER_MANAGEABLE;

    Control.Notify = &PwDeviceNotify;
    Control.Request = &PwDeviceRequest;

    Status = BmUnregisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}


ACPI_STATUS 
PwDeviceNotify (
    BM_NOTIFY               NotifyType,
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PwDeviceNotify");

    switch (NotifyType)
    {
    case BM_NOTIFY_DEVICE_ADDED:
        Status = PwDeviceAdd(DeviceHandle, Context);
        break;

    case BM_NOTIFY_DEVICE_REMOVED:
        Status = PwDeviceRemove(Context);
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PwDeviceRequest (
    BM_REQUEST_INFO         *RequestInfo,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    PW_DEVICE_INFO          *PowerDevice;

    FUNCTION_TRACE("PwDeviceRequest");

    /*
     * Must have a valid request structure.
     */
    if (!RequestInfo || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Context contains information specific to this power-manageable device.
     */ 
    PowerDevice = (PW_DEVICE_INFO*)Context;

    /*
     * Handle Request:
     * ---------------
     */
    switch (RequestInfo->Command)
    {
    case PW_COMMAND_GET_INFO:
        Status = PwDeviceGetInfo(PowerDevice, &(RequestInfo->Buffer));
        break;

    case PW_COMMAND_GET_POWER_STATE:
        {
            PW_POWER_STATE PowerState = PW_POWER_STATE_UNKNOWN;

            if (ACPI_SUCCESS(Status = PwDeviceGetPowerState(PowerDevice, &PowerState)))
            {
                Status = UtCopyToBuffer(&(RequestInfo->Buffer), &PowerState, sizeof(PW_POWER_STATE));
            }
        }
        break;

    case PW_COMMAND_SET_POWER_STATE:
        {
            PW_POWER_STATE *PowerState = NULL;

            if (ACPI_SUCCESS(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&PowerState, sizeof(PW_POWER_STATE))))
            {
                Status = PwDeviceSetPowerState(PowerDevice, PowerDevice->PowerState, *PowerState);
            }
        }
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    RequestInfo->Status = Status;

    return_ACPI_STATUS(Status);
}
