/*
 *  ./acpica/control/processor/processor.c - 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __PROCESSOR_C__

#include <acpi.h>
#include <bm.h>
#include <utility.h>

#include "processor.h"

#define _COMPONENT          PROCESSOR_CONTROL
        MODULE_NAME         ("processor");




ACPI_STATUS
PrGetCurrentFrequency(
    PR_INFO                 *Processor,
    UINT32                  *Frequency)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PrGetCurrentFrequency");

    if (!Processor || !Frequency)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    // TODO...

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrCalculatePowerStates(
    PR_INFO                 *Processor)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_BUFFER             Buffer;
    ACPI_CX_STATE           *State = NULL;
    UINT32                  StateCount = 0;
    UINT32                  i = 0;

    FUNCTION_TRACE("PrCalculatePowerStates");

    if (!Processor)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

 /*
  * Set Latency Defaults:
  * ---------------------
  * Default state latency to ACPI_UINT32_MAX -- meaning that this state
  * should not be used by policy.  This value is overriden by states
  * that are present and have usable latencies (e.g. >= 1000us for C3).
  */
 for (i = 0; i < PR_MAX_POWER_STATES; i++)
 {
        Processor->PowerStates.Info[i].Latency = ACPI_UINT32_MAX;
 }

    /*
     * Get Power State Latencies:
     * --------------------------
     */
    Buffer.Length = 0;
    Buffer.Pointer = NULL;

    Status = AcpiGetProcessorCxInfo(Processor->AcpiHandle, &Buffer);
    if (Status == AE_OK)
    {
        DEBUG_PRINT(ACPI_WARN, ("PrCalculatePowerStates: Invalid power data.\n"));
        return_ACPI_STATUS(AE_ERROR);
    }
    if (Status != AE_BUFFER_OVERFLOW)
    {
        return_ACPI_STATUS(Status);
    }

    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    if (ACPI_FAILURE(Status = AcpiGetProcessorCxInfo(Processor->AcpiHandle, &Buffer)))
    {
        goto End;
    }

    State = (ACPI_CX_STATE*)(Buffer.Pointer);
    if (!State)
    {
        /* Sucessful call (above) but no Cx states */
        goto End;
    }

    StateCount = Buffer.Length / sizeof(ACPI_CX_STATE);
    for (i = 0; i < StateCount; i++)
    {
        if ((State[i].StateNumber < PR_MAX_POWER_STATES) && (State[i].Latency <= 1000))
        {
   DEBUG_PRINT(ACPI_INFO, ("C%d Supported (Latency is %d microseconds)\n", i, State[i].Latency));
            Processor->PowerStates.Info[State[i].StateNumber].Latency = State[i].Latency;
        }
    }

    Processor->PowerStates.Count = PR_MAX_POWER_STATES;

 Processor->PowerStates.ActiveState = PR_POWER_STATE_C1;

End:
    OsdFree(Buffer.Pointer);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrCalculatePerformanceStates(
    PR_INFO                 *Processor)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PrCalculatePerformanceStates");

    if (!Processor)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    // TODO...

    return_ACPI_STATUS(Status);
}
    

ACPI_STATUS
PrCalculateThrottlingStates(
    PR_INFO                 *Processor)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_BUFFER             Buffer;
    ACPI_CPU_THROTTLING_STATE *State = NULL;
    UINT32                  StateCount = 0;
    UINT32                  i = 0;

    FUNCTION_TRACE("PrCalculateThrottlingStates");

    if (!Processor)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Get Throttling States:
     * ----------------------
     */
    Buffer.Length = 0;
    Buffer.Pointer = NULL;

    Status = AcpiGetProcessorThrottlingInfo(Processor->AcpiHandle, &Buffer);
    if (Status == AE_OK)
    {
        DEBUG_PRINT(ACPI_WARN, ("PrCalculateThrottlingStates: Invalid throttling data.\n"));
        return_ACPI_STATUS(AE_ERROR);
    }
    if (Status != AE_BUFFER_OVERFLOW)
    {
        return_ACPI_STATUS(Status);
    }

    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    if (ACPI_FAILURE(Status = AcpiGetProcessorThrottlingInfo(Processor->AcpiHandle, &Buffer)))
    {
        goto End;
    }

    State = (ACPI_CPU_THROTTLING_STATE*)(Buffer.Pointer);
    if (!State)
    {
        goto End;
    }

    StateCount = Buffer.Length / sizeof(ACPI_CPU_THROTTLING_STATE);
    for (i = 0; i < StateCount; i++)
    {
        if (State[i].StateNumber < PR_MAX_THROTTLING_STATES)
        {
            // TODO: Verify that state is *really* supported by this chipset/processor (e.g. errata).
            Processor->ThrottlingStates.Percentage[State[i].StateNumber] = State[i].PercentOfClock;
            Processor->ThrottlingStates.Count++;
   DEBUG_PRINT(ACPI_INFO, ("Throttling State [%d:%d%%]\n", i, State[i].PercentOfClock));
        }
    }

End:
    OsdFree(Buffer.Pointer);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrAddDevice(
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    PR_INFO                 *Processor = NULL;
    BM_DEVICE_INFO          *DeviceInfo = NULL;

    FUNCTION_TRACE("PrAddDevice");

    DEBUG_PRINT(ACPI_INFO, ("PrAddDevice: Processor [0x%08X].\n", DeviceHandle));

    if (!Context || *Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Get information on this device.
     */
    if (ACPI_FAILURE(Status = BmGetDeviceInfo(DeviceHandle, &DeviceInfo)))
    {
        return_ACPI_STATUS(Status);
    }

    Processor = OsdCallocate(sizeof(PR_INFO));
    if (!Processor)
    {
        return AE_NO_MEMORY;
    }

    Processor->DeviceHandle = DeviceInfo->DeviceHandle;
    Processor->AcpiHandle = DeviceInfo->AcpiHandle;

    /*
     * Get Information on this processor...
     */
    if (ACPI_FAILURE(Status = PrCalculatePowerStates(Processor)))
    {
        goto End;
    }
    if (ACPI_FAILURE(Status = PrCalculatePerformanceStates(Processor)))
    {
        goto End;
    }
    if (ACPI_FAILURE(Status = PrCalculateThrottlingStates(Processor)))
    {
        goto End;
    }

End:
    if (ACPI_FAILURE(Status))
    {
        OsdFree(Processor);
    }
    else
    {
        *Context = Processor;
  PrPolicyInitialize(Processor);
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrRemoveDevice(
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    PR_INFO                 *Processor= NULL;

    FUNCTION_TRACE("PrRemoveDevice");

    if (!Context || !*Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Context contains information specific to this power-manageable device.
     */ 
    Processor = (PR_INFO*)(*Context);

 PrPolicyTerminate(Processor);

    DEBUG_PRINT(ACPI_INFO, ("PrRemoveDevice: Processor [0x%08X].\n", Processor->DeviceHandle));

    OsdFree(Processor);

    return_ACPI_STATUS(Status);
}




ACPI_STATUS
PrInitialize(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("PrInitialize");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_PROCESSOR;

    Control.Notify = &PrNotify;
    Control.Request = &PrRequest;

    Status = BmRegisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrTerminate(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("PrTerminate");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_PROCESSOR;

    Control.Notify = &PrNotify;
    Control.Request = &PrRequest;

    Status = BmUnregisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS 
PrNotify (
    BM_NOTIFY               NotifyType,
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PrNotify");

    switch (NotifyType)
    {
    case BM_NOTIFY_DEVICE_ADDED:
        Status = PrAddDevice(DeviceHandle, Context);
        break;

    case BM_NOTIFY_DEVICE_REMOVED:
        Status = PrRemoveDevice(Context);
        break;

    case PR_NOTIFY_PERF_STATES_CHANGE:
        // TODO: Rebuild PR_INFO...
        DEBUG_PRINT(ACPI_INFO, ("PrNotify: Performance (_PPC) states change event detected for [0x%08X].\n", DeviceHandle));
        break;

    case PR_NOTIFY_POWER_STATES_CHANGE:
        // TODO: Rebuild PR_INFO...
        DEBUG_PRINT(ACPI_INFO, ("PrNotify: Power (_CST) states change event detected for [0x%08X].\n", DeviceHandle));
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    return_ACPI_STATUS(Status);
}    



ACPI_STATUS
PrRequest (
    BM_REQUEST_INFO         *RequestInfo,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    PR_INFO                 *Processor = NULL;

    FUNCTION_TRACE("PrRequest");

    /*
     * Must have a valid request structure and Context.
     */
    if (!RequestInfo || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Processor = (PR_INFO*)Context;

    /*
     * Handle Request:
     * ---------------
     */
    switch (RequestInfo->Command)
    {
    case PR_COMMAND_GET_INFO:
        Status = UtCopyToBuffer(&(RequestInfo->Buffer), Processor, sizeof(PR_INFO));
        break;

    case PR_COMMAND_SET_CX_STATE_INFO:
  {
   PR_POWER_INFO   *PowerStates = NULL;

   if (ACPI_SUCCESS(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&PowerStates, sizeof(PR_POWER_INFO))))
   {
    MEMCPY(&(Processor->PowerStates), PowerStates, sizeof(PR_POWER_INFO));
   }
  }
        break;

    case PR_COMMAND_GET_THROTTLING_STATE:
        {
            UINT32 ThrottlingState = 0;

            if (ACPI_SUCCESS(Status = AcpiGetProcessorThrottlingState(Processor->AcpiHandle, &ThrottlingState)))
            {
                Status = UtCopyToBuffer(&(RequestInfo->Buffer), &ThrottlingState, sizeof(UINT32));
            }
        }
        break;

    case PR_COMMAND_SET_THROTTLING_STATE:
        {
            UINT32 *ThrottlingState = NULL;

            if (ACPI_SUCCESS(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&ThrottlingState, sizeof(UINT32))))
            {
                Status = AcpiSetProcessorThrottlingState(Processor->AcpiHandle, *ThrottlingState);
            }
        }
        break;

    case PR_COMMAND_GET_PERF_STATE:
        // TODO...
        Status = AE_SUPPORT;
        break;

    case PR_COMMAND_SET_PERF_STATE:
        // TODO...
        Status = AE_SUPPORT;
        break;

    case PR_COMMAND_GET_CURRENT_FREQ:
        {
            UINT32 Frequency = 0;

            if (ACPI_SUCCESS(Status = PrGetCurrentFrequency(Processor, &Frequency)))
            {
                Status = UtCopyToBuffer(&(RequestInfo->Buffer), &Frequency, sizeof(UINT32));
            }
        }
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    RequestInfo->Status = Status;

    return_ACPI_STATUS(Status);
}
