/*
 *  ./acpica/control/processor/prpolicy.c - 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __PRPOLICY_C__

#include <acpi.h>
#include <bm.h>
#include <utility.h>

#include "processor.h"

#define _COMPONENT          PROCESSOR_CONTROL
        MODULE_NAME         ("prpolicy");





void
PrIdleHandler(
 void     *Context)
{
 ACPI_STATUS    Status = AE_OK;
 PR_INFO                 *Processor = NULL;
 PR_CX_STATE_INFO  *CxState = NULL;
 PR_POWER_STATE          ActiveState = PR_POWER_STATE_UNKNOWN;
 PR_POWER_STATE          NextState = PR_POWER_STATE_UNKNOWN;
 UINT32     PmTimerTicks = 0;

 if (!Context)
 {
  DEBUG_PRINT(ACPI_ERROR, ("Invalid (NULL) Context\n", Status));
  return;
 }

 Processor = (PR_INFO*)Context;

 ActiveState = NextState = Processor->PowerStates.ActiveState;

 CxState = &(Processor->PowerStates.Info[ActiveState]);

 CxState->Utilization++;

    /*
     * Invoke Cx State:
     * ----------------
     */
    if (ACPI_FAILURE(Status = AcpiProcessorSleep(Processor->AcpiHandle, &PmTimerTicks)))
    {
  DEBUG_PRINT(ACPI_ERROR, ("AcpiProcessorSleep() returned error [0x%08X]\n", Status));
  PmTimerTicks = ACPI_UINT32_MAX;
    }

    /*
     * Check For State Promotion:
     * --------------------------
     * Only need to check for promotion on C1 and C2, and then only 
  * when the state has a non-zero count threshold and target state.
     */
 if (CxState->PromotionPolicy.CountThreshold && CxState->PromotionPolicy.TargetState && 
  ((ActiveState == PR_POWER_STATE_C1) || (ActiveState == PR_POWER_STATE_C2)))
 {
        /*
   * Check the amount of time we spent in the Cx state against our
   * promotion policy.  If successful (asleep longer than our threshold)
         * increment our count and see if a promotion is in order.
         */
  if (PmTimerTicks > (CxState->PromotionPolicy.TimeThreshold))
  {
   CxState->PromotionPolicy.Count++;
   CxState->DemotionPolicy.Count = 0;

   if (CxState->PromotionPolicy.Count >= CxState->PromotionPolicy.CountThreshold)
   {
    NextState = CxState->PromotionPolicy.TargetState;
   }
  }
    }

    /*
     * Check For State Demotion:
     * -------------------------
     * Only need to check for demotion on C2 and C3, and then only 
  * when the state has a non-zero count threshold and target state.
     */
 if (CxState->DemotionPolicy.CountThreshold && CxState->DemotionPolicy.TargetState && 
  ((ActiveState == PR_POWER_STATE_C2) || (ActiveState == PR_POWER_STATE_C3)))
 {
        /*
   * Check the amount of time we spent in the Cx state against our
   * demotion policy.  If unsuccessful (asleep shorter than our threshold)
         * increment our count and see if a demotion is in order.
         */
  if (PmTimerTicks < (CxState->DemotionPolicy.TimeThreshold))
        {
   CxState->DemotionPolicy.Count++;
   CxState->PromotionPolicy.Count = 0;

   if (CxState->DemotionPolicy.Count >= CxState->DemotionPolicy.CountThreshold)
   {
    NextState = CxState->DemotionPolicy.TargetState;
   }
        }
 }

    /*
     * New Cx State?
     * -------------
     * If so, clean up from the previous Cx state (if necessary).
     */
    if (NextState != Processor->PowerStates.ActiveState)
    {
  if (ACPI_FAILURE(Status = AcpiSetProcessorSleepState(Processor->AcpiHandle, NextState)))
  {
   DEBUG_PRINT(ACPI_ERROR, ("AcpiSetProcessorSleepState() returned error [0x%08X]\n", Status));
  }
  else
  {
   CxState->PromotionPolicy.Count = 0;
   CxState->DemotionPolicy.Count = 0;
   Processor->PowerStates.ActiveState = NextState;
  }
    }
}



ACPI_STATUS
PrPolicyInitialize(
 PR_INFO                 *Processor)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PrPolicyInitialize");

 if (!Processor)
 {
  return_ACPI_STATUS(AE_BAD_PARAMETER);
 }

 if (ACPI_FAILURE(Status = AcpiSetProcessorSleepState(Processor->AcpiHandle, Processor->PowerStates.ActiveState)))
 {
  return_ACPI_STATUS(Status);
 }

 Status = OsdInstallIdleHandler(&PrIdleHandler, Processor);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
PrPolicyTerminate (
 PR_INFO                 *Processor)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("PrPolicyTerminate");

 // TODO: What about multiple processors?

 if (!Processor)
 {
  return_ACPI_STATUS(AE_BAD_PARAMETER);
 }

    Status = OsdRemoveIdleHandler(&PrIdleHandler);

    return_ACPI_STATUS(Status);
}


