/*
 *  ./acpica/control/thermal/thermal.c - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __THERMAL_C__

#include <acpi.h>

#include <utility.h>

#include "thermal.h"


#define _COMPONENT          THERMAL_CONTROL
        MODULE_NAME         ("thermal");




ACPI_STATUS
TzPrint (
    TZ_INFO                 *ThermalZone)
{
    ACPI_BUFFER             Buffer;

    FUNCTION_TRACE("TzPrint");

    if (!ThermalZone)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Buffer.Length = 256;
    Buffer.Pointer = OsdCallocate(Buffer.Length);
    if (!Buffer.Pointer)
    {
        return_ACPI_STATUS(AE_NO_MEMORY);
    }

    /*
     * Get the full pathname for this ACPI object.
     */
    AcpiGetName(ThermalZone->AcpiHandle, ACPI_FULL_PATHNAME, &Buffer);

    /*
     * Print out basic thermal zone information.
     */
    OsdPrintf("-------------------------------------------------------------\n");
    OsdPrintf("\tThermal Zone [%s]\n", (char*)Buffer.Pointer);
    OsdPrintf("\tAcpiHandle[0x%08x] TemperatureHandle[0x%08x]\n", ThermalZone->AcpiHandle, ThermalZone->TemperatureHandle);
    OsdPrintf("\tRPF[%d] TC1[%d] TC2[%d] TSP[%d]\n", ThermalZone->RecommendedPollingFrequency, ThermalZone->ThermalConstant1, ThermalZone->ThermalConstant2, ThermalZone->ThermalSamplingPeriod);
    OsdPrintf("-------------------------------------------------------------\n");

    OsdFree(Buffer.Pointer);

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
TzGetTemperature (
    TZ_INFO                 *ThermalZone,
    UINT32                  *Temperature)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("TzGetTemperature");

    if (!ThermalZone || !Temperature)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Evaluate the _TMP control method to get the current temperature.
     */
    Status = UtEvaluateSimpleInteger(ThermalZone->TemperatureHandle, Temperature);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzSetCoolingPreference (
    TZ_INFO                 *ThermalZone,
    UINT32                  CoolingPreference)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_LIST        ArgList;
    ACPI_OBJECT             Arg0;

    FUNCTION_TRACE("TzSetCoolingPreference");

    if (!ThermalZone || ((CoolingPreference != TZ_PREFERENCE_ACTIVE) && (CoolingPreference != TZ_PREFERENCE_PASSIVE)))
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Build the argument list, which simply consists of the current 
     * cooling preference.
     */
    MEMSET(&ArgList, 0, sizeof(ACPI_OBJECT));
    ArgList.Count = 1;
    ArgList.Pointer = &Arg0;

    MEMSET(&Arg0, 0, sizeof(ACPI_OBJECT));
    Arg0.Type = ACPI_TYPE_Number;
    Arg0.Number.Value = CoolingPreference;

    /*
     * Evaluate "_SCP" - setting the new cooling preference.
     */
    Status = AcpiEvaluateObject(ThermalZone->AcpiHandle, "_SCP", &ArgList, NULL);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzGetSingleThreshold(
    TZ_INFO                 *ThermalZone,
    TZ_THRESHOLD            *Threshold)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             ThresholdHandle = NULL;
    BM_HANDLE_LIST          HandleList;
    UINT32                  i = 0;

    FUNCTION_TRACE("TzGetSingleThreshold");

    if (!ThermalZone || !Threshold)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    Threshold->Temperature = 0;
    Threshold->State = TZ_STATE_UNKNOWN;
    Threshold->CoolingDeviceCount = 0;
    Threshold->CoolingDevices[0] = BM_HANDLE_UNKNOWN;

    switch (Threshold->Type)
    {
    case TZ_THRESHOLD_CRITICIAL:

        Threshold->Index = 0;

        if (ACPI_SUCCESS(Status = AcpiGetHandle(ThermalZone->AcpiHandle, "_CRT", &ThresholdHandle)))
        {
            if (ACPI_SUCCESS(Status = UtEvaluateSimpleInteger(ThresholdHandle, &(Threshold->Temperature))))
            {
                DEBUG_PRINT(ACPI_INFO, ("TzGetSingleThreshold: Found critical threshold @ [%d]\n", Threshold->Temperature));
            }
        }

        break;

    case TZ_THRESHOLD_PASSIVE:

        Threshold->Index = 0;

        if (ACPI_SUCCESS(Status = AcpiGetHandle(ThermalZone->AcpiHandle, "_PSV", &ThresholdHandle)))
        {
            if (ACPI_SUCCESS(Status = UtEvaluateSimpleInteger(ThresholdHandle, &(Threshold->Temperature))))
            {
                /*
                 * Passive Cooling Devices:
                 * ------------------------
                 * Note that we don't return status when evaluating _PSL
                 * to allow a 'success' when _PSV exists but no _PSL.  
                 * In other words, we'll return a threshold but without
                 * any 
                 */
                if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, "_PSL", &ThresholdHandle)))
                {
                    MEMSET(&HandleList, 0, sizeof(BM_HANDLE_LIST));

                    if (ACPI_SUCCESS(UtEvaluateReferenceList(ThresholdHandle, &HandleList)))
                    {
                        Threshold->CoolingDeviceCount = (HandleList.Count > TZ_MAX_COOLING_DEVICES) ? TZ_MAX_COOLING_DEVICES : HandleList.Count;

                        for (i = 0; i < Threshold->CoolingDeviceCount; i++)
                        {
                            Threshold->CoolingDevices[i] = HandleList.Handles[i];
                        }
                    }
                }

                DEBUG_PRINT(ACPI_INFO, ("TzGetSingleThreshold: Found passive threshold @ [%d]\n", Threshold->Temperature));
            }
        }

        break;

    case TZ_THRESHOLD_ACTIVE:
        {
            char ObjectName[5] = {'_', 'A', 'C', ('0' + Threshold->Index), '\0'};

            if (ACPI_SUCCESS(Status = AcpiGetHandle(ThermalZone->AcpiHandle, ObjectName, &ThresholdHandle)))
            {
                if (ACPI_SUCCESS(Status = UtEvaluateSimpleInteger(ThresholdHandle, &(Threshold->Temperature))))
                {
                    ObjectName[2] = 'L';

                    /*
                     * Active Cooling Devices:
                     * -----------------------
                     */
                    if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, ObjectName, &ThresholdHandle)))
                    {
                        MEMSET(&HandleList, 0, sizeof(BM_HANDLE_LIST));

                        if (ACPI_SUCCESS(UtEvaluateReferenceList(ThresholdHandle, &HandleList)))
                        {
                            Threshold->CoolingDeviceCount = (HandleList.Count > TZ_MAX_COOLING_DEVICES) ? TZ_MAX_COOLING_DEVICES : HandleList.Count;

                            for (i = 0; i < Threshold->CoolingDeviceCount; i++)
                            {
                                Threshold->CoolingDevices[i] = HandleList.Handles[i];
                            }
                        }
                    }

                    DEBUG_PRINT(ACPI_INFO, ("TzGetSingleThreshold: Found active threshold [%d] @ [%d]\n", Threshold->Index, Threshold->Temperature));
                }
            }
        }

        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    return_ACPI_STATUS(Status);
}


ACPI_STATUS
TzGetAllThresholds(
    TZ_INFO                 *ThermalZone,
    ACPI_BUFFER             *Buffer)
{
    ACPI_STATUS             Status = AE_OK;
    TZ_THRESHOLD_LIST       ThresholdList;
    TZ_THRESHOLD            *Threshold = NULL;
    UINT32                  i = 0;

    FUNCTION_TRACE("TzGetAllThresholds");

    if (!ThermalZone || !Buffer)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    MEMSET(&ThresholdList, 0, sizeof(TZ_THRESHOLD_LIST));

    /*
     * Critical Threshold:
     * -------------------
     */
    Threshold = &(ThresholdList.Thresholds[ThresholdList.Count]);
    Threshold->Type = TZ_THRESHOLD_CRITICIAL;

    if (ACPI_SUCCESS(TzGetSingleThreshold(ThermalZone, Threshold)))
    {
        ThresholdList.Count++;
    }

    /*
     * Passive Threshold:
     * ------------------
     */
    Threshold = &(ThresholdList.Thresholds[ThresholdList.Count]);
    Threshold->Type = TZ_THRESHOLD_PASSIVE;

    if (ACPI_SUCCESS(TzGetSingleThreshold(ThermalZone, Threshold)))
    {
        ThresholdList.Count++;
    }

    /*
     * Active Threshold:
     * -----------------
     * Note that active thresholds are sorted by index (e.g. _AC0,
     * _AC1, ...), and thus from highest (_AC0) to lowest (_AC9) 
     * temperature.
     */
    for (i = 0; i < TZ_MAX_ACTIVE_THRESHOLDS; i++)
    {
        Threshold = &(ThresholdList.Thresholds[ThresholdList.Count]);
        Threshold->Type = TZ_THRESHOLD_ACTIVE;
        Threshold->Index = i;

        if (ACPI_SUCCESS(TzGetSingleThreshold(ThermalZone, Threshold)))
        {
            ThresholdList.Count++;
        }
        else
        {
            Threshold->Type = TZ_THRESHOLD_UNKNOWN;
            Threshold->Index = 0;
            break;
        }
    }

    Status = UtCopyToBuffer(Buffer, &ThresholdList, sizeof(TZ_THRESHOLD_LIST));

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzCalculateInfo(
    TZ_INFO                 *ThermalZone)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             AcpiHandle = NULL;

    FUNCTION_TRACE("TzCalculateInfo");

    if (!ThermalZone)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * _TMP Handle:
     * ------------
     * Get the ACPI handle for this zone's _TMP (current temperature) object.
     */
    if (ACPI_FAILURE(Status = AcpiGetHandle(ThermalZone->AcpiHandle, "_TMP", &(ThermalZone->TemperatureHandle))))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Recommended Polling Frequency:
     * ------------------------------
     * Evaluate _TZP, if exists.
     */
    if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, "_TZP", &AcpiHandle)))
    {
        UtEvaluateSimpleInteger(AcpiHandle, &(ThermalZone->RecommendedPollingFrequency));
    }
    else
    {
        ThermalZone->RecommendedPollingFrequency = TZ_DEFAULT_POLLING_FREQUENCY;
    }

    /*
     * Passive Cooling Thermal Constants:
     * ----------------------------------
     * Evaluate _TC1 and _TC2, if exist.
     */
    if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, "_TC1", &AcpiHandle)))
    {
        UtEvaluateSimpleInteger(AcpiHandle, &(ThermalZone->ThermalConstant1));
    }
    if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, "_TC2", &AcpiHandle)))
    {
        UtEvaluateSimpleInteger(AcpiHandle, &(ThermalZone->ThermalConstant2));
    }

    /*
     * Passive Cooling Sampling Period:
     * --------------------------------
     * Evaluate _TSP, if exists.
     */
    if (ACPI_SUCCESS(AcpiGetHandle(ThermalZone->AcpiHandle, "_TSP", &AcpiHandle)))
    {
        UtEvaluateSimpleInteger(AcpiHandle, &(ThermalZone->ThermalSamplingPeriod));
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzAddDevice(
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    TZ_INFO                 *ThermalZone = NULL;
    BM_DEVICE_INFO          *DeviceInfo = NULL;

    FUNCTION_TRACE("TzAddDevice");

    DEBUG_PRINT(ACPI_INFO, ("TzAddDevice: Thermal Zone [0x%08X].\n", DeviceHandle));

    if (!Context || *Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Get information on this device.
     */
    if (ACPI_FAILURE(Status = BmGetDeviceInfo(DeviceHandle, &DeviceInfo)))
    {
        return_ACPI_STATUS(Status);
    }

    /*
     * Allocate a new Thermal Zone element and add to our collection.
     */
    ThermalZone = OsdCallocate(sizeof(TZ_INFO));
    if (!ThermalZone)
    {
        return AE_NO_MEMORY;
    }

    ThermalZone->DeviceHandle = DeviceInfo->DeviceHandle;
    ThermalZone->AcpiHandle = DeviceInfo->AcpiHandle;

    /*
     * Get Information on this thermal zone...
     */
    Status = TzCalculateInfo(ThermalZone);

    if (ACPI_FAILURE(Status))
    {
        OsdFree(ThermalZone);
    }
    else
    {
        *Context = ThermalZone;
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzRemoveDevice(
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;
    TZ_INFO                 *ThermalZone = NULL;

    FUNCTION_TRACE("TzRemoveDevice");

    if (!Context || !*Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    /*
     * Context contains information specific to this power-manageable device.
     */ 
    ThermalZone = (TZ_INFO*)(*Context);

    DEBUG_PRINT(ACPI_INFO, ("TzRemoveDevice: Thermal Zone [0x%08X].\n", ThermalZone->DeviceHandle));

    OsdFree(ThermalZone);

    return_ACPI_STATUS(Status);
}




ACPI_STATUS
TzInitialize(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("TzInitialize");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_THERMAL_ZONE;

    Control.Notify = &TzNotify;
    Control.Request = &TzRequest;

    Status = BmRegisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzTerminate(void)
{
    ACPI_STATUS             Status = AE_OK;
    BM_DEVICE_INFO          Criteria;
    BM_CONTROL_INFO         Control;

    FUNCTION_TRACE("TzTerminate");

    MEMSET(&Criteria, 0, sizeof(BM_DEVICE_INFO));
    MEMSET(&Control, 0, sizeof(BM_CONTROL_INFO));

    Criteria.Class = BM_CLASS_THERMAL_ZONE;

    Control.Notify = &TzNotify;
    Control.Request = &TzRequest;

    Status = BmUnregisterControl(&Criteria, &Control);

    return_ACPI_STATUS(Status);
}


ACPI_STATUS 
TzNotify (
    BM_NOTIFY               NotifyType,
    BM_HANDLE               DeviceHandle,
    void                    **Context)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE("TzNotify");

    if (!Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    switch (NotifyType)
    {
    case BM_NOTIFY_DEVICE_ADDED:
        Status = TzAddDevice(DeviceHandle, Context);
        break;

    case BM_NOTIFY_DEVICE_REMOVED:
        Status = TzRemoveDevice(Context);
        break;

    case TZ_NOTIFY_TEMPERATURE_CHANGE:
        // TODO...
        DEBUG_PRINT(ACPI_INFO, ("TzNotify: Temperature change event detected.\n"));
        break;

    case TZ_NOTIFY_THRESHOLD_CHANGE:
        // TODO...
        DEBUG_PRINT(ACPI_INFO, ("TzNotify: Threshold (_SCP) change event detected.\n"));
        break;

    case TZ_NOTIFY_DEVICE_LISTS_CHANGE:
        DEBUG_PRINT(ACPI_INFO, ("TzNotify: Device lists (_ALx, _PSL, _TZD) change event detected.\n"));
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    return_ACPI_STATUS(Status);
}



ACPI_STATUS
TzRequest (
    BM_REQUEST_INFO         *RequestInfo,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    TZ_INFO                 *ThermalZone = NULL;

    FUNCTION_TRACE("TzRequest");

    /*
     * Must have a valid request structure and Context.
     */
    if (!RequestInfo || !Context)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    ThermalZone = (TZ_INFO*)Context;

    /*
     * Handle Request:
     * ---------------
     */
    switch (RequestInfo->Command)
    {
    case TZ_COMMAND_GET_INFO:
        Status = UtCopyToBuffer(&(RequestInfo->Buffer), ThermalZone, sizeof(TZ_INFO));
        break;

    case TZ_COMMAND_GET_THRESHOLDS:
        Status = TzGetAllThresholds(ThermalZone, &(RequestInfo->Buffer));
        break;

    case TZ_COMMAND_GET_TEMPERATURE:
        {
            UINT32 Temperature = 0;

            if (ACPI_SUCCESS(Status = TzGetTemperature(ThermalZone, &Temperature)))
            {
                Status = UtCopyToBuffer(&(RequestInfo->Buffer), &Temperature, sizeof(UINT32));
            }
        }
        break;

    case TZ_COMMAND_SET_COOLING_PREFERENCE:
        {
            UINT32 *CoolingPreference = NULL;

            if (ACPI_SUCCESS(Status = UtCastBuffer(&(RequestInfo->Buffer), (void**)&CoolingPreference, sizeof(UINT32))))
            {
                Status = TzSetCoolingPreference(ThermalZone, *CoolingPreference);
            }
        }
        break;

    default:
        Status = AE_SUPPORT;
        break;
    }

    RequestInfo->Status = Status;

    return_ACPI_STATUS(Status);
}
