/*
 *  ./acpica/control/thermal/thermal.h - <TBD>
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef __THERMAL_H__
#define __THERMAL_H__

#include <busmgr.h>
#include <power.h>



#define TZ_DEFAULT_POLLING_FREQUENCY    60      /* One minute */
#define TZ_MAX_ACTIVE_THRESHOLDS        10      /* _AC0 through _AC9 */
#define TZ_MAX_THRESHOLDS               12      /* <above> + _CRT + _PSV */
#define TZ_MAX_COOLING_DEVICES          10
#define TZ_MAX_THERMAL_ZONES            256

/* 
 * Thermal Zone Commands:
 * ----------------------
 */
#define TZ_COMMAND_GET_INFO             ((BM_COMMAND) 0x80)
#define TZ_COMMAND_GET_THRESHOLDS       ((BM_COMMAND) 0x81)
#define TZ_COMMAND_GET_TEMPERATURE      ((BM_COMMAND) 0x82)
#define TZ_COMMAND_SET_COOLING_PREFERENCE ((BM_COMMAND) 0x83)

/* 
 * Thermal Zone Notifications:
 * ---------------------------
 */
#define TZ_NOTIFY_TEMPERATURE_CHANGE    ((BM_NOTIFY) 0x80)
#define TZ_NOTIFY_THRESHOLD_CHANGE      ((BM_NOTIFY) 0x81)
#define TZ_NOTIFY_DEVICE_LISTS_CHANGE   ((BM_NOTIFY) 0x82)

/*
 * TZ_THRESHOLD_TYPE: 
 * ------------------
 * Indicates the type of threshold (trip point).
 */
typedef UINT32                          TZ_THRESHOLD_TYPE;

#define TZ_THRESHOLD_UNKNOWN            ((TZ_THRESHOLD_TYPE) 0x00)
#define TZ_THRESHOLD_CRITICIAL          ((TZ_THRESHOLD_TYPE) 0x01)
#define TZ_THRESHOLD_PASSIVE            ((TZ_THRESHOLD_TYPE) 0x02)
#define TZ_THRESHOLD_ACTIVE             ((TZ_THRESHOLD_TYPE) 0x03)

/*
 * TZ_THRESHOLD_STATE: 
 * ------------------
 * Indicates the <collective> state for the cooling devices of a threshold.
 */
typedef UINT32                          TZ_THRESHOLD_STATE;

#define TZ_STATE_UNKNOWN                ((TZ_THRESHOLD_TYPE) 0x00)
#define TZ_STATE_ENABLED                ((TZ_THRESHOLD_TYPE) 0x01)
#define TZ_STATE_DISABLED               ((TZ_THRESHOLD_TYPE) 0x02)

/*
 * TZ_COOLING_PREFERENCE: 
 * ----------------------
 * Values for the current user cooling mode preference (re: _SCP).
 */
typedef UINT32                          TZ_COOLING_PREFERENCE;

#define TZ_PREFERENCE_UNKNOWN           ((TZ_COOLING_PREFERENCE) 0xFFFFFFFF)

#define TZ_PREFERENCE_ACTIVE            ((TZ_COOLING_PREFERENCE) 0x00)
#define TZ_PREFERENCE_PASSIVE           ((TZ_COOLING_PREFERENCE) 0x01)

/*
 * TZ_THRESHOLD:
 * -------------
 * Information on an individual threshold.
 */
typedef struct
{
    TZ_THRESHOLD_TYPE   Type;
    UINT32          Index;
    UINT32          Temperature;
    TZ_THRESHOLD_STATE  State;
    UINT32          CoolingDeviceCount;
    BM_HANDLE           CoolingDevices[TZ_MAX_COOLING_DEVICES];
} TZ_THRESHOLD;

/*
 * TZ_THRESHOLD_LIST:
 * ------------------
 * Container for the thresholds of a given thermal zone.
 * Note that thresholds are always ordered by increasing
 * temperature value to simplify use by thermal policy.
 */
typedef struct
{
    UINT32     Count;
    TZ_THRESHOLD        Thresholds[TZ_MAX_THRESHOLDS];
} TZ_THRESHOLD_LIST;

/*
 * TZ_INFO:
 * --------
 * General thermal zone information.
 */
typedef struct
{
    BM_HANDLE           DeviceHandle;
    ACPI_HANDLE         AcpiHandle;
    ACPI_HANDLE         TemperatureHandle;
    UINT32    RecommendedPollingFrequency;
    UINT32    ThermalConstant1;
    UINT32    ThermalConstant2;
    UINT32    ThermalSamplingPeriod;
} TZ_INFO;
 


ACPI_STATUS
TzInitialize(void);

ACPI_STATUS
TzTerminate(void);

ACPI_STATUS 
TzNotify (
    BM_NOTIFY               NotifyType,
    BM_HANDLE               DeviceHandle,
    void                    **Context);

ACPI_STATUS
TzRequest(
    BM_REQUEST_INFO         *RequestInfo,
    void                    *Context);


#endif  /* __THERMAL_H__ */
