/*
 *  ./acpica/subsystem/common/cmalloc.c - local memory allocation routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMALLOC_C__

#include <acpi.h>
#include <parser.h>
#include <interp.h>
#include <namesp.h>
#include <globals.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmalloc");


/*
 * Most of this code is for tracking memory leaks in the subsystem, and it
 * gets compiled out when the ACPI_DEBUG flag is not set.  It works like so:
 * every memory allocation is kept track of in a doubly linked list.  Each
 * element contains the caller's component, module name, function name, and
 * line number.  _CmAllocate and _CmCallocate call CmAddElementToAllocList
 * to add an element to the list; deletion occurs in the bosy of _CmFree.
 */

#ifdef ACPI_DEBUG


ALLOCATION_INFO *
CmSearchAllocList (
    void                    *Address)
{
    ALLOCATION_INFO         *Element = Gbl_HeadAllocPtr;


    /* Search for the address. note - this always searches the entire list...*/

    while (1)
    {
        if (Element == NULL)
        {
            return NULL;
        }
        
        if (Element->Address == Address)
        {
            return Element;
        }

        Element = Element->Next; 
    }
}



void
CmAddElementToAllocList (
    void                    *Address,
    UINT32                  Size,
    UINT8                   AllocType,
    UINT32                  Component,
    ACPI_STRING             Module,
    INT32                   Line)
{
    ALLOCATION_INFO         *Element;


    FUNCTION_TRACE_PTR ("CmAddElementToAllocList", Address);
    

    CmAcquireMutex (MTX_MEMORY);
    
    /* Keep track of the running total of all allocations. */
    Gbl_CurrentAllocCount++;
    Gbl_RunningAllocCount++;
    if (Gbl_MaxConcurrentAllocCount < Gbl_CurrentAllocCount)
    {
     Gbl_MaxConcurrentAllocCount = Gbl_CurrentAllocCount;
    }
    
    Gbl_CurrentAllocSize += Size;
    Gbl_RunningAllocSize += Size;
    if (Gbl_MaxConcurrentAllocSize < Gbl_CurrentAllocSize)
    {
     Gbl_MaxConcurrentAllocSize = Gbl_CurrentAllocSize;
    }
    
    /* If the head pointer is null, create the first element and fill it in. */

    if (NULL == Gbl_HeadAllocPtr)
    {
        Gbl_HeadAllocPtr = (ALLOCATION_INFO *) OsdCallocate (sizeof (ALLOCATION_INFO));
        
        /* error check */
        
        Gbl_TailAllocPtr = Gbl_HeadAllocPtr;
    }

    else
    {
        Gbl_TailAllocPtr->Next = (ALLOCATION_INFO *) OsdCallocate (sizeof (ALLOCATION_INFO));
        
        /* error check */
        
        Gbl_TailAllocPtr->Next->Previous = Gbl_TailAllocPtr;
        Gbl_TailAllocPtr = Gbl_TailAllocPtr->Next;
    }

    /* 
     * Search list for this address to make sure it is not already on the list.
     * This will catch several kinds of problems.
     */

    Element = CmSearchAllocList (Address);
    if (Element)
    {
        REPORT_ERROR ("CmAddElementToAllocList: Address already present in list!");
        DEBUG_PRINT (ACPI_ERROR, ("Element %p Address %p\n", Element, Address));

        BREAKPOINT3;
    }

    /* Fill in the instance data. */    
    
    Gbl_TailAllocPtr->Address   = Address;
    Gbl_TailAllocPtr->Size      = Size;
    Gbl_TailAllocPtr->AllocType = AllocType;
    Gbl_TailAllocPtr->Component = Component;
    Gbl_TailAllocPtr->Line      = Line;

    STRNCPY (Gbl_TailAllocPtr->Module, Module, sizeof (Gbl_TailAllocPtr->Module));
    
    CmReleaseMutex (MTX_MEMORY);
    return_VOID;
}



void
CmDeleteElementFromAllocList (
    void                    *Address,
    UINT32                  Component,
    ACPI_STRING             Module,
    INT32                   Line)
{
    ALLOCATION_INFO         *Element;
    UINT32     Size;
    UINT32                  DwordLen;
    UINT32                  i;
    

    FUNCTION_TRACE ("CmDeleteElementFromAllocList");

    /* cases: none, one, multiple. */

    if (NULL == Gbl_HeadAllocPtr)
    {
        /* Boy we got problems. */

        _REPORT_ERROR (Module, Line, Component, "CmDeleteElementFromAllocList: Empty allocation list, nothing to free!");
        
        return_VOID;
    }

    
    CmAcquireMutex (MTX_MEMORY);
    
    /* Keep track of the amount of memory allocated. */
    Size = 0;
    Gbl_CurrentAllocCount--;

    if (Gbl_HeadAllocPtr == Gbl_TailAllocPtr)
    {   
        if (Address != Gbl_HeadAllocPtr->Address)
        {
            _REPORT_ERROR (Module, Line, Component, "CmDeleteElementFromAllocList: Deleting non-allocated memory...");

            goto Cleanup;
        }
        
        Size = Gbl_HeadAllocPtr->Size;
        
        OsdFree (Gbl_HeadAllocPtr);
        Gbl_HeadAllocPtr = NULL;
        Gbl_TailAllocPtr = NULL;
        
        DEBUG_PRINT (TRACE_ALLOCATIONS, ("_CmFree: Allocation list deleted.  There are no outstanding allocations.\n"));
    
        goto Cleanup;
    }
    

    /* Search list for this address */

    Element = CmSearchAllocList (Address);
    if (Element)
    {
        /* cases: head, tail, other */

        if (Element == Gbl_HeadAllocPtr)
        {
            Element->Next->Previous = NULL;
            Gbl_HeadAllocPtr = Element->Next;
        }

        else
        {
            if (Element == Gbl_TailAllocPtr)
            {
                Element->Previous->Next = NULL;
                Gbl_TailAllocPtr = Element->Previous;
            }

            else
            {
                Element->Previous->Next = Element->Next;
                Element->Next->Previous = Element->Previous;
            }
        }       
        

        /* Mark the segment as deleted */

        if (Element->Size >= 4)
        {
            DwordLen = Element->Size / 4;

            for (i = 0; i < DwordLen; i++)
            {
                ((UINT32 *) Element->Address)[i] = (UINT32) 0x00DEAD00;
            }

            /* Set obj type, desc, and ref count fields to all ones */

            ((UINT32 *) Element->Address)[0] = (UINT32) 0xFFFFFFFF;
            if (Element->Size >= 8)
            {
                ((UINT32 *) Element->Address)[1] = (UINT32) 0xFFFFFFFF;
            }
        }

        Size = Element->Size;
        OsdFree (Element);
    }
            
    else
    {
        _REPORT_ERROR (Module, Line, Component, "_CmFree: Entry not found in list");
        DEBUG_PRINT (TRACE_ALLOCATIONS, ("_CmFree: Entry %p was not found in allocation list\n", Address));
    }


Cleanup:
    
    Gbl_CurrentAllocSize -= Size;
    CmReleaseMutex (MTX_MEMORY);

    return_VOID;
}



void
CmDumpAllocationInfo (
 void)
{
    FUNCTION_TRACE ("CmDumpAllocationInfo");
    

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
                 ("Current outstanding allocations: %d (%d b, %d Kb)\n",
                 Gbl_CurrentAllocCount, Gbl_CurrentAllocSize, Gbl_CurrentAllocSize / 1024));

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
              ("Maximum concurrent allocations thus far: %d (%d b, %d Kb)\n",
              Gbl_MaxConcurrentAllocCount, Gbl_MaxConcurrentAllocSize, Gbl_MaxConcurrentAllocSize / 1024));

 DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
                 ("Current number of allocated internal objects: %d (%d b, %d Kb)\n",
                 Gbl_CurrentObjectCount, Gbl_CurrentObjectSize, Gbl_CurrentObjectSize / 1024));

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
                 ("Maximum concurrent number of allocated internal objects: %d (%d b, %d Kb)\n",
                 Gbl_MaxConcurrentObjectCount, Gbl_MaxConcurrentObjectSize, Gbl_MaxConcurrentObjectSize / 1024));

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
                 ("Total number of allocated internal objects: %d (%d b, %d Kb)\n",
                 Gbl_RunningObjectCount, Gbl_RunningObjectSize, Gbl_RunningObjectSize / 1024));

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES,
                 ("Total number of allocations: %d (%d b, %d Kb)\n",
                 Gbl_RunningAllocCount, Gbl_RunningAllocSize, Gbl_RunningAllocSize / 1024));

 return_VOID;
}



void
CmDumpCurrentAllocations (
    UINT32                  Component,
    ACPI_STRING             Module)
{
    ALLOCATION_INFO         *Element = Gbl_HeadAllocPtr;
    UINT32                  i;
    

    FUNCTION_TRACE ("CmDumpCurrentAllocations");
    

    if (Element == NULL)
    {
        DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES, ("No outstanding allocations.\n"));
        return_VOID;
    }


    /*
     * Walk the allocation list.
     */

    CmAcquireMutex (MTX_MEMORY);

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES, ("Outstanding allocations:\n"));

    for (i = 1; ; i++)  /* Just a counter */
    {
        if ((Element->Component & Component) &&
            ((Module == NULL) || (0 == STRCMP (Module, Element->Module))))
        {
            DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES, ("%p Len %04X %9.9s-%d",
                            Element->Address, Element->Size, Element->Module, Element->Line));

            /* Most of the elements will be internal objects. */

            switch (((ACPI_OBJECT_INTERNAL *)(Element->Address))->Common.DataType)
            {
            case DESC_TYPE_ACPI_OBJ:
                DEBUG_PRINT_RAW (TRACE_ALLOCATIONS | TRACE_TABLES, (" ObjType %s", 
                                    CmGetTypeName (((ACPI_OBJECT_INTERNAL *)(Element->Address))->Common.Type)));
                break;

            case DESC_TYPE_PARSER:
                DEBUG_PRINT_RAW (TRACE_ALLOCATIONS | TRACE_TABLES, (" ParseObj Opcode %04X", 
                                    ((ACPI_GENERIC_OP *)(Element->Address))->Opcode));
                break;

            case DESC_TYPE_NTE:
                DEBUG_PRINT_RAW (TRACE_ALLOCATIONS | TRACE_TABLES, (" NTE Name %4.4s", 
                                    &((NAME_TABLE_ENTRY *)(Element->Address))->Name));
                break;

            case DESC_TYPE_STATE:
                DEBUG_PRINT_RAW (TRACE_ALLOCATIONS | TRACE_TABLES, (" StateObj"));
                break;
            }

            DEBUG_PRINT_RAW (TRACE_ALLOCATIONS | TRACE_TABLES, ("\n")); 
        }
        
        if (Element->Next == NULL)
        {
            break;
        }
        
        Element = Element->Next;
    }

    CmReleaseMutex (MTX_MEMORY);

    DEBUG_PRINT (TRACE_ALLOCATIONS | TRACE_TABLES, ("Total number of unfreed allocations = %d\n", i));
   
    return_VOID;
}   

#endif  /* Debug routines for memory leak detection */



void *
_CmAllocate (
    UINT32                  Size,
    UINT32                  Component,
    ACPI_STRING             Module,
    INT32                   Line)
{
    void                    *Address = NULL;


    FUNCTION_TRACE_U32 ("_CmAllocate", Size);
    

    /* Check for an inadvertent size of zero bytes */

    if (!Size)
    {
        DEBUG_PRINT (ACPI_ERROR, ("CmAllocate: ** ERROR: Attempt to allocate zero bytes! (%s line %d)\n",
                                    Module, Line));
        REPORT_ERROR ("CmAllocate: Attempt to allocate zero bytes");
        Size = 1;
    }

    Address = OsdAllocate (Size);
    if (!Address)
    {
        /* Report allocation error */

        _REPORT_ERROR (Module, Line, Component,
            "CmAllocate: Memory allocation failure");

        DEBUG_PRINT (ACPI_ERROR, ("CmAllocate: Could not allocate size 0x%x\n", Size));

        return_VALUE (NULL);
    }
    
    CmAddElementToAllocList (Address, Size, MEM_MALLOC, Component, Module, Line);

    DEBUG_PRINT (TRACE_ALLOCATIONS, ("CmAllocate: %p Size 0x%x\n", Address, Size));

    return_VALUE (Address);
}



void *
_CmCallocate (
    UINT32                  Size,
    UINT32                  Component,
    ACPI_STRING             Module,
    INT32                   Line)
{
    void                    *Address = NULL;


    FUNCTION_TRACE_U32 ("_CmCallocate", Size);
 
    /* Check for an inadvertent size of zero bytes */

    if (!Size)
    {
        DEBUG_PRINT (ACPI_ERROR, ("CmCallocate: ** ERROR: Attempt to allocate zero bytes! (%s line %d)\n",
                                    Module, Line));
        REPORT_ERROR ("CmCallocate: Attempt to allocate zero bytes");
        return_VALUE (NULL);
    }


    Address = OsdCallocate (Size);
    
    if (!Address)
    {
        /* Report allocation error */

        _REPORT_ERROR (Module, Line, Component,
            "CmCallocate: Memory allocation failure");

        DEBUG_PRINT (ACPI_ERROR, ("CmCallocate: Could not allocate size 0x%x\n", Size));

        return_VALUE (NULL);
    }

    CmAddElementToAllocList (Address, Size, MEM_CALLOC, Component, Module, Line);
    
    DEBUG_PRINT (TRACE_ALLOCATIONS, ("CmCallocate: %p Size 0x%x\n", Address, Size));

    return_VALUE (Address);
}



void
_CmFree (
    void                    *Address,
    UINT32                  Component,
    ACPI_STRING             Module,
    INT32                   Line)
{
    FUNCTION_TRACE_PTR ("_CmFree", Address);
    

    if (NULL == Address)
    {
        _REPORT_ERROR (Module, Line, Component,
            "_CmFree: Trying to delete a NULL address.");

        return_VOID;
    }   
    
    CmDeleteElementFromAllocList (Address, Component, Module, Line);
    OsdFree (Address);
    
    DEBUG_PRINT (TRACE_ALLOCATIONS, ("CmFree: %p freed\n", Address));

    return_VOID;
}


