
/*
 *  ./acpica/subsystem/common/cmapi.c - External interfaces for "global" ACPI functions
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __CMAPI_C__

#include <acpi.h>
#include <events.h>
#include <hardware.h>
#include <namesp.h>
#include <interp.h>
#include <amlcode.h>
#include <debugger.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmapi");



ACPI_STATUS
AcpiInitialize (ACPI_INIT_DATA *InitData)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AcpiInitialize");

    DEBUG_PRINT_RAW (ACPI_OK, ("ACPI Subsystem version [%s]\n", ACPI_CA_VERSION));
    DEBUG_PRINT (ACPI_INFO, ("Initializing ACPI Subsystem...\n"));


    /* Initialize all globals used by the subsystem */

    CmInitGlobals (InitData);

    /* Initialize the OS-Dependent layer */

    Status = OsdInitialize ();
    if (ACPI_FAILURE (Status))
    {
        DEBUG_PRINT (ACPI_ERROR, ("OSD failed to initialize, %s\n", CmFormatException (Status)));
        REPORT_ERROR ("OSD Initialization Failure");
        return_ACPI_STATUS (Status);
    }

    /* Create the default mutex objects */

    Status = CmMutexInitialize ();
    if (ACPI_FAILURE (Status))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Global mutex creation failure, %s\n", CmFormatException (Status)));
        REPORT_ERROR ("Global Mutex Initialization Failure");
        return_ACPI_STATUS (Status);
    }

    /* If configured, initialize the AML debugger */

    DEBUG_EXEC (DbInitialize ());

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiTerminate (void)
{

    FUNCTION_TRACE ("AcpiTerminate");

    /* Terminate the AML Debuger if present */

    Gbl_DbTerminateThreads = TRUE;
    CmReleaseMutex (MTX_DEBUG_CMD_READY);


    /* Shutdown and free all resources */

    CmSubsystemShutdown ();


    /* Free the mutex objects */

    CmMutexTerminate ();


    /* Now we can shutdown the OS-dependent layer */

    OsdTerminate ();

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiGetSystemInfo (
    ACPI_BUFFER             *OutBuffer)
{
    ACPI_SYSTEM_INFO        *InfoPtr;
    UINT32                  i;


    FUNCTION_TRACE ("AcpiGetSystemInfo");


    /*
     *  Must have a valid buffer
     */
    if ((!OutBuffer)          ||
        (!OutBuffer->Pointer))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    if (OutBuffer->Length < sizeof (ACPI_SYSTEM_INFO)) 
    {
        /*
         *  Caller's buffer is too small
         *  fail the call
         */
        OutBuffer->Length = sizeof (ACPI_SYSTEM_INFO);

        return_ACPI_STATUS (AE_BUFFER_OVERFLOW);
    }


    /*
     *  Set return length and get data
     */
    OutBuffer->Length = sizeof (ACPI_SYSTEM_INFO);
    InfoPtr = (ACPI_SYSTEM_INFO *) OutBuffer->Pointer;

    InfoPtr->AcpiCAVersion      = 0x1234;   /* TBD: need a version number, or use the version string */

    /* System flags (ACPI capabilities) */

    InfoPtr->Flags              = Gbl_SystemFlags;

    /* Timer resolution - 24 or 32 bits  */
    
    InfoPtr->TimerResolution    = HwPmtResolution ();

    /* Clear the reserved fields */

    InfoPtr->Reserved1          = 0;
    InfoPtr->Reserved2          = 0;

    /* Current debug levels */

    InfoPtr->DebugLayer         = DebugLayer;
    InfoPtr->DebugLevel         = DebugLevel;

    /* Current status of the ACPI tables, per table type */

    InfoPtr->NumTableTypes = NUM_ACPI_TABLES;
    for (i = 0; i < NUM_ACPI_TABLES; i++);
    {
        InfoPtr->TableInfo[i].Count = Gbl_AcpiTables[i].Count;
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiFormatException (
    ACPI_STATUS             Exception,
    ACPI_BUFFER             *OutBuffer)
{
    UINT32                  Length;
    char                    *FormattedException;


    FUNCTION_TRACE ("AcpiFormatException");


    /*
     *  Must have a valid buffer
     */
    if ((!OutBuffer)          ||
        (!OutBuffer->Pointer))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* Exception must be within range */

    if (Exception > ACPI_MAX_STATUS)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* Convert the exception code */

    FormattedException = CmFormatException (Exception);

    /*
     * Get length of string and check if it will fit in caller's buffer
     */

    Length = STRLEN (FormattedException);
    if (OutBuffer->Length < Length)
    {
        OutBuffer->Length = Length;
        return_ACPI_STATUS (AE_BUFFER_OVERFLOW);
    }


    /* Copy the string, all done */

    STRCPY (OutBuffer->Pointer, FormattedException);

    return_ACPI_STATUS (AE_OK);
}

