 
/*
 *  ./acpica/subsystem/common/cmcopy.c - Internal to external object translation utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMCOPY_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <pnp.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmcopy");


typedef struct Search_st 
{
    ACPI_OBJECT_INTERNAL        *InternalObj;
    UINT32                      Index;
    ACPI_OBJECT                 *ExternalObj;

} PKG_SEARCH_INFO;


/* Used to traverse nested packages */

PKG_SEARCH_INFO                 Level[MAX_PACKAGE_DEPTH];


ACPI_STATUS
CmBuildExternalSimpleObject (
    ACPI_OBJECT_INTERNAL    *InternalObj,
    ACPI_OBJECT             *ExternalObj,
    UINT8                   *DataSpace,
    UINT32                  *BufferSpaceUsed)
{
    UINT32                  Length = 0;
    char                    *SourcePtr = NULL;


    FUNCTION_TRACE ("CmBuildExternalSimpleObject");

    MEMSET(ExternalObj , 0 , sizeof(ACPI_OBJECT));

    /* In general, the external object will be the same type as the internal object */

    ExternalObj->Type = InternalObj->Common.Type;

    /* However, only a limited number of external types are supported */

    switch (ExternalObj->Type)
    {

    case ACPI_TYPE_String:

        Length = InternalObj->String.Length;
        ExternalObj->String.Length = InternalObj->String.Length;
        ExternalObj->String.Pointer = (char *) DataSpace;
        SourcePtr = InternalObj->String.Pointer;
        break;


    case ACPI_TYPE_Buffer:

        Length = InternalObj->Buffer.Length;
        ExternalObj->Buffer.Length = InternalObj->Buffer.Length;
        ExternalObj->Buffer.Pointer = DataSpace;
        SourcePtr = (char *) InternalObj->Buffer.Pointer;
        break;


    case ACPI_TYPE_Number:

        ExternalObj->Number.Value= InternalObj->Number.Value;
        break;


    case INTERNAL_TYPE_Reference:

        /* 
         * This is an object reference.  We use the object type of "Any" to
         * indicate a reference object containing a handle to an ACPI
         * named object.
         */

        ExternalObj->Type = ACPI_TYPE_Any;
        ExternalObj->Reference.Handle = InternalObj->Reference.Nte;
        break;


    case ACPI_TYPE_Processor:

        ExternalObj->Processor.ProcId = InternalObj->Processor.ProcId;
        ExternalObj->Processor.PBLKAddress = InternalObj->Processor.PBLKAddress;
        ExternalObj->Processor.PBLKLength = InternalObj->Processor.PBLKLength;
        break;

    case ACPI_TYPE_Power:

        ExternalObj->PowerResource.SystemLevel = InternalObj->PowerResource.SystemLevel;
        ExternalObj->PowerResource.ResourceOrder = InternalObj->PowerResource.ResourceOrder;
        break;

    default:
        return_ACPI_STATUS (AE_CTRL_RETURN_VALUE);
        break;
    }


    /* Copy data if necessary (strings or buffers) */

    if (Length)
    {
        /*
         * Copy the return data to the caller's buffer
         */
        MEMCPY ((void *) DataSpace, (void *) SourcePtr, Length);
    }


    *BufferSpaceUsed = Length;
    
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
CmBuildExternalPackageObject (
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    UINT8                   *Buffer, 
    UINT32                  *SpaceUsed)
{
    UINT8                   *FreeSpace;
    ACPI_OBJECT             *ExternalObj;
    UINT32                  CurrentDepth = 0;
    ACPI_STATUS             Status;
    UINT32                  Length = 0;
    UINT32                  ThisIndex;
    UINT32                  ObjectSpace;
    ACPI_OBJECT_INTERNAL    *ThisInternalObj;
    ACPI_OBJECT             *ThisExternalObj;
    PKG_SEARCH_INFO         *LevelPtr;
    

    FUNCTION_TRACE ("CmBuildExternalPackageObject");


    /*
     * First package at head of the buffer
     */
    ExternalObj = (ACPI_OBJECT *) Buffer;

    /*
     * Free space begins right after the first package
     */
    FreeSpace = Buffer + sizeof (ACPI_OBJECT);


    /*
     * Initialize the working variables
     */

    MEMSET ((void *) Level, 0, sizeof (Level));

    Level[0].InternalObj    = InternalObj;
    Level[0].ExternalObj    = ExternalObj;
    Level[0].Index          = 0;
    LevelPtr                = &Level[0];
    CurrentDepth            = 0;

    ExternalObj->Type               = InternalObj->Common.Type;
    ExternalObj->Package.Count      = InternalObj->Package.Count;
    ExternalObj->Package.Elements   = (ACPI_OBJECT *) FreeSpace;


    /*
     * Build an array of ACPI_OBJECTS in the buffer
     * and move the free space past it
     */

    FreeSpace += ExternalObj->Package.Count * sizeof (ACPI_OBJECT);


    while (1)
    {
        ThisIndex       = LevelPtr->Index;
        ThisInternalObj = (ACPI_OBJECT_INTERNAL *) LevelPtr->InternalObj->Package.Elements[ThisIndex];
        ThisExternalObj = (ACPI_OBJECT *) &LevelPtr->ExternalObj->Package.Elements[ThisIndex];

        if ((VALID_DESCRIPTOR_TYPE (ThisInternalObj, DESC_TYPE_ACPI_OBJ)) &&
            (IS_THIS_OBJECT_TYPE (ThisInternalObj, ACPI_TYPE_Package)))
        {
            /*
             * If this object is a package then we go one deeper
             */
            if (CurrentDepth >= MAX_PACKAGE_DEPTH-1)
            {
                /*
                 * Too many nested levels of packages for us to handle
                 */
                DEBUG_PRINT (ACPI_ERROR, ("CmBuildPackageObject: Pkg nested too deep (max %d)\n",
                                            MAX_PACKAGE_DEPTH));
                return_ACPI_STATUS (AE_LIMIT);
            }

            /*
             * Build the package object
             */
            ThisExternalObj->Type              = ACPI_TYPE_Package;
            ThisExternalObj->Package.Count     = ThisInternalObj->Package.Count;
            ThisExternalObj->Package.Elements  = (ACPI_OBJECT *) FreeSpace;

            /*
             * Save space for the array of objects (Package elements)
             * update the buffer length counter
             */
            ObjectSpace             = ThisExternalObj->Package.Count * sizeof(ACPI_OBJECT);
            FreeSpace               += ObjectSpace;
            Length                  += ObjectSpace;

            CurrentDepth++;
            LevelPtr                = &Level[CurrentDepth];
            LevelPtr->InternalObj   = ThisInternalObj;
            LevelPtr->ExternalObj   = ThisExternalObj;
            LevelPtr->Index         = 0;

        }   /* if object is a package */
        
        else
        {
            Status = CmBuildExternalSimpleObject (ThisInternalObj, ThisExternalObj, FreeSpace, &ObjectSpace);
            if (Status != AE_OK) 
            {
                /*
                 * Failure get out
                 */
                return_ACPI_STATUS (Status);
            }

            FreeSpace   += ObjectSpace;
            Length      += ObjectSpace;

            LevelPtr->Index++;
            while (LevelPtr->Index >= LevelPtr->InternalObj->Package.Count)
            {
                /*
                 * We've handled all of the objects at this level,  This means that we
                 * have just completed a package.  That package may have contained one
                 * or more packages itself
                 */
                if (CurrentDepth == 0)
                {
                    /*
                     * We have handled all of the objects in the top level package
                     * just add the length of the package objects and get out
                     */
                    *SpaceUsed = Length;
                    return_ACPI_STATUS (AE_OK);
                }

                /*
                 * go back up a level and move the index past the just completed
                 * package object.
                 */
                CurrentDepth--;
                LevelPtr = &Level[CurrentDepth];
                LevelPtr->Index++;
            }
        }   /* else object is NOT a package */
    }   /* while (1)  */


    /*
     * We'll never get here, but the compiler whines about return value
     */
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
CmBuildExternalObject (
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    ACPI_BUFFER             *RetBuffer)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("CmBuildExternalObject");


    if (IS_THIS_OBJECT_TYPE (InternalObj, ACPI_TYPE_Package))
    {
        /*
         * Package objects contain other objects (which can be objects)
         * buildpackage does it all
         */
        Status = CmBuildExternalPackageObject(InternalObj, RetBuffer->Pointer,
                        &RetBuffer->Length);
    }

    else
    {
        /*
         * Build a simple object (no nested objects)
         */
        Status = CmBuildExternalSimpleObject(InternalObj,
                        (ACPI_OBJECT *) RetBuffer->Pointer,
                        ((UINT8 *) RetBuffer->Pointer + sizeof (ACPI_OBJECT)),
                        &RetBuffer->Length);
        /*
         * build simple does not include the object size in the length
         * so we add it in here
         */
        RetBuffer->Length += sizeof (ACPI_OBJECT);
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
CmBuildInternalSimpleObject (
    ACPI_OBJECT             *ExternalObj,
    ACPI_OBJECT_INTERNAL    *InternalObj)
{

    FUNCTION_TRACE ("CmBuildInternalSimpleObject");


    InternalObj->Common.Type = (UINT8) ExternalObj->Type;

    switch (ExternalObj->Type)
    {

    case ACPI_TYPE_String:

        InternalObj->String.Length      = ExternalObj->String.Length;
        InternalObj->String.Pointer     = ExternalObj->String.Pointer;
        break;


    case ACPI_TYPE_Buffer:

        InternalObj->Buffer.Length      = ExternalObj->Buffer.Length;
        InternalObj->Buffer.Pointer     = ExternalObj->Buffer.Pointer;
        break;


    case ACPI_TYPE_Number:
        /*
         * Number is included in the object itself
         */
        InternalObj->Number.Value           = ExternalObj->Number.Value;
        break;


    default:
        return_ACPI_STATUS (AE_CTRL_RETURN_VALUE);
        break;
    }


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
CmBuildInternalPackageObject (
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    UINT8                   *Buffer, 
    UINT32                  *SpaceUsed)
{
    UINT8                   *FreeSpace;
    ACPI_OBJECT             *ExternalObj;
    UINT32                  CurrentDepth = 0;
    ACPI_STATUS             Status = AE_OK;
    UINT32                  Length = 0;
    UINT32                  ThisIndex;
    UINT32                  ObjectSpace = 0;
    ACPI_OBJECT_INTERNAL    *ThisInternalObj;
    ACPI_OBJECT             *ThisExternalObj;
    PKG_SEARCH_INFO         *LevelPtr;
    

    
    FUNCTION_TRACE ("CmBuildInternalPackageObject");


    /*
     * First package at head of the buffer
     */
    ExternalObj = (ACPI_OBJECT *)Buffer;

    /*
     * Free space begins right after the first package
     */
    FreeSpace = Buffer + sizeof(ACPI_OBJECT);


    /*
     * Initialize the working variables
     */
    
    MEMSET ((void *) Level, 0, sizeof(Level));

    Level[0].InternalObj    = InternalObj;
    Level[0].ExternalObj    = ExternalObj;
    LevelPtr                = &Level[0];
    CurrentDepth            = 0;

    ExternalObj->Type               = InternalObj->Common.Type;
    ExternalObj->Package.Count      = InternalObj->Package.Count;
    ExternalObj->Package.Elements   = (ACPI_OBJECT *)FreeSpace;


    /*
     * Build an array of ACPI_OBJECTS in the buffer
     * and move the free space past it
     */

    FreeSpace += ExternalObj->Package.Count * sizeof(ACPI_OBJECT);


    while (1)
    {
        ThisIndex       = LevelPtr->Index;
        ThisInternalObj = (ACPI_OBJECT_INTERNAL *) &LevelPtr->InternalObj->Package.Elements[ThisIndex];
        ThisExternalObj = (ACPI_OBJECT *) &LevelPtr->ExternalObj->Package.Elements[ThisIndex];

        if (IS_THIS_OBJECT_TYPE (ThisInternalObj, ACPI_TYPE_Package))
        {
            /*
             * If this object is a package then we go one deeper
             */
            if (CurrentDepth >= MAX_PACKAGE_DEPTH-1)
            {
                /*
                 * Too many nested levels of packages for us to handle
                 */
                DEBUG_PRINT (ACPI_ERROR, ("CmBuildPackageObject: Pkg nested too deep (max %d)\n",
                                            MAX_PACKAGE_DEPTH));
                return_ACPI_STATUS (AE_LIMIT);
            }

            /*
             * Build the package object
             */
            ThisExternalObj->Type               = ACPI_TYPE_Package;
            ThisExternalObj->Package.Count      = ThisInternalObj->Package.Count;
            ThisExternalObj->Package.Elements   = (ACPI_OBJECT *) FreeSpace;

            /*
             * Save space for the array of objects (Package elements)
             * update the buffer length counter
             */
            ObjectSpace             = ThisExternalObj->Package.Count * sizeof (ACPI_OBJECT);
            FreeSpace               += ObjectSpace;
            Length                  += ObjectSpace;

            CurrentDepth++;
            LevelPtr                = &Level[CurrentDepth];
            LevelPtr->InternalObj   = ThisInternalObj;
            LevelPtr->ExternalObj   = ThisExternalObj;
            LevelPtr->Index         = 0;

        }   /* if object is a package */
        
        else
        {
/*            Status = CmBuildSimpleObject(ThisInternalObj, ThisExternalObj, FreeSpace, &ObjectSpace);
*/
            if (Status != AE_OK) 
            {
                /*
                 * Failure get out
                 */
                return_ACPI_STATUS (Status);
            }

            FreeSpace   += ObjectSpace;
            Length      +=ObjectSpace;

            LevelPtr->Index++;
            while (LevelPtr->Index >= LevelPtr->InternalObj->Package.Count)
            {
                /*
                 * We've handled all of the objects at this level,  This means that we
                 * have just completed a package.  That package may have contained one
                 * or more packages itself
                 */
                if (CurrentDepth == 0)
                {
                    /*
                     * We have handled all of the objects in the top level package
                     * just add the length of the package objects and get out
                     */
                    *SpaceUsed = Length;
                    return_ACPI_STATUS (AE_OK);
                }

                /*
                 * go back up a level and move the index past the just completed
                 * package object.
                 */
                CurrentDepth--;
                LevelPtr = &Level[CurrentDepth];
                LevelPtr->Index++;
            }
        }   /* else object is NOT a package */
    }   /* while (1)  */


    /*
     * We'll never get here, but the compiler whines about return value
     */
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
CmBuildInternalObject (
    ACPI_OBJECT             *ExternalObj, 
    ACPI_OBJECT_INTERNAL    *InternalObj)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("CmBuildInternalObject");


    if (ExternalObj->Type == ACPI_TYPE_Package)
    {
        /*
         * Package objects contain other objects (which can be objects)
         * buildpackage does it all
         */
/*
        Status = CmBuildInternalPackageObject(InternalObj, RetBuffer->Pointer,
                        &RetBuffer->Length);
*/
        DEBUG_PRINT (ACPI_ERROR, ("CmBuildInternalObject: Packages as parameters not implemented!\n"));
        
        return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
    }

    else
    {
        /*
         * Build a simple object (no nested objects)
         */
        Status = CmBuildInternalSimpleObject (ExternalObj, InternalObj);
        /*
         * build simple does not include the object size in the length
         * so we add it in here
         */
    }

    return_ACPI_STATUS (Status);
}

