
/*
 *  ./acpica/subsystem/common/cmdebug.c - Debug print routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMDEBUG_C__

#include <acpi.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmdebug");





INT32
GetDebugLevel (void)
{

    return DebugLevel;
}

void
SetDebugLevel (
    INT32                   NewDebugLevel)
{

    DebugLevel = NewDebugLevel;
}



void
FunctionTrace (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName)
{
        
    Gbl_NestingLevel++;  
    
    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Entered Function: %s\n", Gbl_NestingLevel, FunctionName);
}



void
FunctionTracePtr (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    void                    *Pointer)
{

    Gbl_NestingLevel++;
    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Entered Function: %s, 0x%p\n", 
                Gbl_NestingLevel, FunctionName, Pointer);
}



void
FunctionTraceStr (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    char                    *String)
{

    Gbl_NestingLevel++;
    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Entered Function: %s, %s\n", 
                Gbl_NestingLevel, FunctionName, String);
}



void
FunctionTraceU32 (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    UINT32                  Integer)
{

    Gbl_NestingLevel++;
    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Entered Function: %s, 0x%lX\n", 
                Gbl_NestingLevel, FunctionName, Integer);
}



void
FunctionExit (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName)
{

    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Exiting Function: %s\n", Gbl_NestingLevel, FunctionName);
    Gbl_NestingLevel--;
}



void
FunctionStatusExit (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    ACPI_STATUS             Status)
{

    if (Status > ACPI_MAX_STATUS)
    {
        DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                    " %2.2ld Exiting Function: %s, [Unknown Status] 0x%X\n", 
                    Gbl_NestingLevel, FunctionName, Status);
    }

    else
    { 
        DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                    " %2.2ld Exiting Function: %s, %s\n", 
                    Gbl_NestingLevel, FunctionName, CmFormatException (Status));
    }

    Gbl_NestingLevel--;
}



void
FunctionValueExit (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    NATIVE_UINT             Value)
{

    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Exiting Function: %s, 0x%X\n", 
                Gbl_NestingLevel, FunctionName, Value);
    Gbl_NestingLevel--;
}



void
FunctionPtrExit (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    char                    *FunctionName,
    char                    *Ptr)
{

    DebugPrint (ModuleName, LineNumber, ComponentId, TRACE_FUNCTIONS,
                " %2.2ld Exiting Function: %s, 0x%p\n", 
                Gbl_NestingLevel, FunctionName, Ptr);
    Gbl_NestingLevel--;
}



void
DebugPrint (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId, 
    INT32                   PrintLevel, 
    char                    *Format,
    ...)
{
    va_list                 args;
     
     
    /* Both the level and the component must be enabled */

    if ((PrintLevel & DebugLevel) && (ComponentId & DebugLayer))
    {
        va_start (args, Format);

        OsdPrintf ("%8s-%04d: ", ModuleName, LineNumber);
        OsdVprintf (Format, args);  
    }
}



void
DebugPrintPrefix (
    char                    *ModuleName, 
    INT32                   LineNumber)
{


    OsdPrintf ("%8s-%04d: ", ModuleName, LineNumber);
}



void
DebugPrintRaw (
    char                    *Format, 
    ...)
{
    va_list                 args;


    va_start (args, Format);

    OsdVprintf (Format, args);

    va_end (args);
}




void
CmDumpBuffer (
    char                    *Buffer, 
    UINT32                  Count, 
    UINT32                  Display,
    INT32                   ComponentId)
{
    UINT32                  i = 0;
    UINT32                  j;
    UINT8                   BufChar;


    /* Only dump the buffer if tracing is enabled */

    if (!((TRACE_TABLES & DebugLevel) && (ComponentId & DebugLayer)))
    {
        return;
    }


    /*
     * Nasty little dump buffer routine!
     */
    while (i < Count)
    {
        /* Print current offset */

        OsdPrintf ("%05X    ", i);


        /* Print 16 hex chars */

        for (j = 0; j < 16;)
        {
            if (i + j >= Count)
            {
                OsdPrintf ("\n");
                return;
            }

            /* Make sure that the char doesn't get sign-extended! */

            switch (Display)
            {
            /* Default is BYTE display */

            default:
                OsdPrintf ("%02X ", *((UINT8 *) &Buffer[i + j]));
                j += 1;
                break;

            case DB_WORD_DISPLAY:
                OsdPrintf ("%04X ", *((UINT16 *) &Buffer[i + j]));
                j += 2;
                break;

            case DB_DWORD_DISPLAY:
                OsdPrintf ("%08X ", *((UINT32 *) &Buffer[i + j]));
                j += 4;
                break;

            case DB_QWORD_DISPLAY:
                OsdPrintf ("%08X", *((UINT32 *) &Buffer[i + j]));
                OsdPrintf ("%08X ", *((UINT32 *) &Buffer[i + j + 4]));
                j += 8;
                break;
            }
        }


        /* 
         * Print the ASCII equivalent characters
         * But watch out for the bad unprintable ones...
         */

        for (j = 0; j < 16; j++)
        {
            if (i + j >= Count)
            {
                OsdPrintf ("\n");
                return;
            }

            BufChar = Buffer[i + j];
            if ((BufChar > 0x1F && BufChar < 0x2E) ||
                (BufChar > 0x2F && BufChar < 0x61) ||
                (BufChar > 0x60 && BufChar < 0x7F))
            {
                OsdPrintf ("%c", BufChar);
            }
            else
            {
                OsdPrintf (".");
            }
        }

        /* Done with that line. */

        OsdPrintf ("\n");
        i += 16;
    }

    return;
}





