 
/*
 *  ./acpica/subsystem/common/cmdelete.c - object deletion utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMDELETE_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <parser.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmdelete");




void
CmDeleteInternalObj (
    ACPI_OBJECT_INTERNAL    *Object)
{
    void                    *ObjPointer = NULL;


    FUNCTION_TRACE_PTR ("CmDeleteInternalObj", Object);


    if (!Object)
    {
        return_VOID;
    }

    /* 
     * Must delete or free any pointers within the object that are not
     * actual ACPI objects (for example, a raw buffer pointer).
     */

    switch (Object->Common.Type)
    {

    case ACPI_TYPE_String:

        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: **** String %p, ptr %p\n", 
                                Object, Object->String.Pointer));

        /* Free the actual string buffer */

        ObjPointer = Object->String.Pointer;
        break;


    case ACPI_TYPE_Buffer:
        
        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: **** Buffer %p, ptr %p\n", 
                                Object, Object->Buffer.Pointer));

        /* Free the actual buffer */

        ObjPointer = Object->Buffer.Pointer;
        break;


    case ACPI_TYPE_Package:
        
        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: **** Package of count %d\n", 
                                Object->Package.Count));

        /* Elements of the package are not handled here, they are deleted separately */

        /* Free the (variable length) element pointer array */

        ObjPointer = Object->Package.Elements;
        break;


    case ACPI_TYPE_Mutex:

        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: ***** Mutex %p, Semaphore %p\n", 
                                Object, Object->Mutex.Semaphore));

        OsdDeleteSemaphore (Object->Mutex.Semaphore);
        break;


    case ACPI_TYPE_Event:

        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: ***** Event %p, Semaphore %p\n", 
                                Object, Object->Event.Semaphore));

        OsdDeleteSemaphore (Object->Event.Semaphore);
        Object->Event.Semaphore = NULL;
        break;


    case ACPI_TYPE_Method:

        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: ***** Method %p, ParserOp %p\n", 
                                Object, Object->Method.ParserOp));

        /* Delete parse tree if it exists */

        if (Object->Method.ParserOp)
        {
            PsDeleteParseTree (Object->Method.ParserOp);
            Object->Method.ParserOp = NULL;
        }

        /* Delete semaphore if it exists */

        if (Object->Method.Semaphore)
        {
            OsdDeleteSemaphore (Object->Method.Semaphore);
            Object->Method.Semaphore = NULL;
        }

        break;


    default:
        break;
    }


    /*
     * Delete any allocated memory found above 
     */

    if (ObjPointer)
    {
        if (!TbSystemTablePointer (ObjPointer))
        {
            DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: Deleting Obj Ptr %p \n", 
                                    ObjPointer));

            CmFree (ObjPointer);
        }
    }

    
    /* Only delete the object if it was dynamically allocated */

    if (Object->Common.Flags & AO_STATIC_ALLOCATION)
    {
        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: Object %p [%s] static allocation, no delete\n",
                                Object, CmGetTypeName (Object->Common.Type)));
    }

    else
    {
        DEBUG_PRINT (ACPI_INFO, ("CmDeleteInternalObj: Deleting object %p [%s]\n",
                                Object, CmGetTypeName (Object->Common.Type)));

        CmDeleteObjectDesc (Object);
        
    }

    return_VOID;
}






ACPI_STATUS
CmDeleteInternalObjectList (
    ACPI_OBJECT_INTERNAL    **ObjList)
{
    ACPI_OBJECT_INTERNAL    **InternalObj;
 
    
    FUNCTION_TRACE ("CmDeleteInternalObjectList");


    /* Walk the null-terminated internal list */

    for (InternalObj = ObjList; *InternalObj; InternalObj++)
    {
        /* 
         * Check for a package
         * Simple objects are simply stored in the array and do not
         * need to be deleted separately.
         */

        if (IS_THIS_OBJECT_TYPE ((*InternalObj), ACPI_TYPE_Package))
        {
            /* Delete the package */
            
            /*
             * TBD: This might not be the right thing to do, depending
             * on how the internal package object was allocated!!!
             */

            CmDeleteInternalObj (*InternalObj);
        }

    }


    /* Free the combined parameter pointer list and object array */

    CmFree (ObjList);

    return_ACPI_STATUS (AE_OK);
}


