
/*
 *  ./acpica/subsystem/common/cminit.c - Common ACPI subsystem initialization
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __CMINIT_C__

#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>
#include <parser.h>
#include <dispatch.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cminit");




void
CmFacpRegisterError (
    char                    *RegisterName, 
    UINT32                  Value)
{
    
    REPORT_ERROR ("Invalid FACP register value");
    
    DEBUG_PRINT (ACPI_ERROR, ("Invalid FACP register value, %s = 0x%X\n", RegisterName, Value));
}



ACPI_STATUS
CmHardwareInitialize (void)
{
    ACPI_STATUS             Status = AE_OK;
    INT32                   Index;


    
    FUNCTION_TRACE ("CmHardwareInitialize");


    /* Are we running on the actual hardware */

    if (!Gbl_AcpiHardwarePresent)
    {
        /* No, just return */

        return_ACPI_STATUS (AE_OK);
    }

    /* We must have the ACPI tables by the time we get here */

    if (!Gbl_FACP)
    {
        Gbl_RestoreAcpiChipset = FALSE;

        DEBUG_PRINT (ACPI_ERROR, ("CmHardwareInitialize: No FACP!\n"));

        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    /* Must support *some* mode! */
/*
    if (!(SystemFlags & SYS_MODES_MASK))
    {
        RestoreAcpiChipset = FALSE;

        DEBUG_PRINT (ACPI_ERROR, ("CmHardwareInitialize: Supported modes uninitialized!\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

*/


    switch (Gbl_SystemFlags & SYS_MODES_MASK)
    {
        /* Identify current ACPI/legacy mode   */

    case (SYS_MODE_ACPI):

        Gbl_OriginalMode = SYS_MODE_ACPI;
        DEBUG_PRINT (ACPI_INFO, ("System supports ACPI mode only.\n"));
        break;


    case (SYS_MODE_LEGACY):

        Gbl_OriginalMode = SYS_MODE_LEGACY;
        DEBUG_PRINT (ACPI_INFO,
                    ("Tables loaded from buffer, hardware assumed to support LEGACY mode only.\n"));
        break;


    case (SYS_MODE_ACPI | SYS_MODE_LEGACY):

        if (HwGetMode () == SYS_MODE_ACPI)
        {
            Gbl_OriginalMode = SYS_MODE_ACPI;
        }
        else
        {
            Gbl_OriginalMode = SYS_MODE_LEGACY;
        }

        DEBUG_PRINT (ACPI_INFO, ("System supports both ACPI and LEGACY modes.\n"));

        DEBUG_PRINT (ACPI_INFO, ("System is currently in %s mode.\n",
                                (Gbl_OriginalMode == SYS_MODE_ACPI) ? "ACPI" : "LEGACY"));
        break;
    }


    if (Gbl_SystemFlags & SYS_MODE_ACPI)
    {
        /* Target system supports ACPI mode */

        /* 
         * The purpose of this block of code is to save the initial state
         * of the ACPI event enable registers. An exit function will be
         * registered which will restore this state when the application
         * exits. The exit function will also clear all of the ACPI event
         * status bits prior to restoring the original mode.
         *
         * The location of the PM1aEvtBlk enable registers is defined as the
         * base of PM1aEvtBlk + PM1aEvtBlkLength / 2. Since the spec further
         * fully defines the PM1aEvtBlk to be a total of 4 bytes, the offset
         * for the enable registers is always 2 from the base. It is hard
         * coded here. If this changes in the spec, this code will need to
         * be modified. The PM1bEvtBlk behaves as expected.
         */

        Gbl_Pm1EnableRegisterSave = OsdIn16 ((Gbl_FACP->Pm1aEvtBlk + 2));
        if (Gbl_FACP->Pm1bEvtBlk)
        {
            Gbl_Pm1EnableRegisterSave |= OsdIn16 ((Gbl_FACP->Pm1bEvtBlk + 2));
        }


        /* 
         * The GPEs behave similarly, except that the length of the register
         * block is not fixed, so the buffer must be allocated with malloc 
         */

        if (Gbl_FACP->Gpe0Blk && Gbl_FACP->Gpe0BlkLen)
        {
            /* GPE0 specified in FACP  */

            Gbl_Gpe0EnableRegisterSave = CmAllocate ((ACPI_SIZE) (Gbl_FACP->Gpe0BlkLen / 2));
            if (!Gbl_Gpe0EnableRegisterSave)
            {
                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            /* Save state of GPE0 enable bits */

            for (Index = 0; Index < Gbl_FACP->Gpe0BlkLen / 2; Index++)
            {
                Gbl_Gpe0EnableRegisterSave[Index] =
                    OsdIn8 ((Gbl_FACP->Gpe0Blk + Gbl_FACP->Gpe0BlkLen / 2));
            }
        }
        
        else
        {
            Gbl_Gpe0EnableRegisterSave = NULL;
        }

        if (Gbl_FACP->Gpe1Blk && Gbl_FACP->Gpe1BlkLen)
        {
            /* GPE1 defined */

            Gbl_Gpe1EnableRegisterSave = CmAllocate ((ACPI_SIZE) (Gbl_FACP->Gpe1BlkLen / 2));
            if (!Gbl_Gpe1EnableRegisterSave)
            {
                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            /* save state of GPE1 enable bits */
    
            for (Index = 0; Index < Gbl_FACP->Gpe1BlkLen / 2; Index++)
            {
                Gbl_Gpe1EnableRegisterSave[Index] =
                    OsdIn8 ((Gbl_FACP->Gpe1Blk + Gbl_FACP->Gpe1BlkLen / 2));
            }
        }
        
        else
        {
            Gbl_Gpe1EnableRegisterSave = NULL;
        }
    
    
        /* 
         * Verify Fixed ACPI Description Table fields,
         * but don't abort on any problems, just display error
         */

        if (Gbl_FACP->Pm1EvtLen < 4)
        {
            CmFacpRegisterError ("PM1_EVT_LEN", (UINT32) Gbl_FACP->Pm1EvtLen);
        }

        if (!Gbl_FACP->Pm1CntLen)
        {
            CmFacpRegisterError ("PM1_CNT_LEN", (UINT32) Gbl_FACP->Pm1CntLen);
        }

        if (!Gbl_FACP->Pm1aEvtBlk)
        {
            CmFacpRegisterError ("PM1a_EVT_BLK", Gbl_FACP->Pm1aEvtBlk);
        }

        if (!Gbl_FACP->Pm1aCntBlk)
        {
            CmFacpRegisterError ("PM1a_CNT_BLK", Gbl_FACP->Pm1aCntBlk);
        }

        if (!Gbl_FACP->PmTmrBlk)
        {
            CmFacpRegisterError ("PM_TMR_BLK", Gbl_FACP->PmTmrBlk);
        }

        if (Gbl_FACP->Pm2CntBlk && !Gbl_FACP->Pm2CntLen)
        {
            CmFacpRegisterError ("PM2_CNT_LEN", (UINT32) Gbl_FACP->Pm2CntLen);
        }

        if (Gbl_FACP->PmTmLen < 4)
        {
            CmFacpRegisterError ("PM_TM_LEN", (UINT32) Gbl_FACP->PmTmLen);
        }

        if (Gbl_FACP->Gpe0Blk && (Gbl_FACP->Gpe0BlkLen & 1))    /* length not multiple of 2    */
        {
            CmFacpRegisterError ("GPE0_BLK_LEN", (UINT32) Gbl_FACP->Gpe0BlkLen);
        }

        if (Gbl_FACP->Gpe1Blk && (Gbl_FACP->Gpe1BlkLen & 1))    /* length not multiple of 2    */
        {
            CmFacpRegisterError ("GPE1_BLK_LEN", (UINT32) Gbl_FACP->Gpe1BlkLen);
        }
    }


    return_ACPI_STATUS (Status);
}



void
CmTerminate (void)
{

    FUNCTION_TRACE ("CmTerminate");


    /* Free global tables, etc. */

    if (Gbl_Gpe0EnableRegisterSave)
    {
        CmFree (Gbl_Gpe0EnableRegisterSave);
    }

    if (Gbl_Gpe1EnableRegisterSave)
    {
        CmFree (Gbl_Gpe1EnableRegisterSave);
    }


    return_VOID;
}

 

ACPI_STATUS
CmSubsystemShutdown (void)
{

    FUNCTION_TRACE ("CmSubsystemShutdown");
    
    /* Just exit if subsystem is already shutdown */

    if (Gbl_Shutdown)
    {
        DEBUG_PRINT (ACPI_ERROR, ("ACPI Subsystem is already terminated\n"));
        return_ACPI_STATUS (AE_OK);
    }

    /* Subsystem appears active, go ahead and shut it down */

    Gbl_Shutdown = TRUE;
    DEBUG_PRINT (ACPI_INFO, ("Shutting down ACPI Subsystem...\n"));


    /* Close the Namespace */

    NsTerminate ();

    /* Close the Event Handling */

    EvTerminate ();

    /* Close the globals */

    CmTerminate ();

    /* Flush the local cache(s) */

    CmDeleteGenericStateCache ();
    CmDeleteObjectCache ();
    DsDeleteWalkStateCache ();

    /* Close the Parser */

    /* TBD: PsTerminate () */

    PsDeleteParseCache ();

    /* Debug only - display leftover memory allocation, if any */

    CmDumpCurrentAllocations (ACPI_UINT32_MAX, NULL);

    BREAKPOINT3;

    return_ACPI_STATUS (AE_OK);
}




