
/*
 *  ./acpica/subsystem/common/cmmutex.c - local mutual exclusion mechanisms
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMMUTEX_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmmutex");




ACPI_STATUS
CmMutexInitialize (
    void)
{
    UINT32                  i;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("CmMutexInitialize");


    /*
     * Create each of the predefined mutex objects
     */
    for (i = 0; i < NUM_MTX; i++)
    {
        Status = CmCreateMutex (i);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }

    return_ACPI_STATUS (AE_OK);
}



void
CmMutexTerminate (
    void)
{
    UINT32                  i;


    FUNCTION_TRACE ("CmMutexTerminate");

    
    /*
     * Delete each predefined mutex object
     */
    for (i = 0; i < NUM_MTX; i++)
    {
        CmDeleteMutex (i);
    }

    return_VOID;
}



ACPI_STATUS
CmCreateMutex (
    ACPI_MUTEX_HANDLE       MutexId)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_U32 ("CmCreateMutex", MutexId);


    if (MutexId > MAX_MTX)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    if (!Gbl_AcpiMutexInfo[MutexId].Mutex)
    {
        Status = OsdCreateSemaphore (1, &Gbl_AcpiMutexInfo[MutexId].Mutex);
        Gbl_AcpiMutexInfo[MutexId].Locked = FALSE;
        Gbl_AcpiMutexInfo[MutexId].UseCount = 0;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
CmDeleteMutex (
    ACPI_MUTEX_HANDLE       MutexId)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE_U32 ("CmDeleteMutex", MutexId);


    if (MutexId > MAX_MTX)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    Status = OsdDeleteSemaphore (Gbl_AcpiMutexInfo[MutexId].Mutex);

    Gbl_AcpiMutexInfo[MutexId].Mutex = NULL;
    Gbl_AcpiMutexInfo[MutexId].Locked = FALSE;

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
CmAcquireMutex (
    ACPI_MUTEX_HANDLE       MutexId)
{
    ACPI_STATUS             Status;


    DEBUG_PRINT (TRACE_MUTEX, ("Acquiring Mutex [%s]\n", CmGetMutexName (MutexId)));

    if (MutexId > MAX_MTX)
    {
        return (AE_BAD_PARAMETER);
    }


    Status = OsdWaitSemaphore (Gbl_AcpiMutexInfo[MutexId].Mutex, 1, WAIT_FOREVER);

    DEBUG_PRINT (TRACE_MUTEX, ("Acquired Mutex  [%s] Status %s\n", 
                    CmGetMutexName (MutexId), CmFormatException (Status)));

    if (ACPI_SUCCESS (Status))
    {
        Gbl_AcpiMutexInfo[MutexId].Locked = TRUE;
        Gbl_AcpiMutexInfo[MutexId].UseCount++;
    }

    return (Status);
}



ACPI_STATUS
CmReleaseMutex (
    ACPI_MUTEX_HANDLE       MutexId)
{
    ACPI_STATUS             Status;


    DEBUG_PRINT (TRACE_MUTEX, ("Releasing Mutex [%s]\n", CmGetMutexName (MutexId)));

    if (MutexId > MAX_MTX)
    {
        return (AE_BAD_PARAMETER);
    }


    Gbl_AcpiMutexInfo[MutexId].Locked = FALSE;  /* Mark before unlocking */

    Status = OsdSignalSemaphore (Gbl_AcpiMutexInfo[MutexId].Mutex, 1);

    DEBUG_PRINT (TRACE_MUTEX, ("Released Mutex  [%s] Status %s\n", 
                    CmGetMutexName (MutexId), CmFormatException (Status)));

    return (Status);
}



