 
/*
 *  ./acpica/subsystem/common/cmobject.c - ACPI object create/delete/cache routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMOBJECT_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <pnp.h>
#include <tables.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmobject");



ACPI_OBJECT_INTERNAL *
_CmCreateInternalObject (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId,
    ACPI_OBJECT_TYPE        Type)
{
    ACPI_OBJECT_INTERNAL    *Object;


    FUNCTION_TRACE_STR ("CmCreateInternalObject", CmGetTypeName (Type));


    /* Allocate the raw object descriptor */

    Object = _CmAllocateObjectDesc (ModuleName, LineNumber, ComponentId);
    if (!Object)
    {   
        /* Allocation failure */

        return_VALUE (NULL);
    }

    /* Save the object type in the object descriptor */

    Object->Common.Type = (UINT8) Type;
    Object->Common.Size = (UINT8) sizeof (ACPI_OBJECT_INTERNAL);

    /* Init the reference count */

    Object->Common.ReferenceCount = 1;

    /* Any per-type initialization should go here */
    

    /* Memory allocation metrics - compiled out in non debug mode. */

    INCREMENT_OBJECT_METRICS (sizeof (ACPI_OBJECT_INTERNAL));

    return_VALUE (Object);
}




BOOLEAN
CmValidInternalObject (
    void                    *Object)
{

    /* Check for a null pointer */

    if (!Object)
    {
        DEBUG_PRINT (ACPI_INFO, ("CmValidInternalObject: **** Null Object Ptr\n"));
        return FALSE;
    }

    /* Check for a pointer within one of the ACPI tables */

    if (TbSystemTablePointer (Object))
    {
        DEBUG_PRINT (ACPI_INFO, ("CmValidInternalObject: **** Object %p is a Pcode Ptr\n",
                        Object));
        return FALSE;
    }

    /* Check the descriptor type field */

    if (!VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_ACPI_OBJ))
    {
        /* Not an ACPI internal object, do some further checking */

        if (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_NTE))
        {
            DEBUG_PRINT (ACPI_INFO, ("CmValidInternalObject: **** Obj %p is NTE, not ACPI obj\n",
                            Object));
        }

        else if (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_PARSER))
        {
            DEBUG_PRINT (ACPI_INFO, ("CmValidInternalObject: **** Obj %p is a parser obj, not ACPI obj\n",
                            Object));
        }

        else
        {
            DEBUG_PRINT (ACPI_INFO, ("CmValidInternalObject: **** Obj %p is of unknown type\n",
                            Object));
        }

        return FALSE;
    }


    /* The object appears to be a valid ACPI_OBJECT_INTERNAL */

    return TRUE;
}



void *
_CmAllocateObjectDesc (
    char                    *ModuleName, 
    INT32                   LineNumber, 
    INT32                   ComponentId)
{
    ACPI_OBJECT_INTERNAL    *Object;
    

    FUNCTION_TRACE ("_AllocateObjectDesc");


    CmAcquireMutex (MTX_CACHES);

    Gbl_ObjectCacheRequests++;

    /* Check the cache first */

    if (Gbl_ObjectCache)
    {
        /* There is an object available, use it */

        Object = Gbl_ObjectCache;
        Gbl_ObjectCache = Object->Common.Next;
        Object->Common.Next = NULL;

        Gbl_ObjectCacheHits++;
        Gbl_ObjectCacheDepth--;

        CmReleaseMutex (MTX_CACHES);
    }

    else
    {
        /* The cache is empty, create a new object */

        CmReleaseMutex (MTX_CACHES);

        /* Attempt to allocate new descriptor */

        Object = _CmCallocate (sizeof (ACPI_OBJECT_INTERNAL), ComponentId,
                                    ModuleName, LineNumber);
        
        if (!Object)
        {
            /* Allocation failed */
        
            _REPORT_ERROR (ModuleName, LineNumber, ComponentId, 
                            "Could not allocate Object Descriptor");

            return_VALUE (NULL);
        }
    }

    /* Mark the descriptor type */

    Object->Common.DataType = DESC_TYPE_ACPI_OBJ;

    DEBUG_PRINT (TRACE_ALLOCATIONS, ("AllocateObjectDesc: %p Size 0x%x\n",
                    Object, sizeof (ACPI_OBJECT_INTERNAL)));

    return_VALUE (Object);
}




void
CmDeleteObjectDesc (
    ACPI_OBJECT_INTERNAL    *Object)
{


    /* Object must be an ACPI_OBJECT_INTERNAL */

    if (Object->Common.DataType != DESC_TYPE_ACPI_OBJ)
    {
        DEBUG_PRINT (ACPI_ERROR, ("CmDeleteObjectDesc: Obj %p is not an ACPI object\n", Object));
        return;
    }

    /* Make sure that the object isn't already in the cache */

    if (Object->Common.Next)
    {
        DEBUG_PRINT (ACPI_ERROR, ("CmDeleteObjectDesc: Obj %p is already in the object cache\n", Object));
        return;
    }


    /* If cache is full, just free this object */

    if (Gbl_ObjectCacheDepth >= MAX_OBJECT_CACHE_DEPTH)
    {
        /* 
         * Memory allocation metrics.  Call the macro here since we only
         * care about dynamically allocated objects.
         */
        DECREMENT_OBJECT_METRICS (Gbl_ObjectCache->Common.Size);

        CmFree (Object);
        return;
    }

    CmAcquireMutex (MTX_CACHES);

    /* Clear the entire object.  This is important! */

    MEMSET (Object, 0, sizeof (ACPI_OBJECT_INTERNAL));
    Object->Common.DataType = DESC_TYPE_ACPI_OBJ;

    /* Put the object at the head of the global cache list */

    Object->Common.Next = Gbl_ObjectCache;
    Gbl_ObjectCache = Object;
    Gbl_ObjectCacheDepth++;


    CmReleaseMutex (MTX_CACHES);
}




void
CmDeleteObjectCache (
    void)
{
    ACPI_OBJECT_INTERNAL    *Next;


    FUNCTION_TRACE ("CmDeleteObjectCache");


    /* Traverse the global cache list */

    while (Gbl_ObjectCache)
    {
        /* Delete one cached state object */

        Next = Gbl_ObjectCache->Common.Next;
        Gbl_ObjectCache->Common.Next = NULL;

        /* 
         * Memory allocation metrics.  Call the macro here since we only
         * care about dynamically allocated objects.
         */
        DECREMENT_OBJECT_METRICS (Gbl_ObjectCache->Common.Size);
        
        CmFree (Gbl_ObjectCache);
        Gbl_ObjectCache = Next;
    }

    return_VOID;
}
 


void
CmInitStaticObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{

    FUNCTION_TRACE_PTR ("CmInitStaticObject", ObjDesc);


    if (!ObjDesc)
    {
        return_VOID;
    }


    /*
     * Clear the entire descriptor 
     */
    MEMSET ((void *) ObjDesc, 0, sizeof (ACPI_OBJECT_INTERNAL));


    /*
     * Initialize the header fields
     * 1) This is an ACPI_OBJECT_INTERNAL descriptor
     * 2) The size is the full object (worst case)
     * 3) The flags field indicates static allocation 
     * 4) Reference count starts at one (not really necessary since the
     *    object can't be deleted, but keeps everything sane)
     */

    ObjDesc->Common.DataType        = DESC_TYPE_ACPI_OBJ;
    ObjDesc->Common.Size            = sizeof (ACPI_OBJECT_INTERNAL);
    ObjDesc->Common.Flags           = AO_STATIC_ALLOCATION;
    ObjDesc->Common.ReferenceCount  = 1;

    return_VOID;
}

