 
/*
 *  ./acpica/subsystem/common/cmrefcnt.c - Object reference count support
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMREFCNT_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <parser.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmrefcnt");



void
CmUpdateRefCount (
    ACPI_OBJECT_INTERNAL    *Object,
    INT32                   Action)
{
    UINT16                  Count;
    UINT16                  NewCount;


    if (!Object)
    {
        return;
    }


    Count = Object->Common.ReferenceCount;
    NewCount = Count;
    
    /*
     * Reference count action (increment, decrement, or force delete)
     */

    switch (Action)
    {

    case REF_INCREMENT:

        NewCount++;
        Object->Common.ReferenceCount = NewCount;

        DEBUG_PRINT (ACPI_INFO, ("CmUpdateRefCount: Obj %p Refs=%d, [Incremented]\n",
                        Object, NewCount));
        break;


    case REF_DECREMENT:

        if (Count < 1)
        {
            DEBUG_PRINT (ACPI_INFO, ("CmUpdateRefCount: Obj %p Refs=%d, can't decrement! (Set to 0)\n",
                            Object, NewCount));

            NewCount = 0;
        }

        else
        {
            NewCount--;
            
            DEBUG_PRINT (ACPI_INFO, ("CmUpdateRefCount: Obj %p Refs=%d, [Decremented]\n",
                            Object, NewCount));
        }

        if (Object->Common.Type == ACPI_TYPE_Method)
        {
            DEBUG_PRINT (ACPI_INFO, ("CmUpdateRefCount: Method Obj %p Refs=%d, [Decremented]\n",
                            Object, NewCount));
        }

        Object->Common.ReferenceCount = NewCount;
        if (NewCount == 0)
        {   
            CmDeleteInternalObj (Object);
        }

        break;


    case REF_FORCE_DELETE:

        DEBUG_PRINT (ACPI_INFO, ("CmUpdateRefCount: Obj %p Refs=%d, Force delete! (Set to 0)\n",
                        Object, Count));
        
        NewCount = 0;
        Object->Common.ReferenceCount = NewCount;
        CmDeleteInternalObj (Object);
        break;


    default: 

        DEBUG_PRINT (ACPI_ERROR, ("CmUpdateRefCount: Unknown action (%d)\n", 
                        Action));
        break;
    }


    /*
     * Sanity check the reference count, for debug purposes only.
     * (A deleted object will have a huge reference count)
     */

    if (Count > MAX_REFERENCE_COUNT)
    {

        DEBUG_PRINT (ACPI_ERROR, ("CmUpdateRefCount: **** AE_ERROR **** Invalid Reference Count (0x%X) in object %p\n\n", 
                        Count, Object));
    }

    return;
}



ACPI_STATUS
CmUpdateObjectReference (
    ACPI_OBJECT_INTERNAL    *Object,
    UINT16                  Action)
{
    ACPI_STATUS             Status;
    UINT32                  i;
    ACPI_OBJECT_INTERNAL    *Next;
    ACPI_OBJECT_INTERNAL    *New;
    ACPI_GENERIC_STATE       *StateList = NULL;
    ACPI_GENERIC_STATE       *State;


    FUNCTION_TRACE_PTR ("CmUpdateObjectReference", Object);


    /* Ignore a null object ptr */

    if (!Object)
    {
        return_ACPI_STATUS (AE_OK);
    }


    /* 
     * Make sure that this isn't a namespace handle or an AML pointer
     */

    if (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_NTE))
    {
        DEBUG_PRINT (ACPI_INFO, ("CmUpdateObjectReference: Object %p is NS handle\n",
                        Object));
        return_ACPI_STATUS (AE_OK);
    }

    if (TbSystemTablePointer (Object))
    {
        DEBUG_PRINT (ACPI_INFO, ("CmUpdateObjectReference: **** Object %p is Pcode Ptr\n",
                        Object));
        return_ACPI_STATUS (AE_OK);
    }


    State = CmCreateUpdateState (Object, Action);

    while (State)
    {

        Object = State->Update.Object;
        Action = State->Update.Value;
        CmDeleteGenericState (State);

        /*
         * All sub-objects must have their reference count incremented also.
         * Different object types have different subobjects.
         */
        switch (Object->Common.Type)
        {

        case ACPI_TYPE_Device:

            //CmUpdateRefCount (Object->Device.AddrHandler, Action);
            Status = CmCreateUpdateStateAndPush (Object->Device.AddrHandler, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            CmUpdateRefCount (Object->Device.SysHandler, Action);
            CmUpdateRefCount (Object->Device.DrvHandler, Action);
            break;


        case INTERNAL_TYPE_AddressHandler:

            /* Must walk list of address handlers */

            Next = Object->AddrHandler.Link;
            while (Next)
            {
                New = Next->AddrHandler.Link;
                CmUpdateRefCount (Next, Action);

                Next = New;
            }
            break;


        case ACPI_TYPE_Package:

            /* 
             * We must update all the sub-objects of the package
             * (Each of whom may have their own sub-objects, etc.
             */
            for (i = 0; i < Object->Package.Count; i++)
            {
                /* 
                 * Push each element onto the stack for later processing.
                 * Note: There can be null elements within the package,
                 * these are simply ignored
                 */

                Status = CmCreateUpdateStateAndPush (Object->Package.Elements[i], Action, &StateList);
                if (ACPI_FAILURE (Status))
                {
                    return_ACPI_STATUS (Status);
                }
            }
            break;


        case ACPI_TYPE_FieldUnit:

            Status = CmCreateUpdateStateAndPush (Object->FieldUnit.Container, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }
            break;


        case INTERNAL_TYPE_DefField:

            Status = CmCreateUpdateStateAndPush (Object->Field.Container, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }
           break;


        case INTERNAL_TYPE_BankField:

            Status = CmCreateUpdateStateAndPush (Object->BankField.BankSelect, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            Status = CmCreateUpdateStateAndPush (Object->BankField.Container, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }
            break;


        case ACPI_TYPE_Region:

            CmUpdateRefCount (Object->Region.Method, Action);
    /* TBD:
            CmUpdateRefCount (Object->Region.AddrHandler, Action);
    */
/*
            Status = CmCreateUpdateStateAndPush (Object->Region.AddrHandler, Action, &StateList);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }
*/
            break;


        case INTERNAL_TYPE_Reference:

            break;
        }


        /*
         * Now we can update the count in the main object.  This can only happen after
         * we update the sub-objects in case this causes the main object to be deleted.
         */

        CmUpdateRefCount (Object, Action);


        /* Move on to the next object to be updated */

        State = CmPopGenericState (&StateList);
    }



    return_ACPI_STATUS (AE_OK);
}



void
CmAddReference (
    ACPI_OBJECT_INTERNAL    *Object)
{

    FUNCTION_TRACE_PTR ("CmAddReference", Object);


    /*
     * Ensure that we have a valid object
     */

    if (!CmValidInternalObject (Object))
    {
        return_VOID;
    }


    /*
     * We have a valid ACPI internal object, now increment the reference count
     */

    CmUpdateObjectReference  (Object, REF_INCREMENT);

    return_VOID;
}



void
CmRemoveReference (
    ACPI_OBJECT_INTERNAL    *Object)
{

    FUNCTION_TRACE_PTR ("CmRemoveReference", Object);


    /*
     * Ensure that we have a valid object
     */

    if (!CmValidInternalObject (Object))
    {
        return_VOID;
    }

    DEBUG_PRINT (ACPI_INFO, ("CmRemoveReference: Obj %p Refs=%d\n", 
                                Object, Object->Common.ReferenceCount));

    /*
     * Decrement the reference count, and only actually delete the object
     * if the reference count becomes 0.  (Must also decrement the ref count
     * of all subobjects!)
     */

    CmUpdateObjectReference  (Object, REF_DECREMENT);

    /*
     * If the reference count has reached zero,
     * delete the object and all sub-objects contained within it
     */
/*
    if (Object->Common.ReferenceCount == 0)
    {
        CmDeleteInternalObj (Object);
    }
*/
    return_VOID;
}


