 
/*
 *  ./acpica/subsystem/common/cmsize.c - Object size routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMSIZE_C__

#include <acpi.h>
#include <acobject.h>
#include <amlcode.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmsize");



ACPI_STATUS
CmGetSimpleObjectSize (
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    UINT32                  *ObjLength)
{
    UINT32                  Length;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("CmGetSimpleObjectSize", InternalObj);


    Length = sizeof (ACPI_OBJECT);

    if (VALID_DESCRIPTOR_TYPE (InternalObj, DESC_TYPE_NTE))
    {
        *ObjLength = Length;
        return_ACPI_STATUS (Status);
    }

    switch (InternalObj->Common.Type)
    {

    case ACPI_TYPE_String:

        Length += InternalObj->String.Length;
        break;


    case ACPI_TYPE_Buffer:

        Length += InternalObj->Buffer.Length;
        break;

    case ACPI_TYPE_Number:
    case ACPI_TYPE_Processor:
    case ACPI_TYPE_Power:
        /*
         * These are included in the object itself
         */
        break;


    case INTERNAL_TYPE_Reference:

        /* 
         * The only type that should be here is opcode AML_NAMEPATH_OP -- since
         * this means an object reference 
         */
        if (InternalObj->Reference.OpCode && InternalObj->Reference.OpCode != AML_NAMEPATH_OP)
        {
            DEBUG_PRINT (ACPI_ERROR, ("CmGetSimpleObjectSize: Unsupported Reference opcode=0x%X in object %p\n",
                                        InternalObj->Reference.OpCode, InternalObj));
            Status = AE_TYPE;
        }

        break;

    default:
        DEBUG_PRINT (ACPI_ERROR, ("CmGetSimpleObjectSize: Unsupported type=0x%X in object %p\n",
                                    InternalObj->Common.Type, InternalObj));
        Status = AE_TYPE;
        break;
    }


    *ObjLength = Length;
   
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
CmGetPackageObjectSize (
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    UINT32                  *ObjLength)
{

    ACPI_OBJECT_INTERNAL    *ThisInternalObj;
    ACPI_OBJECT_INTERNAL    *ParentObj[MAX_PACKAGE_DEPTH] = { 0,0,0,0,0 };
    ACPI_OBJECT_INTERNAL    *ThisParent;
    UINT32                  ThisIndex;
    UINT32                  Index[MAX_PACKAGE_DEPTH] = { 0,0,0,0,0 };
    UINT32                  Length = 0;
    UINT32                  ObjectSpace;
    UINT32                  CurrentDepth = 0;
    UINT32                  PackageCount = 1;
    ACPI_STATUS             Status;
    

    FUNCTION_TRACE_PTR ("CmGetPackageObjectSize", InternalObj);


    ParentObj[0] = InternalObj;

    while (1)
    {
        ThisParent      = ParentObj[CurrentDepth];
        ThisIndex       = Index[CurrentDepth];
        ThisInternalObj = ThisParent->Package.Elements[ThisIndex];

        /* 
         * Check for an unitialized package element.  It is completely
         * legal to declare a package and leave it uninitialized
         */

        if (!ThisInternalObj)
        {
            /* TBD: the real question is: what to do with a null entry, both here and when we copy the package */

            DEBUG_PRINT (ACPI_INFO, ("GetPackageSize: Found a null package element in %p\n", ThisParent));
            return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
        }

        if (IS_THIS_OBJECT_TYPE (ThisInternalObj, ACPI_TYPE_Package))
        {
            /*
             * If this object is a package then we go one deeper
             */
            PackageCount++;
            if (CurrentDepth < MAX_PACKAGE_DEPTH-1)
            {
                CurrentDepth++;
                ParentObj[CurrentDepth] = ThisInternalObj;
                Index[CurrentDepth]     = 0;
            }

            else
            {
                /*
                 * Too many nested levels of packages for us to handle
                 */
                
                DEBUG_PRINT (ACPI_ERROR, ("CmGetPackageObjectSize: Pkg nested too deep (max %d)\n",
                                            MAX_PACKAGE_DEPTH));
                return_ACPI_STATUS (AE_LIMIT);
            }
        }   /* if object is a package */

        else
        {
            Status = CmGetSimpleObjectSize (ThisInternalObj, &ObjectSpace);
            if (Status != AE_OK) 
            {
                return_ACPI_STATUS (Status);
            }

            Length += ObjectSpace;

            Index[CurrentDepth]++;
            while (Index[CurrentDepth] >= ParentObj[CurrentDepth]->Package.Count)
            {
                /*
                 * We've handled all of the objects at this level,  This means that we
                 * have just completed a package.  That package may have contained one
                 * or more packages itself.
                 */
                if (CurrentDepth == 0)
                {
                    /*
                     * We have handled all of the objects in the top level package
                     * just add the length of the package objects and get out.
                     */
                    Length += sizeof (ACPI_OBJECT) * PackageCount;
                    *ObjLength = Length;
                    
                    return_ACPI_STATUS (AE_OK);
                }

                /*
                 * Go back up a level and move the index past the just completed
                 * package object.
                 */
                CurrentDepth--;
                Index[CurrentDepth]++;
            }
        }   /* else object is NOT a package */
    }   /* while (1)  */

    /*
     * We'll never get here, but the compiler whines about return value
     */
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
CmGetObjectSize(
    ACPI_OBJECT_INTERNAL    *InternalObj, 
    UINT32                  *ObjLength)
{
    ACPI_STATUS             Status;


    if ((VALID_DESCRIPTOR_TYPE (InternalObj, DESC_TYPE_ACPI_OBJ)) &&
        (IS_THIS_OBJECT_TYPE (InternalObj, ACPI_TYPE_Package)))
    {
        Status = CmGetPackageObjectSize (InternalObj, ObjLength);
    }

    else
    {
        Status = CmGetSimpleObjectSize (InternalObj, ObjLength);
    }

    return Status;
}




