 
/*
 *  ./acpica/subsystem/common/cmstate.c - Generic "state" data struct manipulation
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __CMSTATE_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <parser.h>

#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("cmstate");




ACPI_STATUS
CmCreateUpdateStateAndPush (
    ACPI_OBJECT_INTERNAL    *Object,
    UINT16                  Action,
    ACPI_GENERIC_STATE      **StateList)
{
    ACPI_GENERIC_STATE       *State;


    /* Ignore null objects; these are expected */

    if (!Object)
    {
        return AE_OK;
    }

    State = CmCreateUpdateState (Object, Action);
    if (!State)
    {
        return AE_NO_MEMORY;
    }


    CmPushGenericState (StateList, State);
    return AE_OK;
}
 


void
CmPushGenericState (
    ACPI_GENERIC_STATE      **ListHead,
    ACPI_GENERIC_STATE      *State)
{
    FUNCTION_TRACE ("CmPushGenericState");
    
    /* Push the state object onto the front of the list (stack) */

    State->Common.Next = *ListHead;
    *ListHead = State;

    return_VOID;
}



ACPI_GENERIC_STATE *
CmPopGenericState (
    ACPI_GENERIC_STATE      **ListHead)
{
    ACPI_GENERIC_STATE      *State;

    
    FUNCTION_TRACE ("DsPopGenericState");


    /* Remove the state object at the head of the list (stack) */

    State = *ListHead;
    if (State)
    {
        /* Update the list head */

        *ListHead = State->Common.Next;
    }

    return_PTR (State);
}



ACPI_GENERIC_STATE *
CmCreateGenericState (void)
{
    ACPI_GENERIC_STATE      *State;


    CmAcquireMutex (MTX_CACHES);

    Gbl_StateCacheRequests++;

    /* Check the cache first */

    if (Gbl_GenericStateCache)
    {
        /* There is an object available, use it */

        State = Gbl_GenericStateCache;
        Gbl_GenericStateCache = State->Common.Next;
        State->Common.Next = NULL;

        Gbl_StateCacheHits++;
        Gbl_GenericStateCacheDepth--;

        CmReleaseMutex (MTX_CACHES);
    }

    else
    {
        /* The cache is empty, create a new object */

        CmReleaseMutex (MTX_CACHES);

        State = CmCallocate (sizeof (ACPI_GENERIC_STATE));
    }

    /* Initialize */

    if (State)
    {
        State->Common.DataType = DESC_TYPE_STATE;
    }

    return State;
}



ACPI_GENERIC_STATE *
CmCreateUpdateState (
    ACPI_OBJECT_INTERNAL    *Object,
    UINT16                  Action)
{
    ACPI_GENERIC_STATE      *State;


    FUNCTION_TRACE_PTR ("CmCreateUpdateState", Object);


    /* Create the generic state object */

    State = CmCreateGenericState ();
    if (!State)
    {
        return NULL;
    }

    /* Init fields specific to the update struct */

    State->Update.Object = Object;
    State->Update.Value  = Action;

    return_PTR (State);
}



ACPI_GENERIC_STATE *
CmCreateControlState (
    void)
{
    ACPI_GENERIC_STATE      *State;


    FUNCTION_TRACE ("CmCreateControlState");

    /* Create the generic state object */

    State = CmCreateGenericState ();
    if (!State)
    {
        return NULL;
    }


    /* Init fields specific to the control struct */

    State->Common.State = CONTROL_CONDITIONAL_EXECUTING;

    return_PTR (State);
}



void
CmDeleteGenericState (
    ACPI_GENERIC_STATE      *State)
{
    FUNCTION_TRACE ("CmDeleteGenericState");


    /* If cache is full, just free this state object */

    if (Gbl_GenericStateCacheDepth >= MAX_STATE_CACHE_DEPTH)
    {
        CmFree (State);
    }

    /* Otherwise put this object back into the cache */

    else
    {
        CmAcquireMutex (MTX_CACHES);

        /* Clear the state */

        MEMSET (State, 0, sizeof (ACPI_GENERIC_STATE));
        State->Common.DataType = DESC_TYPE_STATE;

        /* Put the object at the head of the global cache list */

        State->Common.Next = Gbl_GenericStateCache;
        Gbl_GenericStateCache = State;
        Gbl_GenericStateCacheDepth++;


        CmReleaseMutex (MTX_CACHES);
    }
    return_VOID;
}



void
CmDeleteGenericStateCache (
    void)
{
    ACPI_GENERIC_STATE      *Next;


    FUNCTION_TRACE ("CmDeleteGenericStateCache");


    /* Traverse the global cache list */

    while (Gbl_GenericStateCache)
    {
        /* Delete one cached state object */

        Next = Gbl_GenericStateCache->Common.Next;
        CmFree (Gbl_GenericStateCache);
        Gbl_GenericStateCache = Next;
    }

    return_VOID;
}
 