/*
 *  ./acpica/subsystem/debugger/dbapi.c - AML Debugger external interfaces
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>
#include <events.h>
#include <interp.h>
#include <debugger.h>


#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbcmds");



ACPI_STATUS
DbSingleStep (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    UINT8                   OpType)
{
    ACPI_GENERIC_OP         *Next;
    ACPI_STATUS             Status = AE_OK;
    UINT32                  OriginalDebugLevel;


    /* Is there a breakpoint set? */

    if (Gbl_MethodBreakpoint)
    {
        /* Check if the breakpoint has been reached or passed */

        if ((Gbl_BreakpointWalk == WalkState) &&
            (Gbl_MethodBreakpoint <= Op->AmlOffset))
        {
            /* Hit the breakpoint, resume single step, reset breakpoint */

            OsdPrintf ("***Break*** at AML offset 0x%X\n", Op->AmlOffset);
            Gbl_CmSingleStep = TRUE;
            Gbl_StepToNextCall = FALSE;
            Gbl_MethodBreakpoint = 0;
        }
    }


    /* 
     * Check if this is an opcode that we are interested in --
     * namely, opcodes that have arguments
     */

    if (Op->Opcode == AML_NAMEDFIELD_OP)
    {
        return AE_OK;
    }

    switch (OpType)
    {
    case OPTYPE_UNDEFINED:
    case OPTYPE_CONSTANT:           /* argument type only */
    case OPTYPE_LITERAL:            /* argument type only */
    case OPTYPE_DATA_TERM:          /* argument type only */
    case OPTYPE_LOCAL_VARIABLE:     /* argument type only */
    case OPTYPE_METHOD_ARGUMENT:    /* argument type only */
        return (AE_OK);
        break;

    case OPTYPE_NAMED_OBJECT:
        switch (Op->Opcode)
        {
        case AML_NAMEPATH_OP:
            return (AE_OK);
            break;
        }
    }


    /*
     * Under certain debug conditions, display this opcode and its operands
     */

    if ((OutputToFile)                  ||
        (Gbl_CmSingleStep)              ||
        (DebugLevel & TRACE_PARSE))
    {
        if ((OutputToFile)                  ||
            (DebugLevel & TRACE_PARSE))
        {
            OsdPrintf ("\n[AmlDebug] Next AML Opcode to execute:\n");
        }

        /* 
         * Display this op (and only this op - zero out the NEXT field temporarily,
         * and disable parser trace output for the duration of the display because
         * we don't want the extraneous debug output)
         */

        OriginalDebugLevel = DebugLevel;
        DebugLevel &= ~(TRACE_PARSE | TRACE_FUNCTIONS);
        Next = Op->Next;
        Op->Next = NULL;

        /* Now we can display it */

        DbDisplayOp (Op, ACPI_UINT32_MAX);

        /* Restore everything */

        Op->Next = Next;
        OsdPrintf ("\n");
        DebugLevel = OriginalDebugLevel;
   }


    /* If we are not single stepping, just continue executing the method */

    if (!Gbl_CmSingleStep)
    {
        return (AE_OK);
    }


    /*
     * If we are executing a step-to-call command, 
     * Check if this is a method call.
     */

    if (Gbl_StepToNextCall)
    {
        if (Op->Opcode != AML_METHODCALL_OP)
        {
            /* Not a method call, just keep executing */

            return (AE_OK);
        }

        /* Found a method call, stop executing */

        Gbl_StepToNextCall = FALSE;
    }


    /*
     * If the next opcode is a method call, we will "step over" it
     * by default.
     */

    if (Op->Opcode == AML_METHODCALL_OP)
    {
        Gbl_CmSingleStep = FALSE;  /* No more single step while executing called method */

        /* Set the breakpoint on the call, it will stop execution as soon as we return */

        /* TBD: don't kill the user breakpoint! */

        Gbl_MethodBreakpoint = Op->AmlOffset + 1;  /* Must be non-zero! */
        Gbl_BreakpointWalk = WalkState;
    }



    CmReleaseMutex (MTX_NAMESPACE);

    /* Go into the command loop and await next user command */
    
    Gbl_MethodExecuting = TRUE;
    Status = AE_CTRL_TRUE;
    while (Status == AE_CTRL_TRUE)
    {
        if (Gbl_DebuggerConfiguration == DEBUGGER_MULTI_THREADED)
        {
            /* Handshake with the front-end that gets user command lines */

            CmReleaseMutex (MTX_DEBUG_CMD_COMPLETE);
            CmAcquireMutex (MTX_DEBUG_CMD_READY);
        }

        else
        {
            /* Single threaded, we must get a command line ourselves */

            /* Force output to console until a command is entered */

            DbSetOutputDestination (DB_CONSOLE_OUTPUT);

            /* Different prompt if method is executing */

            if (!Gbl_MethodExecuting)
            {
                OsdPrintf ("%1c ", DB_COMMAND_PROMPT);
            }
            else
            {
                OsdPrintf ("%1c ", DB_EXECUTE_PROMPT);
            }

            /* Get the user input line */

            OsdGetLine (LineBuf);
        }

        Status = DbCommandDispatch (LineBuf, WalkState, Op);
    }

    CmAcquireMutex (MTX_NAMESPACE);


    /* User commands complete, continue execution of the interrupted method */

    return Status;
}



int
DbInitialize (void)
{      


    /* Init globals */

    Buffer = OsdAllocate (BUFFER_SIZE);

    /* Initial scope is the root */

    ScopeBuf [0] = '\\';
    ScopeBuf [1] =  0;


    /*
     * If configured for multi-thread support, the debug executor runs in
     * a separate thread so that the front end can be in another address
     * space, environment, or even another machine.
     */

    if (Gbl_DebuggerConfiguration & DEBUGGER_MULTI_THREADED)
    {
        /* These were created with one unit, grab it */

        CmAcquireMutex (MTX_DEBUG_CMD_COMPLETE);
        CmAcquireMutex (MTX_DEBUG_CMD_READY);

        /* Create the debug execution thread to execute commands */

        OsdQueueForExecution (0, DbExecuteThread, NULL);
    }

 if (!opt_verbose)
 {
  INDENT_STRING = "    ";
        opt_disasm = TRUE;
        opt_stats = FALSE;
 }


    return 0;
}


#endif /* ACPI_DEBUG */
