/*
 *  ./acpica/subsystem/debugger/dbcmds.c - debug commands and output routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>
#include <events.h>
#include <interp.h>
#include <debugger.h>
#include <tables.h>

#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbcmds");



/* These object types map directly to the ACPI_TYPES */


ARGUMENT_INFO               DbObjectTypes [] = 
{
    "ANY",
    "NUMBERS",
    "STRINGS",
    "BUFFERS",
    "PACKAGES",
    "FIELDS",
    "DEVICES",
    "EVENTS",
    "METHODS",
    "MUTEXES",
    "REGIONS",
    "POWERRESOURCES",
    "PROCESSORS",
    "THERMALZONES",
    "BUFFERFIELDS",
    "DDBHANDLES",
    NULL            /* Must be null terminated */
};



void
DbDisplayTableInfo (
    char                    *TableArg)
{
    UINT32                  i;



    for (i = 0; i < NUM_ACPI_TABLES; i++)
    {
        if (Gbl_AcpiTables[i].Pointer)
        {
            OsdPrintf ("%s at 0x%p length 0x%X\n", Gbl_AcpiTableData[i].Name, 
                        Gbl_AcpiTables[i].Pointer, Gbl_AcpiTables[i].Length);
        }
    }
}



void
DbUnloadAcpiTable (
    char                    *TableArg,
    char                    *InstanceArg)
{
    UINT32                  i;
    ACPI_STATUS             Status;



    for (i = 0; i < NUM_ACPI_TABLES; i++)
    {
        if (!STRNCMP (TableArg, Gbl_AcpiTableData[i].Signature, Gbl_AcpiTableData[i].SigLength))
        {
            Status = AcpiUnloadTable (i);
            if (ACPI_SUCCESS (Status))
            {
                OsdPrintf ("[%s] unloaded and uninstalled\n", TableArg);
            }
            else
            {
                OsdPrintf ("%s, while unloading [%s]\n", CmFormatException (Status), TableArg);
            }

            return;
        }
    }

    OsdPrintf ("Unknown table type [%s]\n", TableArg);
}



void
DbSetMethodBreakpoint (
    char                    *Location,
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    UINT32                  Address;


    if (!Op)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    /* Get and verify the breakpoint address */

    Address = STRTOUL (Location, NULL, 16);
    if (Address <= Op->AmlOffset)
    {
        OsdPrintf ("Breakpoint 0x%X is beyond current address 0x%X\n", Address, Op->AmlOffset);
    }

    /* Just save the breakpoint in a global */

    Gbl_MethodBreakpoint = Address;
    Gbl_BreakpointWalk = WalkState;
    OsdPrintf ("Breakpoint set at AML offset 0x%X\n", Address);
}



void
DbSetMethodCallBreakpoint (
    ACPI_GENERIC_OP         *Op)
{


    if (!Op)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }


    Gbl_StepToNextCall = TRUE;
}




void
DbDisassembleAml (
    char                    *Statements,
    ACPI_GENERIC_OP         *Op)
{
    UINT32                  NumStatements = 8;


    if (!Op)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    if (Statements)
    {
        NumStatements = STRTOUL (Statements, NULL, 0);
    }


    DbDisplayOp (Op, NumStatements);
}




void
DbDumpNamespace (
    char                    *StartArg,
    char                    *DepthArg)
{
    ACPI_HANDLE             SubtreeEntry = Gbl_RootObject;
    UINT32                  MaxDepth = ACPI_UINT32_MAX;



    /* No argument given, just start at the root and dump entire namespace */

    if (StartArg)
    {

        /* Check if numeric argument, must be an NTE */

        if ((StartArg[0] >= 0x30) && (StartArg[0] <= 0x39))
        {
            SubtreeEntry = (ACPI_HANDLE) STRTOUL (StartArg, NULL, 16);
            if (!OsdReadable (SubtreeEntry, sizeof (NAME_TABLE_ENTRY)))
            {
                OsdPrintf ("Address %p is invalid in this address space\n", SubtreeEntry);
                return;
            }
            if (!VALID_DESCRIPTOR_TYPE ((SubtreeEntry), DESC_TYPE_NTE))
            {
                OsdPrintf ("Address %p is not a valid NTE\n", SubtreeEntry);
                return;
            }
        }

        /* Alpha argument */

        else
        {

            /* The parameter is a name string that must be resolved to an NTE */

            SubtreeEntry = DbLocalNsLookup (StartArg);
            if (!SubtreeEntry)
            {
                SubtreeEntry = Gbl_RootObject;
            }
        }


        /* Now we can check for the depth argument */

        if (DepthArg)
        {
            MaxDepth = STRTOUL (DepthArg, NULL, 0);
        }
    }


    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);
    OsdPrintf ("ACPI Namespace (from %p subtree):\n", SubtreeEntry);

    /* Display the subtree */

    DbSetOutputDestination (DB_REDIRECTABLE_OUTPUT);
    NsDumpObjects (ACPI_TYPE_Any, MaxDepth, ACPI_UINT32_MAX, SubtreeEntry);
    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
}



void
DbDumpNamespaceByOwner (
    char                    *OwnerArg,
    char                    *DepthArg)
{
    ACPI_HANDLE             SubtreeEntry = Gbl_RootObject;
    UINT32                  MaxDepth = ACPI_UINT32_MAX;
    UINT16                  OwnerId;



    OwnerId = (UINT16) STRTOUL (OwnerArg, NULL, 0);


    /* Now we can check for the depth argument */

    if (DepthArg)
    {
        MaxDepth = STRTOUL (DepthArg, NULL, 0);
    }


    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);
    OsdPrintf ("ACPI Namespace by owner 0x%X:\n", OwnerId);

    /* Display the subtree */

    DbSetOutputDestination (DB_REDIRECTABLE_OUTPUT);
    NsDumpObjects (ACPI_TYPE_Any, MaxDepth, OwnerId, SubtreeEntry);
    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
}



void
DbSendNotify (
    char                    *Name,
    UINT32                  Value)
{
    NAME_TABLE_ENTRY        *Entry;



    /* Translate name to an NTE */

    Entry = DbLocalNsLookup (Name);
    if (!Entry)
    {
        return;
    }

    /* Decode NTE type */

    switch (Entry->Type)
    {
    case ACPI_TYPE_Device:
    case ACPI_TYPE_Thermal:
            
         /* Send the notify */

        EvNotifyDispatch (Entry, Value);
        break;

    default:
        OsdPrintf ("Named object is not a device or a thermal object\n");
        break;
    }

}



void
DbSetMethodData (
    char                    *TypeArg,
    char                    *IndexArg,
    char                    *ValueArg)
{
    char                    Type;
    UINT32                  Index;
    UINT32                  Value;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;


    STRUPR (TypeArg);
    Type = TypeArg[0];
    if ((Type != 'L') &&
        (Type != 'A'))
    {
        OsdPrintf ("Invalid SET operand: %s\n", TypeArg);
        return;
    }

    Index = STRTOUL (IndexArg, NULL, 16);
    Value = STRTOUL (ValueArg, NULL, 16);

    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }



    ObjDesc = CmCreateInternalObject (ACPI_TYPE_Number);
    if (!ObjDesc)
    {
        OsdPrintf ("Could not create an internal object\n");
        return;
    }

    ObjDesc->Number.Value = Value;

    switch (Type)
    {
    case 'A':
        
        /* Set a method argument */

        if (Index > MTH_NUM_ARGS)
        {
            OsdPrintf ("Arg%d - Invalid argument name\n", Index);
            return;
        }

        DsMethodDataSetValue (MTH_TYPE_ARG, Index, ObjDesc);
        ObjDesc = WalkState->Arguments[Index].Object;
        OsdPrintf ("Arg%d: ", Index);
        DbDisplayInternalObject (ObjDesc);
        break;

    case 'L':

        /* Set a method local */

        if (Index > MTH_NUM_LOCALS)
        {
            OsdPrintf ("Local%d - Invalid local variable name\n", Index);
            return;
        }

        DsMethodDataSetValue (MTH_TYPE_LOCAL, Index, ObjDesc);
        ObjDesc = WalkState->LocalVariables[Index].Object;
        OsdPrintf ("Local%d: ", Index);
        DbDisplayInternalObject (ObjDesc);
        break;

    default:
        break;
    }
}



ACPI_STATUS
DbWalkForSpecificObjects (
    ACPI_HANDLE             ObjHandle,
    UINT32                  NestingLevel,
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;
    UINT32                  BufSize;
    char                    buffer[64];


    ObjDesc = ((NAME_TABLE_ENTRY *)ObjHandle)->Object;

    /* Get the full pathname to this method */

    Status = NsHandleToPathname (ObjHandle, &BufSize, buffer);

    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Could Not get pathname for object %p\n", ObjHandle);
        return AE_OK;
    }

    OsdPrintf ("%32s", buffer);

    if (ObjDesc)
    {
        switch (ObjDesc->Common.Type)
        {
        case ACPI_TYPE_Method:
            OsdPrintf ("  #Args %d  Concurrency %d", ObjDesc->Method.ParamCount, ObjDesc->Method.Concurrency);
            break;

        case ACPI_TYPE_Number:
            OsdPrintf ("  Value 0x%X", ObjDesc->Number.Value);
            break;
    
        case ACPI_TYPE_String:
            OsdPrintf ("  \"%s\"", ObjDesc->String.Pointer);
            break;

        case ACPI_TYPE_Region:
            OsdPrintf ("  SpaceId %d Address %X Length %X", ObjDesc->Region.SpaceId, ObjDesc->Region.Address, ObjDesc->Region.Length);
            break;

        case ACPI_TYPE_Package:
            OsdPrintf ("  #Elements %d", ObjDesc->Package.Count);
            break;

        case ACPI_TYPE_Buffer:
            OsdPrintf ("  Length %d", ObjDesc->Buffer.Length);
            break;
        }
    }

    OsdPrintf ("\n");
    return AE_OK;
}



ACPI_STATUS
DbDisplayObjects (
    char                    *ObjTypeArg,
    char                    *DisplayCountArg)
{
    UINT32                  DisplayCount;
    UINT32                  Type;

    
    STRUPR (ObjTypeArg);
    Type = DbMatchArgument (ObjTypeArg, DbObjectTypes);
    if (Type == (UINT32) -1)
    {
        OsdPrintf ("Invalid or unsupported argument\n");
        return AE_OK;
    }


    if (DisplayCountArg)
    {
        DisplayCount = STRTOUL (DisplayCountArg, NULL, 0);
    }
    
    else
    {
        DisplayCount = ACPI_UINT32_MAX;
    }

    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);
    OsdPrintf ("Objects of type [%s] defined in the current ACPI Namespace: \n", CmGetTypeName (Type));

    DbSetOutputDestination (DB_REDIRECTABLE_OUTPUT);

    AcpiWalkNamespace (Type, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX,
                        DbWalkForSpecificObjects, (void *) Type, NULL);

    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
    return AE_OK;
}



ACPI_STATUS
DbWalkAndMatchName (
    ACPI_HANDLE             ObjHandle,
    UINT32                  NestingLevel,
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;
    char                    *RequestedName = (char *) Context;
    UINT32                  i;
    UINT32                  BufSize;
    char                    Buffer[96];


    ObjDesc = ((NAME_TABLE_ENTRY *)ObjHandle)->Object;


    /* Check for a name match */

    for (i = 0; i < 4; i++)
    {
        /* Wildcard support */

        if ((RequestedName[i] != '?') &&
            (RequestedName[i] != ((char *)(&((NAME_TABLE_ENTRY *)ObjHandle)->Name))[i]))
        {
            /* No match, just exit */

            return AE_OK;
        }
    }


    /* Get the full pathname to this object */

    BufSize = sizeof (Buffer);

    Status = NsHandleToPathname (ObjHandle, &BufSize, Buffer);
    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Could Not get pathname for object %p\n", ObjHandle);
    }

    else
    {
        OsdPrintf ("%32s (0x%p) - %s\n", Buffer, ObjHandle, CmGetTypeName (((NAME_TABLE_ENTRY *)ObjHandle)->Type));
    }

    return AE_OK;
}



ACPI_STATUS
DbFindNameInNamespace (
    char                    *NameArg)
{
    
    if (STRLEN (NameArg) > 4)
    {
        OsdPrintf ("Name must be no longer than 4 characters\n");
        return (AE_OK);
    }

    AcpiWalkNamespace (ACPI_TYPE_Any, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX,
        DbWalkAndMatchName, NameArg, NULL);

    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
    return AE_OK;
}



void
DbSetScope (
    char                    *Name)
{

    if (!Name || Name[0] == 0)
    {
        OsdPrintf ("Current scope: %s\n", ScopeBuf);
        return;
    }

    DbPrepNamestring (Name);

    /* TBD: Validate scope here */

    if (Name[0] == '\\')
    {
        STRCPY (ScopeBuf, Name);
        STRCAT (ScopeBuf, "\\");
    }

    else
    {
        STRCAT (ScopeBuf, Name);
        STRCAT (ScopeBuf, "\\");
    }
  
    OsdPrintf ("New scope: %s\n", ScopeBuf);
}

#endif
