/*
 *  ./acpica/subsystem/debugger/dbdisasm.c - parser op tree display routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <namesp.h>
#include <debugger.h>


#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbdisasm");


#define MAX_SHOW_ENTRY      128


char      *INDENT_STRING = "....";




#define BLOCK_PAREN 1
#define BLOCK_BRACE 2

INT32
DbBlockType (
    ACPI_GENERIC_OP *Op)
{

    switch (Op->Opcode)
    {
    case AML_MethodOp:
        return BLOCK_BRACE;
        break;

    default:
        break;
    }

    return BLOCK_PAREN;

}

#ifdef PARSER_ONLY

ACPI_STATUS
PsDisplayObjectPathname (
    ACPI_GENERIC_OP         *Op,
    char                    *Buffer,
    UINT32                  BufferSize)
{
    ACPI_GENERIC_OP         *TargetOp;


    OsdPrintf ("  (Path ");

    /* Search parent tree up to the root if necessary */

    TargetOp = PsFind (Op, Op->Value.Name, 0, 0);
    if (!TargetOp)
    {
        OsdPrintf ("****Not Found****");
        return AE_NOT_FOUND;
    }

    /* If target was found, print the name and complete path */

    DbSprintPath (Buffer, sizeof (Buffer), TargetOp);
    OsdPrintf (")");
    return AE_OK;
}

#else

ACPI_STATUS
PsDisplayObjectPathname (
    ACPI_GENERIC_OP         *Op,
    char                    *Buffer,
    UINT32                  BufferSize)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *Nte;


    OsdPrintf ("  (Path ");

    /* Just get the NTE out of the Op object */

    Nte = Op->NameTableEntry;
    if (!Nte)
    {
        STRCPY (Buffer, "****No NTE, could not get pathname****");
        return AE_OK;
    }

    /* Convert NTE/handle to a full pathname */

    Status = NsHandleToPathname (Nte, &BufferSize, Buffer);
    if (ACPI_FAILURE (Status))
    {
       return Status;
    }

    OsdPrintf ("%s)", Buffer);
    return AE_OK;
}

#endif


#define DB_NO_OP_INFO       "            [%2.2d]  "
#define DB_FULL_OP_INFO     "%5.5X #%4.4X [%2.2d]  "

void
DbDisplayOp (
    ACPI_GENERIC_OP         *Origin,
    UINT32                  NumOpcodes)
{
    static char             Buffer[MAX_SHOW_ENTRY];
    ACPI_GENERIC_OP         *Op = Origin;
    ACPI_GENERIC_OP         *arg;
    ACPI_GENERIC_OP         *depth;
    UINT32                  DepthCount = 0;
    UINT32                  LastDepth = 0;
    UINT32                  i;
    UINT32                  j;


    if (Op)
    {
        while (Op)
        {
   /* indentation */

   DepthCount = 0;
            if (!opt_verbose)
            {
                DepthCount++;
            }

            /* Determine the nesting depth of this argument */

   for (depth = Op->Parent; depth; depth = depth->Parent)
   {
    arg = PsGetArg (depth, 0);
    while (arg && arg != Origin)
    {
     arg = arg->Next;
    }

    if (arg)
    {
     break;
    }

    DepthCount++;
   }
            

            /* Open a new block if we are nested further than last time */

   if (DepthCount > LastDepth)
   {
    VERBOSE_PRINT ((DB_NO_OP_INFO, LastDepth));
    for (i = 0; i < LastDepth; i++)
    {
     OsdPrintf (INDENT_STRING);
    }

                if (DbBlockType (Op) == BLOCK_PAREN)
        OsdPrintf ("(\n");
                else
        OsdPrintf ("{\n");
   }

            /* Close a block if we are nested less than last time */

   else if (DepthCount < LastDepth)
   {
    for (j = 0; j < (LastDepth - DepthCount); j++)
    {
                    VERBOSE_PRINT ((DB_NO_OP_INFO, LastDepth - j));
     for (i = 0; i < (LastDepth - j - 1); i++)
     {
      OsdPrintf (INDENT_STRING);
     }

                    if (DbBlockType (Op) == BLOCK_PAREN)
         OsdPrintf (")\n");
                    else
         OsdPrintf ("}\n");
    }
   }

            /* In verbose mode, print the AML offset, opcode and depth count */

   VERBOSE_PRINT ((DB_FULL_OP_INFO, (unsigned) Op->AmlOffset, Op->Opcode, DepthCount));


            /* Indent the output according to the depth count */

            for (i = 0; i < DepthCount; i++)
            {
                OsdPrintf (INDENT_STRING);
            }


            /* Now print the opcode */

            if (DbSprintOp (Buffer, sizeof (Buffer), Op) > 0)
            {
                /* Resolve a name reference */

                if (Op->Opcode == AML_NAMEPATH_OP && Op->Value.Name)
                {
     if (opt_verbose)
     {
                        PsDisplayObjectPathname (Op, Buffer, sizeof (Buffer));
                    }
                }

                OsdPrintf ("\n");
           }

            /* Get the next node in the tree */

            Op = PsGetDepthNext (Origin, Op);
            LastDepth = DepthCount;

            NumOpcodes--;
            if (!NumOpcodes)
            {
                Op = NULL;
            }
        }

        /* Close the last block(s) */

        DepthCount = LastDepth -1;
        for (i = 0; i < LastDepth; i++)
        {
   VERBOSE_PRINT ((DB_NO_OP_INFO, LastDepth - i));
   for (j = 0; j < DepthCount; j++)
   {
    OsdPrintf (INDENT_STRING);
   }
   OsdPrintf ("}\n");
            DepthCount--;
        }

    }

    else
    {
        if (DbSprintOp (Buffer, sizeof (Buffer), Op) > 0)
        {
            OsdPrintf ("%s\n", Buffer);
        }
    }
}



INT32
DbSprintNamestring (
    char                    *BufferStart, 
    INT32                   BufferSize, 
    UINT8                   *Name)
{
    INT32                   SegCount;
    BOOLEAN                 DoDot = FALSE;


    if (Name)
    {
        if (BufferSize < 2)
        {
            return -1;
        }

        if (PsIsPrefixChar (GET8 (Name)))
        {
            /* append prefix character */

            OsdPrintf ("%1c", GET8 (Name));
            Name++;
        }

        switch (GET8 (Name))
        {
        case AML_DualNamePrefix:
            SegCount = 2;
            Name++;
            break;

        case AML_MultiNamePrefixOp:
            SegCount = (INT32) GET8 (Name + 1);
            Name += 2;
            break;

        default:
            SegCount = 1;
            break;
        }

        while (SegCount--)
        {
            /* append Name segment */

            if (DoDot)
            {
                /* append dot */

                OsdPrintf (".");
            }

            OsdPrintf ("%4.4s", Name);
            DoDot = TRUE;

            Name += 4;
        }
    }

    else
    {
        OsdPrintf ("<NULL NAME PTR>");
    }
    
    return (1);
}



INT32
DbSprintPath (
    char                    *BufferStart,
    UINT32                  BufferSize,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_GENERIC_OP         *Prev;
    ACPI_GENERIC_OP         *Search;
    UINT32                  Name;
    BOOLEAN                 DoDot = FALSE;


    /* TBD: was simply PsIsNamedOp */

    if (PsIsNamedObjectOp (Op->Opcode))
    {
        Prev = NULL;
        while (Prev != Op)
        {
            /* find scope with "prev" as its parent */

            Search = Op;
            for (; ;)
            {
                if (Search->Parent == Prev)
                {
                    break;
                }
                Search = Search->Parent;
            }
                
            if (Prev && !PsIsFieldOp (Search->Opcode))
            {
                /* below root scope, append scope name */

                if (DoDot)
                {
                    /* append dot */

                    OsdPrintf (".");
                }

                Name = PsGetName (Search);
                OsdPrintf ("%4.4s", &Name);
                DoDot = TRUE;
            }
            
            Prev = Search;
        }
    }

    return (1);
}




INT32
DbSprintOp (
    char                    *BufferStart, 
    UINT32                  BufferSize, 
    ACPI_GENERIC_OP         *Op)
{
    UINT32                  Size = 0;
    UINT8                   *ByteData;
    UINT32                  ByteCount;
    UINT32                  i;
    char                    *Buffer = BufferStart;
    ACPI_OP_INFO            *Opc = NULL;
    UINT32                  Name;



    if (Op)
    {
        /* op and arguments */

        switch (Op->Opcode)
        {

        case AML_ByteOp:

   if (opt_verbose)
    OsdPrintf ("(UINT8)  0x%2.2X", Op->Value.Integer & 0xff);
   else
    OsdPrintf ("0x%2.2X", Op->Value.Integer & 0xff);

            break;


        case AML_WordOp:

   if (opt_verbose)
    OsdPrintf ("(UINT16) 0x%4.4X",
           Op->Value.Integer & 0xffff);
   else
    OsdPrintf ("0x%4.4X",
           Op->Value.Integer & 0xffff);
            break;


        case AML_DWordOp:

   if (opt_verbose)
    OsdPrintf ("(UINT32) 0x%8.8X",
           Op->Value.Integer);
   else
    OsdPrintf ("0x%8.8X",
           Op->Value.Integer);
            break;


        case AML_StringOp:

            if (Op->Value.String)
            {
                OsdPrintf ("\"%s\"", Op->Value.String);
            }
            else
            {
                OsdPrintf ("<\"NULL STRING PTR\">");
            }
            break;


        case AML_STATICSTRING_OP:

            if (Op->Value.String)
            {
                OsdPrintf ("\"%s\"", Op->Value.String);
            }
            else
            {
                OsdPrintf ("\"<NULL STATIC STRING PTR>\"");
            }
            break;


        case AML_NAMEPATH_OP:

            Size = DbSprintNamestring (Buffer, BufferSize, Op->Value.Buffer);
            break;


        case AML_NAMEDFIELD_OP:

            OsdPrintf (Buffer, BufferSize, 0, "NamedField    (Length 0x%8.8X)  ", Op->Value.Integer);
            break;


        case AML_RESERVEDFIELD_OP:

            OsdPrintf (Buffer, BufferSize, 0, "ReservedField (Length 0x%8.8X)  ", Op->Value.Integer);
            break;


        case AML_ACCESSFIELD_OP:

            OsdPrintf (Buffer, BufferSize, 0, "AccessField   (Length 0x%8.8X)  ", Op->Value.Integer);
            break;


        case AML_BYTELIST_OP:

            if (opt_verbose)
            {
                OsdPrintf ("ByteList      (Length 0x%8.8X)  ", Op->Value.Integer);
            }
            else
            {
                OsdPrintf ("0x%2.2X", Op->Value.Integer);

                ByteCount = Op->Value.Integer;
                ByteData = ((ACPI_BYTELIST_OP *) Op)->Data;

                for (i = 0; i < ByteCount; i++)
                {
                    OsdPrintf (", 0x%2.2X", ByteData[i]);
                }
            }
            break;


        default:

            Opc = PsGetOpcodeInfo (Op->Opcode);
            if (Opc)
            {
                DEBUG_EXEC ((OsdPrintf ("%s", Opc->Name)));
            }
            else
            {
                OsdPrintf ("<Opcode 0x%04x>", Op->Opcode);
            }

            break;
        }


        if (!Opc)
        {
            if (Op->Next)
            {
    OsdPrintf (",");
            }
        }

    }

    else
    {
        OsdPrintf ("<NULL OP PTR>");
    }

    if (Op && PsIsNamedOp (Op->Opcode))
    {
        Name = PsGetName (Op);
        OsdPrintf (" %4.4s", &Name);

  if (opt_verbose)
  {
   OsdPrintf ("  (Path \\");
   Size = DbSprintPath (Buffer, BufferSize, Op);
            OsdPrintf (")");
  }
        
    }

    return (1);
}


#endif  /* ACPI_DEBUG */

