/*
 *  ./acpica/subsystem/debugger/dbdisply.c - debug display commands
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <namesp.h>
#include <parser.h>
#include <events.h>
#include <interp.h>
#include <debugger.h>

#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbdisply");




void
DbDumpParserDescriptor (
    ACPI_GENERIC_OP         *Op)
{
    ACPI_OP_INFO            *Info;


    Info = PsGetOpcodeInfo (Op->Opcode);

    OsdPrintf ("Parser Op Descriptor:\n");
    OsdPrintf ("%20.20s : %4.4X\n", "Opcode", Op->Opcode);
    OsdPrintf ("%20.20s : %s\n", "Opcode Name", Info->Name);
    OsdPrintf ("%20.20s : %p\n", "Value/ArgList", Op->Value);

    OsdPrintf ("%20.20s : %p\n", "Parent", Op->Parent);
    OsdPrintf ("%20.20s : %p\n", "NextOp", Op->Next);
}



void
DbDecodeAndDisplayObject (
    char                    *Target,
    char                    *OutputType)
{
    void                    *ObjPtr;
    NAME_TABLE_ENTRY        *Entry;
    UINT32                  Display = DB_BYTE_DISPLAY;
    char                    Buffer[80];
    ACPI_BUFFER             RetBuf;
    ACPI_STATUS             Status;
    UINT32                  Size;

    
    
    if (!Target)
    {
        return;
    }

    if (OutputType)
    {
        STRUPR (OutputType);
        if (OutputType[0] == 'W')
        {
            Display = DB_WORD_DISPLAY;
        }
        else if (OutputType[0] == 'D')
         {
            Display = DB_DWORD_DISPLAY;
        }
        else if (OutputType[0] == 'Q')
         {
            Display = DB_QWORD_DISPLAY;
        }
    }
    

    RetBuf.Length = sizeof (Buffer);
    RetBuf.Pointer = Buffer;

    /* Differentiate between a number and a name */

    if ((Target[0] >= 0x30) && (Target[0] <= 0x39))
    {
        ObjPtr = (void *) STRTOUL (Target, NULL, 16);
        if (!OsdReadable (ObjPtr, 16))
        {
            OsdPrintf ("Address %p is invalid in this address space\n", ObjPtr);
            return;
        }

        if (VALID_DESCRIPTOR_TYPE ((ObjPtr), DESC_TYPE_NTE))
        {
            /* This is an NTE */

            if (!OsdReadable (ObjPtr, sizeof (NAME_TABLE_ENTRY)))
            {
                OsdPrintf ("Cannot read entire NTE at address %p\n", ObjPtr);
                return;
            }

            Entry = ObjPtr;
            goto DumpNte;
        }

        else if (VALID_DESCRIPTOR_TYPE ((ObjPtr), DESC_TYPE_ACPI_OBJ))
        {
            /* This is an ACPI OBJECT */

            if (!OsdReadable (ObjPtr, sizeof (ACPI_OBJECT_INTERNAL)))
            {
                OsdPrintf ("Cannot read entire ACPI object at address %p\n", ObjPtr);
                return;
            }

            CmDumpBuffer (ObjPtr, sizeof (ACPI_OBJECT_INTERNAL), Display, ACPI_UINT32_MAX);
            AmlDumpObjectDescriptor (ObjPtr, 1);
        }

        else if (VALID_DESCRIPTOR_TYPE ((ObjPtr), DESC_TYPE_PARSER))
        {
            /* This is an Parser Op object */

            if (!OsdReadable (ObjPtr, sizeof (ACPI_GENERIC_OP)))
            {
                OsdPrintf ("Cannot read entire Parser object at address %p\n", ObjPtr);
                return;
            }


            CmDumpBuffer (ObjPtr, sizeof (ACPI_GENERIC_OP), Display, ACPI_UINT32_MAX);
            DbDumpParserDescriptor ((ACPI_GENERIC_OP *) ObjPtr);
        }

        else
        {
            Size = 16;
            if (OsdReadable (ObjPtr, 64))
            {
                Size = 64;
            }

            /* Just dump some memory */

            CmDumpBuffer (ObjPtr, Size, Display, ACPI_UINT32_MAX);
        }

        return;
    }
                    

    /* The parameter is a name string that must be resolved to an NTE */

    Entry = DbLocalNsLookup (Target);
    if (!Entry)
    {
        return;
    }

DumpNte:
    /* Now dump the NTE */

    Status = AcpiGetName (Entry, ACPI_FULL_PATHNAME, &RetBuf);
    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Could not convert name to pathname\n");
        return;
    }

    OsdPrintf ("Object Pathname:  %s\n", RetBuf.Pointer);
    if (!OsdReadable (Entry, sizeof (NAME_TABLE_ENTRY)))
    {
        OsdPrintf ("Invalid NTE at address %p\n", Entry);
        return;
    }

    CmDumpBuffer ((void *) Entry, sizeof (NAME_TABLE_ENTRY), Display, ACPI_UINT32_MAX);
    AmlDumpNameTableEntry (Entry, 1);

    if (Entry->Object)
    {
        OsdPrintf ("\nAttached Object (0x%X):\n", Entry->Object);
        if (!OsdReadable (Entry->Object, sizeof (ACPI_OBJECT_INTERNAL)))
        {
            OsdPrintf ("Invalid internal ACPI Object at address %p\n", Entry->Object);
            return;
        }

        CmDumpBuffer (Entry->Object, sizeof (ACPI_OBJECT_INTERNAL), Display, ACPI_UINT32_MAX);
        AmlDumpObjectDescriptor (Entry->Object, 1);
    }
}



void
DbDecodeInternalObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{

    if (!ObjDesc)
    {
        return;
    }

    OsdPrintf (" %9.9s ", CmGetTypeName (ObjDesc->Common.Type));

    switch (ObjDesc->Common.Type)
    {
    case ACPI_TYPE_Number:
        OsdPrintf ("0x%.8X", ObjDesc->Number.Value, ObjDesc->Number.Value);
        break;

    case ACPI_TYPE_String:
        OsdPrintf ("\"%.16s\"...", ObjDesc->String.Pointer);
        break;
    }
}



void 
DbDisplayInternalObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    UINT8                   Type;
    ACPI_WALK_STATE         *WalkState;


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);

    OsdPrintf ("%p ", ObjDesc);

    if (!ObjDesc)
    {
        OsdPrintf ("<NullObj>\n");
        return;
    }

    else if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_PARSER))
    {
        OsdPrintf ("<Parser>  ");
    }

    else if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
    {
        OsdPrintf ("<NTE>             Name %4.4s Type %s", &((NAME_TABLE_ENTRY *)ObjDesc)->Name, 
                                                            CmGetTypeName (((NAME_TABLE_ENTRY *)ObjDesc)->Type));
    }

    else if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_ACPI_OBJ))
    {
        OsdPrintf ("<Obj> ");
        Type = ObjDesc->Common.Type;
        if (Type > INTERNAL_TYPE_MAX)
        {
            OsdPrintf (" Type %x [Invalid Type]", Type);
            return;
        }

        switch (ObjDesc->Common.Type)
        {
        case INTERNAL_TYPE_Reference:
            switch (ObjDesc->Reference.OpCode)
            {
            case AML_ZeroOp:
                OsdPrintf ("[Const]     Number 0x%.8X", 0);
                break;

            case AML_OnesOp:
                OsdPrintf ("[Const]     Number 0x%.8X", 0xFFFFFFFFF);
                break;

            case AML_OneOp:
                OsdPrintf ("[Const]     Number 0x%.8X", 1);
                break;

            case AML_LocalOp:
                OsdPrintf ("[Local%d]", ObjDesc->Reference.Offset);
                if (WalkState)
                {
                    ObjDesc = WalkState->LocalVariables[ObjDesc->Reference.Offset].Object;
                    DbDecodeInternalObject (ObjDesc);
                }
                break;

            case AML_ArgOp:
                OsdPrintf ("[Arg%d]  ", ObjDesc->Reference.Offset);
                if (WalkState)
                {
                    ObjDesc = WalkState->Arguments[ObjDesc->Reference.Offset].Object;
                    DbDecodeInternalObject (ObjDesc);
                }
                break;

            case AML_DebugOp:
                OsdPrintf ("[Debug]  ");
                break;

            case AML_IndexOp:
                OsdPrintf ("[Index]  ");
                DbDecodeInternalObject (ObjDesc->Reference.Object);
                break;

            default:
                break;

            }
            break;
        
        default:
            OsdPrintf ("        ");
            DbDecodeInternalObject (ObjDesc);
            break;
        }
    }

    else
    {
        OsdPrintf ("<Not a valid ACPI Object Descriptor> ");
    }

    OsdPrintf ("\n");
}



void
DbDisplayMethodInfo (
    ACPI_GENERIC_OP         *StartOp)
{
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_GENERIC_OP         *RootOp;
    ACPI_GENERIC_OP         *Op;
    ACPI_OP_INFO            *OpInfo;
    UINT32                  NumOps = 0;
    UINT32                  NumOperands = 0;
    UINT32                  NumOperators = 0;
    UINT32                  NumRemainingOps = 0;
    UINT32                  NumRemainingOperands = 0;
    UINT32                  NumRemainingOperators = 0;
    UINT32                  NumArgs;
    UINT32                  Concurrency;
    BOOLEAN                 CountRemaining = FALSE;



    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    ObjDesc = WalkState->MethodDesc;
    Entry = WalkState->Origin->NameTableEntry;

    NumArgs = ObjDesc->Method.ParamCount;
    Concurrency = ObjDesc->Method.Concurrency;

    OsdPrintf ("Currently executing control method is [%4.4s]\n", &Entry->Name);
    OsdPrintf ("%d arguments, max concurrency = %d\n", NumArgs, Concurrency);


    RootOp = StartOp;
    while (RootOp->Parent)
    {
        RootOp = RootOp->Parent;
    }

    Op = RootOp;

    while (Op)
    {
        if (Op == StartOp)
        {
            CountRemaining = TRUE;
        }

        NumOps++;
        if (CountRemaining)
            NumRemainingOps++;

        OpInfo = PsGetOpcodeInfo (Op->Opcode);
        if (!OpInfo)
        {
            continue;
        }


        /* Decode the opcode */

        switch ((OpInfo->Flags & OP_INFO_TYPE))
        {
        case OPTYPE_CONSTANT:           /* argument type only */
        case OPTYPE_LITERAL:            /* argument type only */
        case OPTYPE_DATA_TERM:          /* argument type only */
        case OPTYPE_LOCAL_VARIABLE:     /* argument type only */
        case OPTYPE_METHOD_ARGUMENT:    /* argument type only */
            if (CountRemaining)
                NumRemainingOperands++;

            NumOperands++;
            break;

        default:
            if (CountRemaining)
                NumRemainingOperators++;

            NumOperators++;
            break;
        }


        Op = PsGetDepthNext (StartOp, Op);
    }

    OsdPrintf ("Method contains:       %d AML Opcodes - %d Operators, %d Operands\n", 
                NumOps, NumOperators, NumOperands);

    OsdPrintf ("Remaining to execute:  %d AML Opcodes - %d Operators, %d Operands\n", 
                NumRemainingOps, NumRemainingOperators, NumRemainingOperands);
}



void
DbDisplayLocals (void)
{
    UINT32                  i;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *Entry;



    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    ObjDesc = WalkState->MethodDesc;
    Entry = WalkState->Origin->NameTableEntry;


    OsdPrintf ("Local Variables for method [%4.4s]:\n", &Entry->Name);

    for (i = 0; i < MTH_NUM_LOCALS; i++)
    {
        ObjDesc = WalkState->LocalVariables[i].Object;
        OsdPrintf ("Local%d: ", i);
        DbDisplayInternalObject (ObjDesc);
    }
}




void
DbDisplayArguments (void)
{  
    UINT32                  i;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    UINT32                  NumArgs;
    UINT32                  Concurrency;
    NAME_TABLE_ENTRY        *Entry;


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    ObjDesc = WalkState->MethodDesc;
    Entry = WalkState->Origin->NameTableEntry;

    NumArgs = ObjDesc->Method.ParamCount;
    Concurrency = ObjDesc->Method.Concurrency;

    OsdPrintf ("Method [%4.4s] has %d arguments, max concurrency = %d\n", &Entry->Name, NumArgs, Concurrency);

    for (i = 0; i < NumArgs; i++)
    {
        ObjDesc = WalkState->Arguments[i].Object;
        OsdPrintf ("Arg%d: ", i);
        DbDisplayInternalObject (ObjDesc);
    }
}




void
DbDisplayResults (void)
{
    UINT32                  i;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    UINT32                  NumResults;
    NAME_TABLE_ENTRY        *Entry;


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    ObjDesc = WalkState->MethodDesc;
    Entry = WalkState->Origin->NameTableEntry;
    NumResults = WalkState->NumResults - WalkState->CurrentResult;

    OsdPrintf ("Method [%4.4s] has %d stacked result objects\n", &Entry->Name, NumResults);

    for (i = WalkState->CurrentResult; i < WalkState->NumResults; i++)
    {
        ObjDesc = WalkState->Results[i];
        OsdPrintf ("Result%d: ", i);
        DbDisplayInternalObject (ObjDesc);
    }
}




void
DbDisplayCallingTree (void)
{
    UINT32                  i;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *Entry;


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);
    if (!WalkState)
    {
        OsdPrintf ("There is no method currently executing\n");
        return;
    }

    ObjDesc = WalkState->MethodDesc;
    Entry = WalkState->Origin->NameTableEntry;

    OsdPrintf ("Current Control Method Call Tree\n");

    for (i = 0; WalkState; i++)
    {
        ObjDesc = WalkState->MethodDesc;
        Entry = WalkState->Origin->NameTableEntry;

        OsdPrintf ("    [%4.4s]\n", &Entry->Name);

        WalkState = WalkState->Next;
    }
}




void
DbDisplayResultObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{

    /* TBD:
     * We don't always want to display the result.
     * For now, only display if single stepping
     * however, this output is very useful in other contexts also
     */

    if (!Gbl_CmSingleStep)
    {
        return;
    }

    OsdPrintf ("ResultObj: ");
    DbDisplayInternalObject (ObjDesc);
    OsdPrintf ("\n");
}



void
DbDisplayArgumentObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{


    if (!Gbl_CmSingleStep)
    {
        return;
    }

    OsdPrintf ("ArgObj:    ");
    DbDisplayInternalObject (ObjDesc);
}






#endif /* ACPI_DEBUG */
