/*
 *  ./acpica/subsystem/debugger/dbstats.c - Generation and display of ACPI table statistics
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <debugger.h>
#include <amlcode.h>
#include <parser.h>

#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbstats");

ARGUMENT_INFO               DbStatTypes [] = 
{
    "OBJECTS",
    "MEMORY",
    "MISC",
    "TABLES",
    NULL            /* Must be null terminated */
};

#define CMD_OBJECTS         0
#define CMD_MEMORY          1
#define CMD_MISC            2
#define CMD_TABLES          3

/*
 * Statistics
 */

UINT32                      NumNames = 0;
UINT32                      NumMethods = 0;
UINT32                      NumRegions = 0;
UINT32                      NumPackages = 0;
UINT32                      NumAliases = 0;
UINT32                      NumDevices = 0;
UINT32                      NumFieldDefs = 0;
UINT32                      NumThermalZones =  0;
UINT32                      NumNamedObjects = 0;
UINT32                      NumGrammarElements = 0;
UINT32                      NumMethodElements = 0;
UINT32                      NumMutexes = 0;
UINT32                      NumPowerResources = 0;
UINT32                      NumBankFields = 0;
UINT32                      NumIndexFields = 0;
UINT32                      NumEvents = 0;

UINT32                      SizeOfParseTree;
UINT32                      SizeOfMethodTrees;
UINT32                      SizeOfNTEs;
UINT32                      SizeOfAcpiObjects;





ACPI_STATUS
DbDisplayStatistics (
    char                    *TypeArg)
{
    UINT32                  i;
    UINT32                  Type;


    if (!Gbl_DSDT)
    {
        OsdPrintf ("*** Warning:  There is no DSDT loaded\n");
    }

    if (!TypeArg)
    {
        OsdPrintf ("Use subcommand OBJECTS, MEMORY, MISC, or TABLES\n");
        return AE_OK;
    }

    STRUPR (TypeArg);
    Type = DbMatchArgument (TypeArg, DbStatTypes);
    if (Type == (UINT32) -1)
    {
        OsdPrintf ("Invalid or unsupported argument\n");
        return AE_OK;
    }


    switch (Type)
    {
    case CMD_TABLES:

        OsdPrintf ("ACPI Table Information:\n\n");
        if (Gbl_DSDT)
        {
            OsdPrintf ("DSDT Length:................% 7ld (0x%X)\n", Gbl_DSDT->Length, Gbl_DSDT->Length);
        }
        break;

    case CMD_OBJECTS:

        OsdPrintf ("\nObjects defined in the current namespace:\n\n");
        OsdPrintf ("Names:......................% 7ld\n", NumNames);
        OsdPrintf ("Events:.....................% 7ld\n", NumEvents);
        OsdPrintf ("Devices:....................% 7ld\n", NumDevices);
        OsdPrintf ("Aliases:....................% 7ld\n", NumAliases);
        OsdPrintf ("Mutexes:....................% 7ld\n", NumMutexes);
        OsdPrintf ("Packages:...................% 7ld\n", NumPackages);
        OsdPrintf ("Bank Fields.................% 7ld\n", NumBankFields);
        OsdPrintf ("Index Fields................% 7ld\n", NumIndexFields);
        OsdPrintf ("Thermal Zones:..............% 7ld\n", NumThermalZones);
        OsdPrintf ("Power Resources:............% 7ld\n", NumPowerResources);
        OsdPrintf ("Control Methods:............% 7ld\n", NumMethods);
        OsdPrintf ("Operation Regions:..........% 7ld\n", NumRegions);
        OsdPrintf ("Field Definitions:..........% 7ld\n", NumFieldDefs);
        OsdPrintf ("Total Named objects:........% 7ld\n", NumNamedObjects);

        OsdPrintf ("\n");

        OsdPrintf ("ASL/AML Grammar Usage:\n\n");
        OsdPrintf ("Elements Inside Methods:....% 7ld\n", NumMethodElements);
        OsdPrintf ("Elements Outside Methods:...% 7ld\n", NumGrammarElements - NumMethodElements);
        OsdPrintf ("Total Grammar Elements:.....% 7ld\n", NumGrammarElements);
        break;

    case CMD_MEMORY:

        OsdPrintf ("\nDynamic Memory Estimates:\n\n");
        OsdPrintf ("Parse Tree without Methods:.% 7ld\n", SizeOfParseTree);
        OsdPrintf ("Control Method Parse Trees:.% 7ld (If parsed simultaneously)\n", SizeOfMethodTrees);
        OsdPrintf ("Named Object NTEs:..........% 7ld (%d objects)\n", SizeOfNTEs, NumNamedObjects);
        OsdPrintf ("Named Internal Objects......% 7ld\n", SizeOfAcpiObjects);
        OsdPrintf ("State Cache size............% 7ld\n", Gbl_GenericStateCacheDepth * sizeof (ACPI_GENERIC_STATE));
        OsdPrintf ("Parse Cache size............% 7ld\n", Gbl_ParseCacheDepth * sizeof (ACPI_GENERIC_OP));
        OsdPrintf ("Object Cache size...........% 7ld\n", Gbl_ObjectCacheDepth * sizeof (ACPI_OBJECT_INTERNAL));
        OsdPrintf ("WalkState Cache size........% 7ld\n", Gbl_WalkStateCacheDepth * sizeof (ACPI_WALK_STATE));

        OsdPrintf ("\n");

        OsdPrintf ("Cache Statistics:\n\n");
        OsdPrintf ("State Cache requests........% 7ld\n", Gbl_StateCacheRequests);
        OsdPrintf ("State Cache hits............% 7ld\n", Gbl_StateCacheHits);
        OsdPrintf ("State Cache depth...........% 7ld\n", Gbl_GenericStateCacheDepth);
        OsdPrintf ("Parse Cache requests........% 7ld\n", Gbl_ParseCacheRequests);
        OsdPrintf ("Parse Cache hits............% 7ld\n", Gbl_ParseCacheHits);
        OsdPrintf ("Parse Cache depth...........% 7ld\n", Gbl_ParseCacheDepth);
        OsdPrintf ("Object Cache requests.......% 7ld\n", Gbl_ObjectCacheRequests);
        OsdPrintf ("Object Cache hits...........% 7ld\n", Gbl_ObjectCacheHits);
        OsdPrintf ("Object Cache depth..........% 7ld\n", Gbl_ObjectCacheDepth);
        OsdPrintf ("WalkState Cache requests....% 7ld\n", Gbl_WalkStateCacheRequests);
        OsdPrintf ("WalkState Cache hits........% 7ld\n", Gbl_WalkStateCacheHits);
        OsdPrintf ("WalkState Cache depth.......% 7ld\n", Gbl_WalkStateCacheDepth);
        break;

    case CMD_MISC:

        OsdPrintf ("\nMiscellaneous Statistics:\n\n");
        OsdPrintf ("Calls to PsFind:..  ........% 7ld\n", Gbl_PsFindCount);
        OsdPrintf ("Calls to NsLookup:..........% 7ld\n", Gbl_NsLookupCount);

        OsdPrintf ("\n");

        OsdPrintf ("Mutex usage:\n\n");
        for (i = 0; i < NUM_MTX; i++)
        {
            OsdPrintf ("%-20s:       % 7ld\n", CmGetMutexName (i), Gbl_AcpiMutexInfo[i].UseCount);
        }
        break;
    }

    OsdPrintf ("\n");
    return AE_OK;   
}



void
DbGenerateStatistics (
    ACPI_GENERIC_OP         *Root,
    BOOLEAN                 IsMethod)
{
    ACPI_GENERIC_OP         *Op;



    Op = PsGetChild (Root);

    while (Op)
    {
        switch (Op->Opcode)
        {
        case AML_MethodOp:
            NumMethods++;
            break;

        case AML_NameOp:
            NumNames++;
            break;

        case AML_RegionOp:
            NumRegions++;
            break;

        case AML_AliasOp:
            NumAliases++;
            break;

        case AML_PackageOp:
            NumPackages++;
            break;

        case AML_DeviceOp:
            NumDevices++;
            break;

        case AML_ThermalZoneOp:
            NumThermalZones++;
            break;

        case AML_DefFieldOp:
            NumFieldDefs++;
            break;

        case AML_PowerResOp:
            NumPowerResources++;
            break;

        case AML_MutexOp:
            NumMutexes++;
            break;

        case AML_BankFieldOp:
            NumBankFields++;
            break;

        case AML_IndexFieldOp:
            NumIndexFields++;
            break;

        case AML_EventOp:
            NumEvents++;
            break;
        }

        if (PsIsNamedOp (Op->Opcode))
        {
            NumNamedObjects++;
        }

        if (IsMethod)
        {
            NumMethodElements++;
        }

        NumGrammarElements++;
        Op = PsGetDepthNext (Root, Op);
    }



    SizeOfParseTree     = (NumGrammarElements - NumMethodElements) * (UINT32) sizeof (ACPI_GENERIC_OP);
    SizeOfMethodTrees   = NumMethodElements * (UINT32) sizeof (ACPI_GENERIC_OP);
    SizeOfNTEs          = NumNamedObjects * (UINT32) sizeof (NAME_TABLE_ENTRY);
    SizeOfAcpiObjects   = NumNamedObjects * (UINT32) sizeof (ACPI_OBJECT_INTERNAL);

}

#endif /* ACPI_DEBUG  */
