/*
 *  ./acpica/subsystem/debugger/dbutils.c - AML debugger utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>
#include <events.h>
#include <interp.h>
#include <debugger.h>


#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbutils");



void
DbSetOutputDestination (
    INT32                   OutputFlags)
{
    
    Gbl_DbOutputFlags = (UINT8) OutputFlags;

    if (OutputFlags & DB_REDIRECTABLE_OUTPUT)
    {
        if (OutputToFile)
        {
            DebugLevel = Gbl_DbDebugLevel;
        }
    }
    else
    {
        DebugLevel = Gbl_DbConsoleDebugLevel;
    }
}


void
DbDumpBuffer (
    UINT32                  Address)
{

    OsdPrintf ("\nLocation 0x%X:\n", Address);

    DebugLevel |= TRACE_TABLES;
    CmDumpBuffer ((char *) Address, 64, DB_BYTE_DISPLAY, ACPI_UINT32_MAX);
}



void
DbDumpObject (
    ACPI_OBJECT             *ObjDesc,
    UINT32                  Level)
{
    UINT32                  i;


    if (!ObjDesc)
    {
        return;
    }

    for (i = 0; i < Level; i++)
    {
        OsdPrintf ("  ");
    }

    switch (ObjDesc->Type)
    {
    case ACPI_TYPE_Any:
        OsdPrintf ("[Object Reference]  Value: %p\n", ObjDesc->Reference.Handle);
        break;

    case ACPI_TYPE_Number:
        OsdPrintf ("[Number]  Value: %ld (0x%lX)\n", ObjDesc->Number.Value, ObjDesc->Number.Value);
        break;

    case ACPI_TYPE_String:
        OsdPrintf ("[String]  Value: ");
        for (i = 0; i < ObjDesc->String.Length; i++)
        {
            OsdPrintf ("%c", ObjDesc->String.Pointer[i]);
        }
        OsdPrintf ("\n");
        break;

    case ACPI_TYPE_Buffer:
        OsdPrintf ("[Buffer]  Value: ");
        CmDumpBuffer ((char *) ObjDesc->Buffer.Pointer, ObjDesc->Buffer.Length, DB_DWORD_DISPLAY, _COMPONENT);
        break;

    case ACPI_TYPE_Package:
        OsdPrintf ("[Package]  Contains %d Elements: \n", ObjDesc->Package.Count);

        for (i = 0; i < ObjDesc->Package.Count; i++)
        {
            DbDumpObject (&ObjDesc->Package.Elements[i], Level+1);
        }
        break;

    default:
        OsdPrintf ("[Unknown Type] 0x%X \n", ObjDesc->Type);
        break;
    }
}




void
DbPrepNamestring (
    char                    *Name)
{


    if (!Name)
    {
        return;
    }

    STRUPR (Name);

    /* Convert a leading forward slash to a backslash */

    if (*Name == '/')
    {
        *Name = '\\';
    }

    /* Ignore a leading backslash, this is the root prefix */

    if (*Name == '\\')
    {
        *Name++;
    }

    /* Convert all slash path separators to dots */

    while (*Name)
    {
        if ((*Name == '/') ||
            (*Name == '\\'))
        {
            *Name = '.';
        }

        Name++;
    }
}



ACPI_STATUS
DbSecondPassParse (
    ACPI_GENERIC_OP         *Root)
{
    ACPI_GENERIC_OP         *Op = Root;
    ACPI_DEFERRED_OP        *Method;
    ACPI_GENERIC_OP         *SearchOp;
    ACPI_GENERIC_OP         *StartOp;
    ACPI_STATUS             Status = AE_OK;
    UINT32                  BaseAmlOffset;


    OsdPrintf ("Pass two parse ....\n");

    while (Op)
    {
        if (Op->Opcode == AML_MethodOp)
        {
            Method = (ACPI_DEFERRED_OP *) Op;
            Status = PsParseAml (Op, Method->Body, Method->BodyLength);

          
            BaseAmlOffset = (Method->Value.Arg)->AmlOffset + 1;
            StartOp = (Method->Value.Arg)->Next;
            SearchOp = StartOp;

            while (SearchOp)
            {
                SearchOp->AmlOffset += BaseAmlOffset;
                SearchOp = PsGetDepthNext (StartOp, SearchOp);
            }

        }

        if (Op->Opcode == AML_RegionOp)
        {
            /* TBD: this isn't quite the right thing to do! */

            // Method = (ACPI_DEFERRED_OP *) Op;
            // Status = PsParseAml (Op, Method->Body, Method->BodyLength);
        }

        if (ACPI_FAILURE (Status))
        {
            return Status;
        }

        Op = PsGetDepthNext (Root, Op);
    }

    return Status;
}



NAME_TABLE_ENTRY *
DbLocalNsLookup (
    char                    *Name)
{
    char                    *InternalPath;
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *Entry = NULL;


    DbPrepNamestring (Name);

    /* Build an internal namestring */

    Status = NsInternalizeName (Name, &InternalPath);
    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Invalid namestring: %s\n", Name);
        return NULL;
    }

    /* Lookup the name */

    /* TBD: what scope do we use? */
    /* Use the root scope for the start of the search */

    Status = NsLookup (NULL, InternalPath, ACPI_TYPE_Any, IMODE_Execute, 
                                    NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &Entry);

    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Could not locate name: %s %s\n", Name, CmFormatException (Status));
    }


    CmFree (InternalPath);

    return Entry;
}




#endif /* ACPI_DEBUG */


