
/*
 *  ./acpica/subsystem/dispatcher/dscontrl.c - Dispatcher handling of "control" opcodes
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DPCONTRL_C__

#include <acpi.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dscontrl");





ACPI_STATUS 
DsExecBeginControlOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_GENERIC_STATE      *ControlState;


    DEBUG_PRINT (TRACE_DISPATCH, ("BeginControlOp: Op=%p Opcode=%2.2X State=%p\n", Op, Op->Opcode, WalkState));

    switch (Op->Opcode)
    {
    case AML_IfOp:
    case AML_WhileOp:

        /* 
         * IF/WHILE: Create a new control state to manage these constructs.
         * We need to manage these as a stack, in order to handle nesting.
         */

        ControlState = CmCreateControlState ();
        if (!ControlState)
        {
            Status = AE_NO_MEMORY;
            break;
        }

        CmPushGenericState (&WalkState->ControlState, ControlState);
        break;
    

    case AML_ElseOp:

        /* Predicate is in the state object */
        /* If predicate is true, the IF was executed, ignore ELSE part */

        if (WalkState->LastPredicate)     
        {
            Status = AE_CTRL_TRUE;
        }

        break;


    case AML_ReturnOp:

        break;


    default:
        break;
    }

    return Status;
}



ACPI_STATUS 
DsExecEndControlOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_GENERIC_STATE      *ControlState;



    switch (Op->Opcode)
    {
    case AML_IfOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("EndControlOp: [IF_OP] Op=%p\n", Op));

        /* Save the result of the predicate in case there is an ELSE to come */

        WalkState->LastPredicate = (BOOLEAN) WalkState->ControlState->Common.Value;

        /* Pop the control state that was created at the start of the IF and free it */

        ControlState = CmPopGenericState (&WalkState->ControlState);
        CmDeleteGenericState (ControlState);

        break;


    case AML_ElseOp:

        break;


    case AML_WhileOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("EndControlOp: [WHILE_OP] Op=%p\n", Op));

        if (WalkState->ControlState->Common.Value)
        {
            /* Predicate was true, go back and evaluate it again! */
            
            Status = AE_CTRL_TRUE;
        }

        else
        {
            DEBUG_PRINT (TRACE_DISPATCH, ("EndControlOp: [WHILE_OP] termination! Op=%p\n", Op));

            /* Pop this control state and free it */

            ControlState = CmPopGenericState (&WalkState->ControlState);
            CmDeleteGenericState (ControlState);
        }

        break;


    case AML_ReturnOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("EndControlOp: [RETURN_OP] Op=%p Arg=%p\n",Op, Op->Value.Arg));


        /* One optional operand -- the return value */

        if (Op->Value.Arg)
        {
            Status = DsCreateOperands (WalkState, Op->Value.Arg);
            if (ACPI_FAILURE (Status))
            {
                return Status;
            }

            /* TBD: Just check for NULL arg to signify no return value??? */

            /* 
             * If value being returned is a Reference (such as an arg or local), 
             * resolve it now because it may cease to exist at the end of the method. 
             */

            Status = AmlResolveToValue (&WalkState->Operands [0]);
            if (ACPI_FAILURE (Status))
            {
                return Status;
            }

            /* 
             * Get the return value and save as the last result value
             * This is the only place where WalkState->ReturnDesc is set to anything other
             * than zero!
             */

            WalkState->ReturnDesc = WalkState->Operands[0];
        }

        else
        {
            /* No return operand */

            CmRemoveReference (WalkState->Operands [0]);

            WalkState->Operands [0]     = NULL;
            WalkState->NumOperands      = 0;
            WalkState->ReturnDesc       = NULL;
        }


        DEBUG_PRINT (TRACE_DISPATCH, ("EndControlOp: Completed RETURN_OP State=%p, RetVal=%p\n", WalkState, WalkState->ReturnDesc));

        Status = AE_CTRL_TERMINATE;      /* End the control method execution right now */
        break;


    case AML_NoopCode:

        /* Just do nothing! */
        break;


    case AML_BreakPointOp:

        /* Call up to the OS dependent layer to handle this */

        OsdBreakpoint (NULL);

        /* If it returns, we are done! */

        break;


    case AML_BreakOp:

        DEBUG_PRINT (ACPI_INFO, ("EndControlOp: Break to end of current package, Op=%p\n", Op));

        /* 
         * As per the ACPI specification:
         *      "The break operation causes the current package execution to complete"
         *      "Break -- Stop executing the current code package at this point"
         *
         * Returning AE_FALSE here will cause termination of the current package,
         * and execution will continue one level up, starting with the completion of
         * the parent Op.
         */

        Status = AE_CTRL_FALSE;
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("EndControlOp: Unknown control opcode=%X Op=%p\n", Op->Opcode, Op));

        Status = AE_AML_BAD_OPCODE;
        break;
    }


    return Status;
}






