
/*
 *  ./acpica/subsystem/dispatcher/dsfield.c - Dispatcher field routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSFIELD_C__

#include <acpi.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dsfield");


/*
 * Field flags: Bits 00 - 03 : AccessType (AnyAcc, ByteAcc, etc.)
 *                   04      : LockRule (1 == Lock)
 *                   05 - 06 : UpdateRule
 */

#define FIELD_ACCESS_TYPE_MASK      0x0F
#define FIELD_LOCK_RULE_MASK        0x10
#define FIELD_UPDATE_RULE_MASK      0x60




ACPI_STATUS
DsCreateField (
    ACPI_GENERIC_OP         *Op,
    ACPI_HANDLE             Region,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status = AE_AML_ERROR;
    ACPI_GENERIC_OP         *Arg;
    NAME_TABLE_ENTRY        *Entry;
    UINT8                   FieldFlags;
    UINT8                   AccessAttribute = 0;
    UINT32                  FieldBitPosition = 0;


    FUNCTION_TRACE_PTR ("DsCreateField", Op);
    


    /* First arg is the name of the parent OpRegion */

    Arg = Op->Value.Arg;

    /* Second arg is the field flags */

    Arg = Arg->Next;
    FieldFlags = (UINT8) Arg->Value.Integer;

    /* Each remaining arg is a Named Field */

    Arg = Arg->Next;
    while (Arg)
    {
        switch (Arg->Opcode)
        {
        case AML_RESERVEDFIELD_OP:

            FieldBitPosition += Arg->Value.Size;
            break;


        case AML_ACCESSFIELD_OP:

            /* Get a new AccessType and AccessAttribute for all entries (until end or another AccessAs keyword) */

            AccessAttribute = (UINT8) Arg->Value.Integer;
            FieldFlags      = (UINT8) ((FieldFlags & FIELD_ACCESS_TYPE_MASK) || ((UINT8) (Arg->Value.Integer >> 8)));
            break;


        case AML_NAMEDFIELD_OP:

            Status = NsLookup (WalkState->ScopeInfo, (char *) &((ACPI_NAMED_OP *)Arg)->Name, INTERNAL_TYPE_DefField, IMODE_LoadPass1, 
                                        NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &Entry);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Initialize an object for the new NTE that is on the object stack */

            Status = AmlPrepDefFieldValue (Entry, Region, FieldFlags, AccessAttribute, FieldBitPosition, Arg->Value.Size);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Keep track of bit position for the *next* field */

            FieldBitPosition += Arg->Value.Size;
            break;
        }

        Arg = Arg->Next;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsCreateBankField (
    ACPI_GENERIC_OP         *Op,
    ACPI_HANDLE             Region,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status = AE_AML_ERROR;
    ACPI_GENERIC_OP         *Arg;
    NAME_TABLE_ENTRY        *BankReg;
    NAME_TABLE_ENTRY        *Entry;
    UINT32                  BankValue;
    UINT8                   FieldFlags;
    UINT8                   AccessAttribute = 0;
    UINT32                  FieldBitPosition = 0;

    
    FUNCTION_TRACE_PTR ("DsCreateBankField", Op);


    /* First arg is the name of the parent OpRegion */

    Arg = Op->Value.Arg;

    /* Socond arg is the Bank Register */

    Arg = Arg->Next;

    Status = NsLookup (WalkState->ScopeInfo, Arg->Value.String, INTERNAL_TYPE_BankFieldDefn, IMODE_LoadPass1, 
                                NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &BankReg);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Third arg is the BankValue */
    
    Arg = Arg->Next;
    BankValue = Arg->Value.Integer;


    /* Next arg is the field flags */

    Arg = Arg->Next;
    FieldFlags = (UINT8) Arg->Value.Integer;

    /* Each remaining arg is a Named Field */

    Arg = Arg->Next;
    while (Arg)
    {
        switch (Arg->Opcode)
        {
        case AML_RESERVEDFIELD_OP:

            FieldBitPosition += Arg->Value.Size;
            break;


        case AML_ACCESSFIELD_OP:

            /* Get a new AccessType and AccessAttribute for all entries (until end or another AccessAs keyword) */

            AccessAttribute = (UINT8) Arg->Value.Integer;
            FieldFlags      = (UINT8) ((FieldFlags & FIELD_ACCESS_TYPE_MASK) || ((UINT8) (Arg->Value.Integer >> 8)));
            break;


        case AML_NAMEDFIELD_OP:

            Status = NsLookup (WalkState->ScopeInfo, (char *) &((ACPI_NAMED_OP *)Arg)->Name, INTERNAL_TYPE_DefField, IMODE_LoadPass1, 
                                        NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &Entry);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Initialize an object for the new NTE that is on the object stack */

            Status = AmlPrepBankFieldValue (Entry, Region, BankReg, BankValue, FieldFlags, AccessAttribute, FieldBitPosition, Arg->Value.Size);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Keep track of bit position for the *next* field */

            FieldBitPosition += Arg->Value.Size;
            break;

        }

        Arg = Arg->Next;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsCreateIndexField (
    ACPI_GENERIC_OP         *Op,
    ACPI_HANDLE             Region,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status;
    ACPI_GENERIC_OP         *Arg;
    NAME_TABLE_ENTRY        *Entry;
    NAME_TABLE_ENTRY        *IndexReg;
    NAME_TABLE_ENTRY        *DataReg;
    UINT8                   FieldFlags;
    UINT8                   AccessAttribute = 0;
    UINT32                  FieldBitPosition = 0;

    
    FUNCTION_TRACE_PTR ("DsCreateIndexField", Op);


    Arg = Op->Value.Arg;

    /* First arg is the name of the Index register */

    Status = NsLookup (WalkState->ScopeInfo, Arg->Value.String, ACPI_TYPE_Any, IMODE_LoadPass1, 
                                NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &IndexReg);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Second arg is the data register */

    Arg = Arg->Next;

    Status = NsLookup (WalkState->ScopeInfo, Arg->Value.String, INTERNAL_TYPE_IndexFieldDefn, IMODE_LoadPass1, 
                                NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &DataReg);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }


    /* Next arg is the field flags */

    Arg = Arg->Next;
    FieldFlags = (UINT8) Arg->Value.Integer;


    /* Each remaining arg is a Named Field */

    Arg = Arg->Next;
    while (Arg)
    {
        switch (Arg->Opcode)
        {
        case AML_RESERVEDFIELD_OP:

            FieldBitPosition += Arg->Value.Size;
            break;


        case AML_ACCESSFIELD_OP:

            /* Get a new AccessType and AccessAttribute for all entries (until end or another AccessAs keyword) */

            AccessAttribute = (UINT8) Arg->Value.Integer;
            FieldFlags      = (UINT8) ((FieldFlags & FIELD_ACCESS_TYPE_MASK) || ((UINT8) (Arg->Value.Integer >> 8)));
            break;


        case AML_NAMEDFIELD_OP:

            Status = NsLookup (WalkState->ScopeInfo, (char *) &((ACPI_NAMED_OP *)Arg)->Name, INTERNAL_TYPE_IndexField, IMODE_LoadPass1, 
                                        NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, NULL, &Entry);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Initialize an object for the new NTE that is on the object stack */

            Status = AmlPrepIndexFieldValue (Entry, IndexReg, DataReg, FieldFlags, AccessAttribute, FieldBitPosition, Arg->Value.Size);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            /* Keep track of bit position for the *next* field */

            FieldBitPosition += Arg->Value.Size;
            break;


        default:

            DEBUG_PRINT (ACPI_ERROR, ("DsEnterIndexField: Invalid opcode in field list: %X\n", Arg->Opcode));
            Status = AE_AML_ERROR;
            break;
        }

        Arg = Arg->Next;
    }

    return_ACPI_STATUS (Status);
}





