
/*
 *  ./acpica/subsystem/dispatcher/dsmethod.c - Parser/Interpreter interface - control method parsing
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSMETHOD_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <debugger.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dsmethod");



ACPI_STATUS
DsParseMethod (
    ACPI_HANDLE             ObjHandle)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_GENERIC_OP         *Op;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_OWNER_ID           OwnerId;


    FUNCTION_TRACE_PTR ("DsParseMethod", ObjHandle);


    /* Parameter Validation */

    if (!ObjHandle)
    {
        return_ACPI_STATUS (AE_NULL_ENTRY);
    }

    DEBUG_PRINT (ACPI_INFO, ("DsParseMethod: **** Parsing [%4.4s] **** Nte=%p\n", 
                    &((NAME_TABLE_ENTRY *)ObjHandle)->Name, ObjHandle));


    /* Extract the method object from the method NTE */

    Entry = (NAME_TABLE_ENTRY *) ObjHandle;
    ObjDesc = Entry->Object;
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_NULL_OBJECT);
    }

     /* Create a mutex for the method if there is a concurrency limit */

    if ((ObjDesc->Method.Concurrency != INFINITE_CONCURRENCY) &&
        (!ObjDesc->Method.Semaphore))
    {
        Status = OsdCreateSemaphore (ObjDesc->Method.Concurrency, &ObjDesc->Method.Semaphore);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }

    /* Allocate a new parser op to be the root of the parsed method tree */

    Op = PsAllocOp (AML_MethodOp);
    if (!Op)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Init new op with the method name and pointer back to the NTE */

    PsSetName (Op, Entry->Name);
    Op->NameTableEntry = Entry;


    /* 
     * Parse the method, creating a parse tree.
     *
     * The parse also includes a first pass load of the namespace where
     * newly declared named objects are added into the namespace.  Actual evaluation
     * of the named objects (what would be called a "second pass") happens during the
     * actual execution of the method so that operands to the named objects can take
     * on dynamic run-time values.
     */

    Status = PsParseAml (Op, ObjDesc->Method.Pcode, ObjDesc->Method.PcodeLength);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }


    /* Get a new OwnerId for objects created by this method */

    OwnerId = CmAllocateOwnerId (OWNER_TYPE_METHOD);

    /* Install the parsed tree in the method object */

    ObjDesc->Method.ParserOp = Op;
    ObjDesc->Method.OwningId = OwnerId;

    DEBUG_EXEC (DbGenerateStatistics (Op, 1));

    DEBUG_PRINT (ACPI_INFO, ("DsParseMethod: **** [%4.4s] Parsed **** Nte=%p Op=%p\n", 
                    &((NAME_TABLE_ENTRY *)ObjHandle)->Name, ObjHandle, Op));

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsBeginMethodExecution (
    NAME_TABLE_ENTRY        *MethodEntry,
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("DsBeginMethodExecution", MethodEntry);


    if (!MethodEntry)
    {
        return_ACPI_STATUS (AE_NULL_ENTRY);
    }

    ObjDesc = NsGetAttachedObject (MethodEntry);
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_NULL_OBJECT);
    }

    /*
     * Lock the parser while we check for and possibly parse the control method
     */

    CmAcquireMutex (MTX_PARSER);


    /* If method is not parsed at this time, we must parse it first */

    if (!ObjDesc->Method.ParserOp)
    {

        DEBUG_PRINT (ACPI_INFO, ("PsxExecute: **** Parsing Method **** obj=%p\n", ObjDesc));

        Status = DsParseMethod (MethodEntry);
        if (ACPI_FAILURE (Status))
        {
            CmReleaseMutex (MTX_PARSER);
            return_ACPI_STATUS (Status);
        }
    }


    /* 
     * Increment the method parse tree thread count since there is one additional thread executing
     * in it.  If configured for deletion-on-exit, the parse tree will be deleted when the last
     * thread completes execution of the method
     */

    ((ACPI_DEFERRED_OP *) ObjDesc->Method.ParserOp)->ThreadCount++;

    /* 
     * Parsing is complete, we can unlock the parser.  Parse tree cannot be deleted
     * at least until this thread completes.
     */

    CmReleaseMutex (MTX_PARSER);
    
    /* 
     * If there is a concurrency limit on this method, we need to obtain a unit
     * from the method semaphore.  This releases the interpreter if we block
     */

    if (ObjDesc->Method.Semaphore)
    {
        Status = OsLocalWaitSemaphore (ObjDesc->Method.Semaphore, WAIT_FOREVER);
    }


    return_ACPI_STATUS (Status);

}



ACPI_STATUS
DsCallControlMethod (
    ACPI_WALK_LIST          *WalkList,
    ACPI_WALK_STATE         *ThisWalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status;
    ACPI_DEFERRED_OP        *Method;
    NAME_TABLE_ENTRY        *MethodEntry;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_WALK_STATE         *NextWalkState;
    UINT32                  i;


    FUNCTION_TRACE_PTR ("DsCallControlMethod", ThisWalkState);

    DEBUG_PRINT (TRACE_DISPATCH, ("DsCall, execute method %p, currentstate=%p\n", 
                        ThisWalkState->PrevOp, ThisWalkState));


    /*
     * PrevOp points to the METHOD_CALL Op.
     * Get the NTE entry (in the METHOD_CALL->NAME Op) and the corresponding METHOD Op
     */

    MethodEntry = (ThisWalkState->PrevOp->Value.Arg)->NameTableEntry;
    if (!MethodEntry)
    {
        return_ACPI_STATUS (AE_NULL_ENTRY);
    }

    ObjDesc = NsGetAttachedObject (MethodEntry);
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_NULL_OBJECT);
    }

    /* Parse method if necessary, wait on concurrency semaphore */

    Status = DsBeginMethodExecution (MethodEntry, ObjDesc);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Save the (current) Op for when this walk is restarted */

    ThisWalkState->MethodCallOp = ThisWalkState->PrevOp;
    ThisWalkState->PrevOp       = Op;
    Method                      = ObjDesc->Method.ParserOp;

    /* Create a new state for the preempting walk */

    NextWalkState = DsCreateWalkState ((ACPI_GENERIC_OP *) Method, ObjDesc, WalkList);
    if (!NextWalkState)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* The NextOp of the NextWalk will be the beginning of the method */
    
    NextWalkState->NextOp = (ACPI_GENERIC_OP *) Method;

    /* Open a new scope */

    Status = DsScopeStackPush (MethodEntry->Scope, ACPI_TYPE_Method, NextWalkState);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }


    /* 
     * Initialize the arguments for the method.  The resolved arguments were put 
     * on the previous walk state's operand stack.  Operands on the previous walk state
     * stack always start at index 0.
     */

    Status = DsMethodDataInitArgs (&ThisWalkState->Operands[0], ThisWalkState->NumOperands);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    /* Delete the operands on the previous walkstate operand stack (they were copied to new objects) */

    for (i = 0; i < ObjDesc->Method.ParamCount; i++)
    {
        CmRemoveReference (ThisWalkState->Operands [i]);
    }

    /* Clear the operand stack */

    ThisWalkState->NumOperands = 0;

 
    DEBUG_PRINT (TRACE_DISPATCH, ("DsCall, starting nested execution, newstate=%p\n", NextWalkState));

    return_ACPI_STATUS (AE_OK);


    /* On error, we must delete the new walk state */

Cleanup:
    DsTerminateControlMethod (NextWalkState);
    DsDeleteWalkState (NextWalkState);
    return_ACPI_STATUS (Status);

}



ACPI_STATUS
DsRestartControlMethod (
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    *ReturnDesc)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("DsRestartControlMethod", WalkState);


    if (ReturnDesc)
    {
        /* Get the return value (if any) from the previous method.  NULL if no return value */

        Status = DsResultStackPush (ReturnDesc, WalkState);
        if (ACPI_FAILURE (Status))
        {
            CmRemoveReference (ReturnDesc);
            return_ACPI_STATUS (Status);
        }

        /* Delete the return value if it will not be used by the calling method */

        DsDeleteResultIfNotUsed (WalkState->MethodCallOp, ReturnDesc, WalkState);
    }

    DEBUG_PRINT (TRACE_DISPATCH, ("DsRestart: Method=%p Return=%p State=%p\n", 
                        WalkState->MethodCallOp, ReturnDesc, WalkState));


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsTerminateControlMethod (
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_DEFERRED_OP        *Op;
    NAME_TABLE_ENTRY        *MethodEntry;


    FUNCTION_TRACE_PTR ("DsTerminateControlMethod", WalkState);


    /* The method object should be stored in the walk state */

    ObjDesc = WalkState->MethodDesc;
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_OK);
    }

    /* Delete all arguments and locals */

    DsMethodDataDeleteAll (WalkState);

    /* 
     * Lock the parser while we terminate this method.
     * If this is the last thread executing the method, we have
     * additional cleanup to perform
     */

    CmAcquireMutex (MTX_PARSER);

    /* The root of the method parse tree should be stored in the method object */

    Op = ObjDesc->Method.ParserOp;
    if (!Op)
    {
        goto UnlockAndExit;
    }
 
    /* Signal completion of the execution of this method if necessary */

    if (WalkState->MethodDesc->Method.Semaphore)
    {
        Status = OsdSignalSemaphore (WalkState->MethodDesc->Method.Semaphore, 1);
    }

    /* Decrement the thread count on the method parse tree */

    Op->ThreadCount--;
    if (!Op->ThreadCount)
    {
        /*
         * There are no more threads executing this method.  Perform additional cleanup.
         *
         * The method NTE is stored in the method Op 
         */
        MethodEntry = Op->NameTableEntry;

        /*
         * Delete any namespace entries created immediately underneath the method
         */
        CmAcquireMutex (MTX_NAMESPACE);
        if (MethodEntry->Scope)
        {
            NsDeleteNamespaceSubtree (MethodEntry);
        }

        /* 
         * Delete any namespace entries created anywhere else within the namespace 
         */

        NsDeleteNamespaceByOwner (WalkState->MethodDesc->Method.OwningId);
        CmReleaseMutex (MTX_NAMESPACE);


        /* 
         * Delete the method's parse tree if asked to 
         */
        if (Gbl_WhenToParseMethods & METHOD_DELETE_AT_COMPLETION)
        {
            PsDeleteParseTree (WalkState->MethodDesc->Method.ParserOp);
            WalkState->MethodDesc->Method.ParserOp = NULL;
        }
    }


UnlockAndExit:

    CmReleaseMutex (MTX_PARSER);
    return_ACPI_STATUS (AE_OK);
}





