
/*
 *  ./acpica/subsystem/dispatcher/dsmthdat.c - control method arguments and local variables
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSMTHDAT_C__

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dsmthdat");




ACPI_STATUS
DsMethodDataInit (
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  i;


    FUNCTION_TRACE ("DsMethodDataInit");

    /*
     * WalkState fields are initialized to zero by the CmCallocate().
     *
     * An NTE is assigned to each argument and local so that RefOf() can return
     * a pointer to the NTE.
     */

    /* Init the method arguments */

    for (i = 0; i < MTH_NUM_ARGS; i++)
    {
        WalkState->Arguments[i].DataType        = DESC_TYPE_NTE;
        WalkState->Arguments[i].Type            = INTERNAL_TYPE_MethodArgument;
        WalkState->Arguments[i].Name            = *((UINT32 *) NAMEOF_ARG_NTE) | (i << 24);
    }

    /* Init the method locals */

    for (i = 0; i < MTH_NUM_LOCALS; i++)
    {
        WalkState->LocalVariables[i].DataType   = DESC_TYPE_NTE;
        WalkState->LocalVariables[i].Type       = INTERNAL_TYPE_MethodLocalVar;
        WalkState->LocalVariables[i].Name       = *((UINT32 *) NAMEOF_LOCAL_NTE) | (i << 24);
    }


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsMethodDataDeleteAll (
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  Index;
    ACPI_OBJECT_INTERNAL    *Object;


    FUNCTION_TRACE ("DsMethodDataDeleteAll");


    /* Delete the locals */

    DEBUG_PRINT (ACPI_INFO, ("MethodDeleteAll: Deleting local variables in %p\n", WalkState));

    for (Index = 0; Index < MTH_NUM_LOCALS; Index++)
    {
        Object = WalkState->LocalVariables[Index].Object;
        if (Object)
        {
            DEBUG_PRINT (TRACE_EXEC, ("MethodDeleteAll: Deleting Local%d=%p\n", Index, Object));

            WalkState->LocalVariables[Index].Object = NULL; /* Remove first */
            CmRemoveReference (Object);                     /* Was given a ref when stored */
       }
    }


    /* Delete the arguments */

    DEBUG_PRINT (ACPI_INFO, ("MethodDeleteAll: Deleting arguments in %p\n", WalkState));

    for (Index = 0; Index < MTH_NUM_ARGS; Index++)
    {
        Object = WalkState->Arguments[Index].Object;
        if (Object)
        {
            DEBUG_PRINT (TRACE_EXEC, ("MethodDeleteAll: Deleting Arg%d=%p\n", Index, Object));

            WalkState->Arguments[Index].Object = NULL;      /* Remove first */
            CmRemoveReference (Object);                     /* Was given a ref when stored */
        }
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsMethodDataInitArgs (
    ACPI_OBJECT_INTERNAL    **Params,
    UINT32                  MaxParamCount)
{
    ACPI_STATUS             Status;
    UINT32                  Mindex;
    UINT32                  Pindex;


    FUNCTION_TRACE_PTR ("DsMethodDataInitArgs", Params);


    if (!Params)
    {
        DEBUG_PRINT (TRACE_EXEC, ("DsMethodDataInitArgs: No param list passed to method\n"));
        return_ACPI_STATUS (AE_OK);
    }

    /* Copy passed parameters into the new method stack frame  */
        
    for (Pindex = Mindex = 0; (Mindex < MTH_NUM_ARGS) && (Pindex < MaxParamCount); Mindex++)
    {   
        if (Params[Pindex])   
        {
            /*
             * A valid parameter.
             * Set the current method argument to the Params[Pindex++] argument object descriptor   
             */
            Status = DsMethodDataSetValue (MTH_TYPE_ARG, Mindex, Params[Pindex]);
            if (ACPI_FAILURE (Status))
            {
                break;
            }

            Pindex++;
        }

        else
        {
            break;
        }
    }

    DEBUG_PRINT (TRACE_EXEC, ("DsMethodDataInitArgs: %d args passed to method\n", Pindex));
    return_ACPI_STATUS (AE_OK);
}




ACPI_STATUS
DsMethodDataGetEntry (
    UINT32                  Type,
    UINT32                  Index,
    ACPI_OBJECT_INTERNAL    ***Entry)
{
    ACPI_WALK_STATE         *WalkState;


    FUNCTION_TRACE_U32 ("DsMethodDataGetEntry", Index);


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);

    /* 
     * Get the requested object.
     * The stack "Type" is either a LocalVariable or an Argument 
     */

    switch (Type)
    {

    case MTH_TYPE_LOCAL:

        if (Index > MTH_MAX_LOCAL)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: LocalVar index %d is invalid (max %d)\n",
                                    Index, MTH_MAX_LOCAL));
            return_ACPI_STATUS (AE_BAD_PARAMETER);
        }
        
        *Entry = (ACPI_OBJECT_INTERNAL **) &WalkState->LocalVariables[Index].Object;
        break;


    case MTH_TYPE_ARG:

        if (Index > MTH_MAX_ARG)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: Argument index %d is invalid (max %d)\n",
                                    Index, MTH_MAX_ARG));
            return_ACPI_STATUS (AE_BAD_PARAMETER);
        }
        
        *Entry = (ACPI_OBJECT_INTERNAL **) &WalkState->Arguments[Index].Object;
        break;


    default:
        DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: Stack type %d is invalid\n",
                                Type));
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsMethodDataSetEntry (
    UINT32                  Type,
    UINT32                  Index,
    ACPI_OBJECT_INTERNAL    *Object)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **Entry;


    FUNCTION_TRACE ("DsMethodDataSetEntry");


    /* Get a pointer to the stack entry to set */

    Status = DsMethodDataGetEntry (Type, Index, &Entry);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Increment ref count so object can't be deleted while installed */

    CmAddReference (Object);
    
    /* Install the object into the stack entry */

    *Entry = Object;

    return_ACPI_STATUS (AE_OK);
}



ACPI_OBJECT_TYPE
DsMethodDataGetType (
    UINT32                  Type,
    UINT32                  Index)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **Entry;
    ACPI_OBJECT_INTERNAL    *Object;


    FUNCTION_TRACE ("DsMethodDataGetType");


    /* Get a pointer to the requested stack entry */

    Status = DsMethodDataGetEntry (Type, Index, &Entry);
    if (ACPI_FAILURE (Status))
    {
        return_VALUE ((ACPI_OBJECT_TYPE) -1);
    }

    /* Get the object from the method stack */

    Object = *Entry;

    /* Get the object type */

    if (!Object)
    {
        return_VALUE (ACPI_TYPE_Any); /* Any == 0 => "uninitialized" -- see spec 15.2.3.5.2.28 */
    }

    return_VALUE (Object->Common.Type);
}



NAME_TABLE_ENTRY *
DsMethodDataGetNte (
    UINT32                  Type,
    UINT32                  Index)
{
    NAME_TABLE_ENTRY        *Entry = NULL;
    ACPI_WALK_STATE         *WalkState;


    FUNCTION_TRACE ("DsMethodDataGetNte");


    WalkState = DsGetCurrentWalkState (Gbl_CurrentWalkList);


    switch (Type)
    {

    case MTH_TYPE_LOCAL:

        if (Index > MTH_MAX_LOCAL)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: LocalVar index %d is invalid (max %d)\n",
                                    Index, MTH_MAX_LOCAL));
            return_VALUE (Entry);
        }
        
        Entry =  &WalkState->LocalVariables[Index];
        break;


    case MTH_TYPE_ARG:

        if (Index > MTH_MAX_ARG)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: Argument index %d is invalid (max %d)\n",
                                    Index, MTH_MAX_ARG));
            return_VALUE (Entry);
        }
        
        Entry = &WalkState->Arguments[Index];
        break;


    default:
        DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetEntry: Stack type %d is invalid\n",
                                Type));
        break;
    }


    return_VALUE (Entry);
}



ACPI_STATUS
DsMethodDataGetValue (
    UINT32                  Type,
    UINT32                  Index, 
    ACPI_OBJECT_INTERNAL    **DestDesc)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **Entry;
    ACPI_OBJECT_INTERNAL    *Object;


    FUNCTION_TRACE ("DsMethodDataGetValue");


    /* Validate the object descriptor */

    if (!DestDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetValue: NULL object descriptor pointer\n"));
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    

    /* Get a pointer to the requested method stack entry */

    Status = DsMethodDataGetEntry (Type, Index, &Entry);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Get the object from the method stack */

    Object = *Entry;


    /* Examine the returned object, it must be valid. */

    if (!Object)
    {
        /* 
         * Index points to uninitialized object stack value.  This means
         * that either 1) The expected argument was not passed to the method,
         * or 2) A local variable was referenced by the method (via the ASL)
         * before it was initialized.  Either case is an error.
         */ 

        switch (Type)
        {
        case MTH_TYPE_ARG:
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetValue: Uninitialized Arg[%d] at entry %X\n",
                            Index, Entry));
            return_ACPI_STATUS (AE_AML_UNINITIALIZED_ARG);
            break;

        case MTH_TYPE_LOCAL:
            DEBUG_PRINT (ACPI_ERROR, ("DsMethodDataGetValue: Uninitialized Local[%d] at entry %X\n",
                            Index, Entry));
            return_ACPI_STATUS (AE_AML_UNINITIALIZED_LOCAL);
            break;
        }
    }


    /* 
     * Index points to initialized and valid object stack value.
     * Return an additional reference to the object
     */

    *DestDesc = Object;
    CmAddReference (Object);

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsMethodDataDeleteValue (
    UINT32                  Type,
    UINT32                  Index) 
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **Entry;
    ACPI_OBJECT_INTERNAL    *Object;


    FUNCTION_TRACE ("DsMethodDataDeleteValue");


    /* Get a pointer to the requested entry */

    Status = DsMethodDataGetEntry (Type, Index, &Entry);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Get the current entry in this slot k */

    Object = *Entry;

    /* 
     * Undefine the Arg or Local by setting its descriptor pointer to NULL.
     * Locals/Args can contain both ACPI_OBJECT_INTERNALs and NAME_TABLE_ENTRYs
     */
    *Entry = NULL;


    if ((Object) &&
        (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_ACPI_OBJ)))
    {
        /*
         * There is a valid object in this slot
         * Decrement the reference count by one to balance the increment when the 
         * object was stored in the slot.
         */

        CmRemoveReference (Object);
    }


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsMethodDataSetValue (
    UINT32                  Type,
    UINT32                  Index, 
    ACPI_OBJECT_INTERNAL    *SrcDesc) 
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **Entry;


    FUNCTION_TRACE ("DsMethodDataSetValue");
    DEBUG_PRINT (TRACE_EXEC, ("DsMethodDataSetValue: Type=%d Idx=%d Obj=%p\n",
                    Type, Index, SrcDesc));


    /* Parameter validation */

    if (!SrcDesc)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* Get a pointer to the requested method stack entry */

    Status = DsMethodDataGetEntry (Type, Index, &Entry);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    if (*Entry == SrcDesc)
    {
        DEBUG_PRINT (TRACE_EXEC, ("DsMethodDataSetValue: Obj=%p already installed!\n",
                        SrcDesc));
        goto Cleanup;
    }


    /*
     * If there is an object already in this slot, we either have to delete it, or if
     * this is an argument and there is an object reference stored there, we have to do
     * an indirect store!
     */

    if (*Entry)
    {
        /*
         * Check for an indirect store if an argument contains an object reference (stored as an NTE).
         * We don't allow this automatic dereferencing for locals, since a store to a local should overwrite
         * anything there, including an object reference.
         *
         * If both Arg0 and Local0 contain RefOf (Local4):
         *
         * Store (1, Arg0)                  - Causes indirect store to local4
         * Store (1, Local0)                - Stores 1 in local0, overwriting the reference to local4
         * Store (1, DeRefof (Local0))      - Causes indirect store to local4
         *
         * Weird, but true.
         */

        if ((Type == MTH_TYPE_ARG) &&
            (VALID_DESCRIPTOR_TYPE (*Entry, DESC_TYPE_NTE)))
        {
            DEBUG_PRINT (TRACE_EXEC, ("DsMethodDataSetValue: Arg (%p) is an ObjRef(NTE), storing in %p\n",
                            SrcDesc, *Entry));

            /* Detach an existing object from the NTE */

            NsDetachObject (*Entry);

            /* Store this object into the NTE (do the indirect store) */

            Status = NsAttachObject (*Entry, SrcDesc, SrcDesc->Common.Type);
            return_ACPI_STATUS (Status);
        }


        /* Otherwise, just delete the existing object before storing the new one */

        DsMethodDataDeleteValue (Type, Index);
    }


    /* 
     * Install the ObjStack descriptor (*SrcDesc) into the descriptor for the
     * Arg or Local.
     * Install the new object in the stack entry
     * (increments the object reference count by one)
     */

    Status = DsMethodDataSetEntry (Type, Index, SrcDesc);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    /* Normal exit */

    return_ACPI_STATUS (AE_OK);


    /* Error exit */

Cleanup:

    return_ACPI_STATUS (Status);
}

