
/*
 *  ./acpica/subsystem/dispatcher/dsobject.c - Dispatcher object management routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSOBJECT_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>

#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dsobject");




ACPI_STATUS
DsInitOneObject (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_OBJECT_TYPE        Type;
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    INIT_WALK_INFO          *Info = (INIT_WALK_INFO *) Context;


    /* We are only interested in objects owned by the table that was just loaded */

    if (((NAME_TABLE_ENTRY *) ObjHandle)->OwnerId != Info->TableDesc->TableId)
    {
        return AE_OK;
    }


    /* And even then, we are only interested in a few object types */

    Type = NsGetType (ObjHandle);

    switch (Type)
    {

    case ACPI_TYPE_Region:

        DsInitializeRegion (ObjHandle);

        Info->OpRegionCount++;
        break;


    case ACPI_TYPE_Method:

        Info->MethodCount++;

        //DEBUG_PRINT_RAW (ACPI_OK, ("."));

        /* Always parse methods to detect errors, we may delete the parse tree below */

        Status = DsParseMethod (ObjHandle);

        /* TBD: what do we do with an error? */

        if (ACPI_FAILURE (Status))
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsInitOneObject: Method %p [%4.4s] parse failed! %s\n", 
                                        ObjHandle, &((NAME_TABLE_ENTRY *)ObjHandle)->Name, CmFormatException (Status)));
            break;
        }

        /* Keep the parse tree only if we are parsing all methods at init time (versus just-in-time) */

        if (Gbl_WhenToParseMethods != METHOD_PARSE_AT_INIT)
        {                
            
            NsDeleteNamespaceSubtree (ObjHandle);

            ObjDesc = ((NAME_TABLE_ENTRY *)ObjHandle)->Object; 
            PsDeleteParseTree (ObjDesc->Method.ParserOp);
            ObjDesc->Method.ParserOp = NULL;
        }

        break;

    default:
        break;
    }

    /*
     * We ignore errors from above, and always return OK, since
     * we don't want to abort the walk on a single error.
     */
    return (AE_OK);
}



ACPI_STATUS
DsInitializeObjects (
    ACPI_TABLE_DESC         *TableDesc,
    NAME_TABLE_ENTRY        *StartEntry)
{
    ACPI_STATUS             Status;
    INIT_WALK_INFO          Info;


    FUNCTION_TRACE ("DsInitializeObjects");


    DEBUG_PRINT (TRACE_DISPATCH, ("DsInitializeObjects: **** Starting initialization of namespace objects ****\n"));
    //DEBUG_PRINT_RAW (ACPI_OK, ("Parsing Methods:"));

    Info.MethodCount = 0;
    Info.OpRegionCount = 0;
    Info.TableDesc = TableDesc;


    /* Walk entire namespace from the supplied root */

    Status = AcpiWalkNamespace (ACPI_TYPE_Any, StartEntry, ACPI_INT32_MAX, DsInitOneObject, 
                                &Info, NULL);
    if (ACPI_FAILURE (Status))
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsInitializeObjects: WalkNamespace failed! %x\n", Status));
    }

    DEBUG_PRINT_RAW (ACPI_OK, ("%d Control Methods found and parsed\n", Info.MethodCount));
    DEBUG_PRINT (TRACE_DISPATCH, ("DsInitializeObjects: %d Control Methods found\n", Info.MethodCount));
    DEBUG_PRINT (TRACE_DISPATCH, ("DsInitializeObjects: %d Op Regions found\n", Info.OpRegionCount));

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsInitObjectFromOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    UINT16                  Opcode,
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status;
    ACPI_GENERIC_OP         *Arg;
    ACPI_BYTELIST_OP        *ByteList;
    ACPI_OBJECT_INTERNAL    *ArgDesc;
    ACPI_OP_INFO            *OpInfo;



    OpInfo = PsGetOpcodeInfo (Opcode);
    if (!OpInfo)
    {
        /* Unknown opcode */

        return AE_TYPE;
    }



    /* Get and prepare the first argument */

    switch (ObjDesc->Common.Type)
    {
    case ACPI_TYPE_Buffer:

        /* First arg is a number */

        DsCreateOperand (WalkState, Op->Value.Arg);
        ArgDesc = WalkState->Operands [WalkState->NumOperands - 1];
        DsObjStackPop (1, WalkState);

        /* Resolve the object (could be an arg or local) */

        Status = AmlResolveToValue (&ArgDesc);
        if (ACPI_FAILURE (Status))
        {
            CmRemoveReference (ArgDesc);
            return Status;
        }

        /* We are expecting a number */

        if (ArgDesc->Common.Type != ACPI_TYPE_Number)
        {
            DEBUG_PRINT (ACPI_ERROR, ("InitObject: Expecting number, got obj: %p type %X\n",
                            ArgDesc, ArgDesc->Common.Type));
            CmRemoveReference (ArgDesc);
            return AE_TYPE;
        }
 
        /* Get the value, delete the internal object */

        ObjDesc->Buffer.Length = ArgDesc->Number.Value;
        CmRemoveReference (ArgDesc);

        /* Allocate the buffer */

        ObjDesc->Buffer.Pointer = CmCallocate (ObjDesc->Buffer.Length);
        if (!ObjDesc->Buffer.Pointer)
        {
            return AE_NO_MEMORY;
        }

        /* Second arg is the buffer data (optional) */
        /* ByteList can be either individual bytes or a string initializer! */

        Arg = Op->Value.Arg;        /* skip first arg */
        ByteList = (ACPI_BYTELIST_OP *) Arg->Next;
        if (ByteList)
        {
            if (ByteList->Opcode != AML_BYTELIST_OP)
            {
                DEBUG_PRINT (ACPI_ERROR, ("InitObject: Expecting bytelist, got: %x\n",
                                ByteList));
                return AE_TYPE;
            }

            MEMCPY (ObjDesc->Buffer.Pointer, ByteList->Data, ObjDesc->Buffer.Length);
        }

        break;


    case ACPI_TYPE_Number:
        ObjDesc->Number.Value = Op->Value.Integer;
        break;


    case ACPI_TYPE_String:
        ObjDesc->String.Pointer = Op->Value.String;
        ObjDesc->String.Length = STRLEN (Op->Value.String);
        break;


    case ACPI_TYPE_Method:
        break;


    case INTERNAL_TYPE_Reference:

        switch (OpInfo->Flags & OP_INFO_TYPE)
        {
        case OPTYPE_LOCAL_VARIABLE:

            /* Split the opcode into a base opcode + offset */

            ObjDesc->Reference.OpCode = AML_LocalOp;
            ObjDesc->Reference.Offset = Opcode - AML_LocalOp;
            break;

        case OPTYPE_METHOD_ARGUMENT:

            /* Split the opcode into a base opcode + offset */

            ObjDesc->Reference.OpCode = AML_ArgOp;
            ObjDesc->Reference.Offset = Opcode - AML_ArgOp;
            break;

        default: /* Constants, Literals, etc.. */

            if (Op->Opcode == AML_NAMEPATH_OP)
            {
                /* Nte was saved in Op */

                ObjDesc->Reference.Nte = Op->NameTableEntry;
            }

            ObjDesc->Reference.OpCode = Opcode;
            break;
        }

        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("InitObject: Unimplemented data type: %x\n",
                        ObjDesc->Common.Type));

        break;
    }

    return AE_OK;
}




ACPI_STATUS
DsBuildInternalSimpleObj (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    ACPI_OBJECT_INTERNAL    **ObjDescPtr)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_TYPE        Type;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("DsBuildInternalSimpleObj");


    if (Op->Opcode == AML_NAMEPATH_OP)
    {
        /*
         * This is an object reference.  If The name was previously looked up in the NS,
         * it is stored in this op.  Otherwise, go ahead and look it up now
         */

        if (!Op->NameTableEntry)
        {
            Status = NsLookup (WalkState->ScopeInfo, Op->Value.String, ACPI_TYPE_Any, IMODE_Execute, 
                                        NS_SEARCH_PARENT | NS_DONT_OPEN_SCOPE, NULL, (NAME_TABLE_ENTRY **)&(Op->NameTableEntry));
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }
        }

        /*
         * The reference will be a Reference
         * TBD: unless we really need a separate type of INTERNAL_TYPE_reference
         * TBD: change DsMapOpcodeToDataType to handle this case 
         */
        Type = INTERNAL_TYPE_Reference;
    }

    else
    {
        Type = DsMapOpcodeToDataType (Op->Opcode, NULL);
    }


    /* Create and init the internal ACPI object */

    ObjDesc = CmCreateInternalObject (Type);
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    Status = DsInitObjectFromOp (WalkState, Op, Op->Opcode, ObjDesc);
    if (ACPI_FAILURE (Status))
    {
        CmRemoveReference (ObjDesc);
        return_ACPI_STATUS (Status);
    }

    *ObjDescPtr = ObjDesc;

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
DsBuildInternalPackageObj (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    ACPI_OBJECT_INTERNAL    **ObjDescPtr)
{
    ACPI_GENERIC_OP         *Arg;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("DsBuildInternalPackageObj");


    ObjDesc = CmCreateInternalObject (ACPI_TYPE_Package);
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* The first argument must be the package length */

    Arg = Op->Value.Arg;
    ObjDesc->Package.Count = Arg->Value.Integer;

    /* 
     * Allocate the array of pointers (ptrs to the individual objects)
     * Add an extra pointer slot so that the list is always null terminated.
     */

    ObjDesc->Package.Elements   = CmCallocate ((ObjDesc->Package.Count + 1) * sizeof (void *));
    if (!ObjDesc->Package.Elements)
    {
        /* Package vector allocation failure   */

        REPORT_ERROR ("DsBuildInternalPackageObj: Package vector allocation failure");
        CmFree (ObjDesc);
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    ObjDesc->Package.NextElement = ObjDesc->Package.Elements;


    /*
     * Now init the elements of the package
     */

    Arg = Arg->Next;
    while (Arg)
    {
        if (Arg->Opcode == AML_PackageOp)
        {
            Status = DsBuildInternalPackageObj (WalkState, Arg, ObjDesc->Package.NextElement);
        }
        
        else
        {
            Status = DsBuildInternalSimpleObj (WalkState, Arg, ObjDesc->Package.NextElement);
        }

        ObjDesc->Package.NextElement++;
        Arg = Arg->Next;
    }

    *ObjDescPtr = ObjDesc;
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsBuildInternalObject (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    ACPI_OBJECT_INTERNAL    **ObjDescPtr)
{
    ACPI_STATUS             Status;


    if (Op->Opcode == AML_PackageOp)
    {
        Status = DsBuildInternalPackageObj (WalkState, Op, ObjDescPtr);
    }
    
    else
    {
        Status = DsBuildInternalSimpleObj (WalkState, Op, ObjDescPtr);
    }

    return (Status);
}




ACPI_STATUS
DsCreateNamedObject (
    ACPI_WALK_STATE         *WalkState,
    NAME_TABLE_ENTRY        *Entry,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;



    FUNCTION_TRACE_PTR ("DsCreateNamedObject", Op);


    if (!Op->Value.Arg)
    {
        /* No arguments, there is nothing to do */

        return_ACPI_STATUS (AE_OK);
    }


    /* Build an internal object for the argument(s) */

    Status = DsBuildInternalObject (WalkState, Op->Value.Arg, &ObjDesc);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }


    /* Re-type the object according to it's argument */

    Entry->Type = ObjDesc->Common.Type;
    
    /* Init obj */

    Status = NsAttachObject ((ACPI_HANDLE) Entry, ObjDesc, (UINT8) Entry->Type);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    return_ACPI_STATUS (Status);



Cleanup:

    CmRemoveReference (ObjDesc);

    return_ACPI_STATUS (Status);
}



