
/*
 *  ./acpica/subsystem/dispatcher/dsregion.c - Dispatcher Op Region support
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSREGION_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <events.h>
#include <tables.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dsregion");




ACPI_STATUS
DsGetRegionArguments (
    ACPI_OBJECT_INTERNAL    *RgnDesc)
{
    ACPI_OBJECT_INTERNAL    *MethodDesc;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_GENERIC_OP         *Op;
    ACPI_GENERIC_OP         *RegionOp;
    ACPI_STATUS             Status;
    ACPI_TABLE_DESC         *TableDesc;


    FUNCTION_TRACE_PTR ("DsGetRegionArguments", RgnDesc);


    if (RgnDesc->Region.RegionFlags & REGION_AGRUMENT_DATA_VALID)
    {
        return_ACPI_STATUS (AE_OK);
    }


    MethodDesc = RgnDesc->Region.Method;
    Entry = RgnDesc->Region.Nte;


    /* Allocate a new parser op to be the root of the parsed OpRegion tree */

    Op = PsAllocOp (AML_RegionOp);
    if (!Op)
    {
        return AE_NO_MEMORY;
    }

    /* Save the NTE for use in PsParseAml */

    Op->NameTableEntry = Entry->ParentEntry;

    /* Get a handle to the parent ACPI table */

    Status = TbHandleToObject (Entry->OwnerId, &TableDesc);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Parse the entire OpRegion declaration, creating a parse tree */

    Status = PsParseAml (Op, MethodDesc->Method.Pcode, MethodDesc->Method.PcodeLength);
    if (ACPI_SUCCESS (Status))
    {
        /* Get and init the actual RegionOp created above */

        RegionOp = Op->Value.Arg;
        RegionOp->NameTableEntry = Entry;

        /* Evaluate the address and length arguments for the OpRegion */

        PsWalkParsedAml (RegionOp, RegionOp, NULL, NULL, NULL, NULL, TableDesc->TableId, DsExecBeginOp, DsExecEndOp);
    }

    /* All done with the parse tree, delete it */

    PsDeleteParseTree (Op);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsInitializeRegion (
    ACPI_HANDLE             ObjHandle)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;


    ObjDesc = NsGetAttachedObject (ObjHandle);
    Status = EvInitializeRegion (ObjDesc, FALSE);       /* Namespace is NOT locked */

    return Status;
}



ACPI_STATUS
DsEvalRegionOperands (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *RegionDesc;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_GENERIC_OP         *NextOp;


    FUNCTION_TRACE_PTR ("DsEvalRegionOperands", Op);



    /*
     * This is where we evaluate the address and length fields of the OpRegion declaration
     */

    Entry =  Op->NameTableEntry; 

    NextOp = Op->Value.Arg;     /* NextOp points to the op that holds the SpaceID */
    NextOp = NextOp->Next;      /* NextOp points to address op */

    /* Evaluate/create the address and length operands */

    Status = DsCreateOperands (WalkState, NextOp);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    RegionDesc = NsGetAttachedObject (Entry);
    if (!RegionDesc)
    {
        return_ACPI_STATUS (AE_NOT_EXIST);
    }

    /* Get the length and save it */

    ObjDesc = WalkState->Operands[WalkState->NumOperands -1];       /* Top of stack */

    RegionDesc->Region.Length = ObjDesc->Number.Value;
    CmRemoveReference (ObjDesc);

    /* Get the address and save it */

    ObjDesc = WalkState->Operands[WalkState->NumOperands -2];       /* Top of stack -1 */

    RegionDesc->Region.Address = ObjDesc->Number.Value;
    CmRemoveReference (ObjDesc);
    

    DEBUG_PRINT (TRACE_EXEC, ("DsEvalRegionOperands: RgnObj %p Addr %X Len %X\n", 
                    RegionDesc, RegionDesc->Region.Address, RegionDesc->Region.Length));

    /* Now the address and length are valid for this opregion */

    RegionDesc->Region.RegionFlags |= REGION_AGRUMENT_DATA_VALID;

    return_ACPI_STATUS (Status);
}
