
/*
 *  ./acpica/subsystem/dispatcher/dsutils.c - Dispatcher utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSUTILS_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <debugger.h>

#define _COMPONENT          PARSER
        MODULE_NAME         ("dsutils");




void
DsDeleteResultIfNotUsed (
    ACPI_GENERIC_OP         *Op,
    ACPI_OBJECT_INTERNAL    *ResultObj,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_OP_INFO            *ParentInfo;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("DsDeleteResultIfNotUsed", ResultObj);


    if (!Op)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsDeleteResultIfNotUsed: Null Op=%X\n",
                        Op));
        return_VOID;
    }

    if (!ResultObj)
    {
        return_VOID;
    }

    if (!Op->Parent)
    {
        /* 
         * If there is no parent, the result can't possibly be used!
         * (An executing method typically has no parent, since each method is parsed separately
         */

        /* Must pop the result stack (ObjDesc should be equal to ResultObj) */

        Status = DsResultStackPop (&ObjDesc, WalkState);
        if (ACPI_FAILURE (Status))
        {
            return;
        }

        CmRemoveReference (ResultObj);

        return_VOID;
    }


    /*
     * Get info on the parent.  The root Op is AML_Scope
     */

    ParentInfo = PsGetOpcodeInfo (Op->Parent->Opcode);
    if (!ParentInfo)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsDeleteResultIfNotUsed: Unknown parent opcode. Op=%X\n",
                        Op));

        return_VOID;
    }


    /* Never delete the return value associated with a return opcode */

    if (Op->Parent->Opcode == AML_ReturnOp)
    {
        DEBUG_PRINT (TRACE_DISPATCH, ("DsDeleteResultIfNotUsed: No delete, [RETURN] opcode=%X Op=%X\n",
                        Op->Opcode, Op));
        return_VOID;
    }


    /*
     * Decide what to do with the result based on the parent.  If the parent opcode
     * will not use the result, delete the object.  Otherwise leave it as is, it will
     * be deleted when it is used as an operand later.
     */

    switch (ParentInfo->Flags & OP_INFO_TYPE)
    {
    /*
     * In these cases, the parent will never use the return object, so delete it 
     * here and now.
     */
    case OPTYPE_CONTROL:        /* IF, ELSE, WHILE only */
    case OPTYPE_NAMED_OBJECT:   /* Scope, method, etc. */

        DEBUG_PRINT (TRACE_DISPATCH, ("DsDeleteResultIfNotUsed: Deleting result, Parent opcode=%X Op=%X\n",
                        Op->Opcode, Op));

        /* Must pop the result stack (ObjDesc should be equal to ResultObj) */

        Status = DsResultStackPop (&ObjDesc, WalkState);
        if (ACPI_FAILURE (Status))
        {
            return_VOID;
        }

        CmRemoveReference (ResultObj);
        break;

    /* 
     * In all other cases. the parent will actually use the return object, so keep it.
     */
    default:
        break;
    }

    return_VOID;
}





ACPI_STATUS
DsCreateOperand (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Arg)
{
    ACPI_STATUS             Status = AE_OK;
    char                    *NameString;
    UINT32                  NameLength;
    ACPI_OBJECT_TYPE        DataType; 
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_GENERIC_OP         *ParentOp;
    UINT16                  Opcode;
    UINT32                  Flags;
    OPERATING_MODE          InterpreterMode;


    FUNCTION_TRACE_PTR ("DsCreateOperand", Arg);


    /* A valid name must be looked up in the namespace */

    if ((Arg->Opcode == AML_NAMEPATH_OP) &&
        (Arg->Value.String))
    {
        DEBUG_PRINT (TRACE_DISPATCH, ("DsCreateOperand: Getting a name: Arg=%p\n", Arg));

        /* Get the entire name string from the AML stream */

        Status = AmlGetNameString (ACPI_TYPE_Any, Arg->Value.Buffer, &NameString, &NameLength);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        /* All prefixes have been handled, and the name is in NameString */

        /*
         * Differentiate between a namespace "create" operation versus a "lookup" operation 
         * (IMODE_LoadPass2 vs. IMODE_Execute) in order to support the creation of namespace
         * objects during the execution of control methods.
         */

        ParentOp = Arg->Parent;
        if ((PsIsNamedObjectOp (ParentOp->Opcode)) &&
            (ParentOp->Opcode != AML_METHODCALL_OP) &&
            (ParentOp->Opcode != AML_NAMEPATH_OP))
        {
            /* Enter name into namespace if not found */

            InterpreterMode = IMODE_LoadPass2;
        }

        else
        {
            /* Return a failure if name not found */

            InterpreterMode = IMODE_Execute;
        }

        Status = NsLookup (WalkState->ScopeInfo, NameString, ACPI_TYPE_Any, InterpreterMode, 
                                    NS_SEARCH_PARENT | NS_DONT_OPEN_SCOPE, WalkState, (NAME_TABLE_ENTRY **) &ObjDesc);

        /* Free the namestring created above */

        CmFree (NameString);

        /*
         * The only case where we pass through (ignore) a NOT_FOUND error is for the 
         * CondRefOf opcode.
         */

        if (Status == AE_NOT_FOUND) 
        {
            if (ParentOp->Opcode == AML_CondRefOfOp)
            {
                /*
                 * For the Conditional Reference op, it's OK if the name is not found;  We
                 * just need a way to indicate this to the interpreter, set the object to the root
                 */
                ObjDesc = (ACPI_OBJECT_INTERNAL *) Gbl_RootObject;
                Status = AE_OK;
            }

            else
            {
                /* We just plain didn't find it -- which is a very serious error at this point */

                Status = AE_AML_NAME_NOT_FOUND;
            }
        }

        /* Check status from the lookup */
 
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        /* Put the resulting object onto the current object stack */

        Status = DsObjStackPush (ObjDesc, WalkState);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
        DEBUG_EXEC (DbDisplayArgumentObject (ObjDesc));
    }


    else
    {
        /* Check for null name case */

        if (Arg->Opcode == AML_NAMEPATH_OP)
        {
            /*
             * If the name is null, this means that this is an optional result parameter that was
             * not specified in the original ASL.  Create an Reference for a placeholder 
             */
            Opcode = AML_ZeroOp;        /* Has no arguments! */

            DEBUG_PRINT (TRACE_DISPATCH, ("DsCreateOperand: Null namepath: Arg=%p\n", Arg));

            /* TBD: anything else needed for the zero op lvalue? */
        }

        else
        {
            Opcode = Arg->Opcode;
        }


        /* Get the data type of the argument */

        DataType = DsMapOpcodeToDataType (Opcode, &Flags);
        if (DataType == INTERNAL_TYPE_Invalid)
        {
            return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
        }

        if (Flags & OP_HAS_RETURN_VALUE)
        {
            DEBUG_PRINT (TRACE_DISPATCH, ("DsCreateOperand: Argument previously created, already stacked \n"));

//            DEBUG_EXEC (DbDisplayArgumentObject (WalkState->Operands [WalkState->NumOperands - 1]));

            /* 
             * Use value that was already previously returned by the evaluation of this argument
             */

            Status = DsResultStackPop (&ObjDesc, WalkState);
            if (ACPI_FAILURE (Status))
            {
                /*
                 * Only error is underflow, and this indicates a missing or null operand!
                 */
                DEBUG_PRINT (ACPI_ERROR, ("DsCreateOperand: Could not pop result\n"));
                return_ACPI_STATUS (Status);
            }

        }

        else
        {
            /* Create an ACPI_INTERNAL_OBJECT for the argument */

            ObjDesc = CmCreateInternalObject (DataType);
            if (!ObjDesc)
            {
                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            /* Initialize the new object */

            Status = DsInitObjectFromOp (WalkState, Arg, Opcode, ObjDesc);
            if (ACPI_FAILURE (Status))
            {
                CmFree (ObjDesc);
                return_ACPI_STATUS (Status);
            }
       }

        /* Put the operand object on the object stack */

        Status = DsObjStackPush (ObjDesc, WalkState);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        DEBUG_EXEC (DbDisplayArgumentObject (ObjDesc));

    }


    return_ACPI_STATUS (AE_OK);
}




ACPI_STATUS
DsCreateOperands (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *FirstArg)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_GENERIC_OP         *Arg;
    UINT32                  ArgsPushed = 0;
 

    FUNCTION_TRACE_PTR ("DsCreateOperands", FirstArg);

    Arg = FirstArg;


    /* For all arguments in the list... */

    while (Arg)
    {

        Status = DsCreateOperand (WalkState, Arg);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        DEBUG_PRINT (TRACE_DISPATCH, ("DsCreateOperands: Arg #%d (%p) done, Arg1=%p\n", 
                        ArgsPushed, Arg, FirstArg));

        /* Move on to next argument, if any */

        Arg = Arg->Next;
        ArgsPushed++;
    }

    return_ACPI_STATUS (Status);


Cleanup:
    /* 
     * We must undo everything done above; meaning that we must pop everything off
     * of the operand stack and delete those objects 
     */

    DsObjStackPopAndDelete (ArgsPushed, WalkState);

    DEBUG_PRINT (ACPI_ERROR, ("DsCreateOperands: Error while creating Arg %d - %s\n",
                    (ArgsPushed+1), CmFormatException (Status)));
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsResolveOperands (
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  i;
    ACPI_STATUS             Status = AE_OK;

    
    FUNCTION_TRACE_PTR ("DsResolveOperands", WalkState);


    /* 
     * Attempt to resolve each of the valid operands 
     * Method arguments are passed by value, not by reference
     */
    
    /*
     * TBD: Note from previous parser:
     *
     * TBD: RefOf problem with AmlResolveToValue() conversion.
     */

    for (i = 0; i < WalkState->NumOperands; i++)
    {
        Status = AmlResolveToValue (&WalkState->Operands[i]);
        if (ACPI_FAILURE (Status))
        {
            break;
        }
    }

    return_ACPI_STATUS (Status);
}



ACPI_OBJECT_TYPE
DsMapOpcodeToDataType (
    UINT16                  Opcode,
    UINT32                  *OutFlags)
{
    ACPI_OBJECT_TYPE        DataType = INTERNAL_TYPE_Invalid;
    ACPI_OP_INFO            *OpInfo;
    UINT32                  Flags = 0;


    OpInfo = PsGetOpcodeInfo (Opcode);
    if (!OpInfo)
    {
        /* Unknown opcode */

        return DataType;
    }

    switch (OpInfo->Flags & OP_INFO_TYPE)
    {

    case OPTYPE_LITERAL:

        switch (Opcode)
        {
        case AML_ByteOp:
        case AML_WordOp:
        case AML_DWordOp:

            DataType = ACPI_TYPE_Number;
            break;


        case AML_StringOp:

            DataType = ACPI_TYPE_String;
            break;

        case AML_NAMEPATH_OP:
            DataType = INTERNAL_TYPE_Reference;
            break;
        }

        break;


    case OPTYPE_DATA_TERM:

        switch (Opcode)
        {
        case AML_BufferOp:

            DataType = ACPI_TYPE_Buffer;
            break;

        case AML_PackageOp:

            DataType = ACPI_TYPE_Package;
            break;
        }

        break;


    case OPTYPE_CONSTANT:
    case OPTYPE_METHOD_ARGUMENT:
    case OPTYPE_LOCAL_VARIABLE:

        DataType = INTERNAL_TYPE_Reference;
        break;


    case OPTYPE_MONADIC2:
    case OPTYPE_MONADIC2R:
    case OPTYPE_DYADIC2:
    case OPTYPE_DYADIC2R:
    case OPTYPE_DYADIC2S:
    case OPTYPE_INDEX:
    case OPTYPE_MATCH:

        Flags = OP_HAS_RETURN_VALUE;
        DataType = ACPI_TYPE_Any;

        break;

    case OPTYPE_METHOD_CALL:

        Flags = OP_HAS_RETURN_VALUE;
        DataType = ACPI_TYPE_Method;

        break;


    case OPTYPE_NAMED_OBJECT:

        DataType = DsMapNamedOpcodeToDataType (Opcode);

        break;


    case OPTYPE_DYADIC1:
    case OPTYPE_CONTROL:

        /* No mapping needed at this time */

        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("MapOpcode: Unimplemented data type opcode: %x\n",
                        Opcode));
        break;
    }

    /* Return flags to caller if requested */

    if (OutFlags)
    {
        *OutFlags = Flags;
    }

    return DataType;
}



ACPI_OBJECT_TYPE 
DsMapNamedOpcodeToDataType (
    UINT16                  Opcode)
{
    ACPI_OBJECT_TYPE        DataType; 


    /* Decode Opcode */

    switch (Opcode)
    {
    case AML_ScopeOp:
        DataType = INTERNAL_TYPE_Scope;
        break;

    case AML_DeviceOp:
        DataType = ACPI_TYPE_Device;
        break;

    case AML_ThermalZoneOp:
        DataType = ACPI_TYPE_Thermal;
        break;

    case AML_MethodOp:
        DataType = ACPI_TYPE_Method;
        break;

    case AML_PowerResOp:
        DataType = ACPI_TYPE_Power;
        break;

    case AML_ProcessorOp:
        DataType = ACPI_TYPE_Processor;
        break;

    case AML_DefFieldOp:                            /* DefFieldOp */
        DataType = INTERNAL_TYPE_DefFieldDefn;
        break;

    case AML_IndexFieldOp:                          /* IndexFieldOp */
        DataType = INTERNAL_TYPE_IndexFieldDefn;
        break;

    case AML_BankFieldOp:                           /* BankFieldOp */
        DataType = INTERNAL_TYPE_BankFieldDefn;
        break;

    case AML_NAMEDFIELD_OP:                         /* NO CASE IN ORIGINAL  */
        DataType = ACPI_TYPE_Any;
        break;

    case AML_NameOp:                                /* NameOp - special code in original */
    case AML_NAMEPATH_OP:                    
        DataType = ACPI_TYPE_Any;
        break;

    case AML_AliasOp:
        DataType = INTERNAL_TYPE_Alias;
        break;

    case AML_MutexOp:
        DataType = ACPI_TYPE_Mutex;
        break;

    case AML_EventOp:
        DataType = ACPI_TYPE_Event;
        break;

    case AML_RegionOp:
        DataType = ACPI_TYPE_Region;
        break;


    default:
        DataType = ACPI_TYPE_Any;
        break;

    }

    return DataType;
}



