
/*
 *  ./acpica/subsystem/dispatcher/dswexec.c - Dispatcher method execution callbacks;  
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSWEXEC_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <debugger.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dswexec");





ACPI_STATUS
DsExecBeginOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_OP_INFO            *OpInfo;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("DsExecBeginOp", Op);


    if (Op == WalkState->Origin)
    {
        return_ACPI_STATUS (AE_OK);
    }
    
    /*
     * If the previous opcode was a conditional, this opcode
     * must be the beginning of the associated predicate.
     * Save this knowledge in the current scope descriptor
     */

    if ((WalkState->ControlState) &&
        (WalkState->ControlState->Common.State == CONTROL_CONDITIONAL_EXECUTING))
    {
        DEBUG_PRINT (TRACE_EXEC, ("BeginOp: Exec predicate Op=%X State=%X\n",
                        Op, WalkState));

        WalkState->ControlState->Common.State          = CONTROL_PREDICATE_EXECUTING;
        WalkState->ControlState->Control.PredicateOp    = Op;         /* Save start of predicate */
    }

/*
    if (Op->Opcode != AML_MethodOp)
    {
        return_ACPI_STATUS (AE_OK);
    }
    
*/

    OpInfo = PsGetOpcodeInfo (Op->Opcode);

    /* We want to send name paths to the load code */

    if (Op->Opcode == AML_NAMEPATH_OP)
    {
        OpInfo->Flags = OPTYPE_NAMED_OBJECT;
    }

    switch (OpInfo->Flags & OP_INFO_TYPE)
    {
    case OPTYPE_CONTROL:

        Status = DsExecBeginControlOp (WalkState, Op);
        break;

    case OPTYPE_NAMED_OBJECT:

        if (WalkState->Origin->Opcode == AML_MethodOp)
        {
            /*
             * Found a named object declaration during method execution;  we must enter
             * this object into the namespace.
             *
             * TBD: make this a temporary namespace object
             */

            Status = DsLoad2BeginOp (WalkState, Op);
        }
        break;

    default:
        break;
    }

    /* Nothing to do here during method execution */

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
DsExecEndOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status = AE_OK;
    UINT16                  Opcode;
    UINT8                   Optype;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_GENERIC_OP         *NextOp;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_GENERIC_OP         *FirstArg;
    ACPI_OBJECT_INTERNAL    *ResultObj = NULL;
    ACPI_OP_INFO            *OpInfo;
    UINT32                  OperandIndex;


    FUNCTION_TRACE_PTR ("DsExecEndOp", Op);


    Opcode = (UINT16) Op->Opcode;


    OpInfo = PsGetOpcodeInfo (Op->Opcode);
    if (!OpInfo)
    {
        DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: Unknown opcode. Op=%X\n",
                        Op));

        return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
    }

    Optype = (UINT8) (OpInfo->Flags & OP_INFO_TYPE);
    FirstArg = Op->Value.Arg;

    /* Init the walk state */

    WalkState->NumOperands = 0;
    WalkState->ReturnDesc = NULL;


    /* Call debugger for single step support (DEBUG build only) */

    DEBUG_EXEC (Status = DbSingleStep (WalkState, Op, Optype));
    DEBUG_EXEC (if (Status != AE_OK) {return_ACPI_STATUS (Status);});


    /* Decode the opcode */

    switch (Optype)
    {
    case OPTYPE_UNDEFINED:

        DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: Undefined opcode type Op=%X\n",
                        Op));
        return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
        break;


    case OPTYPE_BOGUS:
        DEBUG_PRINT (TRACE_DISPATCH, ("ExecEndOp: Internal opcode=%X type Op=%X\n",
                        Opcode, Op));
        break;

    case OPTYPE_CONSTANT:           /* argument type only */
    case OPTYPE_LITERAL:            /* argument type only */
    case OPTYPE_DATA_TERM:          /* argument type only */
    case OPTYPE_LOCAL_VARIABLE:     /* argument type only */
    case OPTYPE_METHOD_ARGUMENT:    /* argument type only */
        break;


    /* most operators with arguments */

    case OPTYPE_MONADIC1:
    case OPTYPE_DYADIC1:
    case OPTYPE_MONADIC2:
    case OPTYPE_MONADIC2R:
    case OPTYPE_DYADIC2:
    case OPTYPE_DYADIC2R:
    case OPTYPE_DYADIC2S:
    case OPTYPE_RECONFIGURATION:
    case OPTYPE_INDEX:
    case OPTYPE_MATCH:
    case OPTYPE_CREATE_FIELD:
    case OPTYPE_FATAL:

        Status = DsCreateOperands (WalkState, FirstArg);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        OperandIndex = WalkState->NumOperands - 1;

        switch (Optype)
        {

        case OPTYPE_MONADIC1:

            /* 1 Operand, 0 ExternalResult, 0 InternalResult */

            Status = AmlExecMonadic1 (Opcode, WalkState);
            break;


        case OPTYPE_MONADIC2:

            /* 1 Operand, 0 ExternalResult, 1 InternalResult */

            Status = AmlExecMonadic2 (Opcode, WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_MONADIC2R:

            /* 1 Operand, 1 ExternalResult, 1 InternalResult */

            Status = AmlExecMonadic2R (Opcode, WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_DYADIC1:

            /* 2 Operands, 0 ExternalResult, 0 InternalResult */

            Status = AmlExecDyadic1 (Opcode, WalkState);

            break;


        case OPTYPE_DYADIC2:

            /* 2 Operands, 0 ExternalResult, 1 InternalResult */

            Status = AmlExecDyadic2 (Opcode, WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_DYADIC2R:

            /* 2 Operands, 1 or 2 ExternalResults, 1 InternalResult */


            /* NEW INTERFACE:
             * Pass in WalkState, keep result obj  but let interpreter push the result */

            Status = AmlExecDyadic2R (Opcode, WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_DYADIC2S:   /* Synchronization Operator */

            /* 2 Operands, 0 ExternalResult, 1 InternalResult */

            Status = AmlExecDyadic2S (Opcode, WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_RECONFIGURATION:

            /* 1 or 2 operands, 0 Internal Result */

            Status = AmlExecReconfiguration (Opcode, WalkState);
            break;


        case OPTYPE_CREATE_FIELD:

            /* 3 or 4 Operands, 0 ExternalResult, 0 InternalResult */

            Status = AmlExecCreateField (Opcode, WalkState);
            break;


        case OPTYPE_FATAL:

            /* 3 Operands, 0 ExternalResult, 0 InternalResult */

            Status = AmlExecFatal (WalkState);
            break;


        case OPTYPE_INDEX:  /* Type 2 opcode with 3 operands */

            /* 3 Operands, 1 ExternalResult, 1 InternalResult */

            Status = AmlExecIndex (WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;


        case OPTYPE_MATCH:  /* Type 2 opcode with 6 operands */

            /* 6 Operands, 0 ExternalResult, 1 InternalResult */

            Status = AmlExecMatch (WalkState, &ResultObj);
            if (ACPI_SUCCESS (Status))
            {
                Status = DsResultStackPush (ResultObj, WalkState);
            }

            break;
        }

        break;



    case OPTYPE_CONTROL:    /* Type 1 opcode, IF/ELSE/WHILE/NOOP */

        /* 1 Operand, 0 ExternalResult, 0 InternalResult */

        Status = DsExecEndControlOp (WalkState, Op);

        break;


    case OPTYPE_METHOD_CALL: 

        DEBUG_PRINT (TRACE_DISPATCH, ("ExecEndOp: Method invocation, Op=%X\n", Op));

        /*
         * (AML_METHODCALL) Op->Value->Arg->NameTableEntry contains the method NTE pointer
         */
        NextOp = FirstArg;          /* NextOp points to the op that holds the method name */
        Entry = NextOp->NameTableEntry;

        NextOp = NextOp->Next;      /* NextOp points to first argument op */


        /*
         * Get the method's arguments and put them on the operand stack
         */

        Status = DsCreateOperands (WalkState, NextOp);
        if (ACPI_FAILURE (Status))
        {
            break;
        }

        /*
         * Since the operands will be passed to another control method, we must
         * resolve all local references here (Local variables, arguments to *this* method, etc.)
         */

        Status = DsResolveOperands (WalkState);
        if (ACPI_FAILURE (Status))
        {
            break;
        }

        /* Open new scope on the scope stack */
/*
        Status = NsScopeStackPushEntry (Entry);
        if (ACPI_FAILURE (Status))
        {
            DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: Could not push Scope Stack\n"));
            break;
        }
*/

        /* Tell the walk loop to preempt this running method and execute the new method */

        Status = AE_CTRL_PENDING;

        /* Return now; we don't want to disturb anything, especially the operand count! */

        return_ACPI_STATUS (Status);
        break;


    case OPTYPE_NAMED_OBJECT:


        if ((WalkState->Origin->Opcode == AML_MethodOp) &&
            (WalkState->Origin != Op))
        {
            Status = DsLoad2EndOp (WalkState, Op);
            if (ACPI_FAILURE (Status))
            {
                break;
            }
        }

        switch (Op->Opcode)
        {
        case AML_RegionOp:

            DEBUG_PRINT (TRACE_EXEC, ("ExecEndOp: Executing OpRegion Address/Length Op=%X\n",
                            Op));

            Status = DsEvalRegionOperands (WalkState, Op);

            break;


        case AML_MethodOp:
            

            break;


        case AML_AliasOp:

            /* Alias creation was already handled by call to psxload above */

            break;


        default:
            /* TBD: Nothing to do here at this time */

            Status = AE_OK;
            break;
        }
        
        break;

    default:

        DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: Unimplemented opcode, type=%X Opcode=%X Op=%X\n",
                        Optype, Op->Opcode, Op));

        Status = AE_NOT_IMPLEMENTED;
        break;
    }



    /*
     * Check if we just completed the evaluation of a conditional predicate 
     */

    if ((WalkState->ControlState) &&
        (WalkState->ControlState->Common.State == CONTROL_PREDICATE_EXECUTING) &&
        (WalkState->ControlState->Control.PredicateOp == Op))
    {
        /* Completed the predicate, the result must be a number */

        WalkState->ControlState->Common.State = 0;

/* TBD: REDO now that we have the resultobj mechanism */

        if (ResultObj)
        {
            Status = DsResultStackPop (&ObjDesc, WalkState);
            if (ACPI_FAILURE (Status))
            {
                goto Cleanup;
            }
        }

        else
        {
            Status = DsCreateOperand (WalkState, Op);
            if (ACPI_FAILURE (Status))
            {
                goto Cleanup;
            }

            Status = AmlResolveToValue (&WalkState->Operands [0]);
            if (ACPI_FAILURE (Status))
            {
                goto Cleanup;
            }

            ObjDesc = WalkState->Operands [0];
        }

        if (!ObjDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: No predicate ObjDesc=%X State=%X\n",
                            ObjDesc, WalkState));

            Status = AE_AML_NO_OPERAND;
            goto Cleanup;
        }

        if (ObjDesc->Common.Type != ACPI_TYPE_Number)
        {
            DEBUG_PRINT (ACPI_ERROR, ("ExecEndOp: Bad predicate ObjDesc=%X State=%X\n",
                            ObjDesc, WalkState));

            Status = AE_AML_OPERAND_TYPE;
            goto Cleanup;
        }
        /* Save the result of the predicate evaluation on the control stack */

        if (ObjDesc->Number.Value)
        {
            WalkState->ControlState->Common.Value = TRUE;
        }
        else
        {
            /* Predicate is FALSE, we will just toss the rest of the package */

            WalkState->ControlState->Common.Value = FALSE;
            Status = AE_CTRL_FALSE;
        }

        DEBUG_PRINT (TRACE_EXEC, ("ExecEndOp: Completed a predicate eval=%X Op=%X\n",
                        WalkState->ControlState->Common.Value, Op));

         /* Break to debugger to display result */

        DEBUG_EXEC (DbDisplayResultObject (ObjDesc));

        /* Delete the predicate result object (we know that we don't need it anymore) and cleanup the stack */

        CmRemoveReference (ObjDesc);
        ResultObj = NULL;

        //DsObjStackPop (1, WalkState);

        WalkState->ControlState->Common.State = CONTROL_NORMAL;
    }


Cleanup:

    if (ResultObj)
    {
        /* Break to debugger to display result */

        DEBUG_EXEC (DbDisplayResultObject (ResultObj));

        /* Delete the result op IFF:
         * Parent will not use the result -- such as any non-nested type2 op in a method (parent will be method)
         */
        DsDeleteResultIfNotUsed (Op, ResultObj, WalkState);
    }

    /* Always clear the object stack */

    WalkState->NumOperands = 0; /* TBD Clear stack of return value, but don't delete it */
    
    return_ACPI_STATUS (Status);
}



