
/*
 *  ./acpica/subsystem/dispatcher/dswload2.c - Dispatcher namespace load callbacks
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __DSWLOAD2_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <events.h>


#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("psxload");





ACPI_STATUS
DsLoad2BeginOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    NAME_TABLE_ENTRY        *NewEntry;
    ACPI_STATUS             Status;
    ACPI_OBJECT_TYPE        DataType; 
    char                    Buffer[8];
    char                    *BufferPtr;
    void                    *Original = NULL;



    DEBUG_PRINT (TRACE_DISPATCH, ("Load2BeginOp: Op=%p State=%p\n", Op, WalkState));

    if (!PsIsNamespaceOp (Op->Opcode) &&
        Op->Opcode != AML_NAMEPATH_OP)
    {
        return AE_OK;
    }

    if (Op->Opcode == AML_NAMEPATH_OP)
    {
        BufferPtr = Op->Value.String;
        if (!BufferPtr)
        {
            return AE_OK;
        }
    }

    else
    {
        /* Convert the name to a null-terminated string */

        *(UINT32 *) Buffer = (((ACPI_NAMED_OP *)Op)->Name);
        Buffer[4] = 0;
        BufferPtr = Buffer;
    }


    /* Map the raw opcode into an internal object type */

    DataType = DsMapNamedOpcodeToDataType (Op->Opcode);

    DEBUG_PRINT (TRACE_DISPATCH, ("Load2BeginOp: State=%p Op=%p Type=%x\n", WalkState, Op, DataType));


    if (Op->Opcode == AML_DefFieldOp        ||
        Op->Opcode == AML_BankFieldOp       ||
        Op->Opcode == AML_IndexFieldOp)
    {
        NewEntry = NULL;
        Status = AE_OK;
    }

    else if (Op->Opcode == AML_NAMEPATH_OP)
    {
        /* 
         * The NamePath is an object reference to an existing object.  Don't enter the
         * name into the namespace, but look it up for use later
         */
        Status = NsLookup (WalkState->ScopeInfo, BufferPtr, DataType, IMODE_Execute, 
                                    NS_SEARCH_PARENT, WalkState, &(NewEntry));
    }

    else
    {
        if (Op->NameTableEntry)
        {
            Original = Op->NameTableEntry;
            NewEntry = Op->NameTableEntry;

            if (NsOpensScope (DataType))
            {
                Status = DsScopeStackPush (NewEntry->Scope, DataType, WalkState);
                if (ACPI_FAILURE (Status))
                {
                    return (Status);
                }

            }
            return AE_OK;
        }

        /* 
         * Enter the named type into the internal namespace.  We enter the name
         * as we go downward in the parse tree.  Any necessary subobjects that involve
         * arguments to the opcode must be created as we go back up the parse tree later.
         */
        Status = NsLookup (WalkState->ScopeInfo, BufferPtr, DataType, IMODE_Execute, 
                                        NS_NO_UPSEARCH, WalkState, &(NewEntry));
    }

    if (ACPI_SUCCESS (Status))
    {
        /* 
         * Put the NTE in the "op" object that the parser uses, so we
         * can get it again quickly when this scope is closed
         */
        Op->NameTableEntry = NewEntry;

        if (Original)
        {
            DEBUG_PRINT (ACPI_INFO, ("Lookup: old %p new %p\n", Original, NewEntry));
            if (Original != NewEntry)
            {
                DEBUG_PRINT (ACPI_INFO, ("Lookup match error: old %p new %p\n", Original, NewEntry));
            }
        }
    }


    return (Status);
}



ACPI_STATUS
DsLoad2EndOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_TYPE        DataType; 
    NAME_TABLE_ENTRY        *Entry;
    ACPI_GENERIC_OP         *Arg;
    NAME_TABLE_ENTRY        *NewEntry;



    DEBUG_PRINT (TRACE_DISPATCH, ("Load2EndOp: Op=%p State=%p\n", Op, WalkState));

    if (!PsIsNamespaceObjectOp (Op->Opcode))
    {
        return AE_OK;
    }

    if (Op->Opcode == AML_ScopeOp)
    {
        DEBUG_PRINT (TRACE_DISPATCH, ("Load2EndOp: ending scope Op=%p State=%p\n", Op, WalkState));
        if (((ACPI_NAMED_OP *)Op)->Name == -1)
        {
            DEBUG_PRINT (ACPI_ERROR, ("Load2EndOp: Un-named scope! Op=%p State=%p\n", Op, WalkState));
            return AE_OK;
        }
    }


    DataType = DsMapNamedOpcodeToDataType (Op->Opcode);

    /* 
     * Get the NTE/name from the earlier lookup
     * (It was saved in the *op structure)
     */
    Entry = Op->NameTableEntry;

    /* Put the NTE on the object stack (Contains the ACPI Name of this object) */

//    Status = DsObjStackPush (Entry, WalkState);
    WalkState->Operands[0] = (void *) Entry;
    WalkState->NumOperands = 1;

    /* Pop the scope stack */

    if (NsOpensScope (DataType))
    {

        DEBUG_PRINT (TRACE_DISPATCH, ("AmlEndNamespaceScope/%s: Popping scope for Op %p\n",
                                        CmGetTypeName (DataType), Op));
        DsScopeStackPop (WalkState);
    }



    /*
     * Named operations are as follows:
     *
     * AML_SCOPE
     * AML_DEVICE
     * AML_THERMALZONE
     * AML_METHOD
     * AML_POWERRES
     * AML_PROCESSOR
     * AML_FIELD
     * AML_INDEXFIELD
     * AML_BANKFIELD
     * AML_NAMEDFIELD
     * AML_NAME
     * AML_ALIAS
     * AML_MUTEX
     * AML_EVENT
     * AML_OPREGION
     * AML_CREATEFIELD
     * AML_CREATEBITFIELD
     * AML_CREATEBYTEFIELD
     * AML_CREATEWORDFIELD
     * AML_CREATEDWORDFIELD
     * AML_METHODCALL
     */


    /* Decode the opcode */

    Arg = Op->Value.Arg;

    switch (Op->Opcode)
    {

    case AML_CreateFieldOp:
    case AML_BitFieldOp:
    case AML_ByteFieldOp:
    case AML_WordFieldOp:
    case AML_DWordFieldOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-CreateXxxField: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        /* Get the NameString argument */

        if (Op->Opcode == AML_CreateFieldOp)
        {
            Arg = PsGetArg (Op, 3);
        }
        else
        {
            /* Create Bit/Byte/Word/Dword field */

            Arg = PsGetArg (Op, 2);
        }

        /*
         * Enter the NameString into the namespace
         */

        Status = NsLookup (WalkState->ScopeInfo, Arg->Value.String, INTERNAL_TYPE_DefAny, IMODE_LoadPass1, 
                                    NS_NO_UPSEARCH | NS_DONT_OPEN_SCOPE, WalkState, &(NewEntry));
        if (ACPI_SUCCESS (Status))
        {
            /* We could put the returned object (NTE) on the object stack for later, but
             * for now, we will put it in the "op" object that the parser uses, so we
             * can get it again at the end of this scope
             */
            Op->NameTableEntry = NewEntry;
        }

        break;


    case AML_METHODCALL_OP:

        DEBUG_PRINT (TRACE_DISPATCH, ("RESOLVING-MethodCall: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        /* 
         * Lookup the method name and save the NTE
         */

        Status = NsLookup (WalkState->ScopeInfo, Arg->Value.String, ACPI_TYPE_Any, IMODE_LoadPass2, 
                                    NS_SEARCH_PARENT | NS_DONT_OPEN_SCOPE, WalkState, &(NewEntry));
        if (ACPI_SUCCESS (Status))
        {

/* has name already been resolved by here ??*/

            /* TBD: Make sure that what we found is indeed a method! */
            /* We didn't search for a method on purpose, to see if the name would resolve! */

            /* We could put the returned object (NTE) on the object stack for later, but
             * for now, we will put it in the "op" object that the parser uses, so we
             * can get it again at the end of this scope
             */
            Op->NameTableEntry = NewEntry;
        }

         
         break;


    case AML_ProcessorOp:
        
        /* Nothing to do other than enter object into namespace */

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Processor: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        Status = AmlExecCreateProcessor (Op, (ACPI_HANDLE) Entry);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        DEBUG_PRINT (TRACE_DISPATCH, ("Completed Processor Init, Op=%p State=%p entry=%p\n", Op, WalkState, Entry));
        break;


    case AML_PowerResOp:
        
        /* Nothing to do other than enter object into namespace */

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-PowerResource: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        Status = AmlExecCreatePowerResource (Op, (ACPI_HANDLE) Entry);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        DEBUG_PRINT (TRACE_DISPATCH, ("Completed PowerResource Init, Op=%p State=%p entry=%p\n", Op, WalkState, Entry));
        break;


    case AML_ThermalZoneOp:
        
        /* Nothing to do other than enter object into namespace */

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-ThermalZone: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        break;


    case AML_DefFieldOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Field: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        Arg = Op->Value.Arg;
        if (!Arg->NameTableEntry)
        {
        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Field: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        }

        Status = DsCreateField (Op, (ACPI_HANDLE) Arg->NameTableEntry, WalkState);
        break;


    case AML_IndexFieldOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-IndexField: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        Arg = Op->Value.Arg;

        Status = DsCreateIndexField (Op, (ACPI_HANDLE) Arg->NameTableEntry, WalkState);
        break;


    case AML_BankFieldOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-BankField: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        Arg = Op->Value.Arg;
        Status = DsCreateBankField (Op, (ACPI_HANDLE) Arg->NameTableEntry, WalkState);
        break;


    /*
     * MethodOp PkgLength NamesString MethodFlags TermList
     */
    case AML_MethodOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Method: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));

        if (!Entry->Object)
        {
            Status = AmlExecCreateMethod (((ACPI_DEFERRED_OP *) Op)->Body, ((ACPI_DEFERRED_OP *) Op)->BodyLength,
                                            Arg->Value.Integer, (ACPI_HANDLE) Entry);
        }

        break;


    case AML_MutexOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Mutex: Op=%p State=%p\n", Op, WalkState));

        Status = DsCreateOperands (WalkState, Arg);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        Status = AmlExecCreateMutex (WalkState);
        break;


    case AML_EventOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Event: Op=%p State=%p\n", Op, WalkState));

        Status = DsCreateOperands (WalkState, Arg);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        Status = AmlExecCreateEvent (WalkState);
        break;


    case AML_RegionOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Opregion: Op=%p State=%p Nte=%p\n", Op, WalkState, Entry));


        /* 
         * The OpRegion is not fully parsed at this time.  Only valid argument is the SpaceId.
         * (We must save the address of the AML of the address and length operands)
         */

        Status = AmlExecCreateRegion (((ACPI_DEFERRED_OP *) Op)->Body, ((ACPI_DEFERRED_OP *) Op)->BodyLength, 
                                        Arg->Value.Integer, WalkState);

        DEBUG_PRINT (TRACE_DISPATCH, ("Completed OpRegion Init, Op=%p State=%p entry=%p\n", Op, WalkState, Entry));
        break;


    /* Namespace Modifier Opcodes */

    case AML_AliasOp:
        
        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Alias: Op=%p State=%p\n", Op, WalkState));

        Status = DsCreateOperands (WalkState, Arg);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        Status = AmlExecCreateAlias (WalkState);
        break;


    case AML_NameOp:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-Name: Op=%p State=%p\n", Op, WalkState));

        Status = DsCreateNamedObject (WalkState, Entry, Op);

        break;


    case AML_NAMEPATH_OP:

        DEBUG_PRINT (TRACE_DISPATCH, ("LOADING-NamePath object: State=%p Op=%p nte=%p\n",
                                WalkState, Op, Entry));
        break;


    default:
        break;
    }


Cleanup:

    DsObjStackPop (1, WalkState);      /* Remove the NTE pushed at the very beginning */
    return (Status);
}





