/*
 *  ./acpica/subsystem/dispatcher/dswstate.c - Dispatcher parse tree walk management routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __DSWSTATE_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <namesp.h>
#include <interp.h>

#define _COMPONENT          DISPATCHER
        MODULE_NAME         ("dswstate");





ACPI_STATUS
DsResultStackClear (
    ACPI_WALK_STATE         *WalkState)
{

    WalkState->NumResults = 0;
    WalkState->CurrentResult = 0;

    return AE_OK;
}



ACPI_STATUS
DsResultStackPush (
    void                    *Object,
    ACPI_WALK_STATE         *WalkState)
{


    if (WalkState->NumResults >= OBJ_NUM_OPERANDS)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsResultStackPush: overflow! Obj=%p State=%p Num=%X\n", 
                        Object, WalkState, WalkState->NumResults));
        return AE_STACK_OVERFLOW;
    }

    WalkState->Results [WalkState->NumResults] = Object;
    WalkState->NumResults++;

    DEBUG_PRINT (TRACE_EXEC, ("DsResultStackPush: Obj=%p State=%p Num=%X Cur=%X\n", 
                    Object, WalkState, WalkState->NumResults, WalkState->CurrentResult));

    return AE_OK;
}



ACPI_STATUS
DsResultStackPop (
    ACPI_OBJECT_INTERNAL    **Object,
    ACPI_WALK_STATE         *WalkState)
{


    /* Check for stack underflow */

    if (WalkState->NumResults == 0)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsResultStackPop: Underflow! State=%p Cur=%X Num=%X\n", 
                        WalkState, WalkState->CurrentResult, WalkState->NumResults));
        return AE_AML_NO_OPERAND;
    }


    /* Pop the stack */

    WalkState->NumResults--;

    /* Check for a valid result object */

    if (!WalkState->Results [WalkState->NumResults])
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsResultStackPop: Null operand! State=%p #Ops=%X\n", 
                        WalkState, WalkState->NumResults));
        return AE_AML_NO_OPERAND;
    }


    *Object = WalkState->Results [WalkState->NumResults];   
    WalkState->Results [WalkState->NumResults] = NULL;


    DEBUG_PRINT (TRACE_EXEC, ("DsResultStackPop: Obj=%p State=%p Num=%X Cur=%X\n", 
                    *Object, WalkState, WalkState->NumResults, WalkState->CurrentResult));

    return AE_OK;
}



ACPI_STATUS
DsObjStackDeleteAll (
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  i;


    FUNCTION_TRACE_PTR ("DsObjStackDeleteAll", WalkState);


    /* The stack size is configurable, but fixed */

    for (i = 0; i < OBJ_NUM_OPERANDS; i++)
    {
        if (WalkState->Operands[i])
        {
            CmRemoveReference (WalkState->Operands[i]);
            WalkState->Operands[i] = NULL;
        }
    }

    return_ACPI_STATUS (AE_OK);
}




ACPI_STATUS
DsObjStackPush (
    void                    *Object,
    ACPI_WALK_STATE         *WalkState)
{


    /* Check for stack overflow */

    if (WalkState->NumOperands >= OBJ_NUM_OPERANDS)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsObjStackPush: overflow! Obj=%p State=%p #Ops=%X\n", 
                        Object, WalkState, WalkState->NumOperands));
        return AE_STACK_OVERFLOW;
    }

    /* Put the object onto the stack */

    WalkState->Operands [WalkState->NumOperands] = Object;
    WalkState->NumOperands++;

    DEBUG_PRINT (TRACE_EXEC, ("DsObjStackPush: Obj=%p State=%p #Ops=%X\n", 
                    Object, WalkState, WalkState->NumOperands));

    return AE_OK;
}



ACPI_STATUS
DsObjStackPopObject (
    ACPI_OBJECT_INTERNAL    **Object,
    ACPI_WALK_STATE         *WalkState)
{


    /* Check for stack underflow */

    if (WalkState->NumOperands == 0)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsObjStackPop: Missing operand/stack empty! State=%p #Ops=%X\n", 
                        WalkState, WalkState->NumOperands));
        return AE_AML_NO_OPERAND;
    }


    /* Pop the stack */

    WalkState->NumOperands--;

    /* Check for a valid operand */

    if (!WalkState->Operands [WalkState->NumOperands])
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsObjStackPop: Null operand! State=%p #Ops=%X\n", 
                        WalkState, WalkState->NumOperands));
        return AE_AML_NO_OPERAND;
    }

    /* Get operand and set stack entry to null */

    *Object = WalkState->Operands [WalkState->NumOperands];
    WalkState->Operands [WalkState->NumOperands] = NULL;

    DEBUG_PRINT (TRACE_EXEC, ("DsObjStackPopObject: State=%p #Ops=%X\n", 
                    WalkState, WalkState->NumOperands));

    return AE_OK;
}

        

ACPI_STATUS
DsObjStackPop (
    UINT32                  PopCount,
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  i;



    for (i = 0; i < PopCount; i++)
    {
        /* Check for stack underflow */

        if (WalkState->NumOperands == 0)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsObjStackPop: Underflow! Count=%X State=%p #Ops=%X\n", 
                            PopCount, WalkState, WalkState->NumOperands));
            return AE_STACK_UNDERFLOW;
        }

        /* Just set the stack entry to null */

        WalkState->NumOperands--;
        WalkState->Operands [WalkState->NumOperands] = NULL;
    }

    DEBUG_PRINT (TRACE_EXEC, ("DsObjStackPop: Count=%X State=%p #Ops=%X\n", 
                    PopCount, WalkState, WalkState->NumOperands));

    return AE_OK;
}

    

ACPI_STATUS
DsObjStackPopAndDelete (
    UINT32                  PopCount,
    ACPI_WALK_STATE         *WalkState)
{
    UINT32                  i;
    ACPI_OBJECT_INTERNAL    *ObjDesc;


    for (i = 0; i < PopCount; i++)
    {
        /* Check for stack underflow */

        if (WalkState->NumOperands == 0)
        {
            DEBUG_PRINT (ACPI_ERROR, ("DsObjStackPop: Underflow! Count=%X State=%p #Ops=%X\n", 
                            PopCount, WalkState, WalkState->NumOperands));
            return AE_STACK_UNDERFLOW;
        }

        /* Pop the stack and delete an object if present in this stack entry */

        WalkState->NumOperands--;
        ObjDesc = WalkState->Operands [WalkState->NumOperands];
        if (ObjDesc)
        {
            CmRemoveReference (WalkState->Operands [WalkState->NumOperands]);
            WalkState->Operands [WalkState->NumOperands] = NULL;
        }
    }

    DEBUG_PRINT (TRACE_EXEC, ("DsObjStackPop: Count=%X State=%p #Ops=%X\n", 
                    PopCount, WalkState, WalkState->NumOperands));

    return AE_OK;
}

    

void *
DsObjStackGetValue (
    UINT32                  Index,
    ACPI_WALK_STATE         *WalkState)
{

    FUNCTION_TRACE_PTR ("DsObjStackGetValue", WalkState);


    /* Can't do it if the stack is empty */

    if (WalkState->NumOperands == 0)
    {
        return_VALUE (NULL);
    }

    /* or if the index is past the top of the stack */

    if (Index > (WalkState->NumOperands - (UINT32) 1))
    {
        return_VALUE (NULL);
    }


    return_PTR (WalkState->Operands [(WalkState->NumOperands - 1) - Index]);
}





ACPI_WALK_STATE *
DsGetCurrentWalkState (
    ACPI_WALK_LIST          *WalkList)

{

    DEBUG_PRINT (TRACE_PARSE, ("DsGetCurrentWalkState, =%p\n", WalkList->WalkState));

    if (!WalkList)
    {
        return NULL;
    }

    return WalkList->WalkState;
}



void
DsPushWalkState (
    ACPI_WALK_STATE         *WalkState,
    ACPI_WALK_LIST          *WalkList)
{


    FUNCTION_TRACE ("DsPushWalkState");


    WalkState->Next     = WalkList->WalkState;
    WalkList->WalkState = WalkState;

    return_VOID;
}



ACPI_WALK_STATE *
DsPopWalkState (
    ACPI_WALK_LIST          *WalkList)
{
    ACPI_WALK_STATE         *WalkState;


    FUNCTION_TRACE ("DsPopWalkState");


    WalkState = WalkList->WalkState;

    if (WalkState)
    {
        /* Next walk state becomes the current walk state */

        WalkList->WalkState = WalkState->Next;

        /* 
         * Don't clear the NEXT field, this serves as an indicator
         * that there is a parent WALK STATE 
         *     WalkState->Next = NULL;
         */
    }

    return_PTR (WalkState);
}



ACPI_WALK_STATE *
DsCreateWalkState (
    ACPI_GENERIC_OP         *Origin,
    ACPI_OBJECT_INTERNAL    *MthDesc,
    ACPI_WALK_LIST          *WalkList)
{
    ACPI_WALK_STATE         *WalkState;


    FUNCTION_TRACE ("DsCreateWalkState");


    CmAcquireMutex (MTX_CACHES);
    Gbl_WalkStateCacheRequests++;

    /* Check the cache first */

    if (Gbl_WalkStateCache)
    {
        /* There is an object available, use it */

        WalkState = Gbl_WalkStateCache;
        Gbl_WalkStateCache = WalkState->Next;

        Gbl_WalkStateCacheHits++;
        Gbl_WalkStateCacheDepth--;

        CmReleaseMutex (MTX_CACHES);
   }

    else
    {
        /* The cache is empty, create a new object */

        CmReleaseMutex (MTX_CACHES);                        /* Avoid deadlock with CmCallocate */

        WalkState = CmCallocate (sizeof (ACPI_WALK_STATE));
        if (!WalkState)
        {
            return_VALUE (NULL);
        }
    }

    WalkState->DataType         = DESC_TYPE_WALK;
    WalkState->Origin           = Origin;
    WalkState->MethodDesc       = MthDesc;

    /* Init the method args/local */

    DsMethodDataInit (WalkState);

    /* Put the new state at the head of the walk list */

    DsPushWalkState (WalkState, WalkList);

    return_PTR (WalkState);
}



void
DsDeleteWalkState (
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_GENERIC_STATE      *State;


    FUNCTION_TRACE_PTR ("DsDeleteWalkState", WalkState);


    if (!WalkState)
    {
        return;
    }

    if (WalkState->DataType != DESC_TYPE_WALK)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DsDeleteWalkState: **** %p not a valid walk state\n", WalkState));
        return;
    }

    /* Always must free any linked control states */

    while (WalkState->ControlState)
    {
        State = WalkState->ControlState;
        WalkState->ControlState = State->Common.Next;

        CmDeleteGenericState (State);
    }


    /* Always must free any linked parse states */

    while (WalkState->ScopeInfo)
    {
        State = WalkState->ScopeInfo;
        WalkState->ScopeInfo = State->Common.Next;

        CmDeleteGenericState (State);
    }

    /* If walk cache is full, just free this wallkstate object */

    if (Gbl_WalkStateCacheDepth >= MAX_WALK_CACHE_DEPTH)
    {
        CmFree (WalkState);
    }

    /* Otherwise put this object back into the cache */

    else
    {
        CmAcquireMutex (MTX_CACHES);

        /* Clear the state */

        MEMSET (WalkState, 0, sizeof (ACPI_WALK_STATE));
        WalkState->DataType = DESC_TYPE_WALK;

        /* Put the object at the head of the global cache list */

        WalkState->Next = Gbl_WalkStateCache;
        Gbl_WalkStateCache = WalkState;
        Gbl_WalkStateCacheDepth++;


        CmReleaseMutex (MTX_CACHES);
    }

    return_VOID;
}



void
DsDeleteWalkStateCache (
    void)
{
    ACPI_WALK_STATE         *Next;


    FUNCTION_TRACE ("DsDeleteWalkStateCache");


    /* Traverse the global cache list */

    while (Gbl_WalkStateCache)
    {
        /* Delete one cached state object */

        Next = Gbl_WalkStateCache->Next;
        CmFree (Gbl_WalkStateCache);
        Gbl_WalkStateCache = Next;
    }

    return_VOID;
}
 


