/*
 *  ./acpica/subsystem/events/evapi.c - External interfaces for ACPI events
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __EVAPI_C__

#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>
#include <amlcode.h>
#include <interp.h>

#undef AcpiInstallFixedEventHandler
#undef AcpiRemoveFixedEventHandler
#undef AcpiInstallGpeHandler
#undef AcpiRemoveGpeHandler
#undef AcpiEnableEvent
#undef AcpiDisableEvent
#undef AcpiClearEvent
#undef AcpiGetEventStatus

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evapi");




ACPI_STATUS
AcpiInstallFixedEventHandler (
    UINT32                  Event,
    FIXED_EVENT_HANDLER     Handler,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AcpiInstallFixedEventHandler");


    /* Sanity check the parameters. */

    if (Event >= NUM_FIXED_EVENTS)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    CmAcquireMutex (MTX_EVENTS);

    /* Don't allow two handlers. */

    if (NULL != Gbl_FixedEventHandlers[Event].Handler)
    {
        Status = AE_EXIST;
        goto Cleanup;
    }
    

    /* Install the handler before enabling the event - just in case... */

    Gbl_FixedEventHandlers[Event].Handler = Handler;
    Gbl_FixedEventHandlers[Event].Context = Context;
    
    if (1 != HwRegisterIO (ACPI_WRITE, MTX_LOCK, Event + TMR_EN, 1))
    {
        DEBUG_PRINT (ACPI_WARN, ("Could not write to fixed event enable register.\n"));
        
        /* Remove the handler */

        Gbl_FixedEventHandlers[Event].Handler = NULL;
        Gbl_FixedEventHandlers[Event].Context = NULL;

        Status = AE_ERROR;
        goto Cleanup;
    }

    DEBUG_PRINT (ACPI_INFO, ("Enabled fixed event %d.  Handler: %x\n", Event, Handler));    


Cleanup:
    CmReleaseMutex (MTX_EVENTS);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiRemoveFixedEventHandler (
    UINT32                  Event,
    FIXED_EVENT_HANDLER     Handler)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AcpiRemoveFixedEventHandler");


    /* Sanity check the parameters. */

    if (Event >= NUM_FIXED_EVENTS)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    CmAcquireMutex (MTX_EVENTS);

    /* Disable the event before removing the handler - just in case... */

    if (0 != HwRegisterIO (ACPI_WRITE, MTX_LOCK, Event + TMR_EN, 0))
    {
        DEBUG_PRINT (ACPI_WARN, ("Could not write to fixed event enable register.\n"));
        Status = AE_ERROR;
        goto Cleanup;
    }

    /* Remove the handler */

    Gbl_FixedEventHandlers[Event].Handler = NULL;
    Gbl_FixedEventHandlers[Event].Context = NULL;
    
    DEBUG_PRINT (ACPI_INFO, ("Disabled fixed event %d.\n", Event));    

Cleanup:
    CmReleaseMutex (MTX_EVENTS);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiInstallNotifyHandler (
    ACPI_HANDLE             Device,
    UINT32                  HandlerType, 
    NOTIFY_HANDLER          Handler, 
    void                    *Context)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *NotifyObj;
    NAME_TABLE_ENTRY        *ObjEntry;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AcpiInstallNotifyHandler");


    /* Parameter validation */

    if ((!Handler) ||
        (HandlerType > ACPI_MAX_NOTIFY_HANDLER_TYPE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Convert and validate the device handle */

    CmAcquireMutex (MTX_NAMESPACE);

    ObjEntry = NsConvertHandleToEntry (Device);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }


    /*
     * Support for global notify handlers.  These handlers are invoked for
     * every notifiy of the type specifiec
     */

    if (Device == ACPI_ROOT_OBJECT)
    {
        /*
         *  Make sure the handler is not already installed.
         */

        if (((HandlerType == ACPI_SYSTEM_NOTIFY) && Gbl_SysNotify.Handler) ||
            ((HandlerType == ACPI_DEVICE_NOTIFY) && Gbl_DrvNotify.Handler))
        {
            Status = AE_EXIST;
            goto UnlockAndExit;
        }

        if (HandlerType == ACPI_SYSTEM_NOTIFY)
        {
            Gbl_SysNotify.Nte = ObjEntry;
            Gbl_SysNotify.Handler = Handler;
            Gbl_SysNotify.Context = Context;
        }

        else
        {
            Gbl_DrvNotify.Nte = ObjEntry;
            Gbl_DrvNotify.Handler = Handler;
            Gbl_DrvNotify.Context = Context;
        }


        /* Global notify handler installed */

        goto UnlockAndExit;
    }


    /*
     * These are the ONLY objects that can receive ACPI notifications
     */

    if ((ObjEntry->Type != ACPI_TYPE_Device)     &&
        (ObjEntry->Type != ACPI_TYPE_Processor)  &&
        (ObjEntry->Type != ACPI_TYPE_Power)      &&
        (ObjEntry->Type != ACPI_TYPE_Thermal))
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /* Check for an existing internal object */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) ObjEntry);
    if (ObjDesc)
    {
        /*
         *  The object exists.
         *  Make sure the handler is not already installed.
         */

        if (((HandlerType == ACPI_SYSTEM_NOTIFY) && ObjDesc->Device.SysHandler) ||
            ((HandlerType == ACPI_DEVICE_NOTIFY) && ObjDesc->Device.DrvHandler))
        {
            Status = AE_EXIST;
            goto UnlockAndExit;
        }
    }

    else
    {
        /* Create a new object */

        ObjDesc = CmCreateInternalObject (ObjEntry->Type);
        if (!ObjDesc)
        {
            Status = AE_NO_MEMORY;
            goto UnlockAndExit;
        }

        /* Attach new object to the NTE */

        Status = NsAttachObject (Device, ObjDesc, (UINT8) ObjEntry->Type);

        if (ACPI_FAILURE (Status))
        {
            goto UnlockAndExit;
        }
    }


    /* 
     *  If we get here, we know that there is no handler installed
     *  so let's party
     */
    NotifyObj = CmCreateInternalObject (INTERNAL_TYPE_Notify);
    if (!NotifyObj)
    {
        Status = AE_NO_MEMORY;
        goto UnlockAndExit;
    }

    NotifyObj->NotifyHandler.Nte = ObjEntry;
    NotifyObj->NotifyHandler.Handler = Handler;
    NotifyObj->NotifyHandler.Context = Context;


    if (HandlerType == ACPI_SYSTEM_NOTIFY)
    {
        ObjDesc->Device.SysHandler = NotifyObj;
    }

    else
    {
        ObjDesc->Device.DrvHandler = NotifyObj;
    }


UnlockAndExit:
    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiRemoveNotifyHandler (
    ACPI_HANDLE             Device,
    UINT32                  HandlerType, 
    NOTIFY_HANDLER          Handler)
{
    ACPI_OBJECT_INTERNAL    *NotifyObj;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *ObjEntry;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AcpiRemoveNotifyHandler");


    /* Parameter validation */

    if ((!Handler) ||
        (HandlerType > ACPI_MAX_NOTIFY_HANDLER_TYPE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* Convert and validate the device handle */

    ObjEntry = NsConvertHandleToEntry (Device);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /*
     * These are the ONLY objects that can receive ACPI notifications
     */

    if ((ObjEntry->Type != ACPI_TYPE_Device)     &&
        (ObjEntry->Type != ACPI_TYPE_Processor)  &&
        (ObjEntry->Type != ACPI_TYPE_Power)      &&
        (ObjEntry->Type != ACPI_TYPE_Thermal))
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /* Check for an existing internal object */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) ObjEntry);
    if (!ObjDesc)
    {
        Status = AE_NOT_EXIST;
        goto UnlockAndExit;
    }

    /*
     *  The object exists.
     *
     *  Make sure the handler is installed.
     */

    if (HandlerType == ACPI_SYSTEM_NOTIFY)
    {
        NotifyObj = ObjDesc->Device.SysHandler;
    }
    else
    {
        NotifyObj = ObjDesc->Device.DrvHandler;
    }

    if ((!NotifyObj) ||
        (NotifyObj->NotifyHandler.Handler != Handler))
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /* 
     * Now we can remove the handler
     */
    if (HandlerType == ACPI_SYSTEM_NOTIFY)
    {
        ObjDesc->Device.SysHandler = NULL;
    }
    else
    {
        ObjDesc->Device.DrvHandler = NULL;
    }
 
    CmRemoveReference (NotifyObj);

UnlockAndExit:
    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}


ACPI_STATUS
AcpiInstallGpeHandler (
    UINT32                  GpeNumber,
    UINT32                  Type,
    GPE_HANDLER             Handler,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    
    FUNCTION_TRACE ("AcpiInstallGpeHandler");

    /* Parameter validation */

    if (!Handler || (GpeNumber > NUM_GPE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Ensure that we have a valid GPE number */

    if (Gbl_GpeValid[GpeNumber] == GPE_INVALID)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    CmAcquireMutex (MTX_EVENTS);

    /* Make sure that there isn't a handler there already */

    if (Gbl_GpeInfo[GpeNumber].Handler)
    {
        Status = AE_EXIST;
        goto Cleanup;
    }

    /* Install the handler */

    Gbl_GpeInfo[GpeNumber].Handler = Handler;
    Gbl_GpeInfo[GpeNumber].Context = Context;
    Gbl_GpeInfo[GpeNumber].Type = (UINT8) Type;
    
    /* Clear the GPE (of stale events), the enable it */

    HwClearGpe (GpeNumber);
    HwEnableGpe (GpeNumber);

Cleanup:
    CmReleaseMutex (MTX_EVENTS);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiRemoveGpeHandler (
    UINT32                  GpeNumber, 
    GPE_HANDLER             Handler)
{
    ACPI_STATUS             Status = AE_OK;

    
    FUNCTION_TRACE ("AcpiRemoveGpeHandler");


    /* Parameter validation */

    if (!Handler || (GpeNumber > NUM_GPE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Ensure that we have a valid GPE number */

    if (Gbl_GpeValid[GpeNumber] == GPE_INVALID)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Disable the GPE before removing the handler */

    HwDisableGpe (GpeNumber);

    CmAcquireMutex (MTX_EVENTS);

    /* Make sure that the installed handler is the same */

    if (Gbl_GpeInfo[GpeNumber].Handler != Handler)
    {
        HwEnableGpe (GpeNumber);
        Status = AE_BAD_PARAMETER;
        goto Cleanup;
    }

    /* Remove the handler */

    Gbl_GpeInfo[GpeNumber].Handler = NULL;
    Gbl_GpeInfo[GpeNumber].Context = NULL;
 
Cleanup:
    CmReleaseMutex (MTX_EVENTS);
    return_ACPI_STATUS (Status);
}




ACPI_STATUS
AcpiAcquireGlobalLock (
    UINT32                  Timeout,
    UINT32                  *OutHandle)
{
    ACPI_STATUS             Status;


    AmlEnterInterpreter ();

    /* TBD: add timeout param to internal interface, and perhaps INTERPRETER_LOCKED */


    Status = EvAcquireGlobalLock ();
    AmlExitInterpreter ();

    *OutHandle = 0;
    return Status;
}




ACPI_STATUS
AcpiReleaseGlobalLock (
    UINT32                  Handle)
{


    /* TBD: Validate handle */

    EvReleaseGlobalLock ();
    return AE_OK;
}


