/*
 *  ./acpica/subsystem/events/evapievt.c - External Interfaces, ACPI event disable/enable
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __EVAPIEVT_C__

#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>
#include <amlcode.h>
#include <interp.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evapievt");




ACPI_STATUS
AcpiEnable (void)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AcpiEnable");


    /* Make sure we've got ACPI tables */ 

    if (!Gbl_RSDP)
    {
        DEBUG_PRINT (ACPI_WARN, ("No ACPI tables present!\n"));
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    /* Make sure the BIOS supports ACPI mode */

    if (SYS_MODE_LEGACY == HwGetModeCapabilities())
    {   
        DEBUG_PRINT (ACPI_WARN, ("Only legacy mode supported!\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

    Gbl_OriginalMode = HwGetMode();

    /*
     * Initialize the Fixed and General Purpose Events prior.  This is
     * done prior to enabling SCIs to prevent interrupts from occuring
     * before handers are installed.
     */ 
   
    if (ACPI_FAILURE (EvFixedEventInitialize ()))
    {
        DEBUG_PRINT (ACPI_FATAL, ("Unable to initialize fixed events.\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

    if (ACPI_FAILURE (EvGpeInitialize()))
    {
        DEBUG_PRINT (ACPI_FATAL, ("Unable to initialize general purpose events.\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

    /* Install the SCI handler */

    if (ACPI_FAILURE (EvInstallSciHandler ()))
    {   
        DEBUG_PRINT (ACPI_FATAL, ("Unable to install System Control Interrupt Handler\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

    /* Transition to ACPI mode */

    if (AE_OK != HwSetMode (SYS_MODE_ACPI))
    {   
        DEBUG_PRINT (ACPI_FATAL, ("Could not transition to ACPI mode.\n"));
        return_ACPI_STATUS (AE_ERROR);    
    }

    DEBUG_PRINT (ACPI_OK, ("Transition to ACPI mode successful\n"));

    /* Install handlers for control method GPE handlers (_Lxx, _Exx) */

    EvInitGpeControlMethods ();

    Status = EvInitGlobalLockHandler ();

    return_ACPI_STATUS (Status);
}
    


ACPI_STATUS     
AcpiDisable (void)
{

    FUNCTION_TRACE ("AcpiDisable");


    /* Restore original mode  */

    if (AE_OK != HwSetMode (Gbl_OriginalMode))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Unable to transition to original mode"));
        return_ACPI_STATUS (AE_ERROR);    
    }

    /* Unload the SCI interrupt handler  */

    EvRemoveSciHandler ();
    EvRestoreAcpiState ();
    
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiEnableEvent (
    UINT32                  Event,
    UINT32                  Type)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE ("AcpiEnableEvent");

    switch (Type)
    {
    case EVENT_FIXED:
        /*
         * TBD - Fixed events...
         */
        Status = AE_NOT_IMPLEMENTED;
        break;

    case EVENT_GPE:

        /* Ensure that we have a valid GPE number */

        if (Gbl_GpeValid[Event] == GPE_INVALID)
        {
            Status = AE_BAD_PARAMETER;
        }
        else
        {
            HwEnableGpe (Event);
        }

        break;

    default:
        Status = AE_BAD_PARAMETER;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiDisableEvent (
    UINT32                  Event,
    UINT32                  Type)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE ("AcpiDisableEvent");

    switch (Type)
    {
    case EVENT_FIXED:
        /*
         * TBD - Fixed events...
         */
        Status = AE_NOT_IMPLEMENTED;
        break;

    case EVENT_GPE:

        /* Ensure that we have a valid GPE number */

        if (Gbl_GpeValid[Event] == GPE_INVALID)
        {
            Status = AE_BAD_PARAMETER;
        }
        else
        {
            HwDisableGpe (Event);
        }

        break;

    default:
        Status = AE_BAD_PARAMETER;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiClearEvent (
    UINT32                  Event,
    UINT32                  Type)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE ("AcpiClearEvent");

    switch (Type)
    {
    case EVENT_FIXED:
        /*
         * TBD - Fixed events...
         */
        Status = AE_NOT_IMPLEMENTED;
        break;

    case EVENT_GPE:

        /* Ensure that we have a valid GPE number */

        if (Gbl_GpeValid[Event] == GPE_INVALID)
        {
            Status = AE_BAD_PARAMETER;
        }
        else
        {
            HwClearGpe (Event);
        }

        break;

    default:
        Status = AE_BAD_PARAMETER;
    }

    return_ACPI_STATUS (Status);
}




ACPI_STATUS
AcpiGetEventStatus (
    UINT32                  Event,
    UINT32                  Type,
    ACPI_EVENT_STATUS       *EventStatus)
{
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE ("AcpiGetEventStatus");

    if (!EventStatus)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    switch (Type)
    {
    case EVENT_FIXED:
        /*
         * TBD - Fixed events...
         */
        Status = AE_NOT_IMPLEMENTED;
        break;

    case EVENT_GPE:

        /* Ensure that we have a valid GPE number */

        if (Gbl_GpeValid[Event] == GPE_INVALID)
        {
            Status = AE_BAD_PARAMETER;
        }
        else
        {
            HwGetGpeStatus (Event, EventStatus);
        }

        break;

    default:
        Status = AE_BAD_PARAMETER;
    }

    return_ACPI_STATUS (Status);
}

