/*
 *  ./acpica/subsystem/events/evapirgn.c - External Interfaces, ACPI Operation Regions and
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __EVAPIRGN_C__

#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>
#include <amlcode.h>
#include <interp.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evapirgn");



ACPI_STATUS
AcpiInstallAddressSpaceHandler (
    ACPI_HANDLE             Device, 
    ACPI_ADDRESS_SPACE_TYPE SpaceId, 
    ADDRESS_SPACE_HANDLER   Handler, 
    ADDRESS_SPACE_SETUP     Setup,
    void                    *Context)
{
    ACPI_OBJECT_INTERNAL   *ObjDesc;
    ACPI_OBJECT_INTERNAL   *HandlerObj;
    NAME_TABLE_ENTRY       *ObjEntry;
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_TYPE        Type;
    UINT16                  Flags = 0;


    FUNCTION_TRACE ("AcpiInstallAddressSpaceHandler");

    /* Parameter validation */

    if ((!Device)   ||
        ((!Handler)  && (Handler != ACPI_DEFAULT_HANDLER)) ||
        (SpaceId > ACPI_MAX_ADDRESS_SPACE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* Convert and validate the device handle */

    ObjEntry = NsConvertHandleToEntry (Device);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /*
     *  This registration is valid for only the types below
     *  and the root.  This is where the default handlers
     *  get placed.
     */

    if ((ObjEntry->Type != ACPI_TYPE_Device)     &&
        (ObjEntry->Type != ACPI_TYPE_Processor)  &&
        (ObjEntry->Type != ACPI_TYPE_Thermal)    &&
        (ObjEntry != Gbl_RootObject))
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    if (Handler == ACPI_DEFAULT_HANDLER)
    {
        Flags = ADDR_HANDLER_DEFAULT_INSTALLED;

        switch (SpaceId)
        {
        case ADDRESS_SPACE_SYSTEM_MEMORY:
            Handler = AmlSystemMemorySpaceHandler;
            Setup = (ADDRESS_SPACE_SETUP) EvSystemMemoryRegionSetup;
            break;

        case ADDRESS_SPACE_SYSTEM_IO:
            Handler = AmlSystemIoSpaceHandler;
            Setup = (ADDRESS_SPACE_SETUP) EvIoSpaceRegionSetup;
            break;

        case ADDRESS_SPACE_PCI_CONFIG:
            Handler = AmlPciConfigSpaceHandler;
            Setup = (ADDRESS_SPACE_SETUP) EvPciConfigRegionSetup;
            break;

        default:
            Status = AE_NOT_EXIST;
            goto UnlockAndExit;
            break;
        }
    }

    /*
     *  If the caller hasn't specified a setup routine, use the default
     */
    if (!Setup) 
    {
        Setup = (ADDRESS_SPACE_SETUP) EvDefaultRegionSetup;
    }

    /*
     *  Check for an existing internal object
     */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) ObjEntry);
    if (ObjDesc)
    {
        /*
         *  The object exists.
         *  Make sure the handler is not already installed.
         */

        /* check the address handler the user requested */

        HandlerObj = ObjDesc->Device.AddrHandler;
        while (HandlerObj)
        {
            /*
             *  We have an Address handler, see if user requested this
             *  address space.
             */
            if(HandlerObj->AddrHandler.SpaceId == SpaceId)
            {
                Status = AE_EXIST;
                goto UnlockAndExit;
            }

            /*
             *  Move through the linked list of handlers
             */
            HandlerObj = HandlerObj->AddrHandler.Link;
        }
    }

    else
    {
        DEBUG_PRINT (TRACE_OPREGION, ("Creating object on Device 0x%X while installing handler\n", 
                        ObjEntry));

        /* ObjDesc does not exist, create one */

        if (ObjEntry->Type == ACPI_TYPE_Any)
        {
            Type = ACPI_TYPE_Device;
        }

        else
        {
            Type = ObjEntry->Type;
        }

        ObjDesc = CmCreateInternalObject (Type);
        if (!ObjDesc)
        {
            Status = AE_NO_MEMORY;
            goto UnlockAndExit;
        }

        /* Init new descriptor */

        ObjDesc->Common.Type = (UINT8) Type;

        /* Attach the new object to the NTE */

        Status = NsAttachObject (Device, ObjDesc, (UINT8) Type);
        if (ACPI_FAILURE (Status))
        {
            CmRemoveReference (ObjDesc);
            goto UnlockAndExit;
        }

        /* TBD: Will this always be of type DEVICE? */

        if (Type == ACPI_TYPE_Device)
        {
            ObjDesc->Device.Handle = Device;
        }
    }

    DEBUG_PRINT (TRACE_OPREGION, ("Installing address handler for %s on Device 0x%X (0x%X)\n", 
                    Gbl_RegionTypes[SpaceId], ObjEntry, ObjDesc));

    /* 
     *  Now we can install the handler
     *
     *  At this point we know that the handler we are installing DNE. 
     *  So, we just allocate the object for the handler and link it
     *  into the list.
     */
    HandlerObj = CmCreateInternalObject (INTERNAL_TYPE_AddressHandler);
    if (!HandlerObj)
    {
        Status = AE_NO_MEMORY;
        goto UnlockAndExit;
    }

    HandlerObj->AddrHandler.SpaceId             = (UINT16) SpaceId;
    HandlerObj->AddrHandler.Hflags              = Flags;
    HandlerObj->AddrHandler.Link                = ObjDesc->Device.AddrHandler;
    HandlerObj->AddrHandler.RegionList          = NULL;
    HandlerObj->AddrHandler.Nte                 = ObjEntry;
    HandlerObj->AddrHandler.Handler             = Handler;
    HandlerObj->AddrHandler.Context             = NULL;
    HandlerObj->AddrHandler.InstallTimeContext  = Context;
    HandlerObj->AddrHandler.Setup               = Setup;

    /*
     *  Now walk the namespace finding all of the regions this
     *  handler will manage.
     *
     *  We start at the device and search the branch toward
     *  the leaf nodes until either the leaf is encountered or
     *  a device is detected that has an address handler of the
     *  same type.
     *
     *  In either case we back up and search down the remainder
     *  of the branch
     */
    Status = NsWalkNamespace (ACPI_TYPE_Any, Device, ACPI_INT32_MAX, NS_WALK_NO_UNLOCK, EvAddrHandlerHelper, 
                                HandlerObj, NULL);

    /*
     *  Place this handler 1st on the list
     */

    HandlerObj->Common.ReferenceCount = (UINT16) (HandlerObj->Common.ReferenceCount + ObjDesc->Common.ReferenceCount - 1);
    ObjDesc->Device.AddrHandler = HandlerObj;



UnlockAndExit:
    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiRemoveAddressSpaceHandler (
    ACPI_HANDLE             Device, 
    ACPI_ADDRESS_SPACE_TYPE SpaceId, 
    ADDRESS_SPACE_HANDLER   Handler)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *HandlerObj;
    ACPI_OBJECT_INTERNAL    *RegionObj;
    ACPI_OBJECT_INTERNAL    **LastObjPtr;
    NAME_TABLE_ENTRY        *ObjEntry;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AcpiRemoveAddressSpaceHandler");


    /* Parameter validation */

    if ((!Device)   ||
        ((!Handler)  && (Handler != ACPI_DEFAULT_HANDLER)) ||
        (SpaceId > ACPI_MAX_ADDRESS_SPACE))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* Convert and validate the device handle */

    ObjEntry = NsConvertHandleToEntry (Device);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }


    /* Make sure the internal object exists */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) ObjEntry);
    if (!ObjDesc)
    {
        /*
         *  The object DNE.
         */
        Status = AE_NOT_EXIST;
        goto UnlockAndExit;
    }

    /*
     *  find the address handler the user requested
     */

    HandlerObj = ObjDesc->Device.AddrHandler;
    LastObjPtr = &ObjDesc->Device.AddrHandler;
    while (HandlerObj)
    {
        /*
         *  We have a handler, see if user requested this one
         */

        if(HandlerObj->AddrHandler.SpaceId == SpaceId)
        {
            /*
             *  Got it, first dereference this in the Regions
             */
            DEBUG_PRINT (TRACE_OPREGION, ("Removing address handler 0x%X (0x%X) for %s on Device 0x%X (0x%X)\n", 
                            HandlerObj, Handler, Gbl_RegionTypes[SpaceId], ObjEntry, ObjDesc));

            RegionObj = HandlerObj->AddrHandler.RegionList;

            while (RegionObj)
            {
                /*
                 *  First disassociate the handler from the region.
                 *
                 *  NOTE: this doesn't mean that the region goes away
                 *  The region is just inaccessible as indicated to
                 *  the _REG method
                 */
                EvDisassociateRegionFromHandler(RegionObj);

                /*
                 *  Walk the list, since we took the first region and it
                 *  was removed from the list by the dissassociate call
                 *  we just get the first item on the list again
                 */
                RegionObj = HandlerObj->AddrHandler.RegionList;

            } /* walk the handler's region list */

            /*
             *  Remove this Handler object from the list
             */
            *LastObjPtr = HandlerObj->AddrHandler.Link;

            /*
             *  Now we can delete the handler object
             */
            CmRemoveReference (HandlerObj);
            CmRemoveReference (HandlerObj);

            goto UnlockAndExit;

        } /* found the right handler */

        /*
         *  Move through the linked list of handlers
         */
        LastObjPtr = &HandlerObj->AddrHandler.Link;
        HandlerObj = HandlerObj->AddrHandler.Link;
    }

    /*
     *  If we get here the handler DNE, get out with error
     */
    DEBUG_PRINT (TRACE_OPREGION,
        ("Unable to remove address handler xxxx (0x%X) for %s on Device 0x%X (0x%X)\n", 
        Handler, Gbl_RegionTypes[SpaceId], ObjEntry, ObjDesc));


    Status = AE_NOT_EXIST;


UnlockAndExit:
    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}



