/*
 *  ./acpica/subsystem/events/evfixed.c - Fixed Event handling and dispatch
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <hardware.h>
#include <events.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evfixed");



ACPI_STATUS
EvFixedEventInitialize(void)
{
    int                     i = 0;

    /* Initialize the structure that keeps track of fixed event handlers */

    for (i = 0; i < NUM_FIXED_EVENTS; i++)
    {
        Gbl_FixedEventHandlers[i].Handler = NULL;
        Gbl_FixedEventHandlers[i].Context = NULL;
    }

    HwRegisterIO (ACPI_WRITE, MTX_LOCK, EVENT_PMTIMER + TMR_EN, 0);
    HwRegisterIO (ACPI_WRITE, MTX_LOCK, EVENT_GLOBAL + TMR_EN, 0);
    HwRegisterIO (ACPI_WRITE, MTX_LOCK, EVENT_POWER_BUTTON + TMR_EN, 0);
    HwRegisterIO (ACPI_WRITE, MTX_LOCK, EVENT_SLEEP_BUTTON + TMR_EN, 0);
    HwRegisterIO (ACPI_WRITE, MTX_LOCK, EVENT_RTC + TMR_EN, 0);

    return AE_OK;
}



UINT32
EvFixedEventDetect(void)
{
    UINT32                  IntStatus = INTERRUPT_NOT_HANDLED;
    UINT32                  StatusRegister = 0;
    UINT32                  EnableRegister = 0;

    /* 
     * Read the fixed feature status and enable registers, as all the cases
     * depend on their values.  
     */

    StatusRegister = (UINT32) OsdIn16 (Gbl_FACP->Pm1aEvtBlk);
    if (Gbl_FACP->Pm1bEvtBlk)
    {
        StatusRegister |= (UINT32) OsdIn16 (Gbl_FACP->Pm1bEvtBlk);
    }
        
    EnableRegister = (UINT32) OsdIn16 ((Gbl_FACP->Pm1aEvtBlk + Gbl_FACP->Pm1EvtLen / 2));
    if (Gbl_FACP->Pm1bEvtBlk)
    {
        EnableRegister |= (UINT32) OsdIn16 ((Gbl_FACP->Pm1bEvtBlk + Gbl_FACP->Pm1EvtLen / 2));
    }
        
    DEBUG_PRINT (TRACE_INTERRUPTS, ("Fixed Event Block: Enable = %08x\tStatus = %08x\n", EnableRegister, StatusRegister));

    /* power management timer roll over */

    if ((StatusRegister & STATUS_PMTIMER) && (EnableRegister & ENABLE_PMTIMER))
    {
        IntStatus |= EvFixedEventDispatch (EVENT_PMTIMER);
    }

    /* global event (BIOS want's the global lock) */

    if ((StatusRegister & STATUS_GLOBAL) && (EnableRegister & ENABLE_GLOBAL))
    {
        IntStatus |= EvFixedEventDispatch (EVENT_GLOBAL);
    }

    /* power button event */

    if ((StatusRegister & STATUS_POWER_BUTTON) && (EnableRegister & ENABLE_POWER_BUTTON))
    {
        IntStatus |= EvFixedEventDispatch (EVENT_POWER_BUTTON);
    }

    /* sleep button event */

    if ((StatusRegister & STATUS_SLEEP_BUTTON) && (EnableRegister & ENABLE_SLEEP_BUTTON))
    {
        IntStatus |= EvFixedEventDispatch (EVENT_SLEEP_BUTTON);
    }
 
    return IntStatus;
}



UINT32
EvFixedEventDispatch (
    UINT32                  Event)
{
    /* Clear the status bit */

    HwRegisterIO (ACPI_WRITE, MTX_DO_NOT_LOCK, (INT32)TMR_STS + Event, 1);

    /* 
     * Make sure we've got a handler.  If not, report an error.
     * The event is disabled to prevent further interrupts.
     */
    if (NULL == Gbl_FixedEventHandlers[Event].Handler)
    {
        HwRegisterIO(ACPI_WRITE, MTX_DO_NOT_LOCK, TMR_EN + Event, 0);

        REPORT_ERROR("No installed handler for fixed event.");
        DEBUG_PRINT (ACPI_ERROR, ("EvGpeDispatch: No installed handler for fixed event [0x%08X].", Event));

        return INTERRUPT_NOT_HANDLED;
    }
    
    /* Invoke the handler */

    return (Gbl_FixedEventHandlers[Event].Handler) (Gbl_FixedEventHandlers[Event].Context);
}
    
