
/*
 *  ./acpica/subsystem/events/evglock.c - ACPI Global Lock support
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <events.h>
#include <interp.h>
#include <acpiosd.h>
#include <hardware.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evglock");



void
EvGlobalLockThread (
    void                    *Context)
{

    /* Signal threads that are waiting for the lock */

    if (Gbl_GlobalLockThreadCount)
    {
        /* Send sufficient units to the semaphore */

        OsdSignalSemaphore (Gbl_GlobalLockSemaphore, Gbl_GlobalLockThreadCount);
    }
}



UINT32
EvGlobalLockHandler (
    void                    *Context)
{
    BOOLEAN                 Acquired = FALSE;
    void                    *GlobalLock;


    /* 
     * Attempt to get the lock
     * If we don't get it now, it will be marked pending and we will
     * take another interrupt when it becomes free.
     */

    GlobalLock = &Gbl_FACS->GlobalLock;
    ASM_AcquireGL (GlobalLock, Acquired);
    if (Acquired)
    {
        /* Got the lock, now wake all threads waiting for it */

        Gbl_GlobalLockAcquired = TRUE;

        /* Run the Global Lock thread which will signal all waiting threads */

        OsdQueueForExecution (OSD_PRIORITY_HIGH, EvGlobalLockThread, Context);
    }

    return INTERRUPT_HANDLED;
}



ACPI_STATUS
EvInitGlobalLockHandler (void)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("EvInitGlobalLockHandler");



    Status = AcpiInstallFixedEventHandler (EVENT_GLOBAL, EvGlobalLockHandler, NULL);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
EvAcquireGlobalLock(void)
{
    ACPI_STATUS             Status = AE_OK;
    BOOLEAN                 Acquired = FALSE;
    void                    *GlobalLock;


    FUNCTION_TRACE ("EvAcquireGlobalLock");


    /* One more thread wants the global lock */

    Gbl_GlobalLockThreadCount++;


    /* If we (OS side) have the hardware lock already, we are done */

    if (Gbl_GlobalLockAcquired)
    {
        return_ACPI_STATUS (AE_OK);
    }

    /* Only if the FACS is valid */

    if (!Gbl_FACS)
    {
        return_ACPI_STATUS (AE_OK);
    }


    /* We must acquire the actualy hardware lock */

    GlobalLock = &Gbl_FACS->GlobalLock;
    ASM_AcquireGL (GlobalLock, Acquired);
    if (Acquired)
    {
       /* We got the lock */
    
        DEBUG_PRINT (ACPI_INFO, ("Acquired the HW Global Lock\n"));

        Gbl_GlobalLockAcquired = TRUE;

        return_ACPI_STATUS (AE_OK);
    }


    /*
     * Did not get the lock.  The pending bit was set above, and we must now
     * wait until we get the global lock released interrupt.
     */

    DEBUG_PRINT (ACPI_INFO, ("Waiting for the HW Global Lock\n"));

     /* 
      * Acquire the global lock semaphore first.
      * Since this wait will block, we must release the interpreter
      */

    AmlExitInterpreter ();
    Status = OsLocalWaitSemaphore (Gbl_GlobalLockSemaphore, ACPI_UINT32_MAX);
    AmlEnterInterpreter ();


    return_ACPI_STATUS (Status);
}



void
EvReleaseGlobalLock (void)
{
    BOOLEAN                 Pending = FALSE;
    void                    *GlobalLock;
    
    
    FUNCTION_TRACE ("EvReleaseGlobalLock");


    if (!Gbl_FACS)
    {
        return_VOID;
    }

   /* One fewer thread has the global lock */

    Gbl_GlobalLockThreadCount--;


    /* Have all threads released the lock? */

    if (!Gbl_GlobalLockThreadCount)
    {
        /* No more threads holding lock, we can do the actual hardware release */

        GlobalLock = &Gbl_FACS->GlobalLock;
        ASM_ReleaseGL (GlobalLock, Pending);
        Gbl_GlobalLockAcquired = FALSE;

        /*
         * If the pending bit was set, we must write GBL_RLS to the control register
         */
        if (Pending)
        {
            HwRegisterIO (ACPI_WRITE, MTX_LOCK, (INT32)PM1_CONTROL | GBL_RLS, 1);
        }
    }

    return_VOID;
}
