/*
 *  ./acpica/subsystem/events/evgpe.c - General Purpose Event handling and dispatch
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>
#include <common.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evgpe");



ACPI_STATUS
EvGpeInitialize (void)
{
    UINT32                  i;
    UINT32                  j;
    UINT32                  RegisterIndex;
    UINT32                  GpeNumber;
    UINT16                  Gpe0RegisterCount;
    UINT16                  Gpe1RegisterCount;


    FUNCTION_TRACE ("EvGpeInitialize");

    /* 
     * Setup various GPE counts
     */  
    
    Gpe0RegisterCount       = (UINT16) (Gbl_FACP->Gpe0BlkLen / 2);
    Gpe1RegisterCount       = (UINT16) (Gbl_FACP->Gpe1BlkLen / 2);
    Gbl_GpeRegisterCount    = Gpe0RegisterCount + Gpe1RegisterCount;

    /* 
     * Allocate the Gpe information block 
     */

    Gbl_GpeRegisters = CmCallocate (Gbl_GpeRegisterCount * sizeof (GPE_REGISTERS));
    if (!Gbl_GpeRegisters)
    {
        DEBUG_PRINT (ACPI_ERROR, ("Could not allocate the GpeRegisters block\n"));
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* 
     * Allocate the Gpe dispatch handler block 
     * There are eight distinct GP events per register.
     * Initialization to zeros is sufficient
     */

    Gbl_GpeInfo = CmCallocate ((Gbl_GpeRegisterCount * 8) * sizeof (GPE_LEVEL_INFO));
    if (!Gbl_GpeInfo)
    {
        CmFree (Gbl_GpeRegisters);
        DEBUG_PRINT (ACPI_ERROR, ("Could not allocate the GpeInfo block\n"));
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Set the Gpe validation table to GPE_INVALID */

    MEMSET (Gbl_GpeValid, (int) GPE_INVALID, NUM_GPE);

    /* 
     * Initialize the Gpe information and validation blocks.  A goal of these 
     * blocks is to hide the fact that there are two separate GPE register sets
     * In a given block, the status registers occupy the first half, and 
     * the enable registers occupy the second half.
     */

    /* GPE Block 0 */

    RegisterIndex = 0;

    for (i = 0; i < Gpe0RegisterCount; i++)
    {
        Gbl_GpeRegisters[RegisterIndex].StatusAddr  = (UINT16) (Gbl_FACP->Gpe0Blk + i);
        Gbl_GpeRegisters[RegisterIndex].EnableAddr  = (UINT16) (Gbl_FACP->Gpe0Blk + i + Gpe0RegisterCount);
        Gbl_GpeRegisters[RegisterIndex].GpeBase     = (UINT8) (i * 8);

        for (j = 0; j < 8; j++)
        {
            GpeNumber = Gbl_GpeRegisters[RegisterIndex].GpeBase + j;
            Gbl_GpeValid[GpeNumber] = (UINT8) RegisterIndex;
        }

        /* 
         * Clear the status/enable registers.  Note that status registers
         * are cleared by writing a '1', while enable registers are cleared
         * by writing a '0'.
         */
        OsdOut8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr, 0x00);
        OsdOut8 (Gbl_GpeRegisters[RegisterIndex].StatusAddr, 0xFF);
        
        RegisterIndex++;
    }

    /* GPE Block 1 */

    for (i = 0; i < Gpe1RegisterCount; i++)
    {
        Gbl_GpeRegisters[RegisterIndex].StatusAddr  = (UINT16) (Gbl_FACP->Gpe1Blk + i);
        Gbl_GpeRegisters[RegisterIndex].EnableAddr  = (UINT16) (Gbl_FACP->Gpe1Blk + i + Gpe1RegisterCount);
        Gbl_GpeRegisters[RegisterIndex].GpeBase     = (UINT8) (Gbl_FACP->Gpe1Base + (i * 8));

        for (j = 0; j < 8; j++)
        {
            GpeNumber = Gbl_GpeRegisters[RegisterIndex].GpeBase + j;
            Gbl_GpeValid[GpeNumber] = (UINT8) RegisterIndex;
        }

        /* 
         * Clear the status/enable registers.  Note that status registers
         * are cleared by writing a '1', while enable registers are cleared
         * by writing a '0'.
         */
        OsdOut8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr, 0x00);
        OsdOut8 (Gbl_GpeRegisters[RegisterIndex].StatusAddr, 0xFF);
        
        RegisterIndex++;
    }

    DEBUG_PRINT (ACPI_INFO, ("GPE registers: %d@%X (Blk0) %d@%X (Blk1)\n",
        Gpe0RegisterCount, Gbl_FACP->Gpe0Blk, Gpe1RegisterCount, Gbl_FACP->Gpe1Blk));

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
EvSaveMethodInfo (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *ObjDesc,
    void                    **ReturnValue)
{
    UINT32                  GpeNumber;
    char                    Name[ACPI_NAME_SIZE + 1];
    UINT8                   Type;

    /* Extract the name from the object and convert to a string */

    *(UINT32 *) Name = ((NAME_TABLE_ENTRY *) ObjHandle)->Name;
    Name[ACPI_NAME_SIZE] = 0;    

    /* 
     * Edge/Level determination is based on the 2nd char of the method name
     */
    if (Name[1] == 'L')
    {
        Type = EVENT_LEVEL_TRIGGERED;
    }
    else if (Name[1] == 'E')
    {
        Type = EVENT_EDGE_TRIGGERED;
    }
    else
    {
        /* Unknown method type, just ignore it! */
        
        DEBUG_PRINT (ACPI_ERROR, 
            ("EvSaveMethodInfo: Unknown GPE method type: %s (name not of form _Lnn or _Enn)\n", 
                Name));
        return AE_OK;
    }

    /* Convert the last two characters of the name to the Gpe Number */

    GpeNumber = STRTOUL (&Name[2], NULL, 16);
    if (GpeNumber == ACPI_UINT32_MAX)
    {
        /* Conversion failed; invalid method, just ignore it */

        DEBUG_PRINT (ACPI_ERROR, 
            ("EvSaveMethodInfo: Could not extract GPE number from name: %s (name not of form _Lnn or _Enn)\n", 
                Name));
        return AE_OK;
    }

    /* Ensure that we have a valid GPE number */

    if (Gbl_GpeValid[GpeNumber] == GPE_INVALID)
    {
        /* Not valid, all we can do here is ignore it */

        return AE_OK;
    }

    /* 
     * Now we can add this information to the GpeInfo block 
     * for use during dispatch of this GPE.
     */

    Gbl_GpeInfo [GpeNumber].Type            = Type;
    Gbl_GpeInfo [GpeNumber].MethodHandle    = ObjHandle;


    /*
     * Enable the GPE (SCIs should be disabled at this point)
     */

    HwEnableGpe (GpeNumber);

    DEBUG_PRINT (ACPI_INFO, ("EvSaveMethodInfo: Registered GPE method %s as GPE number %d\n", 
                                Name, GpeNumber));
    return AE_OK;
}




ACPI_STATUS
EvInitGpeControlMethods (void)
{
    ACPI_STATUS             Status;

    FUNCTION_TRACE ("EvInitGpeControlMethods");

    /* Get a permanent handle to the _GPE object */

    Status = AcpiGetHandle (NULL, "\\_GPE", &Gbl_GpeObjHandle);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Traverse the namespace under \_GPE to find all methods there */

    Status = AcpiWalkNamespace (ACPI_TYPE_Method, Gbl_GpeObjHandle, ACPI_INT32_MAX, EvSaveMethodInfo, NULL, NULL);

    return_ACPI_STATUS (Status);
}



void
EvGpeCleanup (void)
{
 FUNCTION_TRACE ("EvGpeCleanup");
 
    CmFree (Gbl_GpeRegisters);
    CmFree (Gbl_GpeInfo);

 return_VOID;
}



UINT32
EvGpeDetect (void)
{
    UINT32                  IntStatus = INTERRUPT_NOT_HANDLED;
    UINT32                  i;
    UINT32                  j;
    UINT8                   EnabledStatusByte;
    UINT8                   BitMask;

    /* 
     * Read all of the 8-bit GPE status and enable registers 
     * in both of the register blocks, saving all of it.
     * Find all currently active GP events.
     */

    for (i = 0; i < Gbl_GpeRegisterCount; i++)
    {
        Gbl_GpeRegisters[i].Status = OsdIn8 (Gbl_GpeRegisters[i].StatusAddr);
        Gbl_GpeRegisters[i].Enable = OsdIn8 (Gbl_GpeRegisters[i].EnableAddr);

        DEBUG_PRINT (TRACE_INTERRUPTS,
            ("GPE block at %x - Enable: %08x\tStatus: %08x\n",
            Gbl_GpeRegisters[i].EnableAddr, Gbl_GpeRegisters[i].Status, Gbl_GpeRegisters[i].Enable));

        /* First check if there is anything active at all in this register */

        EnabledStatusByte = (UINT8) (Gbl_GpeRegisters[i].Status & Gbl_GpeRegisters[i].Enable);
        if (!EnabledStatusByte)
        {
            /* No active GPEs in this register, move on */

            continue;
        }

        /* Now look at the individual GPEs in this byte register */

        for (j = 0, BitMask = 1; j < 8; j++, BitMask <<= 1)
        {
            /* Examine one GPE bit */

            if (EnabledStatusByte & BitMask)
            {
                /* 
                 * Found an active GPE.  Dispatch the event to a handler 
                 * or method.
                 */
                IntStatus |= EvGpeDispatch (Gbl_GpeRegisters[i].GpeBase + j);
            }
        }
    }

    return IntStatus;
}



void
EvAsynchExecuteGpeMethod (
    void                    *Context)
{
    UINT32                  GpeNumber = (UINT32) ((NATIVE_UINT) Context);
    GPE_LEVEL_INFO          GpeInfo;


    FUNCTION_TRACE ("EvAsynchExecuteGpeMethod");


    /* Take a snapshot of the GPE info for this level */

    CmAcquireMutex (MTX_EVENTS);
    GpeInfo = Gbl_GpeInfo [GpeNumber];
    CmReleaseMutex (MTX_EVENTS);

    /*
     * Function Handler (e.g. EC):
     * ---------------------------
     * Execute the installed function handler to handle this event.
     */
    if (GpeInfo.Handler)
    {
        GpeInfo.Handler (GpeInfo.Context);
    }

    /*
     * Method Handler (_Lxx, _Exx):
     * ----------------------------
     * Evaluate the _Lxx/_Exx control method that corresponds to this GPE.
     */
    else if (GpeInfo.MethodHandle)
    {
        NsEvaluateByHandle (GpeInfo.MethodHandle, NULL, NULL);
    }

    /* 
     * Level-Triggered?
     * ----------------
     * If level-triggered, clear the GPE status bit after execution.  Note  
     * that edge-triggered events are cleared prior to calling (via DPC)
     * this function.
     */
    if (GpeInfo.Type | EVENT_LEVEL_TRIGGERED)
    {
        HwClearGpe (GpeNumber);
    }

    /* 
     * Enable the GPE.
     */
    HwEnableGpe (GpeNumber);

    return_VOID;
}



UINT32
EvGpeDispatch (
    UINT32                  GpeNumber)
{
    FUNCTION_TRACE ("EvGpeDispatch");

    DEBUG_PRINT (ACPI_INFO, ("GPE [%d] event occurred.\n", GpeNumber));

    /*DEBUG_INCREMENT_EVENT_COUNT (EVENT_GENERAL);*/

    /* Ensure that we have a valid GPE number */

    if (Gbl_GpeValid[GpeNumber] == GPE_INVALID)
    {
        DEBUG_PRINT (ACPI_ERROR, ("Invalid GPE [%d].\n", GpeNumber));
        return_VALUE (INTERRUPT_NOT_HANDLED);
    }

    /* 
     * Disable the GPE.
     */
    HwDisableGpe (GpeNumber);

    /* 
     * Edge-Triggered?
     * ---------------
     * If edge-triggered, clear the GPE status bit now.  Note that
     * level-triggered events are cleared after the GPE is serviced 
     * (see EvAsynchExecuteGpeMethod).
     */
    if (Gbl_GpeInfo [GpeNumber].Type | EVENT_EDGE_TRIGGERED)
    {
        HwClearGpe (GpeNumber);
    }

    /* 
     * Queue-up the Handler:
     * ---------------------
     * Queue the handler, which is either an installable function handler
     * (e.g. EC) or a control method (e.g. _Lxx/_Exx) for later execution.
     */
    if (Gbl_GpeInfo [GpeNumber].Handler || Gbl_GpeInfo [GpeNumber].MethodHandle)
    {
        if (ACPI_FAILURE(OsdQueueForExecution (OSD_PRIORITY_GPE, EvAsynchExecuteGpeMethod, (void*)(NATIVE_UINT)GpeNumber)))
        {
            /*
             * Shoudn't occur, but if it does report an error. Note that 
             * the GPE will remain disabled until the ACPI Core Subsystem
             * is restarted, or the handler is removed/reinstalled.
             */
            REPORT_ERROR("Unable to queue-up handler for GPE.");
            DEBUG_PRINT (ACPI_ERROR, ("EvGpeDispatch: Unable to queue-up handler for GPE [0x%08X].", GpeNumber));
        }
    }

    /*
     * Non Handled GPEs:
     * -----------------
     * GPEs without handlers are disabled and kept that way until a handler 
     * is registered for them.
     */
    else
    {   
        REPORT_ERROR("No installed handler for GPE.");
        DEBUG_PRINT (ACPI_ERROR, ("EvGpeDispatch: No installed handler for GPE [0x%08X].", GpeNumber));
    }

    return_VALUE(INTERRUPT_HANDLED);
}
