/*
 *  ./acpica/subsystem/events/evnotify.c - ACPI device notification handler dispatch
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <events.h>
#include <namesp.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evnotify");




void
EvNotifyDispatch (
    ACPI_HANDLE             Device, 
    UINT32                  NotifyValue)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *HandlerObj;
    NOTIFY_HANDLER          Handler;


    /*
     * For value 1 (Ejection Request), some device method may need to be run.
     * For value 2 (Device Wake) if _PRW exists, the _PS0 method may need to be run.
     * For value 0x80 (Status Change) on the power button or sleep button, 
     * initiate soft-off or sleep operation?
     */


    DEBUG_PRINT (ACPI_INFO, ("Dispatching Notify(%d) on device %p\n", NotifyValue, Device));
    
    switch (NotifyValue)
    {
    case 0:
        DEBUG_PRINT (ACPI_INFO, ("Notify value: Re-enumerate Devices\n"));
        break;

    case 1:
        DEBUG_PRINT (ACPI_INFO, ("Notify value: Ejection Request\n"));
        break;

    case 2:
        DEBUG_PRINT (ACPI_INFO, ("Notify value: Device Wake\n"));
        break;

    case 0x80:
        DEBUG_PRINT (ACPI_INFO, ("Notify value: Status Change\n"));
        break;

    default:
        DEBUG_PRINT (ACPI_INFO, ("Unknown Notify Value: %lx \n", NotifyValue));
        break;
    }


    /* 
     * Invoke a global notify handler if installed.
     * This is done _before_ we invoke the per-device handler attached to the device.
     */

    if (NotifyValue <= MAX_SYS_NOTIFY)
    {
        /* Global system notification handler */

        if (Gbl_SysNotify.Handler)
        {
            Gbl_SysNotify.Handler (Device, NotifyValue, Gbl_SysNotify.Context);
        }
    }

    else
    {
        /* Global driver notification handler */

        if (Gbl_DrvNotify.Handler)
        {
            Gbl_DrvNotify.Handler (Device, NotifyValue, Gbl_DrvNotify.Context);
        }
    }

    
    /*
     * Get the notify object which must be attached to the device NTE
     */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) Device);
    if (!ObjDesc)
    {
        /* There can be no notify handler for this device */

        DEBUG_PRINT (ACPI_INFO, ("No notify handler for device %p \n", Device));
        return;
    }


    /* We have the notify object, Get the right handler */

    if (NotifyValue <= MAX_SYS_NOTIFY)
    {
        HandlerObj = ObjDesc->Device.SysHandler;
    }
    else
    {
        HandlerObj = ObjDesc->Device.DrvHandler;
    }

    /* Validate the handler */

    if (!HandlerObj)
    {
        /* There is no notify handler for this device */

        DEBUG_PRINT (ACPI_INFO, ("No notify handler for device %p \n", Device));
        return;
    }

    /* There is a handler, invoke it */

    Handler = HandlerObj->NotifyHandler.Handler;
    Handler (Device, NotifyValue, HandlerObj->NotifyHandler.Context);

}
