/*
 *  ./acpica/subsystem/events/evregion.c - ACPI AddressSpace / OpRegion handler dispatch
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __EVREGION_C__

#include <acpi.h>
#include <events.h>
#include <namesp.h>
#include <interp.h>
#include <amlcode.h>
#include <methods.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evregion");




#define PCI_ROOT_HID_STRING        "PNP0A03"
#define PCI_ROOT_HID_VALUE         0x030AD041       /* EISAID("PNP0A03") */




ACPI_STATUS
EvFindOnePciRootBus (
    ACPI_HANDLE             ObjHandle,
    UINT32                  NestingLevel,
    void                    *Context,
    void                    **ReturnValue)
{
    NAME_TABLE_ENTRY        *Entry;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;

    
    Entry = (NAME_TABLE_ENTRY *) ObjHandle;
    ObjDesc = ((NAME_TABLE_ENTRY *)ObjHandle)->Object;


    /*
     * We are looking for all valid _HID objects.
     */

    if ((Entry->Name != *(UINT32 *) METHOD_NAME__HID) ||
        (!ObjDesc))
    {
        return AE_OK;
    }


    /*
     * Found an _HID object. 
     * Now we need a HID with the value EISAID("PNP0A03")
     * HID can be either a number or a string.
     */

    switch (ObjDesc->Common.Type)
    {
    case ACPI_TYPE_Number:

        if (ObjDesc->Number.Value != PCI_ROOT_HID_VALUE)
        {
            return AE_OK;
        }

        break;

    case ACPI_TYPE_String:

        if (STRNCMP (ObjDesc->String.Pointer, PCI_ROOT_HID_STRING, sizeof (PCI_ROOT_HID_STRING)))
        {
            return AE_OK;
        }

        break;

    default:

        return AE_OK;
    }


    /*
     * We found a valid PCI_ROOT_HID.
     * The parent of the HID entry is the PCI device;  Install the default PCI
     * handler for this PCI device.
     */

    Status = AcpiInstallAddressSpaceHandler (Entry->ParentEntry, ADDRESS_SPACE_PCI_CONFIG, 
                                                ACPI_DEFAULT_HANDLER, NULL, NULL);

    return AE_OK;
}




ACPI_STATUS
EvFindPciRootBuses (
    void)
{
    
    NsWalkNamespace (ACPI_TYPE_Any, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX,
                        FALSE, EvFindOnePciRootBus, NULL, NULL);

    return AE_OK;
}




ACPI_STATUS
EvInstallDefaultAddressSpaceHandlers (
    void)
{
    ACPI_STATUS             Status;
    

    FUNCTION_TRACE ("EvInstallDefaultAddressSpaceHandlers");

    /*
     * NOTE:    All address spaces (PCI Config, EC, SMBus) are scope dependent
     *          and registration must occur for a specific device.  In the case
     *          system memory and IO address spaces there is currently no device
     *          associated with the address space.  For these we use the root.
     */

    Status = AcpiInstallAddressSpaceHandler (Gbl_RootObject, ADDRESS_SPACE_SYSTEM_MEMORY, 
                                                ACPI_DEFAULT_HANDLER, NULL, NULL);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    Status = AcpiInstallAddressSpaceHandler (Gbl_RootObject,
                ADDRESS_SPACE_SYSTEM_IO, ACPI_DEFAULT_HANDLER, NULL, NULL);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /*
     *  Install PCI config space handler for all PCI root bridges.  A PCI root
     *  bridge is found by searching for devices containing a HID with the value
     *  EISAID("PNP0A03")
     */

    EvFindPciRootBuses ();


    return_ACPI_STATUS (AE_OK);
}


/* TBD: Move to the methods directory */


ACPI_STATUS
EvExecuteRegMethod (
    ACPI_OBJECT_INTERNAL   *RegionObj,
    UINT32                  Function)
{
    ACPI_OBJECT_INTERNAL   *Params[3];
    ACPI_OBJECT_INTERNAL    SpaceID_Obj;
    ACPI_OBJECT_INTERNAL    Function_Obj;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("EvExecuteRegMethod");


    if (RegionObj->Region.REGMethod == NULL)
    {
        return_ACPI_STATUS (AE_OK);
    }

    /*
     *  _REG method has two arguments
     *  Arg0:   Integer: Operation region space ID
     *          Same value as RegionObj->Region.SpaceId
     *  Arg1:   Integer: connection status
     *          1 for connecting the handler,
     *          0 for disconnecting the handler
     *          Passed as a parameter
     */

    CmInitStaticObject (&SpaceID_Obj);
    CmInitStaticObject (&Function_Obj);

    /*
     *  Method requires two parameters.
     */
    Params [0] = &SpaceID_Obj;
    Params [1] = &Function_Obj;
    Params [2] = NULL;

    /*
     *  Set up the parameter objects
     */
    SpaceID_Obj.Common.Type    = ACPI_TYPE_Number;
    SpaceID_Obj.Number.Value   = RegionObj->Region.SpaceId;

    Function_Obj.Common.Type   = ACPI_TYPE_Number;
    Function_Obj.Number.Value  = Function;

    /*
     *  Execute the method, no return value
     */
    Status = NsEvaluateByHandle (RegionObj->Region.REGMethod, Params, NULL);
    return_ACPI_STATUS (Status);
} 



ACPI_STATUS
EvAddressSpaceDispatch (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    UINT32                  Function,
    UINT32                  Address,
    UINT32                  BitWidth,
    UINT32                  *Value)
{
    ACPI_STATUS             Status;
    ADDRESS_SPACE_HANDLER   Handler;
    ADDRESS_SPACE_SETUP     RegionSetup;
    ACPI_OBJECT_INTERNAL    *HandlerDesc;
    void                    *RegionContext;


    FUNCTION_TRACE ("EvAddressSpaceDispatch");


    /*
     *  Check for an installed handler
     */
    HandlerDesc = RegionObj->Region.AddrHandler;

    if (!HandlerDesc)
    {
        DEBUG_PRINT (TRACE_OPREGION, ("Dispatch address access region 0x%X, no handler\n", RegionObj));
        return_ACPI_STATUS(AE_EXIST);
    }

    /*
     *  It may be the case that the region has never been initialized
     *  Some types of regions require special init code
     */
    if (!(RegionObj->Region.RegionFlags & REGION_INITIALIZED)) 
    {
        /*
         *  This region has not been initialized yet, do it
         */
        RegionSetup = HandlerDesc->AddrHandler.Setup;
        if (!RegionSetup) 
        {
            /*
             *  Bad news, no init routine and not init'd
             */
            DEBUG_PRINT (ACPI_ERROR, ("EvAddressSpaceDispatch: No init routine for region %X\n", 
                            RegionObj));
            return_ACPI_STATUS (AE_UNKNOWN_STATUS);
        }

        Status = RegionSetup (RegionObj, ACPI_REGION_ACTIVATE, HandlerDesc->AddrHandler.InstallTimeContext,
                                &RegionContext);
        /*
         *  Init routine may fail
         */
        if (ACPI_FAILURE (Status))
        {
            DEBUG_PRINT (ACPI_ERROR, ("EvAddressSpaceDispatch: %s from region init, SpaceID %d\n", 
                            CmFormatException (Status), RegionObj->Region.SpaceId));
            return_ACPI_STATUS(Status);
        }

        /*
         *  Save the returned context for use in all accesses to the region
         */
        HandlerDesc->AddrHandler.Context = RegionContext;
    }

    /*
     *  We have everything we need, begin the process
     */
    Handler = HandlerDesc->AddrHandler.Handler;

    DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
            ("Addrhandler 0x%X (0x%X), Address 0x%X\n", 
             RegionObj->Region.AddrHandler->AddrHandler, Handler, Address));

    if (!(HandlerDesc->AddrHandler.Flags & ADDR_HANDLER_DEFAULT_INSTALLED))
    {
        /*
         *  For handlers other than the default (supplied) handlers, we must
         *  exit the interpreter because the handler *might* block -- we don't
         *  know what it will do, so we can't hold the lock on the intepreter.
         */
        AmlExitInterpreter();
    }

    /* 
     *  Invoke the handler.
     */
    Status = Handler (Function, Address, BitWidth, Value, HandlerDesc->AddrHandler.Context);

    if (ACPI_FAILURE (Status))
    {
        DEBUG_PRINT (ACPI_ERROR, ("EvAddressSpaceDispatch: %s from handler, SpaceID %d\n", 
                        CmFormatException (Status), RegionObj->Region.SpaceId));
    }

    if (!(HandlerDesc->AddrHandler.Flags & ADDR_HANDLER_DEFAULT_INSTALLED))
    {
        /* We just returned from a non-default handler, we must re-enter the interpreter */

        AmlEnterInterpreter ();
    }

    return_ACPI_STATUS (Status);
}


void
EvDisassociateRegionFromHandler(
    ACPI_OBJECT_INTERNAL    *RegionObj)
{
    ACPI_OBJECT_INTERNAL    *HandlerObj;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    **LastObjPtr;
    ADDRESS_SPACE_SETUP     RegionSetup;
    void                    *RegionContext;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("EvDisassociateRegionAndHandler");


    /*
     *  Get the address handler from the region object
     */

    HandlerObj = RegionObj->Region.AddrHandler;
    if (!HandlerObj)
    {
        /*
         *  This guy has no handler, we're done
         */
        return_VOID;
    }

    /*
     *  Find this region in the handler's list
     */

    ObjDesc = HandlerObj->AddrHandler.RegionList;
    LastObjPtr = &HandlerObj->AddrHandler.RegionList;

    while (ObjDesc)
    {
        /*
         *  See if this is the one
         */
        if (ObjDesc == RegionObj)
        {
            DEBUG_PRINT (TRACE_OPREGION, ("Removing Region 0x%X from address handler 0x%X\n",
                            RegionObj, HandlerObj));
            /*
             *  This is it, remove it from the handler's list
             */
            *LastObjPtr = ObjDesc->Region.Link;

            /*
             *  Now stop region accesses by executing the _REG
             *  method
             */
            EvExecuteRegMethod (RegionObj, 0);

            /*
             *  Tell undo region prep work
             */
            RegionSetup = HandlerObj->AddrHandler.Setup;
            Status = RegionSetup (RegionObj, ACPI_REGION_DEACTIVATE, HandlerObj->AddrHandler.InstallTimeContext,
                                    &RegionContext);
            /*
             *  Init routine may fail
             */
            if (ACPI_FAILURE (Status))
            {
                DEBUG_PRINT (ACPI_ERROR, ("EvDisassociateRegionFromHandler: %s from region init, SpaceID %d\n", 
                                CmFormatException (Status), RegionObj->Region.SpaceId));
                /*
                 *  Just ignore failures for now
                 */
            }

            /*
             *  Remove handler reference in the region
             *
             *  NOTE: this doesn't mean that the region goes away
             *  The region is just inaccessible as indicated to
             *  the _REG method
             *
             *  If the region is on the handler's list
             *  this better be the region's handler
             */
            ACPI_ASSERT (RegionObj->Region.AddrHandler == HandlerObj);

            RegionObj->Region.AddrHandler = NULL;

            return_VOID;

        } /* found the right handler */

        /*
         *  Move through the linked list of handlers
         */
        LastObjPtr = &ObjDesc->Region.Link;
        ObjDesc = ObjDesc->Region.Link;
    }

    /*
     *  If we get here, the region was not not in the handler's region list
     */
    DEBUG_PRINT (TRACE_OPREGION, ("Cannot remove region 0x%X from address handler 0x%X\n",
                    RegionObj, HandlerObj));

    return_VOID;
}




ACPI_STATUS
EvAssociateRegionAndHandler(
    ACPI_OBJECT_INTERNAL    *HandlerObj,
    ACPI_OBJECT_INTERNAL    *RegionObj)
{
    ACPI_STATUS     Status;


    FUNCTION_TRACE ("EvAssociateRegionAndHandler");


    DEBUG_PRINT (TRACE_OPREGION, ("Adding Region 0x%X to address handler 0x%X\n", 
                    RegionObj, HandlerObj));

    ACPI_ASSERT (RegionObj->Region.SpaceId == HandlerObj->AddrHandler.SpaceId);
    ACPI_ASSERT (RegionObj->Region.AddrHandler == 0);

    /*
     *  Link this region to the front of the handler's list
     */

    RegionObj->Region.Link = HandlerObj->AddrHandler.RegionList;
    HandlerObj->AddrHandler.RegionList = RegionObj;

    /*
     *  set the region's handler
     */

/*
    HandlerObj->Common.ReferenceCount = (UINT16) (HandlerObj->Common.ReferenceCount + RegionObj->Common.ReferenceCount - 1);
*/
    RegionObj->Region.AddrHandler = HandlerObj;

    /*
     *  Last thing, tell ASL region is usable
     */
    CmReleaseMutex (MTX_NAMESPACE);
 Status = EvExecuteRegMethod (RegionObj, 1);
 CmAcquireMutex (MTX_NAMESPACE);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
EvAddrHandlerHelper (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_OBJECT_INTERNAL    *HandlerObj;
    ACPI_OBJECT_INTERNAL    *TmpObj;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *ObjEntry;
    ACPI_STATUS             Status;


    HandlerObj = (ACPI_OBJECT_INTERNAL *) Context;

    /* Parameter validation */

    if (!HandlerObj)
    {
        return (AE_OK);
    }

    /* Convert and validate the device handle */

    ObjEntry = NsConvertHandleToEntry (ObjHandle);
    if (!ObjEntry)
    {
        return (AE_BAD_PARAMETER);
    }

    /*
     *  We only care about regions.and objects
     *  that can have address handlers
     */

    if ((ObjEntry->Type != ACPI_TYPE_Device) &&
        (ObjEntry->Type != ACPI_TYPE_Region) &&
        (ObjEntry != Gbl_RootObject))
    {
        return (AE_OK);
    }

    /* Check for an existing internal object */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) ObjEntry);
    if (!ObjDesc)
    {
        /*
         *  The object DNE, we don't care about it
         */
        return (AE_OK);
    }

    /*
     *  Devices are handled different than regions
     */
    if (IS_THIS_OBJECT_TYPE (ObjDesc, ACPI_TYPE_Device))
    {
        /*
         *  See if this guy has any handlers
         */
        TmpObj = ObjDesc->Device.AddrHandler;
        while (TmpObj)
        {
            /*
             *  Now let's see if it's for the same address space.
             */
            if (TmpObj->AddrHandler.SpaceId == HandlerObj->AddrHandler.SpaceId)
            {
                /*
                 *  It's for the same address space
                 */

                DEBUG_PRINT (TRACE_OPREGION, ("Found handler for %s in device 0x%X (0x%X) handler 0x%X\n",
                                Gbl_RegionTypes[HandlerObj->AddrHandler.SpaceId], ObjDesc, TmpObj, HandlerObj));
                
                /*
                 *  Since the object we found it on was a device, then it
                 *  means that someone has already installed a handler for
                 *  the branch of the namespace from this device on.  Just
                 *  bail out telling the walk routine to not traverse this
                 *  branch.  This preserves the scoping rule for handlers.
                 */
                return (AE_CTRL_DEPTH);
            }

            /*
             *  Move through the linked list of handlers
             */
            TmpObj = TmpObj->AddrHandler.Link;
        }

        /*
         *  As long as the device didn't have a handler for this
         *  space we don't care about it.  We just ignore it and
         *  proceed.
         */
        return (AE_OK);
    }

    /*
     *  Only here if it was a region
     */
    ACPI_ASSERT (ObjDesc->Common.Type == ACPI_TYPE_Region);

    if (ObjDesc->Region.SpaceId != HandlerObj->AddrHandler.SpaceId)
    {
        /*
         *  This region is for a different address space
         *  ignore it
         */
        return (AE_OK);
    }

    /*
     *  Now we have a region and it is for the handler's address
     *  space type.
     *
     *  First disconnect region for any previous handler (if any)
     */
    EvDisassociateRegionFromHandler (ObjDesc);

    /*
     *  Then connect the region to the new handler
     */
    Status = EvAssociateRegionAndHandler (HandlerObj, ObjDesc);

    return (Status);
}


