/*
 *  ./acpica/subsystem/events/evrgnini.c - ACPI AddressSpace / OpRegion init
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __EVRGNINI_C__

#include <acpi.h>
#include <events.h>
#include <namesp.h>
#include <interp.h>
#include <amlcode.h>
#include <methods.h>

#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evrgnini");




ACPI_STATUS
EvSystemMemoryRegionSetup (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    UINT32                  Function,
    void                    *HandlerContext,
    void                    **ReturnContext)
{

    FUNCTION_TRACE ("EvSystemMemoryRegionSetup");


    if (Function == ACPI_REGION_DEACTIVATE) 
    {
        RegionObj->Region.RegionFlags &= ~(REGION_INITIALIZED);
        *ReturnContext = NULL;
    } 
    
    else 
    {
        RegionObj->Region.RegionFlags |= REGION_INITIALIZED;
        *ReturnContext = HandlerContext;
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
EvIoSpaceRegionSetup (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    UINT32                  Function,
    void                    *HandlerContext,
    void                    **ReturnContext)
{

    FUNCTION_TRACE ("EvIoSpaceRegionSetup");


    if (Function == ACPI_REGION_DEACTIVATE) 
    {
        RegionObj->Region.RegionFlags &= ~(REGION_INITIALIZED);
        *ReturnContext = NULL;
    } 
    else 
    {
        RegionObj->Region.RegionFlags |= REGION_INITIALIZED;
        *ReturnContext = HandlerContext;
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
EvPciConfigRegionSetup (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    UINT32                  Function,
    void                    *HandlerContext,
    void                    **ReturnContext)
{
    ACPI_STATUS             Status = AE_OK;
    NATIVE_UINT             Temp;
    PCI_HANDLER_CONTEXT    *PCIContext;
    ACPI_OBJECT_INTERNAL   *HandlerObj;
    NAME_TABLE_ENTRY       *SearchScope;


    FUNCTION_TRACE ("EvPciConfigRegionSetup");


    HandlerObj = RegionObj->Region.AddrHandler;

    if(!HandlerObj) 
    {
        /*
         *  No Handler ... bummer....
         *
         *  This cannot happen because the dispatch routine checks before
         *  we get here, but just in case.
         */
        DEBUG_PRINT (TRACE_OPREGION, ("Attempting to init a region 0x%X, with no handler\n", RegionObj));
        return_ACPI_STATUS(AE_EXIST);
    }

    if (Function == ACPI_REGION_DEACTIVATE) 
    {
        /*
         *  For right now we just free the allocated space
         */
        RegionObj->Region.RegionFlags &= ~(REGION_INITIALIZED);
//        CmFree(HandlerObj->AddrHandler.Context)
        *ReturnContext = NULL;
        return_ACPI_STATUS (Status);
    }

    PCIContext = CmAllocate (sizeof(PCI_HANDLER_CONTEXT));

    if (!PCIContext)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /*
     *  For PCI Config space access, we have to pass the segment, bus,
     *  device and function numbers.  This routine must acquire those.
     */

    /*
     *  First get device and function numbers from the _ADR object
     *  in the parent's scope.
     */
    ACPI_ASSERT(RegionObj->Region.Nte);

    SearchScope = RegionObj->Region.Nte->ParentEntry;


    CmReleaseMutex (MTX_NAMESPACE);

    Status = EvaluateNumeric (METHOD_NAME__ADR, SearchScope, &Temp);
    /*
     *  The default is zero, since the allocation above zeroed the data, just
     *  do nothing on failures.
     */
    if (ACPI_SUCCESS (Status))
    {
        /*
         *  Got it..
         */
        PCIContext->DevFunc = Temp;
    }

    /*
     *  Get the _SEG and _BBN values from the device the handler is installed on
     *
     *  We need to get the _SEG and _BBN objects relative to the PCI BUS device.
     *  This is the device the handler has been registered to handle.
     */

    SearchScope = HandlerObj->AddrHandler.Nte;

    Status = EvaluateNumeric (METHOD_NAME__SEG, SearchScope, &Temp);
    if (ACPI_SUCCESS (Status))
    {
        /*
         *  Got it..
         */
        PCIContext->Seg = Temp;
    }

    Status = EvaluateNumeric (METHOD_NAME__BBN, SearchScope, &Temp);
    if (ACPI_SUCCESS (Status))
    {
        /*
         *  Got it..
         */
        PCIContext->Bus = Temp;
    }

    CmAcquireMutex (MTX_NAMESPACE);

    *ReturnContext = PCIContext;

    RegionObj->Region.RegionFlags |= REGION_INITIALIZED;
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
EvDefaultRegionSetup (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    UINT32                  Function,
    void                    *HandlerContext,
    void                    **ReturnContext)
{

    FUNCTION_TRACE ("EvDefaultRegionSetup");


    if (Function == ACPI_REGION_DEACTIVATE) 
    {
        RegionObj->Region.RegionFlags &= ~(REGION_INITIALIZED);
        *ReturnContext = NULL;
    } 
    else 
    {
        RegionObj->Region.RegionFlags |= REGION_INITIALIZED;
        *ReturnContext = HandlerContext;
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
EvInitializeRegion (
    ACPI_OBJECT_INTERNAL    *RegionObj,
    BOOLEAN                 NsLocked)
{
    ACPI_OBJECT_INTERNAL   *HandlerObj;
    ACPI_OBJECT_INTERNAL   *ObjDesc;
    UINT32                  SpaceId; 
    NAME_TABLE_ENTRY       *Nte;        /* Namespace Object */
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY       *RegEntry;
    ACPI_NAME              *RegNamePtr = (ACPI_NAME *) METHOD_NAME__REG;


    FUNCTION_TRACE ("EvInitializeRegion");


    if (!RegionObj)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    ACPI_ASSERT(RegionObj->Region.Nte);

    Nte = RegionObj->Region.Nte->ParentEntry;
    SpaceId = RegionObj->Region.SpaceId;

    RegionObj->Region.AddrHandler = NULL;
    RegionObj->Region.REGMethod = NULL;
    RegionObj->Region.RegionFlags = INITIAL_REGION_FLAGS;

    /*
     *  Find any "_REG" associated with this region definition
     */
    Status = NsSearchOneScope (*RegNamePtr, Nte->Scope, ACPI_TYPE_Method, &RegEntry, NULL);
    if (Status == AE_OK)
    {
        /*
         *  The _REG method is optional and there can be only one per region
         *  definition.  This will be executed when the handler is attached
         *  or removed
         */
        RegionObj->Region.REGMethod = RegEntry;
    }

    /*
     *  The following loop depends upon the root nte having no parent
     *  ie: Gbl_RootObject->ParentEntry being set to NULL
     */
    while (Nte)
    {
        /*
         *  Check to see if a handler exists
         */
        HandlerObj = NULL;
        ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) Nte);
        if (ObjDesc) 
        {
            /*
             *  can only be a handler if the object exists
             */
            switch (Nte->Type)
            {
            case ACPI_TYPE_Device:

                HandlerObj = ObjDesc->Device.AddrHandler;
                break;

            case ACPI_TYPE_Processor:

                HandlerObj = ObjDesc->Processor.AddrHandler;
                break;

            case ACPI_TYPE_Thermal:

                HandlerObj = ObjDesc->ThermalZone.AddrHandler;
                break;
            }

            while (HandlerObj)
            {
                /*
                 *  This guy has at least one address handler
                 *  see if it has the type we want 
                 */
                if (HandlerObj->AddrHandler.SpaceId == SpaceId)
                {
                    DEBUG_PRINT (TRACE_OPREGION, ("Found handler (0x%X) for region 0x%X in obj 0x%X\n",
                                    HandlerObj, RegionObj, ObjDesc));
                    
                    /*
                     *  Found it! Now update the region and the handler
                     */
                    EvAssociateRegionAndHandler(HandlerObj, RegionObj);
                    return_ACPI_STATUS (AE_OK);
                }

                HandlerObj = HandlerObj->AddrHandler.Link;

            } /* while handlerobj */
        }

        /*
         *  This one does not have the handler we need
         *  Pop up one level
         */
        Nte = Nte->ParentEntry;

    } /* while Nte != ROOT */

    /*
     *  If we get here, there is no handler for this region
     */
    DEBUG_PRINT (TRACE_OPREGION, ("Unable to find handler for region 0x%X\n", RegionObj));

    return_ACPI_STATUS (AE_NOT_EXIST);
}

