
/*
 *  ./acpica/subsystem/events/evsci.c - System Control Interrupt configuration and
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <namesp.h>
#include <hardware.h>
#include <events.h>


#define _COMPONENT          EVENT_HANDLING
        MODULE_NAME         ("evsci");


/* 
 * Elements correspond to counts for
 * TMR, NOT_USED, GBL, PWR_BTN, SLP_BTN, RTC,
 * and GENERAL respectively.  These counts
 * are modified by the ACPI interrupt handler...
 * Note that GENERAL should probably be split out
 * into one element for each bit in the GPE
 * registers 
 */



UINT32 
EvSciHandler (void *Context)
{
    UINT32 InterruptHandled = INTERRUPT_NOT_HANDLED;

    FUNCTION_TRACE("EvSciHandler");

    /*
     * ACPI Enabled?
     * -------------
     * Make sure that ACPI is enabled by checking SCI_EN.  Note that we are
     * required to treat the SCI interrupt as sharable, level, active low.
     */
    if (!HwRegisterIO (ACPI_READ, MTX_DO_NOT_LOCK, (INT32)SCI_EN))
    {
        REPORT_ERROR ("Received and SCI but ACPI is not enabled.");
        return_VALUE (INTERRUPT_NOT_HANDLED);
    }

    /*
     * Fixed Events:
     * -------------
     * Check for and dispatch any Fixed Events that have occurred
     */
    InterruptHandled |= EvFixedEventDetect ();

    /*
     * GPEs:
     * -----
     * Check for and dispatch any GPEs that have occurred
     */
    InterruptHandled |= EvGpeDetect ();

    return_VALUE (InterruptHandled);
}



UINT32 
EvInstallSciHandler (void)
{
    UINT32 Except = AE_OK;
     
    FUNCTION_TRACE ("EvInstallSciHandler");
   
    Except = OsdInstallInterruptHandler (
        (UINT32) Gbl_FACP->SciInt,
        EvSciHandler,
        NULL);
    
    return_ACPI_STATUS (Except);
}



ACPI_STATUS
EvRemoveSciHandler (void)
{
    FUNCTION_TRACE ("EvRemoveSciHandler");
    
#if 0
        /* Disable all events first ???  TBD:  Figure this out!! */

        if (OriginalFixedEnableBitStatus ^ 1 << EventIndex (TMR_FIXED_EVENT))
        {
            AcpiEventDisableEvent (TMR_FIXED_EVENT);
        }
            
        if (OriginalFixedEnableBitStatus ^ 1 << EventIndex (GBL_FIXED_EVENT))
        {
            AcpiEventDisableEvent (GBL_FIXED_EVENT);
        }
            
        if (OriginalFixedEnableBitStatus ^ 1 << EventIndex (PWR_BTN_FIXED_EVENT))
        {
            AcpiEventDisableEvent (PWR_BTN_FIXED_EVENT);
        }
            
        if (OriginalFixedEnableBitStatus ^ 1 << EventIndex (SLP_BTN_FIXED_EVENT))
        {
            AcpiEventDisableEvent (SLP_BTN_FIXED_EVENT);
        }
            
        if (OriginalFixedEnableBitStatus ^ 1 << EventIndex (RTC_FIXED_EVENT))
        {
            AcpiEventDisableEvent (RTC_FIXED_EVENT);
        }
        
        OriginalFixedEnableBitStatus = 0;

#endif      
            
    OsdRemoveInterruptHandler (
        (UINT32) Gbl_FACP->SciInt,
        EvSciHandler);

    return_ACPI_STATUS (AE_OK);
}



#ifdef ACPI_DEBUG

INT32 
EvSciCount (
    UINT32                  Event)
{
    INT32                   Count;

    FUNCTION_TRACE ("EvSciCount");

    /* 
     * Elements correspond to counts for TMR, NOT_USED, GBL, 
     * PWR_BTN, SLP_BTN, RTC, and GENERAL respectively. 
     */

    if (Event >= NUM_FIXED_EVENTS)
    {
        Count = -1; 
    }
    else
    {
        Count = Gbl_EventCount[Event];
    }
    
    return_VALUE (Count);
}

#endif



void
EvRestoreAcpiState (void)
{
    INT32                   Index;


    FUNCTION_TRACE ("EvRestoreAcpiState");


    /* Restore the state of the chipset enable bits. */
    
    if (Gbl_RestoreAcpiChipset == TRUE)
    {
        /* Restore the fixed events */
        
        if (OsdIn16 (Gbl_FACP->Pm1aEvtBlk + 2) != Gbl_Pm1EnableRegisterSave)
        {
            OsdOut16 ((Gbl_FACP->Pm1aEvtBlk + 2), Gbl_Pm1EnableRegisterSave);
        }
        
        if (Gbl_FACP->Pm1bEvtBlk)
        {
            if (OsdIn16 (Gbl_FACP->Pm1bEvtBlk + 2) != Gbl_Pm1EnableRegisterSave)
            {
                OsdOut16 ((Gbl_FACP->Pm1bEvtBlk + 2), Gbl_Pm1EnableRegisterSave);
            }
        }


        /* Ensure that all status bits are clear */
        
        HwClearAcpiStatus ();


        /* Now restore the GPEs */
        
        for (Index = 0; Index < Gbl_FACP->Gpe0BlkLen / 2; Index++)
        {
            if (OsdIn8 ((Gbl_FACP->Gpe0Blk + Gbl_FACP->Gpe0BlkLen / 2)) != Gbl_Gpe0EnableRegisterSave[Index])
            {
                OsdOut8 ((Gbl_FACP->Gpe0Blk + Gbl_FACP->Gpe0BlkLen / 2), Gbl_Gpe0EnableRegisterSave[Index]);
            }
        }

        if (Gbl_FACP->Gpe1Blk && Gbl_FACP->Gpe1BlkLen)
        {
            for (Index = 0; Index < Gbl_FACP->Gpe1BlkLen / 2; Index++)
            {
                if (OsdIn8 ((Gbl_FACP->Gpe1Blk + Gbl_FACP->Gpe1BlkLen / 2)) != Gbl_Gpe1EnableRegisterSave[Index])
                {
                    OsdOut8 ((Gbl_FACP->Gpe1Blk + Gbl_FACP->Gpe1BlkLen / 2), Gbl_Gpe1EnableRegisterSave[Index]);
                }
            }
        }
        
        if (HwGetMode() != Gbl_OriginalMode)
        {
            HwSetMode (Gbl_OriginalMode);
        }
    }
    
    return_VOID;
}




void
EvTerminate (void)
{

    FUNCTION_TRACE ("EvTerminate");


    /* Free global tables, etc. */


    if (Gbl_GpeRegisters)
    {
        CmFree (Gbl_GpeRegisters);
    }

    if (Gbl_GpeInfo)
    {
        CmFree (Gbl_GpeInfo);
    }

    return_VOID;
}

 
