
/*
 *  ./acpica/subsystem/hardware/hw32cx.c - Handle IA32 Cx states
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <actypes.h>
#include <acpiosd.h>
#include <namesp.h>

#include "hardware.h"

#define _COMPONENT          HARDWARE
        MODULE_NAME         ("Hw32cx");


// TODO: Move following into <some> include file (actypes.h was checked out when this was built).

typedef ACPI_STATUS (*ACPI_C_STATE_HANDLER) (ACPI_IO_ADDRESS, UINT32*);



ACPI_C_STATE_HANDLER  HwIa32CxHandlers[MAX_CX_STATES] = {NULL, HwIa32EnterC1, NULL, NULL};
UINT32                      HwIa32ActiveCxState = 1;


ACPI_STATUS
HwIa32EnterC1(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks)
{
    UINT32                  Timer = 0;

 if (!PmTimerTicks)
 {
  /*
   * Enter C1:
   * ---------
   */
  enable();
  halt();
  *PmTimerTicks = ACPI_UINT32_MAX;
 }
 else
 {
  Timer = HwPmtTicks();

  /*
   * Enter C1:
   * ---------
   */
  enable();
  halt();

  /*
   * Compute Time in C1:
   * -------------------
   */
  Timer = HwPmtTicks() - Timer;

  *PmTimerTicks = Timer;
 }

    return AE_OK;
}



ACPI_STATUS
HwIa32EnterC2(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks)
{
    UINT32                  Timer = 0;

    if (!PBlkAddress || !PmTimerTicks) 
    {
        return AE_BAD_PARAMETER;
    }

 /*
  * Disable interrupts before all C2/C3 transitions.
  */
 disable();

 /* 
  * TBD: Check the BM_STS bit before and after entering the Cx state: If
  * Bus Mastering activity, demote to C2.
  */
    Timer = HwPmtTicks();

    /*
     * Enter C2:
     * ---------
     * Read from the P_LVL2 (P_BLK+4) register to invoke a C2 transition.
     */
    OsdIn8((ACPI_IO_ADDRESS)(PBlkAddress + 4));

    /*
     * Perform Dummy Op:
     * -----------------
     * We have to do something useless after reading LVL2 because chipsets
     * cannot guarantee that STPCLK# gets asserted in time to freeze execution.
     */
    OsdIn8((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk);

    /*
     * Compute Time in C2:
     * -------------------
     */
    Timer = HwPmtTicks() - Timer;

 *PmTimerTicks = Timer;

 /*
  * Re-enable interrupts after coming out of C2/C3.
  */
 enable();

    return AE_OK;
}



ACPI_STATUS
HwIa32EnterC3(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks)
{
    UINT32                  Timer = 0;
    UINT8                   Pm2CntBlk = 0;

    if (!PBlkAddress || !PmTimerTicks) 
    {
        return AE_BAD_PARAMETER;
    }

 /*
  * Disable interrupts before all C2/C3 transitions.
  */
 disable();

 /* 
  * TBD: Check the BM_STS bit before and after entering the Cx state: If
  * Bus Mastering activity, demote to C2.
  */
    Timer = HwPmtTicks();

    /*
     * Disable Bus Mastering:
     * ----------------------
     * Set the PM2_CNT.ARB_DIS bit (bit #0), preserving all other bits.
     */
    Pm2CntBlk = OsdIn8((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk);
    Pm2CntBlk |= 0x01;
    OsdOut8 ((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk, Pm2CntBlk);

    /*
     * Enter C3:
     * ---------
     * Read from the P_LVL3 (P_BLK+5) register to invoke a C3 transition.
     */
    OsdIn8((ACPI_IO_ADDRESS)(PBlkAddress + 5));

    /*
     * Perform Dummy Op:
     * -----------------
     * We have to do something useless after reading LVL3 because chipsets
     * cannot guarantee that STPCLK# gets asserted in time to freeze execution.
     */
    OsdIn8((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk);

    /*
     * Re-Enable Bus Mastering:
     * ------------------------
     * Clear the PM2_CNT.ARB_DIS bit (bit #0), preserving all other bits.
     */
    Pm2CntBlk = OsdIn8((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk);
    Pm2CntBlk &= 0xFE;
    OsdOut8 ((ACPI_IO_ADDRESS) Gbl_FACP->Pm2CntBlk, Pm2CntBlk);

    /*
     * Compute Time in C3:
     * -------------------
     */
 /* TBD: TODO: Support 24-bit timers (this algorithm assumes 32-bit) */
    Timer = HwPmtTicks() - Timer;

 *PmTimerTicks = Timer;

 /*
  * Re-enable interrupts after coming out of C2/C3.
  */
 enable();

    return AE_OK;
}


ACPI_STATUS
HwIa32EnterCx (
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks)
{
 if (!HwIa32CxHandlers[HwIa32ActiveCxState])
 {
  return AE_SUPPORT;
 }

 return (HwIa32CxHandlers[HwIa32ActiveCxState](PBlkAddress, PmTimerTicks));
}


ACPI_STATUS
HwIa32SetCx (
    UINT32                  CxState)
{
 /*
  * Supported State?
  * ----------------
  */
 if ((CxState < 1) || (CxState > 3))
 {
  return AE_BAD_PARAMETER;
 }

 if (!HwIa32CxHandlers[CxState])
 {
  return AE_SUPPORT;
 }

    /*
     * New Cx State?
     * -------------
     * We only care when moving from one state to another...
     */
    if (HwIa32ActiveCxState == CxState)
    {
        return AE_OK;
    }

    /*
     * Prepare to Use New State:
     * -------------------------
     */
    switch (CxState)
    {
    case 3:
        HwRegisterIO (ACPI_WRITE, MTX_LOCK, (INT32)BM_RLD, 1);
        break;
    }

    /*
     * Clean up from Old State:
     * ------------------------
     */
    switch (HwIa32ActiveCxState)
    {
    case 3:
        HwRegisterIO (ACPI_WRITE, MTX_LOCK, (INT32)BM_RLD, 0);
        break;
    }

 /*
  * Enable:
  * -------
  */
 HwIa32ActiveCxState = CxState;

    return AE_OK;
}



ACPI_STATUS
HwIa32GetCxInfo (
    UINT32    CxStates[MAX_CX_STATES])
{
    BOOLEAN SMP_system = FALSE;

    FUNCTION_TRACE("HwIa32GetCxInfo");

 if (!CxStates)
 {
  return_ACPI_STATUS(AE_BAD_PARAMETER);
 }

    /*
     *  BUGBUG: need to init SMP_system using info from the MAPIC table.
     */

 /*
  * Set Defaults:
  * -------------
  * C0 and C1 support is implied (but what about that PROC_C1 register 
  * in the FADT?!?!).  Set C2/C3 to max. latency (not supported until 
  * proven otherwise).
  */
 CxStates[0] = 0;
 CxStates[1] = 0;
 CxStates[2] = MAX_CX_STATE_LATENCY;
 CxStates[3] = MAX_CX_STATE_LATENCY;

    /*
     * C2 Supported?
     * -------------
     * We're only supporting C2 when the latency is <= 100 microseconds,
     * and on SMP systems when P_LVL2_UP (which indicates C2 only on UP)
     * is not set.
     */
    if (Gbl_FACP->PLvl2Lat <= 100) 
    {
        if (!SMP_system)
        {
            HwIa32CxHandlers[2] = HwIa32EnterC2;
   CxStates[2] = Gbl_FACP->PLvl2Lat;
        }
        else if (!Gbl_FACP->PLvl2Up) 
        {
            HwIa32CxHandlers[2] = HwIa32EnterC2;
   CxStates[2] = Gbl_FACP->PLvl2Lat;
        }
    }

    /*
     * C3 Supported?
     * -------------
     * We're only supporting C3 on UP systems when the latency is 
     * <= 1000 microseconds and that include the ability to disable 
     * Bus Mastering while in C3 (ARB_DIS) but allows Bus Mastering 
     * requests to wake the system from C3 (BM_RLD).  Note his method 
     * of maintaining cache coherency (disabling of bus mastering) 
     * cannot be used on SMP systems, and flushing caches (e.g. WBINVD)
     * is simply too costly (at this time).
     */
    if (Gbl_FACP->PLvl3Lat <= 1000)
    {
        if (!SMP_system && (Gbl_FACP->Pm2CntBlk && Gbl_FACP->Pm2CntLen))
        {
            HwIa32CxHandlers[3] = HwIa32EnterC3;
   CxStates[3] = Gbl_FACP->PLvl3Lat;
        }
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
HwIa32GetCxHandler(
    UINT32                  CxState,
    ACPI_C_STATE_HANDLER *Handler)
{
    FUNCTION_TRACE ("HwIa32GetCxHandler");

    if ((CxState == 0) || (CxState >= MAX_CX_STATES) || !Handler)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    *Handler = HwIa32CxHandlers[CxState];

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
HwIa32SetCxHandler(
    UINT32                  CxState,
    ACPI_C_STATE_HANDLER Handler)
{
    FUNCTION_TRACE ("HwIa32SetCxHandler");

    if ((CxState == 0) || (CxState >= MAX_CX_STATES) || !Handler)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

    HwIa32CxHandlers[CxState] = Handler;

    return_ACPI_STATUS(AE_OK);
}
