
/*
 *  ./acpica/subsystem/hardware/hw32thrt.c - handle IA32 processor throttling
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <actypes.h>
#include <acpiosd.h>
#include <namesp.h>

#define _COMPONENT          HARDWARE
        MODULE_NAME         ("Hw32thrt");

#define BIT_4           0x10



NATIVE_UINT
HwLocalPow(
    NATIVE_UINT             x, 
    NATIVE_UINT             y)
{
    NATIVE_UINT             i;
    NATIVE_UINT             Result = 1;

    for (i=0; i < y; i++)
    {
        Result = Result * x;
    }

    return Result;
}



void 
HwEnableThrottling(
    ACPI_IO_ADDRESS         PBlkAddress)
{
    UINT32                  PBlkValue;

    FUNCTION_TRACE ("EnableThrottling");

    PBlkValue = OsdIn32 (PBlkAddress);
    PBlkValue = PBlkValue | BIT_4;
    OsdOut32 (PBlkAddress, PBlkValue);

    return_VOID;
}



void 
HwDisableThrottling(
    ACPI_IO_ADDRESS         PBlkAddress)
{
    UINT32                  PBlkValue;

    FUNCTION_TRACE ("DisableThrottling");

    PBlkValue = OsdIn32 (PBlkAddress);
    PBlkValue = PBlkValue & (~(UINT32)BIT_4);
    OsdOut32 (PBlkAddress, PBlkValue);

    return_VOID;
}



UINT32
HwGetDutyCycle (
    UINT8                   DutyOffset,
    ACPI_IO_ADDRESS         PBlkAddress,
    UINT32                  NumThrottleStates)
{
    NATIVE_UINT             Index;
    NATIVE_UINT             Duty32Value;
    NATIVE_UINT             P_CNTMaskOffDutyField;

    FUNCTION_TRACE ("GetDutyCycle");

    /*
     *  Use NumThrottleStates - 1 as mask [ex. 8 - 1 = 7 (Fh)]
     *  and then shift it into the right position
     */
    P_CNTMaskOffDutyField = NumThrottleStates - 1;

    /*
     *  Read in the current value from the port
     */
    Duty32Value = OsdIn32 ((ACPI_IO_ADDRESS) PBlkAddress);
    
    /*
     *  Shift the the value to LSB
     */
    for (Index = 0; Index < (int)DutyOffset; Index++)
    {
        Duty32Value = Duty32Value >> 1;
    }

    /*
     *  Get the duty field only
     */
    Duty32Value = Duty32Value & P_CNTMaskOffDutyField;
    
    return_VALUE((UINT32) Duty32Value);
}



void
HwProgramDutyCycle (
    UINT8                   DutyOffset,
    UINT32                  DutyCycle,
    ACPI_IO_ADDRESS         PBlkAddress,
    UINT32                  NumThrottleStates)
{
    NATIVE_UINT             Index;
    NATIVE_UINT             Duty32Value;
    NATIVE_UINT             P_CNTMaskOffDutyField;
    NATIVE_UINT             PortValue;

    FUNCTION_TRACE ("HwProgramDutyCycle");

    /*
     *  valid DutyCycle passed
     */
    Duty32Value = DutyCycle;
    
    /*
     *  use NumThrottleStates - 1 as mask [ex. 8 - 1 = 7 (Fh)]
     *  and then shift it into the right position
     */
    P_CNTMaskOffDutyField = NumThrottleStates - 1;

    /*
     *  Shift the mask
     */
    for (Index = 0; Index < (int)DutyOffset; Index++)
    {
        P_CNTMaskOffDutyField = P_CNTMaskOffDutyField << 1;
        Duty32Value = Duty32Value << 1;
    }

    /*
     *  Read in the current value from the port
     */
    PortValue = OsdIn32 ((ACPI_IO_ADDRESS) PBlkAddress);
    
    /*
     *  Mask off the duty field so we don't OR in junk!
     */
    PortValue = PortValue & (~P_CNTMaskOffDutyField);
    
    /*
     *  OR in the bits we want to write out to the port
     */
    PortValue = (PortValue | Duty32Value) & (~(UINT32)BIT_4);
    
    /*
     *  write it to the port
     */
    OsdOut32 ((ACPI_IO_ADDRESS)PBlkAddress, PortValue);
    
    return_VOID;
}
