
/*
 *  ./acpica/subsystem/hardware/hwapi.c - ACPI Api's for hardware access
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <actypes.h>
#include <acpiosd.h>
#include <namesp.h>
#include <hardware.h>

#define _COMPONENT          HARDWARE
        MODULE_NAME         ("hwapi");







ACPI_STATUS
AcpiGetProcessorThrottlingInfo(
    ACPI_HANDLE             ProcessorHandle, 
    ACPI_BUFFER             *UserBuffer)
{
    NATIVE_UINT             PercentStep;
    NATIVE_UINT             NextPercent;
    NATIVE_UINT             NumThrottleStates;
    NATIVE_UINT             BufferSpaceNeeded;
    NATIVE_UINT             i;
    UINT8                   DutyOffset;
    UINT8                   DutyWidth;
    NAME_TABLE_ENTRY        *CpuNTE;
    ACPI_OBJECT_INTERNAL    *CpuObj;
    ACPI_CPU_THROTTLING_STATE *StatePtr;


    FUNCTION_TRACE ("AcpiGetProcessorThrottlingInfo");


    /*
     *  Have to at least have a buffer to return info in
     */
    if (!UserBuffer)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }
    
    /*
     *  Convert and validate the device handle
     */

    CpuNTE = NsConvertHandleToEntry (ProcessorHandle);
    if (!CpuNTE)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

   /*
    *   Check for an existing internal object
    */

    CpuObj = NsGetAttachedObject ((ACPI_HANDLE) CpuNTE);
    if (!CpuObj) 
    {
        return_ACPI_STATUS(AE_NOT_FOUND);
    }

    DutyOffset = Gbl_FACP->DutyOffset;
    DutyWidth = Gbl_FACP->DutyWidth;

    /*
     *  P0 must always have a P_BLK all others may be null
     *  in either case, we can't thottle a processor that has no P_BLK
     *
     *  Also if no Duty width, one state and it is 100%
     *
     */
    if (!CpuObj->Processor.PBLKLength || !DutyWidth || (0xFFFF < CpuObj->Processor.PBLKAddress)) 
    {
        /*
         *  Even though we can't throttle, we still have one state (100%)
         */
        NumThrottleStates = 1;
    } 
    else 
    {
        NumThrottleStates = (int) HwLocalPow (2,DutyWidth);
    }

    BufferSpaceNeeded = NumThrottleStates * sizeof(ACPI_CPU_THROTTLING_STATE);

    if ((UserBuffer->Length < BufferSpaceNeeded) || !UserBuffer->Pointer) 
    {
        UserBuffer->Length = BufferSpaceNeeded;
        return_ACPI_STATUS(AE_BUFFER_OVERFLOW);
    }

    UserBuffer->Length = BufferSpaceNeeded;
    StatePtr = (ACPI_CPU_THROTTLING_STATE *)UserBuffer->Pointer;
    PercentStep = 1000 / NumThrottleStates;

    /*
     * Build each entry in the buffer.  Note that we're using the value 
     * 1000 and dividing each state by 10 to better avoid round-off
     * accumulation.  Also note that the throttling STATES are ordered
     * sequentially from 100% (state 0) on down (e.g. 87.5% = state 1),
     * which is exactly opposite from duty cycle values (12.5% = state 1).
     */
    for (i=0, NextPercent=1000; i < NumThrottleStates; i++) 
    {
        StatePtr[i].StateNumber = i;
        StatePtr[i].PercentOfClock = NextPercent / 10;
        NextPercent -= PercentStep;
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
AcpiGetProcessorThrottlingState(ACPI_HANDLE ProcessorHandle, UINT32 *ThrottleState)
{
    NATIVE_UINT             NumThrottleStates;
    UINT8                   DutyOffset;
    UINT8                   DutyWidth;
    UINT32                  DutyCycle;
    NAME_TABLE_ENTRY       *CpuNTE;
    ACPI_OBJECT_INTERNAL   *CpuObj;


    FUNCTION_TRACE ("AcpiGetProcessorThrottlingState");


    /* Convert and validate the device handle */

    CpuNTE = NsConvertHandleToEntry (ProcessorHandle);
    if (!CpuNTE || !ThrottleState)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

   /* Check for an existing internal object */

    CpuObj = NsGetAttachedObject ((ACPI_HANDLE) CpuNTE);
    if (!CpuObj) 
    {
        return_ACPI_STATUS(AE_NOT_FOUND);
    }

    DutyOffset = Gbl_FACP->DutyOffset;
    DutyWidth = Gbl_FACP->DutyWidth;

    /*
     *  Must have a valid P_BLK P0 must have a P_BLK all others may be null
     *  in either case, we can't thottle a processor that has no P_BLK
     *  that means we are in the only supported state (0 - 100%)
     *
     *  also, if DutyWidth is zero there are no additional states
     */
    if (!CpuObj->Processor.PBLKLength || !DutyWidth || (0xFFFF < CpuObj->Processor.PBLKAddress)) 
    {
        *ThrottleState = 0;
        return_ACPI_STATUS(AE_OK);
    }

    NumThrottleStates = (int) HwLocalPow (2,DutyWidth);

    /*
     *  Get the current duty cycle value.
     */
    DutyCycle = HwGetDutyCycle (DutyOffset, CpuObj->Processor.PBLKAddress, NumThrottleStates);

    /* 
     * Convert duty cycle to throttling state (invert).
     */
    if (DutyCycle == 0)
    {
        *ThrottleState = 0;
    }
    else
    {
        *ThrottleState = NumThrottleStates - DutyCycle;
    }

    return_ACPI_STATUS(AE_OK);
}



ACPI_STATUS
AcpiSetProcessorThrottlingState(
    ACPI_HANDLE             ProcessorHandle, 
    UINT32                  ThrottleState)
{
    NATIVE_UINT             NumThrottleStates = 0;
    UINT8                   DutyOffset = 0;
    UINT8                   DutyWidth = 0;
    UINT32                  DutyCycle = 0;
    NAME_TABLE_ENTRY       *CpuNTE;
    ACPI_OBJECT_INTERNAL   *CpuObj;


    FUNCTION_TRACE ("AcpiSetProcessorThrottlingState");


    /* Convert and validate the device handle */

    CpuNTE = NsConvertHandleToEntry (ProcessorHandle);
    if (!CpuNTE)
    {
        return_ACPI_STATUS(AE_BAD_PARAMETER);
    }

   /* Check for an existing internal object */

    CpuObj = NsGetAttachedObject ((ACPI_HANDLE) CpuNTE);
    if (!CpuObj) 
    {
        return_ACPI_STATUS(AE_NOT_FOUND);
    }

    DutyOffset = Gbl_FACP->DutyOffset;
    DutyWidth = Gbl_FACP->DutyWidth;

    /*
     *  Must have a valid P_BLK P0 must have a P_BLK all others may be null
     *  in either case, we can't thottle a processor that has no P_BLK
     *  that means we are in the only supported state (0 - 100%)
     *
     *  also, if DutyWidth is zero there are no additional states
     */
    if (!CpuObj->Processor.PBLKLength || !DutyWidth || (0xFFFF < CpuObj->Processor.PBLKAddress)) 
    {
        /*
         *  If caller wants to set the state to the only state we handle
         *  we're done.
         */
        if (ThrottleState == 0) 
        {
            return_ACPI_STATUS(AE_OK);
        }

        /*
         *  Can't set this state
         */
        return_ACPI_STATUS(AE_SUPPORT);
    }

    NumThrottleStates = (int) HwLocalPow (2,DutyWidth);

    /* 
     * Convert throttling state to duty cycle (invert).
     */
    if (ThrottleState > 0)
    {
        DutyCycle = NumThrottleStates - ThrottleState;
    }

    /*
     *  Turn off throttling (don't muck with the h/w while throttling).
     */
    HwDisableThrottling(CpuObj->Processor.PBLKAddress);

    /*
     *  Program the throttling state.
     */
    HwProgramDutyCycle (DutyOffset, DutyCycle, CpuObj->Processor.PBLKAddress, NumThrottleStates);

    /*
     *  Only enable throttling for non-zero states (0 - 100%)
     */
    if (ThrottleState) 
    {
        HwEnableThrottling(CpuObj->Processor.PBLKAddress);
    }

    return_ACPI_STATUS(AE_OK);
}








ACPI_STATUS
AcpiGetProcessorCxInfo (
 ACPI_HANDLE             ProcessorHandle,
    ACPI_BUFFER             *UserBuffer)
{
 ACPI_STATUS    Status = AE_OK;
 UINT32     CxStateLatencies[4] = {0, 0, 0, 0};
    NATIVE_UINT             BufferSpaceNeeded = 0;
    ACPI_CX_STATE           *StatePtr = NULL;
    NATIVE_UINT             i = 0;


    FUNCTION_TRACE ("AcpiGetProcessorCxInfo");


    /*
     *  Have to at least have a buffer to return info in
     */
    if (!UserBuffer)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    Status = HwIa32GetCxInfo (CxStateLatencies);
 if (ACPI_FAILURE (Status))
 {
  return_ACPI_STATUS (Status);
 }

    BufferSpaceNeeded = 4 * sizeof (ACPI_CX_STATE);

    if ((UserBuffer->Length < BufferSpaceNeeded) || !UserBuffer->Pointer) 
    {
        UserBuffer->Length = BufferSpaceNeeded;
        return_ACPI_STATUS (AE_BUFFER_OVERFLOW);
    }

    UserBuffer->Length = BufferSpaceNeeded;

    StatePtr = (ACPI_CX_STATE *) UserBuffer->Pointer;

    for (i=0; i < 4; i++) 
    {
        StatePtr[i].StateNumber = i;
        StatePtr[i].Latency = CxStateLatencies[i];
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiSetProcessorSleepState (
    ACPI_HANDLE             ProcessorHandle,
    UINT32                  CxState)
{
    ACPI_STATUS             Status;


 FUNCTION_TRACE ("AcpiSetProcessorSleepState");


    Status = HwIa32SetCx (CxState);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiProcessorSleep (
    ACPI_HANDLE             ProcessorHandle,
 UINT32     *PmTimerTicks)
{
    NAME_TABLE_ENTRY  *CpuNTE = NULL;
    ACPI_OBJECT_INTERNAL *CpuObj = NULL;
    ACPI_IO_ADDRESS   PBlkAddress = 0;


 /*
  * Convert ProcessorHandle to PBlkAddres...
  */

    /* Convert and validate the device handle */

    CpuNTE = NsConvertHandleToEntry (ProcessorHandle);
    if (!CpuNTE)
    {
        return AE_BAD_PARAMETER;
    }

   /* Check for an existing internal object */

    CpuObj = NsGetAttachedObject ((ACPI_HANDLE) CpuNTE);
    if (!CpuObj) 
    {
        return AE_NOT_FOUND;
    }

    /* Get the processor register block (P_BLK) address */

    PBlkAddress = CpuObj->Processor.PBLKAddress;
    if (!CpuObj->Processor.PBLKLength) 
    {
        /* Ensure a NULL addresss (note that P_BLK isn't required for C1) */

        PBlkAddress = 0;
    }

 /*
  * Enter the currently active Cx sleep state.
  */
    return HwIa32EnterCx (PBlkAddress, PmTimerTicks);
}



ACPI_STATUS
AcpiGetTimer (
    UINT32                  *OutTicks)
{
    FUNCTION_TRACE ("AcpiGetTimer");


    if (!OutTicks)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    *OutTicks = HwPmtTicks ();

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiSetFirmwareWakingVector (
    void                    *PhysicalAddress)
{
    FUNCTION_TRACE ("AcpiSetFirmwareWakingVector");
    

    /* Make sure that we have an FACS */

    if (!Gbl_FACS)
    {
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    /* Set the vector */

    * ((void **) Gbl_FACS->FirmwareWakingVector) = PhysicalAddress;

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiGetFirmwareWakingVector (
    void                    **PhysicalAddress)
{
    FUNCTION_TRACE ("AcpiGetFirmwareWakingVector");


    if (!PhysicalAddress)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Make sure that we have an FACS */

    if (!Gbl_FACS)
    {
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    /* Get the vector */

    *PhysicalAddress = * ((void **) Gbl_FACS->FirmwareWakingVector);

    
    return_ACPI_STATUS (AE_OK);
}


