/*
 *  ./acpica/subsystem/hardware/hwgpe.c - Low level GPE enable/disable/clear functions
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <acpi.h>
#include <hardware.h>
#include <namesp.h>
#include <events.h>

#define _COMPONENT          HARDWARE
        MODULE_NAME         ("hwgpe");


UINT8 DecodeTo8bit [8] = {1,2,4,8,16,32,64,128};



void
HwEnableGpe (
    UINT32                  GpeNumber)
{
    UINT8                   InByte;
    UINT32                  RegisterIndex;
    UINT8                   BitMask;

    /* 
     * Translate GPE number to index into global registers array.
     */
    RegisterIndex = Gbl_GpeValid[GpeNumber];

    /* 
     * Figure out the bit offset for this GPE within the target register.
     */
    BitMask = DecodeTo8bit [MOD_8 (GpeNumber)];

    /* 
     * Read the current value of the register, set the appropriate bit
     * to enable the GPE, and write out the new register.
     */
    InByte = OsdIn8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr);
    OsdOut8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr, (UINT8)(InByte | BitMask));
}



void
HwDisableGpe (
    UINT32                  GpeNumber)
{
    UINT8                   InByte;
    UINT32                  RegisterIndex;
    UINT8                   BitMask;

    /* 
     * Translate GPE number to index into global registers array.
     */
    RegisterIndex = Gbl_GpeValid[GpeNumber];

    /* 
     * Figure out the bit offset for this GPE within the target register.
     */
    BitMask = DecodeTo8bit [MOD_8 (GpeNumber)];

    /* 
     * Read the current value of the register, clear the appropriate bit,
     * and write out the new register value to disable the GPE.
     */
    InByte = OsdIn8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr);
    OsdOut8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr, (UINT8)(InByte & ~BitMask));
}



void
HwClearGpe (
    UINT32                  GpeNumber)
{
    UINT32                  RegisterIndex;
    UINT8                   BitMask;

    /* 
     * Translate GPE number to index into global registers array.
     */
    RegisterIndex = Gbl_GpeValid[GpeNumber];

    /* 
     * Figure out the bit offset for this GPE within the target register.
     */
    BitMask = DecodeTo8bit [MOD_8 (GpeNumber)];

    /* 
     * Write a one to the appropriate bit in the status register to 
     * clear this GPE.
     */
    OsdOut8 (Gbl_GpeRegisters[RegisterIndex].StatusAddr, BitMask);
}



void
HwGetGpeStatus (
    UINT32                  GpeNumber,
    ACPI_EVENT_STATUS       *EventStatus)
{
    UINT8                   InByte = 0;
    UINT32                  RegisterIndex = 0;
    UINT8                   BitMask = 0;

    if (!EventStatus)
    {
        return;   
    }

    (*EventStatus) = 0;

    /* 
     * Translate GPE number to index into global registers array.
     */
    RegisterIndex = Gbl_GpeValid[GpeNumber];

    /* 
     * Figure out the bit offset for this GPE within the target register.
     */
    BitMask = DecodeTo8bit [MOD_8 (GpeNumber)];

    /* 
     * Enabled?:
     */
    InByte = OsdIn8 (Gbl_GpeRegisters[RegisterIndex].EnableAddr);

    if (BitMask & InByte)
    {
        (*EventStatus) |= EVENT_FLAG_ENABLED;
    }

    /* 
     * Set?
     */
    InByte = OsdIn8 (Gbl_GpeRegisters[RegisterIndex].StatusAddr);

    if (BitMask & InByte)
    {
        (*EventStatus) |= EVENT_FLAG_SET;
    }
}
