
/*
 *  ./acpica/subsystem/hardware/hwregs.c - Read/write access functions for the various ACPI
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __HWREGS_C__

#include <acpi.h>
#include <hardware.h>
#include <namesp.h>

#define _COMPONENT          HARDWARE
        MODULE_NAME         ("hwregs");



/* This matches the #defines in actypes.h. */

ACPI_STRING SleepStateTable[] = {"\\_S0_","\\_S1_","\\_S2_","\\_S3_",
                                 "\\_S4_","\\_S4B","\\_S5_"};




INT32
HwGetBitShift (
    UINT32                  Mask)
{
    INT32                   Shift;


    FUNCTION_TRACE ("HwGetBitShift");


    for (Shift = 0; ((Mask >> Shift) & 1) == 0; Shift++)
    { ; }

    return_VALUE (Shift);
}



void 
HwClearAcpiStatus (void)
{
    UINT16                  GpeLength;
    UINT16                  Index;


    FUNCTION_TRACE ("HwClearAcpiStatus");


    DEBUG_PRINT (TRACE_IO, ("About to write %04X to %04X\n", 
                    ALL_FIXED_STS_BITS, (UINT16) Gbl_FACP->Pm1aEvtBlk));


    CmAcquireMutex (MTX_HARDWARE);

    OsdOut16 (Gbl_FACP->Pm1aEvtBlk, (UINT16) ALL_FIXED_STS_BITS);
    
    if (Gbl_FACP->Pm1bEvtBlk)
    {
        OsdOut16 ((UINT16) Gbl_FACP->Pm1bEvtBlk, (UINT16) ALL_FIXED_STS_BITS);
    }

    /* now clear the GPE Bits */
    
    if (Gbl_FACP->Gpe0BlkLen)
    {
        GpeLength = (UINT16) (Gbl_FACP->Gpe0BlkLen / 2);

        for (Index = 0; Index < GpeLength; Index++)
        {
            OsdOut8 ((Gbl_FACP->Gpe0Blk + Index), (UINT8) 0xff);
        }
    }

    if (Gbl_FACP->Gpe1BlkLen)
    {
        GpeLength = (UINT16) (Gbl_FACP->Gpe1BlkLen / 2);

        for (Index = 0; Index < GpeLength; Index++)
        {
            OsdOut8 ((Gbl_FACP->Gpe1Blk + Index), (UINT8) 0xff);
        }
    }

    CmReleaseMutex (MTX_HARDWARE);
    return_VOID;
}



ACPI_STATUS
HwObtainSleepTypeRegisterData (
    UINT8                   SleepState, 
    UINT8                   *Slp_TypA,
    UINT8                   *Slp_TypB)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_INTERNAL    *ObjDesc;

    
    FUNCTION_TRACE ("HwObtainSleepTypeRegisterData");


    /*
     *  Validate parameters
     */

    if ((SleepState > ACPI_S_STATES_MAX) ||
        !Slp_TypA || !Slp_TypB)
    {   
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     *  Evaluate the namespace object containing the values for this state
     */
  
    Status = NsEvaluateByName (SleepStateTable[SleepState], NULL, &ObjDesc);
    if (AE_OK == Status)
    {  
        if (ObjDesc)
        {
            /*
             *  We got something, now ensure it is correct.  The object must
             *  be a package and must have at least 2 numeric values as the
             *  two elements
             */

            if ((ObjDesc->Common.Type != ACPI_TYPE_Package) ||
                ((ObjDesc->Package.Elements[0])->Common.Type != ACPI_TYPE_Number) ||
                ((ObjDesc->Package.Elements[1])->Common.Type != ACPI_TYPE_Number))
            {   
                /* Invalid _Sx_ package type or value  */
            
                REPORT_ERROR ("Object type returned from interpreter differs from expected value");
                Status = AE_ERROR;
            }
            else
            {   
                /*
                 *  Valid _Sx_ package size, type, and value
                 */
                *Slp_TypA = (UINT8) (ObjDesc->Package.Elements[0])->Number.Value;
                *Slp_TypB = (UINT8) (ObjDesc->Package.Elements[1])->Number.Value;
            }

            CmRemoveReference (ObjDesc);
        }
    }

    return_ACPI_STATUS (Status);
}



UINT32
HwRegisterIO (
    INT32                   ReadWrite,
 UINT32     UseLock,
    INT32                   RegisterId, 
    ...)
{
    UINT32                  RegisterValue = 0;
    UINT32                  Mask = 0;
    UINT32                  Value = 0;
    UINT32                  GpeReg = 0;


    FUNCTION_TRACE ("HwRegisterIO");

    
    if (ReadWrite == ACPI_WRITE)
    {
        va_list         marker;

        va_start (marker, RegisterId);
        Value = va_arg (marker, INT32);
        va_end (marker);
    }

    
    /*
     * Decode the Register ID
     */

    switch (RegisterId & REGISTER_BLOCK_MASK)
    {
    case PM1_EVT:

        if (RegisterId < (INT32) TMR_EN)
        {   
            if(MTX_LOCK == UseLock)
   {
    CmAcquireMutex (MTX_HARDWARE);
   }

            /* status register */
            
            RegisterValue = (UINT32) OsdIn16 (Gbl_FACP->Pm1aEvtBlk);
            DEBUG_PRINT (TRACE_IO, ("PM1a status: Read 0x%X from 0x%X\n", 
                            RegisterValue, Gbl_FACP->Pm1aEvtBlk));
            
            if (Gbl_FACP->Pm1bEvtBlk)
            {
                RegisterValue |= (UINT32) OsdIn16 (Gbl_FACP->Pm1bEvtBlk);
                DEBUG_PRINT (TRACE_IO, ("PM1b status: Read 0x%X from 0x%X\n", 
                                RegisterValue, Gbl_FACP->Pm1bEvtBlk));
            }

            switch (RegisterId)
            {
            case TMR_STS:
                Mask = TMR_STS_MASK;
                break;
            
            case BM_STS:
                Mask = BM_STS_MASK;
                break;
            
            case GBL_STS:
                Mask = GBL_STS_MASK;
                break;
            
            case PWRBTN_STS:
                Mask = PWRBTN_STS_MASK;
                break;
            
            case SLPBTN_STS:
                Mask = SLPBTN_STS_MASK;
                break;
            
            case RTC_STS:
                Mask = RTC_STS_MASK;
                break;
            
            case WAK_STS:
                Mask = WAK_STS_MASK;
                break;
            
            default:
                Mask = 0;
                break;
            }
            
            if (ReadWrite == ACPI_WRITE)
            {
                /* 
                 * Status registers are different from the rest.  Clear by writing 1, writing 0
                 * has no effect.  So, the only relevent information is the single bit we're
                 * interested in, all others should be written as 0 so they will be left
                 * unchanged 
                 */
                
                Value <<= HwGetBitShift (Mask);
                Value &= Mask;
                
                if (Value)
                {
                    DEBUG_PRINT (TRACE_IO, ("About to write %04X to %04X\n", Value, Gbl_FACP->Pm1aEvtBlk));

                    OsdOut16 (Gbl_FACP->Pm1aEvtBlk, (UINT16) Value);
                    
                    if (Gbl_FACP->Pm1bEvtBlk)
                    {
                        OsdOut16 (Gbl_FACP->Pm1bEvtBlk, (UINT16) Value);
                    }
                    
                    RegisterValue = 0;
                }
            }

            if(MTX_LOCK == UseLock)
   {
    CmReleaseMutex (MTX_HARDWARE);
   }
        }
        
        else
        {   
            /* enable register */
            
            if(MTX_LOCK == UseLock)
   {
             CmAcquireMutex (MTX_HARDWARE);
   }

            RegisterValue = (UINT32) OsdIn16 ((Gbl_FACP->Pm1aEvtBlk + Gbl_FACP->Pm1EvtLen / 2));

            DEBUG_PRINT (TRACE_IO, ("PM1a enable: Read 0x%X from 0x%X\n", 
                            RegisterValue, (Gbl_FACP->Pm1aEvtBlk + Gbl_FACP->Pm1EvtLen / 2)));
            
            if (Gbl_FACP->Pm1bEvtBlk)
            {
                RegisterValue |= (UINT32) OsdIn16 ((Gbl_FACP->Pm1bEvtBlk + Gbl_FACP->Pm1EvtLen / 2));

                DEBUG_PRINT (TRACE_IO, ("PM1b enable: Read 0x%X from 0x%X\n", 
                                RegisterValue, (Gbl_FACP->Pm1bEvtBlk + Gbl_FACP->Pm1EvtLen / 2)));
            }

            switch (RegisterId)
            {
            case TMR_EN:
                Mask = TMR_EN_MASK;
                break;
            
            case GBL_EN:
                Mask = GBL_EN_MASK;
                break;
            
            case PWRBTN_EN:
                Mask = PWRBTN_EN_MASK;
                break;
            
            case SLPBTN_EN:
                Mask = SLPBTN_EN_MASK;
                break;
            
            case RTC_EN:
                Mask = RTC_EN_MASK;
                break;
            
            default:
                Mask = 0;
                break;
            }
            
            if (ReadWrite == ACPI_WRITE)
            {
                RegisterValue  &= ~Mask;
                Value          <<= HwGetBitShift (Mask);
                Value          &= Mask;
                RegisterValue  |= Value;

                DEBUG_PRINT (TRACE_IO, ("About to write %04X to %04X\n", RegisterValue, 
                            (Gbl_FACP->Pm1aEvtBlk + Gbl_FACP->Pm1EvtLen / 2)));

                OsdOut16 ((Gbl_FACP->Pm1aEvtBlk + Gbl_FACP->Pm1EvtLen / 2), 
                            (UINT16) RegisterValue);
                
                if (Gbl_FACP->Pm1bEvtBlk)
                {
                    OsdOut16 ((Gbl_FACP->Pm1bEvtBlk + Gbl_FACP->Pm1EvtLen / 2), 
                                (UINT16) RegisterValue);
                }
            }
            if(MTX_LOCK == UseLock)
   {
             CmReleaseMutex (MTX_HARDWARE);
   }
        }
        break;
    

    case PM1_CONTROL:

        RegisterValue = 0;
        
        if(MTX_LOCK == UseLock)
  {
   CmAcquireMutex (MTX_HARDWARE);
  }

        if (RegisterId != (INT32) SLP_TYPb)   
        {
            /* 
             * SLP_TYPx registers are written differently
             * than any other control registers with
             * respect to A and B registers.  The value
             * for A may be different than the value for B 
             */

            RegisterValue = (UINT32) OsdIn16 (Gbl_FACP->Pm1aCntBlk);
            DEBUG_PRINT (TRACE_IO, ("PM1a control: Read 0x%X from 0x%X\n", 
                            RegisterValue, Gbl_FACP->Pm1aCntBlk));
        }

        if (Gbl_FACP->Pm1bCntBlk && RegisterId != (INT32) SLP_TYPa)
        {
            RegisterValue |= (UINT32) OsdIn16 (Gbl_FACP->Pm1bCntBlk);
            DEBUG_PRINT (TRACE_IO, ("PM1b control: Read 0x%X from 0x%X\n", 
                            RegisterValue, Gbl_FACP->Pm1bCntBlk));
        }

        switch (RegisterId)
        {
        case SCI_EN:
            Mask = SCI_EN_MASK;
            break;
        
        case BM_RLD:
            Mask = BM_RLD_MASK;
            break;
        
        case GBL_RLS:
            Mask = GBL_RLS_MASK;
            break;
        
        case SLP_TYPa:
        case SLP_TYPb:
            Mask = SLP_TYPx_MASK;
            break;
        
        case SLP_EN:
            Mask = SLP_EN_MASK;
            break;

        default:
            Mask = 0;
            break;
        }
        
        if (ReadWrite == ACPI_WRITE)
        {
            RegisterValue  &= ~Mask;
            Value          <<= HwGetBitShift (Mask);
            Value          &= Mask;
            RegisterValue  |= Value;
            
            /* 
             * SLP_TYPx registers are written differently
             * than any other control registers with
             * respect to A and B registers.  The value
             * for A may be different than the value for B
             */
            
            if (RegisterId != (INT32) SLP_TYPb)
            {
                if (Mask == SLP_EN_MASK)
                {
                    disable();  /* disable interrupts */
                }

                OsdOut16 (Gbl_FACP->Pm1aCntBlk, (UINT16) RegisterValue);
                
                if (Mask == SLP_EN_MASK)
                {
                    /* 
                     * Enable interrupts, the SCI handler is likely going to be invoked as
                     * soon as interrupts are enabled, since gpe's and most fixed resume
                     * events also generate SCI's. 
                     */
                    enable();
                }
            }
                
            if (Gbl_FACP->Pm1bCntBlk && RegisterId != (INT32) SLP_TYPa)
            {
                OsdOut16 (Gbl_FACP->Pm1bCntBlk, (UINT16) RegisterValue);
            }
        }

        if(MTX_LOCK == UseLock)
  {
         CmReleaseMutex (MTX_HARDWARE);
  }
        break;
    

    case PM2_CONTROL:

        if(MTX_LOCK == UseLock)
  {
      CmAcquireMutex (MTX_HARDWARE);
  }
        RegisterValue = (UINT32) OsdIn16 (Gbl_FACP->Pm2CntBlk);
        DEBUG_PRINT (TRACE_IO, ("PM2 control: Read 0x%X from 0x%X\n", 
                        RegisterValue, Gbl_FACP->Pm2CntBlk));
        
        switch (RegisterId)
        {
        case ARB_DIS:
            Mask = ARB_DIS_MASK;
            break;
        
        default:
            Mask = 0;
            break;
        }
        
        if (ReadWrite == ACPI_WRITE)
        {
            RegisterValue  &= ~Mask;
            Value          <<= HwGetBitShift (Mask);
            Value          &= Mask;
            RegisterValue  |= Value;

            DEBUG_PRINT (TRACE_IO, ("About to write %04X to %04X\n", RegisterValue, Gbl_FACP->Pm2CntBlk));

            OsdOut16 (Gbl_FACP->Pm2CntBlk, (UINT16) RegisterValue);
        }

        if(MTX_LOCK == UseLock)
  {
         CmReleaseMutex (MTX_HARDWARE);
  }
        break;
    

    case PM_TIMER:

        RegisterValue = OsdIn32 (Gbl_FACP->PmTmrBlk);
        DEBUG_PRINT (TRACE_IO, ("PM_TIMER: Read 0x%X from 0x%X\n", 
                        RegisterValue, Gbl_FACP->PmTmrBlk));

        Mask = 0xFFFFFFFF;
        break;
    

    case GPE1_EN_BLOCK:

        GpeReg = (Gbl_FACP->Gpe1Blk + (UINT32) Gbl_FACP->Gpe1Base) + (GpeReg + 
                    ((UINT32) ((Gbl_FACP->Gpe1BlkLen) / 2)));
    

    case GPE1_STS_BLOCK:

        if (!GpeReg)
        {
            GpeReg = (Gbl_FACP->Gpe1Blk + (UINT32) Gbl_FACP->Gpe1Base);
        }


    case GPE0_EN_BLOCK:

        if (!GpeReg)
        {
            GpeReg = Gbl_FACP->Gpe0Blk + ((UINT32) ((Gbl_FACP->Gpe0BlkLen) / 2));
        }
    

    case GPE0_STS_BLOCK:

        if (!GpeReg)
        {
            GpeReg = Gbl_FACP->Gpe0Blk;
        }

        /* Determine the bit to be accessed */
        
        Mask = (((UINT32) RegisterId) & BIT_IN_REGISTER_MASK);
        Mask = 1 << (Mask-1);
        
        /* The base address of the GPE 0 Register Block */
        /* Plus 1/2 the length of the GPE 0 Register Block */
        /* The enable register is the register following the Status Register */
        /* and each register is defined as 1/2 of the total Register Block */
        
        /* This sets the bit within EnableBit that needs to be written to */
        /* the register indicated in Mask to a 1, all others are 0 */
        
        if (Mask > LOW_BYTE)
        {
            /* Shift the value 1 byte to the right and add 1 to the register */
            
            Mask >>= ONE_BYTE;
            GpeReg++;
        }
        
        /* Now get the current Enable Bits in the selected Reg */
        
        if(MTX_LOCK == UseLock)
  {
         CmAcquireMutex (MTX_HARDWARE);
  }

        RegisterValue = (UINT32) OsdIn8 (GpeReg);
        DEBUG_PRINT (TRACE_IO, ("GPE Enable bits: Read 0x%X from 0x%X\n", RegisterValue, GpeReg));
        
        if (ReadWrite == ACPI_WRITE)
        {               
            RegisterValue  &= ~Mask;
            Value          <<= HwGetBitShift (Mask);
            Value          &= Mask;
            RegisterValue  |= Value;
           
            /* This write will put the Action state into the General Purpose */

            /* Enable Register indexed by the value in Mask */

            DEBUG_PRINT (TRACE_IO, ("About to write %04X to %04X\n", RegisterValue, GpeReg));

            OsdOut8 (GpeReg, (UINT8) RegisterValue);
            RegisterValue = (UINT32) OsdIn8 (GpeReg);          
        }
        
        if(MTX_LOCK == UseLock)
  {
   CmReleaseMutex (MTX_HARDWARE);
  }
        break;
    

    case PROCESSOR_BLOCK:
    default:

        Mask = 0;
        break;
    }


    RegisterValue &= Mask;
    RegisterValue >>= HwGetBitShift (Mask);

    DEBUG_PRINT (TRACE_IO, ("Register I/O: returning 0x%X\n", RegisterValue));
    return_VALUE (RegisterValue);
}
