
/*
 *  ./acpica/subsystem/include/acobject.h - Definition of ACPI_OBJECT_INTERNAL (Internal object only)
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _ACOBJECT_H
#define _ACOBJECT_H

#include <actypes.h>
#include <macros.h>
#include <internal.h>

/* 
 * TBD: This wastes memory.
 * TBD: Allow variable sized acpi objects
 * TBD: Must find and fix all code that assumes constant sized object
 *
 * All variants of the ACPI_OBJECT_INTERNAL are defined with the same
 * sequence of field types, with fields that are not used in a particular
 * variant being named "Reserved".  This is not strictly necessary, but
 * may in some circumstances simplify understanding if these structures
 * need to be displayed in a debugger having limited (or no) support for
 * union types.  It also simplifies some debug code in NsDumpTable() which
 * dumps multi-level values: fetching Buffer.Pointer suffices to pick up
 * the value or next level for any of several types.
 */


/*
 * Common area for all objects.
 *
 * DataType is used to differentiate between internal descriptors, and MUST
 * be the first byte in this structure.
 */


#define ACPI_OBJECT_COMMON                  /* Two 32-bit fields */\
    UINT8                   DataType;           /* To differentiate various internal objs */\
    UINT8                   Type;               /* ACPI_OBJECT_TYPE */\
    UINT8                   Size;               /* Size of entire descriptor */\
    UINT8                   Flags;\
    UINT16                  ReferenceCount;     /* For object deletion management */\
    UINT16                  CmFill2;\
    union AcpiObjInternal  *Next; \

/* Defines for flag byte above */

#define AO_STATIC_ALLOCATION        0x1



/* 
 * Common bitfield for the field objects
 */
#define ACPI_COMMON_FIELD_INFO              /* Three 32-bit values */\
    UINT32                  Offset;             /* Byte offset within containing object */\
    UINT16                  Length;             /* # of bits in buffer */ \
    UINT8                   Granularity;\
    UINT8                   BitOffset;          /* Bit offset within min read/write data unit */\
    UINT8                   Access;             /* AccessType */\
    UINT8                   LockRule;\
    UINT8                   UpdateRule;\
    UINT8                   AccessAttribute;





typedef struct /* COMMON */
{
    ACPI_OBJECT_COMMON
    UCHAR   FirstNonCommonByte;
        
} ACPI_OBJECT_Common;


typedef struct /* NUMBER - has value */
{
    ACPI_OBJECT_COMMON

    UINT32                  Value;
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    void                    *Reserved_p1;
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Number;


typedef struct /* STRING - has length and pointer */
{
    ACPI_OBJECT_COMMON

    UINT32                  Length;         /* # of bytes in string, excluding trailing null */
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    char                    *Pointer;       /* String value in AML stream or in allocated space */
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_String;


typedef struct /* BUFFER - has length, sequence, and pointer */
{
    ACPI_OBJECT_COMMON

    UINT32                  Length;         /* # of bytes in buffer */
    UINT32                  Sequence;       /* Sequential count of buffers created */
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    UINT8                   *Pointer;       /* points to the buffer in allocated space */
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Buffer;


typedef struct /* PACKAGE - has count, elements, next element */
{
    ACPI_OBJECT_COMMON

    UINT32                  Count;          /* # of elements in package */
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    union AcpiObjInternal   **Elements;     /* Array of pointers to AcpiObjects */
    union AcpiObjInternal   **NextElement;  /* used only while initializing */
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Package;


typedef struct /* FIELD UNIT */
{
    ACPI_OBJECT_COMMON

    ACPI_COMMON_FIELD_INFO        
    UINT32                  Sequence;           /* Container's sequence number */

    union AcpiObjInternal   *Container;         /* Containing object (Buffer) */
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_FieldUnit;


typedef struct /* DEVICE - has handle and notification handler/context */
{
    ACPI_OBJECT_COMMON

    UINT32                  Reserved1;
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    ACPI_HANDLE             Handle;
    union AcpiObjInternal  *SysHandler;         /* Handler for system notifies */
    union AcpiObjInternal  *DrvHandler;         /* Handler for driver notifies */
    union AcpiObjInternal  *AddrHandler;        /* Handler for Address space */
    void                    *Reserved_p5;

} ACPI_OBJECT_Device;


typedef struct /* EVENT */
{
    ACPI_OBJECT_COMMON

    UINT16                  LockCount;
    UINT16                  ThreadId;
    UINT16                  SignalCount;
    UINT16                  Fill1;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    void                    *Semaphore;
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Event;


#define INFINITE_CONCURRENCY        0xFF

typedef struct /* METHOD */
{
    ACPI_OBJECT_COMMON

    UINT8                   MethodFlags;
    UINT8                   ParamCount;
    UINT8                   Concurrency;
    UINT8                   Fill1;
    UINT32                  PcodeLength;
    UINT32                  TableLength;
    ACPI_OWNER_ID           OwningId;
    UINT16                  Reserved4;

    UINT8                   *Pcode;
    UINT8                   *AcpiTable;
    void                    *ParserOp;
    void                    *Semaphore;
    void                    *Reserved_p5;

} ACPI_OBJECT_Method;


typedef struct /* MUTEX */
{
    ACPI_OBJECT_COMMON

    UINT16                  LockCount;
    UINT16                  ThreadId;
    UINT16                  SyncLevel;
    UINT16                  Fill1;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    void                    *Semaphore;
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Mutex;

/*  Flags for Region */

#define INITIAL_REGION_FLAGS        0x0000  /* value set when the region is created */
#define REGION_AGRUMENT_DATA_VALID  0x0001  /* Addr/Len are set */
#define REGION_INITIALIZED          0x0002  /* region init handler has been called */
                                            /* this includes _REG method, if any */

typedef struct /* REGION */
{
    ACPI_OBJECT_COMMON

    UINT16                  SpaceId;
    UINT16                  RegionFlags;        /* bits defined above */
    UINT32                  Address;
    UINT32                  Length;
    UINT32                  Reserved4;          /* Region Specific data (PCI _ADR) */

    union AcpiObjInternal  *Method;             /* Associated control method */
    union AcpiObjInternal  *AddrHandler;        /* Handler for system notifies */
    union AcpiObjInternal  *Link;               /* Link in list of regions */
                                                /* list is owned by AddrHandler */
    NAME_TABLE_ENTRY       *REGMethod;          /* _REG method for this region (if any) */
    NAME_TABLE_ENTRY       *Nte;                /* containing object */

} ACPI_OBJECT_Region;


typedef struct /* POWER RESOURCE - has Handle and notification handler/context*/
{
    ACPI_OBJECT_COMMON

    UINT32                  SystemLevel;
    UINT32                  ResourceOrder;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    ACPI_HANDLE             Handle;
    union AcpiObjInternal   *SysHandler;        /* Handler for system notifies */
    union AcpiObjInternal   *DrvHandler;        /* Handler for driver notifies */
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_PowerResource;


typedef struct /* PROCESSOR - has Handle and notification handler/context*/
{
    ACPI_OBJECT_COMMON

    UINT32                  ProcId;
    UINT32                  PBLKAddress;
    UINT32                  PBLKLength;
    UINT32                  Reserved4;

    ACPI_HANDLE             Handle;
    union AcpiObjInternal   *SysHandler;        /* Handler for system notifies */
    union AcpiObjInternal   *DrvHandler;        /* Handler for driver notifies */
    union AcpiObjInternal   *AddrHandler;       /* Handler for Address space */
    void                    *Reserved_p5;

} ACPI_OBJECT_Processor;


typedef struct /* THERMAL ZONE - has Handle and Handler/Context */
{
    ACPI_OBJECT_COMMON

    UINT32                  Reserved1;
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    ACPI_HANDLE             Handle;
    union AcpiObjInternal   *SysHandler;        /* Handler for system notifies */
    union AcpiObjInternal   *DrvHandler;        /* Handler for driver notifies */
    union AcpiObjInternal   *AddrHandler;       /* Handler for Address space */
    void                    *Reserved_p5;

} ACPI_OBJECT_ThermalZone;


/*
 * Internal types
 */

typedef struct /* FIELD */
{
    ACPI_OBJECT_COMMON

    ACPI_COMMON_FIELD_INFO        
    UINT32                  Reserved4;

    union AcpiObjInternal   *Container;         /* Containing object */
    void                    *Reserved_p2;
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Field;


typedef struct /* BANK FIELD */
{
    ACPI_OBJECT_COMMON

    ACPI_COMMON_FIELD_INFO        
    UINT32                  Value;              /* Value to store into BankSelect */

    ACPI_HANDLE             BankSelect;         /* Bank select register */
    union AcpiObjInternal   *Container;         /* Containing object */
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_BankField;


typedef struct /* INDEX FIELD */
{
    /* 
     * No container pointer needed since the index and data register definitions
     * will define how to access the respective registers
     */
    ACPI_OBJECT_COMMON

    ACPI_COMMON_FIELD_INFO
    UINT32                  Value;              /* Value to store into Index register */
    
    ACPI_HANDLE             Index;              /* Index register */
    ACPI_HANDLE             Data;               /* Data register */
    void                    *Reserved_p3;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_IndexField;


typedef struct /* NOTIFY HANDLER */
{
    ACPI_OBJECT_COMMON

    UINT32                  Reserved1;
    UINT32                  Reserved2;
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    NAME_TABLE_ENTRY        *Nte;               /* Parent device */
    NOTIFY_HANDLER          Handler;
    void                    *Context;
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_NotifyHandler;


/* Flags for address handler */

#define ADDR_HANDLER_DEFAULT_INSTALLED  0x1

typedef struct /* ADDRESS HANDLER */
{
    ACPI_OBJECT_COMMON

    UINT16                  SpaceId;
    UINT16                  Hflags;
    ADDRESS_SPACE_HANDLER   Handler;
    ADDRESS_SPACE_SETUP     Setup;
    UINT32                  Reserved4;

    NAME_TABLE_ENTRY        *Nte;               /* Parent device */
    void                    *Context;
    void                    *InstallTimeContext;
    union AcpiObjInternal   *Link;              /* Link to next handler on device */
    union AcpiObjInternal   *RegionList;        /* regions using this handler */

} ACPI_OBJECT_AddrHandler;


/*
 * The Reference object type is used for these opcodes:
 * Arg[0-6], Local[0-7], IndexOp, NameOp, ZeroOp, OneOp, OnesOp, DebugOp
 */

typedef struct /* Reference - Local object type */
{
    ACPI_OBJECT_COMMON

    UINT16                  OpCode;  
    UINT8                   Fill1;
    UINT8                   TargetType;         /* Used for IndexOp */
    UINT32                  Offset;             /* Used for ArgOp, LocalOp, and IndexOp */
    UINT32                  Reserved3;
    UINT32                  Reserved4;

    void                    *Object;            /* NameOp=>HANDLE to obj, IndexOp=>ACPI_OBJECT_INTERNAL */
    NAME_TABLE_ENTRY        *Nte;
    union AcpiObjInternal   **Where;  
    void                    *Reserved_p4;
    void                    *Reserved_p5;

} ACPI_OBJECT_Reference;



typedef union AcpiObjInternal           
{
    ACPI_OBJECT_Common          Common;
    ACPI_OBJECT_Number          Number;
    ACPI_OBJECT_String          String;
    ACPI_OBJECT_Buffer          Buffer;
    ACPI_OBJECT_Package         Package;
    ACPI_OBJECT_FieldUnit       FieldUnit;
    ACPI_OBJECT_Device          Device;
    ACPI_OBJECT_Event           Event;
    ACPI_OBJECT_Method          Method;
    ACPI_OBJECT_Mutex           Mutex;
    ACPI_OBJECT_Region          Region;
    ACPI_OBJECT_PowerResource   PowerResource;
    ACPI_OBJECT_Processor       Processor;
    ACPI_OBJECT_ThermalZone     ThermalZone;
    ACPI_OBJECT_Field           Field;
    ACPI_OBJECT_BankField       BankField;
    ACPI_OBJECT_IndexField      IndexField;
    ACPI_OBJECT_Reference       Reference;
    ACPI_OBJECT_NotifyHandler   NotifyHandler;
    ACPI_OBJECT_AddrHandler     AddrHandler;

} ACPI_OBJECT_INTERNAL;

#endif /* _ACOBJECT_H */
