
/*
 *  ./acpica/subsystem/include/acpiosd.h - All interfaces to the OS-dependent layer.  These 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#ifndef __ACPIOSD_H__
#define __ACPIOSD_H__

#include <acenv.h>
#include <actypes.h>


/* Priorities for OsdQueueForExecution */

#define OSD_PRIORITY_HIGH   1
#define OSD_PRIORITY_MED    2
#define OSD_PRIORITY_LO     3
#define OSD_PRIORITY_GPE    OSD_PRIORITY_HIGH


/* 
 * Types specific to the OS-dependent layer interfaces 
 */

typedef
UINT32 (*OSD_HANDLER) (
    void                    *Context);

typedef
void (*OSD_EXECUTION_CALLBACK) (
    void                    *Context);


/*
 * Initialization and shutdown primitives  (Optional)
 */

ACPI_STATUS
OsdInitialize (
 void);

ACPI_STATUS
OsdTerminate (
 void);

/*
 * Synchronization primitives
 */
 
ACPI_STATUS
OsdCreateSemaphore (
 UINT32        InitialUnits,
 ACPI_HANDLE       *OutHandle);

ACPI_STATUS
OsdDeleteSemaphore (
 ACPI_HANDLE       Handle);

ACPI_STATUS
OsdWaitSemaphore (
 ACPI_HANDLE       Handle,
 UINT32        Units,
 UINT32        Timeout);

ACPI_STATUS
OsdSignalSemaphore (
 ACPI_HANDLE       Handle,
 UINT32        Units);

/* 
 * Memory allocation and mapping 
 */

void *
OsdAllocate (
    UINT32                  Size);

void *
OsdCallocate (
    UINT32                  Size);

void 
OsdFree (
    void *                  Memory);

ACPI_STATUS
OsdMapMemory (
    void                    *PhysicalAddress,
    UINT32                  Length,
 void     **LogicalAddress);

void 
OsdUnMapMemory (
    void                    *LogicalAddress, 
    UINT32                  Length);


/* 
 * Interrupt handlers 
 */

ACPI_STATUS
OsdInstallInterruptHandler (
    UINT32                  InterruptNumber,
    OSD_HANDLER             ServiceRoutine,
    void                    *Context);

ACPI_STATUS
OsdRemoveInterruptHandler (
    UINT32                  InterruptNumber,
    OSD_HANDLER             ServiceRoutine);


/* 
 * Scheduling 
 */

ACPI_STATUS
OsdQueueForExecution (
    UINT32                  Priority,
    OSD_EXECUTION_CALLBACK  Function,
    void                    *Context);

void
OsdSleep (
    UINT32                  Seconds,
    UINT32                  Milliseconds);

void
OsdSleepUsec (
    UINT32                  Microseconds);

/*
 * Platform/Hardware independent I/O interfaces
 */

UINT8
OsdIn8 (
    ACPI_IO_ADDRESS         InPort);


UINT16
OsdIn16 (
    ACPI_IO_ADDRESS         InPort);

UINT32
OsdIn32 (
    ACPI_IO_ADDRESS         InPort);

void
OsdOut8 (
    ACPI_IO_ADDRESS         OutPort, 
    UINT8                   Value);

void
OsdOut16 (
    ACPI_IO_ADDRESS         OutPort, 
    UINT16                  Value);

void
OsdOut32 (
    ACPI_IO_ADDRESS         OutPort, 
    UINT32                  Value);


/*
 * Standard access to PCI configuration space
 */

ACPI_STATUS 
OsdReadPciCfgByte (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT8                   *Value);

ACPI_STATUS 
OsdReadPciCfgWord (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT16                  *Value);

ACPI_STATUS 
OsdReadPciCfgDword (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT32                  *Value);

ACPI_STATUS 
OsdWritePciCfgByte (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT8                   Value);

ACPI_STATUS 
OsdWritePciCfgWord (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT16                  Value);


ACPI_STATUS 
OsdWritePciCfgDword (
    UINT32                  Bus, 
    UINT32                  DeviceFunction, 
    UINT32                  Register, 
    UINT32                  Value);


/*
 * Miscellaneous
 */

ACPI_STATUS
OsdBreakpoint (
    char                    *Message);

BOOLEAN
OsdReadable (
    void                    *Pointer,
    UINT32                  Length);


BOOLEAN
OsdWritable (
    void                    *Pointer,
    UINT32                  Length);



/*
 * Debug print routines 
 */

INT32 
OsdPrintf (
    const char              *Format, 
    ...);

INT32
OsdVprintf (
    const char              *Format, 
    va_list                 Args);

/*
 * Debug input
 */

UINT32
OsdGetLine (
    char                    *Buffer);


/*
 * Debug 
 */

void
OsdDbgAssert(
    void                    *FailedAssertion,
    void                    *FileName,
    UINT32                  LineNumber,
 char                    *Message);


#endif /* __ACPIOSD_H__ */
