
/*
 *  ./acpica/subsystem/include/actables.h - Table data structures defined in ACPI specification
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef __ACTABLES_H__
#define __ACTABLES_H__

typedef NATIVE_UINT         IO_ADDRESS;             /* Only for clarity in declarations */

/*
 *  Values for description table header signatures
 */

#define RSDP_SIG            "RSD PTR "              /* RSDT Pointer signature */
#define APIC_SIG            "APIC"                  /* Multiple APIC Description Table */
#define DSDT_SIG            "DSDT"                  /* Differentiated System Description Table */
#define FACP_SIG            "FACP"                  /* Fixed ACPI Description Table */
#define FACS_SIG            "FACS"                  /* Firmware ACPI Control Structure */
#define PSDT_SIG            "PSDT"                  /* Persistent System Description Table */
#define RSDT_SIG            "RSDT"                  /* Root System Description Table */
#define SSDT_SIG            "SSDT"                  /* Secondary System Description Table */
#define SBST_SIG            "SBST"                  /* Smart Battery Specification Table */
#define BOOT_SIG            "BOOT"                  /* Boot table */


#define GL_OWNED            0x02                    /* Ownership of global lock is bit 1 */

/* values of Mapic.Model */

#define DUAL_PIC            0
#define MULTIPLE_APIC       1

/* values of Type in APIC_HEADER */

#define APIC_PROC           0
#define APIC_IO             1


/*
 * Architecture-independent tables
 * The architecture dependent tables are in separate files
 */

typedef struct  /* Root System Descriptor Pointer */
{                                           
    char                    Signature [8];          /* contains "RSD PTR " */
    UINT8                   Checksum;               /* to make sum of struct == 0 */
    char                    OemId [6];              /* OEM identification */
    UINT8                   Reserved;               /* reserved - must be zero */
    UINT32                  RsdtPhysicalAddress;    /* physical address of RSDT */

} ROOT_SYSTEM_DESCRIPTOR_POINTER;


typedef struct  /* ACPI common table header */
{                                           
    char                    Signature [4];          /* identifies type of table */
    UINT32                  Length;                 /* length of table, in bytes,
                                                     * including header */
    UINT8                   Revision;               /* specification minor version # */
    UINT8                   Checksum;               /* to make sum of entire table == 0 */
    char                    OemId [6];              /* OEM identification */
    char                    OemTableId [8];         /* OEM table identification */
    UINT32                  OemRevision;            /* OEM revision number */
    char                    AslCompilerId [4];      /* ASL compiler vendor ID */
    UINT32                  AslCompilerRevision;    /* ASL compiler revision number */

} ACPI_TABLE_HEADER;


typedef struct  /* APIC Table */
{                                           
    ACPI_TABLE_HEADER       header;                 /* table header */
    UINT32                  LocalApicAddress;       /* Physical address for accessing local APICs */
    UINT32_BIT              PCATCompat      : 1;    /* a one indicates system also has dual 8259s */
    UINT32_BIT              Reserved1       : 31;

} APIC_TABLE;


typedef struct  /* APIC Header */
{
    UINT8                   Type;                   /* APIC type.  Either APIC_PROC or APIC_IO */
    UINT8                   Length;                 /* Length of APIC structure */

} APIC_HEADER;


typedef struct  /* Processor APIC */
{
    APIC_HEADER             header;
    UINT8                   ProcessorApicId;        /* ACPI processor id */
    UINT8                   LocalApicId;            /* processor's local APIC id */
    UINT32_BIT              ProcessorEnabled: 1;    /* Processor is usable if set */
    UINT32_BIT              Reserved1       : 32;

} PROCESSOR_APIC;


typedef struct  /* IO APIC */
{
    APIC_HEADER             header;
    UINT8                   IoApicId;               /* I/O APIC ID */
    UINT8                   Reserved;               /* reserved - must be zero */
    UINT32                  IoApicAddress;          /* APIC's physical address */
    UINT32                  Vector;                 /* interrupt vector index where INTI
                                                     * lines start */
} IO_APIC;


typedef struct  /* Smart Battery Description Table */
{
    ACPI_TABLE_HEADER       header;
    UINT32                  WarningLevel;
    UINT32                  LowLevel;
    UINT32                  CriticalLevel;

} SMART_BATTERY_DESCRIPTION_TABLE;




/*
 * ACPI Table information.  We save the table address, length,
 * and type of memory allocation (mapped or allocated) for each
 * table for 1) when we exit, and 2) if a new table is installed
 */

#define ACPI_MEM_NOT_ALLOCATED  0
#define ACPI_MEM_ALLOCATED      1
#define ACPI_MEM_MAPPED         2

#define ACPI_TABLE_SINGLE       0
#define ACPI_TABLE_MULTIPLE     1




/* Data about each known table type */

typedef struct _AcpiTableSupport
{
    char                    *Name;
    char                    *Signature;
    UINT8                   SigLength;
    UINT8                   Flags;
    UINT16                  Status;
    void                    **GlobalPtr;

} ACPI_TABLE_SUPPORT; 



/* 
 * Get the architecture-specific tables
 */

#include <actbl32.h>

#endif /* __ACTABLES_H__ */
