
/*
 *  ./acpica/subsystem/include/hardware.h - hardware specific interfaces
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef __HARDWARE_H__
#define __HARDWARE_H__

/* Sleep states */

#define SLWA_DEBUG_LEVEL    4
#define GTS_CALL            0
#define GTS_WAKE            1

/* Cx States */

#define MAX_CX_STATE_LATENCY 0xFFFFFFFF
#define MAX_CX_STATES  4

/* 
 * The #define's and enum below establish an abstract way of identifying what
 * register block and register is to be accessed.  Do not change any of the
 * values as they are used in switch statements and offset calculations. 
 */
    
#define REGISTER_BLOCK_MASK     0xFF00
#define BIT_IN_REGISTER_MASK    0x00FF
#define PM1_EVT                 0x0100
#define PM1_CONTROL             0x0200
#define PM2_CONTROL             0x0300
#define PM_TIMER                0x0400
#define PROCESSOR_BLOCK         0x0500
#define GPE0_STS_BLOCK          0x0600
#define GPE0_EN_BLOCK           0x0700
#define GPE1_STS_BLOCK          0x0800
#define GPE1_EN_BLOCK           0x0900

enum
{
    /* PM1 status register ids */
    
    TMR_STS =   (PM1_EVT        | 0x01),
    BM_STS,
    GBL_STS,
    PWRBTN_STS,
    SLPBTN_STS,
    RTC_STS,
    WAK_STS,

    /* PM1 enable register ids */
    
    TMR_EN,
    /* need to skip 1 enable number since there's no bus master enable register */  
    GBL_EN =    (PM1_EVT        | 0x0A),
    PWRBTN_EN,
    SLPBTN_EN,
    RTC_EN,

    /* PM1 control register ids */

    SCI_EN =    (PM1_CONTROL    | 0x01),
    BM_RLD,
    GBL_RLS,
    SLP_TYPa,
    SLP_TYPb,
    SLP_EN,

    /* PM2 control register ids */

    ARB_DIS =   (PM2_CONTROL    | 0x01),

    /* PM Timer register ids */

    TMR_VAL =   (PM_TIMER       | 0x01),

    GPE0_STS =  (GPE0_STS_BLOCK | 0x01),
    GPE0_EN =   (GPE0_EN_BLOCK  | 0x01),

    GPE1_STS =  (GPE1_STS_BLOCK | 0x01),
    GPE1_EN =   (GPE0_EN_BLOCK  | 0x01),

    /* Last register value is one less than LAST_REG */

    LAST_REG
};


#define TMR_STS_MASK        0x0001
#define BM_STS_MASK         0x0010
#define GBL_STS_MASK        0x0020
#define PWRBTN_STS_MASK     0x0100
#define SLPBTN_STS_MASK     0x0200
#define RTC_STS_MASK        0x0400
#define WAK_STS_MASK        0x8000
#define ALL_FIXED_STS_BITS  \
                            (TMR_STS_MASK      |   \
                            BM_STS_MASK        |   \
                            GBL_STS_MASK       |   \
                            PWRBTN_STS_MASK    |   \
                            SLPBTN_STS_MASK    |   \
                            RTC_STS_MASK       |   \
                            WAK_STS_MASK)

#define TMR_EN_MASK         0x0001
#define GBL_EN_MASK         0x0020
#define PWRBTN_EN_MASK      0x0100
#define SLPBTN_EN_MASK      0x0200
#define RTC_EN_MASK         0x0400

#define SCI_EN_MASK         0x0001
#define BM_RLD_MASK         0x0002
#define GBL_RLS_MASK        0x0004
#define SLP_TYPx_MASK       0x1C00
#define SLP_EN_MASK         0x2000

#define ARB_DIS_MASK        0x0001

#define GPE0_STS_MASK
#define GPE0_EN_MASK

#define GPE1_STS_MASK
#define GPE1_EN_MASK


#define ACPI_READ           1
#define ACPI_WRITE          2

#define LOW_BYTE            0x00FF
#define ONE_BYTE            0x08

#ifndef SET 
    #define SET             1
#endif
#ifndef CLEAR
    #define CLEAR           0
#endif

ACPI_STATUS
HwInitialize();

ACPI_STATUS
HwShutdown();

ACPI_STATUS
HwInitializeSystemInfo();

ACPI_STATUS
HwSetMode (
    UINT32                  Mode);

UINT32
HwGetMode (
    void);

UINT32
HwGetModeCapabilities (
    void);

/* Register I/O Prototypes */

UINT32
HwRegisterIO (
    INT32                   ReadWrite,
 UINT32     UseLock, 
    INT32                   RegisterId, ... /* DWORD Value */);

void 
HwClearAcpiStatus (
   void);


/* GPE support */

void
HwEnableGpe (
    UINT32                  GpeIndex);

void
HwDisableGpe (
    UINT32                  GpeIndex);

void
HwClearGpe (
    UINT32                  GpeIndex);

void
HwGetGpeStatus (
    UINT32                  GpeNumber,
    ACPI_EVENT_STATUS       *EventStatus);

/* Sleep Prototypes */

ACPI_STATUS
HwObtainSleepTypeRegisterData (
    UINT8                   SleepState,
    UINT8                   *Slp_TypA,
    UINT8                   *Slp_TypB);


/* Cx State Prototypes */

ACPI_STATUS
HwIa32EnterC1(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks);

ACPI_STATUS
HwIa32EnterC2(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks);

ACPI_STATUS
HwIa32EnterC3(
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks);

ACPI_STATUS
HwIa32EnterCx (
    ACPI_IO_ADDRESS         PBlkAddress,
 UINT32     *PmTimerTicks);

ACPI_STATUS
HwIa32SetCx (
    UINT32                  CxState);

ACPI_STATUS
HwIa32GetCxInfo (
    UINT32     CxStates[MAX_CX_STATES]);


/* Throttling Prototypes */

void 
HwEnableThrottling (
    ACPI_IO_ADDRESS         PBlkAddress);

void 
HwDisableThrottling (
    ACPI_IO_ADDRESS         PBlkAddress);

UINT32
HwGetDutyCycle (
    UINT8                   DutyOffset,
    ACPI_IO_ADDRESS         PBlkAddress,
    UINT32                  NumThrottleStates);

void
HwProgramDutyCycle (
    UINT8                   DutyOffset,
    UINT32                  DutyCycle,
    ACPI_IO_ADDRESS         PBlkAddress,
    UINT32                  NumThrottleStates);

NATIVE_UINT
HwLocalPow (
    NATIVE_UINT             x,
    NATIVE_UINT             y);


/* ACPI Timer prototypes */

UINT32
HwPmtTicks (
    void);

UINT32
HwPmtResolution (
    void);


#endif /* __HARDWARE_H__ */
