
/*
 *  ./acpica/subsystem/include/internal.h - Internal data types used across the ACPI subsystem
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _ACPI_INTERNAL_H
#define _ACPI_INTERNAL_H

#include <config.h>


#define WAIT_FOREVER            ((UINT32) -1)

typedef void*                   ACPI_MUTEX;
typedef UINT32                  ACPI_MUTEX_HANDLE;


/* Object descriptor types */

#define DESC_TYPE_ACPI_OBJ      0xAA
#define DESC_TYPE_PARSER        0xBB
#define DESC_TYPE_STATE         0xCC
#define DESC_TYPE_WALK          0xDD
#define DESC_TYPE_NTE           0xEE





/*
 * Predefined handles for the mutex objects used within the subsystem
 * All mutex objects are automatically created by CmMutexInitialize.
 * NOTE: any changes here must be reflected in the Gbl_MutexNames table also!
 */

#define MTX_HARDWARE            0
#define MTX_MEMORY              1
#define MTX_CACHES              2
#define MTX_TABLES              3
#define MTX_PARSER              4
#define MTX_DISPATCHER          5
#define MTX_INTERPRETER         6
#define MTX_EXECUTE             7
#define MTX_NAMESPACE           8
#define MTX_EVENTS              9
#define MTX_OP_REGIONS          10
#define MTX_DEBUG_CMD_READY     11
#define MTX_DEBUG_CMD_COMPLETE  12

#define MAX_MTX                 12
#define NUM_MTX                 MAX_MTX+1


#ifdef ACPI_DEBUG
#ifdef DEFINE_ACPI_GLOBALS

/* Names for the mutexes used in the subsystem */

static char                 *Gbl_MutexNames[] =
{
    "MTX_Hardware",
    "MTX_Memory",
    "MTX_Caches",
    "MTX_Tables",
    "MTX_Parser",
    "MTX_Dispatcher",
    "MTX_Interpreter",
    "MTX_Execute",
    "MTX_Namespace",
    "MTX_Events",
    "MTX_OpRegions",
    "MTX_DebugCmdReady",
    "MTX_DebugCmdComplete"
};

#endif
#endif


/* Table for the global mutexes */

typedef struct AcpiMutexInfo
{
    ACPI_MUTEX                  Mutex;
    UINT32                      UseCount;
    BOOLEAN                     Locked;

} ACPI_MUTEX_INFO;


#define MTX_DO_NOT_LOCK         0
#define MTX_LOCK                1


typedef UINT16                  ACPI_OWNER_ID;       
#define OWNER_TYPE_TABLE        0x0
#define OWNER_TYPE_METHOD       0x1  
#define FIRST_METHOD_ID         0x0000
#define FIRST_TABLE_ID          0x8000

/* TBD: get rid of the need for this! */

#define TABLE_ID_DSDT           0xD1D1



/* Operational modes of the AML interpreter/scanner */

typedef enum 
{
    IMODE_LoadPass1 = 0x01,
    IMODE_LoadPass2 = 0x02,
    IMODE_Execute   = 0x0E

} OPERATING_MODE;


/* 
 * Typedef nte (name table entry) is private to avoid global
 * impact in the event of changes to it.  The externally-known type ACPI_HANDLE
 * is actually an (nte *).  If an external program needs to extract a field
 * from the nte, it should use an access function defined in acpinmsp.c
 *
 * If you need an access function not provided herein, add it to this module
 * rather than exporting the nte typedef.
 *
 * (nte *) are actually used in two different and not entirely compatible
 * ways: as pointer to an individual nte and as pointer to an entire name
 * table (which is an array of nte, sometimes referred to as a scope).  In
 * the latter case, the specific nte pointed to may be unused; however its
 * ParentScope member will be valid.
 *
 * DataType is used to differentiate between internal descriptors, and MUST
 * be the first byte in this structure.
 */

typedef struct NameTableEntry
{
    UINT8                   DataType;
    UINT8                   Type;           /* Type associated with this name */
    UINT8                   Fill1;
    UINT8                   Flags;
    UINT32                  Name;           /* ACPI Name, always 4 chars per ACPI spec */


    void                    *Object;        /* Pointer to attached ACPI object (optional) */
    struct NameTableEntry   *Scope;         /* Scope owned by this name (optional) */
    struct NameTableEntry   *ParentEntry;   /* Parent NTE */
    struct NameTableEntry   *NextEntry;     /* Next NTE within this scope */
    struct NameTableEntry   *PrevEntry;     /* Previous NTE within this scope */

    ACPI_OWNER_ID           OwnerId;        /* ID of owner - either an ACPI table or a method */
    UINT16                  ReferenceCount; /* Current count of references and children */

    /* Align on 16-byte boundary for memory dump readability */
/*
    DEBUG_ONLY_MEMBERS (
    char                    *FillDebug) 
*/
} NAME_TABLE_ENTRY;


#define ENTRY_NOT_FOUND     NULL
#define INVALID_HANDLE      0
#define NULL_HANDLE         INVALID_HANDLE


/* NTE flags */

#define NTE_AML_ATTACHMENT  0x1


/*
 * ACPI Table Descriptor.  One per ACPI table
 */
typedef struct AcpiTableDesc
{
    struct AcpiTableDesc    *Prev;
    struct AcpiTableDesc    *Next;
    struct AcpiTableDesc    *InstalledDesc;
    ACPI_TABLE_HEADER       *Pointer;
    void                    *BasePointer;
    UINT8                   *AmlPointer;
    UINT32                  AmlLength;
    UINT32                  Length;
    UINT32                  Count;
    ACPI_OWNER_ID           TableId;
    UINT8                   Type;
    UINT8                   Allocation;
    BOOLEAN                 LoadedIntoNamespace;

} ACPI_TABLE_DESC;


typedef struct 
{
    char                    *SearchFor;
    ACPI_HANDLE             *List;
    INT32                   *Count;

} FIND_CONTEXT;


typedef struct
{
    NAME_TABLE_ENTRY        *PreviousEntry;
    NAME_TABLE_ENTRY        *NameTable;
    UINT32                  Position;
    BOOLEAN                 TableFull;

} NS_SEARCH_DATA;


/*
 * Predefined Namespace items
 */
#define ACPI_MAX_ADDRESS_SPACE      255
#define ACPI_NUM_ADDRESS_SPACES     256


typedef struct
{
    char                    *Name;
    ACPI_OBJECT_TYPE        Type;
    char                    *Val;

} PREDEFINED_NAMES;





/* 
 * Entry in the AddressSpace (AKA Operation Region) table
 */

typedef struct
{
    ADDRESS_SPACE_HANDLER   Handler;
    void                    *Context;

} ADDRESS_SPACE_INFO;


/* Values and addresses of the GPE registers (both banks) */

typedef struct 
{
    UINT8                   Status;         /* Current value of status reg */
    UINT8                   Enable;         /* Current value of enable reg */
    UINT16                  StatusAddr;     /* Address of status reg */
    UINT16                  EnableAddr;     /* Address of enable reg */
    UINT8                   GpeBase;        /* Base GPE number */

} GPE_REGISTERS;


#define GPE_LEVEL_TRIGGERED         1
#define GPE_EDGE_TRIGGERED          2


/* Information about each particular GPE level */

typedef struct
{
    UINT8                   Type;           /* Level or Edge */

    ACPI_HANDLE             MethodHandle;   /* Method handle for direct (fast) execution */
    GPE_HANDLER             Handler;        /* Address of handler, if any */
    void                    *Context;       /* Context to be passed to handler */

} GPE_LEVEL_INFO;


/* Information about each particular fixed event */

typedef struct
{
 FIXED_EVENT_HANDLER  Handler;  /* Address of handler. */
 void     *Context;  /* Context to be passed to handler */

} FIXED_EVENT_INFO;



/* Information used during field processing */

typedef struct
{
    UINT8                   SkipField;
    UINT8                   FieldFlag;
    UINT32                  PkgLength;

} FIELD_INFO;







#define OP_INFO_TYPE                0x1F
#define OP_INFO_HAS_ARGS            0x20
#define OP_INFO_CHILD_LOCATION      0xC0

/*
 * AML opcode, name, and argument layout
 */
typedef struct acpi_op_info
{
    UINT16                  Opcode;         /* AML opcode */
    UINT8                   Flags;          /* Opcode type, HasArgs flag */
    UINT32                  ParseArgs;      /* Grammar/Parse time arguments */
    UINT32                  RuntimeArgs;    /* Interpret time arguments */

    DEBUG_ONLY_MEMBERS (
    char                    *Name)          /* op name (debug only) */

} ACPI_OP_INFO;


typedef union acpi_op_value
{
    UINT32                  Integer;        /* integer constant */
    UINT32                  Size;           /* bytelist or field size */
    char                    *String;        /* NULL terminated string */
    UINT8                   *Buffer;        /* buffer or string */
    char                    *Name;          /* NULL terminated string */
    struct acpi_generic_op  *Arg;           /* arguments and contained ops */
    NAME_TABLE_ENTRY        *Entry;         /* entry in interpreter namespace tbl */ 

} ACPI_OP_VALUE;


#define ACPI_COMMON_OP \
    UINT8                   DataType;       /* To differentiate various internal objs */\
    UINT8                   Flags;          /* Type of Op */\
    UINT16                  Opcode;         /* AML opcode */\
    UINT32                  AmlOffset;      /* offset of declaration in AML */\
    struct acpi_generic_op  *Parent;        /* parent op */\
    struct acpi_generic_op  *Next;          /* next op */\
    DEBUG_ONLY_MEMBERS (\
    char                    OpName[16])     /* op name (debug only) */\
                                            /* NON-DEBUG members below: */\
    void                    *NameTableEntry;/* for use by interpreter */\
    ACPI_OP_VALUE           Value;          /* Value or args associated with the opcode */\


/*
 * generic operation (eg. If, While, Store)
 */
typedef struct acpi_generic_op
{
    ACPI_COMMON_OP
} ACPI_GENERIC_OP;


/*
 * operation with a name (eg. Scope, Method, Name, NamedField, ...)
 */
typedef struct acpi_named_op
{
    ACPI_COMMON_OP
    UINT32                  Name;           /* 4-byte name or zero if no name */

} ACPI_NAMED_OP;


/*
 * special operation for methods and regions (parsing must be deferred
 * until a first pass parse is completed)
 */
typedef struct acpi_deferred_op
{
    ACPI_COMMON_OP
    UINT32                  Name;           /* 4-byte name or 0 if none */
    UINT32                  BodyLength;     /* AML body size */
    UINT8                   *Body;          /* AML body */
    UINT16                  ThreadCount;    /* Count of threads currently executing a method */

} ACPI_DEFERRED_OP;


/*
 * special operation for bytelists (ByteList only)
 */
typedef struct acpi_bytelist_op
{
    ACPI_COMMON_OP
    UINT8                   *Data;          /* bytelist data */

} ACPI_BYTELIST_OP;



/*
 * Parse state - one state per parser invocation and each control
 * method.
 */

typedef struct acpi_parse_state
{
    UINT8                   *AmlStart;      /* first AML byte */
    UINT8                   *Aml;           /* next AML byte */
    UINT8                   *AmlEnd;        /* (last + 1) AML byte */
    UINT8                   *PkgEnd;        /* current package end */
    ACPI_GENERIC_OP         *StartOp;       /* root of parse tree */
    struct acpi_parse_scope *Scope;         /* current scope */
    struct acpi_parse_scope *ScopeAvail;    /* unused (extra) scope structs */
    struct acpi_parse_state *Next;

} ACPI_PARSE_STATE;


/*
 * Parse scope - one per ACPI scope
 */

typedef struct acpi_parse_scope
{
    ACPI_GENERIC_OP         *Op;            /* current op being parsed */
    UINT8                   *ArgEnd;        /* current argument end */
    UINT8                   *PkgEnd;        /* current package end */
    struct acpi_parse_scope *Parent;        /* parent scope */
    UINT32                  ArgList;        /* next argument to parse */
    UINT32                  ArgCount;       /* Number of fixed arguments */

} ACPI_PARSE_SCOPE;






#define CONTROL_NORMAL                        0xC0
#define CONTROL_CONDITIONAL_EXECUTING         0xC1
#define CONTROL_PREDICATE_EXECUTING           0xC2
#define CONTROL_PREDICATE_FALSE               0xC3
#define CONTROL_PREDICATE_TRUE                0xC4 



#define ACPI_STATE_COMMON                  /* Two 32-bit fields and a pointer */\
    UINT8                   DataType;           /* To differentiate various internal objs */\
    UINT8                   Flags; \
    UINT16                  Value; \
    UINT16                  State; \
    UINT16                  Eval;  \
    void                    *Next; \

typedef struct acpi_common_state
{
    ACPI_STATE_COMMON
} ACPI_COMMON_STATE;


/*
 * Update state - used to traverse complex objects such as packages
 */
typedef struct acpi_update_state
{
    ACPI_STATE_COMMON
    union AcpiObjInternal   *Object;

} ACPI_UPDATE_STATE;

/*
 * Control state - one per if/else and while constructs.
 * Allows nesting of these constructs 
 */
typedef struct acpi_control_state
{
    ACPI_STATE_COMMON
    ACPI_GENERIC_OP         *PredicateOp;   /* Start of if/while predicate */

} ACPI_CONTROL_STATE;


/*
 * Scope state - current scope during namespace lookups
 */

typedef struct acpi_scope_state
{
    ACPI_STATE_COMMON
    NAME_TABLE_ENTRY        *Entry;

} ACPI_SCOPE_STATE;


typedef union acpi_gen_state
{
    ACPI_COMMON_STATE       Common;
    ACPI_CONTROL_STATE      Control;
    ACPI_UPDATE_STATE       Update;
    ACPI_SCOPE_STATE        Scope;

} ACPI_GENERIC_STATE;






/*
 * Walk state - current state of a parse tree walk.  Used for both a leisurely stroll through
 * the tree (for whatever reason), and for control method execution.
 */

#define NEXT_OP_DOWNWARD    1
#define NEXT_OP_UPWARD      2

typedef struct acpi_walk_state
{
    UINT8                   DataType;                           /* To differentiate various internal objs */\
    ACPI_OWNER_ID           OwnerId;                            /* Owner of objects created during the walk */
    BOOLEAN                 LastPredicate;                      /* Result of last predicate */
    UINT8                   NextOpInfo;                         /* Info about NextOp */
    UINT8                   NumOperands;                        /* Stack pointer for Operands[] array */
    UINT8                   NumResults;                         /* Stack pointer for Results[] array */
    UINT8                   CurrentResult;                      /* */

    struct acpi_walk_state  *Next;                              /* Next WalkState in list */
    ACPI_GENERIC_OP         *Origin;                            /* Start of walk */
    ACPI_GENERIC_OP         *PrevOp;                            /* Last op that was processed */
    ACPI_GENERIC_OP         *NextOp;                            /* next op to be processed */
    ACPI_GENERIC_STATE      *ControlState;                      /* List of control states (nested IFs) */
 ACPI_GENERIC_STATE  *ScopeInfo;       /* Stack of nested scopes */
    union AcpiObjInternal   *ReturnDesc;                        /* Return object, if any */
    union AcpiObjInternal   *MethodDesc;                        /* Method descriptor if running a method */
    ACPI_GENERIC_OP         *MethodCallOp;                      /* MethodCall Op if running a method */
    union AcpiObjInternal   *Operands[OBJ_NUM_OPERANDS];        /* Operands passed to the interpreter */
    union AcpiObjInternal   *Results[OBJ_NUM_OPERANDS];         /* Accumulated results */
    struct NameTableEntry   Arguments[MTH_NUM_ARGS];            /* Control method arguments */
    struct NameTableEntry   LocalVariables[MTH_NUM_LOCALS];     /* Control method locals */


} ACPI_WALK_STATE;


/*
 * Walk list - head of a tree of walk states.  Multiple walk states are created when there
 * are nested control methods executing.
 */
typedef struct acpi_walk_list
{

    ACPI_WALK_STATE         *WalkState;

} ACPI_WALK_LIST;



typedef
ACPI_STATUS (*INTERPRETER_CALLBACK) (
    ACPI_WALK_STATE         *State,
    ACPI_GENERIC_OP         *Op);



/* Info used by PsInitObjects */

typedef struct InitWalkInfo
{
    UINT32                  MethodCount;
    UINT32                  OpRegionCount;
    ACPI_TABLE_DESC         *TableDesc;

} INIT_WALK_INFO;


/* TBD: Merge with struct above */

typedef struct AcpiWalkInfo
{
    UINT32                  DebugLevel;
    UINT32                  OwnerId;

} ACPI_WALK_INFO;




/* Entry for a memory allocation (debug only) */

#ifdef ACPI_DEBUG

#define MEM_MALLOC          0
#define MEM_CALLOC          1

typedef struct AllocationInfo 
{
    void                    *Address;
    UINT32                  Size;
    UINT8                   AllocType;
    UINT32                  Component;
    char                    Module[32];
    INT32                   Line;
    struct AllocationInfo   *Previous;
    struct AllocationInfo   *Next;

} ALLOCATION_INFO;

#endif

#endif
