
/*
 *  ./acpica/subsystem/interpreter/icopy.c - AML Interpreter object copy support
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ICOPY_C__

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <tables.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("icopy");


typedef struct Internal_Search_st 
{
    ACPI_OBJECT_INTERNAL        *DestObj;
    UINT32                      Index;
    ACPI_OBJECT_INTERNAL        *SourceObj;

} INTERNAL_PKG_SEARCH_INFO;


/* Used to traverse nested packages when copying*/

INTERNAL_PKG_SEARCH_INFO        CopyLevel[MAX_PACKAGE_DEPTH];



ACPI_STATUS
AmlBuildCopyInternalPackageObject (
    ACPI_OBJECT_INTERNAL    *SourceObj, 
    ACPI_OBJECT_INTERNAL    *DestObj)
{
    UINT32                      CurrentDepth = 0;
    ACPI_STATUS                 Status = AE_OK;
    UINT32                      Length = 0;
    UINT32                      ThisIndex;
    UINT32                      ObjectSpace = 0;
    ACPI_OBJECT_INTERNAL        *ThisDestObj;
    ACPI_OBJECT_INTERNAL        *ThisSourceObj;
    INTERNAL_PKG_SEARCH_INFO    *LevelPtr;
    

    
    FUNCTION_TRACE ("AmlBuildCopyInternalPackageObject");

    /*
     * Initialize the working variables
     */
    
    MEMSET ((void *) CopyLevel, 0, sizeof(CopyLevel));

    CopyLevel[0].DestObj    = DestObj;
    CopyLevel[0].SourceObj  = SourceObj;
    LevelPtr                = &CopyLevel[0];
    CurrentDepth            = 0;

    DestObj->Common.Type        = SourceObj->Common.Type;
    DestObj->Package.Count      = SourceObj->Package.Count;


    /*
     * Build an array of ACPI_OBJECTS in the buffer
     * and move the free space past it
     */

    DestObj->Package.Elements   = CmCallocate ((ACPI_SIZE) (DestObj->Package.Count + 1) *
                                                     sizeof (void *));
    if (!DestObj->Package.Elements)
    {
        /* Package vector allocation failure   */

        REPORT_ERROR ("AmlBuildCopyInternalPackageObject: Package vector allocation failure");
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    DestObj->Package.NextElement = DestObj->Package.Elements;


    while (1)
    {
        ThisIndex       = LevelPtr->Index;
        ThisDestObj     = (ACPI_OBJECT_INTERNAL *) LevelPtr->DestObj->Package.Elements[ThisIndex];
        ThisSourceObj   = (ACPI_OBJECT_INTERNAL *) LevelPtr->SourceObj->Package.Elements[ThisIndex];

        if (IS_THIS_OBJECT_TYPE (ThisSourceObj, ACPI_TYPE_Package))
        {
            /*
             * If this object is a package then we go one deeper
             */
            if (CurrentDepth >= MAX_PACKAGE_DEPTH-1)
            {
                /*
                 * Too many nested levels of packages for us to handle
                 */
                DEBUG_PRINT (ACPI_ERROR, ("AmlBuildCopyInternalPackageObject: Pkg nested too deep (max %d)\n",
                                            MAX_PACKAGE_DEPTH));
                return_ACPI_STATUS (AE_LIMIT);
            }

            /*
             * Build the package object
             */
            ThisDestObj = CmCreateInternalObject (ACPI_TYPE_Package);
            LevelPtr->DestObj->Package.Elements[ThisIndex] = ThisDestObj;


            ThisDestObj->Common.Type        = ACPI_TYPE_Package;
            ThisDestObj->Package.Count      = ThisDestObj->Package.Count;

            /*
             * Save space for the array of objects (Package elements)
             * update the buffer length counter
             */
            ObjectSpace             = ThisDestObj->Package.Count * sizeof (ACPI_OBJECT_INTERNAL);
            Length                  += ObjectSpace;

            CurrentDepth++;
            LevelPtr                = &CopyLevel[CurrentDepth];
            LevelPtr->DestObj       = ThisDestObj;
            LevelPtr->SourceObj     = ThisSourceObj;
            LevelPtr->Index         = 0;

        }   /* if object is a package */
        
        else
        {
            
            ThisDestObj = CmCreateInternalObject (ThisSourceObj->Common.Type);
            LevelPtr->DestObj->Package.Elements[ThisIndex] = ThisDestObj;

            Status = AmlStoreObjectToObject(ThisSourceObj, ThisDestObj);

            if (Status != AE_OK) 
            {
                /*
                 * Failure get out
                 */
                return_ACPI_STATUS (Status);
            }

            Length      +=ObjectSpace;

            LevelPtr->Index++;
            while (LevelPtr->Index >= LevelPtr->DestObj->Package.Count)
            {
                /*
                 * We've handled all of the objects at this level,  This means that we
                 * have just completed a package.  That package may have contained one
                 * or more packages itself
                 */
                if (CurrentDepth == 0)
                {
                    /*
                     * We have handled all of the objects in the top level package
                     * just add the length of the package objects and get out
                     */
                    // *SpaceUsed = Length;
                    return_ACPI_STATUS (AE_OK);
                }

                /*
                 * go back up a level and move the index past the just completed
                 * package object.
                 */
                CurrentDepth--;
                LevelPtr = &CopyLevel[CurrentDepth];
                LevelPtr->Index++;
            }
        }   /* else object is NOT a package */
    }   /* while (1)  */


    /*
     * We'll never get here, but the compiler whines about return value
     */
    return_ACPI_STATUS (AE_OK);
}


