
/*
 *  ./acpica/subsystem/interpreter/ieconfig.c - Namespace reconfiguration (Load/Unload opcodes)
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IECONFIG_C__

#include <acpi.h>
#include <parser.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <events.h>
#include <tables.h>
#include <dispatch.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("ieconfig");




ACPI_STATUS
AmlExecLoadTable (
    ACPI_OBJECT_INTERNAL    *RgnDesc,
    ACPI_HANDLE             *DdbHandle)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *TableDesc = NULL;
    char                    *TablePtr;
    char                    *TableDataPtr;
    ACPI_TABLE_HEADER       TableHeader;
    ACPI_TABLE_DESC         TableInfo;
    UINT32                  i;


    FUNCTION_TRACE ("AmlExecLoadTable");

    /* TBD: Object can be either a field or an opregion */


    /* Get the table header */

    TableHeader.Length = 0;
    for (i = 0; i < sizeof (ACPI_TABLE_HEADER); i++)
    {
        Status = EvAddressSpaceDispatch (RgnDesc, ADDRESS_SPACE_READ, i, 8, 
                                            (UINT32 *) ((char *) &TableHeader + i));
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }

    /* Allocate a buffer for the entire table */

    TablePtr = CmAllocate (TableHeader.Length);
    if (!TablePtr)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Copy the header to the buffer */

    MEMCPY (TablePtr, &TableHeader, sizeof (ACPI_TABLE_HEADER));
    TableDataPtr = TablePtr + sizeof (ACPI_TABLE_HEADER);


    /* Get the table from the op region */

    for (i = 0; i < TableHeader.Length; i++)
    {
        Status = EvAddressSpaceDispatch (RgnDesc, ADDRESS_SPACE_READ, i, 8, 
                                            (UINT32 *) (TableDataPtr + i));
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }
    }


    /* Table must be either an SSDT or a PSDT */

    if ((!STRNCMP (TableHeader.Signature, Gbl_AcpiTableData[TABLE_PSDT].Signature, Gbl_AcpiTableData[TABLE_PSDT].SigLength)) &&
        (!STRNCMP (TableHeader.Signature, Gbl_AcpiTableData[TABLE_SSDT].Signature, Gbl_AcpiTableData[TABLE_SSDT].SigLength)))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Table has invalid signature [%4.4s], must be SSDT or PSDT\n", TableHeader.Signature));
        Status = AE_BAD_SIGNATURE;
        goto Cleanup;
    }

    /* Create an object to be the table handle */

    TableDesc = CmCreateInternalObject (INTERNAL_TYPE_Reference);
    if (!TableDesc)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;
    }


    /* Install the new table into the local data structures */

    TableInfo.Pointer      = (ACPI_TABLE_HEADER *) TablePtr;
    TableInfo.Length       = TableHeader.Length;
    TableInfo.Allocation   = ACPI_MEM_ALLOCATED;
    TableInfo.BasePointer  = TablePtr;

    Status = TbInstallTable (NULL, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    /* Add the table to the namespace */

    Status = AcpiLoadNamespace ();
    if (ACPI_FAILURE (Status))
    {
        /* TBD: Unload the table on failure ? */

        goto Cleanup;
    }

    /* TBD: we need a pointer to the table desc */

    /* Init the table handle */

    TableDesc->Reference.OpCode = AML_LoadOp;
    TableDesc->Reference.Object = TableInfo.InstalledDesc;

    *DdbHandle = TableDesc;

    return_ACPI_STATUS (Status);


Cleanup:

    CmFree (TableDesc);
    CmFree (TablePtr);
    return_ACPI_STATUS (Status);

}
                




ACPI_STATUS
AmlExecUnloadTable (
    ACPI_HANDLE             DdbHandle)
{
    ACPI_STATUS             Status = AE_NOT_IMPLEMENTED;
    ACPI_OBJECT_INTERNAL    *TableDesc = (ACPI_OBJECT_INTERNAL *) DdbHandle;
    ACPI_TABLE_DESC         *TableInfo;


    FUNCTION_TRACE ("AmlExecUnloadTable");


    /* Validate the handle */
    /* TBD: Wasn't this done earlier? */

    if ((!DdbHandle) ||
        (!VALID_DESCRIPTOR_TYPE (DdbHandle, DESC_TYPE_ACPI_OBJ)) ||
        (((ACPI_OBJECT_INTERNAL *)DdbHandle)->Common.Type != INTERNAL_TYPE_Reference))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* Get the actual table descriptor from the DdbHandle */

    TableInfo = (ACPI_TABLE_DESC *) TableDesc->Reference.Object;

    /* 
     * Delete the entire namespace under this table NTE 
     * (Offset contains the TableId)
     */

    Status = NsDeleteNamespaceByOwner (TableInfo->TableId);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Delete the table itself */

    TbDeleteSingleTable (TableInfo->InstalledDesc);

    /* Delete the table descriptor (DdbHandle) */

    CmRemoveReference (TableDesc);

    return_ACPI_STATUS (Status);
}
 



ACPI_STATUS
AmlExecReconfiguration (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *RegionDesc = NULL;
    ACPI_HANDLE             *DdbHandle;


    FUNCTION_TRACE ("AmlExecReconfiguration");


    /* Resolve the operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 2, "after AmlResolveOperands");

    /* Get the table handle, common for both opcodes */

    Status |= DsObjStackPopObject ((ACPI_OBJECT_INTERNAL **) &DdbHandle, WalkState);

    switch (Opcode)
    {

    case AML_LoadOp:

        /* Get the region or field descriptor */

        Status |= DsObjStackPopObject (&RegionDesc, WalkState);
        if (Status != AE_OK)
        {
            AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 2);
            goto Cleanup2;
        }

        Status = AmlExecLoadTable (RegionDesc, DdbHandle);
        break;


    case AML_UnLoadOp:

        if (Status != AE_OK)
        {
            AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 1);
            goto Cleanup1;
        }

        Status = AmlExecUnloadTable (DdbHandle);
        break;


    default:
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecReconfiguration: bad opcode=%X\n", Opcode));

        Status = AE_AML_BAD_OPCODE;
        break;
    }

Cleanup2:
    CmRemoveReference (RegionDesc);

Cleanup1:
//    CmRemoveReference (DdbHandle);


    return_ACPI_STATUS (Status);
}

