
/*
 *  ./acpica/subsystem/interpreter/iedyad.c - ACPI AML (p-code) execution for dyadic operators
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __IEDYAD_C__

#include <acpi.h>
#include <parser.h>
#include <namesp.h>
#include <interp.h>
#include <events.h>
#include <amlcode.h>
#include <dispatch.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iedyad");



ACPI_STATUS
AmlExecDyadic1 (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc = NULL;
    ACPI_OBJECT_INTERNAL    *ValDesc = NULL;
    NAME_TABLE_ENTRY        *Entry;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("AmlExecDyadic1", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 2, "after AmlResolveOperands");

    /* Get the operands */

    Status |= DsObjStackPopObject (&ValDesc, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (ACPI_FAILURE (Status))
    {
        /* Invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 2);
        goto Cleanup;
    }



    /* Examine the opcode */

    switch (Opcode)
    {

    /* DefNotify   :=  NotifyOp    NotifyObject    NotifyValue */

    case AML_NotifyOp:

        /* The ObjDesc is actually an NTE */

        Entry = (NAME_TABLE_ENTRY *) ObjDesc;
        ObjDesc = NULL;

        /* Object must be a device or thermal zone */

        if (Entry && ValDesc)
        {
            switch (Entry->Type)
            {
            case ACPI_TYPE_Device:
            case ACPI_TYPE_Thermal:
            
                /* Requires that Device and ThermalZone be compatible mappings */

                /* Dispatch the notify to the appropriate handler */

                EvNotifyDispatch (Entry, ValDesc->Number.Value);
                break;

            default:
                DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic1/NotifyOp: unexpected notify object type %d\n",
                                ObjDesc->Common.Type));
                
                Status = AE_AML_OPERAND_TYPE;
            }
        }
        break;

    default:
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic1: Unknown dyadic opcode %02x\n", Opcode));
        Status = AE_AML_BAD_OPCODE;
    }


Cleanup:

    /* Always delete both operands */

    CmRemoveReference (ValDesc);
    CmRemoveReference (ObjDesc);
    

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecDyadic2R (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc    = NULL;
    ACPI_OBJECT_INTERNAL    *ObjDesc2   = NULL;
    ACPI_OBJECT_INTERNAL    *ResDesc    = NULL;
    ACPI_OBJECT_INTERNAL    *ResDesc2   = NULL;
    ACPI_OBJECT_INTERNAL    *RetDesc    = NULL;
    ACPI_OBJECT_INTERNAL    *RetDesc2   = NULL;
    ACPI_STATUS             Status      = AE_OK;
    UINT32                  Remainder;
    INT32                   NumOperands = 3;
    char                    *NewBuf;


    FUNCTION_TRACE_U32 ("AmlExecDyadic2R", Opcode);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), NumOperands, "after AmlResolveOperands");

    /* Get all operands */

    if (AML_DivideOp == Opcode)
    {
        NumOperands = 4;
        Status |= DsObjStackPopObject (&ResDesc2, WalkState);
    }

    Status |= DsObjStackPopObject (&ResDesc, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc2, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, &(WalkState->Operands [WalkState->NumOperands -1]), NumOperands);
        goto Cleanup;
    }




    /* Create an internal return object if necessary */

    switch (Opcode)
    {
    case AML_AddOp:
    case AML_BitAndOp:
    case AML_BitNandOp:
    case AML_BitOrOp:
    case AML_BitNorOp:
    case AML_BitXorOp:
    case AML_DivideOp:
    case AML_MultiplyOp:
    case AML_ShiftLeftOp:
    case AML_ShiftRightOp:
    case AML_SubtractOp:

        RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }

        break;
    }


    /* 
     * Execute the opcode 
     */

    switch (Opcode)
    {

    /* DefAdd  :=  AddOp   Operand1    Operand2    Result  */

    case AML_AddOp:

        RetDesc->Number.Value = ObjDesc->Number.Value + ObjDesc2->Number.Value;
        break;
 
        
    /* DefAnd  :=  AndOp   Operand1    Operand2    Result  */

    case AML_BitAndOp:

        RetDesc->Number.Value = ObjDesc->Number.Value & ObjDesc2->Number.Value;
        break;

        
    /* DefNAnd :=  NAndOp  Operand1    Operand2    Result  */

    case AML_BitNandOp:

        RetDesc->Number.Value = ~(ObjDesc->Number.Value & ObjDesc2->Number.Value);
        break;
   
       
    /* DefOr   :=  OrOp    Operand1    Operand2    Result  */
        
    case AML_BitOrOp:

        RetDesc->Number.Value = ObjDesc->Number.Value | ObjDesc2->Number.Value;
        break;

        
    /* DefNOr  :=  NOrOp   Operand1    Operand2    Result  */

    case AML_BitNorOp:

        RetDesc->Number.Value = ~(ObjDesc->Number.Value | ObjDesc2->Number.Value);
        break;

        
    /* DefXOr  :=  XOrOp   Operand1    Operand2    Result  */

    case AML_BitXorOp:

        RetDesc->Number.Value = ObjDesc->Number.Value ^ ObjDesc2->Number.Value;
        break;

        
    /* DefDivide   :=  DivideOp Dividend Divisor Remainder Quotient    */

    case AML_DivideOp:

        if ((UINT32) 0 == ObjDesc2->Number.Value)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic2R/DivideOp: Divide by zero\n"));
            REPORT_ERROR ("AmlExecDyadic2R/DivideOp: Divide by zero");

            Status = AE_AML_DIVIDE_BY_ZERO;
            goto Cleanup;
        }

        RetDesc2 = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc2)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }

        Remainder               = ObjDesc->Number.Value % ObjDesc2->Number.Value;
        RetDesc2->Number.Value  = ObjDesc->Number.Value / ObjDesc2->Number.Value;   /* Result (what we used to call the quotient) */
        RetDesc->Number.Value   = Remainder;                                        /* Remainder */
        break;

        
    /* DefMultiply :=  MultiplyOp  Operand1    Operand2    Result  */

    case AML_MultiplyOp:

        RetDesc->Number.Value = ObjDesc->Number.Value * ObjDesc2->Number.Value;
        break;

        
    /* DefShiftLeft    :=  ShiftLeftOp Operand ShiftCount  Result  */

    case AML_ShiftLeftOp:

        RetDesc->Number.Value = ObjDesc->Number.Value << ObjDesc2->Number.Value;
        break;

        
    /* DefShiftRight   :=  ShiftRightOp    Operand ShiftCount  Result  */

    case AML_ShiftRightOp:

        RetDesc->Number.Value = ObjDesc->Number.Value >> ObjDesc2->Number.Value;
        break;

        
    /* DefSubtract :=  SubtractOp  Operand1    Operand2    Result  */

    case AML_SubtractOp:

        RetDesc->Number.Value = ObjDesc->Number.Value - ObjDesc2->Number.Value;
        break;


    /* DefConcat   :=  ConcatOp    Data1   Data2   Result  */

    case AML_ConcatOp:

        if (ObjDesc2->Common.Type != ObjDesc->Common.Type)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic2R/ConcatOp: operand type mismatch %d %d\n",
                            ObjDesc->Common.Type, ObjDesc2->Common.Type));
            Status = AE_AML_OPERAND_TYPE;
            goto Cleanup;
        }

        /* Both operands are now known to be the same */
        
        if (ACPI_TYPE_String == ObjDesc->Common.Type)
        {
            RetDesc = CmCreateInternalObject (ACPI_TYPE_String);
            if (!RetDesc)
            {
                Status = AE_NO_MEMORY;
                goto Cleanup;
            }

            /* Operand1 is string  */

            NewBuf = CmAllocate ((ACPI_SIZE) (ObjDesc->String.Length
                                                + ObjDesc2->String.Length + 1));
            if (!NewBuf)
            {
                REPORT_ERROR ("AmlExecDyadic2R/ConcatOp: String allocation failure");
                Status = AE_NO_MEMORY;
                goto Cleanup;
            }
            
            STRCPY (NewBuf, (char *) ObjDesc->String.Pointer);
            STRCPY (NewBuf + ObjDesc->String.Length, (char *) ObjDesc2->String.Pointer);
            
            /* Point the return object to the new string */
            
            RetDesc->String.Pointer = NewBuf;
            RetDesc->String.Length = ObjDesc->String.Length += ObjDesc2->String.Length;
        }
        
        else
        {
            /* Operand1 is not a string ==> must be a buffer */

            RetDesc = CmCreateInternalObject (ACPI_TYPE_Buffer);
            if (!RetDesc)
            {
                Status = AE_NO_MEMORY;
                goto Cleanup;
            }

            NewBuf = CmAllocate ((ACPI_SIZE) (ObjDesc->Buffer.Length
                                                + ObjDesc2->Buffer.Length));
            if (!NewBuf)
            {
                /* Only bail out if the buffer is small */
                
                /* TBD: what is the point of this code? */

                if (ObjDesc->Buffer.Length + ObjDesc2->Buffer.Length < 1024)
                {
                    REPORT_ERROR ("AmlExecDyadic2R/ConcatOp: Buffer allocation failure");
                    return_ACPI_STATUS (AE_NO_MEMORY);
                }

                DEBUG_PRINT (ACPI_ERROR, (
                            "AmlExecDyadic2R/ConcatOp: Buffer allocation failure %d\n",
                            ObjDesc->Buffer.Length + ObjDesc2->Buffer.Length));
                Status = AE_NO_MEMORY;
                goto Cleanup;
            }

            MEMCPY (NewBuf, ObjDesc->Buffer.Pointer, 
                            (ACPI_SIZE) ObjDesc->Buffer.Length);
            MEMCPY (NewBuf + ObjDesc->Buffer.Length, ObjDesc2->Buffer.Pointer,
                            (ACPI_SIZE) ObjDesc2->Buffer.Length);
            
            /*
             * Point the return object to the new buffer
             */
            
            RetDesc->Buffer.Pointer     = (UINT8 *) NewBuf;
            RetDesc->Buffer.Length      = ObjDesc->Buffer.Length + ObjDesc2->Buffer.Length;
        }
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic2R: Unknown dyadic opcode %02x\n", Opcode));
        Status = AE_AML_BAD_OPCODE;
        goto Cleanup;
    }

    
    /*
     * Store the result of the operation (which is now in ObjDesc) into
     * the result descriptor, or the location pointed to by the result
     * descriptor (ResDesc).
     */

    if ((Status = AmlExecStore (RetDesc, ResDesc)) != AE_OK)
    {
        goto Cleanup;
    }
    
    if (AML_DivideOp == Opcode)
    {
        Status = AmlExecStore (RetDesc2, ResDesc2);

        /*
         * Since the remainder is not returned, remove a reference to
         * the object we created earlier
         */

        CmRemoveReference (RetDesc2);
    }


Cleanup:

    /* Always delete the operands */

    CmRemoveReference (ObjDesc);
    CmRemoveReference (ObjDesc2);


    /* Delete return object on error */

    if (ACPI_FAILURE (Status))
    {
        /* On failure, delete the result ops */

        CmRemoveReference (ResDesc);
        CmRemoveReference (ResDesc2);

        if (RetDesc)
        {
            /* And delete the internal return object */

            CmRemoveReference (RetDesc);
            RetDesc = NULL;
        }
    }

    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecDyadic2S (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *TimeDesc;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("AmlExecDyadic2S", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 2, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&TimeDesc, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {   
        /* Invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 2);
        goto Cleanup;
    }



    /* Create the internal return object */

    RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
    if (!RetDesc)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;
    }
    RetDesc->Number.Value = 0;              /* Default return value is FALSE, operation did not time out */

    
    /* Examine the opcode */

    switch (Opcode)
    {

    /* DefAcquire  :=  AcquireOp   MutexObject Timeout */

    case AML_AcquireOp:

        Status = OsAcquireMutex (TimeDesc, ObjDesc);
        break;


    /* DefWait :=  WaitOp  EventObject Timeout */

    case AML_WaitOp:

        Status = OsWaitEvent (TimeDesc, ObjDesc);
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic2S: Unknown dyadic synchronization opcode %02x\n",
                        Opcode));
        Status = AE_AML_BAD_OPCODE;
        goto Cleanup;
    }


    /* Return a boolean indicating if operation timed out (TRUE) or not (FALSE) */

    if (Status == AE_TIME)
    {
        RetDesc->Number.Value = (UINT32)(-1);   /* TRUE, operation timed out */
        Status = AE_OK;
    }



Cleanup:

    /* Delete params */

    CmRemoveReference (TimeDesc);
    CmRemoveReference (ObjDesc);

    /* Delete return object on error */

    if (ACPI_FAILURE (Status) &&
        (RetDesc))
    {
        CmRemoveReference (RetDesc);
        RetDesc = NULL;
    }


    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecDyadic2 (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *ObjDesc2;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_STATUS             Status;
    BOOLEAN                 Lboolean;


    FUNCTION_TRACE_PTR ("AmlExecDyadic2", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 2, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&ObjDesc2, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        /* Invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 2);
        goto Cleanup;
    }




    /* Create the internal return object */

    RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
    if (!RetDesc)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;
    }

    /*
     * Execute the Opcode
     */

    Lboolean = FALSE;
    switch (Opcode)
    {

    /* DefLAnd :=  LAndOp  Operand1    Operand2    */

    case AML_LAndOp:
        
        Lboolean = (BOOLEAN) (ObjDesc->Number.Value && ObjDesc2->Number.Value);
        break;


    /* DefLEqual   :=  LEqualOp    Operand1    Operand2    */

    case AML_LEqualOp:

        Lboolean = (BOOLEAN) (ObjDesc->Number.Value == ObjDesc2->Number.Value);
        break;


    /* DefLGreater :=  LGreaterOp  Operand1    Operand2    */

    case AML_LGreaterOp:

        Lboolean = (BOOLEAN) (ObjDesc->Number.Value > ObjDesc2->Number.Value);
        break;


    /* DefLLess    :=  LLessOp Operand1    Operand2    */

    case AML_LLessOp:

        Lboolean = (BOOLEAN) (ObjDesc->Number.Value < ObjDesc2->Number.Value);
        break;


    /* DefLOr  :=  LOrOp   Operand1    Operand2    */

    case AML_LOrOp:

        Lboolean = (BOOLEAN) (ObjDesc->Number.Value || ObjDesc2->Number.Value);
        break;
    

    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecDyadic2: Unknown dyadic opcode %02x\n", Opcode));
        Status = AE_AML_BAD_OPCODE;
        goto Cleanup;
        break;
    }


    /* Set return value to logical TRUE (all ones) or FALSE (zero) */  
    
    if (Lboolean)
    {
        RetDesc->Number.Value = 0xffffffff;
    }
    else
    {
        RetDesc->Number.Value = 0;
    }


Cleanup:

    /* Always delete operands */

    CmRemoveReference (ObjDesc);
    CmRemoveReference (ObjDesc2);

 
    /* Delete return object on error */

    if (ACPI_FAILURE (Status) &&
        (RetDesc))
    {
        CmRemoveReference (RetDesc);
        RetDesc = NULL;
    }


    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}


