/*
 *  ./acpica/subsystem/interpreter/iefield.c - ACPI AML (p-code) execution - field manipulation
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __IEFIELD_C__

#include <acpi.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <hardware.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iefield");





ACPI_STATUS
AmlSetupField (
    ACPI_OBJECT_INTERNAL    *ObjDesc, 
    ACPI_OBJECT_INTERNAL    *RgnDesc, 
    INT32                   FieldBitWidth)
{
    ACPI_STATUS             Status = AE_OK;
    INT32                   FieldByteWidth;


    FUNCTION_TRACE ("AmlSetupField");


    /* Parameter validation */

    if (!ObjDesc || !RgnDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlSetupField: Internal error - null handle\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }

    if (ACPI_TYPE_Region != RgnDesc->Common.Type)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlSetupField: Needed Region, found type %x %s\n",
                        RgnDesc->Common.Type, CmGetTypeName (RgnDesc->Common.Type)));
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }


    /* Init and validate Field width */
    
    FieldByteWidth = FieldBitWidth / 8;     /*  possible values are 1, 2, 4 */

    if ((FieldBitWidth != 8) && 
        (FieldBitWidth != 16) && 
        (FieldBitWidth != 32))
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlSetupField: Internal error - bad width %d\n", FieldBitWidth));
        return_ACPI_STATUS (AE_AML_OPERAND_VALUE);
    }


    /* 
     * If the address and length have not been previously evaluated,
     * evaluate them and save the results.
     */
    if (!(RgnDesc->Region.RegionFlags & REGION_AGRUMENT_DATA_VALID))
    {

        Status = DsGetRegionArguments (RgnDesc);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }


    /* 
     * If (offset rounded up to next multiple of field width)
     * exceeds region length, indicate an error.
     */

    if (RgnDesc->Region.Length <
       (ObjDesc->Field.Offset & ~((UINT32) FieldByteWidth - 1)) + FieldByteWidth)
    {
        /*  
         * Offset rounded up to next multiple of field width
         * exceeds region length, indicate an error
         */

        DUMP_STACK_ENTRY (RgnDesc);
        DUMP_STACK_ENTRY (ObjDesc);

        DEBUG_PRINT (ACPI_ERROR,
                ("AmlSetupField: Operation at %08lX width %d bits exceeds len %08lX field=%p region=%p\n",
                ObjDesc->Field.Offset, FieldBitWidth, RgnDesc->Region.Length, ObjDesc, RgnDesc));

        return_ACPI_STATUS (AE_AML_REGION_LIMIT);
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AmlAccessNamedField (
    INT32                   Mode, 
    ACPI_HANDLE             NamedField, 
    void                    *Buffer,
    UINT32                  BufferLength)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc = NULL;
    ACPI_STATUS             Status = AE_OK;
    BOOLEAN                 Locked = FALSE;
    UINT32                  BitGranularity = 0;
    UINT32                  ByteGranularity;
    UINT32                  DatumLength;
    UINT32                  ActualByteLength;
    UINT32                  ByteFieldLength;


    FUNCTION_TRACE_PTR ("AmlAccessNamedField", NamedField);


    /* Get the attached field object */

    ObjDesc = NsGetAttachedObject (NamedField);
    if (!ObjDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlAccessNamedField: Internal error - null value pointer\n"));
        return_ACPI_STATUS (AE_AML_INTERNAL);
    }

    /* Check the type */

    if (INTERNAL_TYPE_DefField != NsGetType (NamedField))
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlAccessNamedField: Name %4.4s type %x is not a defined field\n",
                        &(((NAME_TABLE_ENTRY *) NamedField)->Name), NsGetType (NamedField)));
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }

    /* ObjDesc valid and NamedField is a defined field  */

    DEBUG_PRINT (ACPI_INFO, ("AccessNamedField: Obj=%p Type=%x Buf=%p Len=%x\n",
                    ObjDesc, ObjDesc->Common.Type, Buffer, BufferLength));
    DEBUG_PRINT (ACPI_INFO, ("AccessNamedField: Mode=%d FieldLen=%d, BitOffset=%d\n",
                    Mode, ObjDesc->FieldUnit.Length, ObjDesc->FieldUnit.BitOffset));
    DUMP_ENTRY (NamedField, ACPI_INFO);


    /* Double-check that the attached object is also a field */

    if (INTERNAL_TYPE_DefField != ObjDesc->Common.Type)
    {
        DEBUG_PRINT (ACPI_ERROR, (
                "AmlAccessNamedField: Internal error - Name %4.4s type %x does not match value-type %x at %p\n",
                &(((NAME_TABLE_ENTRY *) NamedField)->Name), NsGetType (NamedField), ObjDesc->Common.Type, ObjDesc));
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }


    /* Granularity was decoded from the field access type (AnyAcc will be the same as ByteAcc) */

    BitGranularity = ObjDesc->FieldUnit.Granularity;
    ByteGranularity = BitGranularity / 8;

    /* Check if request is too large for the field, and silently truncate if necessary */

    /* TBD: should an error be returned in this case? */

    ByteFieldLength = (UINT32) (ObjDesc->FieldUnit.Length + 7) / 8;


    ActualByteLength = BufferLength;
    if (BufferLength > ByteFieldLength)
    {
        ActualByteLength = ByteFieldLength;

        DEBUG_PRINT (ACPI_INFO, ("AmlAccessNamedField: Byte length too large, truncated to %x\n",
                           ActualByteLength));
    }


    /* Convert byte count to datum count, round up if necessary */

    DatumLength = (ActualByteLength + (ByteGranularity-1)) / ByteGranularity;

    DEBUG_PRINT (ACPI_INFO, ("ByteLen=%x, DatumLen=%x, BitGran=%x, ByteGran=%x\n",
                    ActualByteLength, DatumLength, BitGranularity, ByteGranularity));


    /* Get the global lock if needed */

    Locked = AmlAcquireGlobalLock (ObjDesc->FieldUnit.LockRule);


    /* Perform the actual read or write of the buffer */

    switch (Mode)
    {
    case ACPI_READ:

        Status = AmlReadField (ObjDesc, Buffer, BufferLength, ActualByteLength, DatumLength,
                                BitGranularity, ByteGranularity);
        break;


    case ACPI_WRITE:

        Status = AmlWriteField (ObjDesc, Buffer, BufferLength, ActualByteLength, DatumLength,
                                BitGranularity, ByteGranularity);
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("AccessNamedField: Unknown I/O Mode: %X\n", Mode));
        Status = AE_BAD_PARAMETER;
        break;
    }



    /* Release global lock if we acquired it earlier */

    AmlReleaseGlobalLock (Locked);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlSetNamedFieldValue (
    ACPI_HANDLE             NamedField,
    void                    *Buffer,
    UINT32                  BufferLength)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("AmlSetNamedFieldValue", NamedField);


    if (!NamedField)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlSetNamedFieldValue: Internal error - null handle\n"));
        return_ACPI_STATUS (AE_AML_INTERNAL);
    }

    Status = AmlAccessNamedField (ACPI_WRITE, NamedField, Buffer, BufferLength);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlGetNamedFieldValue (
    ACPI_HANDLE             NamedField, 
    void                    *Buffer,
    UINT32                  BufferLength)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("AmlGetNamedFieldValue", NamedField);


    if ((!NamedField) || (!Buffer))
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetNamedFieldValue: Internal error - null parameter\n"));
        return_ACPI_STATUS (AE_AML_INTERNAL);
    }

    Status = AmlAccessNamedField (ACPI_READ, NamedField, Buffer, BufferLength);
    return_ACPI_STATUS (Status);
}
  
