/*
 *  ./acpica/subsystem/interpreter/iefread.c - read from a field 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __IEFREAD_C__

#include <acpi.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <hardware.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iefread");



ACPI_STATUS
AmlReadFieldData (
    ACPI_OBJECT_INTERNAL    *ObjDesc, 
    UINT32                  FieldByteOffset,
    UINT32                  FieldBitWidth,
    UINT32                  *Value)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *RgnDesc = NULL;
    UINT32                  Address;
    UINT32                  LocalValue = 0;
    INT32                   FieldByteWidth;
    

    FUNCTION_TRACE ("AmlReadFieldData");


    /* ObjDesc is validated by callers */

    if (ObjDesc)
    {
        RgnDesc = ObjDesc->Field.Container;
    }


    FieldByteWidth = FieldBitWidth / 8;
    Status = AmlSetupField (ObjDesc, RgnDesc, FieldBitWidth);
    if (AE_OK != Status)
    {
        return_ACPI_STATUS (Status);
    }

    /* SetupField validated RgnDesc and FieldBitWidth  */

    if (!Value)
    {
        Value = &LocalValue;    /*  support reads without saving value  */
    }


    /* 
     * Round offset down to next multiple of
     * field width, add region base address and offset within the field
     */

    Address = RgnDesc->Region.Address +
              (ObjDesc->Field.Offset & ~((UINT32) FieldByteWidth - 1)) +
              FieldByteOffset;

    if (RgnDesc->Region.SpaceId >= NUM_REGION_TYPES)
    {
        DEBUG_PRINT (TRACE_OPREGION,
                    ("AmlReadFieldData: **** Unknown OpRegion SpaceID %d at %08lx width %d\n",
                    RgnDesc->Region.SpaceId, Address, FieldBitWidth));
    }

    else
    {
        DEBUG_PRINT (TRACE_OPREGION,
                    ("AmlReadFieldData: OpRegion %s at %08lx width %d\n",
                    Gbl_RegionTypes[RgnDesc->Region.SpaceId], Address, FieldBitWidth));
    }


    /* Invoke the appropriate AddressSpace/OpRegion handler */

    Status = EvAddressSpaceDispatch (RgnDesc, ADDRESS_SPACE_READ, 
                                        Address, FieldBitWidth, Value);

    if (Status == AE_NOT_IMPLEMENTED)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlReadFieldData: **** OpRegion type %s not implemented\n",
                Gbl_RegionTypes[RgnDesc->Region.SpaceId]));
    }

    else if (Status == AE_EXIST)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlReadFieldData: **** Unknown OpRegion SpaceID %d\n",
                RgnDesc->Region.SpaceId));
    }

    DEBUG_PRINT (TRACE_OPREGION, ("AmlReadField: Returned value=%08lx \n", *Value));

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlReadField (
    ACPI_OBJECT_INTERNAL    *ObjDesc, 
    void                    *Buffer,
    UINT32                  BufferLength,
    UINT32                  ByteLength,
    UINT32                  DatumLength,
    UINT32                  BitGranularity,
    UINT32                  ByteGranularity)
{
    ACPI_STATUS             Status;
    UINT32                  ThisFieldByteOffset;
    UINT32                  ThisFieldDatumOffset;
    UINT32                  PreviousRawDatum;
    UINT32                  ThisRawDatum;
    UINT32                  ValidFieldBits;
    UINT32                  Mask;
    UINT32                  MergedDatum = 0;



    FUNCTION_TRACE ("AmlReadField");

    /* 
     * Clear the caller's buffer (the whole buffer length as given) 
     * This is very important, especially in the cases where a byte is read,
     * but the buffer is really a UINT32 (4 bytes).
     */

    MEMSET (Buffer, 0, BufferLength);

    /* Read the first raw datum to prime the loop */

    ThisFieldByteOffset = 0;
    ThisFieldDatumOffset= 0;

    Status = AmlReadFieldData (ObjDesc, ThisFieldByteOffset, BitGranularity, &PreviousRawDatum);
    if (ACPI_FAILURE (Status))
    {
        goto Cleanup;
    }

    /* We might actually be done if the request fits in one datum */

    if ((DatumLength == 1) && ((ObjDesc->Field.BitOffset + ObjDesc->FieldUnit.Length) <= (UINT16) BitGranularity))
    {
        MergedDatum = PreviousRawDatum;

        MergedDatum = (MergedDatum >> ObjDesc->Field.BitOffset);

        ValidFieldBits = ObjDesc->FieldUnit.Length % BitGranularity;
        if (ValidFieldBits)
        {
            Mask = (((UINT32) 1 << ValidFieldBits) - (UINT32) 1);
            MergedDatum &= Mask;
        }


        /* Place the MergedDatum into the proper format and return buffer field */
        switch (ByteGranularity)
        {
        case 1:
            ((UINT8 *) Buffer) [ThisFieldDatumOffset] = (UINT8) MergedDatum;
            break;

        case 2:
            ((UINT16 *) Buffer) [ThisFieldDatumOffset] = (UINT16) MergedDatum;
            break;

        case 4:
            ((UINT32 *) Buffer) [ThisFieldDatumOffset] = (UINT32) MergedDatum;
            break;
        }

        ThisFieldByteOffset = 1;
        ThisFieldDatumOffset = 1;
    }

    else
    {
        /* We need to get more raw data to complete one or more field data */

        while (ThisFieldDatumOffset < DatumLength)
        {
            /* Get the next raw datum, it contains bits of the current field datum... */

            Status = AmlReadFieldData (ObjDesc, ThisFieldByteOffset + ByteGranularity, 
                                        BitGranularity, &ThisRawDatum);
            if (ACPI_FAILURE (Status))
            {
                goto Cleanup;
            }

            /* Before merging the data, make sure the unused bits are clear */

            switch (ByteGranularity)
            {
            case 1:
                ThisRawDatum &= 0x000000FF;
                PreviousRawDatum &= 0x000000FF;
                break;

            case 2:
                ThisRawDatum &= 0x0000FFFF;
                PreviousRawDatum &= 0x0000FFFF;
                break;
            }

            /* Put together bits of the two raw data to make a complete field datum */

            if (ObjDesc->Field.BitOffset != 0)
            {
                MergedDatum = (PreviousRawDatum >> ObjDesc->Field.BitOffset) |
                                (ThisRawDatum << (BitGranularity - ObjDesc->Field.BitOffset));
            }

            else
            {
                MergedDatum = PreviousRawDatum;
            }

            /* Now store the datum in the caller's buffer, according to the data type */

            switch (ByteGranularity)
            {
            case 1:
                ((UINT8 *) Buffer) [ThisFieldDatumOffset] = (UINT8) MergedDatum;
                break;

            case 2:
                ((UINT16 *) Buffer) [ThisFieldDatumOffset] = (UINT16) MergedDatum;
                break;

            case 4:
                ((UINT32 *) Buffer) [ThisFieldDatumOffset] = (UINT32) MergedDatum;
                break;
            }


            /* Save the most recent datum since it contains bits of the *next* field datum */

            PreviousRawDatum = ThisRawDatum;

            ThisFieldByteOffset += ByteGranularity;
            ThisFieldDatumOffset++;

        }  /* while */
    }

Cleanup:

    return_ACPI_STATUS (Status);
}






