
/*
 *  ./acpica/subsystem/interpreter/iemonad.c - ACPI AML (p-code) execution for monadic operators
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IEMONAD_C__

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iemonad");






ACPI_STATUS
AmlGetObjectReference (
    ACPI_OBJECT_INTERNAL    *ObjDesc,
    ACPI_OBJECT_INTERNAL    **RetDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("AmlGetObjectReference", ObjDesc);


    if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_ACPI_OBJ))
    {
        if (ObjDesc->Common.Type != INTERNAL_TYPE_Reference)
        {
            *RetDesc = NULL;
            Status = AE_TYPE;
            goto Cleanup;
        }

        /* 
         * Not a Name -- an indirect name pointer would have
         * been converted to a direct name pointer in AmlResolveOperands
         */
        switch (ObjDesc->Reference.OpCode)
        {
        case AML_LocalOp:

            *RetDesc = (void *) DsMethodDataGetNte (MTH_TYPE_LOCAL, (ObjDesc->Reference.Offset));
            break;


        case AML_ArgOp:

            *RetDesc = (void *) DsMethodDataGetNte (MTH_TYPE_ARG, (ObjDesc->Reference.Offset));
            break;


        default:

            DEBUG_PRINT (ACPI_ERROR, ("AmlGetObjectReference: (Internal) Unknown Reference subtype %02x\n",
                            ObjDesc->Reference.OpCode));
            *RetDesc = NULL;
            Status = AE_AML_INTERNAL;
            goto Cleanup;
        }

    }

    else if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
    {
        /* Must be a named object;  Just return the NTE */

        *RetDesc = ObjDesc;
    }
    
    else
    {
        *RetDesc = NULL;
        Status = AE_TYPE;
    }


Cleanup:
 
    DEBUG_PRINT (TRACE_EXEC, ("AmlGetObjectReference: Obj=%p Ref=%p\n", ObjDesc, *RetDesc));
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecMonadic1 (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("AmlExecMonadic1", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS); 
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 1, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 1);
        goto Cleanup;
    }


    /* Examine the opcode */

    switch (Opcode)
    {

    /*  DefRelease  :=  ReleaseOp   MutexObject */

    case AML_ReleaseOp:

        Status = OsReleaseMutex (ObjDesc);
        break;


    /*  DefReset    :=  ResetOp     EventObject */

    case AML_ResetOp:

        Status = OsResetEvent (ObjDesc);
        break;


    /*  DefSignal   :=  SignalOp    EventObject */
    
    case AML_SignalOp:

        Status = OsSignalEvent (ObjDesc);
        break;


    /*  DefSleep    :=  SleepOp     MsecTime    */
    
    case AML_SleepOp:

        OsDoSuspend (ObjDesc->Number.Value);
        break;


    /*  DefStall    :=  StallOp     UsecTime    */
    
    case AML_StallOp:

        OsDoStall (ObjDesc->Number.Value);
        break;


    /*  Unknown opcode  */
    
    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic1: Unknown monadic opcode %02x\n", Opcode));
        Status = AE_AML_BAD_OPCODE;
        break;
    
    } /* switch */



Cleanup:

    /* Always delete the operand */

    CmRemoveReference (ObjDesc);

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AmlExecMonadic2R (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *ResDesc;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_OBJECT_INTERNAL    *RetDesc2 = NULL;
    UINT32                  ResVal;
    ACPI_STATUS             Status;
    INT32                   d0;
    INT32                   d1;
    INT32                   d2;
    INT32                   d3;


    FUNCTION_TRACE_PTR ("AmlExecMonadic2R", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 2, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&ResDesc, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 2);
        goto Cleanup;
    }


    /* Create a return object of type NUMBER for most opcodes */

    switch (Opcode)
    {
    case AML_BitNotOp:
    case AML_FindSetLeftBitOp:
    case AML_FindSetRightBitOp:
    case AML_FromBCDOp:
    case AML_ToBCDOp:
    case AML_CondRefOfOp:

        RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }

        break;
    }


    switch (Opcode)
    {
    /*  DefNot  :=  NotOp   Operand Result  */
    
    case AML_BitNotOp:

        RetDesc->Number.Value = ~ObjDesc->Number.Value;
        break;


    /*  DefFindSetLeftBit   :=  FindSetLeftBitOp    Operand Result  */

    case AML_FindSetLeftBitOp:

        RetDesc->Number.Value = ObjDesc->Number.Value;
        for (ResVal = 0; RetDesc->Number.Value && ResVal < 33; ++ResVal)
        {
            RetDesc->Number.Value >>= 1;
        }

        RetDesc->Number.Value = ResVal;
        break;


    /*  DefFindSetRightBit  :=  FindSetRightBitOp   Operand Result  */

    case AML_FindSetRightBitOp:

        RetDesc->Number.Value = ObjDesc->Number.Value;
        for (ResVal = 0; RetDesc->Number.Value && ResVal < 33; ++ResVal)
        {
            RetDesc->Number.Value <<= 1;
        }

        RetDesc->Number.Value = ResVal == 0 ? 0 : 33 - ResVal;
        break;


    /*  DefFromBDC  :=  FromBCDOp   BCDValue    Result  */

    case AML_FromBCDOp:

        d0 = (INT32) (ObjDesc->Number.Value & 15);
        d1 = (INT32) (ObjDesc->Number.Value >> 4 & 15);
        d2 = (INT32) (ObjDesc->Number.Value >> 8 & 15);
        d3 = (INT32) (ObjDesc->Number.Value >> 12 & 15);
        
        if (d0 > 9 || d1 > 9 || d2 > 9 || d3 > 9)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2R/FromBCDOp: BCD digit too large %d %d %d %d\n",
                            d3, d2, d1, d0));
            Status = AE_AML_NUMERIC_OVERFLOW;
            goto Cleanup;
        }
        
        RetDesc->Number.Value = d0 + d1 * 10 + d2 * 100 + d3 * 1000;
        break;


    /*  DefToBDC    :=  ToBCDOp Operand Result  */
    
    case AML_ToBCDOp:


        if (ObjDesc->Number.Value > 9999)
        {
            DEBUG_PRINT (ACPI_ERROR, ("ExecMonadic2R/ToBCDOp: BCD overflow: %d\n",
                            ObjDesc->Number.Value));
            Status = AE_AML_NUMERIC_OVERFLOW;
            goto Cleanup;
        }
        
        RetDesc->Number.Value
            = ObjDesc->Number.Value % 10
            + (ObjDesc->Number.Value / 10 % 10 << 4)
            + (ObjDesc->Number.Value / 100 % 10 << 8)
            + (ObjDesc->Number.Value / 1000 % 10 << 12);
        
        break;


    /*  DefCondRefOf        :=  CondRefOfOp         SourceObject    Result  */

    case AML_CondRefOfOp:

        /*
         * This op is a little strange because the internal return value is different 
         * than the return value stored in the result descriptor (There are really 
         * two return values)
         */

        if ((NAME_TABLE_ENTRY *) ObjDesc == Gbl_RootObject)
        {
            /* This means that the object does not exist in the namespace, return FALSE */

            RetDesc->Number.Value = 0;

            /* Must delete the result descriptor since there is no reference being returned */

            CmRemoveReference (ResDesc);
            goto Cleanup;
        }

        /* Get the object reference and store it */

        Status = AmlGetObjectReference (ObjDesc, &RetDesc2);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

        Status = AmlExecStore (RetDesc2, ResDesc);

        /* The object exists in the namespace, return TRUE */

        RetDesc->Number.Value = (UINT32) -1;
        goto Cleanup;
        break;


    case AML_StoreOp:

        /*
         * A store operand is typically a number, string, buffer or lvalue
         * TBD: What about a store to a package?
         */

        /* 
         * Do the store, and be careful about deleting the source object,
         * since the object itself may have been stored.
         */
        
        Status = AmlExecStore (ObjDesc, ResDesc);
        if (ACPI_FAILURE (Status))
        {
            /* On failure, just delete the ObjDesc */

            CmRemoveReference (ObjDesc);
        }

        else
        {
            /*
             * Normally, we would remove a reference on the ObjDesc parameter;  But since it
             * is being used as the internal return object (meaning we would normally increment it),
             * the two cancel out, and we simply don't do anything.
             */
            *ReturnDesc = ObjDesc;
        }

        ObjDesc = NULL;
        return_ACPI_STATUS (Status);

        break;


    case AML_DebugOp:
    
        /* Reference, returning an Reference */

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2R: DebugOp should never get here!\n"));
        return_ACPI_STATUS (AE_OK);
        break;


    /*
     * These are obsolete opcodes 
     */

    /*  DefShiftLeftBit     :=  ShiftLeftBitOp      Source          BitNum  */
    /*  DefShiftRightBit    :=  ShiftRightBitOp     Source          BitNum  */

    case AML_ShiftLeftBitOp:
    case AML_ShiftRightBitOp:
        
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2R: %s unimplemented\n", 
                        PsGetOpcodeName (Opcode)));
        Status = AE_SUPPORT;
        goto Cleanup;
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2R: internal error: Unknown monadic opcode %02x\n",
                        Opcode));
        Status = AE_AML_BAD_OPCODE;
        goto Cleanup;
    }
    

    Status = AmlExecStore (RetDesc, ResDesc);



Cleanup:
    /* Always delete the operand object */

    CmRemoveReference (ObjDesc);

    /* Delete return object(s) on error */

    if (ACPI_FAILURE (Status))
    {
        CmRemoveReference (ResDesc);     /* Result descriptor */
        if (RetDesc)
        {
            CmRemoveReference (RetDesc);
            RetDesc = NULL;
        }
    }

    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecMonadic2 (
    UINT16                  Opcode,
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *TmpDesc;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_STATUS             Status;
    UINT32                  Type;
    UINT32                  Value;


    FUNCTION_TRACE_PTR ("AmlExecMonadic2", WALK_OPERANDS);


    /* Resolve all operands */

    Status = AmlResolveOperands (Opcode, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (Opcode), 1, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 1);
        goto Cleanup;
    }



    /* Get the operand and decode the opcode */


    switch (Opcode)
    {

    /*  DefLNot :=  LNotOp  Operand */

    case AML_LNotOp:

        RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }

        RetDesc->Number.Value = !ObjDesc->Number.Value;
        break;


    /*  DefDecrement    :=  DecrementOp Target  */
    /*  DefIncrement    :=  IncrementOp Target  */

    case AML_DecrementOp:
    case AML_IncrementOp:

        /* 
         * Since we are expecting an Reference on the top of the stack, it
         * can be either an NTE or an internal object.
         *
         * TBD: This may be the prototype code for all cases where an Reference
         * is expected!! 10/99
         */

       if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
       {
           RetDesc = ObjDesc;
       }

       else
       {
            /* 
             * Duplicate the Reference in a new object so that we can resolve it
             * without destroying the original Reference object
             */
        
            RetDesc = CmCreateInternalObject (INTERNAL_TYPE_Reference);
            if (!RetDesc)
            {
              Status = AE_NO_MEMORY;
               goto Cleanup;
            }

            RetDesc->Reference.OpCode = ObjDesc->Reference.OpCode;
            RetDesc->Reference.Offset = ObjDesc->Reference.Offset;
            RetDesc->Reference.Object = ObjDesc->Reference.Object;
        }
        

        /* 
         * Convert the RetDesc Reference to a Number
         * (This deletes the original RetDesc)
         */
    
        Status = AmlResolveOperands (AML_LNotOp, &RetDesc);
        if (Status != AE_OK)
        {
            AmlAppendOperandDiag (_THIS_MODULE, __LINE__, Opcode, WALK_OPERANDS, 1);
            goto Cleanup;
        }

        /* Do the actual increment or decrement */
    
        if (AML_IncrementOp == Opcode)
        {
            RetDesc->Number.Value++;
        }
        else
        {
            RetDesc->Number.Value--;
        }

        /* Store the result back in the original descriptor */

        Status = AmlExecStore (RetDesc, ObjDesc);

        /* Objdesc was just deleted (because it is an Reference) */

        ObjDesc = NULL;

        break;


    /*  DefObjectType   :=  ObjectTypeOp    SourceObject    */

    case AML_TypeOp:
        
        if (INTERNAL_TYPE_Reference == ObjDesc->Common.Type)
        {
            /* 
             * Not a Name -- an indirect name pointer would have
             * been converted to a direct name pointer in AmlResolveOperands
             */
            switch (ObjDesc->Reference.OpCode)
            {
            case AML_ZeroOp: 
            case AML_OneOp: 
            case AML_OnesOp:
                
                /* Constants are of type Number */
                
                Type = ACPI_TYPE_Number;
                break;


            case AML_DebugOp:
                
                /* Per 1.0b spec, Debug object is of type DebugObject */
                
                Type = ACPI_TYPE_DebugObject;
                break;


            case AML_IndexOp:

                /* Get the type of this reference (index into another object) */

                Type = ObjDesc->Reference.TargetType;
                if (Type == ACPI_TYPE_Package)
                {
                    /*
                     * The main object is a package, we want to get the type
                     * of the individual package element that is referenced by
                     * the index.
                     */
                    Type = (*(ObjDesc->Reference.Where))->Common.Type;
                }

                break;


            case AML_LocalOp:

                Type = DsMethodDataGetType (MTH_TYPE_LOCAL, (ObjDesc->Reference.Offset));
                break;


            case AML_ArgOp:

                Type = DsMethodDataGetType (MTH_TYPE_ARG, (ObjDesc->Reference.Offset));
                break;


            default:

                DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2/TypeOp:internal error: Unknown Reference subtype %02x\n",
                                ObjDesc->Reference.OpCode));
                Status = AE_AML_INTERNAL;
                goto Cleanup;
            }
        }
        
        else
        {
            /* 
             * Since we passed AmlResolveOperands("l") and it's not an Reference,
             * it must be a direct name pointer. 
             */
            Type = NsGetType ((ACPI_HANDLE) ObjDesc);
        }

        /*  Allocate a descriptor to hold the type. */

        RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }
        
        RetDesc->Number.Value = Type;
        break;


    /*  DefSizeOf   :=  SizeOfOp    SourceObject    */

    case AML_SizeOfOp:

        if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
        {
            ObjDesc = NsGetAttachedObject (ObjDesc);
        }

        if (!ObjDesc)
        {
            Value = 0;
        }

        else
        {
            switch (ObjDesc->Common.Type)
            {

            case ACPI_TYPE_Buffer:

                Value = ObjDesc->Buffer.Length;
                break;


            case ACPI_TYPE_String:

                Value = ObjDesc->String.Length;
                break;


            case ACPI_TYPE_Package:

                Value = ObjDesc->Package.Count;
                break;

            case INTERNAL_TYPE_Reference:

                Value = 4;
                break;

            default:

                DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2: Not Buf/Str/Pkg - found type 0x%X\n", 
                                ObjDesc->Common.Type));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
        }

        /* 
         * Now that we have the size of the object, create a result
         * object to hold the value
         */

        RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!RetDesc)
        {
            Status = AE_NO_MEMORY;
            goto Cleanup;
        }

        RetDesc->Number.Value = Value;
        break;


    /*  DefRefOf    :=  RefOfOp     SourceObject    */

    case AML_RefOfOp:

        Status = AmlGetObjectReference (ObjDesc, &RetDesc);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }
        break;


    /*  DefDerefOf  :=  DerefOfOp   ObjReference    */

    case AML_DerefOfOp:


        /* Check for a method local or argument */

        if (!VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
        {
            /* 
             * Must resolve/dereference the local/arg reference first
             */
            switch (ObjDesc->Reference.OpCode)
            {
            /* Set ObjDesc to the value of the local/arg */

            case AML_LocalOp:

                DsMethodDataGetValue (MTH_TYPE_LOCAL, (ObjDesc->Reference.Offset), &TmpDesc);
                CmRemoveReference (ObjDesc);    /* Delete our ref to the input object */
                ObjDesc = TmpDesc;              /* Point to the object just retrieved */
                break;


            case AML_ArgOp:

                DsMethodDataGetValue (MTH_TYPE_ARG, (ObjDesc->Reference.Offset), &TmpDesc);
                CmRemoveReference (ObjDesc);    /* Delete our ref to the input object */
                ObjDesc = TmpDesc;              /* Point to the object just retrieved */
                break;

            default:

                /* Index op - handled below */
                break;
            }
        }


        /* ObjDesc may have changed from the code above */

        if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
        {
            /* Extract the actual object from the NTE (This is the dereference) */

            RetDesc = ((NAME_TABLE_ENTRY *) ObjDesc)->Object;

            /* Returning a pointer to the object, add another reference! */

            CmAddReference (RetDesc);
        }

        else
        {   /* This must be a reference object produced by the Index ASL operation */
            /* Must be an lvalue, make sure opcode is correct */

            if ((ObjDesc->Reference.OpCode != AML_IndexOp) &&
                (ObjDesc->Reference.OpCode != AML_RefOfOp))
            {
                DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2: DerefOf, invalid obj ref %p\n", ObjDesc));

                Status = AE_TYPE;
                goto Cleanup;
            }


            switch (ObjDesc->Reference.OpCode)
            {
            case AML_IndexOp:

                /*
                 * Supported target types for the Index operator are
                 * 1) A Buffer
                 * 2) A Package
                 */

                if (ObjDesc->Reference.TargetType == ACPI_TYPE_BufferField)
                {
                    /*
                     * The target is a buffer, we must create a new object that contains 
                     * one element of the buffer, the element pointed to by the index.  
                     * NOTE: index into a buffer is NOT a pointer to a sub-buffer of the
                     * main buffer, it is only a pointer to a single element (byte) of the buffer!
                     */
                    RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
                    if (!RetDesc)
                    {
                        Status = AE_NO_MEMORY;
                        goto Cleanup;
                    }

                    TmpDesc = ObjDesc->Reference.Object;
                    RetDesc->Number.Value = TmpDesc->Buffer.Pointer[ObjDesc->Reference.Offset];

                    /* TBD: (see below) Don't add an additional ref! */
                }

                else if (ObjDesc->Reference.TargetType == ACPI_TYPE_Package)
                {
                    /*
                     * The target is a package, we want to return the referenced element of the
                     * package.  We must add another reference to this object, however.
                     */

                    RetDesc = *(ObjDesc->Reference.Where);
                    if (!RetDesc)
                    {
                        /*
                         * We can't return a NULL dereferenced value.  This is an uninitialized package
                         * element and is thus a severe error.
                         */

                        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2: DerefOf, NULL package element obj %p\n", 
                                        ObjDesc));
                        Status = AE_AML_UNINITIALIZED_ELEMENT;
                        goto Cleanup;
                    }

                    CmAddReference (RetDesc);
                }
                
                else
                {
                    DEBUG_PRINT (ACPI_ERROR, ("AmlExecMonadic2: DerefOf, Unknown TargetType %X in obj %p\n", 
                                    ObjDesc->Reference.TargetType, ObjDesc));
                    Status = AE_AML_OPERAND_TYPE;
                    goto Cleanup;
                }

                break;


            case AML_RefOfOp:
           
                RetDesc = ObjDesc->Reference.Object;

                /* Add another reference to the object! */

                CmAddReference (RetDesc);
                break;
            }
        }

        break;
        

    default:

        DEBUG_PRINT (ACPI_ERROR, ( "AmlExecMonadic2: Internal error, unknown monadic opcode %02x\n",
                        Opcode));
        Status = AE_AML_BAD_OPCODE;
        goto Cleanup;
    }



Cleanup:

    if (ObjDesc)
    {
        CmRemoveReference (ObjDesc);
    }

    /* Delete return object on error */

    if (ACPI_FAILURE (Status) &&
        (RetDesc))
    {
        CmRemoveReference (RetDesc);
        RetDesc = NULL;
    }

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}
 
