
/*
 *  ./acpica/subsystem/interpreter/ieopexec.c - ACPI AML (p-code) execution - specific opcodes
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IEOPEXEC_C__

#include <acpi.h>
#include <parser.h>
#include <interp.h>
#include <amlcode.h>
#include <dispatch.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("ieopexec");




ACPI_STATUS
AmlExecFatal (    
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_OBJECT_INTERNAL    *TypeDesc;
    ACPI_OBJECT_INTERNAL    *CodeDesc;
    ACPI_OBJECT_INTERNAL    *ArgDesc;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AmlExecFatal");


    /* Resolve operands */

    Status = AmlResolveOperands (AML_FatalOp, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (AML_FatalOp), 3, "after AmlResolveOperands");

    /* Get operands */

    Status |= DsObjStackPopObject (&ArgDesc, WalkState);
    Status |= DsObjStackPopObject (&CodeDesc, WalkState);
    Status |= DsObjStackPopObject (&TypeDesc, WalkState);
    if (Status != AE_OK)
    {
        /* invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, (UINT16) AML_FatalOp, WALK_OPERANDS, 3);
        goto Cleanup;
    }



    /* DefFatal    :=  FatalOp FatalType   FatalCode   FatalArg    */


    DEBUG_PRINT (ACPI_INFO, ("FatalOp: Type %x Code %x Arg %x <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n",
                    TypeDesc->Number.Value, CodeDesc->Number.Value, ArgDesc->Number.Value));


    /* TBD: call OSD interface to notify OS of fatal error requiring shutdown! */


Cleanup:

    /* Free the operands */

    CmRemoveReference (ArgDesc);
    CmRemoveReference (CodeDesc);
    CmRemoveReference (TypeDesc);


    /* If we get back from the OS call, we might as well keep going. */

    DEBUG_PRINT (ACPI_ERROR, ("AmlExecFatal: FatalOp executed\n"));
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AmlExecIndex (
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *IdxDesc;
    ACPI_OBJECT_INTERNAL    *ResDesc;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_OBJECT_INTERNAL    *TmpDesc;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AmlExecIndex");


    /* Resolve operands */
    /* First operand can be either a package or a buffer */

    Status = AmlResolveOperands (AML_IndexOp, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (AML_IndexOp), 3, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&ResDesc, WalkState);
    Status |= DsObjStackPopObject (&IdxDesc, WalkState);
    Status |= DsObjStackPopObject (&ObjDesc, WalkState);
    if (Status != AE_OK)
    {
        /* Invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, (UINT16) AML_IndexOp, WALK_OPERANDS, 3);
        goto Cleanup;
    }


    /* Create the internal return object */

    RetDesc = CmCreateInternalObject (INTERNAL_TYPE_Reference);
    if (!RetDesc)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;
    }


    /*
     * At this point, the ObjDesc operand is either a Package or a Buffer
     */

    if (ObjDesc->Common.Type == ACPI_TYPE_Package)
    {
        /* Object to be indexed is a Package */

        if (IdxDesc->Number.Value >= ObjDesc->Package.Count)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlExecIndex: Index value out of range\n"));
            Status = AE_AML_PACKAGE_LIMIT;
            goto Cleanup;
        }

        if ((ResDesc->Common.Type == INTERNAL_TYPE_Reference) &&
            (ResDesc->Reference.OpCode == AML_ZeroOp))
        {
            /* 
             * There is no actual result descriptor (the ZeroOp Result descriptor is a placeholder),
             * so just delete the placeholder and return a reference to the package element
             */

            CmRemoveReference (ResDesc);
        }

        else
        {
            /* 
             * TBD - possible dangling reference: if the package vector changes
             * TBD - before this pointer is used, the results may be surprising.
             */

            /* 
             * Each element of the package is an internal object.  Get the one
             * we are after.
             */

            TmpDesc                         = ObjDesc->Package.Elements[IdxDesc->Number.Value];
            RetDesc->Reference.OpCode       = AML_IndexOp;
            RetDesc->Reference.TargetType   = TmpDesc->Common.Type;
            RetDesc->Reference.Object       = TmpDesc;

            Status = AmlExecStore (RetDesc, ResDesc);
            RetDesc->Reference.Object       = NULL;
        }

        /*
         * The local return object must always be a reference to the package element,
         * not the element itself.
         */
        RetDesc->Reference.OpCode       = AML_IndexOp;
        RetDesc->Reference.TargetType   = ACPI_TYPE_Package;
        RetDesc->Reference.Where        = &ObjDesc->Package.Elements[IdxDesc->Number.Value];
    }

    else
    {
        /* Object to be indexed is a Buffer */

        if (IdxDesc->Number.Value >= ObjDesc->Buffer.Length)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlExecIndex: Index value out of range\n"));
            Status = AE_AML_BUFFER_LIMIT;
            goto Cleanup;
        }

        /* 
         * TBD - possible dangling reference: if the package vector changes
         * TBD - before this pointer is used, the results may be surprising.
         */
        RetDesc->Reference.OpCode       = AML_IndexOp;
        RetDesc->Reference.TargetType   = ACPI_TYPE_BufferField;
        RetDesc->Reference.Object       = ObjDesc;
        RetDesc->Reference.Offset       = IdxDesc->Number.Value;

        Status = AmlExecStore (RetDesc, ResDesc);
    }


Cleanup:

    /* Always delete operands */

    CmRemoveReference (ObjDesc);
    CmRemoveReference (IdxDesc);

    /* Delete return object on error */

    if (ACPI_FAILURE (Status))
    {
        CmRemoveReference (ResDesc);

        if (RetDesc)
        {
            CmRemoveReference (RetDesc);
            RetDesc = NULL;
        }
    }

    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlExecMatch (
    ACPI_WALK_STATE         *WalkState,
    ACPI_OBJECT_INTERNAL    **ReturnDesc)
{
    ACPI_OBJECT_INTERNAL    *PkgDesc;
    ACPI_OBJECT_INTERNAL    *Op1Desc;
    ACPI_OBJECT_INTERNAL    *V1Desc;
    ACPI_OBJECT_INTERNAL    *Op2Desc;
    ACPI_OBJECT_INTERNAL    *V2Desc;
    ACPI_OBJECT_INTERNAL    *StartDesc;
    ACPI_OBJECT_INTERNAL    *RetDesc = NULL;
    ACPI_STATUS             Status;
    UINT32                  Index;
    UINT32                  MatchValue = (UINT32) -1;


    FUNCTION_TRACE ("AmlExecMatch");


    /* Resolve all operands */

    Status = AmlResolveOperands (AML_MatchOp, WALK_OPERANDS);
    DUMP_OPERANDS (WALK_OPERANDS, IMODE_Execute, PsGetOpcodeName (AML_MatchOp), 6, "after AmlResolveOperands");

    /* Get all operands */

    Status |= DsObjStackPopObject (&StartDesc, WalkState);
    Status |= DsObjStackPopObject (&V2Desc, WalkState);
    Status |= DsObjStackPopObject (&Op2Desc, WalkState);
    Status |= DsObjStackPopObject (&V1Desc, WalkState);
    Status |= DsObjStackPopObject (&Op1Desc, WalkState);
    Status |= DsObjStackPopObject (&PkgDesc, WalkState);

    if (Status != AE_OK)
    {
        /* invalid parameters on object stack  */

        AmlAppendOperandDiag (_THIS_MODULE, __LINE__, (UINT16) AML_MatchOp, WALK_OPERANDS, 6);
        goto Cleanup;
    }

    /* Validate match comparison sub-opcodes */
    
    if ((Op1Desc->Number.Value > MAX_MATCH_OPERATOR) || 
        (Op2Desc->Number.Value > MAX_MATCH_OPERATOR))
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMatch: operation encoding out of range\n"));
        Status = AE_AML_OPERAND_VALUE;
        goto Cleanup;
    }

    Index = StartDesc->Number.Value;
    if (Index >= (UINT32) PkgDesc->Package.Count)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecMatch: start position value out of range\n"));
        Status = AE_AML_PACKAGE_LIMIT;
        goto Cleanup;
    }

    RetDesc = CmCreateInternalObject (ACPI_TYPE_Number);
    if (!RetDesc)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;

    }

    /* 
     * Examine each element until a match is found.  Within the loop,
     * "continue" signifies that the current element does not match
     * and the next should be examined.
     * Upon finding a match, the loop will terminate via "break" at
     * the bottom.  If it terminates "normally", MatchValue will be -1
     * (its initial value) indicating that no match was found.  When
     * returned as a Number, this will produce the Ones value as specified.
     */

    for ( ; Index < PkgDesc->Package.Count; ++Index)
    {
        /* 
         * Treat any NULL or non-numeric elements as non-matching.
         * XXX - if an element is a Name, should we examine its value?
         */
        if (!PkgDesc->Package.Elements[Index] ||
            ACPI_TYPE_Number != PkgDesc->Package.Elements[Index]->Common.Type)
        {
            continue;
        }

        /* 
         * Within these switch statements:
         *      "break" (exit from the switch) signifies a match;
         *      "continue" (proceed to next iteration of enclosing
         *          "for" loop) signifies a non-match.
         */
        switch (Op1Desc->Number.Value)
        {

        case MATCH_MTR:   /* always true */

            break;


        case MATCH_MEQ:   /* true if equal   */

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 != V1Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MLE:   /* true if less than or equal  */

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 > V1Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MLT:   /* true if less than   */

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 >= V1Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MGE:   /* true if greater than or equal   */

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 < V1Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MGT:   /* true if greater than    */

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 <= V1Desc->Number.Value)
            {
                continue;
            }
            break;


        default:    /* undefined   */

            continue;
        }
 
        
        switch(Op2Desc->Number.Value)
        {

        case MATCH_MTR:

            break;


        case MATCH_MEQ:

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 != V2Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MLE:

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 > V2Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MLT:

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 >= V2Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MGE:

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 < V2Desc->Number.Value)
            {
                continue;
            }
            break;


        case MATCH_MGT:

            if (PkgDesc->Package.Elements[Index]->Number.Value
                 <= V2Desc->Number.Value)
            {
                continue;
            }
            break;


        default:

            continue;
        }

        /* Match found: exit from loop */
        
        MatchValue = Index;
        break;
    }

    /* MatchValue is the return value */

    RetDesc->Number.Value = MatchValue;


Cleanup:

    /* Free the operands */

    CmRemoveReference (StartDesc);
    CmRemoveReference (V2Desc);
    CmRemoveReference (Op2Desc);
    CmRemoveReference (V1Desc);
    CmRemoveReference (Op1Desc);
    CmRemoveReference (PkgDesc);

    
    /* Delete return object on error */

    if (ACPI_FAILURE (Status) &&
        (RetDesc))
    {
        CmRemoveReference (RetDesc);
        RetDesc = NULL;
    }


    /* Set the return object and exit */

    *ReturnDesc = RetDesc;
    return_ACPI_STATUS (Status);
}
