
/*
 *  ./acpica/subsystem/interpreter/ieprep.c - ACPI AML (p-code) execution - field prep utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IEPREP_C__

#include <acpi.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("ieprep");




UINT32
AmlDecodeFieldAccessType (
    UINT32                  Access)
{
 
    switch (Access)
    {
    case ACCESS_AnyAcc:
        return 8;
        break;

    case ACCESS_ByteAcc:
        return 8;
        break;

    case ACCESS_WordAcc:
        return 16;
        break;

    case ACCESS_DWordAcc:
        return 32;
        break;

    default:
        /* Invalid field access type */

        DEBUG_PRINT (ACPI_ERROR, ("AmlDecodeFieldAccessType: Unknown field access type %x\n",
                        Access));
        return 0;
    }
}




ACPI_STATUS
AmlPrepCommonFieldObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc,
    UINT8                   FieldFlags, 
    UINT8                   FieldAttribute,
    UINT32                  FieldPosition, 
    UINT32                  FieldLength)
{
    UINT32                  Granularity;


    FUNCTION_TRACE ("AmlPrepCommonFieldObject");


    /*
     * Note: the structure being initialized is the
     * ACPI_COMMON_FIELD_INFO;  Therefore, we can just use the Field union to
     * access this common area.  No structure fields outside of the common area
     * are initialized by this procedure.
     */

    /* Decode the FieldFlags */

    ObjDesc->Field.Access           = (UINT8) ((FieldFlags & ACCESS_TYPE_MASK) >> ACCESS_TYPE_SHIFT);
    ObjDesc->Field.LockRule         = (UINT8) ((FieldFlags & LOCK_RULE_MASK) >> LOCK_RULE_SHIFT);
    ObjDesc->Field.UpdateRule       = (UINT8) ((FieldFlags & UPDATE_RULE_MASK) >> UPDATE_RULE_SHIFT);

    /* Other misc fields */

    ObjDesc->Field.Length           = (UINT16) FieldLength;
    ObjDesc->Field.AccessAttribute  = FieldAttribute;

    /* Decode the access type so we can compute offsets */

    Granularity = AmlDecodeFieldAccessType (ObjDesc->Field.Access);
    if (!Granularity)
    {
        return_ACPI_STATUS (AE_AML_OPERAND_VALUE);
    }

    /* Access granularity based fields */

    ObjDesc->Field.Granularity      = (UINT8) Granularity;
    ObjDesc->Field.BitOffset        = (UINT8) (FieldPosition % Granularity);
    ObjDesc->Field.Offset           = (UINT32) FieldPosition / Granularity;


    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AmlPrepDefFieldValue (
    NAME_TABLE_ENTRY        *ThisEntry,
    ACPI_HANDLE             Region, 
    UINT8                   FieldFlags,
    UINT8                   FieldAttribute,
    UINT32                  FieldPosition, 
    UINT32                  FieldLength)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    INT32                   Type;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AmlPrepDefFieldValue");


    /* Parameter validation */

    if (!Region)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlPrepDefFieldValue: null Region\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }

    Type = NsGetType (Region);
    if (Type != ACPI_TYPE_Region)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlPrepDefFieldValue: Needed Region, found %d %s\n",
                        Type, CmGetTypeName (Type)));
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }

    /* Allocate a new object */

    ObjDesc = CmCreateInternalObject (INTERNAL_TYPE_DefField);
    if (!ObjDesc)
    {   
        return_ACPI_STATUS (AE_NO_MEMORY);
    }


    /* ObjDesc and Region valid */

    DUMP_OPERANDS ((ACPI_OBJECT_INTERNAL **) &ThisEntry, IMODE_Execute, "AmlPrepDefFieldValue", 1, "case DefField");
    DUMP_OPERANDS ((ACPI_OBJECT_INTERNAL **) &Region, IMODE_Execute, "AmlPrepDefFieldValue", 1, "case DefField");

    /* Initialize areas of the object that are common to all fields */

    Status = AmlPrepCommonFieldObject (ObjDesc, FieldFlags, FieldAttribute, FieldPosition, FieldLength);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Initialize areas of the object that are specific to this field type */

    ObjDesc->Field.Container = NsGetAttachedObject (Region);

    /* An additional reference for the container */

    CmAddReference (ObjDesc->Field.Container);


    /* Debug info */

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepDefFieldValue: bitoff=%X off=%X gran=%X\n",
                    ObjDesc->Field.BitOffset, ObjDesc->Field.Offset, ObjDesc->Field.Granularity));

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepDefFieldValue: set nte %p (%4.4s) val = %p\n",
                    ThisEntry, &(ThisEntry->Name), ObjDesc));

    DUMP_STACK_ENTRY (ObjDesc);
    DUMP_ENTRY (Region, ACPI_INFO);
    DEBUG_PRINT (ACPI_INFO, ("\t%p \n", ObjDesc->Field.Container));
    if (ObjDesc->Field.Container)
    {
        DUMP_STACK_ENTRY (ObjDesc->Field.Container);
    }
    DEBUG_PRINT (ACPI_INFO, ("============================================================\n"));

    /* 
     * Store the constructed descriptor (ObjDesc) into the nte whose
     * handle is on TOS, preserving the current type of that nte.
     */
    Status = NsAttachObject ((ACPI_HANDLE) ThisEntry, ObjDesc,
                                (UINT8) NsGetType ((ACPI_HANDLE) ThisEntry));

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlPrepBankFieldValue (
    NAME_TABLE_ENTRY        *ThisEntry,
    ACPI_HANDLE             Region, 
    ACPI_HANDLE             BankReg, 
    UINT32                  BankVal,
    UINT8                   FieldFlags, 
    UINT8                   FieldAttribute,
    UINT32                  FieldPosition, 
    UINT32                  FieldLength)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    INT32                   Type;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AmlPrepBankFieldValue");


    /* Parameter validation */

    if (!Region)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlPrepBankFieldValue: null Region\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }

    Type = NsGetType (Region);
    if (Type != ACPI_TYPE_Region)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlPrepBankFieldValue: Needed Region, found %d %s\n",
                        Type, CmGetTypeName (Type)));
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }

    /* Allocate a new object */

    ObjDesc = CmCreateInternalObject (INTERNAL_TYPE_BankField);
    if (!ObjDesc)
    {   
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /*  ObjDesc and Region valid    */

    DUMP_OPERANDS ((ACPI_OBJECT_INTERNAL **) &ThisEntry, IMODE_Execute, "AmlPrepBankFieldValue", 1, "case BankField");
    DUMP_OPERANDS ((ACPI_OBJECT_INTERNAL **) &Region, IMODE_Execute, "AmlPrepBankFieldValue", 1, "case BankField");

    /* Initialize areas of the object that are common to all fields */

    Status = AmlPrepCommonFieldObject (ObjDesc, FieldFlags, FieldAttribute, FieldPosition, FieldLength);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Initialize areas of the object that are specific to this field type */

    ObjDesc->BankField.Value        = BankVal;
    ObjDesc->BankField.Container    = NsGetAttachedObject (Region);
    ObjDesc->BankField.BankSelect   = NsGetAttachedObject (BankReg);

    /* An additional reference for the container and bank select */
    /* TBD: is "BankSelect" ever a real internal object?? */

    CmAddReference (ObjDesc->BankField.Container);
    CmAddReference (ObjDesc->BankField.BankSelect);

    /* Debug info */

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepBankFieldValue: bitoff=%X off=%X gran=%X\n",
                    ObjDesc->BankField.BitOffset, ObjDesc->BankField.Offset, ObjDesc->Field.Granularity));

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepBankFieldValue: set nte %p (%4.4s) val = %p\n",
                    ThisEntry, &(ThisEntry->Name), ObjDesc));
    
    DUMP_STACK_ENTRY (ObjDesc);
    DUMP_ENTRY (Region, ACPI_INFO);
    DUMP_ENTRY (BankReg, ACPI_INFO);
    DEBUG_PRINT (ACPI_INFO, ("============================================================\n"));

    /* 
     * Store the constructed descriptor (ObjDesc) into the nte whose
     * handle is on TOS, preserving the current type of that nte.
     */
    Status = NsAttachObject ((ACPI_HANDLE) ThisEntry, ObjDesc,
                                (UINT8) NsGetType ((ACPI_HANDLE) ThisEntry));

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlPrepIndexFieldValue (
    NAME_TABLE_ENTRY        *ThisEntry,
    ACPI_HANDLE             IndexReg, 
    ACPI_HANDLE             DataReg,
    UINT8                   FieldFlags, 
    UINT8                   FieldAttribute,
    UINT32                  FieldPosition, 
    UINT32                  FieldLength)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AmlPrepIndexFieldValue");


    /* Parameter validation */

    if (!IndexReg || !DataReg)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlPrepIndexFieldValue: null handle\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }

    /* Allocate a new object descriptor */

    ObjDesc = CmCreateInternalObject (INTERNAL_TYPE_IndexField);
    if (!ObjDesc)
    {   
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Initialize areas of the object that are common to all fields */

    Status = AmlPrepCommonFieldObject (ObjDesc, FieldFlags, FieldAttribute, FieldPosition, FieldLength);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Initialize areas of the object that are specific to this field type */

    ObjDesc->IndexField.Value       = (UINT32) (FieldPosition / ObjDesc->Field.Granularity);
    ObjDesc->IndexField.Index       = IndexReg;
    ObjDesc->IndexField.Data        = DataReg;

    /* Debug info */

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepIndexFieldValue: bitoff=%X off=%X gran=%X\n",
                    ObjDesc->IndexField.BitOffset, ObjDesc->IndexField.Offset, ObjDesc->Field.Granularity));

    DEBUG_PRINT (ACPI_INFO, ("AmlPrepIndexFieldValue: set nte %p (%4.4s) val = %p\n",
                    ThisEntry, &(ThisEntry->Name), ObjDesc));

    DUMP_STACK_ENTRY (ObjDesc);
    DUMP_ENTRY (IndexReg, ACPI_INFO);
    DUMP_ENTRY (DataReg, ACPI_INFO);
    DEBUG_PRINT (ACPI_INFO, ("============================================================\n"));

    /* 
     * Store the constructed descriptor (ObjDesc) into the nte whose
     * handle is on TOS, preserving the current type of that nte.
     */
    Status = NsAttachObject ((ACPI_HANDLE) ThisEntry, ObjDesc,
                                (UINT8) NsGetType ((ACPI_HANDLE) ThisEntry));

    return_ACPI_STATUS (Status);
}

