/*
 *  ./acpica/subsystem/interpreter/ieregion.c - ACPI default OpRegion (address space) handlers
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __IEREGION_C__

#include <acpi.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <hardware.h>
#include <events.h>
#include <methods.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("ieregion");



ACPI_STATUS
AmlSystemMemorySpaceHandler (
    UINT32                  Function,
    UINT32                  Address,
    UINT32                  BitWidth,
    UINT32                  *Value,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    void                    *PhysicalAddrPtr = NULL;


    FUNCTION_TRACE ("AmlSystemMemorySpaceHandler");


    /* Decode the function parameter */

    switch (Function)
    {

    case ADDRESS_SPACE_READ:

        /* 
         * TBD: This may be too high an overhead to do every time.
         * Probably should have a mapping cached.
         */

        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
            ("Read(%d width) Address:0x%08x\n", BitWidth, Address));

        Status = OsdMapMemory ((void *) Address, 4, &PhysicalAddrPtr);

  if (ACPI_FAILURE (Status))
  {
      return_ACPI_STATUS (Status);
  }

        switch (BitWidth)
        {
        /* System memory width */

        case 8:
            *Value = (UINT32)* (UINT8 *) PhysicalAddrPtr;
            break;

        case 16:
            *Value = (UINT32)* (UINT16 *) PhysicalAddrPtr;
            break;

        case 32:
            *Value = * (UINT32 *) PhysicalAddrPtr;
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR,
                    ("AmlSystemMemorySpaceHandler: Invalid SystemMemory width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;
        }

        OsdUnMapMemory (PhysicalAddrPtr, 4);
        break;


    case ADDRESS_SPACE_WRITE:


        /* TBD: This may be too high an overhead to do every time.
         * Probably should have a mapping cached.
         */
        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
            ("Write(%d width) Address:0x%08x Value 0x%08x\n", BitWidth, Address, *Value));

        Status = OsdMapMemory ((void *) Address, 4, &PhysicalAddrPtr);
                
  if (ACPI_FAILURE (Status))
  {
      return_ACPI_STATUS (Status);
  }

        switch (BitWidth)
        {
        case 8:
            *(UINT8 *) PhysicalAddrPtr = (UINT8) *Value;
            break;

        case 16:
            *(UINT16 *) PhysicalAddrPtr = (UINT16) *Value;
            break;

        case 32:
            *(UINT32 *) PhysicalAddrPtr = *Value;
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlSystemMemorySpaceHandler: Invalid SystemMemory width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;
        }

        OsdUnMapMemory (PhysicalAddrPtr, 4);
        break;


    default:
        Status = AE_BAD_PARAMETER;
        break;
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlSystemIoSpaceHandler (
    UINT32                  Function,
    UINT32                  Address,
    UINT32                  BitWidth,
    UINT32                  *Value,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("AmlSystemIoSpaceHandler");


    /* Decode the function parameter */

    switch (Function)
    {

    case ADDRESS_SPACE_READ:

        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
            ("Read(%d width) Address:0x%08x\n", BitWidth, Address));

        switch (BitWidth)
        {
        /* I/O Port width */

        case 8:
            *Value = (UINT32) OsdIn8 (Address);
            break;

        case 16:
            *Value = (UINT32) OsdIn16 (Address);
            break;

        case 32:
            *Value = OsdIn32 (Address);
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR,
                    ("AmlSystemIoSpaceHandler: Invalid SystemIO width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;
        }

        break;


    case ADDRESS_SPACE_WRITE:

        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
            ("Write(%d width) Address:0x%08x Value 0x%08x\n", BitWidth, Address, *Value));

        switch (BitWidth)
        {
        /* I/O Port width */
        case 8:
            OsdOut8 (Address, (UINT8) *Value);
            break;

        case 16:
            OsdOut16 (Address, (UINT16) *Value);
            break;

        case 32:
            OsdOut32 (Address, *Value);
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlSystemIoSpaceHandler: Invalid SystemIO width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;
        }

        break;


    default:
        Status = AE_BAD_PARAMETER;
        break;
    }

    return_ACPI_STATUS (Status);
}


ACPI_STATUS
AmlPciConfigSpaceHandler (
    UINT32                  Function,
    UINT32                  Address,
    UINT32                  BitWidth,
    UINT32                  *Value,
    void                    *Context)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  PciBus;
    UINT32                  DevFunc;
    UINT8                   PciReg;
    PCI_HANDLER_CONTEXT    *PCIContext;


    FUNCTION_TRACE ("AmlPciConfigSpaceHandler");

    /*
     *  The arguments to Osd(Read|Write)PciCfg(Byte|Word|Dword) are:
     *
     *  SegBus - 0xSSSSBBBB     - SSSS is the PCI bus segment
     *                            BBBB is the PCI bus number
     *
     *  DevFunc - 0xDDDDFFFF    - DDDD is the PCI device number
     *                            FFFF is the PCI device function number
     *
     *  RegNum - Config space register must be < 40h
     *
     *  Value - input value for write, output for read
     *
     */

    PCIContext = (PCI_HANDLER_CONTEXT *) Context;

    PciBus = LOWORD(PCIContext->Seg) << 16;
    PciBus |= LOWORD(PCIContext->Bus);

    DevFunc = PCIContext->DevFunc;

    PciReg  = (UINT8) Address;

    switch (Function)
    {

    case ADDRESS_SPACE_READ:

        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
                        ("R%d S(%04x) B(%04x) DF(%08x) R(%04x)\n", BitWidth,
                        PCIContext->Seg,PCIContext->Bus,PCIContext->DevFunc, PciReg));

        *Value  = 0;

        switch (BitWidth)
        {
        /* PCI Register width */

        case 8:
            Status = OsdReadPciCfgByte (PciBus, DevFunc, PciReg, (UINT8 *) Value);
            break;

        case 16:
            Status = OsdReadPciCfgWord (PciBus, DevFunc, PciReg, (UINT16 *) Value);
            break;

        case 32:
            Status = OsdReadPciCfgDword (PciBus, DevFunc, PciReg, Value);
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR,
                    ("AmlPciConfigSpaceHandler: Invalid PCIConfig width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;

        } /* Switch bitWidth */

        break;


    case ADDRESS_SPACE_WRITE:

        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
                        ("R%d S(%04x) B(%04x) DF(%08x) R(%04x)\n", BitWidth,
                        PCIContext->Seg,PCIContext->Bus,PCIContext->DevFunc, PciReg));


        DEBUG_PRINT ((TRACE_OPREGION | VERBOSE_INFO),
                        ("W%d S(%04x) B(%04x) DF(%08x) R(%04x) D(%08x)\n", BitWidth,
                        PCIContext->Seg,PCIContext->Bus,PCIContext->DevFunc,
                        PciReg,*Value));

        switch (BitWidth)
        {
        /* PCI Register width */

        case 8:
            Status = OsdWritePciCfgByte (PciBus, DevFunc, PciReg, *(UINT8 *) Value);
            break;

        case 16:
            Status = OsdWritePciCfgWord (PciBus, DevFunc, PciReg, *(UINT16 *) Value);
            break;

        case 32:
            Status = OsdWritePciCfgDword (PciBus, DevFunc, PciReg, *Value);
            break;

        default:
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlPciConfigSpaceHandler: Invalid PCIConfig width %d\n", BitWidth));
            Status = AE_AML_OPERAND_VALUE;

        } /* Switch bitWidth */

        break;


    default:

        Status = AE_BAD_PARAMETER;
        break;

    }

    return_ACPI_STATUS (Status);
}

