
/*
 *  ./acpica/subsystem/interpreter/iesystem.c - Interface to OS services
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IESYSTEM_C__

#include <acpi.h>
#include <interp.h>
#include <namesp.h>
#include <hardware.h>
#include <events.h>

#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iesystem");


UINT16 
OsThreadId (void)
{
    return (1);
}



ACPI_STATUS
OsLocalWaitSemaphore (
    ACPI_HANDLE             Semaphore,
    UINT32                  Timeout)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("OsLocalWaitSemaphore");

    Status = OsdWaitSemaphore (Semaphore, 1, 0);
    if (ACPI_SUCCESS (Status))
    {
        return_ACPI_STATUS (Status);
    }

    if (Status == AE_TIME)
    {
        /* We must wait, so unlock the interpreter */

        AmlExitInterpreter ();

        Status = OsdWaitSemaphore (Semaphore, 1, Timeout);

        /* Reacquire the interpreter */

        AmlEnterInterpreter ();

        DEBUG_PRINT (TRACE_EXEC, ("*** Thread awake and inside interpreter after blocking, %s\n", CmFormatException (Status)));
    }

    return_ACPI_STATUS (Status);
}



void
OsDoStall (
    UINT32                  HowLong)
{

    if (HowLong > 1000) /* 1 millisecond */
    {
        /* Since this thread will sleep, we must release the interpreter */

        AmlExitInterpreter ();

        OsdSleepUsec (HowLong);

        /* And now we must get the interpreter again */

        AmlEnterInterpreter ();
    }

    else
    {
        OsdSleepUsec (HowLong);
    }
}



void
OsDoSuspend (
    UINT32                  HowLong)
{
    /* Since this thread will sleep, we must release the interpreter */

    AmlExitInterpreter ();

    OsdSleep ((UINT16) (HowLong / (UINT32) 1000), (UINT16) (HowLong % (UINT32) 1000));

    /* And now we must get the interpreter again */

    AmlEnterInterpreter ();
}



ACPI_STATUS
OsAcquireMutex (
    ACPI_OBJECT_INTERNAL    *TimeDesc, 
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("OsAcquireMutex", ObjDesc);

    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Support for the _GL_ Mutex object -- go get the global lock
     */

    if (ObjDesc->Mutex.Semaphore == Gbl_GlobalLockSemaphore)
    {
        Status = EvAcquireGlobalLock ();
        return_ACPI_STATUS (Status);
    }

    Status = OsLocalWaitSemaphore (ObjDesc->Mutex.Semaphore, TimeDesc->Number.Value);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
OsReleaseMutex (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("OsReleaseMutex");

    
    if (!ObjDesc)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Support for the _GL_ Mutex object -- release the global lock
     */
    if (ObjDesc->Mutex.Semaphore == Gbl_GlobalLockSemaphore)
    {
        EvReleaseGlobalLock ();
        return_ACPI_STATUS (AE_OK);
    }

    Status = OsdSignalSemaphore (ObjDesc->Mutex.Semaphore, 1);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
OsSignalEvent (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("OsSignalEvent");

    
    if (ObjDesc)
    {
        Status = OsdSignalSemaphore (ObjDesc->Event.Semaphore, 1);
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
OsWaitEvent (
    ACPI_OBJECT_INTERNAL    *TimeDesc, 
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("OsWaitEvent");


    if (ObjDesc)
    {
        Status = OsLocalWaitSemaphore (ObjDesc->Event.Semaphore, TimeDesc->Number.Value);
    }

  
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
OsResetEvent (
    ACPI_OBJECT_INTERNAL    *ObjDesc)
{
    ACPI_STATUS             Status = AE_OK;
    void                    *TempSemaphore;


    /* We are going to simply delete the existing semaphore and create a new one! */

    Status = OsdCreateSemaphore (0, &TempSemaphore);
    if (ACPI_SUCCESS (Status))
    {
        OsdDeleteSemaphore (ObjDesc->Mutex.Semaphore);
        ObjDesc->Mutex.Semaphore = TempSemaphore;
    }

    return (Status);
}

