
/*
 *  ./acpica/subsystem/interpreter/iresnte.c - AML Interpreter object resolution
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IRESNTE_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iresnte");



ACPI_STATUS
AmlResolveEntryToValue (
    NAME_TABLE_ENTRY        **StackPtr)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_INTERNAL    *ValDesc = NULL;
    ACPI_OBJECT_INTERNAL    *ObjDesc = NULL;
    NAME_TABLE_ENTRY        *StackEntry;
    UINT8                   *AmlPointer = NULL;
    ACPI_OBJECT_TYPE        EntryType;
    BOOLEAN                 Locked;
    BOOLEAN                 AttachedAmlPointer = FALSE;
    UINT8                   AmlOpcode = 0;
    UINT32                  TempVal;
    ACPI_OBJECT_TYPE        ObjectType;


    FUNCTION_TRACE ("AmlResolveEntryToValue");

    StackEntry = *StackPtr;


    /* 
     * The stack pointer is a "Direct name ptr", and points to a
     * a NAME_TABLE_ENTRY (nte).  Get the pointer that is attached to
     * the nte.
     */

    ValDesc     = NsGetAttachedObject ((ACPI_HANDLE) StackEntry);
    EntryType   = NsGetType ((ACPI_HANDLE) StackEntry);

    DEBUG_PRINT (TRACE_EXEC, ("AmlResolveEntryToValue: Entry=%p ValDesc=%p Type=%X\n", 
                    StackEntry, ValDesc, EntryType));


    /*
     * The ValDesc attached to the NTE can be either:
     * 1) An internal ACPI object 
     * 2) A pointer into the AML stream (into one of the ACPI system tables)
     */

    if (TbSystemTablePointer (ValDesc))
    {
        AttachedAmlPointer = TRUE;
        AmlOpcode = *((UINT8 *) ValDesc);
        AmlPointer = ((UINT8 *) ValDesc) + 1;

        DEBUG_PRINT (TRACE_EXEC, ("AmlResolveEntryToValue: Unparsed AML: %p Len=%X\n", 
                        AmlOpcode, AmlPointer));
    }


    /*
     * Action is based on the type of the NTE, which indicates the type
     * of the attached object or pointer
     */
    switch (EntryType)
    {

    case ACPI_TYPE_Package:

        /* 
         * ValDesc should point to either an ACPI_OBJECT_INTERNAL of
         * type Package, or an initialization in the AML stream.
         */
        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null ValDesc\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }


        if (AttachedAmlPointer)
        {
#if defined  _RPARSER
            if (AML_PackageOp == AmlOpcode)
            {
                /* 
                 * The value pointer in the name table entry
                 * points to a package definition in the AML stream.
                 * Convert it to an object.
                 */

                if (AE_OK != (Status = PsxObjStackPush ()))             /* ObjStack */
                {
                    return_ACPI_STATUS (Status);
                }

                if (AE_OK == (Status = AmlPkgPushExec (AmlPointer, 0L)) &&          /* PkgStack */
                    AE_OK == (Status = AmlDoPkg (ACPI_TYPE_Package, IMODE_Execute)) &&
                    AE_OK == (Status = AmlPkgPopExec ()))                           /* PkgStack */
                {
                    NsAttachObject ((ACPI_HANDLE) StackEntry, PsxObjStackGetValue (0),
                                    (UINT8) ACPI_TYPE_Package);

                    /* Refresh local value pointer to reflect newly set value */
                
                    ValDesc = NsGetAttachedObject ((ACPI_HANDLE) StackEntry);
                    PsxObjStackPop (1);
                }
            
                else
                {
                    PsxObjStackPop (1);
                    return_ACPI_STATUS (Status);
                }
            }

            else
            {
                /* Aml Opcode is not a package op */

                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Not a PackageOp opcode (%x)\n",
                                AmlOpcode));
                return_ACPI_STATUS (AE_AML_BAD_OPCODE);
            }
#else
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Unparsed Packages not supported!\n"));
            return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
#endif
        }
        
        /* ValDesc is an internal object in all cases by the time we get here */

        if (!ValDesc || (ACPI_TYPE_Package != ValDesc->Common.Type))
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - Bad package value\n"));
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }

        /* Return an additional reference to the object */

        ObjDesc = ValDesc;
        CmAddReference (ObjDesc);
        break;


    case ACPI_TYPE_Buffer:

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null Buffer ValuePtr\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }



       if (AttachedAmlPointer)
        {
#if defined _RPARSER
            if (AML_BufferOp == AmlOpcode)
            {
                /* 
                 * The value pointer in the name table entry
                 * points to a buffer definition in the AML stream.
                 * Convert it to an object.
                 */
                if (AE_OK != (Status = PsxObjStackPush ()))                /* ObjStack */
                {
                    return_ACPI_STATUS (Status);
                }

                if (AE_OK == (Status = AmlPkgPushExec (AmlPointer, 0L)) &&          /* PkgStack */
                    AE_OK == (Status = AmlDoPkg (ACPI_TYPE_Buffer, IMODE_Execute)) &&
                    AE_OK == (Status = AmlPkgPopExec ()))                           /* PkgStack */
                {
                    NsAttachObject ((ACPI_HANDLE) StackEntry, PsxObjStackGetValue (0),
                                    (UINT8) ACPI_TYPE_Buffer);
                
                    /* Refresh local value pointer to reflect newly set value */
                
                    ValDesc = NsGetAttachedObject ((ACPI_HANDLE) StackEntry);
                    PsxObjStackPop (1);
                }
            
                else
                {
                    PsxObjStackPop (1);
                    return_ACPI_STATUS (Status);
                }
            }

            else
            {
                /* Aml Opcode is not a buffer op */

                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Not a BufferOp opcode (%x)\n",
                                AmlOpcode));
                return_ACPI_STATUS (AE_AML_BAD_OPCODE);
            }
#else
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Unparsed Buffers not supported!\n"));
            return_ACPI_STATUS (AE_NOT_IMPLEMENTED);
#endif
        }



        /* ValDesc is an internal object in all cases by the time we get here */

        if (!ValDesc || (ACPI_TYPE_Buffer != ValDesc->Common.Type))
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Bad buffer value\n"));
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }
        
        /* Return an additional reference to the object */

        ObjDesc = ValDesc;
        CmAddReference (ObjDesc);

        DEBUG_PRINT (TRACE_BFIELD,
                    ("AmlResolveEntryToValue: New Buffer descriptor seq# %ld @ %p \n",
                    ObjDesc->Buffer.Sequence, ObjDesc));
        break;


    case ACPI_TYPE_String:
      
        if (AttachedAmlPointer)
        {
            /* Allocate a new string object */

            ObjDesc = CmCreateInternalObject (ACPI_TYPE_String);
            if (!ObjDesc)
            {   
                /* Descriptor allocation failure */

                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            /* Init the internal object */

            ObjDesc->String.Pointer = (char *) AmlPointer;
            ObjDesc->String.Length = STRLEN (AmlPointer);
        }

        else
        {
            if (ACPI_TYPE_String != ValDesc->Common.Type)
            {
                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - Bad string value\n"));
                return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
            }

            /* Return an additional reference to the object */

            ObjDesc = ValDesc;
            CmAddReference (ObjDesc);
        }

        break;


    case ACPI_TYPE_Number:

        DEBUG_PRINT (TRACE_EXEC, ("AmlResolveEntryToValue: case Number \n"));

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null Number ValuePtr\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }

        
        /*
         * A ACPI_TYPE_Number can be either an object or an AML pointer
         */

        if (AttachedAmlPointer)
        {
            /* 
             * The attachment points into the AML stream, get the number from there.
             * The actual number is based upon the AML opcode
             * Note: WordOp and DWordOp will not work properly if the
             *       processor's endianness does not match the AML's.
             */

            switch (AmlOpcode)
            {

            case AML_ZeroOp:

                TempVal = 0;
                break;


            case AML_OneOp:

                TempVal = 1;
                break;


            case AML_OnesOp:

                TempVal = 0xFFFFFFFF;
                break;


            case AML_ByteOp:

                TempVal = (UINT32) ((UINT8 *) ValDesc)[1];
                break;


            case AML_WordOp:

                TempVal = (UINT32) *(UINT16 *)&((UINT8 *) ValDesc)[1];
                break;


            case AML_DWordOp:

                TempVal = *(UINT32 *)&((UINT8 *) ValDesc)[1];
                break;


            default:

                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveToValue/Number: Internal error - expected AML number, found %02x\n",
                                AmlOpcode));

                return_ACPI_STATUS (AE_AML_BAD_OPCODE);

            } /* switch */

            
            /* Create and initialize a new object */

            ObjDesc = CmCreateInternalObject (ACPI_TYPE_Number);
            if (!ObjDesc)
            {   
                /* Descriptor allocation failure  */
                
                return_ACPI_STATUS (AE_NO_MEMORY);
            }
            
            ObjDesc->Number.Value = TempVal;
        }
        
        else
        {
            /* The NTE has an attached internal object, make sure it's a number */

            if (ACPI_TYPE_Number != ValDesc->Common.Type)
            {
                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveToValue/Number: Attached obj %p not a Number, type 0x%X\n", ValDesc, ValDesc->Common.Type));
                return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
            }

            /* Return an additional reference to the object */

            ObjDesc = ValDesc;
            CmAddReference (ObjDesc);
        }

        break;


    case INTERNAL_TYPE_DefField:

        /*
         * TBD: Is this the correct solution?
         *
         * This section was extended to convert to generic buffer if 
         *  the return length is greater than 32 bits, but still allows
         *  for returning a type Number for smaller values because the
         *  caller can then apply arithmetic operators on those fields.
         *
         * XXX - Implementation limitation: Fields are implemented as type
         * XXX - Number, but they really are supposed to be type Buffer.
         * XXX - The two are interchangeable only for lengths <= 32 bits.
         */
        if(ValDesc->Field.Length > 32)
        {
            ObjectType = ACPI_TYPE_Buffer;
        }
        else
        {
            ObjectType = ACPI_TYPE_Number;
        }
            
        /*
         * Create the destination buffer object and the buffer space.
         */
        ObjDesc = CmCreateInternalObject (ObjectType);
        if (!ObjDesc)
        {   
            /* Descriptor allocation failure  */
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        /*
         * Fill in the object specific details
         */
        if (ACPI_TYPE_Buffer == ObjectType)
        {
            ObjDesc->Buffer.Pointer = CmCallocate(ValDesc->Field.Length);
            if (!ObjDesc->Buffer.Pointer)
            {   
                CmRemoveReference(ObjDesc);
                /* Buffer allocation failure  */
                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            ObjDesc->Buffer.Length = ValDesc->Field.Length;
                
            Status = AmlGetNamedFieldValue ((ACPI_HANDLE) StackEntry, ObjDesc->Buffer.Pointer, ObjDesc->Buffer.Length);

            if (AE_OK != Status)
            {
                return_ACPI_STATUS (Status);
            }
        }
        else
        {
            Status = AmlGetNamedFieldValue ((ACPI_HANDLE) StackEntry, &TempVal, sizeof (TempVal));

            if (AE_OK != Status)
            {
                return_ACPI_STATUS (Status);
            }

            ObjDesc->Number.Value = TempVal;
        }


        DEBUG_PRINT (TRACE_EXEC, ("AmlResolveEntryToValue: at DefField Entry=%p ValDesc=%p Type=%X\n", 
                        StackEntry, ValDesc, EntryType));
        break;


    case INTERNAL_TYPE_BankField:

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null BankField ValuePtr\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }

        if (AttachedAmlPointer)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - BankField cannot be an Aml ptr\n"));
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }

        if (INTERNAL_TYPE_BankField != ValDesc->Common.Type)
        {
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlResolveToValue/BankField:Internal error - Name %4.4s type %d does not match value-type %d at %p\n",
                    &(((NAME_TABLE_ENTRY *) (*StackPtr))->Name), INTERNAL_TYPE_BankField, ValDesc->Common.Type, ValDesc));
            
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }


        /* Get the global lock if needed */

        ObjDesc = (ACPI_OBJECT_INTERNAL *) *StackPtr;
        Locked = AmlAcquireGlobalLock (ObjDesc->FieldUnit.LockRule);
        {

            /* Set Index value to select proper Data register */
            /* perform the update */
        
            Status = AmlSetNamedFieldValue (ValDesc->BankField.BankSelect,
                                            &ValDesc->BankField.Value, sizeof (ValDesc->BankField.Value));
        }
        AmlReleaseGlobalLock (Locked);


        if (AE_OK != Status)
        {
            return_ACPI_STATUS (Status);
        }
        
        /* Read Data value */
        
        Status = AmlGetNamedFieldValue ((ACPI_HANDLE) ValDesc->BankField.Container, &TempVal, sizeof (TempVal));
        if (AE_OK != Status)
        {
            return_ACPI_STATUS (Status);
        }

        ObjDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!ObjDesc)
        {
            /* Descriptor allocation failure  */

            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        ObjDesc->Number.Value = TempVal;
        break;


    case INTERNAL_TYPE_IndexField:

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null IndexField ValuePtr\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }

        if (AttachedAmlPointer)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - IndexField cannot be an Aml ptr\n"));
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }

        if (INTERNAL_TYPE_IndexField != ValDesc->Common.Type)
        {
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlResolveToValue/IndexField: Internal error - Name %4.4s type %d does not match value-type %d at %p\n",
                    &(((NAME_TABLE_ENTRY *) (*StackPtr))->Name), INTERNAL_TYPE_IndexField, ValDesc->Common.Type, ValDesc));
            
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }


        /* Set Index value to select proper Data register */
        /* Get the global lock if needed */

        ObjDesc = (ACPI_OBJECT_INTERNAL *) *StackPtr;
        Locked = AmlAcquireGlobalLock (ObjDesc->FieldUnit.LockRule);
        {
            /* Perform the update */
          
            Status = AmlSetNamedFieldValue (ValDesc->IndexField.Index,
                                            &ValDesc->IndexField.Value, sizeof (ValDesc->IndexField.Value));
        }
        AmlReleaseGlobalLock (Locked);

        if (AE_OK != Status)
        {
            return_ACPI_STATUS (Status);
        }

        /* Read Data value */
        
        Status = AmlGetNamedFieldValue (ValDesc->IndexField.Data, &TempVal, sizeof (TempVal));
        if (AE_OK != Status)
        {
            return_ACPI_STATUS (Status);
        }

        ObjDesc = CmCreateInternalObject (ACPI_TYPE_Number);
        if (!ObjDesc)
        {
            /* Descriptor allocation failure  */

            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        ObjDesc->Number.Value = TempVal;
        break;


    case ACPI_TYPE_FieldUnit:

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - null FieldUnit ValuePtr\n"));
            return_ACPI_STATUS (AE_AML_NO_OPERAND);
        }

        if (AttachedAmlPointer)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Internal error - FieldUnit cannot be an Aml ptr\n"));
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
        }

        if (ValDesc->Common.Type != (UINT8) EntryType)
        {
            DEBUG_PRINT (ACPI_ERROR, (
                    "AmlResolveToValue/FieldUnit: Internal error - Name %4.4s type %d does not match value-type %d at %p\n",
                      &(((NAME_TABLE_ENTRY *) (*StackPtr))->Name), EntryType,
                      ValDesc->Common.Type, ValDesc));
            
            return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
            break;
        }

        ObjDesc = CmCreateInternalObject (ACPI_TYPE_Any);
        if (!ObjDesc)
        {
            /* Descriptor allocation failure  */

            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        if ((Status = AmlGetFieldUnitValue (ValDesc, ObjDesc)) != AE_OK)
        {
            CmRemoveReference (ObjDesc);
            return_ACPI_STATUS (Status);
        }

        break;



    /* 
     * For these objects, just return the object attached to the NTE
     */
    
    case ACPI_TYPE_Mutex:
    case ACPI_TYPE_Method:
    case ACPI_TYPE_Power:
    case ACPI_TYPE_Processor:
    case ACPI_TYPE_Thermal:
    case ACPI_TYPE_Event:
    case ACPI_TYPE_Region: 


        /* There must be an object attached to this NTE */

        if (!ValDesc)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: NTE %p has no attached object\n",
                            StackEntry));
        
            return_ACPI_STATUS (AE_AML_INTERNAL);
        }

        /* Return an additional reference to the object */

        ObjDesc = ValDesc;
        CmAddReference (ObjDesc);
        break;


    /* Devices rarely have an attached object, return the NTE */

    case ACPI_TYPE_Device:


    /* Method locals and arguments have a pseudo-NTE, just return it */

    case INTERNAL_TYPE_MethodArgument:
    case INTERNAL_TYPE_MethodLocalVar:

        return_ACPI_STATUS (AE_OK);
        break;


    /* TYPE_Any is untyped, and thus there is no object associated with it */

    case ACPI_TYPE_Any:

        DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Untyped entry %p - has no Rvalue\n",
                        StackEntry));
        
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);  /* Cannot be AE_TYPE */


    /* Default case is for unknown types */

    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlResolveEntryToValue: Entry %p - Unknown object type %d\n",
                        StackEntry, EntryType));
        
        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    
    } /* switch (EntryType) */




    /* Put the object descriptor on the stack */

    *StackPtr = (void *) ObjDesc;

    return_ACPI_STATUS (Status);
}



