
/*
 *  ./acpica/subsystem/interpreter/iresobj.c - AML Interpreter object resolution
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IRESOBJ_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iresobj");





ACPI_STATUS
AmlResolveObjectToValue (
    ACPI_OBJECT_INTERNAL    **StackPtr)
{
    ACPI_OBJECT_INTERNAL    *StackDesc;
    ACPI_STATUS             Status = AE_OK;
    ACPI_HANDLE             TempHandle = NULL;
    ACPI_OBJECT_INTERNAL    *ObjDesc = NULL;
    UINT32        Index = 0;
    UINT16                  Opcode;


    FUNCTION_TRACE ("AmlResolveObjectToValue");


    StackDesc = *StackPtr;

    /* This is an ACPI_OBJECT_INTERNAL */

    switch (StackDesc->Common.Type)
    {

    case INTERNAL_TYPE_Reference:

        Opcode = StackDesc->Reference.OpCode;

        switch (Opcode)
        {

        case AML_NameOp:
        
            /* 
             * Convert indirect name ptr to a direct name ptr.
             * Then, AmlResolveEntryToValue can be used to get the value
             */
        
            TempHandle = StackDesc->Reference.Object;
            CmRemoveReference (StackDesc);     /* Delete the Reference Object */

            /* Put direct name pointer onto stack and exit */

            (*StackPtr) = TempHandle;
            Status = AE_OK;
            break;


        case AML_LocalOp:

            Index = StackDesc->Reference.Offset;
            CmRemoveReference (StackDesc);     /* Delete the Reference Object */

            /* 
             * Get the local from the method's state info 
             * Note: this increments the object reference count
             */

            Status = DsMethodDataGetValue (MTH_TYPE_LOCAL, Index, StackPtr);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            StackDesc = *StackPtr;

            DEBUG_PRINT (ACPI_INFO, ("AmlResolveObjectToValue: [Local%d] ValueObj is %p\n",
                            Index, StackDesc));
        
            if (ACPI_TYPE_Number == StackDesc->Common.Type)
            {
                /* Value is a Number */
            
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveObjectToValue: [Local%d] value is [0x%X] \n", 
                                            Index, StackDesc->Number.Value));
            }

            break;


        case AML_ArgOp:

            Index = StackDesc->Reference.Offset;
            CmRemoveReference (StackDesc);     /* Delete the Reference Object*/

            /* 
             * Get the argument from the method's state info 
             * Note: this increments the object reference count
             */

            Status = DsMethodDataGetValue (MTH_TYPE_ARG, Index, StackPtr);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            StackDesc = *StackPtr;
   
            DEBUG_PRINT (TRACE_EXEC, ("AmlResolveObjectToValue: [Arg%d] ValueObj is %p\n",
                            Index, StackDesc));

            if (ACPI_TYPE_Number == StackDesc->Common.Type)
            {
                /* Value is a Number */
            
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveObjectToValue: [Arg%d] value is [0x%X] \n", 
                                            Index, StackDesc->Number.Value));
            }

            break;


        /*
         * TBD:  These next three opcodes change the type of the object,
         * which is actually a no-no.
         */

        case AML_ZeroOp:

            StackDesc->Common.Type = (UINT8) ACPI_TYPE_Number;
            StackDesc->Number.Value = 0;
            break;


        case AML_OneOp:

            StackDesc->Common.Type = (UINT8) ACPI_TYPE_Number;
            StackDesc->Number.Value = 1;
            break;


        case AML_OnesOp:

            StackDesc->Common.Type = (UINT8) ACPI_TYPE_Number;
            StackDesc->Number.Value = 0xFFFFFFFF;
            break;


        case AML_IndexOp:

            switch (StackDesc->Reference.TargetType)
            {
            case ACPI_TYPE_BufferField:

                /* Just return - leave the Reference on the stack */
                break;


            case ACPI_TYPE_Package:
                ObjDesc = *StackDesc->Reference.Where;
                if (ObjDesc)
                {
                    /*
                     * Valid obj descriptor, copy pointer to return value
                     * (i.e., dereference the package index)
                     */
                    CmRemoveReference (StackDesc);     /* Delete the Reference object */
                    CmAddReference (ObjDesc);          /* Increment the return value object */
                    *StackPtr = ObjDesc;
                }

                else
                {
                    /* NULL object descriptor means an unitialized element of the package, can't deref it */

                    DEBUG_PRINT (ACPI_ERROR, ("AmlResolveObjectToValue: Attempt to deref an Index to NULL pkg element Idx=%p\n", StackDesc));
                    Status = AE_AML_UNINITIALIZED_ELEMENT;
                }
                break;

            default:
                /* Invalid reference OBJ*/

                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveObjectToValue: Unknown TargetType %d in Index/Reference obj %p\n", 
                                StackDesc->Reference.TargetType, StackDesc));
                Status = AE_AML_INTERNAL;   
                break;
            }

            break;


        case AML_DebugOp:

            /* Just leave the object as-is */
            break;


        default:

            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveObjectToValue: Unknown Reference object subtype %02x in %p\n",
                            Opcode, StackDesc));
            Status = AE_AML_INTERNAL;

        }   /* switch (Opcode) */


        if (AE_OK != Status)
        {
            return_ACPI_STATUS (Status);
        }

        break; /* case INTERNAL_TYPE_Reference */


    case ACPI_TYPE_FieldUnit:

        ObjDesc = CmCreateInternalObject (ACPI_TYPE_Any);
        if (!ObjDesc)
        {   
            /* Descriptor allocation failure  */
        
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        if ((Status = AmlGetFieldUnitValue (StackDesc, ObjDesc)) != AE_OK)
        {
            CmRemoveReference (ObjDesc);
            ObjDesc = NULL;
        }
    
        *StackPtr = (void *) ObjDesc;
        break;


    case INTERNAL_TYPE_BankField:

        ObjDesc = CmCreateInternalObject (ACPI_TYPE_Any);
        if (!ObjDesc)
        {   
            /* Descriptor allocation failure */
        
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        if ((Status = AmlGetFieldUnitValue (StackDesc, ObjDesc)) != AE_OK)
        {
            CmRemoveReference (ObjDesc);
            ObjDesc = NULL;
        }

        *StackPtr = (void *) ObjDesc;
        break;


    /* TBD: - may need to handle IndexField, and DefField someday */

    default:

        break;

    }   /* switch (StackDesc->Common.Type) */


    return_ACPI_STATUS (Status);
}




