
/*
 *  ./acpica/subsystem/interpreter/iresolve.c - AML Interpreter object resolution
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IRESOLVE_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iresolve");




ACPI_STATUS
AmlGetFieldUnitValue (
    ACPI_OBJECT_INTERNAL    *FieldDesc, 
    ACPI_OBJECT_INTERNAL    *ResultDesc)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  Mask;
    UINT8                   *Location = NULL;
    BOOLEAN                 Locked = FALSE;


    FUNCTION_TRACE ("AmlGetFieldUnitValue");


    if (!FieldDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetFieldUnitValue: Internal error - null field pointer\n"));
        Status = AE_AML_NO_OPERAND;
    }

    else if (!FieldDesc->FieldUnit.Container)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetFieldUnitValue: Internal error - null container pointer\n"));
        Status = AE_AML_INTERNAL;
    }

    else if (ACPI_TYPE_Buffer != FieldDesc->FieldUnit.Container->Common.Type)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetFieldUnitValue: Internal error - container is not a Buffer\n"));
        Status = AE_AML_OPERAND_TYPE;
    }

    else if (FieldDesc->FieldUnit.Sequence
                != FieldDesc->FieldUnit.Container->Buffer.Sequence)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetFieldUnitValue: Internal error - stale Buffer [%lx != %lx]\n",
                        FieldDesc->FieldUnit.Sequence,
                        FieldDesc->FieldUnit.Container->Buffer.Sequence));
        Status = AE_AML_INTERNAL;
    }

    else if (!ResultDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlGetFieldUnitValue: Internal error - null result pointer\n"));
        Status = AE_AML_INTERNAL;
    }

    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }


        /* Get the global lock if needed */

    Locked = AmlAcquireGlobalLock (FieldDesc->FieldUnit.LockRule);

    /* Field location is (base of buffer) + (byte offset) */

    Location = FieldDesc->FieldUnit.Container->Buffer.Pointer
                + FieldDesc->FieldUnit.Offset;

    /* Construct Mask with as many 1 bits as the field width 
     * 
     * NOTE: Only the bottom 5 bits are valid for a shift operation, so
     *  special care must be taken for any shift greater than 31 bits. 
     *
     * TBD: Fields greater than 32-bits will not work.
     */

    if (FieldDesc->FieldUnit.Length < 32)
    {
        Mask = ((UINT32) 1 << FieldDesc->FieldUnit.Length) - (UINT32) 1;
    }
    else
    {
        Mask = 0xFFFFFFFF;
    }

    ResultDesc->Number.Type = (UINT8) ACPI_TYPE_Number;

    /* Shift the word containing the field, and mask the value */

    ResultDesc->Number.Value = *(UINT32 *) Location >> FieldDesc->FieldUnit.BitOffset & Mask;

    DEBUG_PRINT (ACPI_INFO, ("** Read from buffer %p byte %ld bit %d width %d addr %p mask %08lx val %08lx\n",
                    FieldDesc->FieldUnit.Container->Buffer.Pointer,
                    FieldDesc->FieldUnit.Offset,
                    FieldDesc->FieldUnit.BitOffset,
                    FieldDesc->FieldUnit.Length,
                    Location, Mask, ResultDesc->Number.Value));

    /* Release global lock if we acquired it earlier */

    AmlReleaseGlobalLock (Locked);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AmlResolveToValue (
    ACPI_OBJECT_INTERNAL    **StackPtr)
{
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE_PTR ("AmlResolveToValue", StackPtr);


    if (!StackPtr || !*StackPtr)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlResolveToValue: Internal error - null pointer\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }


    /* 
     * The entity pointed to by the StackPtr can be either
     * 1) A valid ACPI_OBJECT_INTERNAL, or
     * 2) A NAME_TABLE_ENTRY (nte)
     */

    if (VALID_DESCRIPTOR_TYPE (*StackPtr, DESC_TYPE_ACPI_OBJ))       
    {

        Status = AmlResolveObjectToValue (StackPtr);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }

    /* 
     * Object on the stack may have changed if AmlResolveObjectToValue() was called
     * (i.e., we can't use an _else_ here.)
     */

    if (VALID_DESCRIPTOR_TYPE (*StackPtr, DESC_TYPE_NTE))       
    {
        Status = AmlResolveEntryToValue ((NAME_TABLE_ENTRY **) StackPtr);
    }


    DEBUG_PRINT (ACPI_INFO, ("AmlResolveToValue: Returning resolved object %p\n", *StackPtr));

    return_ACPI_STATUS (Status);
}
    
