
/*
 *  ./acpica/subsystem/interpreter/iresop.c - AML Interpreter operand/object resolution
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __IRESOP_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>
#include <events.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("iresop");



ACPI_STATUS
AmlResolveOperands (
    UINT16                  Opcode,
    ACPI_OBJECT_INTERNAL    **StackPtr)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_STATUS             Status = AE_OK;
    UINT8                   ObjectType;
    ACPI_HANDLE             TempHandle;
    UINT32                  ArgTypes;
    ACPI_OP_INFO            *OpInfo;
    UINT32                  ThisArgType;


    FUNCTION_TRACE_U32 ("AmlResolveOperands", Opcode);


    OpInfo = PsGetOpcodeInfo (Opcode);
    if (!OpInfo)
    {
        return_ACPI_STATUS (AE_AML_BAD_OPCODE);
    }


    ArgTypes = OpInfo->RuntimeArgs;
    if (ArgTypes == ARGI_INVALID_OPCODE)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlResolveOperands: Internal error - %X is not a runtime opcode\n", Opcode));
        Status = AE_AML_INTERNAL;
        goto Cleanup;
    }

    DEBUG_PRINT (TRACE_EXEC, ("AmlResolveOperands: Opcode %X OperandTypes=%X \n", Opcode, ArgTypes));


   /* 
     * Normal exit is with *Types == '\0' at end of string.
     * Function will return an exception from within the loop upon
     * finding an entry which is not, and cannot be converted
     * to, the required type; if stack underflows; or upon
     * finding a NULL stack entry (which "should never happen").
     */

    while (GET_CURRENT_ARG_TYPE (ArgTypes))
    {
        if (!StackPtr || !*StackPtr)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveOperands: Internal error - null stack entry at %X\n", StackPtr));
            Status = AE_AML_INTERNAL;
            goto Cleanup;
        }

        /* Extract useful items */

        ObjDesc = *StackPtr;

        /* Decode the descriptor type */

        if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))
        {
            /* NTE */

            ObjectType = ((NAME_TABLE_ENTRY *) ObjDesc)->Type;
        }

        else if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_ACPI_OBJ))
        {
            /* ACPI internal object */

            ObjectType = ObjDesc->Common.Type;

            /* Check for bad ACPI_OBJECT_TYPE */

            if (!AmlValidateObjectType (ObjectType))
            {
                DEBUG_PRINT (ACPI_ERROR, ("AmlResolveOperands: Bad operand object type [0x%x]\n", ObjectType));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
        
            if (ObjectType == (UINT8) INTERNAL_TYPE_Reference)
            {
                /*
                 * Decode the Reference 
                 */

                OpInfo = PsGetOpcodeInfo (Opcode);
                if (!OpInfo)
                {
                    return_ACPI_STATUS (AE_AML_BAD_OPCODE);
                }


                switch (ObjDesc->Reference.OpCode)
                {
                case AML_ZeroOp:
                case AML_OneOp:
                case AML_OnesOp:
                case AML_DebugOp:
                case AML_NameOp:
                case AML_IndexOp:
                case AML_ArgOp:
                case AML_LocalOp:

                    DEBUG_PRINT (ACPI_INFO, ("Reference Opcode: %s\n", OpInfo->Name));
                    break;
            
                default:
                    DEBUG_PRINT (ACPI_INFO, ("Reference Opcode: Unknown [%02x]\n", ObjDesc->Reference.OpCode));

                    Status = AE_AML_OPERAND_TYPE;
                    goto Cleanup;
                    break;
                }
            }
        
        }
        
        else
        {
            /* Invalid descriptor */

            DEBUG_PRINT (ACPI_ERROR, ("Bad descriptor type 0x%X in Obj %p\n", ObjDesc->Common.DataType, ObjDesc));

            Status = AE_AML_OPERAND_TYPE;
            goto Cleanup;
        }


        /*
         * Decode a character from the type string
         */

        ThisArgType = GET_CURRENT_ARG_TYPE (ArgTypes);
        INCREMENT_ARG_LIST (ArgTypes);


        switch (ThisArgType)
        {

        case ARGI_REFERENCE:   /* Reference */
        case ARGI_TARGETREF:

            /* Need an operand of type INTERNAL_TYPE_Reference */

            if (VALID_DESCRIPTOR_TYPE (ObjDesc, DESC_TYPE_NTE))             /* direct name ptr OK as-is */
            {
                break;
            }

            if (INTERNAL_TYPE_Reference != ObjectType)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Reference, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }

            if (AML_NameOp == ObjDesc->Reference.OpCode)
            {
                /* Convert an indirect name ptr to direct name ptr and put it on the stack */
                
                TempHandle = ObjDesc->Reference.Object;
                CmRemoveReference (ObjDesc);
                (*StackPtr) = TempHandle;
            }
            break;


        case ARGI_NUMBER:   /* Number */

            /* Need an operand of type ACPI_TYPE_Number */

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Number != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Number, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_STRING:

            /* Need an operand of type ACPI_TYPE_String or ACPI_TYPE_Buffer */

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if ((ACPI_TYPE_String != (*StackPtr)->Common.Type) &&
                (ACPI_TYPE_Buffer != (*StackPtr)->Common.Type))
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed String or Buffer, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_BUFFER:

            /* Need an operand of type ACPI_TYPE_Buffer */

            if ((Status = AmlResolveToValue(StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Buffer != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Buffer, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_MUTEX:

            /* Need an operand of type ACPI_TYPE_Mutex */

            if ((Status = AmlResolveToValue(StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Mutex != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Mutex, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_EVENT:

            /* Need an operand of type ACPI_TYPE_Event */

            if ((Status = AmlResolveToValue(StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Event != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Event, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_REGION:

            /* Need an operand of type ACPI_TYPE_Region */

            if ((Status = AmlResolveToValue(StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Region != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Region, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


         case ARGI_IF:   /* If */

            /* Need an operand of type INTERNAL_TYPE_If */

            if (INTERNAL_TYPE_If != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed If, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_PACKAGE:   /* Package */

            /* Need an operand of type ACPI_TYPE_Package */

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            if (ACPI_TYPE_Package != (*StackPtr)->Common.Type)
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Package, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        case ARGI_ANYTYPE:


            /* 
             * We don't want to resolve IndexOp reference objects during
             * a store because this would be an implicit DeRefOf operation.
             * Instead, we just want to store the reference object.
             */

            if ((Opcode == AML_StoreOp) &&
                ((*StackPtr)->Common.Type == INTERNAL_TYPE_Reference) &&
                ((*StackPtr)->Reference.OpCode == AML_IndexOp))
            {
                break;
            }

            /* All others must be resolved */

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            /* All types OK, so we don't perform any typechecks */

            break;


        case ARGI_DATAOBJECT:
            /*
             * ARGI_DATAOBJECT is only used by the SizeOf operator.
             *
             * The ACPI specification allows SizeOf to return the size of
             *  a Buffer, String or Package.  However, the MS ACPI.SYS AML
             *  Interpreter also allows an NTE reference to return without
             *  error with a size of 4.
             */

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            /* Need a buffer, string, package or NTE reference */

            if (((*StackPtr)->Common.Type != ACPI_TYPE_Buffer) &&
                ((*StackPtr)->Common.Type != ACPI_TYPE_String) &&
                ((*StackPtr)->Common.Type != ACPI_TYPE_Package) &&
                ((*StackPtr)->Common.Type != INTERNAL_TYPE_Reference))
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Buf/Str/Pkg, found %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }

            /*
             * If this is a reference, only allow a reference to an NTE.
             */
            if ((*StackPtr)->Common.Type == INTERNAL_TYPE_Reference)
            {
                if (!(*StackPtr)->Reference.Nte)
                {
                    DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed NTE reference, found %s Obj=%p\n",
                                    CmGetTypeName (ObjectType), *StackPtr));
                    Status = AE_AML_OPERAND_TYPE;
                    goto Cleanup;
                }
            }

            break;


        case ARGI_COMPLEXOBJ:

            if ((Status = AmlResolveToValue (StackPtr)) != AE_OK)
            {
                goto Cleanup;
            }

            /* Need a buffer or package */

            if (((*StackPtr)->Common.Type != ACPI_TYPE_Buffer) &&
                ((*StackPtr)->Common.Type != ACPI_TYPE_Package))
            {
                DEBUG_PRINT (ACPI_INFO, ("AmlResolveOperands: Needed Package, Buf/Pkg %s Obj=%p\n",
                                CmGetTypeName (ObjectType), *StackPtr));
                Status = AE_AML_OPERAND_TYPE;
                goto Cleanup;
            }
            break;


        /* Unknown abbreviation passed in */

        default:
            DEBUG_PRINT (ACPI_ERROR, ("AmlResolveOperands: Internal error - Unknown arg type %X\n",
                        ThisArgType));
            Status = AE_BAD_PARAMETER;
            goto Cleanup;

        }   /* switch (*Types++) */


        /* 
         * If more operands needed, decrement StackPtr to point
         * to next operand on stack (after checking for underflow).
         */
        if (GET_CURRENT_ARG_TYPE (ArgTypes))
        {
            StackPtr--;
        }

    }   /* while (*Types) */


Cleanup:

  return_ACPI_STATUS (Status);
}







