
/*
 *  ./acpica/subsystem/interpreter/isdebug.c - Scanner debug support routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ISDEBUG_C__

#include <acpi.h>

#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("isdebug");


/* 
 *
 * CheckTrash() and CheckTrash are normally defined as macros with null
 * expansions.  If they are being defined as functions instead (for debug),
 * the definitions will be incorporated into this compilation.  This must
 * follow the above definition of PCodeBase, which is referenced in some
 * variants of the function form.
 */
#define DEFINE_CHECKTRASH



/* 
 * CheckTrash and CheckTrashA are for debugging, will normally be null
 *
 * The function forms were originally written to help find a problem where
 * the AML stream was being trashed due to storing via a stale or otherwise
 * bogus pointer, in an environment where the debugger did not provide
 * memory-write breakpoints (if it worked at all).  CheckTrash() was later
 * modified (under an inner #if 1 ... #else ... #endif) to always report its
 * locator parameter -- this was a low-effort method of getting some flow
 * tracing since the calls were already installed in various places.
 */

void
CheckTrash (
    char                    *Where)
{
    static int              Look = 1;
    static                  UINT32 *Trash = (UINT32 *) 0;


#if 1 /* looking for trash */

    if (DebugLevel & TRACE_TRASH) /* use TraceTrash bit of Trace word to enable */
    {
        if ((UINT32 *) 0 == Trash)
        {
            /* 
             * First call after activation: set and report a pointer to the
             * particular place where the trouble was occurring.  This will
             * need to be modified to pursue other storage-overlay bugs.
             */

            /*
             * What is this??
             *
             *  Trash = (UINT32 *) (PCodeBase + 0x284);
             */

            DEBUG_PRINT (TRACE_TRASH,
                        (" Looking for trash at %p [%s] ", Trash, Where));
        }

        /* 
         * Check for expected contents unless we have already reported a
         * problem.  This, too, will need to be modified to pursue other
         * storage-overlay bugs.
         */

        if (Look && *Trash != 0x020a2e0a)
        {
            DEBUG_PRINT (TRACE_TRASH,
                        ("*** Found trash %08lx at %p [%s] *** ",
                        *Trash, Trash, Where));
            Look = 0;
        }
    }

#else /* looking for trash */

    /* flow-tracing */

    LineSet ((int) strlen (Where) + 2, Load);
    DEBUG_PRINT (TRACE_TRASH, (" %s ", Where));

#endif /* looking for trash */
}


/* 
 * This works just like CheckTrash except that it does not report the
 * address where it is looking (which is the same address where CheckTrash()
 * is looking), and it prints out some additional values upon finding a
 * problem.  It has not been modified for flow tracing.
 *
 * The additional values printed -- three pointers, two UINT32s, and two
 * int's -- are meaningful only in the context of the calling logic in
 * nsSearchTable().
 */

void
CheckTrashA (
    char                    *Where, 
    UINT32                  *pulP, 
    UINT32                  ulC1, 
    UINT32                  ulC2,
    int                     *piP, 
    int                     iC1, 
    int                     iC2, 
    void                    *nNt)
{
    static int              Look = 1;
    UINT32                  *Trash = NULL;


    /*
     * What is this?
     *
     * Trash = (UINT32 *) (PCodeBase + 0x284);
     */

    if ((DebugLevel & TRACE_TRASH)
     && Look
     && *Trash != 0x020a2e0a)
    {
        DEBUG_PRINT (TRACE_TRASH,
                    ("*** Found trash %08lx at %p [%s] *** ",
                    *Trash, Trash, Where));
        Look = 0;
        DEBUG_PRINT (TRACE_TRASH,
                    ("\nat %p A=%08lx E=%08lx\nat %08lx A=%04x E=%04x  nt @ %p\n",
                    pulP, ulC1, ulC2, piP, iC1, iC2, nNt));
    }
}

