
/*
 *  ./acpica/subsystem/interpreter/isnames.c - interpreter/scanner name load/execute
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ISNAMES_C__

#include <acpi.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>

#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("isnames");


#define PKG_Type1   64              /*  or 0x40 Max encoding size = 0x3F    */
#define PKG_Type2   16384           /*  or 0x4000 Max encoding size = 0xFFF */
#define PKG_Type3   4194304         /*  or 0x400000 Max encoding size = 0xFFFFF */
#define PKG_Type4   1073741824      /*  or 0x40000000 Max encoding size = 0xFFFFFFF */



char *
AmlAllocateNameString (
    INT32                   PrefixCount, 
    INT32                   NumNameSegs)
{
    char                    *TempPtr;
    char                    *NameString;
    INT32                   SizeNeeded;


    FUNCTION_TRACE ("AmlAllocateNameString");

    /* 
     * Allow room for all \ and ^ prefixes, all segments, and a MultiNamePrefix.
     * This may actually be somewhat longer than needed.
     */

    if (PrefixCount < 0)
    {
        SizeNeeded = 1 /* root */ +  (ACPI_NAME_SIZE * NumNameSegs) + 2 +1; /* zero terminate */
    }
    else
    {
        SizeNeeded = PrefixCount + (ACPI_NAME_SIZE * NumNameSegs) + 2 +1; /* zero terminate */
    }

    /* 
     * Allocate a buffer for the name.
     * This buffer must be deleted by the caller!
     */

    NameString = CmAllocate ((ACPI_SIZE) SizeNeeded);
    if (!NameString)
    {
        /* Allocation failure  */

        REPORT_ERROR ("AmlAllocateNameString: name allocation failure");
        return_VALUE (NULL);
    }

    TempPtr = NameString;

    /* Set up Root or Parent prefixes if needed */

    if (PrefixCount < 0)
    {
        *TempPtr++ = AML_RootPrefix;
    }

    else
    {
        while (PrefixCount--)
        {
            *TempPtr++ = AML_ParentPrefix;
        }
    }


    /* Set up Dual or Multi prefixes if needed */
    
    if (NumNameSegs > 2)
    {
        /* Set up multi prefixes   */

        *TempPtr++ = AML_MultiNamePrefixOp;
        *TempPtr++ = (char) NumNameSegs;
    }

    else if (2 == NumNameSegs)
    {
        /* Set up dual prefixes */

        *TempPtr++ = AML_DualNamePrefix;
    }

    /* Terminate string following prefixes. AmlExecNameSegment() will append the segment(s) */

    *TempPtr = 0;

    return_VALUE (NameString);
}



INT32 
AmlDecodePackageLength (
    INT32                   LastPkgLen)
{
    INT32                   NumBytes = 0;


    FUNCTION_TRACE ("AmlDecodePackageLength");


    if (LastPkgLen < PKG_Type1)
    {
        NumBytes = 1;
    }

    else if (LastPkgLen < PKG_Type2)
    {
        NumBytes = 2;
    }

    else if (LastPkgLen < PKG_Type3)
    {
        NumBytes = 3;
    }

    else if (LastPkgLen < PKG_Type4)
    {
        NumBytes = 4;
    }

    return_VALUE (NumBytes);
}



ACPI_STATUS
AmlExecNameSegment (
    UINT8                   **InAmlAddress,
    char                    *NameString)
{
    UINT8                   *AmlAddress = *InAmlAddress;
    ACPI_STATUS             Status = AE_OK;
    INT32                   Index;
    char                    CharBuf[5];


    FUNCTION_TRACE ("AmlExecNameSegment");


    /* If first character is a digit, we aren't looking at a valid name segment    */

    CharBuf[0] = *AmlAddress;

    if ('0' <= CharBuf[0] && CharBuf[0] <= '9')
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecNameSegment: leading digit: %c\n", CharBuf[0]));
        return_ACPI_STATUS (AE_CTRL_PENDING);
    }

    DEBUG_PRINT (TRACE_LOAD, ("AmlExecNameSegment: Bytes from stream:\n"));

    for (Index = 4; (Index > 0) && (CmValidAcpiCharacter (*AmlAddress)); --Index)
    {
        CharBuf[4 - Index] = *AmlAddress++;
        DEBUG_PRINT (TRACE_LOAD, ("%c\n", CharBuf[4 - Index]));
    }


    /* Valid name segment  */

    if (0 == Index)
    {
        /* Found 4 valid characters */
    
        CharBuf[4] = '\0';

        if (NameString)
        {
            STRCAT (NameString, CharBuf);
            DEBUG_PRINT (TRACE_NAMES, ("AmlExecNameSegment: Appended to - %s \n", NameString));
        }

        else
        {
            DEBUG_PRINT (TRACE_NAMES, ("AmlExecNameSegment: No Name string - %s \n", CharBuf));
        }
    }

    else if (4 == Index)
    {
        /* 
         * First character was not a valid name character,
         * so we are looking at something other than a name.
         */
        DEBUG_PRINT (ACPI_INFO, ("AmlExecNameSegment: Leading char not alpha: %02Xh (not a name)\n", CharBuf[0]));
        Status = AE_CTRL_PENDING;
    }

    else
    {
        /* Segment started with one or more valid characters, but fewer than 4 */
    
        Status = AE_AML_BAD_NAME;
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecNameSegment: Bad char %02x in name, at %p\n", *AmlAddress, AmlAddress));
    }   

    DEBUG_PRINT (TRACE_EXEC, ("Leave AmlExecNameSegment %s \n", CmFormatException (Status)));

    *InAmlAddress = AmlAddress;

    return_ACPI_STATUS (Status);
}





ACPI_STATUS
AmlGetNameString (
    ACPI_OBJECT_TYPE        DataType, 
    UINT8                   *InAmlAddress,
    char                    **OutNameString,
    UINT32                  *OutNameLength)   
{
    ACPI_STATUS             Status = AE_OK;
    UINT8                   *AmlAddress = InAmlAddress;
    char                    *NameString = NULL;
    INT32                   NumSegments;
    INT32                   PrefixCount = 0;
    UINT8                   Prefix = 0;


    FUNCTION_TRACE_PTR ("AmlGetNameString", AmlAddress);


    if (INTERNAL_TYPE_DefField == DataType    || 
        INTERNAL_TYPE_BankField == DataType   || 
        INTERNAL_TYPE_IndexField == DataType)
    {   
        /* Disallow prefixes for types associated with field names */

        NameString = AmlAllocateNameString (0, 1);
        if (!NameString)
        {
            Status = AE_NO_MEMORY;
        }
        else
        {
            Status = AmlExecNameSegment (&AmlAddress, NameString);
        }
    }

    else
    {   
        /* DataType is not a field name   */

        switch (*AmlAddress)
        {   
            /* Examine first character of name for root or parent prefix operators */

        case AML_RootPrefix:

            Prefix = *AmlAddress++;
            DEBUG_PRINT (TRACE_LOAD, ("RootPrefix: %x\n", Prefix));

            /* 
             * Remember that we have a RootPrefix --
             * see comment in AmlAllocateNameString()
             */
            PrefixCount = -1;
            break;


        case AML_ParentPrefix:

            do
            {
                Prefix = *AmlAddress++;
                DEBUG_PRINT (TRACE_LOAD, ("ParentPrefix: %x\n", Prefix));

                ++PrefixCount;

            } while (*AmlAddress == AML_ParentPrefix);
            
            break;


        default:

            break;
        }



        switch (*AmlAddress)
        {
            /* Examine first character of name for name segment prefix operator */
            
        case AML_DualNamePrefix:

            Prefix = *AmlAddress++;
            DEBUG_PRINT (TRACE_LOAD, ("DualNamePrefix: %x\n", Prefix));

            NameString = AmlAllocateNameString (PrefixCount, 2);
            if (!NameString)
            {
                Status = AE_NO_MEMORY;
                break;
            }

            /* Ensure PrefixCount != 0 to remember processing a prefix */
            
            PrefixCount += 2;

            if ((Status = AmlExecNameSegment (&AmlAddress, NameString)) == AE_OK)
            {
                Status = AmlExecNameSegment (&AmlAddress, NameString);
            }
            break;


        case AML_MultiNamePrefixOp:

            Prefix = *AmlAddress++;
            DEBUG_PRINT (TRACE_LOAD, ("MultiNamePrefix: %x\n", Prefix));

            NumSegments = *AmlAddress++;                      /* fetch count of segments */

            NameString = AmlAllocateNameString (PrefixCount, NumSegments);
            if (!NameString)
            {
                Status = AE_NO_MEMORY;
                break;
            }

            /* Ensure PrefixCount != 0 to remember processing a prefix */
            
            PrefixCount += 2;

            while (NumSegments && (Status = AmlExecNameSegment (&AmlAddress, NameString)) == AE_OK)
            {
                --NumSegments;
            }

            break;


        case 0: 
            
            /* NullName valid as of 8-12-98 ASL/AML Grammar Update */
            
            if (-1 == PrefixCount)  
            {
                /* RootPrefix followed by NULL */
            
                DEBUG_PRINT (TRACE_EXEC, ("AmlDoName: NameSeg is \"\\\" followed by NULL\n"));
            }

            AmlAddress++;    /*  consume NULL byte   */
            NameString = AmlAllocateNameString (PrefixCount, 0);
            if (!NameString)
            {
                Status = AE_NO_MEMORY;
                break;
            }

            break;


        default:

            /* Name segment string */

            NameString = AmlAllocateNameString (PrefixCount, 1);
            if (!NameString)
            {
                Status = AE_NO_MEMORY;
                break;
            }

            Status = AmlExecNameSegment (&AmlAddress, NameString);
            break;

        }   /* Switch (PeekOp ())    */
    }


    if (AE_CTRL_PENDING == Status && PrefixCount != 0)
    {
        /* Ran out of segments after processing a prefix */

        DEBUG_PRINT (ACPI_ERROR, ("AmlDoName: Malformed Name at %p\n", NameString));
        REPORT_ERROR ("Ran out of segments after processing a prefix");

        Status = AE_AML_BAD_NAME;
    }


    *OutNameString = NameString;
    *OutNameLength = (UINT32) (AmlAddress - InAmlAddress);

    return_ACPI_STATUS (Status);
}




