
/*
 *  ./acpica/subsystem/interpreter/istore.c - AML Interpreter object store support
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ISTORE_C__

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <tables.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("istore");




ACPI_STATUS
AmlExecStore (
    ACPI_OBJECT_INTERNAL    *ValDesc, 
    ACPI_OBJECT_INTERNAL    *DestDesc)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_INTERNAL    *DeleteDestDesc = NULL;
    ACPI_OBJECT_INTERNAL    *TmpDesc;
    NAME_TABLE_ENTRY        *Entry = NULL;
    UINT8                   Value = 0;


    FUNCTION_TRACE ("AmlExecStore");

    DEBUG_PRINT (ACPI_INFO, ("entered AmlExecStore: Val=%p, Dest=%p\n", 
                    ValDesc, DestDesc));


    /* Validate parameters */

    if (!ValDesc || !DestDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("AmlExecStore: Internal error - null pointer\n"));
        return_ACPI_STATUS (AE_AML_NO_OPERAND);
    }

    /* Examine the datatype of the DestDesc */

    if (VALID_DESCRIPTOR_TYPE (DestDesc, DESC_TYPE_NTE))
    {
        /* Dest is an ACPI_HANDLE, create a new object */

        Entry = (NAME_TABLE_ENTRY *) DestDesc;
        DestDesc = CmCreateInternalObject (INTERNAL_TYPE_Reference);
        if (!DestDesc)
        {   
            /* Allocation failure  */
            
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        /* Build a new Reference wrapper around the handle */

        DestDesc->Reference.OpCode = AML_NameOp;
        DestDesc->Reference.Object = Entry;
    }

    else
    {
        /* DestDesc is not an ACPI_HANDLE  */

        DEBUG_PRINT (ACPI_INFO, ("AmlExecStore: Dest is object (not handle) - may be deleted!\n"));
    }

    /* Destination object must be of type Reference */

    if (DestDesc->Common.Type != INTERNAL_TYPE_Reference)
    {   
        /* Destination is not an Reference */

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecStore: Destination is not an Reference [%p]\n", DestDesc));

        DUMP_STACK_ENTRY (ValDesc);
        DUMP_STACK_ENTRY (DestDesc);
        DUMP_OPERANDS (&DestDesc, IMODE_Execute, "AmlExecStore", 2, "target not Reference");

        return_ACPI_STATUS (AE_AML_OPERAND_TYPE);
    }

    /* Examine the Reference opcode */

    switch (DestDesc->Reference.OpCode)
    {

    case AML_NameOp:

        /*
         *  Storing into a Name
         */
        DeleteDestDesc = DestDesc;
        Status = AmlStoreObjectToNte (ValDesc, DestDesc->Reference.Object);

        break;  /* Case NameOp */


    case AML_IndexOp:

        DeleteDestDesc = DestDesc;
        
        /*
         * Actually, storing to a package is not so simple.  The source must be
         *  evaluated and converted to the type of the destination and then the
         *  source is copied into the destination - we can't just point to the
         *  source object.
         */
        if (DestDesc->Reference.TargetType == ACPI_TYPE_Package)
        {
            /* 
             * The object at *(DestDesc->Reference.Where) is the
             *  element within the package that is to be modified.
             */
            TmpDesc = *(DestDesc->Reference.Where);
            if (TmpDesc)
            {
                /*
                 * If the Destination element is a package, we will delete
                 *  that object and construct a new one.
                 *
                 * TBD: Should both the src and dest be required to be packages?
                 *       && (ValDesc->Common.Type == ACPI_TYPE_Package)
                 */
                if (TmpDesc->Common.Type == ACPI_TYPE_Package)
                {
                    /* Take away the reference for being part of a package and delete  */

                    CmRemoveReference (TmpDesc);
                    CmRemoveReference (TmpDesc);

                    TmpDesc = NULL;
                }
            }

            if (!TmpDesc)
            {
                /*
                 * If the TmpDesc is NULL, that means an uninitialized package
                 * has been used as a destination, therefore, we must create
                 * the destination element to match the type of the source element
                 * NOTE: ValDesc can be of any type.
                 */
                TmpDesc = CmCreateInternalObject (ValDesc->Common.Type);
                if (!TmpDesc)
                {
                    Status = AE_NO_MEMORY;
                    goto Cleanup;
                }

                /*
                 * If the source is a package, copy the source to the new dest
                 */
                if (ACPI_TYPE_Package == TmpDesc->Common.Type)
                {
                    Status = AmlBuildCopyInternalPackageObject (ValDesc, TmpDesc);
                    if (ACPI_FAILURE (Status))
                    {
                        CmRemoveReference (TmpDesc);
                        TmpDesc = NULL;
                        goto Cleanup;
                    }
                }

                /* 
                 * Install the new descriptor into the package and add a reference to the 
                 * newly created descriptor for now being part of the parent package 
                 */

                *(DestDesc->Reference.Where) = TmpDesc;
                CmAddReference (TmpDesc);   
            }

            if (ACPI_TYPE_Package != TmpDesc->Common.Type)
            {
                /* 
                 * The destination element is not a package, so we need to convert the contents of
                 * the source (ValDesc) and copy into the destination (TmpDesc)
                 */
                Status = AmlStoreObjectToObject(ValDesc, TmpDesc);
                if (ACPI_FAILURE (Status))
                {
                    /* 
                     * An error occurrered when copying the internal object
                     *  so delete the reference.
                     */
                    DEBUG_PRINT (ACPI_ERROR, ("AmlExecStore/Index: Unable to copy the internal object\n"));
                    Status = AE_AML_OPERAND_TYPE;
                }
            }

            break;
        }

        /*
         * Check that the destination is a Buffer Field type
         */
        if (DestDesc->Reference.TargetType != ACPI_TYPE_BufferField)
        {
            Status = AE_AML_OPERAND_TYPE;
            break;
        }
        
        /* 
         * Storing into a buffer at a location defined by an Index.
         * This is a little more complex than the package case.
         *
         * If value is not a Number, try to resolve it to one.
         */
        switch (ValDesc->Common.Type) 
        {
        /*
         * If the type is Integer, only the least significant
         *  8-bits are used
         */
        case ACPI_TYPE_Number:
            Value = (UINT8)ValDesc->Number.Value;
            break;

        /*
         * If the type is Buffer, the least significant
         *  8-bits are copied over (last element)
         */
        case ACPI_TYPE_Buffer:
            Value = *(ValDesc->Buffer.Pointer + ValDesc->Buffer.Length - 1);
            break;

        /*
         * If the type is String, the least significant
         *  8-bits are copied over (last element)
         */
        case ACPI_TYPE_String:
            Value = *(ValDesc->String.Pointer + ValDesc->String.Length - 1);
            break;

        /*
         * If value is not a Number, try to resolve it to one.
         */
        default:
            if ((ValDesc->Common.Type != ACPI_TYPE_Number) &&
               ((Status = AmlResolveToValue (&ValDesc)) != AE_OK))
            {
                DeleteDestDesc = DestDesc;
            }

            else if (ValDesc->Common.Type != ACPI_TYPE_Number)
            {
                DEBUG_PRINT (ACPI_ERROR, ("AmlExecStore/Index: Index value must be Number, not %d\n",
                                ValDesc->Common.Type));

                DeleteDestDesc = DestDesc;
                Status = AE_AML_OPERAND_TYPE;
            }
            break;
        }

        /*
         * If we had an error, break out of this case statement.
         */
        if(AE_OK != Status)
        {
            break;
        }

        /* 
         * Delete descriptor that points to name,
         * and point to descriptor for name's value instead.
         */

        DeleteDestDesc = DestDesc;

        /*
         * Valid source value and destination reference pointer.
         *
         * ACPI Specification 1.0B section 15.2.3.4.2.13:
         * Destination should point to either a buffer or a package
         */

        /*
         *  Dereference the Buffer Field and set it equal to the ValDesc value
         */
        TmpDesc = DestDesc->Reference.Object;

        if (TmpDesc->Common.Type != ACPI_TYPE_Buffer)
        {
            Status = AE_AML_OPERAND_TYPE;
            break;
        }

        TmpDesc->Buffer.Pointer[DestDesc->Reference.Offset] = Value;

        DestDesc = TmpDesc;

        break;

    case AML_ZeroOp: 
    case AML_OneOp: 
    case AML_OnesOp:

        /* 
         * Storing to a constant is a no-op -- see spec sec 15.2.3.3.1.
         * Delete the result descriptor.
         */

        DeleteDestDesc = DestDesc;
        break;


    case AML_LocalOp:

        Status = DsMethodDataSetValue (MTH_TYPE_LOCAL, (DestDesc->Reference.Offset), ValDesc);
        DeleteDestDesc = DestDesc;
        break;


    case AML_ArgOp:

        Status = DsMethodDataSetValue (MTH_TYPE_ARG, (DestDesc->Reference.Offset), ValDesc);
        DeleteDestDesc = DestDesc;
        break;


    case AML_DebugOp:

        /* 
         * Storing to the Debug object causes the value stored to be
         * displayed and otherwise has no effect -- see sec. 15.2.3.3.3.
         */
        DEBUG_PRINT (ACPI_INFO, ("**** Write to Debug Object: ****: \n"));
        if (ValDesc->Common.Type == ACPI_TYPE_String)
        {
            DEBUG_PRINT (ACPI_INFO, ("%s\n", ValDesc->String.Pointer));
        }
        else
        {
            DUMP_STACK_ENTRY (ValDesc);
        }

        DeleteDestDesc = DestDesc;
        break;


    default:

        DEBUG_PRINT (ACPI_ERROR, ("AmlExecStore: Internal error - Unknown Reference subtype %02x\n",
                        DestDesc->Reference.OpCode));
        
        /* TBD:  use object dump routine !! */

        DUMP_BUFFER (DestDesc, sizeof (ACPI_OBJECT_INTERNAL));

        DeleteDestDesc = DestDesc;
        Status = AE_AML_INTERNAL;
    
    }   /* switch(DestDesc->Reference.OpCode) */




Cleanup:

    /* Cleanup and exit*/

    if (DeleteDestDesc)
    {
        CmRemoveReference (DeleteDestDesc);
    }

    return_ACPI_STATUS (Status);
}


