
/*
 *  ./acpica/subsystem/interpreter/istoreob.c - AML Interpreter object store support, store to object
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __ISTOREOB_C__

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <interp.h>
#include <amlcode.h>
#include <namesp.h>
#include <tables.h>


#define _COMPONENT          INTERPRETER
        MODULE_NAME         ("istoreob");




ACPI_STATUS
AmlStoreObjectToObject (
    ACPI_OBJECT_INTERNAL    *ValDesc,
    ACPI_OBJECT_INTERNAL    *DestDesc)
{
    ACPI_STATUS             Status = AE_OK;
    UINT8                   *Buffer = NULL;
    UINT32                  Length = 0;
    ACPI_OBJECT_TYPE        DestinationType = DestDesc->Common.Type;


    FUNCTION_TRACE ("AmlStoreObjectToObject");

    DEBUG_PRINT (ACPI_INFO, ("entered AmlStoreObjectToObject: Dest=%p, Val=%p\n", 
                    DestDesc, ValDesc));

    /*
     *  Assuming the parameters are valid!!!
     */
    ACPI_ASSERT((DestDesc) && (ValDesc));

    DEBUG_PRINT (ACPI_INFO, ("AmlStoreObjectToObject: Storing %s into %s\n",
                    CmGetTypeName (ValDesc->Common.Type),
                    CmGetTypeName (DestDesc->Common.Type)));

    /*
     *  First ensure we have a value that can be stored in the target
     */
    switch (DestinationType) 
    {
        /* Type of Name's existing value */

    case ACPI_TYPE_Number:

        /*
         *  These cases all require only number values or values that
         *  can be converted to numbers.
         *
         *  If value is not a Number, try to resolve it to one.
         */

        if (ValDesc->Common.Type != ACPI_TYPE_Number) 
        {
            /*
             *  Initially not a number, convert
             */
            Status = AmlResolveToValue (&ValDesc);
            if ((Status == AE_OK) &&
                (ValDesc->Common.Type != ACPI_TYPE_Number)) 
            {
                /*
                 *  Conversion successful but still not a number
                 */
                DEBUG_PRINT (ACPI_ERROR, ("AmlStoreObjectToObject: Value assigned to %s must be Number, not %s\n",
                                CmGetTypeName (DestinationType),
                                CmGetTypeName (ValDesc->Common.Type)));
                Status = AE_AML_OPERAND_TYPE;
            }
        }

        break;

    case ACPI_TYPE_String:
    case ACPI_TYPE_Buffer:

        /* 
         *  Storing into a Field in a region or into a buffer or into
         *  a string all is essentially the same.
         *
         *  If value is not a valid type, try to resolve it to one.
         */

        if ((ValDesc->Common.Type != ACPI_TYPE_Number) && 
            (ValDesc->Common.Type != ACPI_TYPE_Buffer) &&
            (ValDesc->Common.Type != ACPI_TYPE_String)) 
        {
            /*
             *  Initially not a valid type, convert
             */
            Status = AmlResolveToValue (&ValDesc);
            if ((Status == AE_OK) &&
                (ValDesc->Common.Type != ACPI_TYPE_Number) && 
                (ValDesc->Common.Type != ACPI_TYPE_Buffer) &&
                (ValDesc->Common.Type != ACPI_TYPE_String)) 
            {
                /*
                 *  Conversion successful but still not a valid type
                 */
                DEBUG_PRINT (ACPI_ERROR, ("AmlStoreObjectToObject: Assign wrong type %s to %s (must be type Num/Str/Buf)\n",
                                CmGetTypeName (ValDesc->Common.Type),
                                CmGetTypeName (DestinationType)));
                Status = AE_AML_OPERAND_TYPE;
            }
        }
        break;


    default: 
        
        /* 
         * TBD: What other combinations must be implemented?
         */
        Status = AE_NOT_IMPLEMENTED;
        break;
    }

    /* Exit now if failure above */

    if (Status != AE_OK)
    {
        goto CleanUpAndBailOut;
    }

    /*
     * Everything is ready to execute now,  We have 
     * a value we can handle, just perform the update
     */

    switch (DestinationType) 
    {

    case ACPI_TYPE_String:

        /*
         *  Perform the update
         */
        
        switch (ValDesc->Common.Type)
        {
        case ACPI_TYPE_Number:
            Buffer = (UINT8 *) &ValDesc->Number.Value;
            Length = sizeof (ValDesc->Number.Value);
            break;

        case ACPI_TYPE_Buffer:
            Buffer = (UINT8 *) ValDesc->Buffer.Pointer;
            Length = ValDesc->Buffer.Length; 
            break;

        case ACPI_TYPE_String:
            Buffer = (UINT8 *) ValDesc->String.Pointer;
            Length = ValDesc->String.Length; 
            break;
        }

        /*
         *  Setting a string value replaces the old string
         */

        if (Length < DestDesc->String.Length)
        {
            /*
             *  Zero fill, not willing to do pointer arithmetic for
             *  architecture independence.  Just clear the whole thing
             */
            MEMSET(DestDesc->String.Pointer, 0, DestDesc->String.Length);
            MEMCPY(DestDesc->String.Pointer, Buffer, Length);
        }
        else
        {
            /*
             *  Free the current buffer, then allocate a buffer
             *  large enough to hold the value
             */
            if ( DestDesc->String.Pointer &&
                !TbSystemTablePointer (DestDesc->String.Pointer))
            {
                /*
                 *  Only free if not a pointer into the DSDT
                 */

                CmFree(DestDesc->String.Pointer);
            }

            DestDesc->String.Pointer = CmAllocate ((ACPI_SIZE) (Length + 1));
            DestDesc->String.Length = Length;

   if (!DestDesc->String.Pointer)
   {
    Status = AE_NO_MEMORY;
    goto CleanUpAndBailOut;
   }

            MEMCPY(DestDesc->String.Pointer, Buffer, Length);
        }
        break;


    case ACPI_TYPE_Buffer:

        /*
         *  Perform the update to the buffer
         */
        
        switch (ValDesc->Common.Type)
        {
        case ACPI_TYPE_Number:
            Buffer = (UINT8 *) &ValDesc->Number.Value;
            Length = sizeof (ValDesc->Number.Value);
            break;

        case ACPI_TYPE_Buffer:
            Buffer = (UINT8 *) ValDesc->Buffer.Pointer;
            Length = ValDesc->Buffer.Length; 
            break;

        case ACPI_TYPE_String:
            Buffer = (UINT8 *) ValDesc->String.Pointer;
            Length = ValDesc->String.Length; 
            break;
        }

  /*
   * If the buffer is uninitialized, 
   * memory needs to be allocated for the copy.
   */
  if(0 == DestDesc->Buffer.Length)
  {
   DestDesc->Buffer.Pointer = CmCallocate(Length);
   DestDesc->Buffer.Length = Length;

   if (!DestDesc->Buffer.Pointer)
   {
    Status = AE_NO_MEMORY;
    goto CleanUpAndBailOut;
   }
  }

        /*
         *  Buffer is a static allocation,
         *  only place what will fit in the buffer.
         */
        if (Length <= DestDesc->Buffer.Length)
        {
            /*
             *  Zero fill first, not willing to do pointer arithmetic for
             *  architecture independence.  Just clear the whole thing
             */
            MEMSET(DestDesc->Buffer.Pointer, 0, DestDesc->Buffer.Length);
            MEMCPY(DestDesc->Buffer.Pointer, Buffer, Length);
        }
        else
        {
            /*
             *  truncate, copy only what will fit
             */
            MEMCPY(DestDesc->Buffer.Pointer, Buffer, DestDesc->Buffer.Length);
            DEBUG_PRINT (ACPI_INFO, ("AmlStoreObjectToObject: Truncating src buffer from %d to %d\n", 
                            Length, DestDesc->Buffer.Length));
        }
        break;

    case ACPI_TYPE_Number:

        DestDesc->Number.Value = ValDesc->Number.Value;
        break;

    default: 
            
        /* 
         * All other types than Alias and the various Fields come here.
         * Store ValDesc as the new value of the Name, and set
         * the Name's type to that of the value being stored in it.
         * ValDesc reference count is incremented by AttachObject.
         */

        DEBUG_PRINT (ACPI_WARN, ("AmlStoreObjectToObject: Store into %s not implemented\n",
                        CmGetTypeName (DestDesc->Common.Type)));
    
        Status = AE_NOT_IMPLEMENTED;
        break;
    }

CleanUpAndBailOut:

    return_ACPI_STATUS (Status);
}

