
/*
 *  ./acpica/subsystem/methods/mtwak.c - acpi Execute_WAK()
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __MTWAK_C__

#include <acpi.h>
#include <methods.h>
#include <namesp.h>


#define _COMPONENT          MISCELLANEOUS
        MODULE_NAME         ("mtwak");



ACPI_STATUS 
Execute_WAK (
    INT32                   WakeSleepState)
{
    ACPI_OBJECT_INTERNAL    ObjDesc;
    ACPI_OBJECT_INTERNAL    *InObjDesc[2];
    ACPI_OBJECT_INTERNAL    *OutObjDesc;
    ACPI_STATUS             Status = AE_NOT_FOUND;


    FUNCTION_TRACE ("Execute_WAK");


    ObjDesc.Common.Type = (UINT8) ACPI_TYPE_Number;
    ObjDesc.Number.Value = WakeSleepState;
    
    InObjDesc[0] = &ObjDesc;
    InObjDesc[1] = NULL;

    if (!SystemMethodExists ("_WAK"))
    {   
        /*  _WAK control method does NOT exist in ACPI namespace    */
        
        DEBUG_PRINT (ACPI_ERROR, ("_WAK does not exist\n"));
        return_ACPI_STATUS (AE_NOT_FOUND);
    }

    /*  _WAK control method exists in namespace */
    
    Status = NsEvaluateByName ("\\_WAK", InObjDesc, &OutObjDesc);
    if (ACPI_FAILURE (Status))
    {
        if (Status == AE_NOT_FOUND)
        {
            DEBUG_PRINT (ACPI_INFO, ("_WAK was not found\n"));
        }

        else
        {
            DEBUG_PRINT (ACPI_ERROR, ("_WAK failed with status %4.4x\n", Status));
        }

        return_ACPI_STATUS (Status);
    }

    if (!OutObjDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("No object was returned from _WAK\n"));
        return_ACPI_STATUS (AE_TYPE);
    }

    /*  successful AcpiEvaluateObject(_WAK)  */
    
    if (OutObjDesc->Common.Type != ACPI_TYPE_Package)
    {   
        /*  _WAK did not return a package   */
        
        Status = AE_TYPE;
        DEBUG_PRINT (ACPI_ERROR, ("_WAK was executed.  A return package was expected"
                "\n  but %s was returned\n", 
                ((OutObjDesc->Common.Type > (UINT8) INTERNAL_TYPE_Reference) ? 
                "Unknown" : CmGetTypeName (OutObjDesc->Common.Type))));
    }
    
    else
    {   
        /*  _WAK returned a package */
        
        if ((OutObjDesc->Package.Count != 2) ||
             ((OutObjDesc->Package.Elements[0])->Common.Type != ACPI_TYPE_Number) ||
             ((OutObjDesc->Package.Elements[1])->Common.Type != ACPI_TYPE_Number))
        {   
            /*  _WAK returned an incorrect package  */
            
            Status = AE_ERROR;
            
            if (OutObjDesc->Package.Count != 2)
            {   
                /*  incorrect package size  */
                
                DEBUG_PRINT (ACPI_ERROR, ("_WAK return package must have 2 elements,"
                            "\n  but contains %d elements\n", 
                            OutObjDesc->Package.Count));
            }
            
            else
            {   
                /*  non-numeric package */
                
                DEBUG_PRINT (ACPI_ERROR, ("Element %d of _WAK return package must be a"
                        "\n  number but was a %s\n", (OutObjDesc->Package.Elements[0])->Common.Type != ACPI_TYPE_Number ? 0 : 1,
                        (OutObjDesc->Common.Type > (UINT8) INTERNAL_TYPE_Reference) ? "Unknown" : CmGetTypeName (OutObjDesc->Common.Type)));
            }
        }   
        
        else
        {   
            /*  _WAK returned a numeric package of correct size */
            
            if (BIT0 (OutObjDesc->Package.Elements[0]->Number.Value))
            {
                Status = AE_ERROR;
                DEBUG_PRINT (ACPI_ERROR, ("WAK was signaled but failed due to lack of power\n"));
            }

            if (BIT1 (OutObjDesc->Package.Elements[0]->Number.Value))
            {
                Status = AE_ERROR;
                DEBUG_PRINT (ACPI_ERROR, ("WAK was signaled but failed due to a thermal condition\n"));
            }

            if (OutObjDesc->Package.Elements[1]->Number.Value)
            {
                Status = AE_ERROR;
                DEBUG_PRINT (ACPI_ERROR, ("WAK was executed. The power supply really"
                            "\n  entered the S%d state due to an overdrawn current condition.\n"));
            }
        }
    }

    /* On exit, we must delete the return object */

    CmRemoveReference (OutObjDesc);



    return_ACPI_STATUS (Status);
}
