
/*
 *  ./acpica/subsystem/namespace/nsaccess.c - Top-level functions for accessing ACPI namespace
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSACCESS_C__

#include <acpi.h>
#include <amlcode.h>
#include <interp.h>
#include <namesp.h>
#include <dispatch.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsaccess");




ACPI_STATUS
NsRootCreateScope (
    NAME_TABLE_ENTRY        *Entry)
{
    FUNCTION_TRACE ("NsRootCreateScope");


    /* Allocate a scope table */

    if (Entry->Scope)
    {
        return_ACPI_STATUS (AE_EXIST);
    }

    Entry->Scope = NsAllocateNameTable (NS_TABLE_SIZE);
    if (!Entry->Scope)
    {
        /*  root name table allocation failure  */

        REPORT_ERROR ("Root name table allocation failure");
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* 
     * Init the scope first entry -- since it is the exemplar of 
     * the scope (Some fields are duplicated to new entries!) 
     */
    NsInitializeTable (Entry->Scope, NULL, Entry);
    return_ACPI_STATUS (AE_OK);

}




ACPI_STATUS
NsRootInitialize (void)
{
    ACPI_STATUS             Status = AE_OK;
    PREDEFINED_NAMES        *InitVal = NULL;
    NAME_TABLE_ENTRY        *NewEntry;
    ACPI_OBJECT_INTERNAL    *ObjDesc;


    FUNCTION_TRACE ("NsRootInitialize");


    CmAcquireMutex (MTX_NAMESPACE);
    
    /* 
     * Root is initially NULL, so a non-NULL value indicates
     * that NsRootInitialize() has already been called; just return.
     */

    if (Gbl_RootObject->Scope)
    {
        Status = AE_OK;
        goto UnlockAndExit;
    }


    /* Create the root scope */

    Status = NsRootCreateScope (Gbl_RootObject);
    if (ACPI_FAILURE (Status))
    {
        goto UnlockAndExit;
    }

    /* Enter the pre-defined names in the name table */
    
    DEBUG_PRINT (ACPI_INFO, ("Entering predefined name table entries into namespace\n"));

    for (InitVal = Gbl_PreDefinedNames; InitVal->Name; InitVal++)
    {
        Status = NsLookup (NULL, InitVal->Name, InitVal->Type, 
                                    IMODE_LoadPass2, NS_NO_UPSEARCH, NULL, &NewEntry);

        /* 
         * if name entered successfully
         *  && its entry in PreDefinedNames[] specifies an initial value
         */
        
        if ((Status == AE_OK) && 
            NewEntry && InitVal->Val)
        {
            /* Entry requests an initial value, allocate a descriptor for it. */
            
            ObjDesc = CmCreateInternalObject (InitVal->Type);
            if (!ObjDesc)
            {
                Status = AE_NO_MEMORY;
                goto UnlockAndExit;
            }

            /* 
             * Convert value string from table entry to internal representation.
             * Only types actually used for initial values are implemented here.
             */

            switch (InitVal->Type)
            {

            case ACPI_TYPE_Number:

                ObjDesc->Number.Value = (UINT32) STRTOUL (InitVal->Val, NULL, 10);
                break;


            case ACPI_TYPE_String:

                ObjDesc->String.Length = (UINT16) STRLEN (InitVal->Val);

                /* 
                 * Allocate a buffer for the string.  All String.Pointers must be
                 * allocated buffers! (makes deletion simpler)
                 */
                ObjDesc->String.Pointer = CmAllocate ((ACPI_SIZE) (ObjDesc->String.Length + 1));
                if (!ObjDesc->String.Pointer)
                {
                    REPORT_ERROR ("Initial value string allocation failure");
                    CmRemoveReference (ObjDesc);
                    Status = AE_NO_MEMORY;
                    goto UnlockAndExit;
                }

                STRCPY ((char *) ObjDesc->String.Pointer, InitVal->Val);
                break;


            case ACPI_TYPE_Mutex:

                ObjDesc->Mutex.SyncLevel = (UINT16) STRTOUL (InitVal->Val, NULL, 10);
                Status = OsdCreateSemaphore (1, &ObjDesc->Mutex.Semaphore);
                if (ACPI_FAILURE (Status))
                {
                    goto UnlockAndExit;
                }

                if (STRCMP (InitVal->Name, "_GL_") == 0)
                {
                    /* We just created the mutex for the global lock, save it */

                    Gbl_GlobalLockSemaphore = ObjDesc->Mutex.Semaphore;
                }

                /* TBD: These fields may be obsolete */

                ObjDesc->Mutex.LockCount = 0;
                ObjDesc->Mutex.ThreadId  = 0;
                break;


            default:
                REPORT_ERROR ("Unsupported initial type value");
                CmRemoveReference (ObjDesc);
                ObjDesc = NULL;
                continue;
            }

            /* Store pointer to value descriptor in nte */
        
            NsAttachObject (NewEntry, ObjDesc, ObjDesc->Common.Type);
        }
    }


UnlockAndExit:
    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
NsLookup (
    ACPI_GENERIC_STATE      *ScopeInfo,
    char                    *Name, 
    ACPI_OBJECT_TYPE        Type, 
    OPERATING_MODE          InterpreterMode,
 UINT32     Flags,
    ACPI_WALK_STATE   *WalkState,
    NAME_TABLE_ENTRY        **RetEntry)
{
    NAME_TABLE_ENTRY        *PrefixScope;
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *EntryToSearch = NULL;
    NAME_TABLE_ENTRY        *ThisEntry = NULL;
    NAME_TABLE_ENTRY        *ScopeToPush = NULL;
    UINT32                  NumSegments;
    ACPI_OBJECT_TYPE        TypeToCheckFor;              /* Type To Check For */
    UINT32                  i;
    BOOLEAN                 NullNamePath = FALSE;


    FUNCTION_TRACE ("NsLookup");


    if (!RetEntry)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

   
    Gbl_NsLookupCount++;

    *RetEntry = ENTRY_NOT_FOUND;
    if (!Gbl_RootObject->Scope)
    {
        /* 
         * If the name space has not been initialized:
         * -  In Pass1 of Load mode, we need to initialize it
         *    before trying to define a name.
         * -  In Exec mode, there are no names to be found.
         */

        if (IMODE_LoadPass1 == InterpreterMode)
        {
            if ((Status = NsRootInitialize ()) != AE_OK)
            {
                return_ACPI_STATUS (Status);
            }
        }
        else
        {
            return_ACPI_STATUS (AE_NOT_FOUND);
        }
    }


    /* 
     * Get the prefix scope.
     * A null scope means use the root scope 
     */

    if ((!ScopeInfo) ||
        (!ScopeInfo->Scope.Entry))
    {
        PrefixScope = Gbl_RootObject->Scope;
    }
    else
    {
        PrefixScope = ScopeInfo->Scope.Entry;
    }


    /*
     * This check is explicitly split provide relax the TypeToCheckFor
     * conditions for BankFieldDefn.  Originally, both BankFieldDefn and
     * DefFieldDefn caused TypeToCheckFor to be set to ACPI_TYPE_Region,
     * but the BankFieldDefn may also check for a Field definition as well
     * as an OperationRegion.
     */

    if (INTERNAL_TYPE_DefFieldDefn == Type)
    {
        /* DefFieldDefn defines fields in a Region */

        TypeToCheckFor = ACPI_TYPE_Region;
    }

    else if (INTERNAL_TYPE_BankFieldDefn == Type)
    {
        /* BankFieldDefn defines data fields in a Field Object */

        TypeToCheckFor = ACPI_TYPE_Any;
    }

    else
    {
        TypeToCheckFor = Type;
    }


    /* Examine the name pointer */

    if (!Name)
    {
        /*  8-12-98 ASL Grammar Update supports null NamePath   */

        NullNamePath = TRUE;
        NumSegments = 0;
        ThisEntry = Gbl_RootObject;

        DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Null Name (Zero segments),  Flags=%x\n", Flags));
    }

    else
    {
        /* 
         * Valid name pointer (Internal name format)
         *
         * Check for prefixes.  As represented in the AML stream, a Name consists
         * of an optional scope prefix followed by a segment part.
         *
         * If present, the scope prefix is either a RootPrefix (in which case the
         * name is fully qualified), or zero or more ParentPrefixes (in which case
         * the name's scope is relative to the current scope).
         *
         * The segment part consists of either:
         *  - A single 4-byte name segment, or
         *  - A DualNamePrefix followed by two 4-byte name segments, or
         *  - A MultiNamePrefixOp, followed by a byte indicating the number of
         *    segments and the segments themselves.
         */

        if (*Name == AML_RootPrefix)
        {
            /* Name is fully qualified, look in root name table */
        
            EntryToSearch = Gbl_RootObject->Scope;
            Name++;                 /* point to segment part */

            DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Searching from root [%p]\n", 
                                        EntryToSearch));

            /* Direct reference to root, "\" */

            if (!(*Name))
            {
                *RetEntry = Gbl_RootObject;
                return_ACPI_STATUS (AE_OK);
            }
        }
    
        else
        {
            /* Name is relative to current scope, start there */
        
            EntryToSearch = PrefixScope;

            DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Searching relative to pfx scope [%p]\n",
                                        PrefixScope));

            /* Handle up-prefix (carat).  More than one prefix is supported */

            while (*Name == AML_ParentPrefix)
            {
            
                /* Point to segment part or next ParentPrefix */

                Name++;     

                /*  Backup to the parent's scope  */
            
                EntryToSearch = EntryToSearch->ParentEntry;
                if (!EntryToSearch)
                {
                    /* Current scope has no parent scope */
                
                    REPORT_ERROR ("NsLookup: Too many parent prefixes or scope has no parent");

                    return_ACPI_STATUS (AE_NOT_FOUND);
                }

                EntryToSearch = EntryToSearch->Scope;
            }
        }


        /* Examine the name prefix opcode, if any, to determine the number of segments */

        if (*Name == AML_DualNamePrefix)
        {
            NumSegments = 2;
            Name++;                             /* point to first segment */

            DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Dual Name (2 segments, Flags=%X)\n", Flags));
        }
    
        else if (*Name == AML_MultiNamePrefixOp)
        {
            NumSegments = (INT32)* (UINT8 *) ++Name;
            Name++;                             /* point to first segment */
        
            DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Multi Name (%d Segments, Flags=%X) \n", NumSegments, Flags));
        }

        else
        {
            /* 
             * No Dual or Multi prefix, hence there is only one
             * segment and Name is already pointing to it.
             */
            NumSegments = 1;

            DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Simple Name (1 segment, Flags=%X)\n", Flags));
        }


        /* Debug only: print the entire name that we are about to lookup */

        DEBUG_PRINT (TRACE_NAMES, ("NsLookup: ["));
        for (i = 0; i < NumSegments; i++)
        {
            DEBUG_PRINT_RAW (TRACE_NAMES, ("%4.4s/", &Name[i * 4]));
        }
        DEBUG_PRINT_RAW (TRACE_NAMES, ("]\n"));
    }


    /*
     * Search namespace for each segment of the name.
     * Loop through and verify/add each name segment.
     */


    while (NumSegments-- && EntryToSearch)
    {
        /* 
         * Search for the current segment in the table where it should be.
         * Type is significant only at the last level.
         */

        Status = NsSearchAndEnter (*(UINT32 *) Name, WalkState, EntryToSearch, InterpreterMode,
                                    NumSegments == 0 ? Type : ACPI_TYPE_Any, Flags, &ThisEntry);
        if (Status != AE_OK)
        {
            if (Status == AE_NOT_FOUND)
            {
                /* Name not in ACPI namespace  */

                if (IMODE_LoadPass1 == InterpreterMode || 
                    IMODE_LoadPass2 == InterpreterMode)
                {
                    REPORT_ERROR ("Name table overflow");
                }

                else
                {
                    DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Name not found in this scope\n"));
                }
            }

            return_ACPI_STATUS (Status);
        }

        if (NumSegments         == 0  &&                            /* If last segment                  */
            TypeToCheckFor      != ACPI_TYPE_Any &&                 /* and looking for a specific type  */
            TypeToCheckFor      != INTERNAL_TYPE_DefAny &&          /* which is not a local type        */
            TypeToCheckFor      != INTERNAL_TYPE_Scope &&           /*   "   "   "  "   "     "         */
            TypeToCheckFor      != INTERNAL_TYPE_IndexFieldDefn &&  /*   "   "   "  "   "     "         */
            ThisEntry->Type     != ACPI_TYPE_Any &&                 /* and type of entry is known       */
            ThisEntry->Type     != TypeToCheckFor)                  /* and entry does not match request */
        {                                                           /* complain.                        */
            /* Complain about type mismatch */

            REPORT_WARNING ("Type mismatch");
            DEBUG_PRINT (ACPI_WARN, ("NsLookup: %4.4s, type 0x%X, checking for type 0x%X\n", 
                                        Name, ThisEntry->Type, TypeToCheckFor));
        }

        /*
         * If last segment and not looking for a specific type, but type of
         * found entry is known, use that type to see if it opens a scope.
         */

        if ((0 == NumSegments) && (ACPI_TYPE_Any == Type))
        {
            Type = ThisEntry->Type;
        }

        if ((NumSegments || NsOpensScope (Type)) &&
            (ThisEntry->Scope == NULL))
        {
            /* 
             * More segments or the type implies enclosed scope, 
             * and the next scope has not been allocated.
             */

            DEBUG_PRINT (ACPI_INFO, ("NsLookup: Load mode=%d  ThisEntry=%x\n", InterpreterMode, ThisEntry));

            if ((IMODE_LoadPass1 == InterpreterMode) || 
                (IMODE_LoadPass2 == InterpreterMode))
            {   
                /* First or second pass load mode ==> locate the next scope */
                
                DEBUG_PRINT (TRACE_NAMES, ("NsLookup: Creating and adding a new scope\n"));
                ThisEntry->Scope = NsAllocateNameTable (NS_TABLE_SIZE);

                if (!ThisEntry->Scope)
                {
                    return_ACPI_STATUS (AE_NO_MEMORY);
                }
            }

            /* Now complain if there is no next scope */
            
            if (ThisEntry->Scope == NULL)
            {
                if (IMODE_LoadPass1 == InterpreterMode || 
                    IMODE_LoadPass2 == InterpreterMode)
                {
                    DEBUG_PRINT (ACPI_ERROR, ("NsLookup: ***Error - No child scope at entry %p\n", ThisEntry));
                    REPORT_ERROR ("Name Table allocation failure");
                    return_ACPI_STATUS (AE_NOT_FOUND);
                }

                DEBUG_PRINT (ACPI_INFO, ("NsLookup: No child scope at entry %p\n", ThisEntry));

                return_ACPI_STATUS (AE_NOT_FOUND);
            }


            /* Scope table initialization */

            if (IMODE_LoadPass1 == InterpreterMode || 
                IMODE_LoadPass2 == InterpreterMode)
            {
                /* Initialize the new table */

                NsInitializeTable (ThisEntry->Scope, EntryToSearch, ThisEntry);
            }
        }

        EntryToSearch = ThisEntry->Scope;
        Name += ACPI_NAME_SIZE;                 /* point to next name segment */
    }


    if (!(Flags & NS_DONT_OPEN_SCOPE) && (WalkState))
    {
        /* 
         * If entry is a type which opens a scope,
         * push the new scope on the scope stack.
         */

        if (NsOpensScope (TypeToCheckFor))
        {
            /*  8-12-98 ASL Grammar Update supports null NamePath   */

            if (NullNamePath)
            {
                ScopeToPush = ThisEntry;
            }
            else
            {
                ScopeToPush = ThisEntry->Scope;
            }

            Status = DsScopeStackPush (ScopeToPush, Type, WalkState);
            if (ACPI_FAILURE (Status))
            {
                return_ACPI_STATUS (Status);
            }

            DEBUG_PRINT (ACPI_INFO, ("NsLookup: Set global scope to %p\n", ScopeToPush));
        }
    }

    *RetEntry = ThisEntry;
    return_ACPI_STATUS (AE_OK);
}

