
/*
 *  ./acpica/subsystem/namespace/nsalloc.c - Namespace allocation and deletion utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __NSALLOC_C__

#include <acpi.h>
#include <namesp.h>
#include <interp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsalloc");




NAME_TABLE_ENTRY *
NsAllocateNameTable (
    INT32                   NteCount)
{
    NAME_TABLE_ENTRY        *NameTable = NULL;
    ACPI_SIZE               AllocSize;


    FUNCTION_TRACE ("NsAllocateNameTable");


    AllocSize = (ACPI_SIZE) NteCount * sizeof (NAME_TABLE_ENTRY);

    
    /* Allow room for link to appendage */
    
    AllocSize += sizeof (NAME_TABLE_ENTRY *);

  
    NameTable = CmCallocate (AllocSize);
    if (NameTable)
    {
        /* Move past the appendage pointer */
    
        NameTable = (NAME_TABLE_ENTRY *) (((UINT8 *) NameTable) + 
                        sizeof (NAME_TABLE_ENTRY *));
    }

    DEBUG_PRINT (TRACE_EXEC, ("NsAllocateNameTable: NameTable=%p\n", NameTable));

    return_VALUE (NameTable);
}



ACPI_STATUS
NsDeleteNamespaceSubtree (
    NAME_TABLE_ENTRY        *ParentHandle)
{
    NAME_TABLE_ENTRY        *ChildHandle;
    UINT32                  Level;
    ACPI_OBJECT_INTERNAL    *ObjDesc;


    FUNCTION_TRACE ("NsDeleteNamespaceSubtree");



    ChildHandle     = 0;
    Level           = 1;

    /* 
     * Traverse the tree of objects until we bubble back up 
     * to where we started.
     */

    while (Level > 0)
    {
        /* Get the next typed object in this scope.  Null returned if not found */

        ChildHandle = NsGetNextObject (ACPI_TYPE_Any, ParentHandle, ChildHandle);
        if (ChildHandle)
        {
            /* Found an object - delete the object within the Value field */

            ObjDesc = NsGetAttachedObject (ChildHandle);
            if (ObjDesc)
            {
                NsDetachObject (ChildHandle);
                CmRemoveReference (ObjDesc);
            }


            /* Clear the NTE in case this scope is reused (e.g., a control method scope) */

            ChildHandle->Type = ACPI_TYPE_Any;
            ChildHandle->Name = 0;

            /* Check if this object has any children */

            if (NsGetNextObject (ACPI_TYPE_Any, ChildHandle, 0))
            {
                /* There is at least one child of this object, visit the object */

                Level++;
                ParentHandle    = ChildHandle;
                ChildHandle     = 0;
            }

            else
            {
                /* There may be a name table even if there are no children */

                NsDeleteScope (((NAME_TABLE_ENTRY *) ChildHandle)->Scope);
                ((NAME_TABLE_ENTRY *) ChildHandle)->Scope = NULL;

            }
        }

        else
        {
            /* 
             * No more children in this object.  
             * We will move up to the grandparent.
             */
            Level--;

            /* Delete the scope (Name Table) associated with the parent object */
            /* Don't delete the top level scope, this allows the dynamic deletion of
             * objects created underneath control methods!
             */

            if (Level != 0)
            {
                NsDeleteScope (((NAME_TABLE_ENTRY *) ParentHandle)->Scope);
                ((NAME_TABLE_ENTRY *) ParentHandle)->Scope = NULL;
            }

            /* New "last child" is this parent object */

            ChildHandle = ParentHandle;

            /* Now we can move up the tree to the grandparent */

            ParentHandle = ParentHandle->ParentEntry;
        }
    }


    return_ACPI_STATUS (AE_OK); 
}





void
NsRemoveReference (
    NAME_TABLE_ENTRY        *Entry)
{
    NAME_TABLE_ENTRY        *ThisEntry;


    /* There may be a name table even if there are no children */

    NsDeleteScope (Entry->Scope);
    Entry->Scope = NULL;


    /* 
     * Decrement the reference count(s) of all parents up to the root,
     * And delete anything with zero remaining references.
     */
    ThisEntry = Entry;
    while (ThisEntry)
    {
        /* Decrement the reference */

        ThisEntry->ReferenceCount--;

        /* Delete entry if no more references */

        if (!ThisEntry->ReferenceCount)
        {
            /* Delete the scope if present */

            if (ThisEntry->Scope)
            {
                NsDeleteScope (ThisEntry->Scope);
                ThisEntry->Scope = NULL;
            }

            /* Mark the entry free  (This doesn't deallocate anything) */

            NsFreeTableEntry (ThisEntry);

        }

        /* Move up to parent */

        ThisEntry = ThisEntry->ParentEntry;
    }
}




ACPI_STATUS
NsDeleteNamespaceByOwner (
    UINT16                  OwnerId)
{
    NAME_TABLE_ENTRY        *ChildHandle;
    UINT32                  Level;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *ParentHandle;


    FUNCTION_TRACE ("NsDeleteNamespaceSubtree");


    ParentHandle    = Gbl_RootObject;
    ChildHandle     = 0;
    Level           = 1;

    /* 
     * Traverse the tree of objects until we bubble back up 
     * to where we started.
     */

    while (Level > 0)
    {
        /* Get the next typed object in this scope.  Null returned if not found */

        ChildHandle = NsGetNextObject (ACPI_TYPE_Any, ParentHandle, ChildHandle);
        if (ChildHandle)
        {
            if (ChildHandle->OwnerId == OwnerId)
            {
                /* Found an object - delete the object within the Value field */

                ObjDesc = NsGetAttachedObject (ChildHandle);
                if (ObjDesc)
                {
                    NsDetachObject (ChildHandle);
                    CmRemoveReference (ObjDesc);
                }
            }

            /* Check if this object has any children */

            if (NsGetNextObject (ACPI_TYPE_Any, ChildHandle, 0))
            {
                /* There is at least one child of this object, visit the object */

                Level++;
                ParentHandle    = ChildHandle;
                ChildHandle     = 0;
            }

            else if (ChildHandle->OwnerId == OwnerId)
            {
                NsRemoveReference (ChildHandle);
            }
        }

        else
        {
            /* 
             * No more children in this object.  
             * We will move up to the grandparent.
             */
            Level--;

            /* Delete the scope (Name Table) associated with the parent object */
            /* Don't delete the top level scope, this allows the dynamic deletion of
             * objects created underneath control methods!
             */

            if (Level != 0)
            {
                if (ParentHandle->OwnerId == OwnerId)
                {
                    NsRemoveReference (ParentHandle);
                }
            }


            /* New "last child" is this parent object */

            ChildHandle = ParentHandle;

            /* Now we can move up the tree to the grandparent */

            ParentHandle = ParentHandle->ParentEntry;
        }
    }


    return_ACPI_STATUS (AE_OK); 
}



void
NsDeleteScope (
    ACPI_HANDLE             Scope)
{
    NAME_TABLE_ENTRY        *ThisTable;
    char                    *AllocatedTable;


 FUNCTION_TRACE_PTR ("NsDeleteScope", Scope);


    if (!Scope)
    {
        return_VOID;
    }

    ThisTable = (NAME_TABLE_ENTRY *) Scope;


    /*
     * Deallocate the name table and all appendages
     */
    do
    {
        /* 
         * Create the original allocated pointer (backup to before appendage ptr),
         * and get the next appendage pointer.  This ptr is null at end-of-list.
         */

        AllocatedTable =  (((char *) ThisTable) - sizeof (NAME_TABLE_ENTRY *));
        ThisTable = NEXTSEG (ThisTable);

        DEBUG_PRINT (ACPI_INFO, ("NsDeleteScope: Deleting Name Table at %p \n", AllocatedTable));
        
        /* Now we can free the table */

        CmFree (AllocatedTable); 

    } while (ThisTable);
    
    return_VOID;
}


