
/*
 *  ./acpica/subsystem/namespace/nsapinam.c - Public interfaces to the ACPI subsystem
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSAPINAM_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <methods.h>
#include <amlcode.h>
#include <pnp.h>
#include <parser.h>
#include <dispatch.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsapinam");






ACPI_STATUS
AcpiLoadNamespace (
    void)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AcpiLoadNameSpace");


    /* There must be at least a DSDT installed */

    if (Gbl_DSDT == NULL)
    {
        DEBUG_PRINT (ACPI_ERROR, ("DSDT is not in memory\n"));
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }


    /* Init the hardware */
    /* TBD: Should this should be moved elsewhere, like AcpiEnable! ??*/
    /* TBD: we need to be able to call this interface repeatedly! */
    /* TBD: Does H/W require init before loading the namespace? */

    Status = CmHardwareInitialize ();
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* 
     * Load the namespace.  The DSDT is required,
     * but the SSDT and PSDT tables are optional.
     */

    Status = NsLoadTableByType (TABLE_DSDT);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Ignore exceptions from these */

    NsLoadTableByType (TABLE_SSDT);
    NsLoadTableByType (TABLE_PSDT);


    DEBUG_PRINT_RAW (ACPI_OK, ("ACPI Namespace successfully loaded at root 0x%p\n", 
                    Gbl_RootObject->Scope));


    /* Install the default OpRegion handlers, ignore the return code right now. */

    EvInstallDefaultAddressSpaceHandlers ();

    return_ACPI_STATUS (Status);
}




ACPI_STATUS 
AcpiGetHandle (
    ACPI_HANDLE             Parent, 
    ACPI_STRING             Pathname, 
    ACPI_HANDLE             *RetHandle)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ThisEntry;
    NAME_TABLE_ENTRY        *Scope = NULL;


    if (!RetHandle || !Pathname)
    {
        return AE_BAD_PARAMETER;
    }

    if (Parent)
    {   
        CmAcquireMutex (MTX_NAMESPACE);
        
        ThisEntry = NsConvertHandleToEntry (Parent);
        if (!ThisEntry)
        {   
            CmReleaseMutex (MTX_NAMESPACE);
            return AE_BAD_PARAMETER;
        }

        Scope = ThisEntry->Scope;
        CmReleaseMutex (MTX_NAMESPACE);
    }

    /* Special case for root, since we can't search for it */
    /* TBD: Check for both forward and backslash?? */

    if (STRCMP (Pathname, NS_ROOT_PATH) == 0)
    {
        *RetHandle = NsConvertEntryToHandle (Gbl_RootObject);
        return AE_OK;
    }

    /*
     *  Find the Nte and convert to the user format
     */
    ThisEntry = NULL;
    Status = NsGetNte (Pathname, Scope, &ThisEntry);

   *RetHandle = NsConvertEntryToHandle (ThisEntry);

    return (Status);
}



ACPI_STATUS 
AcpiGetName (
    ACPI_HANDLE             Handle, 
    UINT32                  NameType,
    ACPI_BUFFER             *RetPathPtr)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ObjEntry;


    /* Buffer pointer must be valid always */

    if (!RetPathPtr || (NameType > ACPI_NAME_TYPE_MAX))
    {
        return AE_BAD_PARAMETER;
    }

    /* Allow length to be zero and ignore the pointer */

    if ((RetPathPtr->Length) &&
       (!RetPathPtr->Pointer))
    {
        return AE_BAD_PARAMETER;
    }

    if (NameType == ACPI_FULL_PATHNAME)
    {
        /* Get the full pathname (From the namespace root) */

        Status = NsHandleToPathname (Handle, &RetPathPtr->Length, RetPathPtr->Pointer);
        return Status;
    }

    /* 
     * Wants the single segment ACPI name.  
     * Validate handle and convert to an NTE 
     */

    CmAcquireMutex (MTX_NAMESPACE);
    ObjEntry = NsConvertHandleToEntry (Handle);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

    /* Check if name will fit in buffer */

    if (RetPathPtr->Length < PATH_SEGMENT_LENGTH)
    {
        RetPathPtr->Length = PATH_SEGMENT_LENGTH;
        Status = AE_BUFFER_OVERFLOW;
        goto UnlockAndExit;
    }

    /* Just copy the ACPI name from the NTE and zero terminate it */

    STRNCPY (RetPathPtr->Pointer, (char *) &ObjEntry->Name, ACPI_NAME_SIZE);
    ((char *) RetPathPtr->Pointer) [ACPI_NAME_SIZE] = 0;
    Status = AE_OK;


UnlockAndExit:

    CmReleaseMutex (MTX_NAMESPACE);
    return Status;
}



ACPI_STATUS
AcpiGetObjectInfo (
    ACPI_HANDLE             Device, 
    ACPI_DEVICE_INFO        *Info)
{
    DEVICE_ID               Hid;
    DEVICE_ID               Uid;
    ACPI_STATUS             Status;
    UINT32                  DeviceStatus = 0;
    UINT32                  Address = 0;
    NAME_TABLE_ENTRY        *DeviceEntry;


    /* Parameter validation */

    if (!Device || !Info)
    {
        return AE_BAD_PARAMETER;
    }

    CmAcquireMutex (MTX_NAMESPACE);

    DeviceEntry = NsConvertHandleToEntry (Device);
    if (!DeviceEntry)
    {
        CmReleaseMutex (MTX_NAMESPACE);
        return AE_BAD_PARAMETER;
    }

    Info->Type      = DeviceEntry->Type;
    Info->Name      = DeviceEntry->Name;
    Info->Parent    = NsConvertEntryToHandle (DeviceEntry->ParentEntry);

    CmReleaseMutex (MTX_NAMESPACE);

    /*
     * If not a device, we are all done.
     */
    if (Info->Type != ACPI_TYPE_Device)
    {
        return AE_OK;
    }


    /* Get extra info for ACPI devices */

    Info->Valid     = 0;

    /* Execute the _HID method and save the result */

    Status = Execute_HID (DeviceEntry, &Hid);
    if (ACPI_SUCCESS (Status))
    {
        if (Hid.Type == STRING_PTR_DEVICE_ID)
        {
            STRCPY (Info->HardwareId, Hid.Data.StringPtr);
        }
        else
        {
            STRCPY (Info->HardwareId, Hid.Data.Buffer);
        }

        Info->Valid |= ACPI_VALID_HID;
    }

    /* Execute the _UID method and save the result */

    Status = Execute_UID (DeviceEntry, &Uid);
    if (ACPI_SUCCESS (Status))
    {
        if (Hid.Type == STRING_PTR_DEVICE_ID)
        {
            STRCPY (Info->UniqueId, Uid.Data.StringPtr);
        }
        else
        {
            STRCPY (Info->UniqueId, Uid.Data.Buffer);
        }

        Info->Valid |= ACPI_VALID_UID;
    }

    /* 
     * Execute the _STA method and save the result
     * _STA is not always present 
     */

    Status = Execute_STA (DeviceEntry, &DeviceStatus);
    if (ACPI_SUCCESS (Status))
    {
        Info->CurrentStatus = DeviceStatus;
        Info->Valid |= ACPI_VALID_STA;
    }

    /* 
     * Execute the _ADR method and save result if successful
     * _ADR is not always present
     */

    Status = EvaluateNumeric (METHOD_NAME__ADR, DeviceEntry, &Address);
    if (ACPI_SUCCESS (Status))
    {
        Info->Address = Address;
        Info->Valid |= ACPI_VALID_ADR;
    }

    return AE_OK;
}

