
/*
 *  ./acpica/subsystem/namespace/nsdump.c - table dumping routines for debug
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSDUMP_C__

#include <acpi.h>
#include <interp.h>
#include <namesp.h>
#include <tables.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsdump");



#ifdef ACPI_DEBUG


ACPI_STATUS
NsDumpPathname (
    ACPI_HANDLE             Handle, 
    char                    *Msg, 
    UINT32                  Level, 
    UINT32                  Component)
{
    char                    *Buffer;
    UINT32                  Length;

    
    FUNCTION_TRACE ("NsDumpPathname");

    /* Do this only if the requested debug level and component are enabled */

    if (!(DebugLevel & Level) || !(DebugLayer & Component))
    {
        return_ACPI_STATUS (AE_OK);
    }

    Buffer = CmAllocate (PATHNAME_MAX);
    if (!Buffer)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Convert handle to a full pathname and print it (with supplied message) */

    Length = PATHNAME_MAX;
    if (ACPI_SUCCESS (NsHandleToPathname (Handle, &Length, Buffer)))
    {
        OsdPrintf ("%s %s (%p)\n", Msg, Buffer, Handle);
    }

    CmFree (Buffer);

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
NsDumpOneObject (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_WALK_INFO          *Info = (ACPI_WALK_INFO *) Context;
    NAME_TABLE_ENTRY        *Appendage = NULL;
    NAME_TABLE_ENTRY        *ThisEntry;
    UINT8                   *Value;
    ACPI_OBJECT_INTERNAL    *ObjDesc = NULL;
    ACPI_OBJECT_TYPE        ObjType;
    ACPI_OBJECT_TYPE        Type;
    UINT32                  BytesToDump;
    UINT32                  DownstreamSiblingMask = 0;
    INT32                   LevelTmp;
    UINT32                  WhichBit;
    UINT32                  Size = 0;


    ThisEntry = NsConvertHandleToEntry (ObjHandle);

    LevelTmp    = Level;
    Type        = ThisEntry->Type;
    WhichBit    = 1;


    if (!(DebugLevel & Info->DebugLevel))
    {
        return AE_OK;
    }
    
    if (!ObjHandle)
    {
        DEBUG_PRINT (ACPI_INFO, ("Null object handle\n"));
        return AE_OK;
    }

    /* Check if the owner matches */

    if ((Info->OwnerId != ACPI_UINT32_MAX) &&
        (Info->OwnerId != ThisEntry->OwnerId))
    {
        return AE_OK;
    }


    /* Indent the object according to the level */

    while (LevelTmp--)
    {

        /* Print appropriate characters to form tree structure */

        if (LevelTmp)
        {
            if (DownstreamSiblingMask & WhichBit)
            {    
                DEBUG_PRINT_RAW (TRACE_TABLES, ("|"));
            }

            else
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, (" "));
            }
        
            WhichBit <<= 1;
        }
    
        else
        {
            if (NsExistDownstreamSibling (ThisEntry + 1, Size, Appendage))
            {
                DownstreamSiblingMask |= (1 << (Level - 1));
                DEBUG_PRINT_RAW (TRACE_TABLES, ("+"));
            }

            else
            {
                DownstreamSiblingMask &= 0xffffffff ^ (1 << (Level - 1));
                DEBUG_PRINT_RAW (TRACE_TABLES, ("+"));
            }

            if (ThisEntry->Scope == NULL)
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, ("-"));
            }
        
            else if (NsExistDownstreamSibling (ThisEntry->Scope, NS_TABLE_SIZE,
                                                NEXTSEG (ThisEntry->Scope)))
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, ("+"));
            }
        
            else
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, ("-"));
            }
        }
    }


    /* Check the integrity of our data */

    if (Type > INTERNAL_TYPE_MAX)
    {
        Type = INTERNAL_TYPE_DefAny;                                 /* prints as *ERROR* */
    }
    
    if (!CmValidAcpiName (*(UINT32 *) &ThisEntry->Name))
    {
        REPORT_WARNING ("Invalid Name");
    }

    /*
     * Now we can print out the pertinent information
     */

    DEBUG_PRINT_RAW (TRACE_TABLES,
                (" %4.4s %-9s ", &ThisEntry->Name, CmGetTypeName (Type)));

    DEBUG_PRINT_RAW (TRACE_TABLES, ("%p S:%p O:%p",
                ThisEntry,
                ThisEntry->Scope, 
                ThisEntry->Object));


    if (!ThisEntry->Object)
    {
        /* No attached object, we are done */

        DEBUG_PRINT_RAW (TRACE_TABLES, ("\n"));
        return AE_OK;
    }

    switch (Type)
    {

    case ACPI_TYPE_Method:

        /* Name is a Method and its AML offset/length are set */
        
        DEBUG_PRINT_RAW (TRACE_TABLES, (" M:%p-%X\n",
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->Method.Pcode,
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->Method.PcodeLength));                

        break;
    

    case ACPI_TYPE_Number:
 
        DEBUG_PRINT_RAW (TRACE_TABLES, (" N:%X\n",
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->Number.Value));
        break;


    case ACPI_TYPE_String:

        DEBUG_PRINT_RAW (TRACE_TABLES, (" S:%p-%X\n",
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->String.Pointer,
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->String.Length));
        break;


    case ACPI_TYPE_Buffer:

        DEBUG_PRINT_RAW (TRACE_TABLES, (" B:%p-%X\n",
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->Buffer.Pointer,
                    ((ACPI_OBJECT_INTERNAL *) ThisEntry->Object)->Buffer.Length));
        break;


    default:

        DEBUG_PRINT_RAW (TRACE_TABLES, ("\n"));
        break;
    }
  
    /* If debug turned off, done */

    if (!(DebugLevel & TRACE_VALUES))
    {
        return AE_OK;
    }


    /* If there is an attached object, display it */

    Value = ThisEntry->Object;

    /* Dump attached objects */

    while (Value)
    {
        ObjType = INTERNAL_TYPE_Invalid;

        /* Decode the type of attached object and dump the contents */

        DEBUG_PRINT_RAW (TRACE_TABLES, ("        Attached Object %p: ", Value));

        if (TbSystemTablePointer (Value))
        {
            DEBUG_PRINT_RAW (TRACE_TABLES, ("(Ptr to AML Code)\n"));
            BytesToDump = 16;
        }

        else if (VALID_DESCRIPTOR_TYPE (Value, DESC_TYPE_NTE))
        {
            DEBUG_PRINT_RAW (TRACE_TABLES, ("(Ptr to Name Table Entry)\n"));
            BytesToDump = sizeof (NAME_TABLE_ENTRY);
        }


        else if (VALID_DESCRIPTOR_TYPE (Value, DESC_TYPE_ACPI_OBJ))
        {
            ObjDesc = (ACPI_OBJECT_INTERNAL *) Value;
            ObjType = ObjDesc->Common.Type;

            if (ObjType > INTERNAL_TYPE_MAX)
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, ("(Ptr to ACPI Object type 0x%X [UNKNOWN])\n", ObjType));
                BytesToDump = 32;
            }

            else
            {
                DEBUG_PRINT_RAW (TRACE_TABLES, ("(Ptr to ACPI Object type 0x%X [%s])\n", 
                                    ObjType, CmGetTypeName (ObjType)));
                BytesToDump = ObjDesc->Common.Size;
            }
        }

        else
        {
            DEBUG_PRINT_RAW (TRACE_TABLES, ("(String or Buffer - not descriptor)\n", Value));
            BytesToDump = 16;
        }

        DUMP_BUFFER (Value, BytesToDump);

        /* If value is NOT an internal object, we are done */

        if ((TbSystemTablePointer (Value)) ||
            (VALID_DESCRIPTOR_TYPE (Value, DESC_TYPE_NTE)))
        {
            goto Cleanup;
        }

        /* 
         * Valid object, get the pointer to next level, if any
         */
        switch (ObjType)
        {
        case ACPI_TYPE_String:
            Value = (UINT8 *) ObjDesc->String.Pointer;
            break;

        case ACPI_TYPE_Buffer:
            Value = (UINT8 *) ObjDesc->Buffer.Pointer;
            break;

        case ACPI_TYPE_Package:
            Value = (UINT8 *) ObjDesc->Package.Elements;
            break;

        case ACPI_TYPE_Method:
            Value = (UINT8 *) ObjDesc->Method.Pcode;
            break;

        case ACPI_TYPE_FieldUnit:
            Value = (UINT8 *) ObjDesc->FieldUnit.Container;
            break;

        case INTERNAL_TYPE_DefField:
            Value = (UINT8 *) ObjDesc->Field.Container;
            break;

        case INTERNAL_TYPE_BankField:
            Value = (UINT8 *) ObjDesc->BankField.Container;
            break;

        case INTERNAL_TYPE_IndexField:
            Value = (UINT8 *) ObjDesc->IndexField.Index;
            break;

       default:
            goto Cleanup;
        }

        ObjType = INTERNAL_TYPE_Invalid;     /* Terminate loop after next pass */
    }

Cleanup:
    DEBUG_PRINT_RAW (TRACE_TABLES, ("\n"));
    return AE_OK;
}



void
NsDumpObjects (
    ACPI_OBJECT_TYPE        Type, 
    UINT32                  MaxDepth, 
    UINT32                  OwnerId,
    ACPI_HANDLE             StartHandle)
{
    ACPI_WALK_INFO          Info;


    Info.DebugLevel = TRACE_TABLES;
    Info.OwnerId = OwnerId;

    NsWalkNamespace (Type, StartHandle, MaxDepth, NS_WALK_NO_UNLOCK, NsDumpOneObject, 
                        (void *) &Info, NULL);
}



ACPI_STATUS
NsDumpOneDevice (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *Context,
    void                    **ReturnValue)
{
    ACPI_DEVICE_INFO        Info;
    ACPI_STATUS             Status;
    UINT32                  i;


    Status = NsDumpOneObject (ObjHandle, Level, Context, ReturnValue);

    Status = AcpiGetObjectInfo (ObjHandle, &Info);
    if (ACPI_SUCCESS (Status))
    {
        for (i = 0; i < Level; i++)
        {
            DEBUG_PRINT_RAW (TRACE_TABLES, (" "));
        }

        DEBUG_PRINT_RAW (TRACE_TABLES, ("    HID: %.8X, ADR: %.8X, Status: %x\n",
                        Info.HardwareId, Info.Address, Info.CurrentStatus));

    }

    return Status;
}



void
NsDumpRootDevices (void)
{
    ACPI_HANDLE             SysBusHandle;


    /* Only dump the table if tracing is enabled */

    if (!(TRACE_TABLES & DebugLevel))
    {
        return;
    }

    AcpiGetHandle (0, NS_SYSTEM_BUS, &SysBusHandle);

    DEBUG_PRINT (TRACE_TABLES, ("Display of all devices in the namespace:\n"));
    NsWalkNamespace (ACPI_TYPE_Device, SysBusHandle, ACPI_INT32_MAX, NS_WALK_NO_UNLOCK, 
                        NsDumpOneDevice, NULL, NULL);
}



void
NsDumpTables (
    ACPI_HANDLE             SearchBase, 
    INT32                   MaxDepth)
{
    ACPI_HANDLE             SearchHandle = SearchBase;


    FUNCTION_TRACE ("NsDumpTables");


    if (!Gbl_RootObject->Scope)
    {      
        /* 
         * If the name space has not been initialized,
         * there is nothing to dump.
         */
        DEBUG_PRINT (TRACE_TABLES, ("NsDumpTables: name space not initialized!\n"));
        return_VOID;
    }

    if (NS_ALL == SearchBase)
    {
        /*  entire namespace    */

        SearchHandle = Gbl_RootObject;
        DEBUG_PRINT (TRACE_TABLES, ("\\\n"));
    }


    NsDumpObjects (ACPI_TYPE_Any, MaxDepth, ACPI_UINT32_MAX, SearchHandle);
    return_VOID;
}



void
NsDumpEntry (
    ACPI_HANDLE             Handle,
    UINT32                  DebugLevel)
{
    ACPI_WALK_INFO          Info;


    FUNCTION_TRACE_PTR ("NsDumpEntry", Handle);

    Info.DebugLevel = DebugLevel;
    Info.OwnerId = ACPI_UINT32_MAX;

    NsDumpOneObject (Handle, 1, &Info, NULL);
    
    DEBUG_PRINT (TRACE_EXEC, ("leave NsDumpEntry %p\n", Handle));
    return_VOID;
}

#endif
 
