
/*
 *  ./acpica/subsystem/namespace/nseval.c - Object evaluation interfaces -- includes control
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSEVAL_C__

#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nseval");




ACPI_STATUS
NsEvaluateRelative (
    NAME_TABLE_ENTRY        *Handle, 
    char                    *Pathname, 
    ACPI_OBJECT_INTERNAL    **Params,
    ACPI_OBJECT_INTERNAL    **ReturnObject)
{
    NAME_TABLE_ENTRY        *RelObjEntry;
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ObjEntry = NULL;
    char                    *InternalPath = NULL;
    ACPI_GENERIC_STATE      ScopeInfo;


    FUNCTION_TRACE ("NsEvaluateRelative");


    /*
     * Must have a valid object handle
     */
    if (!Handle) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Build an internal name string for the method */

    Status = NsInternalizeName (Pathname, &InternalPath);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Get the prefix handle and NTE */

    CmAcquireMutex (MTX_NAMESPACE);

    RelObjEntry = NsConvertHandleToEntry (Handle);
    if (!RelObjEntry)
    {
        CmReleaseMutex (MTX_NAMESPACE);
        Status = AE_BAD_PARAMETER;
        goto Cleanup;
    }

    /* Lookup the name in the namespace */

    ScopeInfo.Scope.Entry = RelObjEntry->Scope;
    Status = NsLookup (&ScopeInfo, InternalPath, ACPI_TYPE_Any, IMODE_Execute, 
                                NS_NO_UPSEARCH, NULL, &ObjEntry);
    CmReleaseMutex (MTX_NAMESPACE);

    if (Status != AE_OK)
    {
        DEBUG_PRINT (ACPI_INFO, ("NsEvaluateRelative: Object [%s] not found [%.4X]\n",
                        InternalPath, Status));
        goto Cleanup;
    }

    /*
     * Now that we have a handle to the object, we can attempt
     * to evaluate it.
     */

    DEBUG_PRINT (ACPI_INFO, ("NsEvaluateRelative: %s [%p] Value %p\n",
                                Pathname, ObjEntry, ObjEntry->Object));

    Status = NsEvaluateByHandle (ObjEntry, Params, ReturnObject);

    DEBUG_PRINT (ACPI_INFO, ("NsEvaluateRelative: *** Completed eval of object %s ***\n",
                                Pathname));

Cleanup:

    /* Cleanup */

    CmFree (InternalPath);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
NsEvaluateByName (
    char                    *Pathname, 
    ACPI_OBJECT_INTERNAL    **Params,
    ACPI_OBJECT_INTERNAL    **ReturnObject)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ObjEntry = NULL;
    char                    *InternalPath = NULL;

    
    FUNCTION_TRACE ("NsEvaluateByName");


    /* Build an internal name string for the method */

    if (Pathname[0] != '\\' || Pathname[1] != '/')
    {
        Status = NsInternalizeName (Pathname, &InternalPath);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* Lookup the name in the namespace */

    Status = NsLookup (NULL, InternalPath, ACPI_TYPE_Any, IMODE_Execute, 
                                NS_NO_UPSEARCH, NULL, &ObjEntry);
    CmReleaseMutex (MTX_NAMESPACE);

    if (Status != AE_OK)
    {
        DEBUG_PRINT (ACPI_INFO, ("NsEvaluateByName: Object at [%s] was not found, status=%.4X\n",
                        InternalPath, Status));
        goto Cleanup;
    }

    /*
     * Now that we have a handle to the object, we can attempt
     * to evaluate it.
     */

    DEBUG_PRINT (ACPI_INFO, ("NsEvaluateByName: %s [%p] Value %p\n",
                                Pathname, ObjEntry, ObjEntry->Object));

    Status = NsEvaluateByHandle (ObjEntry, Params, ReturnObject);

    DEBUG_PRINT (ACPI_INFO, ("NsEvaluateByName: *** Completed eval of object %s ***\n",
                                Pathname));


Cleanup:

    /* Cleanup */

    if (InternalPath)
    {
        CmFree (InternalPath);
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
NsEvaluateByHandle (
    NAME_TABLE_ENTRY        *Handle, 
    ACPI_OBJECT_INTERNAL    **Params,
    ACPI_OBJECT_INTERNAL    **ReturnObject)
{
    NAME_TABLE_ENTRY        *ObjEntry; 
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *LocalReturnObject;


    FUNCTION_TRACE ("NsEvaluateByHandle");


    /* Check if namespace has been initialized */

    if (!Gbl_RootObject->Scope)
    {
        return_ACPI_STATUS (AE_NO_NAMESPACE);
    }

    /* Parameter Validation */

    if (!Handle)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    if (ReturnObject)
    {
        /* Initialize the return value to an invalid object */

        *ReturnObject = NULL;
    }

    /* Get the prefix handle and NTE */

    CmAcquireMutex (MTX_NAMESPACE);

    ObjEntry = NsConvertHandleToEntry (Handle);
    if (!ObjEntry)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }



    /*
     * Two major cases here:
     * 1) The object is an actual control method -- execute it.
     * 2) The object is not a method -- just return it's current value
     *
     * In both cases, the namespace is unlocked by the Ns* procedure
     */

    if (NsGetType (ObjEntry) == ACPI_TYPE_Method)
    {
        /* Case 1) We have an actual control method to execute */

        Status = NsExecuteControlMethod (ObjEntry, Params, &LocalReturnObject);
    }

    else
    {
        /* Case 2) Object is NOT a method, just return its current value */
    
        Status = NsGetObjectValue (ObjEntry, &LocalReturnObject);
    }


    /*
     * Check if there is a return value on the stack that must be dealt with 
     */

    if (Status == AE_CTRL_RETURN_VALUE)
    {
        /* 
         * If the Method returned a value and the caller provided a place
         * to store a returned value, Copy the returned value to the object
         * descriptor provided by the caller.
         */

        if (ReturnObject)
        {
            /* Valid return object, copy the pointer to the returned object */

            *ReturnObject = LocalReturnObject;
        }


        /* Map AE_RETURN_VALUE to AE_OK, we are done with it */

        if (Status == AE_CTRL_RETURN_VALUE)
        {
            Status = AE_OK;
        }
    }

    /* Namespace was unlocked by the handling Ns* function, so we just return */

    return_ACPI_STATUS (Status);


UnlockAndExit:

    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
NsExecuteControlMethod (
    NAME_TABLE_ENTRY        *MethodEntry, 
    ACPI_OBJECT_INTERNAL    **Params,
    ACPI_OBJECT_INTERNAL    **ReturnObjDesc)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    *ObjDesc;


    FUNCTION_TRACE ("NsExecuteControlMethod");


    /* Verify that there is a method associated with this object */

    ObjDesc = NsGetAttachedObject ((ACPI_HANDLE) MethodEntry);
    if (!ObjDesc)
    {
        DEBUG_PRINT (ACPI_ERROR, ("Control method is undefined (nil value)\n"));
        return_ACPI_STATUS (AE_ERROR);
    }

    /* 
     * Valid method, Set the current scope to that of the Method, and execute it.
     */

    DEBUG_PRINT (ACPI_INFO, ("Control method at Offset %x Length %lx]\n",
                    ObjDesc->Method.Pcode + 1,
                    ObjDesc->Method.PcodeLength - 1));

    DUMP_PATHNAME (MethodEntry, "NsExecuteControlMethod: Executing", 
                    TRACE_NAMES, _COMPONENT);

    DEBUG_PRINT (TRACE_NAMES, ("At offset %8XH\n", ObjDesc->Method.Pcode + 1));

   
    /* 
     * Unlock the namespace before execution.  This allows namespace access
     * via the external Acpi* interfaces while a method is being executed.
     * However, any namespace deletion must acquire both the namespace and
     * interpter locks to ensure that no thread is using the portion of the
     * namespace that is being deleted.
     */

    CmReleaseMutex (MTX_NAMESPACE);

    /* 
     * Excecute the method via the interpreter
     */
    Status = AmlExecuteMethod (MethodEntry, Params, ReturnObjDesc);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
NsGetObjectValue (
    NAME_TABLE_ENTRY        *ObjectEntry,
    ACPI_OBJECT_INTERNAL    **ReturnObjDesc)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *ValDesc;


    FUNCTION_TRACE ("NsGetObjectValue");


    /*
     *  We take the value from certain objects directly
     */

    if ((ObjectEntry->Type == ACPI_TYPE_Processor) ||
        (ObjectEntry->Type == ACPI_TYPE_Power)) 
    {

        /*
         *  Create a Reference object to contain the object
         */
        ObjDesc = CmCreateInternalObject (ObjectEntry->Type);
        if (!ObjDesc)
        {
           Status = AE_NO_MEMORY;
           goto UnlockAndExit;
        }

        /*
         *  Get the attached object
         */

        ValDesc = NsGetAttachedObject (ObjectEntry);
        if (!ValDesc)
        {
            Status = AE_NULL_OBJECT;
            goto UnlockAndExit;
        }

        /*
         *  Just copy from the original to the return object
         */

        MEMCPY (&ObjDesc->Common.FirstNonCommonByte, &ValDesc->Common.FirstNonCommonByte,
                (sizeof(ACPI_OBJECT_Common) - sizeof(ObjDesc->Common.FirstNonCommonByte)));
    } 
    

    /*
     * Other objects require a reference object wrapper which we then attempt to resolve.
     */
    else 
    {       
        /* Create an Reference object to contain the object */

        ObjDesc = CmCreateInternalObject (INTERNAL_TYPE_Reference);
        if (!ObjDesc)
        {
           Status = AE_NO_MEMORY;
           goto UnlockAndExit;
        }

        /* Construct a descriptor pointing to the name */

        ObjDesc->Reference.OpCode  = (UINT8) AML_NameOp;
        ObjDesc->Reference.Object  = (void *) ObjectEntry;

        /* 
         * Use AmlResolveToValue() to get the associated value.  The call to AmlResolveToValue causes 
         * ObjDesc (allocated above) to always be deleted.
         */

        Status = AmlResolveToValue (&ObjDesc);
    }

    /* 
     * If AmlResolveToValue() succeeded, the return value was placed in ObjDesc.
     */

    if (Status == AE_OK)
    {
        Status = AE_CTRL_RETURN_VALUE;

        *ReturnObjDesc = ObjDesc;
        DEBUG_PRINT (ACPI_INFO, ("NsGetObjectValue: Returning obj %p\n", *ReturnObjDesc));
    }


UnlockAndExit:

    /* Unlock the namespace */

    CmReleaseMutex (MTX_NAMESPACE);
    return_ACPI_STATUS (Status);
}
