
/*
 *  ./acpica/subsystem/namespace/nsload.c - namespace loading procedures
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSLOAD_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <namesp.h>
#include <methods.h>
#include <amlcode.h>
#include <pnp.h>
#include <parser.h>
#include <dispatch.h>
#include <debugger.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsload");



ACPI_STATUS
NsParseTable (
    ACPI_TABLE_DESC         *TableDesc,
    NAME_TABLE_ENTRY        *Scope)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("NsParseTable");


    /* Create the root object */

    Gbl_ParsedNamespaceRoot = PsAllocOp (AML_ScopeOp);
    if (!Gbl_ParsedNamespaceRoot)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Initialize the root object */

    ((ACPI_NAMED_OP *) Gbl_ParsedNamespaceRoot)->Name = * (UINT32 *) NS_ROOT;

    /* Pass 1:  Parse everything except control method bodies */

    Status = PsParseAml (Gbl_ParsedNamespaceRoot, TableDesc->AmlPointer, TableDesc->AmlLength); 
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

#ifndef PARSER_ONLY

    DEBUG_PRINT (TRACE_PARSE, ("NsParseTable: Building Internal Namespace\n"));
BREAKPOINT3;

/* TBD: OBSOLETE, pass 1 is now part of the parse*/
/*
    PsWalkParsedAml (PsGetChild (Gbl_ParsedNamespaceRoot), Gbl_ParsedNamespaceRoot, NULL, Scope, NULL, NULL,
                        TableDesc->TableId, DsLoad1BeginOp, DsLoad1EndOp);
*/

    Status = PsWalkParsedAml (PsGetChild (Gbl_ParsedNamespaceRoot), Gbl_ParsedNamespaceRoot, NULL, Scope, NULL, NULL,
                        TableDesc->TableId, DsLoad2BeginOp, DsLoad2EndOp);


    /*
     * Now that the internal namespace has been constructed, we can delete the
     * parsed namespace, since it is no longer needed
     */

    DEBUG_PRINT (TRACE_PARSE, ("NsParseTable: Deleting Parsed Namespace\n"));
BREAKPOINT3;


    DEBUG_EXEC (DbGenerateStatistics (Gbl_ParsedNamespaceRoot, 0));

    PsDeleteParseTree (Gbl_ParsedNamespaceRoot);
    Gbl_ParsedNamespaceRoot = NULL;
#endif


    return_ACPI_STATUS (Status);
}





ACPI_STATUS
NsLoadTable (
    ACPI_TABLE_DESC         *TableDesc,
    NAME_TABLE_ENTRY        *Entry)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("NsLoadTable");


    if (!TableDesc->AmlPointer)
    {
        DEBUG_PRINT (ACPI_ERROR, ("NsLoadTable: Null AML pointer\n"));
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    DEBUG_PRINT (ACPI_INFO, ("NsLoadTable: AML block at %p\n", TableDesc->AmlPointer));


    if (!TableDesc->AmlLength)
    {
        DEBUG_PRINT (ACPI_ERROR, ("NsLoadTable: Zero-length AML block\n"));
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /*
     * Parse the table and load the namespace with all named objects found within.  
     * Control methods are NOT parsed at this time.  In fact, the control methods 
     * cannot be parsed until the entire namespace is loaded, because if a control 
     * method makes a forward reference (call) to another control method, we can't 
     * continue parsing because we don't know how many arguments to parse next!
     */

    DEBUG_PRINT (ACPI_INFO, ("NsLoadTable: **** Loading table into namespace ****\n"));

    CmAcquireMutex (MTX_NAMESPACE);
    Status = NsParseTable (TableDesc, Entry->Scope);
    CmReleaseMutex (MTX_NAMESPACE);

    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /*
     * Now we can parse the control methods.  We always parse them here for a sanity
     * check, and if configured for just-in-time parsing, we delete the control method
     * parse trees.
     */

    DEBUG_PRINT (ACPI_INFO, ("NsLoadTable: **** Begin Table Method Parsing and Object Initialization ****\n"));
BREAKPOINT3;

    Status = DsInitializeObjects (TableDesc, Entry);

    DEBUG_PRINT (ACPI_INFO, ("NsLoadTable: **** Completed Table Method Parsing and Object Initialization ****\n"));
BREAKPOINT3;

    return_ACPI_STATUS (Status);
}




ACPI_STATUS
NsLoadTableByType (
    ACPI_TABLE_TYPE         TableType)
{
    UINT32                  i;
    ACPI_STATUS             Status = AE_OK;
    ACPI_TABLE_HEADER       *TablePtr;
    ACPI_TABLE_DESC         *TableDesc;

    
    FUNCTION_TRACE ("NsLoadTableByType");


    CmAcquireMutex (MTX_TABLES);


    /*
     * Table types supported are:
     * DSDT (one), SSDT/PSDT (multiple)
     */

    switch (TableType)
    {

    case TABLE_DSDT:

        DEBUG_PRINT (ACPI_INFO, ("NsLoadTableByType: Loading DSDT\n"));

        TableDesc = &Gbl_AcpiTables[TABLE_DSDT];

        /* If table already loaded into namespace, just return */

        if (TableDesc->LoadedIntoNamespace)
        {
            goto UnlockAndExit;
        }

        TableDesc->TableId = TABLE_ID_DSDT;

        /* Initialize the root of the namespace tree */
    
        Status = NsRootInitialize ();
        if (ACPI_FAILURE (Status))
        {
            goto UnlockAndExit;
        }

        /* Now load the single DSDT */

        Status = NsLoadTable (TableDesc, Gbl_RootObject);
        if (ACPI_SUCCESS (Status))
        {
            TableDesc->LoadedIntoNamespace = TRUE;
        }

        break;


    case TABLE_SSDT:

        DEBUG_PRINT (ACPI_INFO, ("NsLoadTableByType: Loading %d SSDTs\n",
                        Gbl_AcpiTables[TABLE_SSDT].Count));

        /*
         * Traverse list of SSDT tables
         */

        TableDesc = &Gbl_AcpiTables[TABLE_SSDT];
        for (i = 0; i < Gbl_AcpiTables[TABLE_SSDT].Count; i++)
        {
            TablePtr = TableDesc->Pointer;

            /* Only attempt to load table if it is not already loaded! */

            if (!TableDesc->LoadedIntoNamespace)
            {
                Status = NsLoadTable (TableDesc, Gbl_RootObject);
                if (ACPI_FAILURE (Status))
                {
                    break;
                }

                TableDesc->LoadedIntoNamespace = TRUE;
            }

            TableDesc = TableDesc->Next;
        }

        break;


    case TABLE_PSDT:

        DEBUG_PRINT (ACPI_INFO, ("NsLoadTableByType: Loading %d PSDTs\n",
                        Gbl_AcpiTables[TABLE_PSDT].Count));

        /*
         * Traverse list of PSDT tables
         */

        TableDesc = &Gbl_AcpiTables[TABLE_PSDT];
        for (i = 0; i < Gbl_AcpiTables[TABLE_PSDT].Count ; i++)
        {
            TablePtr = TableDesc->Pointer;

            /* Only attempt to load table if it is not already loaded! */

            if (!TableDesc->LoadedIntoNamespace)
            {
                Status = NsLoadTable (TableDesc, Gbl_RootObject);
                if (ACPI_FAILURE (Status))
                {
                    break;
                }

                TableDesc->LoadedIntoNamespace = TRUE;
            }

            TableDesc = TableDesc->Next;
        }

        break;


    default:
        Status = AE_SUPPORT;
    }


UnlockAndExit:

    CmReleaseMutex (MTX_TABLES);

    return_ACPI_STATUS (Status);

}





