
/*
 *  ./acpica/subsystem/namespace/nsmodify.c - functions for expanding/contracting the namespace
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSMODIFY_C__

#include <acpi.h>
#include <amlcode.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsmodify");




void
NsFreeTableEntry (
    NAME_TABLE_ENTRY        *Entry)
{
 FUNCTION_TRACE ("NsFreeTableEntry");


    if (!Entry)
    {
        return_VOID;
    }

    /*
     * Need to delete
     * 1) The scope, if any
     * 2) An attached object, if any
     */

    if (Entry->Scope)
    {
        CmFree (Entry->Scope);
        Entry->Scope = NULL;
    }

    if (Entry->Object)
    {
        NsDetachObject (Entry->Object);
        Entry->Object = NULL;
    }


    /* Unlink the NTE from the table */

    if (Entry->PrevEntry)
    {
        Entry->PrevEntry->NextEntry = Entry->NextEntry;
    }

    if (Entry->NextEntry)
    {
        Entry->NextEntry->PrevEntry = Entry->PrevEntry;
    }


    /* Mark the entry unallocated */

    Entry->Name = 0;
 
    return_VOID;
}



ACPI_STATUS
NsDeleteSubtree (
    ACPI_HANDLE             StartHandle)
{
    ACPI_STATUS             Status;
    ACPI_HANDLE             ChildHandle;
    ACPI_HANDLE             ParentHandle;
    ACPI_HANDLE             NextChildHandle;
    ACPI_HANDLE             Dummy;
    UINT32                  Level;


    FUNCTION_TRACE ("NsDeleteSubtree");


    ParentHandle    = StartHandle;
    ChildHandle     = 0;
    Level           = 1;

    /* 
     * Traverse the tree of objects until we bubble back up 
     * to where we started.
     */

    while (Level > 0)
    {
        /* Attempt to get the next object in this scope */

        Status = AcpiGetNextObject (ACPI_TYPE_Any, ParentHandle, ChildHandle, &NextChildHandle);

        /*
         * Regardless of the success or failure of the previous operation, 
         * we are done with the previous object (if there was one), and any
         * children it may have had.  So we can now safely delete it (and 
         * its scope, if any)
         */

        NsFreeTableEntry (ChildHandle);
        ChildHandle = NextChildHandle;


        /* Did we get a new object? */

        if (ACPI_SUCCESS (Status))
        {
            /* Check if this object has any children */

            if (ACPI_SUCCESS (AcpiGetNextObject (ACPI_TYPE_Any, ChildHandle, 0, &Dummy)))
            {
                /* There is at least one child of this object, visit the object */

                Level++;
                ParentHandle    = ChildHandle;
                ChildHandle     = 0;
            }
        }

        else
        {
            /* 
             * No more children in this object, go back up to the object's parent
             */
            Level--;
            ChildHandle = ParentHandle;
            AcpiGetParent (ParentHandle, &ParentHandle);
        }
    }

    /* Now delete the starting object, and we are done */

    NsFreeTableEntry ((NAME_TABLE_ENTRY *) ChildHandle);


    return_ACPI_STATUS (AE_OK);
}




ACPI_STATUS
NsUnloadNamespace (
    ACPI_HANDLE             Handle)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("NsUnloadNameSpace");


    /* Parameter validation */

    if (!Gbl_RootObject->Scope)
    {
        return_ACPI_STATUS (AE_NO_NAMESPACE);
    }

    if (!Handle)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* This function does the real work */

    Status = NsDeleteSubtree (Handle);

    return_ACPI_STATUS (Status);
}




