
/*
 *  ./acpica/subsystem/namespace/nsnames.c - Name manipulation and search
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSNAMES_C__

#include <acpi.h>
#include <amlcode.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsnames");


        

char *
NsNameOfScope (
    NAME_TABLE_ENTRY        *EntryToSearch)
{
    NAME_TABLE_ENTRY        *Temp = NULL;
    char                    *NameBuffer;
    ACPI_SIZE               Size;


    FUNCTION_TRACE ("NsNameOfScope");


    if (!Gbl_RootObject->Scope || !EntryToSearch)
    {
        /* 
         * If the name space has not been initialized,
         * this function should not have been called.
         */
        return_VALUE (NULL);
    }

    /* Calculate required buffer size based on depth below root NT */
    
    for (Size = 1, Temp = EntryToSearch;
            Temp->ParentEntry;
            Temp = Temp->ParentEntry)
    {
        Size += ACPI_NAME_SIZE;
    }


    /* Allocate the buffer */

    NameBuffer = CmCallocate (Size + 1);
    if (!NameBuffer)
    {
        REPORT_ERROR ("NsNameOfScope: allocation failure");
        return_VALUE (NULL);
    }
    

    /* Store terminator byte, then build name backwards */
    
    NameBuffer[Size] = '\0';
    while ((Size > ACPI_NAME_SIZE) && EntryToSearch->ParentEntry)
    {
        Size -= ACPI_NAME_SIZE;
        *(UINT32 *) (NameBuffer + Size) = NsFindParentName (EntryToSearch);
        EntryToSearch = EntryToSearch->ParentEntry;
    }

    NameBuffer[--Size] = AML_RootPrefix;

    if (Size != 0)
    {
        DEBUG_PRINT (ACPI_ERROR, ("NsNameOfScope:  Bad pointer returned; size = %d\n", Size));
    }

    return_VALUE (NameBuffer);
}



char *
NsNameOfCurrentScope (
    ACPI_WALK_STATE         *WalkState)
{
    char                    *ScopeName;


    FUNCTION_TRACE ("NsNameOfCurrentScope");


    if (WalkState && WalkState->ScopeInfo)
    {
        ScopeName = NsNameOfScope (WalkState->ScopeInfo->Scope.Entry);
        return_VALUE (ScopeName);
    }
    
    REPORT_ERROR ("Current scope pointer is invalid");

    return_VALUE (NULL);
}



ACPI_STATUS
NsHandleToPathname (
    ACPI_HANDLE             TargetHandle, 
    UINT32                  *BufSize, 
    char                    *UserBuffer)
{
    ACPI_STATUS             Status = AE_OK;
    NAME_TABLE_ENTRY        *EntryToSearch = NULL;
    NAME_TABLE_ENTRY        *Temp = NULL;
    ACPI_SIZE               PathLength = 0;
    ACPI_SIZE               Size;
    UINT32                  UserBufSize;
    BOOLEAN                 NamespaceWasLocked;


    FUNCTION_TRACE_PTR ("NsHandleToPathname", TargetHandle);


    if (!Gbl_RootObject->Scope || !TargetHandle)
    {
        /* 
         * If the name space has not been initialized,
         * this function should not have been called.
         */

        return_ACPI_STATUS (AE_NO_NAMESPACE);
    }

    NamespaceWasLocked = Gbl_AcpiMutexInfo[MTX_NAMESPACE].Locked;
    if (!NamespaceWasLocked)
    {
        CmAcquireMutex (MTX_NAMESPACE);
    }

    EntryToSearch = NsConvertHandleToEntry (TargetHandle);
    if (!EntryToSearch)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* 
     * Compute length of pathname as 5 * number of name segments.
     * Go back up the parent tree to the root
     */
    for (Size = 0, Temp = EntryToSearch;
          Temp->ParentEntry;
          Temp = Temp->ParentEntry)
    {
        Size += PATH_SEGMENT_LENGTH;
    }

    /* Set return length to the required path length */

    PathLength = Size + 1;
    UserBufSize = *BufSize;
    *BufSize = PathLength;

    /* Check if the user buffer is sufficiently large */

    if (PathLength > UserBufSize)
    {
        Status = AE_BUFFER_OVERFLOW;
        goto UnlockAndExit;
    }
          
    /* Store null terminator */
    
    UserBuffer[Size] = '\0';
    Size -= ACPI_NAME_SIZE;

    /* Put the original ACPI name at the end of the path */

    *(UINT32 *) (UserBuffer + Size) = EntryToSearch->Name;
    UserBuffer[--Size] = PATH_SEPARATOR;

    /* Build name backwards, putting "." between segments */
    
    while ((Size > ACPI_NAME_SIZE) && EntryToSearch)
    {
        Size -= ACPI_NAME_SIZE;
        *(UINT32 *) (UserBuffer + Size) = NsFindParentName (EntryToSearch);
        
        UserBuffer[--Size] = PATH_SEPARATOR;
        EntryToSearch = EntryToSearch->ParentEntry;
    }

    /* Overlay the "." preceding the first segment with the root name "\" */
    
    UserBuffer[Size] = '\\';

    DEBUG_PRINT (TRACE_EXEC, ("NsHandleToPathname: Len=%d, %s \n", 
                                PathLength, UserBuffer));


UnlockAndExit:

    if (!NamespaceWasLocked)
    {
        CmReleaseMutex (MTX_NAMESPACE);
    }

    return_ACPI_STATUS (Status);
}




BOOLEAN
NsPatternMatch (
    NAME_TABLE_ENTRY        *ObjEntry, 
    char                    *SearchFor)
{
    INT32                   i;


    for (i = 0; i < ACPI_NAME_SIZE; i++)
    {
        if (SearchFor[i] != '?' &&
            SearchFor[i] != ((char *) &ObjEntry->Name)[i])
        {
            /* No match */

            return FALSE;
        }
    }

    /* name matches pattern */

    return TRUE;
}



ACPI_STATUS
NsNameCompare (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *Context,
    void                    **ReturnValue)
{
    FIND_CONTEXT            *Find = Context;


    /* Match, yes or no? */

    if (NsPatternMatch ((NAME_TABLE_ENTRY *) ObjHandle, Find->SearchFor))
    {
        /* Name matches pattern */
        
        if (Find->List)
        {
            DEBUG_PRINT (TRACE_NAMES, ("FindName: Match found: %.4s\n",
                            &((NAME_TABLE_ENTRY *) ObjHandle)->Name));

            Find->List[*(Find->Count)] = ObjHandle;
        }
        
        ++*(Find->Count);
    }

    return AE_OK;        /* Don't terminate the walk */
}



void
NsLowFindNames (
    NAME_TABLE_ENTRY        *ThisEntry, 
    char                    *SearchFor,
    INT32                   *Count, 
    ACPI_HANDLE             List[], 
    INT32                   MaxDepth)
{
    FIND_CONTEXT            Find;


    FUNCTION_TRACE ("NsLowFindNames");


    if (0 == MaxDepth || !ThisEntry || !SearchFor || !Count)
    {
        /* 
         * Zero requested depth, nothing to search,
         * nothing to search for, or count pointer bad 
         */
        
        return_VOID;
    }

    /* Init the context structure used by compare routine */

    Find.List       = List;
    Find.Count      = Count;
    Find.SearchFor  = SearchFor;

    /* Walk the namespace and find all matches */

    NsWalkNamespace (ACPI_TYPE_Any, (ACPI_HANDLE) ThisEntry, MaxDepth, NS_WALK_NO_UNLOCK, 
                        NsNameCompare, &Find, NULL);

    if (List)
    {
        /* null-terminate the output array */
        
        List[*Count] = (ACPI_HANDLE) 0;
    }

    return_VOID;
}



ACPI_HANDLE *
NsFindNames (
    char                    *SearchFor, 
    ACPI_HANDLE             StartHandle, 
    INT32                   MaxDepth)
{
    ACPI_HANDLE             *List = NULL;
    INT32                   Count;


    FUNCTION_TRACE ("NsFindNames");


    if (!Gbl_RootObject->Scope)
    {
        /* 
         * If the name space has not been initialized,
         * there surely are no matching names.
         */
        return_VALUE (NULL);
    }

    if (NS_ALL == StartHandle)
    {
        /* base is root */

        StartHandle = Gbl_RootObject;
    }
    
    else if (((NAME_TABLE_ENTRY *) StartHandle)->Scope)
    {
        /* base has children to search */

        StartHandle = ((NAME_TABLE_ENTRY *) StartHandle)->Scope;
    }
    
    else
    {
        /* 
         * If base is not the root and has no children,
         * there is nothing to search.
         */
        return_VALUE (NULL);
    }

    if (!SearchFor)
    {
        /* Search name not specified */

        SearchFor = "????";
    }

    
    /* Pass 1.  Get required buffer size, don't try to build list */
    
    Count = 0;
    NsLowFindNames (StartHandle, SearchFor, &Count, NULL, MaxDepth);
    
    if (0 == Count)
    {
        return_VALUE (NULL);
    }

    Count++;                                            /* Allow for trailing null */
    List = CmCallocate (Count * sizeof(ACPI_HANDLE));
    if (!List)
    {
        REPORT_ERROR ("NsFindNames: allocation failure");
        return_VALUE (NULL);
    }

    /* Pass 2.  Fill buffer */

    Count = 0;
    NsLowFindNames (StartHandle, SearchFor, &Count, List, MaxDepth);

    return_VALUE (List);
}



