
/*
 *  ./acpica/subsystem/namespace/nsobject.c - Utilities for objects attached to namespace table entries
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __NSOBJECT_C__

#include <acpi.h>
#include <amlcode.h>
#include <namesp.h>
#include <interp.h>
#include <tables.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsobject");




ACPI_STATUS
NsAttachObject (
    ACPI_HANDLE             Handle, 
    ACPI_HANDLE             Object, 
    UINT8                   Type)
{
    NAME_TABLE_ENTRY        *ThisEntry = (NAME_TABLE_ENTRY *) Handle;
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *PreviousObjDesc;
    ACPI_OBJECT_TYPE        ObjType = ACPI_TYPE_Any;
    UINT8                   Flags;


    FUNCTION_TRACE ("NsAttachObject");


    /*
     * Parameter validation
     */

    if (!Gbl_RootObject->Scope)
    {
        /* Name space not initialized  */

        REPORT_ERROR ("NsAttachObject: Name space not initialized");
        return_ACPI_STATUS (AE_NO_NAMESPACE);
    }
    
    if (!Handle)
    {
        /* Invalid handle */

        REPORT_ERROR ("NsAttachObject: Null name handle");
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    if (!Object && (ACPI_TYPE_Any != Type))
    {
        /* Null object */

        REPORT_ERROR ("NsAttachObject: Null object, but type not ACPI_TYPE_Any");
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    if (!VALID_DESCRIPTOR_TYPE (Handle, DESC_TYPE_NTE))
    {
        /* Not a name handle */

        REPORT_ERROR ("NsAttachObject: Invalid handle");
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Check if this object is already attached */

    if (ThisEntry->Object == Object)
    {
        DEBUG_PRINT (TRACE_EXEC,("NsAttachObject: Obj %p already installed in NTE %p\n",
                        Object, Handle));
    
        return_ACPI_STATUS (AE_OK);
    }


    /* Get the current flags field of the NTE */

    Flags = ThisEntry->Flags;
    Flags &= ~NTE_AML_ATTACHMENT;


    /* If null object, we will just install it */

    if (!Object)
    {
        ObjDesc = NULL;
        ObjType = ACPI_TYPE_Any;
    }

    /*
     * If the object is an NTE with an attached object, 
     * we will use that (attached) object
     */

    else if (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_NTE) && 
            ((NAME_TABLE_ENTRY *) Object)->Object)
    {
        /* 
         * Value passed is a name handle and that name has a non-null value.
         * Use that name's value and type.
         */

        ObjDesc = ((NAME_TABLE_ENTRY *) Object)->Object;
        ObjType = ((NAME_TABLE_ENTRY *) Object)->Type;

        /*
         * Copy appropriate flags
         */

        if (((NAME_TABLE_ENTRY *) Object)->Flags & NTE_AML_ATTACHMENT)
        {
            Flags |= NTE_AML_ATTACHMENT;
        }
    }


    /*
     * Otherwise, we will use the parameter object, but we must type it first
     */

    else
    {
        ObjDesc = (ACPI_OBJECT_INTERNAL *) Object;


        /* If a valid type (non-ANY) was given, just use it */

        if (ACPI_TYPE_Any != Type)
        {
            ObjType = (ACPI_OBJECT_TYPE) Type;
        }

    
        /* Type is TYPE_Any, we must try to determinte the actual type of the object */

        /*
         * Check if value points into the AML code
         */
        else if (TbSystemTablePointer (Object))
        {
            /* Object points into the AML stream.  Set a flag bit in the NTE to indicate this */
    
            Flags |= NTE_AML_ATTACHMENT;


            /* Check for a recognized OpCode */

            switch (*(UINT8 *) Object)
            {

            case AML_OpPrefix:

                if (*(UINT16 *) Object != AML_RevisionOp)
                {
                    /* OpPrefix is unrecognized unless part of RevisionOp */
            
                    break;
                }

                /* Else fall through to set type as Number */
    

            case AML_ZeroOp: case AML_OnesOp: case AML_OneOp:
            case AML_ByteOp: case AML_WordOp: case AML_DWordOp:

                ObjType = ACPI_TYPE_Number;
                break;


            case AML_StringOp:

                ObjType = ACPI_TYPE_String;
                break;


            case AML_BufferOp:

                ObjType = ACPI_TYPE_Buffer;
                break;


            case AML_MutexOp:

                ObjType = ACPI_TYPE_Mutex;
                break;


            case AML_PackageOp:

                ObjType = ACPI_TYPE_Package;
                break;


            default:

                DEBUG_PRINT (ACPI_ERROR, ("AML Opcode/Type [%x] not supported in attach\n",
                               *(UINT16 *) Object));

                return_ACPI_STATUS (AE_TYPE);

                break;
            }
        }

        else
        {
            /* 
             * Cannot figure out the type -- set to DefAny which will print as an
             * error in the name table dump
             */

            if (GetDebugLevel () > 0)
            {
                DUMP_PATHNAME (Handle, "NsAttachObject confused: setting bogus type for  ", 
                                ACPI_INFO, _COMPONENT);

                if (TbSystemTablePointer (Object))
                {
                    DEBUG_PRINT (ACPI_INFO,
                                ("AML-stream code %02x\n", *(UINT8 *) Object));
                }
        
                else if (VALID_DESCRIPTOR_TYPE (Object, DESC_TYPE_NTE))
                {
                    DUMP_PATHNAME (Object, "name ", ACPI_INFO, _COMPONENT);
                }
        
                else
                {
                    DUMP_PATHNAME (Object, "object ", ACPI_INFO, _COMPONENT);
                    DUMP_STACK_ENTRY (Object);
                }
            }

            ObjType = INTERNAL_TYPE_DefAny;
        }
    }



    DEBUG_PRINT (TRACE_EXEC,("NsAttachObject: Installing obj %p into NTE %p\n",
                    ObjDesc, Handle));
    

    /* Must increment the new value's reference count (if it is an internal object) */

    CmAddReference (ObjDesc);

    /* Save the existing object (if any) for deletion later */

    PreviousObjDesc = ThisEntry->Object;

    /* Install the object and set the type, flags */

    ThisEntry->Object   = ObjDesc;
    ThisEntry->Type     = (UINT8) ObjType;
    ThisEntry->Flags    = Flags;


    /* 
     * Delete an existing attached object. 
     */

    if (PreviousObjDesc)
    {
        CmRemoveReference (PreviousObjDesc);    /* One for the attach to the NTE */
        CmRemoveReference (PreviousObjDesc);    /* Now delete */
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
NsAttachMethod (
    ACPI_HANDLE             Handle, 
    UINT8                   *PcodeAddr, 
    UINT32                  PcodeLength)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    ACPI_OBJECT_INTERNAL    *PreviousObjDesc;
    NAME_TABLE_ENTRY        *ThisEntry = (NAME_TABLE_ENTRY *) Handle;


    FUNCTION_TRACE ("NsAttachMethod");


    /* Parameter validation */

    if (!Gbl_RootObject->Scope)
    {
        /* Name space uninitialized */

        REPORT_ERROR ("NsAttachMethod: name space uninitialized");
        return_ACPI_STATUS (AE_NO_NAMESPACE);
    }
    
    if (!Handle)
    {
        /* Null name handle */

        REPORT_ERROR ("NsAttachMethod: null name handle");
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    
    /* Allocate a method descriptor */

    ObjDesc = CmCreateInternalObject (ACPI_TYPE_Method);
    if (!ObjDesc)
    {
        /* Method allocation failure  */

        REPORT_ERROR ("NsAttachMethod: allocation failure");
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Init the method info */

    ObjDesc->Method.Pcode       = PcodeAddr;
    ObjDesc->Method.PcodeLength = PcodeLength;

    /* Update reference count and install */

    CmAddReference (ObjDesc);

    PreviousObjDesc = ThisEntry->Object;
    ThisEntry->Object = ObjDesc;


    /* 
     * Delete an existing object.  Don't try to re-use in case it is shared
     */
    if (PreviousObjDesc)
    {
        DEBUG_PRINT (ACPI_INFO, ("NsAttachMethod: ***Old: %p Obj %p Pcode %p Len 0x%X\n",
                                    Handle, PreviousObjDesc, PreviousObjDesc->Method.Pcode, PreviousObjDesc->Method.PcodeLength));

        CmRemoveReference (PreviousObjDesc);
    }

    DEBUG_PRINT (ACPI_INFO, ("NsAttachMethod: %p Obj %p Pcode %p Len 0x%X\n",
                            Handle, ObjDesc, ObjDesc->Method.Pcode, ObjDesc->Method.PcodeLength));

    return_ACPI_STATUS (AE_OK);
}



void
NsDetachObject (
    ACPI_HANDLE             Object)
{
    NAME_TABLE_ENTRY        *Entry = Object;
    ACPI_OBJECT_INTERNAL    *ObjDesc;

    
    FUNCTION_TRACE ("NsDetachObject");

    ObjDesc = Entry->Object;
    if (!ObjDesc)
    {
        return_VOID;
    }

    /* Clear the entry in all cases */

    Entry->Object = NULL;
    
    /* Found a valid value */

    DEBUG_PRINT (ACPI_INFO, ("NsDetachObject: Object=%p Value=%p Name %4.4s\n", 
                            Entry, ObjDesc, &Entry->Name));

    /* Not every value is an object allocated via CmCallocate, must check */

    if (!TbSystemTablePointer (ObjDesc))
    {
        /* Attempt to delete the object (and all subobjects) */

        CmRemoveReference (ObjDesc);
    }

    return_VOID;
}



void *
NsGetAttachedObject (
    ACPI_HANDLE             Handle)
{
    FUNCTION_TRACE_PTR ("NsGetAttachedObject", Handle);


    if (!Handle)
    {
        /* handle invalid */

        REPORT_WARNING ("NsGetAttachedObject: Null handle");
        return_VALUE (NULL);
    }

    return_VALUE (((NAME_TABLE_ENTRY *) Handle)->Object);
}



ACPI_STATUS
NsCompareObject (
    ACPI_HANDLE             ObjHandle, 
    UINT32                  Level, 
    void                    *ObjDesc,
    void                    **ReturnValue)
{

    if (((NAME_TABLE_ENTRY *) ObjHandle)->Object == ObjDesc)
    {
        DEBUG_PRINT (ACPI_INFO, ("NsCompareObject: match found, Obj %x Val %x\n",
                        ObjHandle, ((NAME_TABLE_ENTRY *) ObjHandle)->Object));

        if (ReturnValue)
        {
            *ReturnValue = ObjHandle;
        }

        return AE_CTRL_TERMINATE;        /* Stop the walk */
    }

    return AE_OK;   /* Not found, continue the walk */
}



ACPI_HANDLE
NsFindAttachedObject (
    ACPI_OBJECT_INTERNAL    *ObjDesc, 
    ACPI_HANDLE             StartHandle, 
    INT32                   MaxDepth)
{
    ACPI_HANDLE             RetObject;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("NsFindAttachedObject");


    /* Parameter validation */

    if (!ObjDesc)
    {
        return_VALUE (NULL);
    }

    if (0 == MaxDepth)
    {
        return_VALUE (NULL);
    }

    if (!Gbl_RootObject->Scope)
    {
        /* 
         * If the name space has not been initialized,
         * there surely are no matching values.
         */
        return_VALUE (NULL);
    }

    if (NS_ALL == StartHandle)
    {
        StartHandle = Gbl_RootObject;
    }
    
    else
    {
        /* 
         * If base is not the root and has no children,
         * there is nothing to search.
         */
        return_VALUE (NULL);
    }


    /* 
     * Walk namespace until a match is found.
     * Either the matching object is returned, or NULL in case
     * of no match.
     */
    Status = NsWalkNamespace (ACPI_TYPE_Any, StartHandle, MaxDepth, NS_WALK_NO_UNLOCK, NsCompareObject, 
                                ObjDesc, &RetObject);
    if (ACPI_FAILURE (Status))
    {
        DEBUG_PRINT (ACPI_ERROR, ("NsFindAttachedObject: WalkNamespace failed! %x\n", Status));
        RetObject = NULL;
    }

    return_VALUE (RetObject);
}




