
/*
 *  ./acpica/subsystem/namespace/nssearch.c - Namespace search
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSSEARCH_C__

#include <acpi.h>
#include <amlcode.h>
#include <interp.h>
#include <namesp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nssearch");







ACPI_STATUS
NsSearchOneScope (
    UINT32                  EntryName, 
    NAME_TABLE_ENTRY        *NameTable, 
    ACPI_OBJECT_TYPE        Type, 
    NAME_TABLE_ENTRY        **RetEntry, 
    NS_SEARCH_DATA          *RetInfo)
{
    UINT32                  Position;
    UINT32                  Tries;


    FUNCTION_TRACE ("NsSearchOneScope");

    /* Debug only */

    {
        DEBUG_EXEC (char *ScopeName = NsNameOfScope (NameTable));
        DEBUG_PRINT (TRACE_NAMES, ("NsSearchOneScope: Searching %s [%p]\n",
                            ScopeName, NameTable));
        DEBUG_PRINT (TRACE_NAMES, ("NsSearchOneScope: For %4.4s (type 0x%X)\n",
                            &EntryName, Type));
        DEBUG_EXEC (CmFree (ScopeName));
    }



    /* 
     * Name tables are built (and subsequently dumped) in the
     * order in which the names are encountered during the namespace load;
     *
     * All namespace searching will be linear;  If a table overflows an 
     * additional segment will be allocated and added (chained).
     *
     * Start linear search at top of table
     */
    Position = 0;

    /* Init return data */

    if (RetInfo)
    {
        RetInfo->PreviousEntry = NULL;
        RetInfo->NameTable = NameTable;
    }

    /* 
     * Search for name in table, starting at Position.  Stop searching upon
     * finding an unused entry or after examining all entries in the table.
     *
     * Moving to the "next" entry is done at the bottom of the loop instead
     * of in the iteration expression because the method used depends on
     * whether or not USE_HASHING is in effect.
     */

    for (Tries = NS_TABLE_SIZE; Tries && 0 != NameTable[Position].Name; Tries--)
    {
        /* Search for name in table */

        if (NameTable[Position].Name == EntryName)
        {
            /* 
             * Found matching entry.  Capture type if appropriate before
             * returning the entry.
             */

            /* 
             * The DefFieldDefn and BankFieldDefn cases are actually
             * looking up the Region in which the field will be defined
             */

            if ((INTERNAL_TYPE_DefFieldDefn == Type) || 
                (INTERNAL_TYPE_BankFieldDefn == Type))
            {
                Type = ACPI_TYPE_Region;
            }

            /* 
             * Scope, DefAny, and IndexFieldDefn are bogus "types" which do
             * not actually have anything to do with the type of the name
             * being looked up.  For any other value of Type, if the type
             * stored in the entry is Any (i.e. unknown), save the actual type.
             */

            if (Type != INTERNAL_TYPE_Scope && 
                Type != INTERNAL_TYPE_DefAny &&
                Type != INTERNAL_TYPE_IndexFieldDefn && 
                NameTable[Position].Type == ACPI_TYPE_Any)
            {
                NameTable[Position].Type = (UINT8) Type;
            }

            DEBUG_PRINT (TRACE_NAMES, ("NsSearchOneScope: Name %4.4s (actual type 0x%X) found at %p\n", 
                            &EntryName, NameTable[Position].Type, &NameTable[Position]));
            
            *RetEntry = &NameTable[Position];
            return_ACPI_STATUS (AE_OK);
        }

        if (RetInfo)
        {
            /* Save a pointer to this entry, it might become the previous entry */

            RetInfo->PreviousEntry = &NameTable[Position];
        }
        

        /* Done with this table? */

        if ((1 == Tries) && 
            (NEXTSEG (NameTable)))
        {
            /* 
             * Just examined last slot, but table has an appendage.
             * All appendages, even to the root NT, contain NS_DEFAULT_TABLE_SIZE entries.
             */

            NameTable = NEXTSEG (NameTable);
            DEBUG_PRINT (TRACE_EXEC, ("NsSearchOneScope: Search appendage NameTable=%p\n", NameTable));
            Position = 0;
            Tries += NS_TABLE_SIZE;
        }
        
        else
        {
            ++Position;
        }
    }


    /* Searched entire table, not found */

    DEBUG_PRINT (TRACE_NAMES, ("NsSearchOneScope: Name %4.4s (type 0x%X) not found at %p\n", 
                                &EntryName, Type, &NameTable[Position]));


    if (RetInfo)
    {
        /* Save the final information (name was not found) */

        RetInfo->TableFull  = (BOOLEAN) (!(Tries));      /* Table is full if no more tries available */
        RetInfo->Position   = Position;
        RetInfo->NameTable  = NameTable;
    }

    return_ACPI_STATUS (AE_NOT_FOUND);
}




ACPI_STATUS
NsSearchParentTree (
    UINT32                  EntryName, 
    NAME_TABLE_ENTRY        *NameTable, 
    ACPI_OBJECT_TYPE        Type, 
    NAME_TABLE_ENTRY        **RetEntry)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ParentEntry;


    FUNCTION_TRACE ("NsSearchParentTree");


    /* 
     * NameTable[0] will be an unused entry if the table being searched is empty,
     * However, its ParentEntry member will have been filled in
     * when the table was allocated (unless it is the root name table).
     */

   
    if (!NsLocal (Type) && 
        NameTable[0].ParentEntry)
    {
        ParentEntry = NameTable[0].ParentEntry;
        DEBUG_PRINT (TRACE_NAMES, ("NsSearchParentTree: Searching parent for %4.4s\n", 
                                    &EntryName));

        /* Search parents until found or we have backed up to the root */

        while (ParentEntry)
        {
            /* Search parent scope */
            /* TBD: Why ACPI_TYPE_Any? */

            Status = NsSearchOneScope (EntryName, ParentEntry->Scope, ACPI_TYPE_Any, RetEntry, NULL);
            if (Status == AE_OK)
            {
                return_ACPI_STATUS (Status);
            }

            /* Not found here, go up another level (until we reach the root) */

            ParentEntry = ParentEntry->ParentEntry;
        }

        /* Not found in parent tree */
    }
 
    else
    {
        /*
         * No parent, or type is "local".  We won't be searching the parent tree.
         */

        if (!NameTable[0].ParentEntry)
        {
            DEBUG_PRINT (TRACE_NAMES, ("NsSearchParentTree: [%4.4s] has no parent\n", 
                                        &EntryName));
        }

        else if (NsLocal (Type))
        {
            DEBUG_PRINT (TRACE_NAMES, ("NsSearchParentTree: [%4.4s] (type 0x%X) is local (no search)\n", 
                                        &EntryName, Type));
        }
    }

    return_ACPI_STATUS (AE_NOT_FOUND);
}



ACPI_STATUS
NsCreateAndLinkNewTable (
    NAME_TABLE_ENTRY        *NameTable)
{
    NAME_TABLE_ENTRY        *NewTable;
    NAME_TABLE_ENTRY        *ParentEntry;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("NsCreateAndLinkNewTable");


    /* Sanity check on the data structure */

    if (NEXTSEG (NameTable))
    {
        /* We should never get here (an appendage already allocated) */

        DEBUG_PRINT (ACPI_ERROR,
                    ("NsCreateAndLinkNewTable: appendage %p about to be overwritten\n",
                    NEXTSEG (NameTable)));
    }


    /*
     * We can use the parent entries from the start of the current table 
     * Since the parent information remains the same.
     */

    ParentEntry = NameTable[0].ParentEntry;


    /* Allocate and chain an appendage to the filled table */
    
    NewTable = NsAllocateNameTable (NS_TABLE_SIZE);
    if (!NewTable)
    {
        REPORT_ERROR ("Name Table appendage allocation failure");
        Status = AE_NO_MEMORY;
    }

    else
    {
        /* 
         * Allocation successful. Init the new table.
         */
        NEXTSEG (NameTable) = NewTable;
        NsInitializeTable (NewTable, ParentEntry->Scope, ParentEntry);

        DEBUG_PRINT (TRACE_EXEC, 
            ("NsCreateAndLinkNewTable: NewTable=%p, ParentEntry=%p, Scope=%p\n",
                NewTable, ParentEntry, NameTable->Scope));
    }

    return_ACPI_STATUS (Status);
}



void
NsInitializeTable (
    NAME_TABLE_ENTRY        *NewTable, 
    NAME_TABLE_ENTRY        *ParentScope, 
    NAME_TABLE_ENTRY        *ParentEntry)
{


    NewTable->ParentEntry     = ParentEntry;


    DEBUG_PRINT (TRACE_NAMES, ("NsInitializeTable: %x\n", NewTable));
}



void
NsInitializeEntry (
    ACPI_WALK_STATE   *WalkState,
    NAME_TABLE_ENTRY        *NameTable, 
    UINT32                  Position, 
    UINT32                  EntryName, 
    ACPI_OBJECT_TYPE        Type, 
    NAME_TABLE_ENTRY        *PreviousEntry)
{
    NAME_TABLE_ENTRY        *NewEntry;
    UINT16                  OwnerId = TABLE_ID_DSDT;


    FUNCTION_TRACE ("NsInitializeEntry");


    /* 
     * Get the owner ID from the Walk state 
     * The owner ID is used to track table deletion and deletion of objects created by methods 
     */
    if (WalkState)
    {
        OwnerId = WalkState->OwnerId;
    }

    /* The new entry is given by two parameters */

    NewEntry = &NameTable[Position];

    /* Init the new entry */

    NewEntry->DataType       = DESC_TYPE_NTE;
    NewEntry->Name           = EntryName;
    NewEntry->ParentEntry    = NameTable[0].ParentEntry;
    NewEntry->OwnerId        = OwnerId;
    NewEntry->ReferenceCount = 1;

    /* 
     * Set forward and back links.
     * Important:  These are the links that tie the tables together
     * so that when walking the links, it is invisible that their
     * are separate, disjoint tables.
     */
    if (PreviousEntry)
    {
        PreviousEntry->NextEntry = NewEntry;
    }

    NewEntry->PrevEntry = PreviousEntry;
    NewEntry->NextEntry = NULL;

    /* 
     * If adding a name with unknown type, or having to add the region in
     * order to define fields in it, we have a forward reference.
     */

    if ((ACPI_TYPE_Any == Type) || 
        (INTERNAL_TYPE_DefFieldDefn == Type) || 
        (INTERNAL_TYPE_BankFieldDefn == Type))
    {
        /* 
         * We don't want to abort here, however!
         * We will fill in the actual type when the real definition
         * is found later.
         */

        DEBUG_PRINT (ACPI_INFO, ("[%4.4s] is a forward reference into the namespace\n",
                        &EntryName));

    }

    /* 
     * The DefFieldDefn and BankFieldDefn cases are actually
     * looking up the Region in which the field will be defined
     */

    if ((INTERNAL_TYPE_DefFieldDefn == Type) || 
        (INTERNAL_TYPE_BankFieldDefn == Type))
    {
        Type = ACPI_TYPE_Region;
    }

    /* 
     * Scope, DefAny, and IndexFieldDefn are bogus "types" which do
     * not actually have anything to do with the type of the name
     * being looked up.  Save any other value of Type as the type of
     * the entry.
     */

    if ((Type != INTERNAL_TYPE_Scope) && 
        (Type != INTERNAL_TYPE_DefAny) && 
        (Type != INTERNAL_TYPE_IndexFieldDefn))
    {
        NewEntry->Type = (UINT8) Type;
    }

    DEBUG_PRINT (TRACE_NAMES, ("NsInitializeEntry: %4.4s added to %p at %p\n", 
                                &EntryName, NameTable, NewEntry));
    
    return_VOID;
}



ACPI_STATUS
NsSearchAndEnter (
    UINT32                  EntryName, 
    ACPI_WALK_STATE   *WalkState,
    NAME_TABLE_ENTRY        *NameTable,
    OPERATING_MODE          InterpreterMode, 
    ACPI_OBJECT_TYPE        Type, 
    UINT32                  Flags,
    NAME_TABLE_ENTRY        **RetEntry)
{
    UINT32                  Position;       /* position in table */
    ACPI_STATUS             Status;
    NS_SEARCH_DATA          SearchInfo;
    NAME_TABLE_ENTRY        *Entry;


    FUNCTION_TRACE ("NsSearchAndEnter");


    /* Parameter validation */

    if (!NameTable || !EntryName || !RetEntry)
    {
        REPORT_ERROR ("NsSearchAndEnter: bad parameter");
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* Name must consist of printable characters */

    if (!CmValidAcpiName (EntryName))
    {
        DEBUG_PRINT (ACPI_ERROR, ("NsSearchAndEnter:  *** Bad char in name: %08lx *** \n", 
                                    EntryName));

        return_ACPI_STATUS (AE_BAD_CHARACTER);
    }


    /* Try to find the name in the table specified by the caller */

    *RetEntry = ENTRY_NOT_FOUND;
    Status = NsSearchOneScope (EntryName, NameTable, Type, RetEntry, &SearchInfo);
    if (Status != AE_NOT_FOUND)
    {
        /* Either found it or there was an error -- finished either way */

        return_ACPI_STATUS (Status);
    }


    /*
     * Not found in the table.  If we are NOT performing the first pass (name entry) of
     * loading the namespace, search the parent tree (all the way to the root if necessary.)  
     * We don't want to perform the parent search when the namespace is actually being loaded.
     * We want to perform the search when namespace references are being resolved (load pass 2)
     * and during the execution phase.
     */

    if ((InterpreterMode != IMODE_LoadPass1) &&
        (Flags & NS_SEARCH_PARENT))
    {
        /* Not found in table - search parent tree according to ACPI specification */

        Status = NsSearchParentTree (EntryName, NameTable, Type, RetEntry);
        if (Status == AE_OK)
        {
            return_ACPI_STATUS (Status);
        }
    }


    /*
     * In execute mode, just search, never add names.  Exit now.
     */

    if (InterpreterMode == IMODE_Execute)
    {
        DEBUG_PRINT (TRACE_NAMES, ("NsSearchAndEnter: %4.4s Not found in %p [Not adding]\n", 
                                    &EntryName, NameTable));
    
        return_ACPI_STATUS (AE_NOT_FOUND);
    }


    /* 
     * Extract the pertinent info from the search result struct.
     * NameTable and position might now point to an appendage
     */
    NameTable = SearchInfo.NameTable;
    Position = SearchInfo.Position;


    /* 
     * This block handles the case where the existing table is full.
     * we must allocate a new table before we can initialize a new entry
     */

    if (SearchInfo.TableFull)
    {
        Status = NsCreateAndLinkNewTable (NameTable);
        if (Status != AE_OK)
        {
            return_ACPI_STATUS (Status);
        }

        /* Point to the first slot in the new table */

        NameTable = NEXTSEG (NameTable);
        Position = 0;
    }


    /*
     * There is room in the table (or we have just allocated a new one.)
     * Initialize the new entry
     */

    NsInitializeEntry (WalkState, NameTable, Position, EntryName, Type, 
                        SearchInfo.PreviousEntry);


    *RetEntry   = &NameTable[Position];
    Entry       = &NameTable[Position];

    /* Increment the reference count(s) of all parents up to the root! */

    while (Entry->ParentEntry)
    {
        Entry = Entry->ParentEntry;
        Entry->ReferenceCount++;
    }


    return_ACPI_STATUS (AE_OK);
}

