
/*
 *  ./acpica/subsystem/namespace/nsutils.c - Utilities for accessing ACPI namespace
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __NSUTILS_C__

#include <acpi.h>
#include <namesp.h>
#include <interp.h>
#include <amlcode.h>
#include <tables.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsutils");




BOOLEAN
NsValidRootPrefix (
    char                    Prefix)
{

    return ((BOOLEAN) (Prefix == '\\'));
}



BOOLEAN
NsValidPathSeparator (
    char                    Sep)
{

    return ((BOOLEAN) (Sep == '.'));
}



ACPI_OBJECT_TYPE
NsGetType (
    ACPI_HANDLE             handle)
{
    FUNCTION_TRACE ("NsGetType");


    if (!handle)
    {
        /*  Handle invalid  */

        REPORT_WARNING ("NsGetType: Null handle");
        return_VALUE (ACPI_TYPE_Any);
    }

    return_VALUE (((NAME_TABLE_ENTRY *) handle)->Type);
}



INT32
NsLocal (
    ACPI_OBJECT_TYPE        Type)
{
    FUNCTION_TRACE ("NsLocal");


    if (!CmValidObjectType (Type))
    {
        /*  type code out of range  */

        REPORT_WARNING ("NsLocal: Invalid Object Type");
        return_VALUE (NSP_NORMAL);
    }

    return_VALUE ((INT32) Gbl_NsProperties[Type] & NSP_LOCAL);
}





ACPI_STATUS
NsInternalizeName (
    char                    *ExternalName,
    char                    **ConvertedName)
{
    char                    *Result = NULL;
    char                    *InternalName;
    ACPI_SIZE               NumSegments;
    BOOLEAN                 FullyQualified = FALSE;
    UINT32                  i;


    FUNCTION_TRACE ("NsInternalizeName");


    if ((!ExternalName)      || 
        (*ExternalName == 0) ||
        (!ConvertedName))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /* 
     * For the internal name, the required length is 4 bytes per segment, 
     * plus 1 each for RootPrefix, MultiNamePrefixOp, segment count, trailing null 
     * (which is not really needed, but no there's harm in putting it there)
     *
     * strlen() + 1 covers the first NameSeg, which has no path separator
     */

    if (NsValidRootPrefix (ExternalName[0]))
    {
        FullyQualified = TRUE;
        ExternalName++;
    }


    /* 
     * Determine the number of ACPI name "segments" by counting the number
     * of path separators within the string.  Start with one segment since
     * the segment count is (# separators) + 1, and zero separators is ok.
     */

    NumSegments = 1;    
    for (i = 0; ExternalName[i]; i++)
    {
        if (NsValidPathSeparator (ExternalName[i]))
        {
            NumSegments++;
        }
    }



    /* We need a segment to store the internal version of the name */

    InternalName = CmCallocate ((ACPI_NAME_SIZE * NumSegments) + 4);
    if (!InternalName)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }


    /* Setup the correct prefixes, counts, and pointers */

    if (FullyQualified)
    {
        InternalName[0] = '\\';
        InternalName[1] = AML_MultiNamePrefixOp;
        InternalName[2] = (char) NumSegments;
        Result = &InternalName[3];
    }
    else
    {
        InternalName[0] = AML_MultiNamePrefixOp;
        InternalName[1] = (char) NumSegments;
        Result = &InternalName[2];
    }


    /* Build the name (minus path separators) */

    for (; NumSegments; NumSegments--)
    {
        for (i = 0; i < ACPI_NAME_SIZE; i++) //STRNCPY (Result, ExternalName, ACPI_NAME_SIZE);
        {
            if (NsValidPathSeparator (*ExternalName) ||
               (*ExternalName == 0))
            {
                /* Pad the segment with underscore(s) if segment is short */

                Result[i] = '_';
            }

            else
            {
                /* Convert char to uppercase and save it */

                Result[i] = (char) TOUPPER (*ExternalName);
                ExternalName++;
            }

        }

        /* Now we must have a path separator, or the pathname is bad */

        if (!NsValidPathSeparator (*ExternalName) &&
            (*ExternalName != 0))
        {
            CmFree (InternalName);
            return_ACPI_STATUS (AE_BAD_PARAMETER);
        }

        /* Move on the next segment */

        ExternalName++;
        Result += ACPI_NAME_SIZE;
    }

    
    /* Return the completed name */

    *Result = 0;                     /* Terminate the string! */
    *ConvertedName = InternalName;


    if (FullyQualified)
    {
        DEBUG_PRINT (TRACE_EXEC,("NsInternalizeName: returning [%p] (abs) \"\\%s\"\n", 
                                InternalName, &InternalName[3])); 
    }
    else
    {
        DEBUG_PRINT (TRACE_EXEC,("NsInternalizeName: returning [%p] (rel) \"%s\"\n", 
                                InternalName, &InternalName[2])); 
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
NsExternalizeName (
    UINT32                  InternalNameLength,
    char                    *InternalName,
    UINT32                  *ConvertedNameLength,
    char                    **ConvertedName)
{
    UINT32                  PrefixLength = 0;
    UINT32                  NamesIndex = 0;
    UINT32                  NamesCount = 0;
    UINT32                  i = 0;
    UINT32                  j = 0;

    FUNCTION_TRACE ("NsExternalizeName");

    if (InternalNameLength < 0 || !InternalName || !ConvertedNameLength || !ConvertedName)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Check for a prefix (one '\' | one or more '^').
     */
    switch (InternalName[0])
    {
    case '\\':
        PrefixLength = 1;
        break;

    case '^':
        for (i = 0; i < InternalNameLength; i++)
        {
            if (InternalName[i] != '^')
            {
                PrefixLength = i + 1;
            }
        }

        if (i == InternalNameLength)
        {
            PrefixLength = i;
        }

        break;
    }

    /*
     * Check for object names.  Note that there could be 0-255 of these 
     * 4-byte elements.
     */
    if (PrefixLength < InternalNameLength)
    {
        switch (InternalName[PrefixLength])
        {

        /* <count> 4-byte names */

        case AML_MultiNamePrefixOp:
            NamesIndex = PrefixLength + 2;
            NamesCount = (UINT32) InternalName[PrefixLength + 1];
            break;


        /* two 4-byte names */

        case AML_DualNamePrefix:
            NamesIndex = PrefixLength + 1;
            NamesCount = 2;
            break;


        /* NullName */

        case 0:
            NamesIndex = 0;
            NamesCount = 0;
            break;


        /* one 4-byte name */

        default:
            NamesIndex = PrefixLength;
            NamesCount = 1;
            break;
        }
    }

    /* 
     * Calculate the length of ConvertedName, which equals the length
     * of the prefix, length of all object names, length of any required 
     * punctuation ('.') between object names, plus the NULL terminator.
     */
    *ConvertedNameLength = PrefixLength + (4 * NamesCount) + ((NamesCount > 0) ? (NamesCount - 1) : 0) + 1;

    /*
     * Check to see if we're still in bounds.  If not, there's a problem
     * with InternalName (invalid format).
     */ 
    if (*ConvertedNameLength > InternalNameLength)
    {
        REPORT_ERROR ("NsExternalizeName: Invalid internal name.\n");
        return_ACPI_STATUS (AE_BAD_PATHNAME);
    }

    /* 
     * Build ConvertedName...
     */

    (*ConvertedName) = CmCallocate (*ConvertedNameLength);
    if (!(*ConvertedName))
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    j = 0;

    for (i = 0; i < PrefixLength; i++)
    {
        (*ConvertedName)[j++] = InternalName[i];
    }

    if (NamesCount > 0)
    {
        for (i = 0; i < NamesCount; i++)
        {
            if (i > 0)
            {
                (*ConvertedName)[j++] = '.';
            }

            (*ConvertedName)[j++] = InternalName[NamesIndex++];
            (*ConvertedName)[j++] = InternalName[NamesIndex++];
            (*ConvertedName)[j++] = InternalName[NamesIndex++];
            (*ConvertedName)[j++] = InternalName[NamesIndex++];
        }
    }

    return_ACPI_STATUS (AE_OK);
}



NAME_TABLE_ENTRY *
NsConvertHandleToEntry (
    ACPI_HANDLE             Handle)
{

    /* 
     * Simple implementation for now;
     * TBD: Real integer handles allow for more verification 
     * and keep all pointers within this subsystem!
     */

    if (!Handle)
    {
        return NULL;
    }

    if (Handle == ACPI_ROOT_OBJECT)
    {
        return Gbl_RootObject;
    }


    /* We can at least attempt to verify the handle */

    if (!VALID_DESCRIPTOR_TYPE (Handle, DESC_TYPE_NTE))
    {
        return NULL;
    }

    return (NAME_TABLE_ENTRY *) Handle;
}



ACPI_HANDLE
NsConvertEntryToHandle(NAME_TABLE_ENTRY *Nte)
{


    /* 
     * Simple implementation for now;
     * TBD: Real integer handles allow for more verification 
     * and keep all pointers within this subsystem!
     */

    return (ACPI_HANDLE) Nte;


/* ---------------------------------------------------
 
    if (!Nte)
    {
        return NULL;
    }

    if (Nte == Gbl_RootObject)
    {
        return ACPI_ROOT_OBJECT;
    }


    return (ACPI_HANDLE) Nte;
------------------------------------------------------*/
}




void
NsTerminate (void)
{
    ACPI_OBJECT_INTERNAL    *ObjDesc;
    NAME_TABLE_ENTRY        *Entry;


    FUNCTION_TRACE ("NsTerminate");


    Entry = Gbl_RootObject;

    /*
     * 1) Free the entire namespace -- all objects, tables, and stacks
     */
    /* Delete all objects linked to the root (additional table descriptors) */

    NsDeleteNamespaceSubtree (Entry);

    /* Detach any object(s) attached to the root */

    ObjDesc = NsGetAttachedObject (Entry);
    if (ObjDesc)
    {
        NsDetachObject (Entry);
        CmRemoveReference (ObjDesc);
    }

    NsDeleteScope (Entry->Scope);
    Entry->Scope = NULL;


    REPORT_SUCCESS ("Entire namespace and objects deleted");
    DEBUG_PRINT (ACPI_INFO, ("NsTerminate: Namespace freed\n"));


    /* 
     * 2) Now we can delete the ACPI tables 
     */

    TbDeleteAcpiTables ();

    DEBUG_PRINT (ACPI_INFO, ("NsTerminate: ACPI Tables freed\n"));

 return_VOID;
}

 

