
/*
 *  ./acpica/subsystem/namespace/nswalk.c - Functions for walking the APCI namespace
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __NSWALK_C__

#include <acpi.h>
#include <interp.h>
#include <namesp.h>
#include <methods.h>
#include <acobject.h>
#include <pnp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nswalk");



NAME_TABLE_ENTRY *
NsGetNextObject (
    ACPI_OBJECT_TYPE        Type, 
    NAME_TABLE_ENTRY        *Parent, 
    NAME_TABLE_ENTRY        *Child)
{
    NAME_TABLE_ENTRY        *ThisEntry;



    if (!Child)
    {

        /* It's really the parent's _scope_ that we want */

        ThisEntry = Parent->Scope;
    }

    else
    {
        /* Start search at the NEXT object */

        ThisEntry = Child->NextEntry;
    }


    /* If any type is OK, we are done */

    if (Type == ACPI_TYPE_Any)
    {
        /* Make sure this is valid entry first */

        if ((!ThisEntry) ||
            (!ThisEntry->Name))
        {
            return NULL;
        }

        return (ThisEntry);
    }


    /* Must search for the object -- but within this scope only */

    while (ThisEntry)
    {
        /* If type matches, we are done */

        if (ThisEntry->Type == Type)
        {
            return (ThisEntry);
        }

        /* Otherwise, move on to the next object */

        ThisEntry = ThisEntry->NextEntry;
    }


    /* Not found */

    return NULL;
}




ACPI_STATUS
NsWalkNamespace (
    ACPI_OBJECT_TYPE        Type, 
    ACPI_HANDLE             StartObject, 
    UINT32                  MaxDepth,
    BOOLEAN                 UnlockBeforeCallback,
    WALK_CALLBACK           UserFunction, 
    void                    *Context, 
    void                    **ReturnValue)
{
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *ChildHandle;
    NAME_TABLE_ENTRY        *ParentHandle;
    ACPI_OBJECT_TYPE        ChildType;
    UINT32                  Level;


    FUNCTION_TRACE ("NsWalkNamespace");

    /* Special case for the namespace root object */

    if (StartObject == ACPI_ROOT_OBJECT)
    {
        StartObject = Gbl_RootObject;
    }


    /* Null child means "get first object" */

    ParentHandle    = StartObject;
    ChildHandle     = 0;
    ChildType       = ACPI_TYPE_Any;
    Level           = 1;

    /* 
     * Traverse the tree of objects until we bubble back up to where we
     * started. When Level is zero, the loop is done because we have 
     * bubbled up to (and passed) the original parent handle (StartHandle)
     */

    while (Level > 0)
    {
        /* Get the next typed object in this scope.  Null returned if not found */

        Status = AE_OK;
        ChildHandle = NsGetNextObject (ACPI_TYPE_Any, ParentHandle, ChildHandle);
        if (ChildHandle)
        {
            /* Found an object, Get the type if we are not searching for ANY */

            if (Type != ACPI_TYPE_Any)
            {
                ChildType = ChildHandle->Type;
            }

            if (ChildType == Type)
            {
                /* Found a matching object, invoke the user callback function */


                if (UnlockBeforeCallback)
                {
                    CmReleaseMutex (MTX_NAMESPACE);
                }

                Status = UserFunction (ChildHandle, Level, Context, ReturnValue);

                if (UnlockBeforeCallback)
                {
                    CmAcquireMutex (MTX_NAMESPACE);
                }

                switch (Status)
                {
                case AE_OK:
                case AE_CTRL_DEPTH:
                    break;                          /* Just keep going */

                case AE_CTRL_TERMINATE:
                    return_ACPI_STATUS (AE_OK);     /* Exit now, with OK status */
                    break;

                default:
                    return_ACPI_STATUS (Status);    /* All others are valid exceptions */
                    break;
                }
            }

            /* 
             * Depth first search:
             * Attempt to go down another level in the namespace if we are allowed to.
             * Don't go any further if we have reached the caller specified maximum depth
             * or if the user function has specified that the maximum depth has been reached.
             */

            if ((Level < MaxDepth) && (Status != AE_CTRL_DEPTH))
            {
                if (NsGetNextObject (ACPI_TYPE_Any, ChildHandle, 0))
                {
                    /* There is at least one child of this object, visit the object */

                    Level++;
                    ParentHandle    = ChildHandle;
                    ChildHandle     = 0;
                }
            }
        }

        else
        {
            /* 
             * No more children in this object (NsGetNextObject failed), 
             * go back upwards in the namespace tree to the object's parent.
             */
            Level--;
            ChildHandle = ParentHandle;
            ParentHandle = ParentHandle->ParentEntry;
        }
    }

    return_ACPI_STATUS (AE_OK);                   /* Complete walk, not terminated by user function */
}


