/*
 *  ./acpica/subsystem/parser/psargs.c - Parse AML opcode arguments
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __PSARGS_C__

#include <acpi.h>
#include <parser.h>
#include <amlcode.h>
#include <namesp.h>

#define _COMPONENT          PARSER
        MODULE_NAME         ("psargs");



UINT32
PsGetNextPackageLength (
    ACPI_PARSE_STATE        *ParserState)
{
    INT32                   EncodedLength;
    INT32                   Length = 0;


    FUNCTION_TRACE ("PsGetNextPackageLength");


    EncodedLength = (INT32) GET8 (ParserState->Aml);
    ParserState->Aml++;


    switch (EncodedLength >> 6) /* bits 6-7 contain encoding scheme */
    {
    case 0: /* 1-byte encoding (bits 0-5) */

        Length = (EncodedLength & 0x3f);
        break;


    case 1: /* 2-byte encoding (next byte + bits 0-3) */

        Length = (GET8 (ParserState->Aml) << 4) | (EncodedLength & 0xf);
        ParserState->Aml++;
        break;


    case 2: /* 3-byte encoding (next 2 bytes + bits 0-3) */

        Length = ( (GET8 (ParserState->Aml + 1) << 12)
               | (GET8 (ParserState->Aml) << 4)
               | (EncodedLength & 0xf));
        ParserState->Aml += 2;
        break;


    case 3: /* 4-byte encoding (next 3 bytes + bits 0-3) */

        Length = ( (GET8 (ParserState->Aml + 2) << 20)
               | (GET8 (ParserState->Aml + 1) << 12)
               | (GET8 (ParserState->Aml) << 4)
               | (EncodedLength & 0xf));
        ParserState->Aml += 3;
        break;
    }

    return_VALUE (Length);
}



UINT8 *
PsGetNextPackageEnd (
    ACPI_PARSE_STATE        *ParserState)
{
    UINT8                   *Start = ParserState->Aml;
    NATIVE_UINT             Length;


    FUNCTION_TRACE ("PsGetNextPackageEnd");


    Length = (NATIVE_UINT) PsGetNextPackageLength (ParserState);

    return_PTR (Start + Length); /* end of package */
}



char *
PsGetNextNamestring (
    ACPI_PARSE_STATE        *ParserState)
{
    char                    *Start = (char *) ParserState->Aml;
    char                    *End = (char *) ParserState->Aml;
    INT32                   Length;


    FUNCTION_TRACE ("PsGetNextNamestring");


    /* Handle multiple prefix characters */

    while (PsIsPrefixChar (GET8 (End)))
    {
        /* include prefix '\\' or '^' */

        End++;
    }

    /* Decode the path */

    switch (GET8 (End))
    {
    case 0:
        
        /* NullName */

        if (End == Start)
        {
            Start = NULL;
        }
        End++;
        break;


    case AML_DualNamePrefix:

        /* two name segments */

        End += 9;
        break;


    case AML_MultiNamePrefixOp:

        /* multiple name segments */

        Length = (INT32) GET8 (End + 1) * 4;
        End += 2 + Length;
        break;


    default:

        /* single name segment */
        /* assert (PsIsLead (GET8 (End))); */

        End += 4;
        break;
    }

    ParserState->Aml = (UINT8*) End;

    return_PTR (Start);
}



void
PsGetNextNamepath (
    ACPI_PARSE_STATE        *ParserState,
    ACPI_GENERIC_OP         *Arg,
    UINT32                  *ArgCount,
    BOOLEAN                 MethodCall)
{
    char                    *Path;
    ACPI_GENERIC_OP         *Name;



    FUNCTION_TRACE ("PsGetNextNamepath");


    Path = PsGetNextNamestring (ParserState);
    if (!Path || !MethodCall)
    {
        /* Null name case, create a null namepath object */

        PsInitOp (Arg, AML_NAMEPATH_OP);        
        Arg->Value.Name = Path;
        return_VOID;
    }


#ifdef PARSER_ONLY
    if (Gbl_ParsedNamespaceRoot)
    {
        ACPI_GENERIC_OP         *MethodOp;
        ACPI_GENERIC_OP         *Count;

        /*
         * Lookup the name in the parsed namespace 
         */

        MethodOp = MethodCall ? PsFind (PsGetParentScope (ParserState), Path, AML_MethodOp, 0):NULL;
        if (MethodOp)
        {
            DEBUG_PRINT (TRACE_PARSE, ("PsGetNextNamepath: Found a method while parsed namespace still valid!\n"));

            Count = PsGetArg (MethodOp, 0);
            if (Count && Count->Opcode == AML_ByteOp)
            {
                Name = PsAllocOp (AML_NAMEPATH_OP);
                if (Name)
                {
                    /* Change arg into a METHOD CALL and attach the name to it */

                    PsInitOp (Arg, AML_METHODCALL_OP);

                    Name->Value.Name        = Path;
                    Name->NameTableEntry    = MethodOp;          /* Point METHODCALL/NAME to the METHOD NTE */
                    PsAppendArg (Arg, Name);
                    
                    *ArgCount = Count->Value.Integer & METHOD_FLAGS_ARG_COUNT;
                }
            }

            return_VOID;
        }
    }

#else
    {
        /*
         * The full parse tree has already been deleted -- therefore, we are parsing
         * a control method.  We can lookup the name in the namespace instead of
         * the parse tree!
         */

        ACPI_STATUS             Status;
        NAME_TABLE_ENTRY        *Method = NULL;
        NAME_TABLE_ENTRY        *Nte;
        ACPI_GENERIC_STATE      ScopeInfo;


        /* 
         * Lookup the name in the internal namespace
         */
        ScopeInfo.Scope.Entry = NULL;
        Nte = ParserState->StartOp->NameTableEntry;
        if (Nte)
        {
            ScopeInfo.Scope.Entry = Nte->Scope;
        }

        /* 
         * Lookup object.  We don't want to add anything new to the namespace here, however.
         * So we use MODE_EXECUTE.
         */

        Status = NsLookup (&ScopeInfo, Path, ACPI_TYPE_Any, IMODE_Execute, /* MUST BE mode EXECUTE to perform upsearch */ 
                                NS_SEARCH_PARENT | NS_DONT_OPEN_SCOPE, NULL, &Nte);
        if (ACPI_SUCCESS (Status))
        {
            if (Nte->Type == ACPI_TYPE_Method)
            {
                Method = Nte;
                DEBUG_PRINT (TRACE_PARSE, ("PsGetNextNamepath: method - %p Path=%p\n", Method, Path));

                Name = PsAllocOp (AML_NAMEPATH_OP);
                if (Name)
                {
                    /* Change arg into a METHOD CALL and attach the name to it */

                    PsInitOp (Arg, AML_METHODCALL_OP);

                    Name->Value.Name        = Path;
                    Name->NameTableEntry    = Method;           /* Point METHODCALL/NAME to the METHOD NTE */
                    PsAppendArg (Arg, Name);

                    *ArgCount = ((ACPI_OBJECT_INTERNAL *) Method->Object)->Method.ParamCount;
                }
            
                return_VOID;
            }
        }
    }
#endif

    /* Everything else has failed */
    /* variable dereference */

    PsInitOp (Arg, AML_NAMEPATH_OP);        
    Arg->Value.Name = Path;


    return_VOID;
}



void
PsGetNextSimpleArg (
    ACPI_PARSE_STATE        *ParserState,
    INT32                   ArgType,
    ACPI_GENERIC_OP         *Arg)
{


    FUNCTION_TRACE_U32 ("PsGetNextSimpleArg", ArgType);


    switch (ArgType)
    {

    case ARGP_BYTEDATA:

        PsInitOp (Arg, AML_ByteOp);
        Arg->Value.Integer = (UINT32) GET8 (ParserState->Aml);
        ParserState->Aml++;
        break;


    case ARGP_WORDDATA:

        PsInitOp (Arg, AML_WordOp);
        Arg->Value.Integer = (UINT32) GET16 (ParserState->Aml);
        ParserState->Aml += 2;
        break;


    case ARGP_DWORDDATA:

        PsInitOp (Arg, AML_DWordOp);
        Arg->Value.Integer = (UINT32) GET32 (ParserState->Aml);
        ParserState->Aml += 4;
        break;


    case ARGP_CHARLIST:

        PsInitOp (Arg, AML_StringOp);
        Arg->Value.String = (char*) ParserState->Aml;

        while (GET8 (ParserState->Aml) != '\0')
        {
            ParserState->Aml++;
        }
        ParserState->Aml++;
        break;


    case ARGP_NAME:
    case ARGP_NAMESTRING:

        PsInitOp (Arg, AML_NAMEPATH_OP);
        Arg->Value.Name = PsGetNextNamestring (ParserState);
        break;
    }

    return_VOID;
}



ACPI_GENERIC_OP *
PsGetNextField (
    ACPI_PARSE_STATE        *ParserState)
{
    ACPI_PTRDIFF            AmlOffset = ParserState->Aml - ParserState->AmlStart;
    ACPI_GENERIC_OP         *Field;
    UINT16                  Opcode;


    FUNCTION_TRACE ("PsGetNextField");


    /* determine field type */

    switch (GET8 (ParserState->Aml))
    {

    default:

        Opcode = AML_NAMEDFIELD_OP;
        break;


    case 0x00:

        Opcode = AML_RESERVEDFIELD_OP;
        ParserState->Aml++;
        break;


    case 0x01:

        Opcode = AML_ACCESSFIELD_OP;
        ParserState->Aml++;
        break;
    }


    /* Allocate a new field op */

    Field = PsAllocOp (Opcode);
    if (Field)
    {
        Field->AmlOffset = AmlOffset;

        /* Decode the field type */

        switch (Opcode)
        {
        case AML_NAMEDFIELD_OP:

            /* Get the name */

            PsSetName (Field, GET32 (ParserState->Aml));
            ParserState->Aml += 4;

            /* Get the length which is encoded as a package length */

            Field->Value.Size = PsGetNextPackageLength (ParserState);
            break;


        case AML_RESERVEDFIELD_OP:

            /* Get the length which is encoded as a package length */

            Field->Value.Size = PsGetNextPackageLength (ParserState);
            break;


        case AML_ACCESSFIELD_OP:

            /* Get the AccessType and AccessAtrib and merge them into the field Op */

            Field->Value.Integer = ((GET8 (ParserState->Aml) << 8) | 
                                     GET8 (ParserState->Aml));
            ParserState->Aml += 2;
            break;
        }
    }

    return_PTR (Field);
}



ACPI_GENERIC_OP *
PsGetNextArg (
    ACPI_PARSE_STATE        *ParserState, 
    INT32                   ArgType, 
    UINT32                  *ArgCount)
{
    ACPI_GENERIC_OP         *Arg = NULL;
    ACPI_GENERIC_OP         *Prev = NULL;
    ACPI_GENERIC_OP         *Field;
    INT32                   Subop;



    FUNCTION_TRACE_PTR ("PsGetNextArg", ParserState);


    switch (ArgType)
    {
    case ARGP_BYTEDATA:
    case ARGP_WORDDATA:
    case ARGP_DWORDDATA:
    case ARGP_CHARLIST:
    case ARGP_NAME:
    case ARGP_NAMESTRING:

        /* constants, strings, and namestrings are all the same size */

        Arg = PsAllocOp (AML_ByteOp);
        if (Arg)
        {
            PsGetNextSimpleArg (ParserState, ArgType, Arg);
        }
        break;


    case ARGP_PKGLENGTH:

        /* package length, nothing returned */

        ParserState->PkgEnd = PsGetNextPackageEnd (ParserState);
        break;


    case ARGP_FIELDLIST:

        if (ParserState->Aml < ParserState->PkgEnd)
        {
            /* non-empty list */

            while (ParserState->Aml < ParserState->PkgEnd)
            {
                Field = PsGetNextField (ParserState);
                if (!Field)
                {
                    break;
                }

                if (Prev)
                {
                    Prev->Next = Field;
                }

                else
                {
                    Arg = Field;
                }

                Prev = Field;
            }

            /* skip to End of byte data */

            ParserState->Aml = ParserState->PkgEnd;
        }
        break;


    case ARGP_BYTELIST:

        if (ParserState->Aml < ParserState->PkgEnd)
        {
            /* non-empty list */

            Arg = PsAllocOp (AML_BYTELIST_OP);
            if (Arg)
            {
                /* fill in bytelist data */

                Arg->Value.Size = (ParserState->PkgEnd - ParserState->Aml);
                PsToBytelistOp (Arg)->Data = ParserState->Aml;
            }

            /* skip to End of byte data */

            ParserState->Aml = ParserState->PkgEnd;
        }
        break;


    case ARGP_TARGET:
    case ARGP_SUPERNAME:
        {
            Subop = PsPeekOpcode (ParserState);
            if (Subop == 0              || 
                PsIsLeadingChar (Subop) || 
                PsIsPrefixChar (Subop))
            {
                /* NullName or NameString */

                Arg = PsAllocOp (AML_NAMEPATH_OP);
                if (Arg)
                {
                    PsGetNextNamepath (ParserState, Arg, ArgCount, 0);
                }
            }

            else
            {
                /* single complex argument, nothing returned */

                *ArgCount = 1;
            }
        }
        break;
            

    case ARGP_DATAOBJ:
    case ARGP_TERMARG:

        /* single complex argument, nothing returned */

        *ArgCount = 1;
        break;


    case ARGP_DATAOBJLIST:
    case ARGP_TERMLIST:
    case ARGP_OBJLIST:

        if (ParserState->Aml < ParserState->PkgEnd)
        {
            /* non-empty list of variable arguments, nothing returned */

            *ArgCount = ACPI_VAR_ARGS;
        }
        break;
    }

    return_PTR (Arg);
}
