/*
 *  ./acpica/subsystem/parser/psfind.c - Parse tree search routine
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __PSFIND_C__

#include <acpi.h>
#include <parser.h>
#include "amlcode.h"

#define _COMPONENT          PARSER
        MODULE_NAME         ("psfind");



ACPI_GENERIC_OP*
PsGetParent (
    ACPI_GENERIC_OP         *Op)
{
    ACPI_GENERIC_OP         *Parent = Op;


    /* Traverse the tree upward (to root if necessary) */

    while (Parent)
    {
        switch (Parent->Opcode)
        {
        case AML_ScopeOp:
        case AML_PackageOp:
        case AML_MethodOp:
        case AML_DeviceOp:
        case AML_PowerResOp:
        case AML_ThermalZoneOp:

            return Parent->Parent;
        }

        Parent = Parent->Parent;
    }

    return Parent;
}



ACPI_GENERIC_OP *
PsFindName (
    ACPI_GENERIC_OP         *Scope, 
    UINT32                  Name,
    UINT32                  Opcode)
{
    ACPI_GENERIC_OP         *Op;
    ACPI_GENERIC_OP         *Field;


    /* search scope level for matching name segment */

    Op = PsGetChild (Scope);

    while (Op)
    {
        if (PsIsFieldOp (Op->Opcode))
        {
            /* Field, search named fields */

            Field = PsGetChild (Op);
            while (Field)
            {
                if (PsIsNamedOp (Field->Opcode) 
                   && PsGetName (Field) == Name
                   && (!Opcode || Field->Opcode == Opcode))
                {
                    return Field;
                }

                Field = Field->Next;
            }
        }

        else if (PsIsNamedOp (Op->Opcode)
                && PsGetName (Op) == Name
                && (!Opcode || Op->Opcode == Opcode))
        {
            break;
        }
        
        Op = Op->Next;
    }

    return Op;
}



ACPI_GENERIC_OP*
PsFind (
    ACPI_GENERIC_OP         *Scope, 
    char                    *Path,
    UINT16                  Opcode,
    UINT32                  Create)
{
    UINT32                  segCount;
    UINT32                  unprefixed = 1;
    UINT32                  Name;
    UINT32                  NameOp;
    ACPI_GENERIC_OP         *Op = NULL;


    FUNCTION_TRACE_PTR ("PsFind", Scope);



    if (!Scope || !Path)
    {
        DEBUG_PRINT (TRACE_PARSE, ("PsFind: Null path (%p) or scope (%p)!\n", Path, Scope));
        return_VALUE (NULL);
    }


    Gbl_PsFindCount++;


    /* Handle all prefixes in the name path */

    while (PsIsPrefixChar (GET8 (Path)))
    {
        switch (GET8 (Path))
        {

        case '\\':
            
            /* Could just use a global for "root scope" here */

            while (Scope->Parent)
            {
                Scope = Scope->Parent;
            }

            /* get first object within the scope */
            /* TBD: OR - set next in root scope to point to the same value as arg */

            /* Scope = Scope->Value.Arg; */

            break;


        case '^':

            /* Go up to the next valid scoping Op (method, scope, etc.) */

            if (PsGetParent (Scope))
            {
                Scope = PsGetParent (Scope);
            }

            break;
        }

        unprefixed = 0;
        Path++;
    }

    /* get name segment count */

    switch (GET8 (Path))
    {
    case '\0':
        segCount = 0;

        /* Null name case */

        if (unprefixed)
        {
            Op = NULL;
        }
        else
        {
            Op = Scope;
        }


        DEBUG_PRINT (TRACE_PARSE, ("PsFind: Null path, returning current root scope Op=%p\n", Op));
        return_PTR (Op);
        break;

    case AML_DualNamePrefix:
        segCount = 2;
        Path++;
        break;

    case AML_MultiNamePrefixOp:
        segCount = GET8 (Path + 1);
        Path += 2;
        break;

    default:
        segCount = 1;
        break;
    }
        
    DEBUG_PRINT (TRACE_PARSE, ("PsFind: Search scope %p Segs=%d Opcode=%4.4X Create=%d\n", Scope, segCount, Opcode, Create));

    /* match each name segment */

    while (Scope && segCount)
    {
        Name = GET32 (Path);
        Path += 4;
        segCount --;

        if (segCount)
        {
            NameOp = 0;
        }
        else
        {
            NameOp = Opcode;
        }

        Op = PsFindName (Scope, Name, NameOp);
        if (Op)
        {
            DEBUG_PRINT (TRACE_PARSE, ("PsFind: [%4.4s] Found! Op=%X Opcode=%4.4X\n", &Name, Op, Op->Opcode));
        }

        if (!Op)
        {
            if (Create)
            {
                /* Create a new Scope level */

                if (segCount)
                {
                    Op = PsAllocOp (AML_ScopeOp);
                }
                else
                {
                    Op = PsAllocOp (Opcode);
                }

                if (Op)
                {
                    PsSetName (Op, Name);
                    PsAppendArg (Scope, Op);

                    DEBUG_PRINT (TRACE_PARSE, ("PsFind: [%4.4s] Not found, created Op=%X Opcode=%4.4X\n", &Name, Op, Opcode));
                }
            }

            else if (unprefixed)
            {
                /* Search higher scopes for unprefixed name */

                while (!Op && Scope->Parent)
                {
                    Scope = Scope->Parent;
                    Op = PsFindName (Scope, Name, Opcode);
                    if (Op)
                    {
                        DEBUG_PRINT (TRACE_PARSE, ("PsFind: [%4.4s] Found in parent tree! Op=%X Opcode=%4.4X\n", &Name, Op, Op->Opcode));
                    }
                }
            }

            else
            {
                DEBUG_PRINT (TRACE_PARSE, ("PsFind: Segment [%4.4s] Not Found in scope %p!\n", &Name, Scope));
            }
        }

        unprefixed = 0;
        Scope = Op;
    }

    return_PTR (Op);
}


