/*
 *  ./acpica/subsystem/parser/psparse.c - Parser top level AML parse routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



/*
 * Parse the AML and build an operation tree as most interpreters,
 * like Perl, do.  Parsing is done by hand rather than with a YACC
 * generated parser to tightly constrain stack and dynamic memory
 * usage.  At the same time, parsing is kept flexible and the code
 * fairly compact by parsing based on a list of AML opcode
 * templates in Gbl_AmlOpInfo[]
 */

#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <amlcode.h>
#include <namesp.h>
#include <debugger.h>

#define _COMPONENT          PARSER
        MODULE_NAME         ("psparse");


UINT32                      Gbl_Depth = 0;
extern UINT32               Gbl_ScopeDepth;




ACPI_STATUS 
PsDeleteCompletedOp (
    ACPI_WALK_STATE         *State,
    ACPI_GENERIC_OP         *Op)
{

    PsFreeOp (Op);
    return AE_OK;
}


#ifndef PARSER_ONLY

void
PsDeleteParseTree (
    ACPI_GENERIC_OP         *Root)
{
    ACPI_GENERIC_OP         *Op;
    ACPI_WALK_STATE         WalkState;



    WalkState.Origin = Root;
    Op = Root;

    /* TBD: hack for root case */

    if (Op == Gbl_ParsedNamespaceRoot)
    {
        Op = PsGetChild (Op);
    }

    /* Save the root until last, so that we know when the tree has been walked */

    WalkState.NextOp = Op;
    WalkState.NextOpInfo = NEXT_OP_DOWNWARD;

    while (WalkState.NextOp)
    {
        PsGetNextWalkOp (&WalkState, WalkState.NextOp, PsDeleteCompletedOp);
    }
}
#endif


UINT32
PsGetOpcodeSize (
    UINT32                  Opcode)
{
    
    /* Extended (2-byte) opcode if > 255 */

    if (Opcode > 0x00FF)
    {
        return 2;
    }

    /* Otherwise, just a single byte opcode */

    return 1;
}



UINT16
PsPeekOpcode (
    ACPI_PARSE_STATE        *ParserState)
{
    UINT8                   *Aml;
    UINT16                  Opcode;


    Aml = ParserState->Aml;
    Opcode = (UINT16) GET8 (Aml);

    Aml++;


    /* 
     * Original code special cased LNOTEQUAL, LLESSEQUAL, LGREATEREQUAL.  These opcodes are 
     * no longer recognized. Instead, they are broken into two opcodes.
     *
     *
     *    if (Opcode == AML_EXTOP
     *       || (Opcode == AML_LNOT
     *          && (GET8 (Aml) == AML_LEQUAL
     *               || GET8 (Aml) == AML_LGREATER
     *               || GET8 (Aml) == AML_LLESS)))      extended Opcode, !=, <=, or >= 
     */

    if (Opcode == AML_EXTOP)
    {
        /* Extended opcode */

        Opcode = (UINT16) ((Opcode << 8) | GET8 (Aml));
        Aml++;
    }

    /* don't convert bare name to a namepath */

    return Opcode;
}



ACPI_PARSE_STATE *
PsCreateState (
    UINT8                   *Aml, 
    INT32                   AmlSize)
{
    ACPI_PARSE_STATE        *ParserState; 


    FUNCTION_TRACE ("PsInitState");


    ParserState = CmCallocate (sizeof (ACPI_PARSE_STATE));
    if (!ParserState)
    {
        return_VALUE (NULL);
    }

    ParserState->Aml       = Aml;
    ParserState->AmlEnd    = Aml + AmlSize;
    ParserState->PkgEnd    = ParserState->AmlEnd;
    ParserState->AmlStart  = Aml;


    return_PTR (ParserState);
}


#ifdef PARSER_ONLY

ACPI_STATUS
PsFindObject (
    UINT16                  Opcode,
    ACPI_PARSE_STATE        *ParserState,
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         **Op)
{
    char                    *Path;



    /* Find the name in the parse tree */

    Path = PsGetNextNamestring (ParserState);

    *Op = PsFind (PsGetParentScope (ParserState),
                 Path, Opcode, 1);

    if (!(*Op))
    {
        return AE_NOT_FOUND;
    }

    return AE_OK;
}
#else

ACPI_STATUS
PsFindObject (
    UINT16                  Opcode,
    ACPI_PARSE_STATE        *ParserState,
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         **OutOp)
{
    char                    *Path;
    ACPI_GENERIC_OP         *Op;
    ACPI_OBJECT_TYPE        DataType; 
    ACPI_STATUS             Status;
    NAME_TABLE_ENTRY        *Nte = NULL;


    FUNCTION_TRACE ("PsFindInNamespace");

    
    /*
     * The full parse tree has already been deleted -- therefore, we are parsing
     * a control method.  We can lookup the name in the namespace instead of
     * the parse tree!
     */


    Path = PsGetNextNamestring (ParserState);

    /* Map the raw opcode into an internal object type */

    DataType = DsMapNamedOpcodeToDataType (Opcode);

    /* 
     * Enter the object into the namespace
     */

    Status = NsLookup (WalkState->ScopeInfo, Path, DataType, IMODE_LoadPass1,    /* Create if not found */
                            NS_NO_UPSEARCH, WalkState, &Nte);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Create a new op */

    Op = PsAllocOp (Opcode);
    if (!Op)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* Initialize */

    ((ACPI_NAMED_OP *)Op)->Name = Nte->Name;
    Op->NameTableEntry = Nte;


    PsAppendArg (PsGetParentScope (ParserState), Op);

    *OutOp = Op;


    return_ACPI_STATUS (AE_OK);
}
#endif












ACPI_STATUS
PsParseLoop (
    ACPI_PARSE_STATE        *ParserState,
    ACPI_WALK_STATE         *WalkState)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_GENERIC_OP         *Op = NULL;            /* current op */
    ACPI_OP_INFO            *OpInfo;
    ACPI_GENERIC_OP         *Arg = NULL;
    ACPI_DEFERRED_OP        *DeferredOp;
    UINT32                  ArgCount;       /* push for fixed or variable arguments */
    UINT32                  ArgTypes = 0;
    ACPI_PTRDIFF            AmlOffset;
    UINT16                  Opcode;
    ACPI_GENERIC_OP         PreOp;
 


#ifndef PARSER_ONLY
    ACPI_OBJECT_TYPE        DataType; 
#endif

    FUNCTION_TRACE_PTR ("PsParseLoop", ParserState);



    /*
     * Iterative parsing loop, while there is more aml to process:
     */
    while (ParserState->Aml < ParserState->AmlEnd)
    {
        if (!Op)
        {
            /* Get the next opcode from the AML stream */

            AmlOffset   = ParserState->Aml - ParserState->AmlStart;
            Opcode      = PsPeekOpcode (ParserState);
            OpInfo      = PsGetOpcodeInfo (Opcode);

            /*
             * First cut to determine what we have found:
             * 1) A valid AML opcode
             * 2) A name string
             * 3) An unknown/invalid opcode
             */

            if (OpInfo)
            {
                /* Found opcode info, this is a normal opcode */

                ParserState->Aml += PsGetOpcodeSize (Opcode);
                ArgTypes = OpInfo->ParseArgs;
            }

            else if (PsIsPrefixChar (Opcode) || PsIsLeadingChar (Opcode))
            {
                /* 
                 * Starts with a valid prefix or ASCII char, this is a name string.
                 * Convert the bare name string to a namepath. 
                 */

                Opcode = AML_NAMEPATH_OP;
                ArgTypes = ARGP_NAMESTRING;
            }

            else
            {
                /* The opcode is unrecognized.  We simply skip unknown opcodes */

                DEBUG_PRINT (TRACE_PARSE, ("ParseLoop: Found unknown opcode %lX, skipping\n", Opcode));

                ParserState->Aml += PsGetOpcodeSize (Opcode);
                continue;
            }
        

            /* Create Op structure and append to parent's argument list */

            if (PsIsNamedOp (Opcode))
            {
                PreOp.Value.Arg = NULL;
                PreOp.Opcode = Opcode;

                while (GET_CURRENT_ARG_TYPE (ArgTypes) != ARGP_NAME)
                {
                    Arg = PsGetNextArg (ParserState, GET_CURRENT_ARG_TYPE (ArgTypes), &ArgCount);
                    PsAppendArg (&PreOp, Arg);
                    INCREMENT_ARG_LIST (ArgTypes);
                }

      
                /* We know that this arg is a name, move to next arg */

                INCREMENT_ARG_LIST (ArgTypes);

                Status = PsFindObject (Opcode, ParserState, WalkState, &Op);
                if (ACPI_FAILURE (Status))
                {
                    return_ACPI_STATUS (AE_NOT_FOUND);
                }

                PsAppendArg (Op, PreOp.Value.Arg);
                Gbl_Depth++;


                if (Op->Opcode == AML_RegionOp)
                {
                    DeferredOp = PsToDeferredOp (Op);
                    if (DeferredOp)
                    {
                        /* 
                         * Skip parsing of control method or opregion body, because we don't
                         * have enough info in the first pass to parse them correctly.
                         *
                         * Backup to beginning of OpRegion declaration (2 for Opcode, 4 for name)
                         */

                        DeferredOp->Body        = ParserState->Aml - 6;
                        DeferredOp->BodyLength  = 0; /* Unknown until we parse the body */
                    }
                }
            }

            else
            {
                /* Not a named opcode, just allocate Op and append to parent */

                Op = PsAllocOp (Opcode);
                if (!Op)
                {
                    return_ACPI_STATUS (AE_NO_MEMORY);
                }

                PsAppendArg (PsGetParentScope (ParserState), Op);
            }
        
            Op->AmlOffset = AmlOffset;

            if (OpInfo)
            {
                DEBUG_PRINT (TRACE_PARSE, ("ParseLoop:  Op=%p (%s) Opcode=%4.4lX Offset=%5.5lX\n",
                                Op, OpInfo->Name, Op->Opcode, Op->AmlOffset));
            }
        }


        ArgCount = 0;
        if (ArgTypes)   /* Are there any arguments that must be processed? */
        {
            /* get arguments */

            switch (Op->Opcode)
            {
            case AML_ByteOp:        /* AML_BYTEDATA_ARG */
            case AML_WordOp:        /* AML_WORDDATA_ARG */
            case AML_DWordOp:       /* AML_DWORDATA_ARG */
            case AML_StringOp:      /* AML_ASCIICHARLIST_ARG */

                /* fill in constant or string argument directly */

                PsGetNextSimpleArg (ParserState, GET_CURRENT_ARG_TYPE (ArgTypes), Op);
                break;

            case AML_NAMEPATH_OP:   /* AML_NAMESTRING_ARG */

                PsGetNextNamepath (ParserState, Op, &ArgCount, 1);
                ArgTypes = 0;
                break;


            default:

                /* Op is not a constant or string, append each argument */

                while (GET_CURRENT_ARG_TYPE (ArgTypes) && !ArgCount)
                {
                    AmlOffset = ParserState->Aml - ParserState->AmlStart;
                    Arg = PsGetNextArg (ParserState, GET_CURRENT_ARG_TYPE (ArgTypes), &ArgCount);
                
                    if (Arg)
                    {
                        Arg->AmlOffset = AmlOffset;
                    }

                    PsAppendArg (Op, Arg);
                    INCREMENT_ARG_LIST (ArgTypes);
                }


                /* For a method, save the length and address of the body */

                if (Op->Opcode == AML_MethodOp)
                {
                    DeferredOp = PsToDeferredOp (Op);
                    if (DeferredOp)
                    {
                        /* 
                         * Skip parsing of control method or opregion body, because we don't
                         * have enough info in the first pass to parse them correctly.
                         */

                        DeferredOp->Body        = ParserState->Aml;
                        DeferredOp->BodyLength  = ParserState->PkgEnd - ParserState->Aml;

                        /*
                         * Skip body of method.  For OpRegions, we must continue parsing because the
                         * opregion is not a standalone package (We don't know where the end is).
                         */
                        ParserState->Aml        = ParserState->PkgEnd;
                        ArgCount                = 0;
                    }
                }

                break;
            }
        }

        if (!ArgCount)
        {
            /* completed Op, prepare for next */

            if (PsIsNamedOp (Op->Opcode))
            {
                if (Gbl_Depth)
                {
                    Gbl_Depth--;
                }

                if (Op->Opcode == AML_RegionOp)
                {
                    DeferredOp = PsToDeferredOp (Op);
                    if (DeferredOp)
                    {
                        /* 
                         * Skip parsing of control method or opregion body, because we don't
                         * have enough info in the first pass to parse them correctly.
                         *
                         * Completed parsing an OpRegion declaration, we now know the length.
                         */

                        DeferredOp->BodyLength  = ParserState->Aml - DeferredOp->Body;
                    }
                }
            
            
#ifndef PARSER_ONLY
                DataType = DsMapNamedOpcodeToDataType (Op->Opcode);

                if (Op->Opcode == AML_NameOp)
                {
                    if (Op->Value.Arg)
                    {
            
                        DataType = DsMapOpcodeToDataType ((Op->Value.Arg)->Opcode, NULL);
                        ((NAME_TABLE_ENTRY *)Op->NameTableEntry)->Type = (UINT8) DataType;
                    }
                }

                /* Pop the scope stack */

                if (NsOpensScope (DataType))
                {

                    DEBUG_PRINT (TRACE_DISPATCH, ("AmlEndNamespaceScope/%s: Popping scope for Op %p\n",
                                                    CmGetTypeName (DataType), Op));
                    DsScopeStackPop (WalkState);
                }
#endif                
            }


            ParserState->Scope->ArgCount--;

            if (PsHasCompletedScope (ParserState))
            {
                PsPopScope (ParserState, &Op, &ArgTypes);
            }

            else
            {
                Op = NULL;
            }
        }

        else
        {
            /* complex argument, push Op and prepare for argument */

            PsPushScope (ParserState, Op, ArgTypes, ArgCount);
            Op = NULL;
        }

    } /* while ParserState->Aml */
    

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
PsParseAml (
    ACPI_GENERIC_OP         *StartScope,
    UINT8                   *Aml, 
    UINT32                  AmlSize)
{
    ACPI_STATUS             Status;
    ACPI_PARSE_STATE        *ParserState;
    ACPI_WALK_STATE         *WalkState;
    ACPI_WALK_LIST          WalkList;
    NAME_TABLE_ENTRY        *Nte = NULL;


    FUNCTION_TRACE ("PsParseAml");

    DEBUG_PRINT (TRACE_PARSE, ("PsParseAml: Entered with Scope=%p Aml=%p size=%lX\n", StartScope, Aml, AmlSize));


    /* Initialize parser state and scope */

    ParserState = PsCreateState (Aml, AmlSize);
    if (!ParserState)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    PsInitScope (ParserState, StartScope);


    /* Initialize a new walk list */


    WalkList.WalkState = NULL;

    WalkState = DsCreateWalkState (NULL, NULL, &WalkList);
    if (!WalkState)
    {
        Status = AE_NO_MEMORY;
        goto Cleanup;
    }

    /* TBD: move to createWalkSTate */

    WalkState->OwnerId = 0xD1D1; // MAKE THIS A PARAMETER: OwnerId;

    Nte = ParserState->StartOp->NameTableEntry;
    if (Nte)
    {
        /* Push start scope on scope stack and make it current  */

        Status = DsScopeStackPush (Nte->Scope, Nte->Type, WalkState);
        if (ACPI_FAILURE (Status))
        {
            goto Cleanup;
        }

    }


    /* Create the parse tree */

    Status = PsParseLoop (ParserState, WalkState);


Cleanup:

    /* Cleanup */

    DsDeleteWalkState (WalkState);
    PsCleanupScope (ParserState);
    CmFree (ParserState);


    return_ACPI_STATUS (Status);
}


