/*
 *  ./acpica/subsystem/parser/psscope.c - Parser scope stack management routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




#include <acpi.h>
#include <parser.h>

#define _COMPONENT          PARSER
        MODULE_NAME         ("psscope");




ACPI_GENERIC_OP *
PsGetParentScope (
    ACPI_PARSE_STATE        *ParserState)
{
    return ParserState->Scope->Op;
}



BOOLEAN
PsHasCompletedScope (
    ACPI_PARSE_STATE        *ParserState)
{
    return (BOOLEAN) ((ParserState->Aml >= ParserState->Scope->ArgEnd ||
                        !ParserState->Scope->ArgCount));
}



ACPI_STATUS
PsInitScope (
    ACPI_PARSE_STATE        *ParserState,
    ACPI_GENERIC_OP         *Root)
{
    ACPI_PARSE_SCOPE        *Scope;
    
    
    Scope = CmCallocate (sizeof (ACPI_PARSE_SCOPE));
    if (!Scope)
    {
        return AE_NO_MEMORY;
    }

    Scope->Op               = Root;
    Scope->ArgCount         = ACPI_VAR_ARGS;
    Scope->ArgEnd           = ParserState->AmlEnd;
    Scope->PkgEnd           = ParserState->AmlEnd;
    ParserState->Scope      = Scope;
    ParserState->StartOp    = Root;

    return AE_OK;
}



ACPI_STATUS
PsPushScope (
    ACPI_PARSE_STATE        *ParserState,
    ACPI_GENERIC_OP         *Op, 
    UINT32                  RemainingArgs,
    UINT32                  ArgCount)
{
    ACPI_PARSE_SCOPE        *Scope = ParserState->ScopeAvail;


    FUNCTION_TRACE ("PsPushScope");


    if (Scope)
    {
        /* grabbed scope from available list */

        ParserState->ScopeAvail = Scope->Parent;
    }

    else
    {
        /* allocate scope from the heap */

        Scope = (ACPI_PARSE_SCOPE*) CmAllocate (sizeof (ACPI_PARSE_SCOPE));
        if (!Scope)
        {
            return_ACPI_STATUS (AE_NO_MEMORY);
        }
    }

    /* Always zero out the scope before init */

    MEMSET (Scope, 0, sizeof (*Scope));

    Scope->Op           = Op;
    Scope->ArgList      = RemainingArgs;
    Scope->ArgCount     = ArgCount;
    Scope->PkgEnd       = ParserState->PkgEnd;
    Scope->Parent       = ParserState->Scope;
    ParserState->Scope  = Scope;
    
    if (ArgCount == ACPI_VAR_ARGS)
    {
        /* multiple arguments */

        Scope->ArgEnd = ParserState->PkgEnd;
    }
    
    else
    {
        /* single argument */

        Scope->ArgEnd = ACPI_MAX_AML;
    }
 
    return_ACPI_STATUS (AE_OK);
}



void
PsPopScope (
    ACPI_PARSE_STATE        *ParserState,
    ACPI_GENERIC_OP         **Op,
    UINT32                  *ArgList)
{
    ACPI_PARSE_SCOPE        *Scope = ParserState->Scope;


    FUNCTION_TRACE ("PsPopScope");


    if (Scope->Parent)
    {
        /* return to parsing previous op */
        
        *Op                     = Scope->Op;
        *ArgList                = Scope->ArgList;
        ParserState->PkgEnd     = Scope->PkgEnd;
        ParserState->Scope      = Scope->Parent;

        /* add scope to available list */

        Scope->Parent           = ParserState->ScopeAvail;
        ParserState->ScopeAvail = Scope;
    }

    else
    {
        /* empty parse stack, prepare to fetch next opcode */

        *Op                     = NULL;
        *ArgList                = 0;
    }

    return_VOID;
}



void
PsCleanupScope (
    ACPI_PARSE_STATE        *ParserState)
{
    ACPI_PARSE_SCOPE        *Scope; 


    FUNCTION_TRACE_PTR ("PsCleanupScope", ParserState);


    if (!ParserState)
    {
        return;
    }

    /* destroy available list */

    while (ParserState->ScopeAvail)
    {
        Scope = ParserState->ScopeAvail;
        ParserState->ScopeAvail = Scope->Parent;
        CmFree (Scope);
    }

    /* destroy scope stack */

    while (ParserState->Scope)
    {
        Scope = ParserState->Scope;
        ParserState->Scope = Scope->Parent;
        CmFree (Scope);
    }

    return_VOID;
}

