/*
 *  ./acpica/subsystem/parser/pstree.c - Parser op tree manipulation/traversal/search 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __PSTREE_C__

#include <acpi.h>
#include <parser.h>
#include "amlcode.h"

#define _COMPONENT          PARSER
        MODULE_NAME         ("pstree");




ACPI_GENERIC_OP *
PsGetArg (
    ACPI_GENERIC_OP         *Op, 
    UINT32                  Argn)
{
    ACPI_GENERIC_OP         *Arg = NULL;
    ACPI_OP_INFO            *OpInfo;



    /* Get the info structure for this opcode */

    OpInfo = PsGetOpcodeInfo (Op->Opcode);
    if (!OpInfo)
    {
        /* Invalid opcode */

        return NULL;
    }

    /* Check if this opcode requires argument sub-objects */

    if (!(OpInfo->Flags & OP_INFO_HAS_ARGS))
    {
        /* Has no linked argument objects */

        return NULL;
    }

    /* Get the requested argument object */

    Arg = Op->Value.Arg;
    while (Arg && Argn)
    {
        Argn--;
        Arg = Arg->Next;
    }

    return Arg;
}



void
PsAppendArg (
    ACPI_GENERIC_OP         *Op, 
    ACPI_GENERIC_OP         *Arg)
{
    ACPI_GENERIC_OP         *PrevArg;
    ACPI_OP_INFO            *OpInfo;


    if (!Op)
    {
        return;
    }

    /* Get the info structure for this opcode */

    OpInfo = PsGetOpcodeInfo (Op->Opcode);
    if (!OpInfo)
    {
        /* Invalid opcode */

        return;
    }

    /* Check if this opcode requires argument sub-objects */

    if (!(OpInfo->Flags & OP_INFO_HAS_ARGS))
    {
        /* Has no linked argument objects */

        return;
    }

            

    /* Append the argument to the linked argument list */

    if (Op->Value.Arg)
    {
        /* Append to existing argument list */

        PrevArg = Op->Value.Arg;
        while (PrevArg->Next)
        {
            PrevArg = PrevArg->Next;
        }
        PrevArg->Next = Arg;
    }

    else
    {
        /* No argument list, this will be the first argument */

        Op->Value.Arg = Arg;
    }


    /* Set the parent in this arg and any args linked after it */

    while (Arg)
    {
        Arg->Parent = Op;
        Arg = Arg->Next;
    }
}



ACPI_GENERIC_OP *
PsGetChild (
    ACPI_GENERIC_OP         *Op)
{
    ACPI_GENERIC_OP         *Child = NULL;


    switch (Op->Opcode)
    {
    case AML_ScopeOp:
    case AML_ElseOp:
    case AML_DeviceOp:
    case AML_ThermalZoneOp:
    case AML_METHODCALL_OP:

        Child = PsGetArg (Op, 0);
        break;


    case AML_BufferOp:
    case AML_PackageOp:
    case AML_MethodOp:
    case AML_IfOp:
    case AML_WhileOp:
    case AML_DefFieldOp:

        Child = PsGetArg (Op, 1);
        break;


    case AML_PowerResOp:
    case AML_IndexFieldOp:

        Child = PsGetArg (Op, 2);
        break;


    case AML_ProcessorOp:
    case AML_BankFieldOp:

        Child = PsGetArg (Op, 3);
        break;

    }

    return Child;
}



ACPI_GENERIC_OP *
PsGetDepthNext (
    ACPI_GENERIC_OP         *Origin, 
    ACPI_GENERIC_OP         *Op)
{
    ACPI_GENERIC_OP         *Next = NULL;
    ACPI_GENERIC_OP         *Parent;
    ACPI_GENERIC_OP         *Arg;


    if (!Op)
    {
        return NULL;
    }

    /* look for an argument or child */

    Next = PsGetArg (Op, 0);
    if (Next)
    {
        return Next;
    }

    /* look for a sibling */

    Next = Op->Next;
    if (Next)
    {
        return Next;
    }

    /* look for a sibling of parent */

    Parent = Op->Parent;

    while (Parent)
    {
        Arg = PsGetArg (Parent, 0);
        while (Arg && (Arg != Origin) && (Arg != Op))
        {
            Arg = Arg->Next;
        }

        if (Arg == Origin)
        {
            /* reached parent of origin, end search */

            return NULL;
        }

        if (Parent->Next)
        {
            /* found sibling of parent */
            return Parent->Next;
        }

        Op = Parent;
        Parent = Parent->Parent;
    }

    return Next;
}



ACPI_GENERIC_OP *
PsFetchPrefix (
    ACPI_GENERIC_OP         *Scope, 
    char                    **Path, 
    UINT32                  io)
{
    UINT32                  prefix = io ? GET8 (*Path):**Path;


    switch (prefix)
    {
    case '\\':
    case '/':

        /* go to the root */

        *Path += 1;
        while (Scope->Parent)
        {
            Scope = Scope->Parent;
        }
        break;


    case '^':

        /* go up one level */

        *Path += 1;
        Scope = Scope->Parent;
        break;
    }

    if (Scope && !Scope->Parent)
    {
        /* searching from the root, start with its children */

        Scope = PsGetChild (Scope);
    }

    return Scope;
}



UINT32
PsFetchName (
    char                    **Path, 
    UINT32                  io)
{
    UINT32                  Name = 0;
    char                    *nm;
    UINT32                  i;
    char                    ch;


    if (io)
    {
        Name = GET32 (*Path);
        *Path += 4;
    }

    else
    {      
        if (**Path == '.')
        {
            *Path += 1;
        }
        
        nm = (char*) &Name;
        for (i = 0; i < 4; i++)
        {
            ch = **Path;
            if (ch && ch != '.')
            {
                *nm = ch;
                *Path += 1;
            }

            else
            {
                *nm = '_';
            }
            nm++;
        }
    }

    return Name;
}


