/*
 *  ./acpica/subsystem/parser/psutils.c - Parser miscellaneous utilities (Parser only)
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include "amlcode.h"

#define _COMPONENT          PARSER
        MODULE_NAME         ("psutils");


#define PARSEOP_GENERIC     1
#define PARSEOP_NAMED       2
#define PARSEOP_DEFERRED    3
#define PARSEOP_BYTELIST    4
        


void
PsInitOp (
    ACPI_GENERIC_OP         *Op,
    UINT16                  Opcode)
{
    ACPI_OP_INFO             *AmlOp;


    Op->DataType = DESC_TYPE_PARSER;
    Op->Opcode = Opcode;


    AmlOp = PsGetOpcodeInfo (Opcode);
    if (AmlOp)
    {
        /* Debug only! */

        DEBUG_EXEC (STRNCPY (Op->OpName, AmlOp->Name, sizeof (Op->OpName)));
    }
}



ACPI_GENERIC_OP*
PsAllocOp (
    UINT16                  Opcode)
{
    ACPI_GENERIC_OP         *Op = NULL;
    UINT32                  Size;
    UINT8                   Flags;


    /* Allocate the minimum required size object */

    if (PsIsDeferredOp (Opcode))
    {
        Size = sizeof (ACPI_DEFERRED_OP);
        Flags = PARSEOP_DEFERRED;
    }

    else if (PsIsNamedOp (Opcode))
    {
        Size = sizeof (ACPI_NAMED_OP);
        Flags = PARSEOP_NAMED;
    }
    
    else if (PsIsBytelistOp (Opcode))
    {
        Size = sizeof (ACPI_BYTELIST_OP);
        Flags = PARSEOP_BYTELIST;
    }

    else
    {
        Size = sizeof (ACPI_GENERIC_OP);
        Flags = PARSEOP_GENERIC;

        /*
         * The generic op is by far the most common (16 to 1), and therefore the op cache is 
         * implemented with this type.
         *
         * Check if there is an Op already available in the cache
         */

        CmAcquireMutex (MTX_CACHES);
        Gbl_ParseCacheRequests++;
        if (Gbl_ParseCache)
        {
            /* Extract an op from the front of the cache list */

            Gbl_ParseCacheDepth--;
            Gbl_ParseCacheHits++;
            
            Op = Gbl_ParseCache;
            Gbl_ParseCache = Op->Next;

            /* Clear the previously used Op */

            MEMSET (Op, 0, sizeof (ACPI_GENERIC_OP));
        }
        CmReleaseMutex (MTX_CACHES);
    }

    /* Allocate a new Op if necessary */
    
    if (!Op)
    {
        Op = CmCallocate (Size);
    }

    /* Initialize the Op */
    if (Op)
    {
        PsInitOp (Op, Opcode);
        Op->Flags = Flags;
    }

    return Op;
}




void
PsFreeOp (
    ACPI_GENERIC_OP         *Op)
{


    if (Op->Flags == PARSEOP_GENERIC)
    {
        /* Is the cache full? */

        if (Gbl_ParseCacheDepth < MAX_PARSE_CACHE_DEPTH)
        {
            /* Put a GENERIC_OP back into the cache */

            CmAcquireMutex (MTX_CACHES);
            Gbl_ParseCacheDepth++;

            Op->Next = Gbl_ParseCache;
            Gbl_ParseCache = Op;

            CmReleaseMutex (MTX_CACHES);
            return;
        }
    }

    /*
     * Not a GENERIC OP, or the cache is full, just free the Op 
     */

    CmFree (Op);
}




void
PsDeleteParseCache (
    void)
{
    ACPI_GENERIC_OP         *Next;


    FUNCTION_TRACE ("PsDeleteParseCache");


    /* Traverse the global cache list */

    while (Gbl_ParseCache)
    {
        /* Delete one cached state object */

        Next = Gbl_ParseCache->Next;
        CmFree (Gbl_ParseCache);
        Gbl_ParseCache = Next;
    }

    return_VOID;
}








/*
 * Is "c" a namestring lead character?
 */


BOOLEAN
PsIsLeadingChar (
    INT32                   c)
{
    return ((BOOLEAN) (c == '_' || (c >= 'A' && c <= 'Z')));
}


/*
 * Is "c" a namestring prefix character?
 */
BOOLEAN
PsIsPrefixChar (
    INT32                   c)
{
    return ((BOOLEAN) (c == '\\' || c == '^'));
}


BOOLEAN
PsIsNamespaceObjectOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
           (Opcode == AML_ScopeOp           || 
            Opcode == AML_DeviceOp          || 
            Opcode == AML_ThermalZoneOp     || 
            Opcode == AML_MethodOp          || 
            Opcode == AML_PowerResOp        || 
            Opcode == AML_ProcessorOp       || 
            Opcode == AML_DefFieldOp        || 
            Opcode == AML_IndexFieldOp      || 
            Opcode == AML_BankFieldOp       || 
            Opcode == AML_NAMEDFIELD_OP     || 
            Opcode == AML_NameOp            || 
            Opcode == AML_AliasOp           || 
            Opcode == AML_MutexOp           || 
            Opcode == AML_EventOp           || 
            Opcode == AML_RegionOp          ||
            Opcode == AML_CreateFieldOp     ||
            Opcode == AML_BitFieldOp        ||
            Opcode == AML_ByteFieldOp       ||
            Opcode == AML_WordFieldOp       ||
            Opcode == AML_DWordFieldOp      ||
            Opcode == AML_METHODCALL_OP     ||
            Opcode == AML_NAMEPATH_OP));
}

BOOLEAN
PsIsNamespaceOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
           (Opcode == AML_ScopeOp           ||
            Opcode == AML_DeviceOp          ||
            Opcode == AML_ThermalZoneOp     ||
            Opcode == AML_MethodOp          ||
            Opcode == AML_PowerResOp        ||
            Opcode == AML_ProcessorOp       ||
            Opcode == AML_DefFieldOp        ||
            Opcode == AML_IndexFieldOp      || 
            Opcode == AML_BankFieldOp       ||
            Opcode == AML_NameOp            ||
            Opcode == AML_AliasOp           ||
            Opcode == AML_MutexOp           ||
            Opcode == AML_EventOp           ||
            Opcode == AML_RegionOp          ||
            Opcode == AML_NAMEDFIELD_OP)); 
}




/*
 * Is opcode for a named object Op?
 * (Includes all named object opcodes)
 *
 * TBD: Need a better way than this brute force approach!
 */
BOOLEAN
PsIsNamedObjectOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
           (Opcode == AML_ScopeOp           || 
            Opcode == AML_DeviceOp          || 
            Opcode == AML_ThermalZoneOp     || 
            Opcode == AML_MethodOp          || 
            Opcode == AML_PowerResOp        || 
            Opcode == AML_ProcessorOp       || 
            Opcode == AML_NAMEDFIELD_OP     || 
            Opcode == AML_NameOp            || 
            Opcode == AML_AliasOp           || 
            Opcode == AML_MutexOp           || 
            Opcode == AML_EventOp           || 
            Opcode == AML_RegionOp          ||


            Opcode == AML_CreateFieldOp     ||
            Opcode == AML_BitFieldOp        ||
            Opcode == AML_ByteFieldOp       ||
            Opcode == AML_WordFieldOp       ||
            Opcode == AML_DWordFieldOp      ||
            Opcode == AML_METHODCALL_OP     ||
            Opcode == AML_NAMEPATH_OP));
}


/*
 * Is opcode for a named Op?
 */
BOOLEAN
PsIsNamedOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
           (Opcode == AML_ScopeOp           ||
            Opcode == AML_DeviceOp          ||
            Opcode == AML_ThermalZoneOp     ||
            Opcode == AML_MethodOp          ||
            Opcode == AML_PowerResOp        ||
            Opcode == AML_ProcessorOp       ||
            Opcode == AML_NameOp            ||
            Opcode == AML_AliasOp           ||
            Opcode == AML_MutexOp           ||
            Opcode == AML_EventOp           ||
            Opcode == AML_RegionOp          ||
            Opcode == AML_NAMEDFIELD_OP)); 
}


BOOLEAN
PsIsDeferredOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
           (Opcode == AML_MethodOp ||
            Opcode == AML_RegionOp));
}


/*
 * Is opcode for a bytelist?
 */
BOOLEAN
PsIsBytelistOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) (Opcode == AML_BYTELIST_OP));
}


/*
 * Is opcode for a Field, IndexField, or BankField
 */
BOOLEAN
PsIsFieldOp (
    UINT16                  Opcode)
{
    return ((BOOLEAN) 
              (Opcode == AML_CreateFieldOp
            || Opcode == AML_DefFieldOp
            || Opcode == AML_IndexFieldOp
            || Opcode == AML_BankFieldOp));
}


/*
 * Cast an acpi_op to an acpi_deferred_op if possible
 */
ACPI_DEFERRED_OP *
PsToDeferredOp (
    ACPI_GENERIC_OP         *Op)
{
    return (PsIsDeferredOp (Op->Opcode)
            ? ( (ACPI_DEFERRED_OP *) Op) : NULL);
}


/*
 * Cast an acpi_op to an acpi_named_op if possible
 */
ACPI_NAMED_OP*
PsToNamedOp (
    ACPI_GENERIC_OP         *Op)
{
    return (PsIsNamedOp (Op->Opcode)
            ? ( (ACPI_NAMED_OP *) Op) : NULL);
}


/*
 * Cast an acpi_op to an acpi_bytelist_op if possible
 */
ACPI_BYTELIST_OP*
PsToBytelistOp (
    ACPI_GENERIC_OP         *Op)
{
    return (PsIsBytelistOp (Op->Opcode)
            ? ( (ACPI_BYTELIST_OP*) Op) : NULL);
}


/*
 * Get op's name (4-byte name segment) or 0 if unnamed
 */
UINT32
PsGetName (
    ACPI_GENERIC_OP         *Op)
{
    ACPI_NAMED_OP               *Named = PsToNamedOp (Op);

    return (Named ? Named->Name : 0);
}


/*
 * Set op's name
 */
void
PsSetName (
    ACPI_GENERIC_OP         *Op, 
    UINT32                  name)
{
    ACPI_NAMED_OP           *Named = PsToNamedOp (Op);

    if (Named)
    {
        Named->Name = name;
    }
}

