/*
 *  ./acpica/subsystem/parser/pswalk.c - Parser routines to walk parsed op tree(s)
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




#include <acpi.h>
#include <amlcode.h>
#include <parser.h>
#include <dispatch.h>
#include <namesp.h>
#include <interp.h>

#define _COMPONENT          PARSER
        MODULE_NAME         ("pswalk");




ACPI_STATUS
PsGetNextWalkOp (
    ACPI_WALK_STATE         *WalkState,
    ACPI_GENERIC_OP         *Op,
    INTERPRETER_CALLBACK    AscendingCallback)
{
    ACPI_GENERIC_OP         *Next;
    ACPI_GENERIC_OP         *Parent;
    ACPI_GENERIC_OP         *GrandParent;
    ACPI_STATUS             Status;


    FUNCTION_TRACE_PTR ("PsGetNextWalkOp", Op);


    /* Check for a argument only if we are descending in the tree */

    if (WalkState->NextOpInfo != NEXT_OP_UPWARD)
    {
        /* Look for an argument or child of the current op */

        Next = PsGetArg (Op, 0);
        if (Next)
        {
            /* Still going downward in tree (Op is not completed yet) */

            WalkState->PrevOp       = Op;
            WalkState->NextOp       = Next;
            WalkState->NextOpInfo   = NEXT_OP_DOWNWARD;

            return_ACPI_STATUS (AE_OK);
        }


        /* 
         * No more children, this Op is complete.  Save Next and Parent
         * in case the Op object gets deleted by the callback routine
         */

        Next    = Op->Next;
        Parent  = Op->Parent;

        Status = AscendingCallback (WalkState, Op);

        switch (Status)
        {
        case AE_CTRL_TERMINATE:

            /* 
             * A control method was terminated via a RETURN statement.
             * The walk of this method is complete.
             */
            WalkState->PrevOp       = WalkState->Origin;
            WalkState->NextOp       = NULL;

            return_ACPI_STATUS (AE_OK);
            break;


        case AE_CTRL_FALSE:

            /*
             * Either an IF/WHILE Predicate was false or we encountered a BREAK opcode
             *,In both cases, we do not execute the rest of the package;  We simply
             * close out the parent (finishing the walk of this branch of the tree)
             * and continue execution at the parent level.
             */

            Next        = Parent->Next;
            Status      = AE_OK;

            /*
             * If there is a sibling to the parent, we must close out the parent now,
             * because we are going to continue to go downward (to the sibling)
             * in the parse tree.
             */
            if (Next)
            {
                Status = AscendingCallback (WalkState, Parent);

                /* The parent sibling will be next */

                WalkState->PrevOp       = Op;
                WalkState->NextOp       = Next;
                WalkState->NextOpInfo   = NEXT_OP_DOWNWARD;

                /* Continue downward */

                return_ACPI_STATUS (AE_OK);
            }

            /* Drop into the loop below because we are moving upwards in the tree */

            break;


        default:
            /* 
             * Check for a sibling to the current op.  A sibling means
             * we are still going "downward" in the tree.
             */

            if (Next)
            {
                /* There is a sibling, it will be next */

                WalkState->PrevOp       = Op;
                WalkState->NextOp       = Next;
                WalkState->NextOpInfo   = NEXT_OP_DOWNWARD;

                /* Continue downward */

                return_ACPI_STATUS (Status);
            }

            /* 
             * No sibling, but check status.
             * Abort on error from callback routine 
             */
            if (Status != AE_OK)
            {
                /* Next op will be the parent */

                WalkState->PrevOp       = Op;
                WalkState->NextOp       = Parent;
                WalkState->NextOpInfo   = NEXT_OP_UPWARD;

                return_ACPI_STATUS (Status);
            }

            /* Drop into the loop below because we are moving upwards in the tree */

            break;
        }
    }

    else
    {
        /*
         * We are resuming a walk, and we were (are) going upward in the tree.
         * So, we want to drop into the parent loop below.
         */

        Parent = Op;
    }


    /*
     * Look for a sibling of the current Op's parent
     * Continue moving up the tree until we find a node that has not been 
     * visited, or we get back to where we started.
     */
    while (Parent)
    {
        /* We are moving up the tree, therefore this parent Op is complete */

        GrandParent = Parent->Parent;
        Next        = Parent->Next;

        Status = AscendingCallback (WalkState, Parent);


        switch (Status)
        {
        case AE_CTRL_FALSE:

            /*
             * Either an IF/WHILE Predicate was false or we encountered a BREAK opcode
             *,In both cases, we do not execute the rest of the package;  We simply
             * close out the parent (finishing the walk of this branch of the tree)
             * and continue execution at the parent level.
             */

            Parent      = GrandParent;
            Next        = GrandParent->Next;
            GrandParent = GrandParent->Parent;

            Status = AscendingCallback (WalkState, Parent);

            /* Now continue to the next node in the tree */

            break;


        case AE_CTRL_TRUE:

            /* 
             * Predicate of a WHILE was true and the loop just completed an execution.
             * Go back to the start of the loop and reevaluate the predicate.
             */

            Op = WalkState->ControlState->Control.PredicateOp;      /* Points to the predicate */

            WalkState->ControlState->Common.State = CONTROL_PREDICATE_EXECUTING;

            WalkState->PrevOp       = Op->Parent;
            WalkState->NextOp       = Op;                   /* Evaluate the predicate again (next) */
            WalkState->NextOpInfo   = NEXT_OP_DOWNWARD;     /* Because we will traverse WHILE tree again */

            return_ACPI_STATUS (AE_OK);
            break;


        case AE_CTRL_TERMINATE:

            /* 
             * A control method was terminated via a RETURN statement.
             * The walk of this method is complete.
             */
            WalkState->PrevOp       = WalkState->Origin;
            WalkState->NextOp       = NULL;

            return_ACPI_STATUS (AE_OK);
            break;
        }


        /*
         * If we are back to the starting point, the walk is complete.
         */
        if (Parent == WalkState->Origin)
        {
            /* Reached the point of origin, the walk is complete */

            WalkState->PrevOp       = Parent;
            WalkState->NextOp       = NULL;

            return_ACPI_STATUS (Status);
        }


        /*
         * If there is a sibling to this parent (it is not the starting point Op),
         * then we will visit it.
         */
        if (Next)
        {
            /* found sibling of parent */

            WalkState->PrevOp       = Parent;
            WalkState->NextOp       = Next;
            WalkState->NextOpInfo   = NEXT_OP_DOWNWARD;

            return_ACPI_STATUS (Status);
        }

        /*
         * No sibling, check for an error from closing the parent 
         * (Also, AE_PENDING if a method call was encountered)
         */
        if (Status != AE_OK)
        {
            WalkState->PrevOp       = Parent;
            WalkState->NextOp       = GrandParent;
            WalkState->NextOpInfo   = NEXT_OP_UPWARD;
            
            return_ACPI_STATUS (Status);
        }

        /* No siblings, no errors, just move up one more level in the tree */

        Op                  = Parent;
        Parent              = GrandParent;
        WalkState->PrevOp   = Op;
    }


    /* Got all the way to the top of the tree, we must be done! */
    /* However, the code should have terminated in the loop above */

    WalkState->NextOp       = NULL;

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
PsWalkLoop (
    ACPI_WALK_LIST          *WalkList,
    ACPI_GENERIC_OP         *StartOp,
    INTERPRETER_CALLBACK    DescendingCallback,
    INTERPRETER_CALLBACK    AscendingCallback)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_WALK_STATE         *WalkState;
    ACPI_GENERIC_OP         *Op = StartOp;


    FUNCTION_TRACE_PTR ("PsWalkLoop", StartOp);


    WalkState = DsGetCurrentWalkState (WalkList);


    /* Walk entire subtree, visiting all nodes depth-first */

    while (Op)
    {
        if (WalkState->NextOpInfo != NEXT_OP_UPWARD)
        {
            Status = DescendingCallback (WalkState, Op);
        }

        /* 
         * A TRUE exception means that an ELSE was detected, but the IF predicate evaluated TRUE.
         */
        if (Status == AE_CTRL_TRUE)
        {
            /*
             * Ignore the entire ELSE block by moving on to the the next opcode.
             * And we do that by simply going up in the tree (either to the next sibling
             * or to the parent) from here.
             */
          
            WalkState->NextOpInfo = NEXT_OP_UPWARD;
        }

        /* Get the next node (op) in the depth-first walk */

        Status = PsGetNextWalkOp (WalkState, Op, AscendingCallback);

        /* A PENDING exception means that a control method invocation has been detected */

        if (Status == AE_CTRL_PENDING)
        {
            /* Transfer control to the called control method */

            Status = DsCallControlMethod (WalkList, WalkState, Op);

            /* If the method call worked, a new walk state was created -- get it */

            WalkState = DsGetCurrentWalkState (WalkList);
        }

        /* Abort the walk on any exception */

        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        Op = WalkState->NextOp;
    }

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
PsWalkParsedAml (
    ACPI_GENERIC_OP         *StartOp,
    ACPI_GENERIC_OP         *EndOp,
    ACPI_OBJECT_INTERNAL    *MthDesc,
    NAME_TABLE_ENTRY        *StartScope,
    ACPI_OBJECT_INTERNAL    **Params,
    ACPI_OBJECT_INTERNAL    **CallerReturnDesc,
    ACPI_OWNER_ID           OwnerId,
    INTERPRETER_CALLBACK    DescendingCallback,
    INTERPRETER_CALLBACK    AscendingCallback)
{
    ACPI_GENERIC_OP         *Op;
    ACPI_WALK_STATE         *WalkState;
    ACPI_OBJECT_INTERNAL    *ReturnDesc;
    ACPI_STATUS             Status;
    ACPI_WALK_LIST          WalkList;
    ACPI_WALK_LIST          *PrevWalkList;


    FUNCTION_TRACE_PTR ("PsWalkParsedAml", StartOp);


    /* Parameter Validation */

    if (!StartOp || !EndOp)
    {
        return AE_BAD_PARAMETER;
    }

    /* Initialize a new walk list */

    WalkList.WalkState = NULL;

    WalkState = DsCreateWalkState (EndOp, MthDesc, &WalkList);
    if (!WalkState)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }

    /* TBD: move to createWalkSTate */

    WalkState->OwnerId = OwnerId;

    /* TBD:
     * TEMP until we pass WalkState to the interpreter
     */
    PrevWalkList = Gbl_CurrentWalkList;
    Gbl_CurrentWalkList = &WalkList;

    if (StartScope)
    {
        /* Push start scope on scope stack and make it current  */

        Status = DsScopeStackPush (StartScope, ACPI_TYPE_Method, WalkState);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

    }

    if (MthDesc)
    {
        /* Init arguments if this is a control method */
        /* TBD: add walkstate as a param */

        DsMethodDataInitArgs (Params, MTH_NUM_ARGS);
    }

    Op = StartOp;
    Status = AE_OK;
    

    /*
     * Execute the walk loop as long as there is a valid Walk State.  This handles nested
     * control method invocations without recursion.
     */

    DEBUG_PRINT (TRACE_PARSE, ("PsWalkParsedAml: Op=%p EndOp=%p State=%p\n", Op, EndOp, WalkState));

    while (WalkState)
    {
        if (Status == AE_OK)
        {
            Status = PsWalkLoop (&WalkList, Op, DescendingCallback, AscendingCallback);
        }

        DEBUG_PRINT (TRACE_PARSE, ("PsWalkParsedAml: Completed one call to walk loop, State=%p\n", WalkState));

        /* We are done with this walk, move on to the parent if any */

        BREAKPOINT3;

        WalkState = DsPopWalkState (&WalkList);
        ReturnDesc = WalkState->ReturnDesc;     /* Extract return value before we delete WalkState */

        DEBUG_PRINT (TRACE_PARSE, ("PsWalkParsedAml: ReturnValue=%p, State=%p\n", WalkState->ReturnDesc, WalkState));
 
        /* Reset the current scope to the beginning of scope stack */

        DsScopeStackClear (WalkState);

        /* If we just returned from the execution of a control method, there's lots of cleanup to do */

        if (WalkState->MethodDesc &&
            WalkState->MethodDesc->Method.ParserOp)
        {
            DsTerminateControlMethod (WalkState);
        }

         /* Delete this walk state and all linked control states */

        DsDeleteWalkState (WalkState);

       /* Check if we have restarted a preempted walk */

        WalkState = DsGetCurrentWalkState (&WalkList);
        if (WalkState &&
            Status == AE_OK)
        {
            /* There is another walk state, restart it */

            /* 
             * If the method returned value is not used by the parent,
             * The object is deleted 
             */

            DsRestartControlMethod (WalkState, ReturnDesc);

            /* Get the next Op to process */

            Op = WalkState->NextOp;
        }

        /* Just completed a 1st-level method, save the final internal return value (if any) */

        else if (CallerReturnDesc)
        {
            *CallerReturnDesc = ReturnDesc;         /* Will be NULL if no return value */
        }

        else if (ReturnDesc)
        {
            CmRemoveReference (ReturnDesc);    /* Caller doesn't want it, must delete it */
        }
    }


    Gbl_CurrentWalkList = PrevWalkList;

    return_ACPI_STATUS (Status);
}


