/*
 *  ./acpica/subsystem/resources/rsby2lst.c - RsByteStreamToList function
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSBY2LST_C__

#include <acpi.h>
#include <resource.h>
#include <pnp.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rsby2lst");



ACPI_STATUS
RsByteStreamToList (
    UINT8                   *ByteStreamBuffer,
    UINT32                  ByteStreamBufferLength,
    UINT8                   **OutputBuffer)
{
    ACPI_STATUS             Status = AE_UNKNOWN_STATUS;
    UINT32                  BytesParsed = 0;
    UINT8                   ResourceType = 0;
    UINT32                  BytesConsumed = 0;
    UINT8                   **Buffer = OutputBuffer;
    UINT32                  StructureSize = 0;
    BOOLEAN                 EndTagProcessed = FALSE;
    

    FUNCTION_TRACE ("RsByteStreamToList");

    while (BytesParsed < ByteStreamBufferLength &&
            FALSE == EndTagProcessed)
    {
        /*
         * Look at the next byte in the stream
         */
        ResourceType = *ByteStreamBuffer;
        
        /*
         * See if this is a small or large resource
         */
        if(ResourceType & 0x80)
        {
            /*
             * Large Resource Type
             */
            switch (ResourceType)
            {
            case MEMORY_RANGE_24:
                /*
                 * 24-Bit Memory Resource
                 */
                Status = RsMemory24Resource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case LARGE_VENDOR_DEFINED:
                /*
                 * Vendor Defined Resource
                 */
                Status = RsVendorResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case MEMORY_RANGE_32:
                /*
                 * 32-Bit Memory Range Resource
                 */
                Status = RsMemory32RangeResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case FIXED_MEMORY_RANGE_32:
                /*
                 * 32-Bit Fixed Memory Resource
                 */
                Status = RsFixedMemory32Resource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case DWORD_ADDRESS_SPACE:
                /*
                 * 32-Bit Address Resource
                 */
                Status = RsAddress32Resource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;
            
            case WORD_ADDRESS_SPACE:
                /*
                 * 16-Bit Address Resource
                 */
                Status = RsAddress16Resource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;
            
            case EXTENDED_IRQ:
                /*
                 * Extended IRQ
                 */
                Status = RsExtendedIrqResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;
            
/* 64-bit not currently supported */
/*
            case 0x8A: 
                break;
*/
            
            default:
                /*
                 * If we get here, everything is out of sync,
                 *  so exit with an error
                 */
                return_ACPI_STATUS (AE_ERROR);
                break;
            }
        }

        else
        {
            /*
             * Small Resource Type
             *  Only bits 7:3 are valid
             */
            ResourceType >>= 3;

            switch(ResourceType)
            {
            case IRQ_FORMAT:
                /*
                 * IRQ Resource
                 */
                Status = RsIrqResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case DMA_FORMAT:
                /*
                 * DMA Resource
                 */
                Status = RsDmaResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case START_DEPENDENT_TAG:
                /*
                 * Start Dependent Functions Resource
                 */
                Status = RsStartDependentFunctionsResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case END_DEPENDENT_TAG:
                /*
                 * End Dependent Functions Resource
                 */
                Status = RsEndDependentFunctionsResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case IO_PORT_DESCRIPTOR:
                /*
                 * IO Port Resource
                 */
                Status = RsIoResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case FIXED_LOCATION_IO_DESCRIPTOR:
                /*
                 * Fixed IO Port Resource
                 */
                Status = RsFixedIoResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case SMALL_VENDOR_DEFINED:
                /*
                 * Vendor Specific Resource
                 */
                Status = RsVendorResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);

                break;

            case END_TAG:
                /*
                 * End Tag 
                 */
                Status = RsEndTagResource(ByteStreamBuffer, &BytesConsumed, Buffer, &StructureSize);
                EndTagProcessed = TRUE;

                break;

            default:
                /*
                 * If we get here, everything is out of sync,
                 *  so exit with an error
                 */
                return_ACPI_STATUS (AE_ERROR);
                break;

            } /* switch */
        }  /* if(ResourceType & 0x80) */

        /*
         * Update the return value and counter
         */
        BytesParsed += BytesConsumed;
        
        /*
         * Set the byte stream to point to the next resource
         */
        ByteStreamBuffer += BytesConsumed;

        /*
         * Set the Buffer to the next structure
         */
        *Buffer += StructureSize;

    } /*  while (BytesParsed < ByteStreamBufferLength && FALSE == EndTagProcessed)  */
    
    /* 
     * Check the reason for exiting the while loop
     */
    if (ByteStreamBufferLength != BytesParsed || TRUE != EndTagProcessed)
    {
        return_ACPI_STATUS (AE_ERROR);
    }

    return_ACPI_STATUS (AE_OK);

} /* RsByteStreamToList */

