/*
 *  ./acpica/subsystem/resources/rscalcb.c - RsCalculateByteStreamLength function
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSCALCB_C__

#include <acpi.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rscalcb");



ACPI_STATUS
RsCalculateByteStreamLength (
    RESOURCE                *LinkedList,
    UINT32                  *SizeNeeded)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  ByteStreamSizeNeeded = 0;
    UINT32                  SizeOfThisBit;
    EXTENDED_IRQ_RESOURCE   *ExIrq = NULL;
    BOOLEAN                 Done = FALSE;
    

    FUNCTION_TRACE ("RsCalculateByteStreamLength");

    while (!Done)
    {

        /*
         * Init the variable that will hold the size to add to the total.
         */
        SizeOfThisBit = 0;

        switch (LinkedList->Id)
        {
        case Irq:
            /*
             * IRQ Resource
             */
            /* 
             * For an IRQ Resource, Byte 3, although optional, will
             *  always be created - it holds IRQ information.
             */
            SizeOfThisBit = 4;
            break;

        case Dma:
            /*
             * DMA Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 3;
            break;

        case StartDependentFunctions:
            /*
             * Start Dependent Functions Resource
             */
            /* 
             * For a StartDependentFunctions Resource, Byte 1, 
             * although optional, will always be created.
             */
            SizeOfThisBit = 2;
            break;

        case EndDependentFunctions:
            /*
             * End Dependent Functions Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 1;
            break;
        
        case Io:
            /*
             * IO Port Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 8;
            break;
        
        case FixedIo:
            /*
             * Fixed IO Port Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 4;
            break;
        
        case VendorSpecific:
            /*
             * Vendor Defined Resource
             */
            /*
             * For a Vendor Specific resource, if the Length is between 1 and 7
             *  it will be created as a Small Resource data type, otherwise it
             *  is a Large Resource data type.
             */
            if(LinkedList->Data.VendorSpecific.Length > 7)
            {
                SizeOfThisBit = 3;
            }
            else
            {
                SizeOfThisBit = 1;
            }
            SizeOfThisBit += LinkedList->Data.VendorSpecific.Length;
            break;
        
        case EndTag:
            /*
             * End Tag 
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 2;
            Done = TRUE;
            break;
        
        case Memory24:
            /*
             * 24-Bit Memory Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 12;
            break;
        
        case Memory32:
            /*
             * 32-Bit Memory Range Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 20;
            break;
        
        case FixedMemory32:
            /*
             * 32-Bit Fixed Memory Resource
             */
            /*
             * For this resource the size is static
             */
            SizeOfThisBit = 12;
            break;
        
        case Address16:
            /*
             * 16-Bit Address Resource
             */
            /*
             * The base size of this byte stream is 16. If a Resource Source
             *  string is not NULL, add 1 for the Index + the length of the 
             *  null terminated string Resource Source + 1 for the null.
             */
            SizeOfThisBit = 16;
            
            if(NULL != LinkedList->Data.Address16.ResourceSource)
            {
                SizeOfThisBit += (1 + LinkedList->Data.Address16.ResourceSourceStringLength);
            }
            break;
        
        case Address32:
            /*
             * 32-Bit Address Resource
             */
            /*
             * The base size of this byte stream is 26. If a Resource Source
             *  string is not NULL, add 1 for the Index + the length of the 
             *  null terminated string Resource Source + 1 for the null.
             */
            SizeOfThisBit = 26;
            
            if(NULL != LinkedList->Data.Address16.ResourceSource)
            {
                SizeOfThisBit += (1 + LinkedList->Data.Address16.ResourceSourceStringLength);
            }
            break;

        case ExtendedIrq:
            /*
             * Extended IRQ Resource
             */
            /*
             * The base size of this byte stream is 9. This is for an Interrupt
             *  table length of 1.  For each additional interrupt, add 4.
             * If a Resource Source string is not NULL, add 1 for the Index 
             *  + the length of the null terminated string Resource Source 
             *  + 1 for the null.
             */
            SizeOfThisBit = 9;
            
            SizeOfThisBit += (LinkedList->Data.ExtendedIrq.NumberOfInterrupts - 1) * 4;

            if(NULL != ExIrq->ResourceSource)
            {
                SizeOfThisBit += (1 + LinkedList->Data.ExtendedIrq.ResourceSourceStringLength);
            }
            break;
        
        default:
            /*
             * If we get here, everything is out of sync,
             *  so exit with an error
             */
            return_ACPI_STATUS (AE_ERROR);
            break;

        } /* switch (LinkedList->Id) */

        /*
         * Update the total
         */
        ByteStreamSizeNeeded += SizeOfThisBit;

        /*
         * Point to the next object
         */
        LinkedList = (RESOURCE *) ((NATIVE_UINT) LinkedList + (NATIVE_UINT) LinkedList->Length);
    }
    
    /*
     * This is the data the caller needs
     */
    *SizeNeeded = ByteStreamSizeNeeded;
    
    return_ACPI_STATUS (Status);

} /* RsCalculateByteStreamLength */

