/*
 *  ./acpica/subsystem/resources/rscalcl.c - RsCalculateListLength function
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSCALCL_C__

#include <acpi.h>
#include <pnp.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rscalcl");



ACPI_STATUS
RsCalculateListLength (
    UINT8                   *ByteStreamBuffer,
    UINT32                  ByteStreamBufferLength,
    UINT32                  *SizeNeeded)
{
    UINT32                  BufferSize = 0;
    UINT32                  BytesParsed = 0;
    UINT8                   ResourceType = 0;
    UINT32                  StructureSize = 0;
    UINT32                  BytesConsumed = 0;
    UINT8                   *Buffer;
    UINT8                   NumberOfInterrupts = 0;
    UINT8                   Temp8;
    UINT16                  Temp16;
    UINT8                   Index;
    UINT8                   NumberOfChannels = 0;
    UINT8                   AdditionalBytes = 0;
    

    FUNCTION_TRACE ("RsCalculateListLength");

    while (BytesParsed < ByteStreamBufferLength)
    {
        /*
         * Look at the next byte in the stream
         */
        ResourceType = *ByteStreamBuffer;
        
        /*
         * See if this is a small or large resource
         */
        if(ResourceType & 0x80)
        {
            /*
             * Large Resource Type
             */
            switch (ResourceType)
            {
            case MEMORY_RANGE_24:
                /*
                 * 24-Bit Memory Resource
                 */
                BytesConsumed = 12;
    
                StructureSize = sizeof (MEMORY24_RESOURCE) + RESOURCE_LENGTH_NO_DATA;

                break;

            case LARGE_VENDOR_DEFINED:
                /*
                 * Vendor Defined Resource
                 */
                Buffer = ByteStreamBuffer;

                ++Buffer;

                Temp16 = *(UINT16 *)Buffer;

                BytesConsumed = Temp16 + 3;

                /*
                 * Ensure a 32-bit boundry for the structure
                 */ 
                if(0 != Temp16 % 4)
                {
                    Temp16 = (UINT16) (((Temp16 / 4) * 4) + 4);
                }

                StructureSize = sizeof (VENDOR_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA + 
                                (Temp16 * sizeof (UINT8));
                break;

            case MEMORY_RANGE_32:
                /*
                 * 32-Bit Memory Range Resource
                 */

                BytesConsumed = 20;
    
                StructureSize = sizeof (MEMORY32_RESOURCE) + RESOURCE_LENGTH_NO_DATA;
                break;

            case FIXED_MEMORY_RANGE_32:
                /*
                 * 32-Bit Fixed Memory Resource
                 */
                BytesConsumed = 12;
    
                StructureSize = sizeof(FIXED_MEMORY32_RESOURCE) + RESOURCE_LENGTH_NO_DATA;
                break;

            case DWORD_ADDRESS_SPACE:
                /*
                 * 32-Bit Address Resource
                 */
                Buffer = ByteStreamBuffer;

                ++Buffer;

                Temp16 = *(UINT16 *) Buffer;
        
                BytesConsumed = Temp16 + 3;

                /*
                 * Resource Source Index and Resource Source are optional
                 *  elements.  Check the length of the Bytestream.  If
                 *  it is greater than 23, that means that an Index exists
                 *  and is followed by a null termininated string.  Therefore,
                 *  set the temp variable to the length minus the minimum
                 *  byte stream length plus the byte for the Index to determine
                 *  the size of the NULL terminiated string.
                 */ 
                if (23 < Temp16)
                {
                    Temp8 = (UINT8) (Temp16 - 24);
                }
                else
                {
                    Temp8 = 0;
                }

                /*
                 * Ensure a 32-bit boundary for the structure
                 */ 
                if(0 != Temp8 % 4)
                {
                    Temp8 = (UINT8) (((Temp8 / 4) * 4) + 4);
                }

                StructureSize = sizeof (ADDRESS32_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA + 
                                (Temp8 * sizeof (UINT8));
                break;
            
            case WORD_ADDRESS_SPACE:
                /*
                 * 16-Bit Address Resource
                 */
                Buffer = ByteStreamBuffer;

                ++Buffer;

                Temp16 = *(UINT16 *)Buffer;
        
                BytesConsumed = Temp16 + 3;

                /*
                 * Resource Source Index and Resource Source are optional
                 *  elements.  Check the length of the Bytestream.  If
                 *  it is greater than 13, that means that an Index exists
                 *  and is followed by a null termininated string.  Therefore,
                 *  set the temp variable to the length minus the minimum
                 *  byte stream length plus the byte for the Index to determine
                 *  the size of the NULL terminiated string.
                 */ 
                if (13 < Temp16)
                {
                    Temp8 = (UINT8) (Temp16 - 14);
                }
                else
                {
                    Temp8 = 0;
                }

                /*
                 * Ensure a 32-bit boundry for the structure
                 */ 
                if(0 != Temp8 % 4)
                {
                    Temp8 = (UINT8) (((Temp8 / 4) * 4) + 4);
                }

                StructureSize = sizeof (ADDRESS16_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA + 
                                (Temp8 * sizeof (UINT8));
                break;
            
            case EXTENDED_IRQ:
                /*
                 * Extended IRQ
                 */
                Buffer = ByteStreamBuffer;

                ++Buffer;

                Temp16 = *(UINT16 *) Buffer;
        
                BytesConsumed = Temp16 + 3;

                /*
                 * Point past the length field and the
                 *  Interrupt vector flags to save off the
                 *  Interrupt table length to the Temp8 variable.
                 */
                Buffer += 3;

                Temp8 = *Buffer;

                /*
                 * To compensate for multiple interrupt numbers,
                 *  Add 4 bytes for each additional interrupts greater than 1
                 */
                AdditionalBytes = (UINT8) ((Temp8 - 1) * 4);

                /*
                 * Resource Source Index and Resource Source are optional
                 *  elements.  Check the length of the Bytestream.  If
                 *  it is greater than 9 plus the Additional bytes from 
                 *  the interrupt table, that means that an Index exists
                 *  and is followed by a null termininated string.  Therefore,
                 *  set the temp variable to the length minus the minimum
                 *  byte stream length plus the byte for the Index to determine
                 *  the size of the NULL terminiated string.
                 */ 
                if (9 + AdditionalBytes < Temp16)
                {
                    Temp8 = (UINT8) (Temp16 - (9 + AdditionalBytes));
                }

                else
                {
                    Temp8 = 0;
                }

                /*
                 * Ensure a 32-bit boundry for the structure
                 */ 
                if(0 != Temp8 % 4)
                {
                    Temp8 = (UINT8) (((Temp8 / 4) * 4) + 4);
                }

                StructureSize = sizeof (EXTENDED_IRQ_RESOURCE) + 
                                RESOURCE_LENGTH_NO_DATA + 
                                (AdditionalBytes * sizeof (UINT8)) + 
                                (Temp8 * sizeof (UINT8));
               
                break;
            
/* 64-bit not currently supported */
/*
            case 0x8A: 
                break;
*/
            
            default:
                /*
                 * If we get here, everything is out of sync,
                 *  so exit with an error
                 */
                return_ACPI_STATUS (AE_ERROR);
                break;
            }
        }

        else
        {
            /*
             * Small Resource Type
             *  Only bits 7:3 are valid
             */
            ResourceType >>= 3;

            switch(ResourceType)
            {
            case IRQ_FORMAT:
                /*
                 * IRQ Resource
                 */
                /*
                 * Determine if it there are two or three trailing bytes
                 */
                Buffer = ByteStreamBuffer;

                Temp8 = *Buffer;

                if(Temp8 & 0x01)
                {
                    BytesConsumed = 4;
                }
                else
                {
                    BytesConsumed = 3;
                }

                /*
                 * Point past the descriptor
                 */
                ++Buffer;

                /*
                 * Look at the number of bits set
                 */
                Temp16 = *(UINT16 *) Buffer;
    
                for(Index = 0; Index < 16; Index++)
                {
                    if(Temp16 & 0x1)
                    {
                        ++NumberOfInterrupts;
                    }

                    Temp16 >>= 1;
                }
    
                StructureSize = sizeof (IO_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA + 
                                (NumberOfInterrupts * sizeof (UINT32));

                break;

            case DMA_FORMAT:
                /*
                 * DMA Resource
                 */
                Buffer = ByteStreamBuffer;

                BytesConsumed = 3;

                /*
                 * Point past the descriptor
                 */
                ++Buffer;

                /*
                 * Look at the number of bits set
                 */
                Temp8 = *Buffer;

                for(Index = 0; Index < 8; Index++)
                {
                    if(Temp8 & 0x1)
                    {
                        ++NumberOfChannels;
                    }

                    Temp8 >>= 1;
                }

                StructureSize = sizeof (DMA_RESOURCE) + 
                                RESOURCE_LENGTH_NO_DATA +
                                (NumberOfChannels * sizeof (UINT32));

                break;

            case START_DEPENDENT_TAG:
                /*
                 * Start Dependent Functions Resource
                 */
                /*
                 * Determine if it there are two or three trailing bytes
                 */
                Buffer = ByteStreamBuffer;

                Temp8 = *Buffer;

                if(Temp8 & 0x01)
                {
                    BytesConsumed = 2;
                }
                else
                {
                    BytesConsumed = 1;
                }

    
                StructureSize = sizeof (START_DEPENDENT_FUNCTIONS_RESOURCE) + 
                                RESOURCE_LENGTH_NO_DATA;
                break;

            case END_DEPENDENT_TAG:
                /*
                 * End Dependent Functions Resource
                 */
                BytesConsumed = 1;
    
                StructureSize = RESOURCE_LENGTH;
                break;

            case IO_PORT_DESCRIPTOR:
                /*
                 * IO Port Resource
                 */
                BytesConsumed = 8;
    
                StructureSize = sizeof (IO_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA;
                break;

            case FIXED_LOCATION_IO_DESCRIPTOR:
                /*
                 * Fixed IO Port Resource
                 */
                BytesConsumed = 4;

                StructureSize = sizeof (FIXED_IO_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA;
                break;

            case SMALL_VENDOR_DEFINED:
                /*
                 * Vendor Specific Resource
                 */
                Buffer = ByteStreamBuffer;

                Temp8 = *Buffer;
                
                Temp8 = (UINT8) (Temp8 & 0x7);

                BytesConsumed = Temp8 + 1;

                /*
                 * Ensure a 32-bit boundry for the structure
                 */ 
                if(0 != Temp8 % 4)
                {
                    Temp8 = (UINT8) (((Temp8 / 4) * 4) + 4);
                }

                StructureSize = sizeof (VENDOR_RESOURCE) +
                                RESOURCE_LENGTH_NO_DATA + 
                                (Temp8 * sizeof (UINT8));
                break;

            case END_TAG:
                /*
                 * End Tag 
                 */
                BytesConsumed = 2;

                StructureSize = RESOURCE_LENGTH;
                break;

            default:
                /*
                 * If we get here, everything is out of sync,
                 *  so exit with an error
                 */
                return_ACPI_STATUS (AE_ERROR);
                break;

            } /* switch */

        }  /* if(ResourceType & 0x80) */

        /*
         * Update the return value and counter
         */
        BufferSize += StructureSize;
        BytesParsed += BytesConsumed;
        
        /*
         * Set the byte stream to point to the next resource
         */
        ByteStreamBuffer += BytesConsumed;

    }
    
    /*
     * This is the data the caller needs
     */
    *SizeNeeded = BufferSize;
    
    return_ACPI_STATUS (AE_OK);

} /* RsCalculateListLength */

