/*
 *  ./acpica/subsystem/resources/rscrprt.c - RsCreatePciRoutingTable function
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSCRPRT_C__

#include <acpi.h>
#include <resource.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rscrrpt");



ACPI_STATUS
RsCreatePciRoutingTable (
    ACPI_OBJECT_INTERNAL    *PackageObject,
    UINT8                   *OutputBuffer,
    UINT32                  *OutputBufferLength)
{
    UINT8                   *Buffer = OutputBuffer;
    ACPI_OBJECT_INTERNAL    **TopObjectList = NULL;
    ACPI_OBJECT_INTERNAL    **SubObjectList = NULL;
    ACPI_OBJECT_INTERNAL    *PackageElement = NULL;
    UINT32                  BufferSizeNeeded = 0;
    UINT32                  NumberOfElements = 0;
    UINT32                  Index = 0;
    UINT8                   TableIndex = 0;
    BOOLEAN                 NameFound = FALSE;
    PCI_ROUTING_TABLE       *UserPRT = NULL;
    

    FUNCTION_TRACE ("RsCreatePciRoutingTable");

    /*
     * Validate parameters:
     *  
     *  1. If MethodReturnObject is NULL, or
     *  2. If OutputBuffer is NULL and OutputBufferLength is not zero
     *  
     *  Return an error
     */
    if (!PackageObject ||
       (!OutputBuffer && 0 != *OutputBufferLength))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Calculate the buffer size needed for the routing table.
     */
    NumberOfElements = PackageObject->Package.Count;
    
    /*
     * Properly calculate the size of the return buffer.
     *  The base size is the number of elements * the sizes of the
     *  structures.  Additional space for the strings is added below.
     *  The minus one is to subtract the size of the UINT8 Source[1] 
     *  member because it is added below.
     * NOTE: The NumberOfElements is incremented by one to add an end table
     *  structure that is essentially a structure of zeros.
     */
    BufferSizeNeeded = (NumberOfElements + 1) * (sizeof (PCI_ROUTING_TABLE) - 1);

    /*
     * But each PRT_ENTRY structure has a pointer to a string and
     *  the size of that string must be found.
     */
    TopObjectList = PackageObject->Package.Elements;

    for (Index = 0; Index < NumberOfElements; Index++)
    {
        /*
         * Dereference the sub-package
         */
        PackageElement = *TopObjectList;

        /*
         * The SubObjectList will now point to an array of the
         *  four IRQ elements: Address, Pin, Source and SourceIndex
         */
        SubObjectList = PackageElement->Package.Elements;

        /*
         * Scan the IrqTableElements for the Source Name String
         */
        NameFound = FALSE;

        for (TableIndex = 0; TableIndex < 4 && !NameFound; TableIndex++)
        {
            if (ACPI_TYPE_String == (*SubObjectList)->Common.Type)
            {
                NameFound = TRUE;
            }

            else
            {
                /*
                 * Look at the next element
                 */
                SubObjectList++;
            }
        }

        /*
         * Was a String type found?
         */
        if (TRUE == NameFound)
        {
            /*
             * The length String.Length field includes the terminating NULL
             */
            BufferSizeNeeded += ((*SubObjectList)->String.Length);
        }

        else
        {
            /*
             * If no name was found, then this is a NULL, which is translated
             *  as a UINT32 zero.
             */
            BufferSizeNeeded += sizeof(UINT32);
        }

        /*
         * Point to the next ACPI_OBJECT_INTERNAL
         */
        TopObjectList++;
    }

    DEBUG_PRINT (VERBOSE_INFO, ("RsCreatePciRoutingTable: BufferSizeNeeded = %d\n", 
                                    BufferSizeNeeded));

    /*
     * If the data will fit into the available buffer
     *  call to fill in the list
     */
    if (BufferSizeNeeded <= *OutputBufferLength)
    {
        /*
         * Zero out the return buffer before proceeding
         */
        MEMSET (OutputBuffer, 0x00, *OutputBufferLength);

        /*
         * Loop through the ACPI_INTERNAL_OBJECTS - Each object should
         *  contain a UINT32 Address, a UINT8 Pin, a Name and a UINT8 SourceIndex.
         */
        TopObjectList = PackageObject->Package.Elements;

        NumberOfElements = PackageObject->Package.Count;
    
        UserPRT = (PCI_ROUTING_TABLE *)Buffer;

        for (Index = 0; Index < NumberOfElements; Index++)
        {
            /*
             * Point UserPrt past this current structure
             *
             * NOTE: On the first iteration, UserPrt->Length will be zero
             *  because we zero'ed out the return buffer earlier
             */
            Buffer += UserPRT->Length;

            UserPRT = (PCI_ROUTING_TABLE *)Buffer;

            /*
             * Fill in the Length field with the information we
             *  have at this point.
             *  The minus one is to subtract the size of the UINT8 Source[1] 
             *  member because it is added below.
             */
            UserPRT->Length = (sizeof(PCI_ROUTING_TABLE) - 1);

            /*
             * Dereference the sub-package
             */
            PackageElement = *TopObjectList;

            /*
             * The SubObjectList will now point to an array of the
             *  four IRQ elements: Address, Pin, Source and SourceIndex
             */
            SubObjectList = PackageElement->Package.Elements;
            
            /*
             * Dereference the Address
             */
            if (ACPI_TYPE_Number == (*SubObjectList)->Common.Type)
            {
                UserPRT->Data.Address = (*SubObjectList)->Number.Value;
            }

            else
            {
                return_ACPI_STATUS (AE_BAD_DATA);
            }

            /*
             * Dereference the Pin
             */
            SubObjectList++;

            if (ACPI_TYPE_Number == (*SubObjectList)->Common.Type)
            {
                UserPRT->Data.Pin = (*SubObjectList)->Number.Value;
            }

            else
            {
                return_ACPI_STATUS (AE_BAD_DATA);
            }

            /*
             * Dereference the Source Name
             */
            SubObjectList++;

            if (ACPI_TYPE_String == (*SubObjectList)->Common.Type)
            {
                STRCPY(UserPRT->Data.Source, (*SubObjectList)->String.Pointer);

                /*
                 * Add to the Length field the length of the string
                 */
                UserPRT->Length += (*SubObjectList)->String.Length;
            }

            else
            {
                /*
                 * If this is a number, then the Source Name is NULL, since
                 *  the entire buffer was zeroed out, we can leave this alone.
                 */
                if (ACPI_TYPE_Number == (*SubObjectList)->Common.Type)
                {
                    /*
                     * Add to the Length field the length of the UINT32 NULL
                     */
                    UserPRT->Length += sizeof(UINT32);
                }

                else
                {
                    return_ACPI_STATUS (AE_BAD_DATA);
                }
            }

            /*
             * Dereference the Source Index
             */
            SubObjectList++;

            if (ACPI_TYPE_Number == (*SubObjectList)->Common.Type)
            {
                UserPRT->Data.SourceIndex = (*SubObjectList)->Number.Value;
            }

            else
            {
                return_ACPI_STATUS (AE_BAD_DATA);
            }

            /*
             * Point to the next ACPI_OBJECT_INTERNAL
             */
            TopObjectList++;
        }

        DEBUG_PRINT (VERBOSE_INFO, ("RsCreatePciRoutingTable: OutputBuffer = %p\n", OutputBuffer));
    }

    else
    {
        *OutputBufferLength = BufferSizeNeeded;
        
        return_ACPI_STATUS (AE_BUFFER_OVERFLOW);
    }

    /*
     * Report the amount of buffer used
     */
    *OutputBufferLength = BufferSizeNeeded;

    return_ACPI_STATUS (AE_OK);

}    /*    RsCreatePciRoutingTable    */

