/*
 *  ./acpica/subsystem/resources/rseirq.c - RsExtendedIrqResource, 
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSEIRQ_C__

#include <acpi.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rseirq");



ACPI_STATUS
RsExtendedIrqResource (
    UINT8                   *ByteStreamBuffer,
    UINT32                  *BytesConsumed,
    UINT8                   **OutputBuffer,
    UINT32                  *StructureSize)
{
    UINT8                   *Buffer = ByteStreamBuffer;
    RESOURCE                *OutputStruct = (RESOURCE *) * OutputBuffer;
    UINT16                  Temp16 = 0;
    UINT8                   Temp8 = 0;        
    UINT8                   Index;
    UINT32                  StructSize = sizeof (EXTENDED_IRQ_RESOURCE) + RESOURCE_LENGTH_NO_DATA;
    

    FUNCTION_TRACE ("RsExtendedIrqResource");

    /*
     * Point past the Descriptor to get the number of bytes consumed
     */ 
    Buffer += 1;
    Temp16 = *(UINT16 *)Buffer;

    *BytesConsumed = Temp16 + 3;
    OutputStruct->Id = ExtendedIrq;

    /*
     * Point to the Byte3
     */ 
    Buffer += 2;
    Temp8 = *Buffer;

    OutputStruct->Data.ExtendedIrq.ProducerConsumer = Temp8 & 0x01;

    /*
     * Check for HE, LL or HL
     */
    if(Temp8 & 0x02)
    {
        OutputStruct->Data.ExtendedIrq.EdgeLevel = EDGE_SENSITIVE;
        OutputStruct->Data.ExtendedIrq.ActiveHighLow = ACTIVE_HIGH;
    }

    else
    {
        if(Temp8 & 0x4)
        {
            OutputStruct->Data.ExtendedIrq.EdgeLevel = LEVEL_SENSITIVE;
            OutputStruct->Data.ExtendedIrq.ActiveHighLow = ACTIVE_LOW;
        }

        else
        {
            /*
             * Only _LL and _HE polarity/trigger interrupts 
             *  are allowed (ACPI spec v1.0b ection 6.4.2.1),
             *  so an error will occur if we reach this point
             */
            return_ACPI_STATUS (AE_BAD_DATA);
        }
    }

    /*
     * Check for sharable
     */
    OutputStruct->Data.ExtendedIrq.SharedExclusive = (Temp8 >> 3) & 0x01;
    
    /*
     * Point to Byte4 (IRQ Table length)
     */ 
    Buffer += 1;
    Temp8 = *Buffer;

    OutputStruct->Data.ExtendedIrq.NumberOfInterrupts = Temp8;

    /*
     * Add any additional structure size to properly calculate
     *  the next pointer at the end of this function
     */
     StructSize += (Temp8 - 1) * 4;

    /*
     * Point to Byte5 (First IRQ Number)
     */
    Buffer += 1;

    /*
     * Cycle through every IRQ in the table
     */
    for (Index = 0; Index < Temp8; Index++)
    {
        OutputStruct->Data.ExtendedIrq.Interrupts[Index] = (UINT32)*Buffer;
        
        /* Point to the next IRQ */

        Buffer += 4;
    }
    
    /*
     * This will leave us pointing to the Resource Source Index
     *  If it is present, then save it off and calculate the
     *  pointer to where the null terminated string goes:
     *  Each Interrupt takes 32-bits + the 5 bytes of the 
     *  stream that are default.
     */
    if (*BytesConsumed > (UINT32)(OutputStruct->Data.ExtendedIrq.NumberOfInterrupts * 4) + 5)
    {
        /* Dereference the Index */

        Temp8 = *Buffer;
        OutputStruct->Data.ExtendedIrq.ResourceSourceIndex = (UINT32)Temp8;
        
        /* Point to the String */

        Buffer += 1;
        
        /* Copy the string into the buffer */

        Index = 0;

        while (0x00 != *Buffer)
        {
            OutputStruct->Data.ExtendedIrq.ResourceSource[Index] = *Buffer;
            Buffer += 1;
            Index += 1;
        }

        /*
         * Add the terminating null
         */
        OutputStruct->Data.ExtendedIrq.ResourceSource[Index] = 0x00;
        OutputStruct->Data.ExtendedIrq.ResourceSourceStringLength = Index + 1;

        /*
         * In order for the StructSize to fall on a 32-bit boundry,
         *  calculate the length of the string and expand the 
         *  StructSize to the next 32-bit boundry.
         */
        Temp8 = (UINT8) (Index + 1);

        if(0 == Temp8 % 4)
        {
            StructSize += Temp8;
        }

        else
        {
            StructSize += ((Temp8 / 4) * 4) + 4;
        }
    }

    else
    {
        OutputStruct->Data.ExtendedIrq.ResourceSourceIndex = 0x00;
        OutputStruct->Data.ExtendedIrq.ResourceSourceStringLength = 0;
        OutputStruct->Data.ExtendedIrq.ResourceSource[0] = 0x00;
    }

    /*
     * Set the Length parameter
     */
    OutputStruct->Length = StructSize;

    /*
     * Return the final size of the structure
     */
    *StructureSize = StructSize;

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
RsExtendedIrqStream (
    RESOURCE                *LinkedList,
    UINT8                   **OutputBuffer,
    UINT32                  *BytesConsumed)
{
    UINT8                   *Buffer = *OutputBuffer;
    UINT16                  *LengthField;
    UINT32                  Temp32 = 0;
    UINT8                   Temp8 = 0;
    UINT8                   Index;
    UINT8                   *TempPointer = NULL;
    

    FUNCTION_TRACE ("RsExtendedIrqStream");

    /*
     * The descriptor field is static
     */ 
    *Buffer = 0x89;
    Buffer += 1;

    /*
     * Set a pointer to the Length field - to be filled in later
     */

    LengthField = (UINT16 *)Buffer;
    Buffer += 2;

    /*
     * Set the Interrupt vector flags
     */
    Temp8 = (UINT8) (LinkedList->Data.ExtendedIrq.ProducerConsumer & 0x01);

    Temp8 |= ((LinkedList->Data.ExtendedIrq.SharedExclusive & 0x01) << 3);

    if (LEVEL_SENSITIVE == LinkedList->Data.ExtendedIrq.EdgeLevel &&
       ACTIVE_LOW == LinkedList->Data.ExtendedIrq.ActiveHighLow)
    {
        Temp8 |= 0x04;
    }
    else
    {
        Temp8 |= 0x02;
    }

    *Buffer = Temp8;
    Buffer += 1;
    
    /*
     * Set the Interrupt table length
     */
    Temp8 = (UINT8) LinkedList->Data.ExtendedIrq.NumberOfInterrupts;

    *Buffer = Temp8;
    
    Buffer += 1;

    for (Index = 0; Index < LinkedList->Data.ExtendedIrq.NumberOfInterrupts; Index++)
    {
        Temp32 = LinkedList->Data.ExtendedIrq.Interrupts[Index];
        
        *(UINT32 *) Buffer = Temp32;
        Buffer += 4;
    }

    /*
     * Resource Source Index and Resource Source are optional
     */
    if (0 != LinkedList->Data.ExtendedIrq.ResourceSourceStringLength)
    {
        Temp8 = (UINT8) LinkedList->Data.ExtendedIrq.ResourceSourceIndex;

        *Buffer = Temp8;
        Buffer += 1;
        TempPointer = Buffer;
        
        /*
         * Copy the string
         */
        STRCPY (TempPointer, LinkedList->Data.ExtendedIrq.ResourceSource);

        /*
         * Buffer needs to be set to the length of the sting + one for the 
         *  terminating null
         */
        Buffer += (STRLEN (LinkedList->Data.ExtendedIrq.ResourceSource) + 1);
    }
    
    /*
     * Return the number of bytes consumed in this operation
     */
    *BytesConsumed = (UINT32) ((NATIVE_UINT) Buffer - (NATIVE_UINT) *OutputBuffer);

    /*
     * Set the length field to the number of bytes consumed
     *  minus the header size (3 bytes)
     */
    *LengthField = (UINT16) (*BytesConsumed - 3);

    return_ACPI_STATUS (AE_OK);
}

