/*
 *  ./acpica/subsystem/resources/rsirq.c - RsIrqResource,
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __RSIRQ_C__

#include <acpi.h>

#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rsirq");



ACPI_STATUS
RsIrqResource (
    UINT8                   *ByteStreamBuffer,
    UINT32                  *BytesConsumed,
    UINT8                   **OutputBuffer,
    UINT32                  *StructureSize)
{
    UINT8                   *Buffer = ByteStreamBuffer;
    RESOURCE                *OutputStruct = (RESOURCE *) * OutputBuffer;
    UINT16                  Temp16 = 0;
    UINT8                   Temp8 = 0;        
    UINT8                   Index;
    UINT32                  StructSize = sizeof (IRQ_RESOURCE) + RESOURCE_LENGTH_NO_DATA;
    

    FUNCTION_TRACE ("RsIrqResource");

    /*
     * The number of bytes consumed are contained in the descriptor (Bits:0-1)
     */ 
    Temp8 = *Buffer;

    *BytesConsumed = (Temp8 & 0x03) + 1;

    OutputStruct->Id = Irq;

    /*
     * Point to the 16-bits of Bytes 1 and 2
     */ 
    Buffer += 1;
    Temp16 = *(UINT16 *)Buffer;

    OutputStruct->Data.Irq.NumberOfInterrupts = 0;

    /* Decode the IRQ bits */
    for (Index = 0; Index < 16; Index++)
    {
        if((Temp16 >> Index) & 0x01)
        {
            OutputStruct->Data.Irq.Interrupts[OutputStruct->Data.Irq.NumberOfInterrupts] = Index;
            OutputStruct->Data.Irq.NumberOfInterrupts += 1;
        }
    }

    /*
     * Calculate the structure size based upon the number of interrupts
     */
    StructSize += (OutputStruct->Data.Irq.NumberOfInterrupts - 1) * 4;

    /*
     * Point to Byte 3 if it is used
     */
    if (4 == *BytesConsumed)
    {
        Buffer += 2;

        Temp8 = *Buffer;

        /*
         * Check for HE, LL or HL
         */
        if (Temp8 & 0x01)
        {
            OutputStruct->Data.Irq.EdgeLevel = EDGE_SENSITIVE;
            OutputStruct->Data.Irq.ActiveHighLow = ACTIVE_HIGH;
        }

        else
        {
            if (Temp8 & 0x8)
            {
                OutputStruct->Data.Irq.EdgeLevel = LEVEL_SENSITIVE;
                OutputStruct->Data.Irq.ActiveHighLow = ACTIVE_LOW;
            }

            else
            {
                /*
                 * Only _LL and _HE polarity/trigger interrupts 
                 *  are allowed (ACPI spec v1.0b ection 6.4.2.1),
                 *  so an error will occur if we reach this point
                 */
                return_ACPI_STATUS (AE_BAD_DATA);
            }
        }

        /*
         * Check for sharable
         */
        OutputStruct->Data.Irq.SharedExclusive = (Temp8 >> 3) & 0x01;
    }

    else
    {
        /*
         * Assume Edge Sensitive, Active High, Non-Sharable
         *  per ACPI Specification
         */
        OutputStruct->Data.Irq.EdgeLevel = EDGE_SENSITIVE;
        OutputStruct->Data.Irq.ActiveHighLow = ACTIVE_HIGH;
        OutputStruct->Data.Irq.SharedExclusive = EXCLUSIVE;
    }
   
    /*
     * Set the Length parameter
     */
    OutputStruct->Length = StructSize;

    /*
     * Return the final size of the structure
     */
    *StructureSize = StructSize;

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
RsIrqStream (
    RESOURCE                *LinkedList,
    UINT8                   **OutputBuffer,
    UINT32                  *BytesConsumed)
{
    UINT8                   *Buffer = *OutputBuffer;
    UINT16                  Temp16 = 0;
    UINT8                   Temp8 = 0;        
    UINT8                   Index;
    BOOLEAN                 IRQInfoByteNeeded;
   
    
    FUNCTION_TRACE ("RsIrqStream");

    /*
     * The descriptor field is set based upon whether a third byte is needed
     *  to contain the IRQ Information.
     */ 
    if (EDGE_SENSITIVE == LinkedList->Data.Irq.EdgeLevel &&
        ACTIVE_HIGH == LinkedList->Data.Irq.ActiveHighLow &&
        EXCLUSIVE == LinkedList->Data.Irq.SharedExclusive)
    {
        *Buffer = 0x22;
        IRQInfoByteNeeded = FALSE;
    }
    else
    {
        *Buffer = 0x23;
        IRQInfoByteNeeded = TRUE;
    }

    Buffer += 1;

    Temp16 = 0;

    /*
     * Loop through all of the interrupts and set the mask bits
     */
    for(Index = 0; Index < LinkedList->Data.Irq.NumberOfInterrupts; Index++)
    {
        Temp8 = (UINT8) LinkedList->Data.Irq.Interrupts[Index];
        Temp16 |= 0x1 << Temp8;
    }

    *(UINT16 *)Buffer = Temp16;

    Buffer += 2;

    /*
     * Set the IRQ Info byte if needed.
     */
    if (IRQInfoByteNeeded)
    {
        Temp8 = 0;

        Temp8 = (UINT8) ((LinkedList->Data.Irq.SharedExclusive & 0x01) << 4);

        if (LEVEL_SENSITIVE == LinkedList->Data.Irq.EdgeLevel &&
            ACTIVE_LOW == LinkedList->Data.Irq.ActiveHighLow)
        {
            Temp8 |= 0x08;
        }

        else
        {
            Temp8 |= 0x01;
        }

        *Buffer = Temp8;

        Buffer += 1;
    }
    
    /*
     * Return the number of bytes consumed in this operation
     */
    *BytesConsumed = (UINT32) ((NATIVE_UINT) Buffer - (NATIVE_UINT) *OutputBuffer);

    return_ACPI_STATUS (AE_OK);
}

