
/*
 *  ./acpica/subsystem/resources/rsutils.c - Utilities for the resource manager
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __RSUTILS_C__

#include <acpi.h>
#include <namesp.h>
#include <resource.h>


#define _COMPONENT          RESOURCE_MANAGER
        MODULE_NAME         ("rsutils");




ACPI_STATUS
RsGetPRTMethodData (
    ACPI_HANDLE             Handle, 
    ACPI_BUFFER             *RetBuffer)
{
    ACPI_OBJECT_INTERNAL    *RetObj;
    ACPI_STATUS             Status;
    UINT32                  BufferSpaceNeeded = RetBuffer->Length;


    FUNCTION_TRACE ("RsGetPRTMethodData");
    

    /*
     *  Must have a valid handle and buffer, So we have to have a handle
     *  a return buffer structure and if there is a non-zero buffer length
     *  we also need a valid pointer in the buffer
     */
    if ((!Handle)               ||
        (!RetBuffer)            ||
        ((!RetBuffer->Pointer) && (RetBuffer->Length))) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     *  Execute the method, no parameters
     */
    Status = NsEvaluateRelative (Handle, "_PRT", NULL, &RetObj);
    if (Status != AE_OK) 
    {
        return_ACPI_STATUS (Status);
    }

    if (!RetObj)
    {
        /* Return object is required */

        DEBUG_PRINT (ACPI_ERROR, ("No object was returned from _PRT\n"));
        return_ACPI_STATUS (AE_TYPE);
    }


    /*
     * The return object will be a package, so check the
     *  parameters.  If the return object is not a package,
     *  then the underlying AML code is corrupt or improperly
     *  written.
     */
    if (ACPI_TYPE_Package != RetObj->Common.Type)
    {
        Status = AE_AML_OPERAND_TYPE;
        goto Cleanup;
    }
    
    /*
     * Make the call to create a resource linked list from the
     *  byte stream buffer that comes back from the _CRS method
     *  execution.
     */
    Status = RsCreatePciRoutingTable (RetObj,
                                      RetBuffer->Pointer,
                                      &BufferSpaceNeeded);

    /*
     * Tell the user how much of the buffer we have used or is needed
     *  and return the final status.
     */
    RetBuffer->Length = BufferSpaceNeeded;


    /* On exit, we must delete the object returned by evaluateObject */

Cleanup:

    CmRemoveReference (RetObj);

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
RsGetCRSMethodData (
    ACPI_HANDLE             Handle, 
    ACPI_BUFFER             *RetBuffer)
{
    ACPI_OBJECT_INTERNAL    *RetObj;
    ACPI_STATUS             Status;
    UINT32                  BufferSpaceNeeded = RetBuffer->Length;


    FUNCTION_TRACE ("RsGetCRSMethodData");
  
    
    /*
     *  Must have a valid handle and buffer, So we have to have a handle
     *  a return buffer structure and if there is a non-zero buffer length
     *  we also need a valid pointer in the buffer
     */
    if ((!Handle)               ||
        (!RetBuffer)            ||
        ((!RetBuffer->Pointer) && (RetBuffer->Length))) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     *  Execute the method, no parameters
     */
    Status = NsEvaluateRelative (Handle, "_CRS", NULL, &RetObj);
    if (Status != AE_OK) 
    {
        return_ACPI_STATUS (Status);
    }

    if (!RetObj)
    {
        /* Return object is required */

        DEBUG_PRINT (ACPI_ERROR, ("No object was returned from _CRS\n"));
        return_ACPI_STATUS (AE_TYPE);
    }

    /*
     * The return object will be a buffer, but check the
     *  parameters.  If the return object is not a buffer,
     *  then the underlying AML code is corrupt or improperly
     *  written.
     */
    if (ACPI_TYPE_Buffer != RetObj->Common.Type)
    {
        Status = AE_AML_OPERAND_TYPE;
        goto Cleanup;
    }
    
    /*
     * Make the call to create a resource linked list from the
     *  byte stream buffer that comes back from the _CRS method
     *  execution.
     */
    Status = RsCreateResourceList (RetObj,
                                   RetBuffer->Pointer,
                                   &BufferSpaceNeeded);

    if (AE_OK == Status)
    {
        RsDumpResourceList((RESOURCE *)RetBuffer->Pointer);
    }

    /*
     * Tell the user how much of the buffer we have used or is needed
     *  and return the final status.
     */
    RetBuffer->Length = BufferSpaceNeeded;


    /* On exit, we must delete the object returned by evaluateObject */

Cleanup:

    CmRemoveReference (RetObj);

    return_ACPI_STATUS (Status);   
}



ACPI_STATUS
RsGetPRSMethodData (
    ACPI_HANDLE             Handle, 
    ACPI_BUFFER             *RetBuffer)
{
    ACPI_OBJECT_INTERNAL    *RetObj;
    ACPI_STATUS             Status;
    UINT32                  BufferSpaceNeeded = RetBuffer->Length;


    FUNCTION_TRACE ("RsGetPRSMethodData");
  
    
    /*
     *  Must have a valid handle and buffer, So we have to have a handle
     *  a return buffer structure and if there is a non-zero buffer length
     *  we also need a valid pointer in the buffer
     */
    if ((!Handle)               ||
        (!RetBuffer)            ||
        ((!RetBuffer->Pointer) && (RetBuffer->Length))) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     *  Execute the method, no parameters
     */
    Status = NsEvaluateRelative (Handle, "_PRS", NULL, &RetObj);
    if (Status != AE_OK) 
    {
        return_ACPI_STATUS (Status);
    }

    if (!RetObj)
    {
        /* Return object is required */

        DEBUG_PRINT (ACPI_ERROR, ("No object was returned from _PRS\n"));
        return_ACPI_STATUS (AE_TYPE);
    }

    /*
     * The return object will be a buffer, but check the
     *  parameters.  If the return object is not a buffer,
     *  then the underlying AML code is corrupt or improperly
     *  written..
     */
    if (ACPI_TYPE_Buffer != RetObj->Common.Type)
    {
        Status = AE_AML_OPERAND_TYPE;
        goto Cleanup;
    }
    
    /*
     * Make the call to create a resource linked list from the
     *  byte stream buffer that comes back from the _CRS method
     *  execution.
     */
    Status = RsCreateResourceList (RetObj,
                                   RetBuffer->Pointer,
                                   &BufferSpaceNeeded);

    /*
     * Tell the user how much of the buffer we have used or is needed
     *  and return the final status.
     */
    RetBuffer->Length = BufferSpaceNeeded;


    /* On exit, we must delete the object returned by evaluateObject */

Cleanup:

    CmRemoveReference (RetObj);

    return_ACPI_STATUS (Status); 
}



ACPI_STATUS
RsSetSRSMethodData (
    ACPI_HANDLE             Handle, 
    ACPI_BUFFER             *InBuffer)
{
    ACPI_OBJECT_INTERNAL    *Params[2];
    ACPI_OBJECT_INTERNAL    ParamObj;
    ACPI_STATUS             Status;
    UINT8                   *ByteStream = NULL;
    UINT32                  BufferSizeNeeded = 0;                 

    FUNCTION_TRACE ("RsSetSRSMethodData");
    
    /*
     *  Must have a valid handle and buffer
     */
    if ((!Handle)       ||
        (!InBuffer)           ||
        (!InBuffer->Pointer)  ||
        (!InBuffer->Length)) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * The InBuffer parameter will point to a linked list of
     *  resource parameters.  It needs to be formatted into a 
     *  byte stream to be sent in as an input parameter.
     */
    BufferSizeNeeded = 0;

    /*
     * First call is to get the buffer size needed
     */
    Status = RsCreateByteStream (InBuffer->Pointer,
                                 ByteStream,
                                 &BufferSizeNeeded);

    /*
     * We expect a return of AE_BUFFER_OVERFLOW
     *  if not, exit with the error
     */
    if (AE_BUFFER_OVERFLOW != Status)
    {
        return_ACPI_STATUS (Status);
    }

    /*
     * Allocate the buffer needed
     */
    ByteStream = CmCallocate(BufferSizeNeeded);

    if (NULL == ByteStream)
    {
        return_ACPI_STATUS (AE_NO_MEMORY);
    }
    
    /*
     * Now call to convert the linked list into a byte stream
     */
    Status = RsCreateByteStream (InBuffer->Pointer,
                                 ByteStream,
                                 &BufferSizeNeeded);

    if(AE_OK != Status)
    {
        /*
         *  Failed the call
         */
        CmFree (ByteStream);
        return_ACPI_STATUS (Status);
    }

    /*
     *  Init the param object
     */
    CmInitStaticObject (&ParamObj);

    /*
     *  Method requires one parameter.  Set it up
     */
    Params [0] = &ParamObj;
    Params [1] = NULL;

    /*
     *  Set up the parameter object
     */
    ParamObj.Common.Type    = ACPI_TYPE_Buffer;
    ParamObj.Buffer.Length  = BufferSizeNeeded;
    ParamObj.Buffer.Pointer = ByteStream;

    /*
     *  Execute the method, no return value
     */
    Status = NsEvaluateRelative (Handle, "_SRS", Params, NULL);

    /*
     *  Clean up and return the status from NsEvaluateRelative
     */
    CmFree (ByteStream);
    return_ACPI_STATUS (Status);
}

