
/*
 *  ./acpica/subsystem/tables/tbapi.c - Public interfaces to the ACPI subsystem
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __TBAPI_C__

#include <acpi.h>
#include <acobject.h>
#include <interp.h>
#include <tables.h>
#include <methods.h>
#include <pnp.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbapi");



ACPI_STATUS
AcpiLoadFirmwareTables (void)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  NumberOfTables = 0; 


    FUNCTION_TRACE ("AcpiLoadFirmwareTables");


    /* Get the RSDT first */

    Status = TbGetTableRsdt (&NumberOfTables);
    if (Status != AE_OK)
    {
        goto ErrorExit;
    }



    /* Now get the rest of the tables */

    Status = TbGetAllTables (NumberOfTables, NULL);
    if (Status != AE_OK)
    {
        goto ErrorExit;
    }


    DEBUG_PRINT (ACPI_OK, ("ACPI Tables successfully loaded\n"));

    return_ACPI_STATUS (AE_OK);


ErrorExit:    
    DEBUG_PRINT (ACPI_ERROR, ("Failure during ACPI Table Init: %s\n", CmFormatException (Status)));
    
    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiLoadTable (
    ACPI_TABLE_HEADER       *TablePtr)
{
    ACPI_STATUS             Status;
    ACPI_TABLE_DESC         TableInfo;


    FUNCTION_TRACE ("AcpiLoadTable");

    if (!TablePtr)
    {
        return AE_BAD_PARAMETER;
    }

    /* Copy the table to a local buffer */

    Status = TbGetTable (NULL, ((char *) TablePtr), &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }
    
    /* Install the new table into the local data structures */

    Status = TbInstallTable (NULL, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        /* TBD: must free table allocated by TbGetTable */
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiUnloadTable (
    ACPI_TABLE_TYPE         TableType)
{
    ACPI_TABLE_DESC         *ListHead;


    FUNCTION_TRACE ("AcpiUnloadTable");


    /* Parameter validation */

    if (TableType > ACPI_TABLE_MAX)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    
    /* Find all tables of the requested type */

    ListHead = &Gbl_AcpiTables[TableType];
    do
    {            
        /* Delete the entire namespace under this table NTE */

        NsDeleteNamespaceByOwner (ListHead->TableId);

        /* Delete (or unmap) the actual table */

        TbDeleteAcpiTable (TableType);  

    } while (ListHead != &Gbl_AcpiTables[TableType]);

    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
AcpiGetTableHeader (
    ACPI_TABLE_TYPE         TableType, 
    UINT32                  Instance,
    ACPI_TABLE_HEADER       *OutTableHeader)
{
    ACPI_TABLE_HEADER       *TblPtr;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AcpiGetTableHeader");

    Status = AE_OK;

    if ((Instance == 0)                 ||
        (TableType == TABLE_RSDP)       ||
        (!OutTableHeader))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Check the table type and instance */

    if ((TableType > ACPI_TABLE_MAX)    ||
        (Gbl_AcpiTableData[TableType].Flags == ACPI_TABLE_SINGLE &&
         Instance > 1))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    

    /* Get a pointer to the entire table */

    Status = TbGetTablePtr (TableType, Instance, &TblPtr);
    if (Status != AE_OK)
    {
        return_ACPI_STATUS (Status);
    }

    /*
     * The function will return a NULL pointer if the table is not loaded
     */
    if (TblPtr == NULL)
    {
        return_ACPI_STATUS (AE_NOT_EXIST);
    }

    /*
     * Copy the header to the caller's buffer
     */
    MEMCPY ((void *) OutTableHeader, (void *) TblPtr, sizeof (ACPI_TABLE_HEADER));

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiGetTable (
    ACPI_TABLE_TYPE         TableType,
    UINT32                  Instance,
    ACPI_BUFFER             *RetBuffer)
{
    ACPI_TABLE_HEADER       *TblPtr;
    ACPI_STATUS             Status;
    UINT32                  RetBufLen;


    FUNCTION_TRACE ("AcpiGetTable");

    Status = AE_OK;

    /*
     *  Must have a buffer
     */
    if ((Instance == 0)                 ||
        (!RetBuffer)                    ||
        (!RetBuffer->Pointer)           ||
        (!RetBuffer->Length)) 
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    /* Check the table type and instance */

    if ((TableType > ACPI_TABLE_MAX)    ||
        (Gbl_AcpiTableData[TableType].Flags == ACPI_TABLE_SINGLE &&
         Instance > 1))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }
    
    
    /* Get a pointer to the entire table */

    Status = TbGetTablePtr (TableType, Instance, &TblPtr);
    if (Status != AE_OK)
    {
        return_ACPI_STATUS (Status);
    }

    /*
     * The function will return a NULL pointer if the table is not loaded
     */
    if (TblPtr == NULL)
    {
        return_ACPI_STATUS (AE_NOT_EXIST);
    }

    /*
     * Got a table ptr, assume it's ok and copy it to the user's buffer
     */
    if (TableType == TABLE_RSDP)
    {
        /*
         *  RSD PTR is the only "table" without a header
         */
        RetBufLen = sizeof (ROOT_SYSTEM_DESCRIPTOR_POINTER);
    }
    else
    {
        RetBufLen = TblPtr->Length;
    }

    /*
     * Verify we have space in the caller's buffer for the table
     */
    if (RetBuffer->Length < RetBufLen)
    {
        RetBuffer->Length = RetBufLen;
        return_ACPI_STATUS (AE_BUFFER_OVERFLOW);
    }

    RetBuffer->Length = RetBufLen;

    MEMCPY ((void *) RetBuffer->Pointer, (void *) TblPtr, RetBufLen);

    return_ACPI_STATUS (AE_OK);
}

