
/*
 *  ./acpica/subsystem/tables/tbget.c - ACPI Table get* routines
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __TBGET_C__

#include <acpi.h>
#include <hardware.h>
#include <tables.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbget");




ACPI_STATUS
TbGetTablePtr (
    ACPI_TABLE_TYPE         TableType, 
    UINT32                  Instance,
    ACPI_TABLE_HEADER       **TablePtrLoc)
{
    ACPI_TABLE_DESC         *TableDesc;
    UINT32                  i;


    FUNCTION_TRACE ("TbGetTablePtr");
    

    if ((ACPI_TABLE_HEADER *) Gbl_RSDP == NULL)
    {
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    if (TableType > ACPI_TABLE_MAX)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    /*
     * For all table types (Single/Multiple), the first
     * instance is always in the list head.
     */

    if (Instance == 1)
    {
        /* 
         * Just pluck the pointer out of the global table!
         * Will be null if no table is present
         */

        *TablePtrLoc = Gbl_AcpiTables[TableType].Pointer;
        return_ACPI_STATUS (AE_OK);
    }


    /*
     * Check for instance out of range
     */
    if (Instance > Gbl_AcpiTables[TableType].Count)
    {
        return_ACPI_STATUS (AE_NOT_EXIST);
    }

    /* Walk the list to get the table */

    TableDesc = Gbl_AcpiTables[TableType].Next;
    for (i = 1; i < Gbl_AcpiTables[TableType].Count; i++)
    {
        TableDesc = TableDesc->Next;
    }

    /* We are now pointing to the requested table's descriptor */

    *TablePtrLoc = TableDesc->Pointer;

    return_ACPI_STATUS (AE_OK);
}




ACPI_STATUS
TbGetTable (
    void                    *PhysicalAddress, 
    char                    *BufferPtr, 
    ACPI_TABLE_DESC         *TableInfo)
{
    ACPI_TABLE_HEADER       *TableHeader = NULL;
    ACPI_TABLE_HEADER       *FullTable = NULL;
    UINT32                  Size;
    UINT8                   Allocation;
 ACPI_STATUS    Status = AE_OK;


    FUNCTION_TRACE ("TbGetTable");

    if (!TableInfo)
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }


    if (BufferPtr)
    {
        /* 
         * Getting data from a buffer, not BIOS tables 
         */

        TableHeader = (ACPI_TABLE_HEADER *) BufferPtr;
        Status = TbValidateTableHeader (TableHeader);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        /* Valid header and table length, allocate buffer for the entire table */

        FullTable = CmAllocate (TableHeader->Length);
        if (!FullTable)
        {
            return_ACPI_STATUS (AE_NO_MEMORY);
        }
        
        /* Copy the entire table (including header) to the local buffer */

        Size = (ACPI_SIZE) TableHeader->Length;
        MEMCPY (FullTable, BufferPtr, Size);

        /* Save allocation type */

        Allocation = ACPI_MEM_ALLOCATED;
    }
    

    /*
     * Not reading from a buffer, just map the table's physical memory
     * into our address space.
     */
    else
    {
        Size = SIZE_IN_HEADER;

        Status = TbMapAcpiTable (PhysicalAddress, &Size, (void **)&FullTable);
  if (ACPI_FAILURE (Status))
  {
      return_ACPI_STATUS (Status);
  }

        /* Save allocation type */

        Allocation = ACPI_MEM_MAPPED;
    }


    /* Return values */

    TableInfo->Pointer      = FullTable;
    TableInfo->Length       = Size;
    TableInfo->Allocation   = Allocation;
    TableInfo->BasePointer  = FullTable;

    return_ACPI_STATUS (Status);
}



