
/*
 *  ./acpica/subsystem/tables/tbgetall.c - Get firmware ACPI tables
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __TBGETALL_C__

#include <acpi.h>
#include <hardware.h>
#include <tables.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbgetall");




ACPI_STATUS
TbGetAllTables (
    UINT32                  NumberOfTables, 
    char                    *TablePtr)
{
    ACPI_STATUS             Status = AE_OK;
    UINT32                  Index;
    ACPI_TABLE_DESC         TableInfo;

    
    FUNCTION_TRACE ("TbGetAllTables");

    DEBUG_PRINT (ACPI_INFO, ("Number of tables: %d\n", NumberOfTables));


    /* 
     * Loop through all table pointers found in RSDT.
     * This will NOT include the FACS and DSDT - we must get 
     * them after the loop
     */

    for (Index = 0; Index < NumberOfTables; Index++)
    {
        /* Clear the TableInfo each time */

        MEMSET (&TableInfo, 0, sizeof (ACPI_TABLE_DESC));

        /* Get the table via the RSDT */

        Status = TbGetTable ((void *) Gbl_RSDT->TableOffsetEntry[Index], 
                                TablePtr, &TableInfo);
        if (ACPI_FAILURE (Status))
        {
            return_ACPI_STATUS (Status);
        }

        /* Recognize and install the table */

        Status = TbInstallTable (TablePtr, &TableInfo);
        if (ACPI_FAILURE (Status))
        {
            /* 
             * Unrecognized or unsupported table, delete it and ignore the error.
             * Just get as many tables as we can, later we will determine if there
             * are enough tables to continue.
             */

            TbDeleteSingleTable (&TableInfo);
        }
    }



    /*
     * Get the minimum set of ACPI tables, namely:
     *
     * 1) FACP (via RSDT in loop above)
     * 2) FACS
     * 3) DSDT
     *
     */


    /* 
     * Get the FACS (must have the FACP first, from loop above)
     * TbGetTableFacs will fail if FACP pointer is not valid
     */

    Status = TbGetTableFacs (TablePtr, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Install the FACS */

    Status = TbInstallTable (TablePtr, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }


    /* Get the DSDT (We know that the FACP if valid now) */

    Status = TbGetTable ((void *) Gbl_FACP->Dsdt, TablePtr, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Install the DSDT */

    Status = TbInstallTable (TablePtr, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Dump the DSDT Header */

    DEBUG_PRINT (TRACE_TABLES, ("Hex dump of DSDT Header:\n"));
    DUMP_BUFFER ((UINT8 *) Gbl_DSDT, sizeof (ACPI_TABLE_HEADER));
    
    /* Dump the entire DSDT */

    DEBUG_PRINT (TRACE_TABLES, ("Hex dump of DSDT (After header), size %d (0x%x)\n",
                    Gbl_DSDT->Length, Gbl_DSDT->Length));
    DUMP_BUFFER ((UINT8 *) (Gbl_DSDT + 1), Gbl_DSDT->Length);

    /* 
     * Initialize the capabilities flags.
     * Assumes that platform supports ACPI_MODE since we have tables!
     */
        
    Gbl_SystemFlags |= HwGetModeCapabilities ();

    return_ACPI_STATUS (Status);
}
  