
/*
 *  ./acpica/subsystem/tables/tbinstal.c - ACPI table installation
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __TBINSTAL_C__

#include <acpi.h>
#include <hardware.h>
#include <tables.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbinstal");




ACPI_STATUS
TbInstallTable (
    char                    *TablePtr,
    ACPI_TABLE_DESC         *TableInfo)
{
    ACPI_TABLE_TYPE         TableType;
    ACPI_TABLE_HEADER       *TableHeader;
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("TbInstallTable");


    /* 
     * Check the table signature and make sure it is recognized
     * Also checks the header checksum
     */

    Status = TbRecognizeTable (TablePtr, TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    /* Table type is returned by RecognizeTable */

    TableType   = TableInfo->Type;
    TableHeader = TableInfo->Pointer;

    /* Lock tables while installing */

    CmAcquireMutex (MTX_TABLES);

    /* Install the table into the global data structure */

    Status = TbInitTableDescriptor (TableInfo->Type, TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    DEBUG_PRINT (ACPI_INFO, ("%s located at %p\n", 
                                Gbl_AcpiTableData[TableType].Name, TableHeader));

    CmReleaseMutex (MTX_TABLES);
    return_ACPI_STATUS (AE_OK);
}



ACPI_STATUS
TbRecognizeTable (
    char                    *TablePtr,
    ACPI_TABLE_DESC         *TableInfo)
{
    ACPI_TABLE_HEADER       *TableHeader;
    ACPI_STATUS             Status;
    ACPI_TABLE_TYPE         TableType = 0;
    UINT32                  i;

    
    FUNCTION_TRACE ("TbRecognizeTable");


    /* Ensure that we have a valid table pointer */

    TableHeader = (ACPI_TABLE_HEADER *) TableInfo->Pointer;
    if (!TableHeader)
    {   
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /* Search for a signature match among the known table types */

    Status = AE_SUPPORT;
    for (i = 1; i < NUM_ACPI_TABLES; i++)       /* Start at one -> Skip RSDP */
    {
        if (!memcmp (TableHeader->Signature, Gbl_AcpiTableData[i].Signature, Gbl_AcpiTableData[i].SigLength))
        {
            /* Found a signature match, get the pertinent info from the TableData structure */

            TableType       = i;
            Status          = Gbl_AcpiTableData[i].Status;  /* AE_SUPPORT or AE_OK */

            break;
        }
    }

    /* Return the table type and length via the info struct */

    TableInfo->Type     = (UINT8) TableType;
    TableInfo->Length   = TableHeader->Length;


    /* 
     * Validate checksum for _most_ tables,
     * even the ones whose signature we don't recognize
     */

    if (TableType != TABLE_FACS)
    {
        /* But don't abort if the checksum is wrong */
        /* TBD: make this a configuration option? */

        TbVerifyTableChecksum (TableHeader);
    }

    /* 
     * An AE_SUPPORT means that the table was not recognized.
     * We basically ignore this;  just print a debug message
     */

    if (Status == AE_SUPPORT)
    {
        DEBUG_PRINT (ACPI_INFO, ("Unsupported table %s (Type %d) was found and discarded\n",
                            Gbl_AcpiTableData[TableType].Name, TableType));
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
TbInitTableDescriptor (
    ACPI_TABLE_TYPE         TableType,
    ACPI_TABLE_DESC         *TableInfo)
{
    ACPI_TABLE_DESC         *ListHead;
    ACPI_TABLE_DESC         *TableDesc;


    FUNCTION_TRACE_U32 ("TbInitTableDescriptor", TableType);

    /*
     * Install the table into the global data structure 
     */
    
    ListHead    = &Gbl_AcpiTables[TableType];
    TableDesc   = ListHead;
   


    /*
     * Two major types of tables:  1) Only one instance is allowed.  This includes most
     * ACPI tables such as the DSDT.  2) Multiple instances of the table are allowed.  This
     * includes SSDT and PSDTs.
     */

    if (Gbl_AcpiTableData[TableType].Flags == ACPI_TABLE_SINGLE)
    {
        /*
         * Only one table allowed, just update the list head
         */

        if (ListHead->Pointer)
        {
            return_ACPI_STATUS (AE_EXIST);
        }

        TableDesc->Count = 1;
    }


    else 
    {
        /*
         * Multiple tables allowed for this table type, we must link
         * the new table in to the list of tables of this type.
         */

        if (ListHead->Pointer)
        {
            TableDesc = CmCallocate (sizeof (ACPI_TABLE_DESC));
            if (!TableDesc)
            {
                return_ACPI_STATUS (AE_NO_MEMORY);
            }

            ListHead->Count++;

            /* Update the original previous */

            ListHead->Prev->Next = TableDesc;

            /* Update new entry */

            TableDesc->Prev = ListHead->Prev;
            TableDesc->Next = (ACPI_TABLE_DESC *) ListHead;

            /* Update list head */

            ListHead->Prev = TableDesc;
        }

        else
        {
            TableDesc->Count = 1;
        }
    }


    /* Common initialization of the table descriptor */

    TableDesc->Pointer              = TableInfo->Pointer;
    TableDesc->Length               = TableInfo->Length;
    TableDesc->Allocation           = TableInfo->Allocation;
    TableDesc->AmlPointer           = (UINT8 *) (TableDesc->Pointer + 1),
    TableDesc->AmlLength            = (UINT32) (TableDesc->Length - (UINT32) sizeof (ACPI_TABLE_HEADER));
    TableDesc->TableId              = CmAllocateOwnerId (OWNER_TYPE_TABLE);
    TableDesc->LoadedIntoNamespace  = FALSE;


    /* Set the appropriate global pointer (if there is one) to point to the newly installed table */

    if (Gbl_AcpiTableData[TableType].GlobalPtr)
    {
        *(Gbl_AcpiTableData[TableType].GlobalPtr) = TableInfo->Pointer;
    }


    /* Return Data */

    TableInfo->TableId          = TableDesc->TableId;
    TableInfo->InstalledDesc    = TableDesc;

    return_ACPI_STATUS (AE_OK);
}


