
/*
 *  ./acpica/subsystem/tables/tbremove.c - ACPI Table removal and deletion utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __TBREMOVE_C__

#include <acpi.h>
#include <tables.h>
#include <interp.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbremove");




void
TbDeleteAcpiTables (void)
{
    UINT32                      i;


    /*
     * Free memory allocated for ACPI tables
     * Memory can either be mapped or allocated
     */

    for (i = 0; i < ACPI_TABLE_MAX; i++)
    {
        TbDeleteAcpiTable (i);
    }

}



void
TbDeleteAcpiTable (
    ACPI_TABLE_TYPE             Type)
{
 FUNCTION_TRACE ("TbDeleteAcpiTable");


    if (Type > ACPI_TABLE_MAX)
    {
        return_VOID;
    }


    CmAcquireMutex (MTX_TABLES);

    /* Free the table */

    TbFreeAcpiTablesOfType (&Gbl_AcpiTables[Type]);


    /* Clear the appropriate "typed" global table pointer */

    switch (Type)
    {
    case TABLE_RSDP:
        Gbl_RSDP = NULL;
        break;

    case TABLE_APIC:
        Gbl_APIC = NULL;
        break;

    case TABLE_DSDT:
        Gbl_DSDT = NULL;
        break;

    case TABLE_FACP:
        Gbl_FACP = NULL;
        break;

    case TABLE_FACS:
        Gbl_FACS = NULL;
        break;

    case TABLE_PSDT:
        break;

    case TABLE_RSDT:
        Gbl_RSDT = NULL;
        break;

    case TABLE_SSDT:
        break;

    case TABLE_SBST:
        Gbl_SBST = NULL;

    default:
        break;
    }

    CmReleaseMutex (MTX_TABLES);

    return_VOID;
}



ACPI_TABLE_DESC *
TbDeleteSingleTable (
    ACPI_TABLE_DESC         *TableDesc)
{
    ACPI_TABLE_DESC         *NextDesc;


    if (!TableDesc)
    {
        return NULL;
    }


    /* Unlink the descriptor */

    if (TableDesc->Prev)
    {
        TableDesc->Prev->Next = TableDesc->Next;
    }

    if (TableDesc->Next)
    {
        TableDesc->Next->Prev = TableDesc->Prev;
    }


    /* Free the memory allocated for the table itself */

    if (TableDesc->Pointer)
    {
        /* Valid table, determine type of memory allocation */

        switch (TableDesc->Allocation)
        {

        case ACPI_MEM_NOT_ALLOCATED:

            break;


        case ACPI_MEM_ALLOCATED:

            CmFree (TableDesc->BasePointer);
            break;


        case ACPI_MEM_MAPPED:

            OsdUnMapMemory (TableDesc->BasePointer, TableDesc->Length);
            break;
        }
    }


    /* Free the table descriptor (Don't delete the list head, tho) */



    if ((TableDesc->Prev) == (TableDesc->Next))
    {

        NextDesc = NULL;

        /* Clear the list head */

        TableDesc->Pointer   = NULL;
        TableDesc->Length    = 0;
        TableDesc->Count     = 0;

    }

    else
    {
        /* Free the table descriptor */

        NextDesc = TableDesc->Next;
        CmFree (TableDesc);
    }


    return NextDesc;
}





void
TbFreeAcpiTablesOfType (
    ACPI_TABLE_DESC         *ListHead)
{
    ACPI_TABLE_DESC         *TableDesc;
    UINT32                  Count;
    UINT32                  i;


 FUNCTION_TRACE ("TbFreeAcpiTablesOfType");


    /* Get the head of the list */

    TableDesc   = ListHead;
    Count       = ListHead->Count;

    /*
     * Walk the entire list, deleting both the allocated tables
     * and the table descriptors
     */

    for (i = 0; i < Count; i++)
    {
        TableDesc = TbDeleteSingleTable (TableDesc);
    }

    return_VOID;
}


