
/*
 *  ./acpica/subsystem/tables/tbrsd.c - Functions for the ACPI RSDP
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __TBRSD_C__

#include <acpi.h>
#include <hardware.h>
#include <tables.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbrsd");




ACPI_STATUS
TbGetTableRsdt (
    UINT32                  *NumberOfTables)
{
    ACPI_STATUS             Status = AE_OK;
    ACPI_TABLE_DESC         TableInfo;


    FUNCTION_TRACE ("TbGetTableRsdt");

    /* Get the RSDP */

    Status = TbFindRsdp (&TableInfo);
    if (ACPI_FAILURE (Status))
    {
        REPORT_WARNING ("RSDP structure not found");
        return_ACPI_STATUS (AE_NO_ACPI_TABLES);
    }

    /* Save the table pointers and allocation info */

    Status = TbInitTableDescriptor (TABLE_RSDP, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    Gbl_RSDP = (ROOT_SYSTEM_DESCRIPTOR_POINTER *) TableInfo.Pointer;


    /* RSDP structure was found */

    DEBUG_PRINT (ACPI_INFO, ("RSDP located at %p\n", Gbl_RSDP));

    /* Get the RSDT */

    Status = TbGetTable ((void *) Gbl_RSDP->RsdtPhysicalAddress, 
                            NULL, &TableInfo);
    if (Status != AE_OK)
    {
        return_ACPI_STATUS (Status);
    }

    /* Save the table pointers and allocation info */

    Status = TbInitTableDescriptor (TABLE_RSDT, &TableInfo);
    if (ACPI_FAILURE (Status))
    {
        return_ACPI_STATUS (Status);
    }

    Gbl_RSDT = (ROOT_SYSTEM_DESCRIPTION_TABLE *) TableInfo.Pointer;

    /* Valid RSDT pointer */

    if (STRNCMP (Gbl_RSDT->header.Signature, RSDT_SIG, 4))
    {
        /* Invalid RSDT signature */

        REPORT_ERROR ("Invalid signature where RSDP indicates RSDT should be located");

        DEBUG_PRINT (ACPI_INFO,
                    ("RSDP indicates RSDT should be located at %lXh, however the table\n"
                    "  signature at %lXh is incorrect.\n"
                    "  Expected signature: 'RSDT'  Actual signature: '%4.4s'\n",
                    Gbl_RSDP->RsdtPhysicalAddress, Gbl_RSDP->RsdtPhysicalAddress,
                    Gbl_RSDT->header.Signature));

        return_ACPI_STATUS (AE_BAD_SIGNATURE);
    }


    /* Valid RSDT signature, verify the checksum */

    DEBUG_PRINT (ACPI_INFO, ("RSDT located at %p, physical address %lX\n", 
                            Gbl_RSDT, Gbl_RSDP->RsdtPhysicalAddress));

    Status = TbVerifyTableChecksum (Gbl_RSDT);
    
    /* Determine the number of tables pointed to by the RSDT */

    *NumberOfTables = (INT32) (Gbl_RSDT->header.Length - sizeof (ACPI_TABLE_HEADER)) / 4;


    return_ACPI_STATUS (Status);
}



char *
TbScanMemoryForRsdp (
    char                    *StartAddress, 
    UINT32                  Length)
{
    UINT32                  Offset;
    char                    *MemRover;



    /* Search from given start addr for the requested length  */

    for (Offset = 0, MemRover = StartAddress;
         Offset < Length;
         Offset += RSDP_SCAN_STEP, MemRover += RSDP_SCAN_STEP)
    {

        /* The signature and checksum must both be correct */

        if (STRNCMP (MemRover, RSDP_SIG, sizeof (RSDP_SIG)-1) == 0 &&
            TbChecksum (MemRover, sizeof (ROOT_SYSTEM_DESCRIPTOR_POINTER)) == 0)
        {
            /* If so, we have found the RSDP */

            return MemRover;
        }
    }

    /* Searched entire block, no RSDP was found */

    return NULL;
}



ACPI_STATUS
TbFindRsdp (
    ACPI_TABLE_DESC         *TableInfo)
{
    char                    *TablePtr;
    char                    *MemRover;
    ACPI_STATUS             Status = AE_OK;

    FUNCTION_TRACE ("TbFindRsdp");

    if (Gbl_AcpiInitData.RSDP_PhysicalAddress)
    {
        /*
         *  RSDP address was supplied as part of the initialization data
         */

        Status = OsdMapMemory(Gbl_AcpiInitData.RSDP_PhysicalAddress,
                                sizeof (ROOT_SYSTEM_DESCRIPTOR_POINTER),
        (void **)&TablePtr);

  if (ACPI_FAILURE (Status))
  {
            return_ACPI_STATUS (Status);
  }

        if (!TablePtr)
        {
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        /*
         *  The signature and checksum must both be correct
         */

        if (STRNCMP (TablePtr, RSDP_SIG, sizeof (RSDP_SIG)-1) != 0)
        {
            /* Nope, BAD Signature */

            return_ACPI_STATUS (AE_BAD_SIGNATURE);
        }

        /* The signature and checksum must both be correct */

        if (TbChecksum (TablePtr, sizeof (ROOT_SYSTEM_DESCRIPTOR_POINTER)) != 0)
        {
            /* Nope, BAD Checksum */

            return_ACPI_STATUS (AE_BAD_CHECKSUM);
        }

        /* RSDP supplied is OK */
        /* If so, we have found the RSDP */

        TableInfo->Pointer      = (ACPI_TABLE_HEADER *) TablePtr;
        TableInfo->Length       = sizeof (ROOT_SYSTEM_DESCRIPTOR_POINTER);
        TableInfo->Allocation   = ACPI_MEM_MAPPED;
        TableInfo->BasePointer  = TablePtr;

        return_ACPI_STATUS (AE_OK);
    }

    /*
     * Search memory for RSDP.  First map low physical memory.
     */

    Status = OsdMapMemory (LO_RSDP_WINDOW_BASE, LO_RSDP_WINDOW_SIZE, (void **)&TablePtr);

 if (ACPI_FAILURE (Status))
 {
     return_ACPI_STATUS (Status);
 }

    /* 
     * 1) Search EBDA (low memory) paragraphs  
     */

    if (NULL != (MemRover = TbScanMemoryForRsdp (TablePtr, LO_RSDP_WINDOW_SIZE)))
    {
        /* Found it, return pointer and don't delete the mapping */

        TableInfo->Pointer      = (ACPI_TABLE_HEADER *) MemRover;
        TableInfo->Length       = LO_RSDP_WINDOW_SIZE;
        TableInfo->Allocation   = ACPI_MEM_MAPPED;
        TableInfo->BasePointer  = TablePtr;

        return_ACPI_STATUS (AE_OK);
    }

    /* This mapping is no longer needed */

    OsdUnMapMemory (TablePtr, LO_RSDP_WINDOW_SIZE);


    /* 
     * 2) Search upper memory: 16-byte boundaries in E0000h-F0000h  
     */

    Status = OsdMapMemory (HI_RSDP_WINDOW_BASE, HI_RSDP_WINDOW_SIZE, (void **)&TablePtr);

 if (ACPI_FAILURE (Status))
 { 
     return_ACPI_STATUS (Status);
 }

    if (NULL != (MemRover = TbScanMemoryForRsdp (TablePtr, HI_RSDP_WINDOW_SIZE)))
    {
        /* Found it, return pointer and don't delete the mapping */

        TableInfo->Pointer      = (ACPI_TABLE_HEADER *) MemRover;
        TableInfo->Length       = HI_RSDP_WINDOW_SIZE;
        TableInfo->Allocation   = ACPI_MEM_MAPPED;
        TableInfo->BasePointer  = TablePtr;

        return_ACPI_STATUS (AE_OK);
    }

    /* This mapping is no longer needed */

    OsdUnMapMemory (TablePtr, HI_RSDP_WINDOW_SIZE);


    /* RSDP signature was not found */

    return_ACPI_STATUS (AE_NOT_FOUND);
}



