
/*
 *  ./acpica/subsystem/tables/tbutils.c - Table manipulation utilities
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#define __TBUTILS_C__

#include <acpi.h>
#include <tables.h>
#include <interp.h>


#define _COMPONENT          TABLE_MANAGER
        MODULE_NAME         ("tbutils");



ACPI_STATUS
TbHandleToObject (
    UINT16                  TableId,
    ACPI_TABLE_DESC         **TableDesc)
{
    UINT32                  i;
    ACPI_TABLE_DESC         *ListHead;


    for (i = 0; i < ACPI_TABLE_MAX; i++)
    {
        ListHead = &Gbl_AcpiTables[i];
        do
        {
            if (ListHead->TableId == TableId)
            {
                *TableDesc = ListHead;
                return AE_OK;
            }

            ListHead = ListHead->Next;

        } while (ListHead != &Gbl_AcpiTables[i]);
    }


    DEBUG_PRINT (ACPI_ERROR, ("TableId=0x%X does not exist\n", TableId));
    return AE_BAD_PARAMETER;

}



BOOLEAN
TbSystemTablePointer (
    void                    *Where)
{
    UINT32                  i;
    ACPI_TABLE_DESC         *TableDesc;
    ACPI_TABLE_HEADER       *Table;


    /* No function trace, called too often! */


    /* Ignore null pointer */

    if (!Where)
    {
        return (FALSE);
    }


    /* Check for a pointer within the DSDT */

    if (IS_IN_ACPI_TABLE (Where, Gbl_DSDT))
    {
        return (TRUE);
    }


    /* Check each of the loaded SSDTs (if any)*/

    TableDesc = &Gbl_AcpiTables[TABLE_SSDT];

    for (i = 0; i < Gbl_AcpiTables[TABLE_SSDT].Count; i++)
    {
        Table = TableDesc->Pointer;

        if (IS_IN_ACPI_TABLE (Where, Table))
        {
            return (TRUE);
        }

        TableDesc = TableDesc->Next;
    }


    /* TBD: Need to check the PSDTs? */


    return (FALSE);
}




ACPI_STATUS
TbValidateTableHeader (
    ACPI_TABLE_HEADER       *TableHeader)
{

    /* Verify that this is a valid address */

    if (!OsdReadable (TableHeader, sizeof (ACPI_TABLE_HEADER)))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Cannot read table header at %p\n", TableHeader));
        return AE_BAD_ADDRESS;
    }


    /* Ensure that the signature is 4 ASCII characters */

    if (!CmValidAcpiName (*(UINT32 *) TableHeader->Signature))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Table signature [%X] contains one or more invalid characters\n", *(UINT32 *) TableHeader->Signature));
        REPORT_WARNING ("Invalid table signature found");
        return AE_BAD_SIGNATURE;
    }


    /* Validate the table length */

    if (TableHeader->Length < (UINT32) sizeof (ACPI_TABLE_HEADER))
    {
        DEBUG_PRINT (ACPI_ERROR, ("Invalid length in header at %08lXh name %4.4s\n", TableHeader, &TableHeader->Signature));
        REPORT_WARNING ("Invalid table header length found");
        return AE_BAD_HEADER;
    }


    return AE_OK;
}



ACPI_STATUS
TbMapAcpiTable (
    void                    *PhysicalAddress,
    UINT32                  *Size,
 void     **LogicalAddress)
{
    ACPI_TABLE_HEADER       *Table;
    UINT32                  TableSize = *Size;
 ACPI_STATUS    Status = AE_OK;



    /* If size is zero, look at the table header to get the actual size */

    if (!*Size)
    {
        /* Get the table header so we can extract the table length */

        Status = OsdMapMemory (PhysicalAddress, sizeof (ACPI_TABLE_HEADER), (void **)&Table);
  if (ACPI_FAILURE (Status))
  {
   return Status;
  }

        /* Validate the header */

        Status = TbValidateTableHeader (Table);
        if (ACPI_FAILURE (Status))
        {
            OsdUnMapMemory (Table, sizeof (ACPI_TABLE_HEADER));
            return Status;
        }

        /* Now we know how large to make the mapping (table + header) */

        TableSize = Table->Length;
        OsdUnMapMemory (Table, sizeof (ACPI_TABLE_HEADER));
    }


    /* Map the physical memory for the correct length */

    Status = OsdMapMemory (PhysicalAddress, TableSize, (void **)&Table);
 if (ACPI_FAILURE (Status))
 {
  return Status;
 }

 DEBUG_PRINT (ACPI_INFO, ("Mapped memory for ACPI table, length=%d(0x%X) at %p\n",
                    TableSize, TableSize, Table));

    *Size = TableSize;
 *LogicalAddress = Table;

 return Status;
}



ACPI_STATUS
TbVerifyTableChecksum (
    void                    *TableHeader)
{
    UINT8                   CheckSum;
    ACPI_STATUS             Status = AE_OK;


    FUNCTION_TRACE ("TbVerifyTableChecksum");


    /* Compute the checksum on the table */

    CheckSum = TbChecksum (TableHeader,
                        (ACPI_SIZE) ((ACPI_TABLE_HEADER *) TableHeader)->Length);

    /* Return the appropriate exception */

    if (CheckSum)
    {
        REPORT_ERROR ("Invalid ACPI table checksum");
        DEBUG_PRINT (ACPI_INFO, ("TbVerifyTableChecksum: Invalid checksum (%X)\n", CheckSum));

        Status = AE_BAD_CHECKSUM;
    }


    return_ACPI_STATUS (Status);
}



UINT8
TbChecksum (
    void                    *Buffer, 
    UINT32                  Length)
{
    UINT8                   *limit;
    UINT8                   *rover;
    UINT8                   sum = 0;


    if (Buffer && Length)
    {   
        /*  Buffer and Length are valid   */
        
        limit = (UINT8 *) Buffer + Length;

        for (rover = Buffer; rover < limit; rover++)
        {
            sum = (UINT8) (sum + *rover);
        }
    }

    return sum;
}




