/*
 *  acpictl.c - ACPI control utility
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include <popt.h>

#include "ipc.h"

static const char *acpictl_name = "";

/*
 * Log message
 */
void
log(char *fmt,...)
{
	static char buffer[ACPID_MAX_REQUEST];
	va_list args;

	va_start(args, fmt);
	buffer[0] = '\0';
	vsnprintf(buffer, ACPID_MAX_REQUEST-1, fmt, args);
	va_end(args);

	fprintf(stderr, "%s: %s\n", acpictl_name, buffer);
}

/*
 * Open server connection
 */
static int
client_init(void)
{
	int fd, size;
	struct sockaddr_un addr;
	char buffer[1];

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
	{
		fprintf(stderr, "socket: %s\n", strerror(errno));
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = PF_UNIX;
	strcpy(addr.sun_path, ACPID_SERVER_PATH);
	size = strlen(addr.sun_path) + sizeof(addr.sun_family);

	if (connect(fd, (struct sockaddr*) &addr, size) < 0)
	{
		log("acpid is not running.");
		close(fd);
		return -1;
	}

	if ((write(fd, ACPID_PROTOCOL, strlen(ACPID_PROTOCOL))
	    < strlen(ACPID_PROTOCOL)) ||
	    (read(fd, buffer, 1) < 1))
	{
		log("bad protocol version.");
		close(fd);
		return -1;
	}

	return fd;
}

/*
 * Send a simple request and listen for a return
 * Return value points to static struct that will be
 * overwritten next time the function is called.
 */
struct acpid_request *
send_request(int fd, int type, const char *data)
{
	static char buffer[ACPID_MAX_REQUEST];
	struct acpid_request *rqst = (struct acpid_request*) buffer;
	
	rqst->type = type;
	rqst->status = ACPID_STATUS_OK;
	strcpy(rqst->data, data ? data:"");
	rqst->size = sizeof(*rqst) + strlen(rqst->data) + 1;

	if (write(fd, rqst, rqst->size) != rqst->size
	    || read(fd, buffer, sizeof(buffer)) < sizeof(struct acpid_request))
	{
		return NULL;
	}

	return rqst;
}

static void
do_battery_request(int fd)
{
	int i, count;
	char batnum[12]; /* more than enough for an integer */
	struct acpid_request *rqst;

	rqst = send_request(fd, ACPID_BATTERY_GET_COUNT, NULL);
	if (rqst->status != ACPID_STATUS_OK)
	{
		log("no batteries");
		return;
	}

	log("%s batteries", rqst->data);
	count = strtol(rqst->data, NULL, 0);
	for(i = 0; i < count; i++) {
		snprintf(batnum, sizeof(batnum), "%d", i);
		rqst = send_request(fd, ACPID_BATTERY_GET_INFO, batnum);
		log("battery[%d]   info: %s", i, rqst->data);
		rqst = send_request(fd, ACPID_BATTERY_GET_STATUS, batnum);
		log("battery[%d] status: %s", i, rqst->data);
	}
}

static void
do_pid_request(int fd)
{
	struct acpid_request *rqst;

	rqst = send_request(fd, ACPID_PID_GET, NULL);
	if (rqst->status != ACPID_STATUS_OK)
		log("acpid not running");
	else
		log("acpid pid %s", rqst->data);
}


int
main(int argc, char **argv)
{
	int fd;
	int acpid_pid = 0;
	int battery = 0;
	int intermediary = 0;
	int rc;
	struct acpid_request *rqst;

	poptContext optCon;
	struct poptOption optionsTable[] = {
		{ "pid", 'p', POPT_ARG_NONE, &acpid_pid, 0,
		  "print acpid's pid", NULL},
		{ "battery", 'b', POPT_ARG_NONE, &battery, 0,
		  "print battery information", NULL},
		{ "intermediary", 'i', POPT_ARG_NONE, &intermediary, 0,
		  "act as intermediary between another program and acpid", NULL},
#if 0
		{ "", '', POPT_ARG_, &, 0,
		  "", ""},
#endif
		POPT_AUTOHELP
		{ NULL, 0, 0, NULL, 0 },
	};


	acpictl_name = strrchr(argv[0], '/');
	acpictl_name = acpictl_name ? (acpictl_name + 1):argv[0];

	optCon = poptGetContext("acpictl", argc,
				(const char**) argv,
				optionsTable, 0);
	rc = poptReadDefaultConfig(optCon, 0);
	if (rc < 0)
	{
		fprintf(stderr, "%s: %s\n", acpictl_name, poptStrerror(rc));
		exit(1);
	}
	rc = poptGetNextOpt(optCon);
	if (rc < -1)
	{
		fprintf(stderr, "%s: %s: %s\n", acpictl_name,
			poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
			poptStrerror(rc));
		exit(1);
	}
	poptFreeContext(optCon);

	fd = client_init();
	if (fd < 0)
		exit(1);


	if (intermediary)
	{
		/* FIXME */

		/* do not try running anything else */
		exit (0);
	}


	if (acpid_pid)
		do_pid_request(fd);

	if (battery)
		do_battery_request(fd);

	close(fd);
	
	return 0;
}

