/*
 *  acpid.c - ACPI daemon
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ac-config.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/io.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <linux/acpi.h>

#include "acpid.h"
#include "methods.h"
#include "globals.h"
#include "output.h"

static int acpid_debug = 0;

UINT8 *AmlPtr = NULL;
UINT32 AmlLength = 0;

/*
 * Log message to syslog or stdout
 */
void
acpid_log(int priority, char *fmt,...)
{
	static char buffer[256];
	va_list args;

#ifndef ACPI_DEBUG
	if (priority == LOG_DEBUG)
		return;
#endif

	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);

	if (!acpid_debug)
		syslog(priority, buffer);
	else
		fprintf(stderr, "acpid: %s\n", buffer);
}

/*
 * Load an ACPI table from file
 */
static struct acpi_table *
acpid_load_table(const char *path)
{
	struct acpi_table *table = NULL;
	int fd = open(path, O_RDONLY);
	if (fd >= 0)
	{
		struct acpi_table hdr;
		if (read(fd, &hdr, sizeof(hdr)) == sizeof(hdr))
		{
			table = malloc(hdr.length);
			if (table)
			{
				unsigned long size = hdr.length - sizeof(hdr);
				memcpy(table, &hdr, sizeof(hdr));
				if (read(fd, table + 1, size) != size)
				{
					free(table);
					table = NULL;
				}
			}
		}
		close(fd);
	}
	return table;
}

/*
 * Examine sysctl value
 */
unsigned long
acpid_read_sysctl(const char *name)
{
	static char buffer[128];
	int fd, size;

	sprintf(buffer, ACPI_PROC "/%s", name);
	fd = open(buffer, O_RDONLY);
	if (fd < 0)
	{
		acpid_log(LOG_ERR, "%s: %s", buffer, strerror(errno));
		return 0;
	}
	size = read(fd, buffer, sizeof(buffer));
	close(fd);
	return strtoul(buffer, NULL, 0);
}

/*
 * Modify sysctl value
 */
int
acpid_write_sysctl(const char *name, unsigned long value)
{
#ifndef ACPI_STANDALONE
	static char buffer[128];
	int fd, size;

	sprintf(buffer, ACPI_PROC "/%s", name);
	fd = open(buffer, O_RDWR);
	if (fd < 0)
	{
		return -1;
	}
	size = sprintf(buffer, "0x%08lx\n", value);
	write(fd, buffer, size);
	close(fd);
#endif
	return 0;
}

/*
 * Get information from processor block object
 */
static ACPI_STATUS
acpid_find_pblk(ACPI_HANDLE handle, UINT32 level, void *ctx, void **value)
{
	unsigned long *p_blk = (unsigned long *) ctx;
	NAME_TABLE_ENTRY *obj;
	ACPI_OBJECT_INTERNAL *proc;

	obj = NsConvertHandleToEntry(handle);
	proc = NsGetAttachedObject((ACPI_HANDLE) obj);
	if (proc)
	{
		acpid_log(LOG_INFO, "P_BLK %d @ 0x%04x:%d",
		    proc->Processor.ProcId,
		    proc->Processor.PBLKAddress,
		    proc->Processor.PBLKLength);
		if (proc->Processor.PBLKAddress && !*p_blk)
			*p_blk = proc->Processor.PBLKAddress;
	}

	return AE_OK;
}

/*
 * Upload power management values
 */
static int
acpid_set_values()
{
	ACPI_OBJECT values[5];
	ACPI_BUFFER values_buffer;
	ACPI_OBJECT_LIST params;
	ACPI_OBJECT param;
	char sx_name[6], sx_sysctl[32];
	unsigned long p_blk = 0;
	int sx;

	// fetch SLP_TYP values
	for (sx = 0; sx < 6; sx++)
	{
		sprintf(sx_name, "\\_S%d_", sx);
		sprintf(sx_sysctl, "s%d_slp_typ", sx);
		
		values_buffer.Length = sizeof(values);
		values_buffer.Pointer = values;
		if (ACPI_SUCCESS(AcpiEvaluateObject(NULL,
					sx_name,
					NULL,
					&values_buffer))
			&& values[0].Type == ACPI_TYPE_Package
			&& values[0].Package.Count >= 2)
		{
			unsigned value = ((values[1].Number.Value << 8)
				| values[2].Number.Value);
			acpid_log(LOG_INFO, "S%d SLP_TYP (0x%04x)", sx, value);
			acpid_write_sysctl(sx_sysctl, value);
		}
		else
		{
			acpid_log(LOG_INFO, "S%d not supported", sx);
		}
	}

	if (ACPI_SUCCESS(AcpiWalkNamespace(ACPI_TYPE_Processor,
				ACPI_ROOT_OBJECT,
				ACPI_INT32_MAX,
				acpid_find_pblk,
				&p_blk,
				NULL))
		&& p_blk)
	{
		acpid_write_sysctl("p_blk", p_blk);
	}
	else
	{
		acpid_log(LOG_INFO, "no usable processor block");
	}

	return 0;
}

/*
 * Initialize ACPI-CA subsystem
 */
static int
acpid_init(char **argv)
{
	ACPI_TABLE_HEADER *table, hdr;
	unsigned long table_addr;
	ACPI_STATUS status;
	char sig[5];
	int facp_loaded = 0;
	int dsdt_loaded = 0;

	acpid_log(LOG_INFO, "started");

	if (!ACPI_SUCCESS(AcpiInitialize(NULL)))
	{
		acpid_log(LOG_ERR, "cannot initialize ACPI");
		return -1;
	}

	/*
	 * tables specified on the command line override tables
	 * found in /proc/sys/acpi/
	 */
	while (*argv)
	{
		table = (ACPI_TABLE_HEADER *) acpid_load_table(*argv);
		if (!table || !ACPI_SUCCESS(AcpiLoadTable(table)))
		{
			acpid_log(LOG_ERR, "cannot load %s", *argv);
			return -1;
		}
		memcpy(sig, table->Signature, 4);
		sig[4] = '\0';
		if (strcmp(sig, "FACP") == 0)
			facp_loaded = 1;
		else if (strcmp(sig, "DSDT") == 0)
			dsdt_loaded = 1;
		acpid_log(LOG_INFO, "loaded %s from %s", sig, *argv);
		free(table);
		argv++;
	}

	if (!facp_loaded)
	{
		table = (ACPI_TABLE_HEADER *)
			acpid_load_table(ACPI_PROC "/facp");
		if (table)
		{
			acpid_log(LOG_INFO, "FACP loaded from " ACPI_PROC "/");
			if (!ACPI_SUCCESS(AcpiLoadTable(table)))
			{
				acpid_log(LOG_ERR, "cannot load FACP");
				return -1;
			}
		}
	}

	if (!dsdt_loaded)
	{
		table = (ACPI_TABLE_HEADER *)
			acpid_load_table(ACPI_PROC "/dsdt");
		if (table)
		{
			acpid_log(LOG_INFO, "DSDT loaded from " ACPI_PROC "/");
			if (!ACPI_SUCCESS(AcpiLoadTable(table)))
			{
				acpid_log(LOG_ERR, "cannot load DSDT");
				return -1;
			}
		}
	}

	status = AcpiLoadNamespace();
	if (!ACPI_SUCCESS(status) && status != AE_CTRL_PENDING)
	{
		acpid_log(LOG_ERR, "cannot load ACPI namespace");
		return -1;
	}

	if (acpid_set_values())
	{
		return -1;
	}

	return 0;
}

/*
 * Deinitialize ACPI-CA subsystem
 */
void
acpid_exit(int sig)
{
	static int exited = 0;
	if (!exited)
	{
		exited = 1;
		acpid_log(LOG_INFO, "exiting");
		AcpiTerminate();
		acpid_event_exit();
		kill(0, SIGTERM);
	}
	exit(0);
}

/*
 * Parse command line arguments
 */
static char **
acpid_getopts(const char *name, int argc, char **argv)
{
	struct option opts[] =
	{
		{"debug", 0, 0, 'd'},
		{"trace", 0, 0, 't'},
		{"version", 0, 0, 'v'},
		{"help", 0, 0, 'h'},
		{NULL, 0, 0, 0},
	};
	const char *opts_help[] =
	{
		"Run daemon in the foreground.",    /* debug */
		"Create a (sizeable) debug trace.", /* trace */
		"Print version information.",       /* version */
		"Print this message.",              /* help */
	};
	struct option *opt;
	const char **hlp;
	int max, size;

	for (;;)
	{
		int i = getopt_long(argc, argv, "dtvh", opts, NULL);
		if (i == -1)
		{
			break;
		}
		switch (i)
		{
		case 'd':
			acpid_debug++;
			break;
		case 't':
			DebugLevel = ~0;
			DebugLayer = ~0;
			break;
		case 'v':
			printf(PACKAGE "-" VERSION "\n");
			exit(0);
		case 'h':
		default:
			fprintf(stderr,
				"Usage: %s [OPTIONS] [FILE ...]\n",
				name);
			max = 0;
			for (opt = opts; opt->name; opt++)
			{
				size = strlen(opt->name);
				if (size > max)
					max = size;
			}
			for (opt = opts, hlp = opts_help;
			     opt->name;
			     opt++, hlp++)
			{
				fprintf(stderr,
					"  -%c, --%s",
					opt->val,
					opt->name);
				size = strlen(opt->name);
				for (; size < max; size++)
					printf(" ");
				printf("  %s\n", *hlp);
			}
			exit(1);
			break;
		}
	}

	return (argv + optind);
}

int
main(int argc, char **argv)
{
	const char *name;
	int event, server;
	int pid;

	name = (const char *) strrchr(argv[0], '/');
	name = name ? (name + 1) : argv[0];

	argv = acpid_getopts(name, argc, argv);

#ifndef ACPI_STANDALONE
	if (geteuid() != 0)
	{
		fprintf(stderr, "%s: must be run as root\n", name);
		exit(1);
	}
#endif

	if (access(ACPI_PROC, R_OK))
	{
		fprintf(stderr, "%s: ACPI is not present\n", name);
		exit(1);
	}

#ifndef ACPI_STANDALONE
	pid = acpid_server_detect();
	if (pid >= 0)
	{
		fprintf(stderr, "%s: already running (pid %d)\n", name, pid);
		exit(1);
	}
#endif

	if (!acpid_debug)
	{
		pid = fork();
		if (pid < 0)
		{
			fprintf(stderr, "%s: fork: %s\n",
				name, strerror(errno));
			exit(1);
		}
		else if (pid)
		{
			exit(0);
		}

		setsid();
		chdir("/");
		umask(0);

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

#ifndef ACPI_STANDALONE
	event = acpid_event_init();
	if (event < 0)
	{
		exit(1);
	}
#else
	event = open("/dev/null", O_RDONLY);
#endif

#ifndef ACPI_STANDALONE
	server = acpid_server_init();
	if (server < 0)
	{
		exit(1);
	}

	if (iopl(3))
	{
		acpid_log(LOG_ERR, "iopl: %s", strerror(errno));
		exit(1);
	}
#else
	server = open("/dev/null", O_RDONLY);
#endif

	if (acpid_init(argv))
	{
		exit(1);
	}

	signal(SIGINT, acpid_exit);
	signal(SIGQUIT, acpid_exit);
	signal(SIGTERM, acpid_exit);

	acpid_pci_init();

#if 0
	acpid_bm_init();
#endif

	if (acpid_debug
	    && isatty(STDOUT_FILENO)
	    && isatty(STDIN_FILENO))
	{
		acpid_debug_init();
		acpid_event_loop(event, server, STDIN_FILENO);
	}
	else
	{
		acpid_event_loop(event, server, -1);
	}

	close(server);
	close(event);

	acpid_exit(0);

	return 0;
}
