/*
 *  bm.c - Bus manager interface
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include "acpid.h"
#include "busmgr.h"
#include "ec.h"
#include "power.h"
#include "thermal.h"
#include "battery.h"
#include "event.h"
#include "ipc.h"
#include "list.h"

struct acpid_battery
{
	BM_HANDLE handle;
	int id;
	BT_BATTERY_INFO info;
	BT_BATTERY_STATUS status;
	struct list_entry entry;
};

static LIST_HEAD(acpid_batteries);

#define SCRATCH_SIZE 256
static UINT8 scratch[SCRATCH_SIZE];

static struct acpid_battery *
acpid_create_battery(BM_HANDLE handle, int id)
{
	struct acpid_battery *bat;
	ACPI_STATUS status;
	BM_REQUEST_INFO rqst;
	const char *units = "?";
	
	bat = malloc(sizeof(struct acpid_battery));
	if (!bat)
		return NULL;
	memset(bat, 0, sizeof(*bat));
	bat->handle = handle;
	bat->id = id;

	memset(&rqst, 0, sizeof(rqst));
	rqst.Status = AE_OK;
	rqst.DeviceHandle = handle;
	rqst.Command = BT_COMMAND_GET_INFO;
	rqst.Buffer.Length = sizeof(scratch);
	rqst.Buffer.Pointer = scratch;
	
	status = BmRequest(&rqst);
	if (!ACPI_SUCCESS(status))
	{
		acpid_log(LOG_ERR, "BT_COMMAND_GET_INFO failed");
		free(bat);
		return NULL;
	}

	memcpy(&bat->info, scratch, sizeof(bat->info));

	switch (bat->info.PowerUnit)
	{
	case 0: units = "mW"; break;
	case 1: units = "mA"; break;
	}

	acpid_log(LOG_DEBUG, "  design capacity:       %d %sh",
	    bat->info.DesignCapacity, units);
	acpid_log(LOG_DEBUG, "  last full capacity:    %d %sh",
	    bat->info.LastFullChargeCapacity, units);
	acpid_log(LOG_DEBUG, "  technology:            %s",
	    bat->info.BatteryTechnology ? "rechargeable":"non-rechargeable");
	acpid_log(LOG_DEBUG, "  design voltage:        %d mV",
	    bat->info.DesignVoltage);
	acpid_log(LOG_DEBUG, "  warning threshold:     %d %sh",
	    bat->info.DesignCapacityOfWarning, units);
	acpid_log(LOG_DEBUG, "  low threshold:         %d %sh",
	    bat->info.DesignCapacityOfLow, units);
	acpid_log(LOG_DEBUG, "  low-warn granularity:  %d %sh",
	    bat->info.DesignCapacityGranularity1, units);
	acpid_log(LOG_DEBUG, "  full-warn granularity: %d %sh",
	    bat->info.DesignCapacityGranularity2, units);
	acpid_log(LOG_DEBUG, "  model number:          %s",
		  bat->info.ModelNumber);
	acpid_log(LOG_DEBUG, "  serial number:         %s",
		  bat->info.SerialNumber);
	acpid_log(LOG_DEBUG, "  battery type:          %s",
		  bat->info.BatteryType);
	acpid_log(LOG_DEBUG, "  oem information:       %s",
	    bat->info.OEMInformation);

	list_append(&acpid_batteries, &bat->entry);

	return bat;
}

static struct acpid_battery *
acpid_find_battery(int id)
{
	struct list_entry *entry = acpid_batteries.next;
	while (entry != &acpid_batteries)
	{
		struct acpid_battery *bat
			= LIST_ENTRY(entry, struct acpid_battery, entry);
		if (bat->id == id)
			return bat;
		entry = entry->next;
	}
	return NULL;
}

static int
acpid_get_battery_status(struct acpid_battery *bat)
{
	ACPI_STATUS status;
	BM_REQUEST_INFO rqst;
	const char *units = "?";

	if (!bat)
		return -1;

	memset(&bat->status, 0, sizeof(bat->status));

	memset(&rqst, 0, sizeof(rqst));
	rqst.Status = AE_OK;
	rqst.DeviceHandle = bat->handle;
	rqst.Command = BT_COMMAND_GET_STATUS;
	rqst.Buffer.Length = sizeof(BT_BATTERY_STATUS);
	rqst.Buffer.Pointer = &bat->status;

	status = BmRequest(&rqst);
	if (!ACPI_SUCCESS(status))
		acpid_log(LOG_ERR, "BT_COMMAND_GET_STATUS failed");

	switch (bat->info.PowerUnit)
	{
	case 0: units = "mW"; break;
	case 1: units = "mA"; break;
	}

	acpid_log(LOG_DEBUG, "  present rate:       %d %s",
	    bat->status.PresentRate, units);
	acpid_log(LOG_DEBUG, "  remaining capacity: %d %sh",
	    bat->status.RemainingCapacity, units);
	acpid_log(LOG_DEBUG, "  present voltage:    %d mV",
	    bat->status.PresentVoltage);

	return 0;
}

static void
acpid_enum_battery(void)
{
	ACPI_STATUS status;
	BM_REQUEST_INFO rqst;
	BM_SEARCH_INFO search;

	memset(&rqst, 0, sizeof(rqst));
	rqst.Status = AE_OK;
	rqst.DeviceHandle = BM_HANDLE_ROOT;
	rqst.Command = BM_COMMAND_SEARCH;
	rqst.Buffer.Length = sizeof(search);
	rqst.Buffer.Pointer = &search;

	memset(&search, 0, sizeof(search));
	memcpy(&search.Criteria.HardwareId,
	       BM_HID_CM_BATTERY,
	       sizeof(BM_HID_CM_BATTERY));

	status = BmRequest(&rqst);
	if (ACPI_SUCCESS(status))
	{
		int i;

		acpid_log(LOG_INFO,
			  "found %d batteries",
			  search.Results.Count);
		
		for (i = 0; i < search.Results.Count; i++)
		{
			BM_HANDLE handle = search.Results.Handles[i];
			struct acpid_battery *bat
				= acpid_create_battery(handle, i);
			if (bat)
				acpid_get_battery_status(bat);
		}
	}
	else
	{
		acpid_log(LOG_INFO, "found no batteries");
	}
}

int
acpid_battery_request(struct acpid_request *rqst)
{
	switch (rqst->type)
	{
	case ACPID_BATTERY_GET_COUNT:
		{
			int count = 0;
			struct list_entry *entry = acpid_batteries.next;
			while (entry != &acpid_batteries)
			{
				count++;
				entry = entry->next;
			}
			sprintf(rqst->data, "%d", count);
		}
		break;
	case ACPID_BATTERY_GET_INFO:
		{
			int id = strtoul(rqst->data, 0, 0);
			struct acpid_battery *bat = acpid_find_battery(id);
			if (!bat)
			{
				strcpy(rqst->data, "Not found");
				return -1;
			}
			sprintf(rqst->data,
				"%d %d %d %d %d %d %d %d %d %s %s %s %s",
				bat->info.PowerUnit,
				bat->info.DesignCapacity,
				bat->info.LastFullChargeCapacity,
				bat->info.BatteryTechnology,
				bat->info.DesignVoltage,
				bat->info.DesignCapacityOfWarning,
				bat->info.DesignCapacityOfLow,
				bat->info.DesignCapacityGranularity1,
				bat->info.DesignCapacityGranularity2,
				bat->info.ModelNumber,
				bat->info.SerialNumber,
				bat->info.BatteryType,
				bat->info.OEMInformation);
		}
		break;
	case ACPID_BATTERY_GET_STATUS:
		{
			int id = strtoul(rqst->data, 0, 0);
			struct acpid_battery *bat = acpid_find_battery(id);
			if (!bat)
			{
				strcpy(rqst->data, "Not found");
				return -1;
			}
			if (acpid_get_battery_status(bat))
			{
				strcpy(rqst->data, "Bad status");
				return -1;
			}
			sprintf(rqst->data,
				"%d %d %d %d",
				bat->info.PowerUnit,
				bat->status.PresentRate,
				bat->status.RemainingCapacity,
				bat->status.PresentVoltage);
		}
		break;
	default:
		strcpy(rqst->data, "Not implemented (battery)");
		return -1;
	}
	return 0;
}

static void
acpid_enum_thermal(void)
{
	ACPI_STATUS status;
	BM_REQUEST_INFO rqst;
	BM_SEARCH_INFO search;

	memset(&rqst, 0, sizeof(rqst));
	rqst.Status = AE_OK;
	rqst.DeviceHandle = BM_HANDLE_ROOT;
	rqst.Command = BM_COMMAND_SEARCH;
	rqst.Buffer.Length = sizeof(search);
	rqst.Buffer.Pointer = &search;

	memset(&search, 0, sizeof(search));
	search.Criteria.Class = BM_CLASS_THERMAL_ZONE;

	status = BmRequest(&rqst);
	if (ACPI_SUCCESS(status))
	{
		acpid_log(LOG_INFO,
			  "found %d thermal zones",
			  search.Results.Count);
	}
	else
	{
		acpid_log(LOG_INFO, "no thermal zones");
	}
}

int
acpid_bm_init(void)
{
	if (!ACPI_SUCCESS(BmInitialize()))
	{
		acpid_log(LOG_ERR, "BmInitialize failed");
		return -1;
	}

	EcInitialize();
	PwInitialize();
	TzInitialize();
	BtInitialize();

	acpid_enum_battery();
	acpid_enum_thermal();

	return 0;
}

void
acpid_bm_exit(void)
{
	BtTerminate();
	TzTerminate();
	PwTerminate();
	EcTerminate();
}
