/*
 *  debug.c - Debugger support
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "acpid.h"
#include "acobject.h"
#include "common.h"
#include "debugger.h"
#include "internal.h"

#ifndef ACPI_DEBUG
UINT8 Gbl_DbOutputFlags = DB_CONSOLE_OUTPUT;
char LineBuf[80];
#endif
#if !defined(ACPI_APPLICATION) || !defined(ACPI_DEBUG)
FILE *DebugFile = NULL;
#else
extern FILE *DebugFile;
#endif

int acpid_debug_output = 0;



#ifdef ACPI_GDB_STYLE

struct acpid_debug_cmd
{
	const char *name;
	const char *args;
	const char *help;
	const char *mapping;
};

static struct acpid_debug_cmd acpid_cmds[] =
{
	{"print", "NAME [ARGS ...]", "Evaluate expression", "execute"},
	{"x", "NAME|ADDRESS", "Examine object or memory", "dump"},

	{"debug", "NAME", "Debug control method", "debug"},
	{"next", "", "Step over instruction", ""},
	{"step", "", "Step into instruction", "into"},
	{"continue", "", "Continue running", "go"},

	{"namespace", "", "Display namespace", "namespace"},
	{"tables", "", "Display loaded tables", "tables"},

	{"disassemble", "COUNT", "Disassemble AML", "list"},
	{"backtrace", "", "Print backtrace", "tree"},

	{"level", "LEVEL", "Set debug level", "level"},

	{"help", "[COMMAND]", "Print command help", NULL},
	{"quit", "", "Exit debugger", NULL},

	{NULL, NULL, NULL}
};

static struct acpid_debug_cmd *
acpid_debug_match(const char *str)
{
	struct acpid_debug_cmd *cmd;
	for (cmd = acpid_cmds; cmd->name; cmd++)
	{
		const char *s1 = cmd->name, *s2 = str;
		for (; *s1 && *s2 && *s1 == *s2; s1++, s2++) /* empty */ ;
		if (!*s2 || isspace(*s2))
			break;
	}
	if (!cmd->name)
	{
		char name[16];
		memcpy(name, str, sizeof(name));
		name[sizeof(name) - 1] = '\0';
		name[strcspn(name, " \t")] = '\0';
		printf("Undefined command: \"%s\". Try \"help\".\n", name);
		cmd = NULL;
	}
	return cmd;
}

static void
acpid_debug_help(const char *str)
{
	struct acpid_debug_cmd *cmd;
	if (!*str)
	{
		int size, max = 0;
		for (cmd = acpid_cmds; cmd->name; cmd++)
		{
			size = strlen(cmd->name) + strlen(cmd->args);
			if (size > max)
				max = size;
		}
		for (cmd = acpid_cmds; cmd->name; cmd++)
		{
			printf("  %s %s", cmd->name, cmd->args);
			size = strlen(cmd->name) + strlen(cmd->args);
			for (; size < max; size++)
				printf(" ");
			printf("  %s\n", cmd->help);
		}
	}
	else
	{
		cmd = acpid_debug_match(str);
		if (cmd)
			printf("%s\n", cmd->help);
	}
}

#endif /* ACPI_GDB_STYLE */

void
acpid_debug_printf(const char *str)
{
	if ((Gbl_DbOutputFlags & DB_REDIRECTABLE_OUTPUT) && DebugFile)
		fprintf(DebugFile, "%s", str);
	else
		printf("%s", str);
}

static void
acpid_debug_prompt(void)
{
	printf("(debug) ");
	fflush(stdout);
}

int
acpid_debug_init(void)
{
	acpid_debug_prompt();
	acpid_debug_output = 1;
	return 0;
}

int
acpid_debug_request(int fd)
{
	char buffer[64];
	int size;
	
	size = read(fd, buffer, sizeof(buffer));
	if (size <= 0)
		return 0;
	buffer[size - 1] = '\0';

#ifdef ACPI_GDB_STYLE
	{
		static char last[64] = {0};
		struct acpid_debug_cmd *cmd;
		char *bp;

		bp = buffer + strspn(buffer, " \t");
		if (!*bp)
		{
			strcpy(buffer, last);
			bp = buffer + strspn(buffer, " \t");
			if (!*bp)
			{
				acpid_debug_prompt();
				return 0;
			}
		}

		strcpy(last, buffer);

		cmd = acpid_debug_match(bp);
		if (!cmd)
		{
			acpid_debug_prompt();
			return 0;
		}

		bp += strcspn(bp, " \t");
		bp += strspn(bp, " \t");
		
		if (cmd && cmd->mapping)
		{
			strcpy(LineBuf, cmd->mapping);
			if (*bp)
			{
				strcat(LineBuf, " ");
				strcat(LineBuf, bp);
			}
		}
		else if (cmd)
		{
			if (!strcmp(cmd->name, "help"))
				acpid_debug_help(bp);
			if (!strcmp(cmd->name, "quit"))
				acpid_exit(0);

			acpid_debug_prompt();
			return 0;
		}
	}
#else /* ACPI_GDB_STYLE */
	strcpy(LineBuf, buffer);
#endif /* ACPI_GDB_STYLE */

        CmReleaseMutex (MTX_DEBUG_CMD_READY);
        CmAcquireMutex (MTX_DEBUG_CMD_COMPLETE);

	acpid_debug_prompt();
	return 0;
}
