/*
 *  event.c - Event handling
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <linux/acpi.h>
#include "acpid.h"
#include "event.h"

#define ACPI_MAX_FIXED 32
#define ACPI_MAX_GPE 32

/*
 * Event handler and context
 */
struct acpid_action
{
	union
	{
		FIXED_EVENT_HANDLER fixed;
		GPE_HANDLER gpe;
	} handler;
	void *context;
	int level;
};

static pid_t acpid_event_pid = -1;
static int acpid_event_rcv = -1;
static int acpid_event_ack = -1;

static struct acpid_action acpid_fixed[ACPI_MAX_FIXED];
static struct acpid_action acpid_gpe[ACPI_MAX_GPE];

static acpi_sstate_t acpid_event_state = 0;

/*
 * Handle power button press
 */
static UINT32
acpid_pwrbtn(void *context)
{
	acpid_log(LOG_INFO, "power button pressed");
	if (acpid_event_state == ACPI_S0)
	{
		// power button pressed from S0, halt system
		system("/sbin/init 0");
	}
	return 0;
}

/*
 * Handle sleep button press
 */
static UINT32
acpid_slpbtn(void *context)
{
	acpid_log(LOG_INFO, "sleep button pressed");
	if (acpid_event_state == ACPI_S0)
	{
		// sleep button pressed from S0, sleep
		acpid_write_sysctl("sleep", 1);
	}
	return 0;
}

/*
 * Create a child process to monitor for ACPI events
 */
int
acpid_event_init(void)
{
	int child_to_parent[2];
	int parent_to_child[2];
	char buffer[32];
	int fd, size;
	pid_t pid;

	if (pipe(child_to_parent) < 0 || pipe(parent_to_child) < 0)
	{
		acpid_log(LOG_ERR, "pipe: %s", strerror(errno));
		return -1;
	}

	pid = fork();
	if (pid < 0)
	{
		acpid_log(LOG_ERR, "fork: %s", strerror(errno));
		return -1;
	}
	else if (pid > 0)
	{
		close(child_to_parent[1]);
		close(parent_to_child[0]);
		acpid_event_pid = pid;
		acpid_event_rcv = child_to_parent[0];
		acpid_event_ack = parent_to_child[1];

		// enable power button and sleep button
		acpi_install_fixed_event_handler(EVENT_POWER_BUTTON,
						 acpid_pwrbtn, 0);
		acpi_enable_event(EVENT_POWER_BUTTON, EVENT_FIXED);
		acpi_install_fixed_event_handler(EVENT_SLEEP_BUTTON,
						 acpid_slpbtn, 0);
		acpi_enable_event(EVENT_SLEEP_BUTTON, EVENT_FIXED);

		return child_to_parent[0];
	}

	close(child_to_parent[0]);
	close(parent_to_child[1]);
	acpid_event_rcv = child_to_parent[1];
	acpid_event_ack = parent_to_child[0];

	fd = open(ACPI_PROC "/event", O_RDONLY);
	if (fd < 0)
	{
		acpid_log(LOG_ERR, ACPI_PROC "/event: %s", strerror(errno));
		exit(1);
	}
	
	while(1)
	{
		size = read(fd, buffer, sizeof(buffer));
		if (size <= 0
		    || write(acpid_event_rcv, buffer, size) <= 0
		    || read(acpid_event_ack, buffer, sizeof(buffer)) <= 0)
			break;
	}

	close(fd);
	exit(0);
}

/*
 * Kill child process
 */
void
acpid_event_exit(void)
{
	if (acpid_event_pid > 0)
	{
		close(acpid_event_rcv);
		close(acpid_event_ack);
		kill(acpid_event_pid, SIGTERM);
		waitpid(acpid_event_pid, NULL, 0);
	}
}

/*
 * Handle an ACPI event
 */
int
acpid_event_request(int fd)
{
	char buffer[32], *bufend;
	unsigned long pm1_status, gpe_status;
	struct acpid_action *action;
	int size, bit;
	
	size = read(fd, buffer, sizeof(buffer));
	if (size <= 0)
		return -1;

	pm1_status = strtoul(buffer, &bufend, 0);
	gpe_status = strtoul(bufend, &bufend, 0);
	acpid_event_state = (acpi_sstate_t) strtoul(bufend, NULL, 0);
	
	if (pm1_status)
	{
		action = acpid_fixed;
		for (bit = 0; bit < ACPI_MAX_FIXED; bit++)
		{
			if ((pm1_status & 0x1)
			    && action->handler.fixed)
			{
				action->handler.fixed(action->context);
			}
			pm1_status >>= 1;
			action++;
		}
	}
	
	if (gpe_status)
	{
		action = acpid_gpe;
		for (bit = 0; bit < ACPI_MAX_GPE; bit++)
		{
			if ((gpe_status & 0x1)
			    && action->handler.gpe)
			{
				action->handler.gpe(action->context);
			}
			gpe_status >>= 1;
			action++;
		}
	}
	
	if (write(acpid_event_ack, buffer, size) <= 0)
		return -1;

	return 0;
}

/*
 *
 * Replacements for ACPI-CA's event handling functions
 *
 */

static int
acpid_fixed_event_to_bit(UINT32 event)
{
	int bit = -1;
	switch (event)
	{
	case EVENT_GLOBAL:
		bit = 5;
		break;
	case EVENT_POWER_BUTTON:
		bit = 8;
		break;
	case EVENT_SLEEP_BUTTON:
		bit = 9;
		break;
	case EVENT_RTC:
		bit = 10;
		break;
	default:
		break;
	}
	return bit;
}

ACPI_STATUS
acpi_install_fixed_event_handler(UINT32 event,
				 FIXED_EVENT_HANDLER handler,
				 void *context)
{
	int bit = acpid_fixed_event_to_bit(event);
	if (bit >= 0)
	{
		struct acpid_action *action = acpid_fixed + bit;
		action->handler.fixed = handler;
		action->context = context;
		action->level = 0;
	}
	return AE_OK;
}

ACPI_STATUS
acpi_remove_fixed_event_handler(UINT32 event, FIXED_EVENT_HANDLER handler)
{
	int bit = acpid_fixed_event_to_bit(event);
	if (bit >= 0)
	{
		struct acpid_action *action = acpid_fixed + bit;
		action->handler.fixed = NULL;
		action->context = NULL;
		action->level = 0;
	}
	return AE_OK;
}

ACPI_STATUS
acpi_install_gpe_handler(UINT32 gpe,
			 UINT32 type,
			 GPE_HANDLER handler,
			 void *context)
{
	if (gpe < ACPI_MAX_GPE)
	{
		struct acpid_action *action = acpid_gpe + gpe;
		action->handler.gpe = handler;
		action->context = context;
		action->level = (type == EVENT_LEVEL_TRIGGERED);
	}
	return AE_OK;
}

ACPI_STATUS
acpi_remove_gpe_handler(UINT32 gpe, GPE_HANDLER handler)
{
	if (gpe < ACPI_MAX_GPE)
	{
		struct acpid_action *action = acpid_gpe + gpe;
		action->handler.gpe = NULL;
		action->context = NULL;
		action->level = 0;
	}
	return AE_OK;
}

ACPI_STATUS
acpi_enable_event(UINT32 event, UINT32 type)
{
	unsigned long enable, level;
	struct acpid_action *action;
	int bit;
	switch (type)
	{
	case EVENT_FIXED:
		bit = acpid_fixed_event_to_bit(event);
		if (bit >= 0 && bit < ACPI_MAX_FIXED)
		{
			enable = acpid_read_sysctl("pm1_enable");
			enable |= (1 << bit);
			acpid_write_sysctl("pm1_enable", enable);
		}
		break;
	case EVENT_GPE:
		bit = event;
		if (bit < ACPI_MAX_GPE)
		{
			action = acpid_gpe + bit;
			if (action->level)
			{
				level = acpid_read_sysctl("gpe_level");
				level |= (1 << bit);
				acpid_write_sysctl("gpe_level", level);
			}
			enable = acpid_read_sysctl("gpe_enable");
			enable |= (1 << bit);
			acpid_write_sysctl("gpe_enable", enable);
		}
		break;
	}
	return AE_OK;
}

ACPI_STATUS
acpi_disable_event(UINT32 event, UINT32 type)
{
	unsigned long enable, level;
	struct acpid_action *action;
	int bit;
	switch (type)
	{
	case EVENT_FIXED:
		bit = acpid_fixed_event_to_bit(event);
		if (bit >= 0 && bit < ACPI_MAX_FIXED)
		{
			enable = acpid_read_sysctl("pm1_enable");
			enable &= ~(1 << bit);
			acpid_write_sysctl("pm1_enable", enable);
		}
		break;
	case EVENT_GPE:
		bit = event;
		if (bit < ACPI_MAX_GPE)
		{
			enable = acpid_read_sysctl("gpe_enable");
			enable &= ~(1 << bit);
			acpid_write_sysctl("gpe_enable", enable);
			action = acpid_gpe + bit;
			if (action->level)
			{
				level = acpid_read_sysctl("gpe_level");
				level &= ~(1 << bit);
				acpid_write_sysctl("gpe_level", level);
			}
		}
		break;
	}
	return AE_OK;
}

ACPI_STATUS
acpi_clear_event(UINT32 event, UINT32 type)
{
	// event status are cleared by the ACPI driver
	return AE_OK;
}

ACPI_STATUS
acpi_get_event_status(UINT32 event, UINT32 type, ACPI_EVENT_STATUS *status)
{
	*status = 0;
	return AE_OK;
}
