/*
 *  ipc.c - Client-server communications
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "acpid.h"
#include "ipc.h"

#define ACPID_SERVER_BACKLOG 5
#define ACPID_SERVER_TIMEOUT 1

static fd_set acpid_clients;

/*
 * Detect an already running server
 */
int
acpid_server_detect(void)
{
	static char buffer[ACPID_MAX_REQUEST];
	struct acpid_request *rqst = (struct acpid_request*) buffer;
	struct sockaddr_un addr;
	struct timeval timeout;
	fd_set fdset;
	int fd, size;
	int pid = -1;

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
	{
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sun_family = PF_UNIX;
	strcpy(addr.sun_path, ACPID_SERVER_PATH);
	size = strlen(addr.sun_path) + sizeof(addr.sun_family);

	FD_ZERO(&fdset);
	FD_SET(fd, &fdset);
	timeout.tv_sec = ACPID_SERVER_TIMEOUT;
	timeout.tv_usec = 0;

	/* connect, send protocol version, get ack, bail on any error */
	if (connect(fd, (struct sockaddr*) &addr, size) < 0
	    || (write(fd, ACPID_PROTOCOL, strlen(ACPID_PROTOCOL))
		< strlen(ACPID_PROTOCOL))
	    || select(fd + 1, &fdset, NULL, NULL, &timeout) <= 0
	    || read(fd, buffer, 1) < 1)
	{
		close(fd);
		return -1;
	}


	rqst->type = ACPID_PID_GET;
	rqst->status = ACPID_STATUS_OK;
	rqst->size = sizeof(*rqst);

	if (write(fd, rqst, rqst->size) == rqst->size
	    && read(fd, buffer, sizeof(buffer)) >= sizeof(struct acpid_request))
	{
		pid = strtol(rqst->data, NULL, 0);
	}

	close(fd);
	return pid;
}

/*
 * Initialize server
 */
int
acpid_server_init(void)
{
	struct sockaddr_un addr;
	int size, fd;

	FD_ZERO(&acpid_clients);

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
	{
		acpid_log(LOG_ERR, "socket: %s", strerror(errno));
		return -1;
	}

	unlink(ACPID_SERVER_PATH);
	
	memset(&addr, 0, sizeof(addr));
	addr.sun_family = PF_UNIX;
	strcpy(addr.sun_path, ACPID_SERVER_PATH);
	size = strlen(addr.sun_path) + sizeof(addr.sun_family);

	if (bind(fd, (struct sockaddr*) &addr, size) < 0)
	{
		acpid_log(LOG_ERR, "bind: %s", strerror(errno));
		close(fd);
		return -1;
	}

	chmod(ACPID_SERVER_PATH, 0777);

	if (listen(fd, ACPID_SERVER_BACKLOG) < 0)
	{
		acpid_log(LOG_ERR, "listen: %s", strerror(errno));
		close(fd);
		return -1;
	}

	return fd;
}

/*
 * Handle new client
 */
static int
acpid_server_client(int fd)
{
	struct sockaddr_un addr;
	int size, client;
	char protocol[8];

	size = sizeof(addr);
	client = accept(fd, (struct sockaddr*) &addr, &size);
	if (client < 0)
		return -1;

	/* get, check, and ack protocol version */
	if ((read(client, protocol, sizeof(protocol)) < sizeof(protocol)) ||
	    (strncmp(protocol, ACPID_PROTOCOL, sizeof(protocol))) ||
	    (write(client, "\n", 1) < 1))
	{
		close (client);
		return -1;
	}

	return client;
}

/*
 * Handle client request
 */
static int
acpid_server_request(int fd)
{
	static char buffer[ACPID_MAX_REQUEST];
	struct acpid_request *rqst = (struct acpid_request*) buffer;
	int size;

	memset(buffer, 0, sizeof(buffer));
	size = read(fd, buffer, sizeof(buffer));
	if (size <= 0 || size < sizeof(struct acpid_request))
	{
		return -1;
	}

	rqst->status = ACPID_STATUS_ERROR;
	if(rqst->type == ACPID_PID_GET)
	{
		snprintf(rqst->data, ACPID_REQUEST_DATA_SIZE, "%d", getpid());
		rqst->status = ACPID_STATUS_OK;
	}
	else if(rqst->type | ACPID_BATTERY_MASK)
		rqst->status = acpid_battery_request(rqst);
	else 
	{
		strcpy(rqst->data, "Not implemented");
		rqst->status = ACPID_STATUS_ERROR;
	}
	rqst->size = sizeof(*rqst) + strlen(rqst->data) + 1;

	if (write(fd, rqst, rqst->size) != rqst->size)
	{
		return -1;
	}

	return 0;
}

/*
 * Service clients and wait for an ACPI event
 */
int
acpid_event_loop(int event, int server, int debug)
{
	int minfd, maxfd;
	fd_set fdset;
	int fd;

	minfd = server;
	maxfd = server;
	if (event < minfd)
		minfd = event;
	else if (event > maxfd)
		maxfd = event;
	if (debug >= 0 && debug < minfd)
		minfd = debug;
	else if (debug > maxfd)
		maxfd = debug;

	while (1)
	{
		fdset = acpid_clients;
		FD_SET(event, &fdset);
		FD_SET(server, &fdset);
		if (debug >= 0)
			FD_SET(debug, &fdset);

		if (select(maxfd + 1, &fdset, NULL, NULL, NULL) < 0)
		{
			acpid_log(LOG_ERR, "accept: %s", strerror(errno));
			return -1;
		}

		for (fd = minfd; fd <= maxfd; fd++)
		{
			if (!FD_ISSET(fd, &fdset))
				continue;
			if (fd == event)
			{
				acpid_event_request(fd);
			}
			else if (fd == debug)
			{
				acpid_debug_request(fd);
			}
			else if (fd == server)
			{
				// client open
				int client = acpid_server_client(fd);
				if (client >= 0)
				{
					FD_SET(client, &acpid_clients);
					if (client < minfd)
						minfd = client;
					if (client > maxfd)
						maxfd = client;
				}
			}
			else if (acpid_server_request(fd))
			{
				// client close/terminate
				FD_CLR(fd, &acpid_clients);
				close(fd);
			}
		}
	}

	return 0;
}
