/*
 *  osd.c - OS-dependent functions
 *
 *  Copyright (C) 1999-2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/io.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sched.h>
#include <linux/types.h>
#include <linux/pci.h>

#include "acpid.h"
#include "acobject.h"
#include "namesp.h"
#include "methods.h"

#define ACPI_THREAD_STACK_SIZE 0x1000

#ifdef ACPI_STANDALONE
#undef inb
#define inb(port) 0
#undef inw
#define inw(port) 0
#undef inl
#define inl(port) 0
#undef outb
#define outb(val, port) 0
#undef outw
#define outw(val, port) 0
#undef outl
#define outl(val, port) 0
#endif

ACPI_STATUS
OsdInitialize(void)
{
	return AE_OK;
}

ACPI_STATUS
OsdTerminate(void)
{
	return AE_OK;
}

INT32
OsdPrintf(const char *fmt,...)
{
	INT32 size;
	va_list args;
	va_start(args, fmt);
	size = OsdVprintf(fmt, args);
	va_end(args);
	return size;
}

INT32
OsdVprintf(const char *fmt, va_list args)
{
	static char buffer[512];
	int size = vsprintf(buffer, fmt, args);
	if (!acpid_debug_output)
	{
		if (size > 0 && buffer[size - 1] == '\n')
			buffer[size - 1] = '\0';
		acpid_log(LOG_DEBUG, "%s", buffer);
	}
	else
		acpid_debug_printf(buffer);
	return size;
}

void *
OsdAllocate(UINT32 size)
{
	void *ptr = malloc(size);
	if (ptr)
		memset(ptr, 0, size);
	return ptr;
}

void *
OsdCallocate(UINT32 size)
{
	void *ptr = OsdAllocate(size);
	if (NULL != ptr)
	{
		memset(ptr, 0, size);
	}
	return ptr;
}

void
OsdFree(void *ptr)
{
	free(ptr);
}

ACPI_STATUS
OsdMapMemory(void *addr, UINT32 size, void **logical)
{
	void *ptr = NULL;
#ifndef ACPI_STANDALONE
	int mem = open("/dev/mem", O_RDONLY);
	if (mem >= 0)
	{
		// mmap table header to determine size
		unsigned long offset = ((unsigned long) addr
			& (getpagesize() - 1));
		__u8 *mapped = mmap(NULL,
				    size + offset,
				    PROT_READ,
				    MAP_PRIVATE,
				    mem,
				    (unsigned long) addr - offset);
		if (mapped != MAP_FAILED)
			ptr = (void *) (mapped + offset);
		close(mem);
	}
#else
	ptr = malloc(size);
	if (ptr)
		memset(ptr, 0, size);
#endif
	*logical = ptr;
	return AE_OK;
}

void
OsdUnMapMemory(void *ptr, UINT32 size)
{
#ifndef ACPI_STANDALONE
	if (ptr)
	{
		unsigned long offset
		= (unsigned long) ptr & (getpagesize() - 1);
		munmap((void *) ((unsigned long) ptr - offset), size + offset);
	}
#else
	free(ptr);
#endif
}

ACPI_STATUS
OsdInstallInterruptHandler(UINT32 irq, OSD_HANDLER handler, void *context)
{
	// the ACPI driver does its own interrupt handling
	return AE_OK;
}

ACPI_STATUS
OsdRemoveInterruptHandler(UINT32 irq, OSD_HANDLER handler)
{
	// the ACPI driver does its own interrupt handling
	return AE_OK;
}

void
OsdSleep(UINT32 sec, UINT32 ms)
{
	usleep(sec * 1000000 + ms * 1000);
}

void
OsdSleepUsec(UINT32 us)
{
	usleep(us);
}

void
OsdStall(UINT32 us)
{
	usleep(us);
}

UINT8
OsdIn8(ACPI_IO_ADDRESS port)
{
	UINT8 value = inb(port);
	return value;
}

UINT16
OsdIn16(ACPI_IO_ADDRESS port)
{
	UINT16 value = inw(port);
	return value;
}

UINT32
OsdIn32(ACPI_IO_ADDRESS port)
{
	UINT32 value = inl(port);
	return value;
}

void
OsdOut8(ACPI_IO_ADDRESS port, UINT8 val)
{
	outb(val, port);
}

void
OsdOut16(ACPI_IO_ADDRESS port, UINT16 val)
{
	outw(val, port);
}

void
OsdOut32(ACPI_IO_ADDRESS port, UINT32 val)
{
	outl(val, port);
}

ACPI_STATUS
OsdReadPciCfgByte(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT8 * val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_read(bus, devfn, addr, val, 1) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdReadPciCfgWord(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT16 * val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_read(bus, devfn, addr, val, 2) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdReadPciCfgDword(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT32 * val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_read(bus, devfn, addr, val, 4) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdWritePciCfgByte(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT8 val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_write(bus, devfn, addr, &val, 1) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdWritePciCfgWord(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT16 val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_write(bus, devfn, addr, &val, 2) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdWritePciCfgDword(
	UINT32 bus,
	UINT32 func,
	UINT32 addr,
	UINT32 val)
{
	int devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	if (acpid_pci_write(bus, devfn, addr, &val, 4) < 0)
		return AE_ERROR;
	return AE_OK;
}

ACPI_STATUS
OsdQueueForExecution(
	UINT32 priority,
	OSD_EXECUTION_CALLBACK callback,
	void *context)
{
#ifndef ACPI_APPLICATION
	typedef int (*clone_callback_t)(void*);
	void *stack = malloc(ACPI_THREAD_STACK_SIZE);
	if (!stack)
		return AE_NO_MEMORY;

	if (clone((clone_callback_t) callback,
		  stack + ACPI_THREAD_STACK_SIZE - 1,
		  CLONE_VM,
		  context) < 0)
	{
		free(stack);
		return AE_ERROR;
	}
#endif

	return AE_OK;
}

ACPI_STATUS
OsdCreateSemaphore(UINT32 init, ACPI_HANDLE * handle)
{
	int value = (int) init;
	int semid = semget(IPC_PRIVATE, 1, IPC_CREAT | IPC_EXCL | 0666);
	if (semid < 0 || semctl(semid, 0, SETVAL, value) < 0)
	{
		acpid_log(LOG_ERR, "OsdCreateSemaphore: semget: %s",
		    strerror(errno));
		*handle = (ACPI_HANDLE) -1;
		return AE_ERROR;
	}
	*handle = (ACPI_HANDLE) semid;
	return AE_OK;
}

ACPI_STATUS
OsdDeleteSemaphore(ACPI_HANDLE handle)
{
	int semid = (int) handle;
	if (semctl(semid, 0, IPC_RMID, 0) < 0)
	{
		acpid_log(LOG_ERR, "OsdDeleteSemaphore: semctl: %s",
		    strerror(errno));
		return AE_ERROR;
	}
	return AE_OK;
}

ACPI_STATUS
OsdWaitSemaphore(ACPI_HANDLE handle, UINT32 units, UINT32 timeout)
{
	int semid = (int) handle;
	struct sembuf op = {0, -1, 0};
	op.sem_op = -((short) units);
	if (semop(semid, &op, 1) < 0)
	{
		acpid_log(LOG_ERR, "OsdWaitSemaphore: semop: %s",
		    strerror(errno));
		return AE_ERROR;
	}
	return AE_OK;
}

ACPI_STATUS
OsdSignalSemaphore(ACPI_HANDLE handle, UINT32 units)
{
	int semid = (int) handle;
	struct sembuf op = {0, 1, IPC_NOWAIT};
	op.sem_op = (short) units;
	if (semop(semid, &op, 1) < 0)
	{
		acpid_log(LOG_ERR, "OsdSignalSemaphore: semop: %s",
		    strerror(errno));
		return AE_ERROR;
	}
	return AE_OK;
}

ACPI_STATUS
OsdBreakpoint(char *msg)
{
	OsdPrintf("OsdBreakpoint: %s", msg);
	return AE_OK;
}

void
OsdDbgTrap(char *msg)
{
	OsdPrintf("OsdDbgTrap: %s", msg);
}

void
OsdDbgAssert(void *failure, void *file, UINT32 line, char *msg)
{
	acpid_log(LOG_DEBUG, "OsdDbgAssert: `%s' at %s:%d",
		  (char *) msg, (char *) file, line);
}

UINT32
OsdGetLine(char *buffer)
{
	fgets(buffer, 64, stdin);
	return strlen(buffer);
}

char *
strupr(char *str)
{
	char *s = str;
	while (*s)
	{
		*s = toupper(*s);
		s++;
	}
	return str;
}

BOOLEAN
OsdReadable(void *ptr, UINT32 len)
{
	return 1;
}

BOOLEAN
OsdWritable(void *ptr, UINT32 len)
{
	return 1;
}
