/*
 *  pci.c - PCI config space
 *
 *  Copyright (C) 2000 Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/pci.h>
#include "acpid.h"
#include "macros.h"
#include "acobject.h"
#include "events.h"
#include "globals.h"

#define PCI_PROC "/proc/bus/pci"
#define MAX_PCI_PATH_SIZE 64

int
acpid_pci_read(int bus, int devfn, int addr, void *val, int size)
{
	int status = -1;
	char path[MAX_PCI_PATH_SIZE];
	int fd;

	sprintf(path, "%s/%02d/%02d.%d",
		PCI_PROC, bus, PCI_SLOT(devfn), PCI_FUNC(devfn));

	fd = open(path, O_RDONLY);
	if (fd >= 0
	    && lseek(fd, addr, SEEK_SET) == addr
	    && read(fd, val, size) == size)
	{
		status = 0;
	}
	else
	{
		acpid_log(LOG_ERR, "read from %s failed", path);
	}
	close(fd);

	return status;
}

int
acpid_pci_write(int bus, int devfn, int addr, void *val, int size)
{
	int status = -1;
	char path[MAX_PCI_PATH_SIZE];
	int fd;

	sprintf(path, "%s/%02d/%02d.%d",
		PCI_PROC, bus, PCI_SLOT(devfn), PCI_FUNC(devfn));

	fd = open(path, O_WRONLY);
	if (fd >= 0
	    && lseek(fd, addr, SEEK_SET) == addr
	    && write(fd, val, size) == size)
	{
		status = 0;
	}
	else
	{
		acpid_log(LOG_ERR, "write to %s failed", path);
	}
	close(fd);

	return status;
}

static ACPI_STATUS
acpid_pci_handler(
	UINT32 cmd,
	UINT32 address,
	UINT32 bit_width,
	UINT32 *value,
	void *context)
{
	ACPI_STATUS status = AE_OK;
	PCI_HANDLER_CONTEXT *info = (PCI_HANDLER_CONTEXT*) context;
	int size, bus, dev, func;

	bus = info->Bus & 0xffff;
	dev = (info->DevFunc >> 16) & 0xffff;
	func = info->DevFunc & 0xffff;

	switch (bit_width)
	{
	case 8: size = 1; break;
	case 16: size = 2; break;
	case 32: size = 4; break;
	default:
		acpid_log(LOG_ERR, "invalid PCI access (%d bits)", bit_width);
		return AE_BAD_PARAMETER;
	}

	switch (cmd)
	{
	case ADDRESS_SPACE_READ:
		if (acpid_pci_read(bus,
				   PCI_DEVFN(dev, func),
				   address,
				   value,
				   size) < 0)
		{
			status = AE_ERROR;
		}
		break;
	case ADDRESS_SPACE_WRITE:
		if (acpid_pci_write(bus,
				    PCI_DEVFN(dev, func),
				    address,
				    value,
				    size) < 0)
		{
			status = AE_ERROR;
		}
		break;
	default:
		acpid_log(LOG_ERR, "invalid PCI access (cmd %d)", cmd);
		return AE_BAD_PARAMETER;
	}

	return status;
}

int
acpid_pci_init(void)
{
	ACPI_STATUS status;
	int fd;

	if (access(PCI_PROC, R_OK))
	{
		acpid_log(LOG_ERR, "PCI is not present");
	}

	status = AcpiInstallAddressSpaceHandler(Gbl_RootObject,
						ADDRESS_SPACE_PCI_CONFIG,
						acpid_pci_handler,
						NULL,
						NULL);
	if (!ACPI_SUCCESS(status))
	{
		acpid_log(LOG_ERR, "cannot install PCI handler");
	}

	return 0;
}
