/*
 *  ./acpica/subsystem/debugger/dbexec.c - debugger control method execution
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <acpi.h>
#include <parser.h>
#include <dispatch.h>
#include <amlcode.h>
#include <namesp.h>
#include <parser.h>
#include <events.h>
#include <interp.h>
#include <debugger.h>
#include <tables.h>

#ifdef ACPI_DEBUG

#define _COMPONENT          DEBUGGER
        MODULE_NAME         ("dbexec");



typedef struct dbmethodinfo
{
    ACPI_HANDLE             ThreadGate;
    char                    *Name;
    char                    **Args;
    UINT32                  Flags;
    UINT32                  NumLoops;
    char                    Pathname[128];

} DB_METHOD_INFO;


DB_METHOD_INFO              Info;




ACPI_STATUS
DbExecuteMethod (
    DB_METHOD_INFO          *Info,
    ACPI_BUFFER             *ReturnObj)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_LIST        ParamObjects;
    ACPI_OBJECT             Params[MTH_NUM_ARGS];
    UINT32                  i;


    if (OutputToFile && !DebugLevel)
    {
        OsdPrintf ("Warning: debug output is not enabled!\n");
    }

    if (Info->Args && Info->Args[0])
    {
        for (i = 0; Info->Args[i] && i < MTH_NUM_ARGS; i++)
        {
            Params[i].Type              = ACPI_TYPE_Number;
            Params[i].Number.Value      = STRTOUL (Info->Args[i], NULL, 16);
        }

        ParamObjects.Pointer        = Params;
        ParamObjects.Count          = i;
    }

    else
    {
        /* Setup default parameters */

        Params[0].Type              = ACPI_TYPE_Number;
        Params[0].Number.Value      = 0x01020304;

        Params[1].Type              = ACPI_TYPE_String;
        Params[1].String.Length     = 12;
        Params[1].String.Pointer    = "AML Debugger";

        ParamObjects.Pointer        = Params;
        ParamObjects.Count          = 2;
    }

    /* Prepare for a return object of arbitrary size */

    ReturnObj->Pointer           = Buffer;
    ReturnObj->Length            = BUFFER_SIZE;


    /* Do the actual method execution */

    Status = AcpiEvaluateObject (NULL, Info->Pathname, &ParamObjects, ReturnObj);

    Gbl_CmSingleStep = FALSE;
    Gbl_MethodExecuting = FALSE;

    return Status;
}




void
DbExecuteSetup (
    DB_METHOD_INFO          *Info)
{

    /* Catenate the current scope to the supplied name */

    Info->Pathname[0] = 0;
    if ((Info->Name[0] != '\\') &&
        (Info->Name[0] != '/'))
    {
        STRCAT (Info->Pathname, ScopeBuf);
    }

    STRCAT (Info->Pathname, Info->Name);
    DbPrepNamestring (Info->Pathname);

    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);
    OsdPrintf ("Executing %s\n", Info->Pathname);

    if (Info->Flags & EX_SINGLE_STEP)
    {
        Gbl_CmSingleStep = TRUE;
        DbSetOutputDestination (DB_CONSOLE_OUTPUT);
    }

    else
    {
        /* No single step, allow redirection to a file */

        DbSetOutputDestination (DB_REDIRECTABLE_OUTPUT);
    }
}



void
DbExecute (
    char                    *Name,
    char                    **Args,
    UINT32                  Flags)
{
    ACPI_STATUS             Status;
    UINT32                  PreviousAllocations;
    UINT32                  PreviousSize;
    UINT32                  Allocations;
    UINT32                  Size;
    ACPI_BUFFER             ReturnObj;

    

    /* Memory allocation tracking */

    PreviousAllocations = Gbl_CurrentAllocCount;
    PreviousSize = Gbl_CurrentAllocSize;


    Info.Name = Name;
    Info.Args = Args;
    Info.Flags = Flags;

    DbExecuteSetup (&Info);
    Status = DbExecuteMethod (&Info, &ReturnObj);


    /* Memory allocation tracking */

    Allocations = Gbl_CurrentAllocCount - PreviousAllocations;
    Size = Gbl_CurrentAllocSize - PreviousSize;

    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);

    if (Allocations > 0)
    {
        OsdPrintf ("Outstanding: %ld allocations of total size %ld after execution\n",
                        Allocations, Size);
    }


    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Execution of %s failed with status %s\n", Info.Pathname, CmFormatException (Status));
    }

    else
    {
        /* Display a return object, if any */

        if (ReturnObj.Length)
        {
            OsdPrintf ("Execution of %s returned object %p\n", Info.Pathname, ReturnObj.Pointer);
            DbDumpObject (ReturnObj.Pointer, 1);
        }
    }

    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
}



void
DbMethodThread (
    void                    *Context)
{
    ACPI_STATUS             Status;
    DB_METHOD_INFO          *Info = Context;
    UINT32                  i;
    ACPI_BUFFER             ReturnObj;


    for (i = 0; i < Info->NumLoops; i++)
    {
        Status = DbExecuteMethod (Info, &ReturnObj);
        if (ACPI_SUCCESS (Status))
        {
            if (ReturnObj.Length)
            {
                OsdPrintf ("Execution of %s returned object %p\n", Info->Pathname, ReturnObj.Pointer);
                DbDumpObject (ReturnObj.Pointer, 1);
            }
        }
    }


    /* Signal our completion */

    OsdSignalSemaphore (Info->ThreadGate, 1);
}




void
DbCreateExecutionThreads (
    char                    *NumThreadsArg,
    char                    *NumLoopsArg,
    char                    *MethodNameArg)
{
    ACPI_STATUS             Status;
    UINT32                  NumThreads;
    UINT32                  NumLoops;
    UINT32                  i;
    ACPI_HANDLE             ThreadGate;


    /* Get the arguments */

    NumThreads = STRTOUL (NumThreadsArg, NULL, 0);
    NumLoops = STRTOUL (NumLoopsArg, NULL, 0);

    if (!NumThreads || !NumLoops)
    {
        OsdPrintf ("Bad argument: Threads %d, Loops %d\n", NumThreads, NumLoops);
        return;
    }


    /* Create the synchronization semaphore */

    Status = OsdCreateSemaphore (0, &ThreadGate);
    if (ACPI_FAILURE (Status))
    {
        OsdPrintf ("Could not create semaphore, %s\n", CmFormatException (Status));
        return;
    }

    /* Setup the context to be passed to each thread */

    Info.Name = MethodNameArg;
    Info.Args = NULL;
    Info.Flags = 0;
    Info.NumLoops = NumLoops;
    Info.ThreadGate = ThreadGate;

    DbExecuteSetup (&Info);


    /* Create the threads */

    OsdPrintf ("Creating %d threads to execute %d times each\n", NumThreads, NumLoops);

    for (i = 0; i < (NumThreads); i++)
    {
        OsdQueueForExecution (OSD_PRIORITY_MED, DbMethodThread, &Info);
    }


    /* Wait for all threads to complete */

    i = NumThreads;
    while (i)   /* Brain damage for OSD implementations that only support wait of 1 unit */
    {
        Status = OsdWaitSemaphore (ThreadGate, 1, WAIT_FOREVER);
        i--;
    }

    /* Cleanup and exit */

    OsdDeleteSemaphore (ThreadGate);

    DbSetOutputDestination (DB_DUPLICATE_OUTPUT);
    OsdPrintf ("All threads (%d) have completed\n", NumThreads);
    DbSetOutputDestination (DB_CONSOLE_OUTPUT);
}




#endif