
/*
 *  ./acpica/subsystem/namespace/nsapiobj.c - Public interfaces to the ACPI subsystem
 *
 *  Copyright (C) 2000 R. Byron Moore
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#define __NSAPIOBJ_C__

#include <acpi.h>
#include <interp.h>
#include <namesp.h>
#include <methods.h>
#include <acobject.h>
#include <pnp.h>


#define _COMPONENT          NAMESPACE
        MODULE_NAME         ("nsapiobj");



ACPI_STATUS
AcpiEvaluateObject (
    ACPI_HANDLE             Handle, 
    ACPI_STRING             Pathname, 
    ACPI_OBJECT_LIST        *ParamObjects,
    ACPI_BUFFER             *ReturnBuffer)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_INTERNAL    **ParamPtr = NULL;
    ACPI_OBJECT_INTERNAL    *ReturnObj = NULL;
    ACPI_OBJECT_INTERNAL    *ObjectPtr = NULL;
    UINT32                  BufferSpaceNeeded;
    UINT32                  UserBufferLength;
    UINT32                  Count;
    UINT32                  i;
    UINT32                  ParamLength;
    UINT32                  ObjectLength;


    FUNCTION_TRACE ("AcpiEvaluateObject");


    /* 
     * If there are parameters to be passed to the object (which must be a control method), 
     * the external objects must be converted to internal objects
     */

    if (ParamObjects && ParamObjects->Count)
    {
        /* 
         * Allocate a new parameter block for the internal objects
         * Add 1 to count to allow for null terminated internal list
         * TBD: merge into single allocation!
         */

        Count           = ParamObjects->Count;
        ParamLength     = (Count + 1) * sizeof (void *);
        ObjectLength    = Count * sizeof (ACPI_OBJECT_INTERNAL);

        ParamPtr = CmCallocate (ParamLength +       /* Parameter List part */
                                ObjectLength);      /* Actual objects */
        if (!ParamPtr)
        {
            return_ACPI_STATUS (AE_NO_MEMORY);
        }

        ObjectPtr = (ACPI_OBJECT_INTERNAL *) ((UINT8 *) ParamPtr + ParamLength);

        /*
         * Init the param array of pointers and NULL terminate the list
         */

        for (i = 0; i < Count; i++)
        {
            ParamPtr[i] = &ObjectPtr[i];
            CmInitStaticObject (&ObjectPtr[i]);
        }
        ParamPtr[Count] = NULL;                 

        /* 
         * Convert each external object in the list to an internal object
         */
        for (i = 0; i < Count; i++)
        {
            Status = CmBuildInternalObject (&ParamObjects->Pointer[i], ParamPtr[i]);
            if (ACPI_FAILURE (Status))
            {
                CmDeleteInternalObjectList (ParamPtr);
                return_ACPI_STATUS (Status);
            }
        }
    }


    /*
     * Three major cases:
     * 1) Fully qualified pathname 
     * 2) No handle, not fully qualified pathname (error)
     * 3) Valid handle 
     */

    if ((Pathname) && 
        (NsValidRootPrefix (Pathname[0])))
    {
        /*
         *  The path is fully qualified, just evaluate by name
         */
        Status = NsEvaluateByName (Pathname, ParamPtr, &ReturnObj);
    }

    else if (!Handle)
    {
        /*
         * A handle is optional iff a fully qualified pathname is specified.
         * Since we've already handled fully qualified names above, this is an error
         */

        if (!Pathname)
        {
            DEBUG_PRINT (ACPI_ERROR, ("AcpiEvaluateObject: Both Handle and Pathname are NULL\n"));
        }

        else
        {
            DEBUG_PRINT (ACPI_ERROR, ("AcpiEvaluateObject: Handle is NULL and Pathname is relative\n"));
        }

        Status = AE_BAD_PARAMETER;
    }

    else 
    {
        /*
         *  We get here if we have a handle -- and if we have a pathname it is relative.
         *  The handle will be validated in the lower procedures
         */

        if (!Pathname)
        {
            /*
             * The null pathname case means the handle is for the actual object to be evaluated
             */
            Status = NsEvaluateByHandle (Handle, ParamPtr, &ReturnObj);
        }

        else
        {
           /*
            * Both a Handle and a relative Pathname
            */
            Status = NsEvaluateRelative (Handle, Pathname, ParamPtr, &ReturnObj);
        }
    }


    
    /* 
     * If we are expecting a return value, and all went well above, 
     * copy the return value to an external object.
     */

    if (ReturnBuffer)
    {
        UserBufferLength = ReturnBuffer->Length;
        ReturnBuffer->Length = 0;

        if (ReturnObj)
        {
            if (VALID_DESCRIPTOR_TYPE (ReturnObj, DESC_TYPE_NTE))
            {
                /* 
                 * If we got an NTE as a return object, this means the object we are evaluating has nothing
                 * interesting to return (such as a mutex, etc.)  We return an error because these types
                 * are essentially unsupported by this interface.  We don't check up front because this makes
                 * it easier to add support for various types at a later date if necessary.
                 */
                Status = AE_TYPE;
                ReturnObj = NULL;   /* No need to delete an NTE */
            }

            if (ACPI_SUCCESS (Status))
            {
                /*
                 *  Find out how large a buffer is needed to contain the
                 *  returned object
                 */
                Status = CmGetObjectSize (ReturnObj, &BufferSpaceNeeded);
                if (ACPI_SUCCESS (Status))
                {
                    /* Check if there is enough room in the caller's buffer */

                    if (UserBufferLength < BufferSpaceNeeded) 
                    {
                        /*
                         *  Caller's buffer is too small, can't give him partial results
                         *  fail the call but return the buffer size needed
                         */

                        DEBUG_PRINT (ACPI_ERROR, ("AcpiEvaluateObject: Needed buffer size %d, received %d\n",
                                                    BufferSpaceNeeded, UserBufferLength));

                        ReturnBuffer->Length = BufferSpaceNeeded;
                        Status = AE_BUFFER_OVERFLOW;
                    }

                    else
                    {
                        /*
                         *  We have enough space for the object, build it
                         */
                        Status = CmBuildExternalObject (ReturnObj, ReturnBuffer);
                        ReturnBuffer->Length = BufferSpaceNeeded;
                    }
                }
            }
        }
    }


    /* Delete the return and parameter objects */

    if (ReturnObj)
    {
        /*
         * Delete the internal return object. (Or at least
         * decrement the reference count by one)
         */
        CmRemoveReference (ReturnObj);
    }

    /*
     * Free the input parameter list (if we created one),
     */

    if (ParamPtr)
    {
        /* Free the allocated parameter block */

        CmDeleteInternalObjectList (ParamPtr);
    }

    return_ACPI_STATUS (Status);
}



ACPI_STATUS
AcpiGetNextObject (
    ACPI_OBJECT_TYPE        Type, 
    ACPI_HANDLE             Parent, 
    ACPI_HANDLE             Child, 
    ACPI_HANDLE             *RetHandle)
{
    ACPI_STATUS             Status = AE_OK;
    NAME_TABLE_ENTRY        *Entry;
    NAME_TABLE_ENTRY        *ParentEntry = NULL;
    NAME_TABLE_ENTRY        *ChildEntry = NULL;


    /* Parameter validation */

    if (Type > ACPI_TYPE_MAX)
    {
        return AE_BAD_PARAMETER;
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* If null handle, use the parent */

    if (!Child)
    {
        /* Start search at the beginning of the specified scope */

        ParentEntry = NsConvertHandleToEntry (Parent);
        if (!ParentEntry)
        {
            Status = AE_BAD_PARAMETER;
            goto UnlockAndExit;
        }
    }

    /* Non-null handle, ignore the parent */

    else
    {
        /* Convert and validate the handle */

        ChildEntry = NsConvertHandleToEntry (Child);
        if (!ChildEntry)
        {
            Status = AE_BAD_PARAMETER;
            goto UnlockAndExit;
        }
    }


    /* Internal function does the real work */

    Entry = NsGetNextObject (Type, ParentEntry, ChildEntry);
    if (!Entry)
    {
        Status = AE_NOT_FOUND;
        goto UnlockAndExit;
    }

    if (RetHandle)
    {
        *RetHandle = NsConvertEntryToHandle (Entry);
    }


UnlockAndExit:

    CmReleaseMutex (MTX_NAMESPACE);
    return Status;
}



ACPI_STATUS 
AcpiGetType (
    ACPI_HANDLE             Handle,
    ACPI_OBJECT_TYPE        *RetType)
{
    NAME_TABLE_ENTRY        *Object;


    /* Parameter Validation */

    if (!RetType)
    {
        return AE_BAD_PARAMETER;
    }

    /* Special case for the predefined Root Object (return type ANY) */

    if (Handle == ACPI_ROOT_OBJECT)
    {
        *RetType = ACPI_TYPE_Any;
        return AE_OK;
    }

    CmAcquireMutex (MTX_NAMESPACE);

    /* Convert and validate the handle */

    Object = NsConvertHandleToEntry (Handle);
    if (!Object)
    {
        CmReleaseMutex (MTX_NAMESPACE);
        return AE_BAD_PARAMETER;
    }

    *RetType = Object->Type;


    CmReleaseMutex (MTX_NAMESPACE);
    return AE_OK;
}



ACPI_STATUS
AcpiGetParent (
    ACPI_HANDLE             Handle, 
    ACPI_HANDLE             *RetHandle)
{
    NAME_TABLE_ENTRY        *Object;
    ACPI_STATUS             Status = AE_OK;
    

    /* No trace macro, too verbose */


    if (!RetHandle)
    {
        return AE_BAD_PARAMETER;
    }

    /* Special case for the predefined Root Object (no parent) */

    if (Handle == ACPI_ROOT_OBJECT)
    {
        return AE_NULL_ENTRY;
    }


    CmAcquireMutex (MTX_NAMESPACE);

    /* Convert and validate the handle */

    Object = NsConvertHandleToEntry (Handle);
    if (!Object)
    {
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit;
    }

   
    /* Get the parent entry */

    *RetHandle = NsConvertEntryToHandle(Object->ParentEntry);

    /* Return exeption if parent is null */

    if (!Object->ParentEntry)
    {
        Status = AE_NULL_ENTRY;
    }


UnlockAndExit:

    CmReleaseMutex (MTX_NAMESPACE);
    return AE_OK;
}



ACPI_STATUS
AcpiWalkNamespace (
    ACPI_OBJECT_TYPE        Type, 
    ACPI_HANDLE             StartObject, 
    UINT32                  MaxDepth,
    WALK_CALLBACK           UserFunction, 
    void                    *Context, 
    void                    **ReturnValue)
{
    ACPI_STATUS             Status;


    FUNCTION_TRACE ("AcpiWalkNamespace");


    /* Parameter validation */

    if ((Type > ACPI_TYPE_MAX)  ||
        (!MaxDepth)             || 
        (!UserFunction))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Lock the namespace around the walk.
     * The namespace will be unlocked/locked around each call to the user function - since this function
     * must be allowed to make Acpi calls itself.
     */

    CmAcquireMutex (MTX_NAMESPACE);
    Status = NsWalkNamespace (Type, StartObject, MaxDepth, NS_WALK_UNLOCK, UserFunction, Context, ReturnValue);
    CmReleaseMutex (MTX_NAMESPACE);

    return_ACPI_STATUS (Status);
}


